/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import kddml.Core.DataMining.*;
import kddml.Core.*;

/**
 * A factory class that creates an association model from an external resource (e.g. text file,
 * PMML model).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */

public abstract class AssociationExternalFactory extends ModelExternalFactory {

  /**
   * Constructor given the physical file containing the object.
   *
   * @param f File
   * @throws KDDMLCoreException
   */
  public AssociationExternalFactory(java.io.File f)
      throws KDDMLCoreException {
    super(f);
    object = new AssociationModel(f);
  }


  public KDDMLObject newInstance() throws KDDMLCoreException {
    super.newInstance();
    Double min_support = buildMinSupport();
    CheckerUtility.checkMinSupport(min_support);
    ((AssociationModel) object).min_support = min_support;

    Double min_confidence = buildMinConfidence();
    CheckerUtility.checkMinConfidence(min_confidence);
    ((AssociationModel) object).min_confidence = min_confidence;

    Integer number_of_transactions = buildNumberOfTransactions();
    CheckerUtility.checkNumTransactions(number_of_transactions);
    ((AssociationModel) object).number_of_transactions = number_of_transactions;

    Integer max_items = buildMaxNumberOfItemsPerTA();
    if (max_items != null) {
      CheckerUtility.checkMaxItems(max_items);
      ( (AssociationModel) object).max_number_of_items_per_TA = max_items;
    }

    Double avg_items = buildAvgNumberOfItemsPerTA();
    if (avg_items != null) {
      CheckerUtility.checkAvgItems(avg_items);
      ( (AssociationModel) object).avg_number_of_items_per_TA = avg_items;
    }

    ((AssociationModel) object).items = buildItems();
    ((AssociationModel) object).itemsets = buildItemsets();
    ((AssociationModel) object).association_rules = buildAssociationRules();
    return object;
  }

  /**
   * Returns the minimum support of the association rules.
   * Mininmum support is a real in (0,1]
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected abstract Double buildMinSupport() throws AssociationModelException;

  /**
   * Returns the minimum confidence of the association rules.
   * Mininmum confidence is a real in (0,1].
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected abstract Double buildMinConfidence() throws AssociationModelException;

  /**
   * Returns the number of transaction used to build the association rules.
   * Returns a positive integer.
   *
   * @throws AssociationModelException
   * @return Integer
   */
  protected abstract Integer buildNumberOfTransactions() throws AssociationModelException;

  /**
   * Returns the maximum number of iterms per transaction.
   * Returns a positive integer.
   *
   * @throws AssociationModelException
   * @return Integer
   */
  protected abstract Integer buildMaxNumberOfItemsPerTA() throws AssociationModelException;

  /**
   * Returns the average number of items per transaction.
   * Returns a positive double.
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected abstract Double buildAvgNumberOfItemsPerTA() throws AssociationModelException;

  /**
   * Returns the list of association rules.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of AssociationRules
   */
  protected abstract java.util.TreeSet buildAssociationRules() throws AssociationModelException;

  /**
   * Returns the list of itemsets.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of Itemsets
   */
  protected abstract java.util.TreeSet buildItemsets() throws AssociationModelException;

  /**
   * Returns the list of items.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of Item
   */
  protected abstract java.util.TreeSet buildItems() throws AssociationModelException;
}
