/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import java.util.Vector;
import java.util.Iterator;
import java.util.TreeSet;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;
import kddml.Core.*;
import weka.core.*;

/**
 * Association rule mining finds interesting association or correlation relationships
 * among a large set of data items. A typical example of association rule mining is
 * market basket analysis that analyzes customer buying habits by finding association
 * between the different items that customers place in their shopping baskets.<br>
 * KDDML association model contains boolean association rules (or intra-attribute rules)
 * that have the form: <p>spaghetti, tomato --> parmesan,</p>
 * or inter-attribute rules, such as <p>carType = racing, homeInsurance = false -->
 * married = yes.</p>
 * An association rule model consists of four major parts:
 * <ol>
 * <li> Model attributes (such as minimum support, minimum confidence, etc.). </li>
 * <li> A list of items. </li>
 * <li> A list of frequent itemsets. </li>
 * <li> A list of association rules. </li>
 * </ol>
 * Items, itemsets and association rules are ordered using a precise criterium and this
 * objects are mantained in a TreeSet ordered structure.
 * An AssociationModel can contain any number of itemsets and association rules.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class AssociationModel extends kddml.Core.DataMining.MiningModel
    implements AssociationModelManager {

  /**
   * Max number of itemsets that the model can contain.
   */
  public static int MAX_NUMBER_OF_ITEMSETS = 5000;

  /**
   * Max number of association rules that the model can contain.
   */
  public static int MAX_NUMBER_OF_RULES = 5000;

  /**
   * The set of ordered association rules.
   */
  protected TreeSet association_rules;

  /**
   * The set of ordered itemsets.
   */
  protected TreeSet itemsets;

  /**
   * The set of ordered items.
   */
  protected TreeSet items;

  /**
   * The number of transactions (baskets of items) contained in the input data.
   */
  protected Integer number_of_transactions;

  /**
   * Minimum support.
   * Support is calculate as: (#supporting transactions / #total transactions).
   */
  protected Double min_support;

  /**
   * Minimum confidence.
   * Confidence is calculate as: (support(rule) / support(antecedent)).
   */
  protected Double min_confidence;

  /**
   * The number of items contained in the largest transaction.
   */
  protected Integer max_number_of_items_per_TA;

  /**
   * The average number of items contained in a transaction.
   */
  protected Double avg_number_of_items_per_TA;

  /**
   * Proprietary constructor.
   *
   * @param physical_path File the repository physical path containing the model.
   */
  protected AssociationModel(java.io.File physical_path) {
    super(physical_path);
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return kddml.Core.KDDMLObjectType.PMML_ASSOCIATION_MODEL;
  }

  /**
   * Adds a new association rule to this model.
   * Returns true if the rule has been added. Returns false if the model already
   * contains the rule.
   *
   * @param rule AssociationRuleManager the association rule to add to the model.
   * @return boolean true if the rule do not exist in the model, false otherwise.
   */
  public boolean addRule(AssociationRuleManager rule) {
    if (contains(rule))
      return false;
    if ((rule).getSupport() < min_support.doubleValue())
      return false;
    if ((rule).getConfidence() < min_confidence.doubleValue())
      return false;
    association_rules.add(rule);

    // vengono ri-generati gli itemsets frequenti dalle nuove RdA...
    itemsets = RdA2ItemsetUtility.compute(association_rules, MAX_NUMBER_OF_ITEMSETS);

    // scorre gli itemsets per calcolare gli items...
    items = Itemset2ItemUtility.compute(itemsets);
    return true;
  }

  /**
   * Adds a set of association rules to this model.
   *
   * @param rules AssociationRuleManager[] a set of association rules to add
   * to the model
   */
  public void addRules(AssociationRuleManager[] rules) {

    for (int i=0; i<rules.length; i++) {
      AssociationRule rule = (AssociationRule) rules[i];
      if (!contains(rule))
        if (rule.getSupport() >= min_support.doubleValue() &&
            rule.getConfidence() >= min_confidence.doubleValue())
          association_rules.add(rules[i]);
    }

    // vengono ri-generati gli itemsets frequenti dalle nuove RdA...
    itemsets = RdA2ItemsetUtility.compute(association_rules, MAX_NUMBER_OF_RULES);

    // scorre gli itemsets per calcolare gli items...
    items = Itemset2ItemUtility.compute(itemsets);
  }

  /**
   * Returns the list of association rules belonging to the model.
   *
   * @return Iterator the list of association rules.
   */
  public Iterator getAssociationRules() {
    return association_rules.iterator();
  }

  /**
   * Returns the list of itemsets belonging to the model.
   *
   * @return Iterator the list of itemsets.
   */
  public Iterator getItemsets() {
    return itemsets.iterator();
  }

  /**
   * Returns the list of items belonging to the model.
   *
   * @return Iterator the list of items.
   */
  public Iterator getItems() {
    return items.iterator();
  }

  /**
   * Remove a rule from the model. If rebuild_itemsets is true, then the frequent itemsets
   * are computed from the new set of association rules. Otherwise, itemsets will not change.
   * Returns true if and only if the rule has been removed from the model.
   *
   * @param rule AssociationRuleManager the association rule to remove from the model.
   * @param rebuild_itemsets boolean if true computes the frequent itemsets from association
   * rules.
   * @return boolean false if the rule is not removed.
   */
  public boolean removeRule(AssociationRuleManager rule, boolean rebuild_itemsets) {

    boolean ris = association_rules.remove(rule);

    if (rebuild_itemsets && ris) {
      // vengono ri-generati gli itemsets frequenti dalle RdA non rimosse...
      itemsets = RdA2ItemsetUtility.compute(association_rules, MAX_NUMBER_OF_RULES);

      // scorre gli itemsets per calcolare gli items...
      items = Itemset2ItemUtility.compute(itemsets);
    }
    return ris;
  }

  /**
   * Removes a set of association rules from the list of rules. If rebuild_itemsets is true,
   * then the frequent itemsets are computed from the new set of association rules.
   * Otherwise, itemsets will not change.
   *
   * @param rules AssociationRuleManager[] a set of association rules to add
   * to the model
   * @param rebuild_itemsets boolean
   */
  public void removeRules(AssociationRuleManager[] rules, boolean rebuild_itemsets) {
    boolean b1 = false;
    for (int i = 0; i < rules.length; i++) {
      if (association_rules.remove(rules[i]))
        b1 = true;
    }
    if (rebuild_itemsets && b1) {
      // vengono ri-generati gli itemsets frequenti dalle RdA non rimosse...
      itemsets = RdA2ItemsetUtility.compute(association_rules, MAX_NUMBER_OF_RULES);

      // scorre gli itemsets per calcolare gli items...
      items = Itemset2ItemUtility.compute(itemsets);
    }
  }

  /**
   * Removes a frequent itemset from the model. Also removes from the model the
   * association rules referred by the itemset.
   * Returns true if and only if the itemset has been removed from the model.
   *
   * @param itemset ItemsetManager the itemset to be removed.
   * @return boolean
   */
  public boolean removeItemset(ItemsetManager itemset) {

    //System.out.println("Itemset to be removed: "+itemset);
    boolean ris = itemsets.remove(itemset);

    if (ris) {
      // remove association rules refered...
      Iterator iter = association_rules.iterator();
      Vector v = new Vector();
      while (iter.hasNext()) {
        AssociationRule rule = (AssociationRule) iter.next();
        if (rule.contains(itemset))
          v.add(rule);
      }
      for (int j=0; j<v.size();j++)
        association_rules.remove(v.get(j));

      // scorre gli itemsets per calcolare gli items...
      items = Itemset2ItemUtility.compute(itemsets);
    }
    return ris;
  }

  /**
   * Removes a set of frequent itemsets from the model. Also removes from the model the
   * association rules referred by the itemsets.
   *
   * @param itemsets ItemsetManager[] a list of itemsets.
   */
  public void removeItemsets(ItemsetManager[] itemsets) {
    for (int i = 0; i < itemsets.length; i++) {
      removeItemset(itemsets[i]);
    }
    // remove association rules refered...
    Iterator iter = association_rules.iterator();
    Vector v = new Vector();
    while (iter.hasNext()) {
      AssociationRule rule = (AssociationRule) iter.next();
      for (int i = 0; i < itemsets.length; i++) {
        if (rule.contains(itemsets[i])) {
          v.add(rule);
          break;
        }
      }
    }
    for (int j=0; j<v.size();j++)
      association_rules.remove(v.get(j));

    // scorre gli itemsets per calcolare gli items...
    items = Itemset2ItemUtility.compute(this.itemsets);


  }

  /**
   * Returns the minimum support of rules.
   *
   * @return double
   */
  public double getMinSupport() {
    return min_support.doubleValue();
  }

  /**
   * Returns the minimum confidence of rules.
   *
   * @return double
   */
  public double getMinConfidence() {
    return min_confidence.doubleValue();
  }

  /**
   * Returns the  number of transactions used to build the model.
   *
   * @return int
   */
  public int getNumberOfTransactions() {
    return number_of_transactions.intValue();
  }

  /**
   * Returns the  number of items.
   *
   * @return int
   */
  public int getNumberOfItems() {
    return items.size();
  }

  /**
   * Returns the  number of itemsets.
   *
   * @return int
   */
  public int getNumberOfItemsets() {
    return itemsets.size();
  }

  /**
   * Returns the number of rules.
   *
   * @return int
   */
  public int getNumberOfRules() {
    return association_rules.size();
  }

  /**
   * Returns the maximum number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @return Integer
   */
  public Integer getMaxTransactionSize() {
    return max_number_of_items_per_TA;
  }

  /**
   * Sets the maximum number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @param max_number_of_items_per_TA int
   */
  public void setMaxTransactionSize(int max_number_of_items_per_TA) {
    this.max_number_of_items_per_TA = new Integer(max_number_of_items_per_TA);
  }

  /**
   * Returns the average number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @return Double
   */
  public Double getAverageTransactionSize() {
    return avg_number_of_items_per_TA;
  }

  /**
   * Sets the average number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @param avg_number_of_items_per_TA double
   */
  public void setAvgTransactionSize(double avg_number_of_items_per_TA) {
    this.avg_number_of_items_per_TA = new Double(avg_number_of_items_per_TA);
  }

  /**
   * Returns true if the model contains boolean association rules. Boolean association
   * rules (or intra-attribute rules) have the form: <br>
   * <p>spaghetti, tomato --> parmesan.</p><br>
   * Returns false if the model contains inter-attribute rules, such as<br>
   * <p>carType = racing, homeInsurance = false --> married = yes.</p>
   *
   * @return boolean true if the rules is an intra-attribute rules. False otherwise.
   */
  public boolean isBoolean() {
    if (items.size() != 0) {
      Iterator iter = items.iterator();
      while (iter.hasNext()) {
        Item item = (Item) iter.next();
        if (!item.isBoolean())
          return false;
      }
      return true;
    }
    return false;
  }

  /**
   * Returns ture if the model do not contain frequent itemsets and association rules.
   *
   * @return boolean
   */
  public boolean isEmpty() {
    return (itemsets.size() == 0 && association_rules.size() == 0);
  }

  /**
   * Tests if the model contains an item.
   *
   * @param item ItemManager
   * @return boolean true if the model contains the item.
   */
  public boolean contains(ItemManager item) {
    Iterator iter = items.iterator();
    while (iter.hasNext()) {
      Item i = (Item) iter.next();
      if (i.equals(item))
        return true;
    }
    return false;
  }

  /**
   * Tests if the model contains an input itemset.
   *
   * @param itemset ItemsetManager
   * @return boolean true if the model contains the itemset.
   */
  public boolean contains(ItemsetManager itemset) {
    Iterator iter = itemsets.iterator();
    while (iter.hasNext()) {
      Itemset i = (Itemset) iter.next();
      if (i.equals(itemset))
        return true;
    }
    return false;
  }

  /**
   * Tests if the model contains an input association rule.
   *
   * @param rule AssociationRuleManager
   * @return boolean true if the model contains the association rule.
   */
  public boolean contains(AssociationRuleManager rule) {
    Iterator iter = association_rules.iterator();
    while (iter.hasNext()) {
      AssociationRule i = (AssociationRule) iter.next();
      if (i.equals(rule))
        return true;
    }
    return false;
  }

  /**
   * Save the object in the system repository.
   * The destination path is provided by the object_path variable.
   *
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws kddml.Core.KDDMLCoreException {

    Document root = initializeDocument();

    // attribute settings...
    Element association_model;

    association_model = XMLUtility.doc.createElement("AssociationModel");
    if (mining_schema != null)
      association_model = appendMiningSchema((MiningSchema) mining_schema, association_model);

    setAttributes(association_model);
    // costruisce gli elementi PMML che rappresentano gli items...
    Vector ht1 = appendPMMLItems(association_model);
    // costruisce gli elementi PMML che rappresentano gli itemsets...
    Vector ht2 = appendPMMLItemsets(association_model, ht1);
    // costruisce gli elementi PMML che rappresentano le regole di associazione...
    ht1 = null;
    System.gc();
    appendPMMLRules(association_model, ht2);

    root.getDocumentElement().appendChild(association_model);
    ht2 = null;
    System.gc();

    XMLUtility.toFile(object_path.getAbsolutePath(), root);
  }

  /**
   * Clones this model returning a new association model as output.
   *
   * @param output_path File the path in the model system repository containing the cloned
   * model
   * @return AssociationModel
   */
  public AssociationModel clone(java.io.File output_path) {

    AssociationModel output = new AssociationModel(output_path);
    output.data_dictionary = (DataDictionary) this.data_dictionary.clone();
    output.mining_schema = (MiningSchema) this.mining_schema.clone();
    output.min_support = new Double(this.getMinSupport());
    output.min_confidence = new Double(this.getMinConfidence());
    output.number_of_transactions = new Integer(this.getNumberOfTransactions());

    output.items = new TreeSet(this.items);
    output.itemsets = new TreeSet(this.itemsets);
    output.association_rules = new TreeSet(this.association_rules);

    output.model_name = new String(this.getModelName());
    output.algorithm_name = new String(this.getAlgorithmName());
    output.max_number_of_items_per_TA = new Integer(this.getMaxTransactionSize().intValue());
    output.avg_number_of_items_per_TA = new Double(this.getAverageTransactionSize().doubleValue());

    return output;
  }


  /**
   * Returns a representation of the rules contained in the model as set of instances.
   * The number of instances is equal to the number of association rules.
   * The list of attributes is reported below:
   * <ol>
   * <li> the items of the body as string in comma separated format (attribute body); </li>
   * <li> the items of the head as string in comma separated format (attribute head); </li>
   * <li> the body cardinality (attribute body_cardinality); </li>
   * <li> the head cardinality (attribute head_cardinality); </li>
   * <li> the support of the rule (attribute support); </li>
   * <li> the confidence of the rule (attribute confidence). </li>
   *
   * @return Object the set of instances as generic object.
   */
  public Object rdaToInstances() {
    FastVector fv = new FastVector();
    fv = null;
    Attribute attr1 = new Attribute("head", fv);
    Attribute attr2 = new Attribute("body", fv);
    Attribute attr3 = new Attribute("head_cardinality");
    Attribute attr4 = new Attribute("body_cardinality");
    Attribute attr5 = new Attribute("rda_support");
    Attribute attr6 = new Attribute("rda_confidence");

    FastVector attributes = new FastVector();
    attributes.addElement(attr1);
    attributes.addElement(attr2);
    attributes.addElement(attr3);
    attributes.addElement(attr4);
    attributes.addElement(attr5);
    attributes.addElement(attr6);

    Instances ris = new Instances(this.model_name+"_association_rules", attributes, 0);
    // scorre le regole di association per ottenere le singole istanze
    Iterator iter = getAssociationRules();
    while(iter.hasNext()) {
      AssociationRule rule = (AssociationRule) iter.next();
      Itemset head = (Itemset) rule.antecedent;
      Itemset body = (Itemset) rule.consequent;
      Instance instance = new Instance(6);
      instance.setDataset(ris);
      instance.setValue(0, head.toTransaction().toString());
      instance.setValue(1, body.toTransaction().toString());
      instance.setValue(2, head.getSize());
      instance.setValue(3, body.getSize());
      instance.setValue(4, rule.getSupport());
      instance.setValue(5, rule.getConfidence());
      ris.add(instance);
    }

    return ris;
  }

  /**
   * Returns a representation of the itemsets contained in the model as set of instances.
   * The number of instances is equal to the number of itemsets.
   * The list of attributes is reported below:
   * <ol>
   * <li> the items belonging to the itemset as string in comma separated format (attribute itemset); </li>
   * <li> the itemset cardinalty (attribute itemset_cardinality); </li>
   * <li> the itemset support (attribute itemset_support). </li>
   *
   * @return Object the set of instances as generic object.
   */
  public Object itemsetsToInstances() {
    FastVector fv = new FastVector();
    fv = null;
    Attribute attr7 = new Attribute("itemset", fv);
    Attribute attr8 = new Attribute("itemset_cardinality");
    Attribute attr9 = new Attribute("itemset_support");

    FastVector attributes = new FastVector();
    attributes.addElement(attr7);
    attributes.addElement(attr8);
    attributes.addElement(attr9);

    Instances ris = new Instances(this.model_name+"_association_rules", attributes, 0);

    // scorre gli itemsets
    Iterator iter = getItemsets();
    while(iter.hasNext()) {
      Instance instance = new Instance(3);
      Itemset itemset = (Itemset) iter.next();
      instance.setDataset(ris);
      instance.setValue(0, itemset.toTransaction().toString());
      instance.setValue(1, itemset.getSize());
      if (itemset.getSupport() != null)
        instance.setValue(2, itemset.getSupport().doubleValue());
      else
        instance.setMissing(2);
      ris.add(instance);
    }

    return ris;
  }


  protected Element setAttributes(Element association_model)
      throws AssociationModelException {

    association_model.setAttribute("functionName", "associationRules");
    association_model.setAttribute("algorithmName", algorithm_name);
    association_model.setAttribute("modelName", model_name);

    association_model.setAttribute("minimumSupport", min_support.toString());
    association_model.setAttribute("minimumConfidence", min_confidence.toString());

    association_model.setAttribute("numberOfTransactions", number_of_transactions.toString());

    association_model.setAttribute("numberOfItems",
                                   new Integer(items.size()).toString());
    association_model.setAttribute("numberOfItemsets",
                                   new Integer(itemsets.size()).toString());
    association_model.setAttribute("numberOfRules",
                                   new Integer(association_rules.size()).toString());

    if (max_number_of_items_per_TA != null)
      association_model.setAttribute("maxNumberOfItemsPerTA",
                                     max_number_of_items_per_TA.toString());
    if (avg_number_of_items_per_TA != null)
      association_model.setAttribute("avgNumberOfItemsPerTA",
                                     avg_number_of_items_per_TA.toString());

    return association_model;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = super.toString();
    s = s + "---------------------------------\n";
    s = s + "Association Model:\n";
    s = s + "---------------------------------\n";
    s = s + "Number of Items: "+items.size()+"\n";
    s = s + "Number of Itemset: "+itemsets.size()+"\n";
    s = s + "Number of Rules: "+association_rules.size()+"\n";
    Iterator r = items.iterator();
    s = s + "---------\n";
    s = s + "List of items: ";
    while(r.hasNext()) {
      Item i = (Item) r.next();
      s = s+i.toString()+", ";
    }
    s = s + "\n";
    s = s + "---------\n";
    s = s + "List of frequent itemsets: \n";
    r = itemsets.iterator();
    while(r.hasNext()) {
      Itemset i = (Itemset) r.next();
      s = s + i.toString()+"\n";
    }
    s = s + "---------\n";
    s = s + "List of association rules: \n";
    r = association_rules.iterator();
    while(r.hasNext()) {
      AssociationRule i = (AssociationRule) r.next();
      s = s + i.toString()+"\n";
    }
    return s;
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML() throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.PMML_ASSOCIATION_MODEL);

    String s1 = getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    try {
      XMLUtility.formatting(s1, s2, s3, null, null);
    }
    catch(Exception e) {
      String msg = "Exception found during the HTML code generation.";
      msg = e.toString();
      throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
    }
  }


  protected Vector appendPMMLItems(Element association_model) {
    Vector ht_item = new Vector();
    Iterator iter = items.iterator();
    int id = 0;
    while(iter.hasNext()) {
      id++;
      Item item = (Item) iter.next();
      Element el = XMLUtility.doc.createElement("Item");
      XMLUtility.setAttribute(el, "id",new Integer(id).toString());
      XMLUtility.setAttribute(el, "value",item.toString());
      if (item.mapped_value != null)
        XMLUtility.setAttribute(el, "mappedValue", item.mapped_value);
      if (item.weight != null)
        el.setAttribute("weight", item.weight.toString());
      association_model.appendChild(el);
      if (!ht_item.contains(item))
        ht_item.add(item);
    }
    return ht_item;
  }

  protected Vector appendPMMLItemsets(Element association_model, Vector ht_items) {
    Vector ht_itemsets = new Vector();
    Iterator iter = itemsets.iterator();
    int id = 0;
    Itemset itemset;
    while(iter.hasNext()) {
      id++;
      itemset = (Itemset) iter.next();
      Element el = XMLUtility.doc.createElement("Itemset");
      el.setAttribute("id",new Integer(id).toString());
      el.setAttribute("numberOfItems",new Integer(itemset.getSize()).toString());
      if (itemset.getSupport() != null)
        el.setAttribute("support",new Double(itemset.getSupport().doubleValue()).toString());
      // scorre gli items...
      Iterator iter2 = itemset.getItems();
      Item item;
      while (iter2.hasNext()) {
        item = (Item) iter2.next();
        Element ref = XMLUtility.doc.createElement("ItemRef");
        ref.setAttribute("itemRef",(new Integer(ht_items.indexOf(item) + 1)).toString());
        el.appendChild(ref);
      }
      association_model.appendChild(el);
      if (!ht_itemsets.contains(itemset))
        ht_itemsets.add(itemset);
    }
    return ht_itemsets;
  }

  protected void appendPMMLRules(Element association_model, Vector ht_itemsets) {
    Iterator iter = association_rules.iterator();
    while(iter.hasNext()) {
      AssociationRule rule = (AssociationRule) iter.next();
      Element el = XMLUtility.doc.createElement("AssociationRule");
      el.setAttribute("support",new Double(rule.getSupport()).toString());
      el.setAttribute("confidence",new Double(rule.getConfidence()).toString());
      ItemsetManager antecedent = rule.getAntecedent();
      el.setAttribute("antecedent",(new Integer(ht_itemsets.indexOf(antecedent) + 1)).toString());
      ItemsetManager consequent = rule.getConsequent();
      el.setAttribute("consequent",(new Integer(ht_itemsets.indexOf(consequent) + 1)).toString());
      association_model.appendChild(el);
    }
  }

  /**
   * Only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    try {
      java.io.File f = new java.io.File("C:/Documents and Settings/romei/kddml/Repository/Models/Rda/tmp.xml");
      AssociationRepositoryFactory fact = new AssociationRepositoryFactory(f);
      AssociationModel m = (AssociationModel) fact.newInstance();
      AssociationModel m2 = (AssociationModel) m.clone(new java.io.File(""));
      Itemset ant = new Itemset();
      Itemset cons = new Itemset();
      Item i1 = new Item("CIOCCOLATINI");
      Item i2 = new Item("ACQUE_MINERALI");
      Item i3 = new Item("PANDORO_TRADIZIONALE");
      ant.addItem(i1);
      cons.addItem(i2);
      cons.addItem(i3);
      AssociationRule r = new AssociationRule(ant, cons, 0.5, 0.5);
      System.out.println(m.removeRule(r, false));
      System.out.println(m2.contains(r));
    }
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
  }
}
