/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.*;

/**
 * A manager interface for an association model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface AssociationModelManager
    extends kddml.Core.DataMining.MiningModelManager {

  /**
   * Adds a new association rule to this model.
   * Returns true if the rule has been added. Returns false if the model already
   * contains the rule.
   *
   * @param rule AssociationRuleManager the association rule to add to the model.
   * @return boolean true if the rule do not exist in the model, false otherwise.
   */
  public abstract boolean addRule(AssociationRuleManager rule);

  /**
   * Adds a set of association rules to this model.
   *
   * @param rules_array AssociationRuleManager[] a set of association rules to add
   * to the model
   */
  public abstract void addRules(AssociationRuleManager[] rules_array);

  /**
   * Remove a rule from the model. If rebuild_itemsets is true, then the frequent itemsets
   * are computed from the new set of association rules. Otherwise, itemsets will not change.
   * Returns true if and only if the rule has been removed from the model.
   *
   * @param rule AssociationRuleManager the association rule to remove from the model.
   * @param rebuild_itemsets boolean if true computes the frequent itemsets from association
   * rules.
   * @return boolean false if the rule is not removed.
   */
  public abstract boolean removeRule(AssociationRuleManager rule,
                                     boolean rebuild_itemsets);

  /**
   * Removes a set of association rules from the list of rules. If rebuild_itemsets is true,
   * then the frequent itemsets are computed from the new set of association rules.
   * Otherwise, itemsets will not change.
   *
   * @param rules_array AssociationRuleManager[] a set of association rules to add
   * to the model
   * @param rebuild_itemsets boolean
   */
  public abstract void removeRules(AssociationRuleManager[] rules_array,
                                   boolean rebuild_itemsets);

  /**
   * Removes a frequent itemset from the model. Also removes from the model the
   * association rules referred by the itemset.
   * Returns true if and only if the itemset has been removed from the model.
   *
   * @param itemset ItemsetManager the itemset to be removed.
   * @return boolean
   */
  public abstract boolean removeItemset(ItemsetManager itemset);

  /**
   * Removes a set of frequent itemsets from the model. Also removes from the model the
   * association rules referred by the itemsets.
   *
   * @param itemsets_array ItemsetManager[] a list of itemsets.
   */
  public abstract void removeItemsets(ItemsetManager[] itemsets_array);

  /**
   * Tests if the model contains an input itemset.
   *
   * @param itemset ItemsetManager
   * @return boolean true if the model contains the itemset.
   */
  public abstract boolean contains(ItemsetManager itemset);

  /**
   * Tests if the model contains an input association rule.
   *
   * @param rule AssociationRuleManager
   * @return boolean true if the model contains the association rule.
   */
  public abstract boolean contains(AssociationRuleManager rule);

  /**
   * Tests if the model contains an item.
   *
   * @param item ItemManager
   * @return boolean true if the model contains the item.
   */
  public boolean contains(ItemManager item);

  /**
   * Returns the list of items belonging to the model.
   *
   * @return Iterator the list of items.
   */
  public abstract Iterator getItems();

  /**
   * Returns the list of itemsets belonging to the model.
   *
   * @return Iterator the list of itemsets.
   */
  public abstract Iterator getItemsets();

  /**
   * Returns the list of association rules belonging to the model.
   *
   * @return Iterator the list of association rules.
   */
  public abstract Iterator getAssociationRules();

  /**
   * Returns the minimum support of rules.
   *
   * @return double
   */
  public abstract double getMinSupport();

  /**
   * Returns the minimum confidence of rules.
   *
   * @return double
   */
  public abstract double getMinConfidence();

  /**
   * Returns the  number of transactions used to build the model.
   *
   * @return int
   */
  public abstract int getNumberOfTransactions();

  /**
   * Returns the  number of items.
   *
   * @return int
   */
  public abstract int getNumberOfItems();

  /**
   * Returns the  number of itemsets.
   *
   * @return int
   */
  public abstract int getNumberOfItemsets();

  /**
   * Returns the number of rules.
   *
   * @return int
   */
  public abstract int getNumberOfRules();

  /**
   * Returns the average number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @return Double
   */
  public abstract Double getAverageTransactionSize();

  /**
   * Sets the maximum number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @param max_number_of_items_per_TA int
   */
  public void setMaxTransactionSize(int max_number_of_items_per_TA);

  /**
   * Returns the maximum number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @return Integer
   */
  public abstract Integer getMaxTransactionSize();

  /**
   * Sets the average number of items in the transactions contained in the input
   * data used to build the model.
   *
   * @param avg_number_of_items_per_TA double
   */
  public void setAvgTransactionSize(double avg_number_of_items_per_TA);

  /**
   * Returns true if the model contains boolean association rules. Boolean association
   * rules (or intra-attribute rules) have the form: <br>
   * <p>spaghetti, tomato --> parmesan.</p><br>
   * Returns false if the model contains inter-attribute rules, such as<br>
   * <p>carType = racing, homeInsurance = false --> married = yes.</p>
   *
   * @return boolean true if the rules is an intra-attribute rules. False otherwise.
   */
  public abstract boolean isBoolean();

  /**
   * Returns a representation of the rules contained in the model as set of instances.
   * The number of instances is equal to the number of association rules.
   * The list of attributes is reported below:
   * <ol>
   * <li> the items of the body as string in comma separated format (attribute body); </li>
   * <li> the items of the head as string in comma separated format (attribute head); </li>
   * <li> the body cardinality (attribute body_cardinality); </li>
   * <li> the head cardinality (attribute head_cardinality); </li>
   * <li> the support of the rule (attribute support); </li>
   * <li> the confidence of the rule (attribute confidence). </li>
   *
   * @return Object the set of instances as generic object.
   */
  public abstract Object rdaToInstances();

  /**
   * Returns a representation of the itemsets contained in the model as set of instances.
   * The number of instances is equal to the number of itemsets.
   * The list of attributes is reported below:
   * <ol>
   * <li> the items belonging to the itemset as string in comma separated format (attribute itemset); </li>
   * <li> the itemset cardinalty (attribute itemset_cardinality); </li>
   * <li> the itemset support (attribute itemset_support). </li>
   *
   * @return Object the set of instances as generic object.
   */
  public abstract Object itemsetsToInstances();

  /**
   * Clones this model returning a new association model as output.
   *
   * @param output_path File the path in the model system repository containing the cloned
   * model
   * @return AssociationModel
   */
  public AssociationModel clone(java.io.File output_path);
}
