/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataMining.AssociationRules;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.TreeSet;

/**
 * A factory class that creates a generic model from a proprietary (i.e. not external)
 * resource.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class AssociationProprietaryFactory extends ModelProprietaryFactory {

  protected Double min_support;

  protected Double min_confidence;

  protected Integer number_of_transactions;

  protected Integer max_number_of_items_per_TA;

  protected Double avg_number_of_items_per_TA;

  protected TreeSet items;

  protected TreeSet itemsets;

  protected TreeSet association_rules;

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param min_support double the minimum support.
   * @param min_confidence double the minimum confidence.
   * @param number_of_transactions int the number of transactions.
   * @throws KDDMLCoreException
   */
  public AssociationProprietaryFactory(java.io.File f,
                                            DataDictionary dd,
                                            MiningSchema ms,
                                            double min_support,
                                            double min_confidence,
                                            int number_of_transactions)
      throws KDDMLCoreException {

    super(f, dd, ms);
    // controlli sui parametri immessi
    CheckerUtility.checkMinSupport(new Double(min_support));
    this.min_support = new Double(min_support);

    CheckerUtility.checkMinConfidence(new Double(min_confidence));
    this.min_confidence = new Double(min_confidence);

    CheckerUtility.checkNumTransactions(new Integer(number_of_transactions));
    this.number_of_transactions = new Integer(number_of_transactions);
  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    AssociationModel model = new AssociationModel(physical_file);

    model = (AssociationModel) setMiningSchema((MiningModel) model);
    model = (AssociationModel) setDataDictionary((MiningModel) model);

    if (algorithm_name != null)
      model = (AssociationModel) setAlgorithmName((MiningModel) model);
    if (model_name != null)
      model = (AssociationModel) setModelName((MiningModel) model);

    model.min_support = min_support;
    model.min_confidence = min_confidence;
    model.number_of_transactions = number_of_transactions;
    model.max_number_of_items_per_TA = max_number_of_items_per_TA;
    model.avg_number_of_items_per_TA = avg_number_of_items_per_TA;

    return model;
  }

  /**
   * Sets the maximum number of items per transaction. This parameter is optional
   * in the output model.
   *
   * @param i int
   * @throws MiningModelException
   */
  public void setMaxNumberOfItemsPerTA(int i) throws MiningModelException {
    CheckerUtility.checkMaxItems(new Integer(i));
    max_number_of_items_per_TA = new Integer(i);
  }

  /**
   * Sets the average number of items per transaction. This parameter is optional
   * in the output model
   *
   * @param i double
   * @throws AssociationModelException
   */
  public void setAvgNumberOfItemsPerTA(double i) throws AssociationModelException {
    CheckerUtility.checkAvgItems(new Double(i));
    avg_number_of_items_per_TA = new Double(i);
  }
}
