/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.Iterator;

/**
 * Association defines a relationship between two itemsets where the antecendent
 * implies the consequent.
 * An association rule consists of four major parts:
 * <ol>
 * <li> the antecedent itemset (A); </li>
 * <li> the consequent itemset (C); </li>
 * <li> the support of the rule, that is, the relative frequency of transactions that contain A and C.
 * support(A->C) = support(A+C) </li>
 * <li> The confidence of the rule. confidence(A->C) = support(A+C) / support(A) </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class AssociationRule implements AssociationRuleManager {

  protected Itemset antecedent;

  protected Itemset consequent;

  /**
   * The support of the rule. Can not be null.
   * It is a number between 0 and 1, inclusive.
   */
  protected double support;

  /**
   * The confidence of the rule. Can not be null.
   * It is a number between 0 and 1, inclusive.
   */
  protected double confidence;

  /**
   * Constructor give the antecedent, the consequent, the support and the confidence.
   *
   * @param antecedent Itemset
   * @param consequent Itemset
   * @param support double
   * @param confidence double
   */
  public AssociationRule(Itemset antecedent,
                         Itemset consequent,
                         double support,
                         double confidence) {
    this.antecedent = antecedent;
    this.consequent = consequent;
    this.support = support;
    this.confidence = confidence;
  }

  /**
   * Returns the antecedent of the association rule. Cannot return null.
   *
   * @return ItemsetManager
   */
  public ItemsetManager getAntecedent() {
    return antecedent;
  }

  /**
   * Returns the consequent of the association rule. Cannot return null.
   *
   * @return ItemsetManager
   */
  public ItemsetManager getConsequent() {
    return consequent;
  }

  /**
   * Returns the support of this rule in the model as a percentage of the total number of
   * transactions. It is a number between 0 and 1, inclusive.
   *
   * @return double
   */
  public double getSupport() {
    return support;
  }

  /**
   * Returns the support for this rule in terms of absolute count of transactions
   * supporting it.
   *
   * @param num_transactions int the total number of transactions used to build the rules.
   * @return int
   */
  public int getAbsoluteSupport(int num_transactions) {
    double d = support * num_transactions;
    return (int) d;
  }

  /**
   * Returns the confidence of this rule in the model as a percentage of the total number of
   * transactions. It is a number between 0 and 1, inclusive.
   *
   * @return double
   */
  public double getConfidence() {
    return confidence;
  }

  /**
   * Returns the size of this rule in terms of the number of items in antecedent and consequent.
   * Returns a value >= 2.
   *
   * @return int
   */
  public int getSize() {
    return antecedent.getSize() + consequent.getSize();
  }

  /**
   * Tests if this rule satisfy an input transaction.
   * A transaction satisfy a rule I1, ..., In --> In+1, ..., Im if every item Ii with i in [1, m]
   * occours in the transaction.
   *
   * @param transaction TransactionManager
   * @return boolean
   */
  public boolean isSupported(TransactionManager transaction) {
    return ((this.antecedent.isSupported(transaction)) &
            (this.consequent.isSupported(transaction)));
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "";
    Iterator iter1 = antecedent.getItems();
    Iterator iter2 = consequent.getItems();
    while(iter1.hasNext())
      s = s + ((Item) iter1.next()).toString() + ", ";
    s = s + " --> ";
    while(iter2.hasNext())
      s = s + ((Item) iter2.next()).toString() + ", ";
    s = s + "(supp="+support+")";
    s = s + "(conf="+confidence+")";
    return s;
  }

  /**
   * Tests if two rules are equal.
   *
   * @param rule AssociationRuleManager
   * @return boolean
   */
  public boolean equals(Object rule) {
    // due regole sono uguali se hanno lo stesso antecedente e lo stesso conseguente...
    ItemsetManager antecedent1 = this.antecedent;
    ItemsetManager consequent1 = this.consequent;
    ItemsetManager antecedent2 = ((AssociationRule) rule).getAntecedent();
    ItemsetManager consequent2 = ((AssociationRule) rule).getConsequent();
    return ((antecedent1.equals(antecedent2)) && (consequent1.equals(consequent2)));
  }

  /**
   * Test if an association rule contains an input itemset.
   * Return true if the association rule contains the itemset in the body or in the
   * head.
   *
   * @param itemset ItemsetManager
   * @return boolean
   */
  public boolean contains(ItemsetManager itemset) {
    return (antecedent.equals((Itemset) itemset) ||
            consequent.equals((Itemset) itemset));
  }
}
