/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

/**
 * A manager interface for an association rule.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface AssociationRuleManager {

  /**
   * Test if an association rule contains an input itemset.
   * Return true if the association rule contains the itemset in the body or in the
   * head.
   *
   * @param itemset ItemsetManager
   * @return boolean
   */
  public abstract boolean contains(ItemsetManager itemset);

  /**
   * Tests if two rules are equal.
   *
   * @param rule AssociationRuleManager
   * @return boolean
   */
  public abstract boolean equals(Object rule);

  /**
   * Returns the antecedent of the association rule. Cannot return null.
   *
   * @return ItemsetManager
   */
  public abstract ItemsetManager getAntecedent();

  /**
   * Returns the consequent of the association rule. Cannot return null.
   *
   * @return ItemsetManager
   */
  public abstract ItemsetManager getConsequent();

  /**
   * Returns the support of this rule in the model as a percentage of the total number of
   * transactions. It is a number between 0 and 1, inclusive.
   *
   * @return double
   */
  public abstract double getSupport();

  /**
   * Returns the confidence of this rule in the model as a percentage of the total number of
   * transactions. It is a number between 0 and 1, inclusive.
   *
   * @return double
   */
  public abstract double getConfidence();

  /**
   * Returns the size of this rule in terms of the number of items in antecedent and consequent.
   * Returns a value >= 2.
   *
   * @return int
   */
  public int getSize();

  /**
   * Tests if this rule satisfy an input transaction.
   * A transaction satisfy a rule I1, ..., In --> In+1, ..., Im if every item Ii with i in [1, m]
   * occours in the transaction.
   *
   * @param transaction TransactionManager
   * @return boolean
   */
  public boolean isSupported(TransactionManager transaction);

  /**
   * Returns the support for this rule in terms of absolute count of transactions
   * supporting it.
   *
   * @param num_transactions int the total number of transactions used to build the rules.
   * @return int
   */
  public int getAbsoluteSupport(int num_transactions);

}
