/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.HashSet;
import java.util.Iterator;
/**
 * A class used to represent an item in transactional analisys.
 * An item consists of three major parts:
 * <ol>
 * <li> the item name; </li>
 * <li> the item value for multivalue association rules; it can be null for
 * boolean association rules; </li>
 * <li> a mapped value to which the original item value is mapped. For instance, this
 * could be a product name if the original value is an EAN code.  </li>
 * </ol>
 * Items can have additional features such as the price or the quantity of a purchased
 * product.
 */
public class Item implements ItemManager {

  /**
   * The name of the item.
   */
  protected String name;

  /**
   * The value of the item as in the input data.
   */
  protected String value;

  /**
   * Optional, a value to which the original item value is mapped.
   * For instance, this could be a product name if the original value is an EAN code.
   * Null if no mapped value is specified.
   */
  protected String mapped_value;

  /**
   * The weight of the item.
   * For example, the price or value of an item.
   * Null if no weight is specified.
   */
  protected Double weight;

  /**
   * Boolean item. In boolean items, the field value is null.
   */
  protected boolean is_boolean;

  /**
   * The list of features related to this item. Can be empty.
   */
  protected HashSet features;

  /**
   * Constructor.
   * For boolean items, the input value coincides with the name of the item.
   * For no-boolean items, the input value is in the form "name = value".
   *
   * @param value String
   */
  public Item(String value) {
    if (value.startsWith("'"))
      value = value.substring(1, value.length());
    if (value.endsWith("'"))
      value = value.substring(0, value.length()-1);

    int index = value.indexOf("=");
    if (index == -1) {
      this.name = value;
      this.value = null;
      mapped_value = null;
      weight = null;
      is_boolean = true;
    }
    else {
      this.name = value.substring(0, index);
      this.value = value.substring(index + 1, value.length());
      mapped_value = null;
      weight = null;
      is_boolean = false;
    }
    this.features = new HashSet();
  }

  /**
   * Tests if two items are equal.
   *
   * @param item Object
   * @return boolean
   */
  public boolean equals(Object item) {
    String v1, v2;
    if ((((Item) item).value == null) ||
        (this.value == null) ) {
      // items di tipo booleano: si confrontano gli elementi name
      v1 = this.name.replaceAll(" ","");
      v2 = ((Item) item).name.replaceAll(" ","");
      return v1.equalsIgnoreCase(v2);
    }
    else {
      // items di tipo multivalore: si confrontano name e value
      String v3, v4;
      v1 = this.name.replaceAll(" ","");
      v2 = ((Item) item).name.replaceAll(" ","");
      v3 = this.value.replaceAll(" ", "");
      v4 = ( (Item) item).value.replaceAll(" ", "");
      return ((v1.equalsIgnoreCase(v2)) && (v3.equalsIgnoreCase(v4)));
    }
  }

  /**
   * Returns the item name.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the item value. Returns null if the method isBoolean() returns true.
   *
   * @return String
   */
  public String getValue() {
    return value;
  }

  /**
   * Returns the mapped value of the item.
   * Returns null if the item do not have a mapped value.
   *
   * @return String
   */
  public String getMappedValue() {
    return mapped_value;
  }

  /**
   * Sets the mapped value to the item.
   *
   * @param mapped_value String
   */
  public void setMappedValue(String mapped_value) {
    this.mapped_value = mapped_value;
  }

  /**
   * Returns the weight of the item.
   *
   * @return Double
   */
  public Double getWeight() {
    return weight;
  }

  /**
   * Sets the weight of the item.
   *
   * @param weight double
   */
  public void setWeight(double weight) {
    this.weight = new Double(weight);
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    if (value == null)
      return name;
    else
      return name + "=" + value;
  }

  /**
   * Test if an item has the field value (e.g. "outlook = sunny").
   * Returns true if the field value is not null.
   *
   * @return boolean
   */
  public boolean isBoolean() {
    return is_boolean;
  }

  /**
   * Adds a new feature to this item.
   * Returns false if the item already contains this feature.
   *
   * @param attr ItemFeatureManager
   * @return boolean
   */
  public boolean addNewFeature(ItemFeatureManager attr){
    return this.features.add(attr);
  }

  /**
   * Returns the specified feature. Returns null if the feature do not exist.
   *
   * @param name String the name of the wished feature (e.g. price, quantity).
   * @return ItemFeatureManager
   */
  public ItemFeatureManager getFeature(String name) {

    Iterator i = this.getFeatures();
    while( i.hasNext() ){
      ItemFeature ia = (ItemFeature) i.next();
      if( ia.name().equalsIgnoreCase(name) )
        return ia;
    }
    return null;
  }

  /**
   * Returns true if the specified feature do not exist or if his value is missing.
   *
   * @param name String
   * @return boolean
   */
  public boolean isMissingFeature(String name){
    ItemFeatureManager a = this.getFeature(name);
    if( a == null )
    	return true;
    return a.isMissing();
  }

  /**
   * Returns the list of features of the item.
   *
   * @return Iterator
   */
  public Iterator getFeatures(){
  	return this.features.iterator();
  }

  /**
   * Returns all numerical feature values of this item (e.g. if the item has two
   * numeric feature, price and quantity, this method returns a array with lenght 2
   * containing the price).
   *
   * @return double[]
   */
  /*public double[] getAllNumericFeaturesValues(){
    double[] array = new double[ this.features.size() ];
    int index = 0;

    Iterator i = this.getFeatures();
    while( i.hasNext() ){
      ItemFeature a = (ItemFeature)i.next();
      if( a.isNumeric() ) {
        array[index++] = a.value();
      }
    }

    if( index != array.length ) {
      double[] trim_array = new double[index];
      System.arraycopy(array, 0, trim_array, 0, index);
      return trim_array;
    }

    return array;
  }*/

  /**
   * Tests if this item has some features associated.
   *
   * @return boolean
   */
  public boolean hasFeatures() {
    return !this.features.isEmpty();
  }

  public int hashCode() {
    return this.toString().hashCode();
  }
}
