/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

/**
 * ItemFeature represents an item attribute (e.g. the price of a purchased product).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra
 * @version 1.0
 */
public class ItemFeature implements ItemFeatureManager {

  public static final String MISSING_SVALUE = new String("?");

  public static final double MISSING_DVALUE = Double.NaN;

  /**
   * The name of the feature (e.g. price of quantity).
   */
  protected String name;

  /**
   * The value of the feature.
   */
  protected Object value;

  protected boolean is_numeric;

  protected boolean is_missing;

  /**
   * Constructor. Builds a missing object value.
   *
   * @param name String
   * @param is_numeric boolean
   */
  public ItemFeature(String name, boolean is_numeric) {
    this.name = name;
    is_missing = true;
    this.is_numeric = is_numeric;
    value = null;
  }

  /**
   * Tests if the item feature is numeric.
   *
   * @return boolean
   */
  public boolean isNumeric() {
    return this.is_numeric;
  }

  /**
   * Tests if the item feature value is missing.
   *
   * @return boolean
   */
  public boolean isMissing(){
    return this.is_missing;
  }

  /**
   * Returns the name of this feature.
   *
   * @return String
   */
  public String name() {
    return this.name;
  }

  /**
   * Returns the value of this numeric feature.
   * Throws an exception if the feature is not numeric.
   *
   * @throws IllegalArgumentException
   * @return double
   */
  public double value() throws IllegalArgumentException {
    if( !isNumeric() )
      throw new IllegalArgumentException();
    if( isMissing() )
      return MISSING_DVALUE;
    return ((Double)this.value).doubleValue();
  }

  /**
   * Returns the value of this string feature.
   * Throws an exception if the feature is not string.
   *
   * @throws IllegalArgumentException
   * @return double
   */
  public String stringValue() throws IllegalArgumentException {
    if(isNumeric())
      throw new IllegalArgumentException();
    if( isMissing() )
      return MISSING_SVALUE;
    return (String) this.value;
  }

  /**
   * Sets the numeric value to this numeric item feature.
   * Throws an exception if the item is not numerical.
   *
   * @param value double
   * @throws IllegalArgumentException
   */
  public void setValue(double value) throws IllegalArgumentException {
    if (!is_numeric)
      throw new IllegalArgumentException();
    this.value = new Double(value);
    this.is_missing = false;
  }

  /**
   * Sets the value to this string item feature.
   * Throws an exception if the item is numerical.
   *
   * @param value String
   * @throws IllegalArgumentException
   */
  public void setValue(String value) throws IllegalArgumentException {
    if (is_numeric)
      throw new IllegalArgumentException();
    this.value = value;
    this.is_missing = false;
  }

  public int hashCode(){
    return this.name.hashCode();
  }

  /**
   * Tests if two items are equal. Two items are equal if they have the same name.
   *
   * @param o Object
   * @return boolean
   */
  public boolean equals(ItemFeature o){
    return this.name().equalsIgnoreCase(o.name());
  }
}
