/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.TreeSet;
import java.util.Iterator;
import java.util.Comparator;

/**
 * An Itemset object contains a component of an association rule and consists of items.
 * An itemset consists of one or more categorical values. For example, the itemset
 * {milk, bread, butter} is composed of three items. The same categorical values can be
 * referenced by multiple itemsets.
 * Single items that compose the itemset are mantained in a tree set structure.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class Itemset implements ItemsetManager {

  /**
   * The relative support of the itemset. Can be null.
   */
  protected Double support;

  /**
   * The list of items.
   */
  protected TreeSet items;

  /**
   * Empty constructor. Builds a new itemset without items.
   */
  public Itemset() {
    Comparator comp = new CompareItems();
    items = new TreeSet(comp);
    support = null;
  }

  /**
   * Constructor given the list of items as TreeSet.
   *
   * @param items TreeSet
   */
  protected Itemset(TreeSet items) {
    this.items = items;
    support = null;
  }

  public boolean addItem(ItemManager item) {
    return items.add(item);
  }

  /**
   * Sets the support of this itemsets.
   * Support must be a number between 0 and 1, inclusive.
   *
   * @param support double
   */
  public void setSupport(double support) {
    this.support = new Double(support);
  }

  /**
   * Returns the support ot the itemset in the model as a percentage in respect to
   * the total number of transactions. It is a number between 0 and 1, inclusive.
   * Can be null.
   *
   * @return Double
   */
  public Double getSupport() {
    return support;
  }

  /**
   * Returns the support for this itemset in terms of absolute count of transactions
   * supporting it.
   *
   * @param num_transactions int the total number of transactions used to build the itemset.
   * @return int
   */
  public int getAbsoluteSupport(int num_transactions) {
    double d = support.doubleValue() * num_transactions;
    return (int) d;
  }

  /**
   * Returns the number of items. Returns a value >=0.
   *
   * @return int
   */
  public int getSize() {
    return items.size();
  }

  public String toString() {
    if (items.size() == 0)
      return "Empty Itemset!!!";
    else {
      String s="";
      Iterator iter = items.iterator();
      while (iter.hasNext())
        s = s + ((Item) iter.next()).toString() + ", ";
      if (support != null)
        s = s + "(supp="+support+")";
      else
        s = s + "(supp=n/a)";
      return s;
    }
  }

  /**
   * Returns the list of items as iterator.
   *
   * @return Iterator
   */
  public Iterator getItems() {
     return items.iterator();
  }

  public boolean contain(ItemManager item) {
    Iterator iter = items.iterator();
    ItemManager i1 = item;
    while (iter.hasNext()) {
      Item i2 = (Item) iter.next();
      if (i1.equals(i2))
        return true;
    }
    return false;
  }

  /**
   * Tests if this itemset satisfy an input transaction.
   * A transaction satisfy an itemset I1, ..., In if every item Ii with i in [1, m]
   * occours in the transaction.
   *
   * @param transaction TransactionManager
   * @return boolean
   */
  public boolean isSupported(TransactionManager transaction) {
    Iterator iter = items.iterator();
    while(iter.hasNext()) {
      Item i = (Item) iter.next();
      if (!transaction.contains(i))
        return false;
    }
    return true;
  }

  /**
   * Tests if two itemsets are equals, i.e., they contains the same items.
   *
   * @param itemset Object
   * @return boolean
   */
  public boolean equals(Object itemset) {

    if (this.getSize() != ((Itemset) itemset).getSize())
      return false;
    Iterator it1 = this.getItems();
    Iterator it2 = ((Itemset) itemset).getItems();
    boolean b;
    while (it1.hasNext()) {
      Item item1 = (Item) it1.next();
      Item item2 = (Item) it2.next();
      b = item1.equals(item2);
      if (!b)
        return false;
    }
    return true;
  }

  /**
   * Returns the list of items as transaction.
   *
   * @return TransactionManager
   */
  public TransactionManager toTransaction() {
    Iterator iter = getItems();
    Item[] s = new Item[items.size()];
    int i=0;
    while(iter.hasNext()) {
      Item item = (Item) iter.next();
      s[i] = item;
      i++;
    }
    return (TransactionManager) new Transaction(s);
  }

  public static void main(String args[]) {

    Itemset set = new Itemset();
    set.addItem(new Item("A"));
    set.addItem(new Item("B"));
    System.out.println(set.contain(new Item("A")));
    System.out.println(((Transaction) set.toTransaction()).toString());
  }
}
