/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.*;

/**
 * This utility class contains methods and variables used inside this package
 * for model proprieties checking.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Elena Ratti
 * @version 1.0
 */
final class Itemset2RdAUtility {

  /**
   * Dato un insieme di itemsets, calcola le corrispondenti regole di associazione
   * che rispettano la minima confidenza specificata in ingresso.
   * L'operatore si comporta nel seguente modo:
   * per ogni insieme frequente per cui il numero di items sia >1 il supporto relativo viene messo
   * dentro ad una tabella hash con chiave l'insieme di items associato.
   * Una volta riempita la tabella hash ci facciamo restituire l'enumerazione di tutte le chiavi e
   * per ogni chiave C estraggo tutti i suoi sottoinsiemi e durante l'estrazione, per ogni sottoinsieme
   * "Q" estratto, controllo se support(C)/support(Q) >= minima confidenza specificata dall'utente.
   * Se  verificata la disuguaglianza allora inserisco tra le regole di associazione la regola Q => (C-Q).
   * Il supporto di C e di Q vengono ricavati dalla tabella hash passando come chiave rispettivamente
   * gli insieme di items C ed Q.
   */
  protected static TreeSet compute(TreeSet itemsets, double min_conf, int max_rules)
      throws AssociationModelException {

    Comparator comp = new CompareAssociationRules();
    TreeSet ris = new TreeSet(comp);
    // costruisce la tabella hash contenete i supporti degli itemsets...
    Hashtable support = buildHashtable(itemsets);
    Iterator iter = itemsets.iterator();
    int i=0;
    while(iter.hasNext()) {
      if ((i%500) == 0)
        System.out.println("Itemsets parsed: "+i+" of "+itemsets.size());
      i++;
      Itemset itemset = (Itemset) iter.next();
      Iterator iter2 = itemset.getItems();
      // vector contiene la lista degli items associati all'itemset
      Vector C = new Vector();
      while (iter2.hasNext()) {
        Item item = (Item) iter2.next();
        C.add(item.toString());
      }
      if (C.size() > 1) {
        // estrae i sottoinsiemi dell'itemset corrente...
        Vector subsets = extractSubsets(C);
        double support_C = ((Double) support.get(C)).doubleValue();
        for (int j=0; j< subsets.size(); j++) {
          Vector Q = (Vector) subsets.get(j);
          if (C.size() != Q.size()) {
            if (!support.contains(Q)) {
              String msg = "Itemset "+C.toString()+" do not contain subitemset "+C.toString()+"./n"+
                  "Every itemset must contain subitemsets in association model!";
            }
            double support_Q = ((Double) support.get(Q)).doubleValue();
            double confidence = support_C / support_Q;
            if (confidence >= min_conf) {
              // inserimento della regola Q --> C-Q con supporto = supporto(C)
              // e con confidenza = supporto(C) / supporto(Q)
              Vector C_Q = diff(C,Q);
              double support_C_Q = ((Double) support.get(C_Q)).doubleValue();
              AssociationRule rule = buildRule(Q, C_Q, support_C, confidence, support_Q, support_C_Q);
              ris.add(rule);

              // controlla il numero di regole prodotte
              if (ris.size() > max_rules) {
                ris.remove((AssociationRule) ris.last());
              }
            }
          }
        }
      }
    }
    support = null;
    iter = null;
    System.gc();
    return ris;
  }


  /**
   * Costruisce la tabella hash per leggere i supporti associati agli itemset.
   * Come chiave viene utilizzato un vettore contenente la lista degli item associati
   * all'itemset.
   */
  private static Hashtable buildHashtable(TreeSet itemset) throws AssociationModelException {
    Iterator iter = itemset.iterator();
    Hashtable ris = new Hashtable();
    while (iter.hasNext()) {
      Itemset i = (Itemset) iter.next();
      Iterator iter2 = i.getItems();
      Vector v = new Vector();
      while (iter2.hasNext()) {
        Item item = (Item) iter2.next();
        v.add(item.toString());
      }
      if (i.getSupport() == null) {
        String msg = "Itemset support is null; can't calculate RdA.";
        throw new AssociationModelException(msg, msg,
                                            kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
      }
      ris.put(v, i.getSupport());
    }
    return ris;
  }

  /**
   * Dato in ingresso in vettore contenente un insieme di stringhe, restituisce i sottoinsiemi
   * dell'insieme di stringhe.
   */
  private static Vector extractSubsets(Vector set) {
    Vector ris = new Vector();
    Vector tmp = new Vector(set);
    if (tmp.size() == 2) {
      // caso base...
      Vector v1 = new Vector();
      Vector v2 = new Vector();
      Vector v3 = new Vector();
      v1.add(tmp.get(0));
      v2.add(tmp.get(1));
      v3 = new Vector(tmp);
      ris.add(v1);
      ris.add(v2);
      ris.add(v3);
    }
    else {
      // caso induttivo...
      String first = (String) tmp.get(0);
      tmp.remove(0);
      Vector recurse = extractSubsets(tmp);
      ris = new Vector(recurse);
      Vector vect = new Vector();
      vect.add(first);
      ris.add(vect);
      for (int i=0; i<recurse.size(); i++) {
        Vector v = new Vector((Vector) recurse.get(i));
        v.add(0, first);
        ris.add(v);
      }
    }
    return ris;
  }

  /**
   * Calcola la differenza tra due vettori.
   */
  private static Vector diff(Vector v1, Vector v2) {
    Vector tmp = new Vector(v1);
    for (int j = 0; j < v2.size(); j++) {
      String s = (String) v2.get(j);
      if (tmp.contains(s))
        tmp.remove(s);
    }
    return tmp;
  }

  /**
   * Costruisce la regola di associazione.
   */
  private static AssociationRule buildRule(Vector antecedent, Vector consequent,
                                           double support, double confidence,
                                           double support_antecedent, double support_consequent) {

    Comparator comp = new CompareItems();
    TreeSet ant = new TreeSet(comp);
    // costruzione del body
    for (int i= 0; i < antecedent.size(); i++) {
      Item item = new Item((String) antecedent.get(i));
      ant.add(item);
    }
    // costruzione dell'head
    TreeSet cons = new TreeSet(comp);
    for (int i= 0; i < consequent.size(); i++) {
      Item item = new Item((String) consequent.get(i));
      cons.add(item);
    }
    Itemset body = new Itemset(ant);
    Itemset head = new Itemset(cons);
    body.setSupport(support_antecedent);
    head.setSupport(support_consequent);
    AssociationRule ris = new AssociationRule(body, head, support, confidence);
    return ris;
  }
}
