/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.TreeSet;
import java.util.*;
import java.lang.*;

/**
 * Factory used to build an association model given a set of frequent itemsets and
 * a specified minimum confidence.
 * Questa classe dato un insieme di itemsets frequenti ricava delle
 * regole di associazione che soddisfano la minima confidenza specificata dall'utente.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class ItemsetAssociationFactory extends AssociationProprietaryFactory {

  private int max_rules;

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param min_support double the minimum support.
   * @param min_confidence double the minimum confidence.
   * @param number_of_transactions int the number of transactions.
   * @param itemsets Itemset[] the list of frequent itemset. Ordering is mantaining by the class.
   * @throws KDDMLCoreException
   */
  public ItemsetAssociationFactory(java.io.File f,
                                   DataDictionary dd,
                                   MiningSchema ms,
                                   double min_support,
                                   double min_confidence,
                                   int number_of_transactions,
                                   Itemset[] itemsets)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);

    Comparator comp = new CompareItemSets();
    TreeSet set = new TreeSet(comp);
    for (int i=0; i<itemsets.length; i++)
      set.add(itemsets[i]);

    this.itemsets = set;
    max_rules = AssociationModel.MAX_NUMBER_OF_RULES;
  }

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param min_support double the minimum support.
   * @param min_confidence double the minimum confidence.
   * @param number_of_transactions int the number of transactions.
   * @param itemsets TreeSet the list of frequent itemset.
   * @throws KDDMLCoreException
   */
  public ItemsetAssociationFactory(java.io.File f,
                                   DataDictionary dd,
                                   MiningSchema ms,
                                   double min_support,
                                   double min_confidence,
                                   int number_of_transactions,
                                   TreeSet itemsets)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);

    this.itemsets = itemsets;
    max_rules = AssociationModel.MAX_NUMBER_OF_RULES;
  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    AssociationModel model = (AssociationModel) super.newInstance();

    model.MAX_NUMBER_OF_RULES = max_rules;

    // calcola gli items e le regole....
    if (!itemsets.isEmpty()) {
      items = Itemset2ItemUtility.compute(itemsets);
      association_rules = Itemset2RdAUtility.compute
          (itemsets, this.min_confidence.doubleValue(), max_rules);
    }
    else {
      items = new TreeSet();
      association_rules = new TreeSet(new CompareItemSets());
    }
    model.items = items;
    model.itemsets = itemsets;
    model.association_rules = association_rules;
    return model;
  }

  /**
   * Sets the maximum number of rules admitted in the model.
   * This parameter is optional. If omitted, all rules will be computed.
   *
   * @param max_rules int
   */
  public void setMaxNumberOfRules(int max_rules) {
    this.max_rules = max_rules;
  }

  public static void main(String args[]) {
  }
}
