/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.Iterator;

/**
 * A manager interface for itemsets.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface ItemsetManager {

  /**
   * Returns the support ot the itemset in the model as a percentage in respect to
   * the total number of transactions. It is a number between 0 and 1, inclusive.
   * Can be null.
   *
   * @return Double
   */
  public Double getSupport();

  /**
   * Returns the list of items as iterator.
   *
   * @return Iterator
   */
  public Iterator getItems();

  /**
   * Returns the number of items. Returns a value >=0.
   *
   * @return int
   */
  public int getSize();

  /**
   * Tests if two itemsets are equals, i.e., they contains the same items.
   *
   * @param itemset Object
   * @return boolean
   */
  public boolean equals(Object itemset);

  /**
   * Tests if an item is contained in the itemset.
   *
   * @param item Item
   * @return boolean
   */
  public boolean contain(ItemManager item);

  /**
   * Adds a new item to this itemset.
   *
   * @param item Item
   * @return boolean
   */
  public boolean addItem(ItemManager item);

  /**
   * Tests if this itemset satisfy an input transaction.
   * A transaction satisfy an itemset I1, ..., In if every item Ii with i in [1, m]
   * occours in the transaction.
   *
   * @param transaction TransactionManager
   * @return boolean
   */
  public boolean isSupported(TransactionManager transaction);

  /**
   * Sets the support of this itemsets.
   * Support must be a number between 0 and 1, inclusive.
   *
   * @param support double
   */
  public void setSupport(double support);

  /**
   * Returns the support for this itemset in terms of absolute count of transactions
   * supporting it.
   *
   * @param num_transactions int the total number of transactions used to build the itemset.
   * @return int
   */
  public int getAbsoluteSupport(int num_transactions);

  /**
   * Returns the list of items as transaction.
   *
   * @return TransactionManager
   */
  public TransactionManager toTransaction();

}
