/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import org.w3c.dom.*;
import java.io.*;
import java.util.*;

/**
 * Factory class that builds an association model from a PMML model.<br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class PMMLAssociationFactory extends AssociationExternalFactory {

  private String pmml_version;

  private Element header;

  private Element data_dictionary;

  private Element mining_schema;

  private Element association_model;

  private String min_support;

  private String min_confidence;

  private String number_of_transactions;

  private String algorithm_name;

  private String model_name;

  private String max_number_of_items;

  private String avg_number_of_items;

  private Hashtable items_ht;

  private Hashtable itemsets_ht;

  /**
   * Constructor given the pmml_path.
   *
   * @param f File the file in the system repository that will contain the generated model.
   * @param pmml_path String the URI identifying the pmml document.
   * @throws KDDMLCoreException if the source cann't be found or the parser returns
   * errors.
   */
  public PMMLAssociationFactory(java.io.File f, String pmml_path)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    // effettua il parsing del documento XML contenente il modello...
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(pmml_path),
                              new DTDResolver(kddml.Core.KDDMLObjectType.PMML_ASSOCIATION_MODEL));
    }
    catch(java.io.IOException e) {
      String msg = "Invalid PMML reference: "+pmml_path;
      throw new AssociationModelException(msg, e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }

    Element root = (Element) doc.getDocumentElement();

    if (!root.getTagName().equalsIgnoreCase("pmml"))
      throw new AssociationModelException(pmml_path+ " no contains a PMML document.",
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the pmml version
    pmml_version = root.getAttribute("version");

    // reading the copyright
    header = (Element) root.getElementsByTagName("Header").item(0);

     // reading the data dictionary element
    data_dictionary = (Element) root.getElementsByTagName("DataDictionary").item(0);

    association_model = (Element) doc.getElementsByTagName("AssociationModel").item(0);
    if (association_model == null)
      throw new AssociationModelException(pmml_path+ " no contains a PMML association model document.",
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the mining schema element...
    mining_schema = (Element) association_model.getElementsByTagName("MiningSchema").item(0);

    // reading attributes...
    readAttributes();
  }

  /**
   * Returns the data dictionary of the output model.
   *
   * @throws MiningModelException
   * @return DataDictionary
   */
  protected DataDictionary buildDataDictionary() throws MiningModelException {
    if (data_dictionary == null)
      return new DataDictionary();
    else
      return PMMLUtility.buildDataDictionary(data_dictionary);
  }

  /**
   * Returns the mining schema of the output model.
   *
   * @throws MiningModelException
   * @return MiningSchema
   */
  protected MiningSchema buildMiningSchema() throws MiningModelException {
    if (mining_schema == null)
      return new MiningSchema();
    else
      return PMMLUtility.buildMiningSchema(mining_schema);
  }

  /**
   * Returns the name of the algorithm.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildAlgorithmName() throws MiningModelException {
    return algorithm_name;
  }

  /**
   * Returns the name of the model.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildModelName() throws MiningModelException {
    return model_name;
  }

  /**
   * Returns the PMML version supported.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildPMMLVersion() throws MiningModelException {
    return pmml_version;
  }

  /**
   * Returns the copyright.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildCopyright() throws MiningModelException {
    if (header != null)
      return XMLUtility.getAttribute(header, "copyright");
    else
      return null;
  }

  /**
   * Returns the application name.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationName() throws MiningModelException {

    if (header != null) {
      Element el = (Element) header.getElementsByTagName("Application").item(0);
      if (el != null)
        return XMLUtility.getAttribute(el, "name");
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Returns the application version.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationVersion() throws MiningModelException {

    if (header != null) {
      Element el = (Element)header.getElementsByTagName("Application").item(0);
      if (el != null)
        return el.getAttribute("version");
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Returns the minimum support of the association rules.
   * Mininmum support is a real in (0,1]
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected Double buildMinSupport() throws AssociationModelException {
    Double ris;
    try {
      ris = new Double(min_support);
    }
    catch(Exception e) {
      String msg = "Minimum support must be a real between 0 and 1.";
      throw new AssociationModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the minimum confidence of the association rules.
   * Mininmum confidence is a real in (0,1].
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected Double buildMinConfidence() throws AssociationModelException {
    Double ris;
    try {
      ris = new Double(min_confidence);
    }
    catch(Exception e) {
      String msg = "Minimum confidence must be a real between 0 and 1.";
      throw new AssociationModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the number of transaction used to build the association rules.
   * Returns a positive integer.
   *
   * @throws AssociationModelException
   * @return Integer
   */
  protected Integer buildNumberOfTransactions() throws AssociationModelException {
    Integer ris;
    try {
      ris = new Integer(number_of_transactions);
    }
    catch(Exception e) {
      String msg = "Number of transactions must be a positive integer.";
      throw new AssociationModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the maximum number of iterms per transaction.
   * Returns a positive integer.
   *
   * @throws AssociationModelException
   * @return Integer
   */
  protected Integer buildMaxNumberOfItemsPerTA() throws AssociationModelException {
    if (max_number_of_items.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(max_number_of_items);
      }
      catch (Exception e) {
        String msg =
            "The max number of items per transction must be a positive integer.";
        throw new AssociationModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Returns the average number of items per transaction.
   * Returns a positive double.
   *
   * @throws AssociationModelException
   * @return Double
   */
  protected Double buildAvgNumberOfItemsPerTA() throws AssociationModelException {
    if (avg_number_of_items.equals(""))
      return null;
    else {
      Double ris;
      try {
        ris = new Double(avg_number_of_items);
      }
      catch (Exception e) {
        String msg =
            "The average number of items per transaction must be a positive real.";
        throw new AssociationModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Returns the list of association rules.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of AssociationRules
   */
  protected TreeSet buildAssociationRules() throws AssociationModelException {
    try {
      CompareAssociationRules comp = new CompareAssociationRules();
      TreeSet ris = new TreeSet(comp);
      NodeList n = association_model.getElementsByTagName("AssociationRule");
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);
        double support = new Double(el.getAttribute("support")).doubleValue();
        double confidence = new Double(el.getAttribute("confidence")).
            doubleValue();
        String antecedent_id = XMLUtility.getAttribute(el, "antecedent");
        String consequent_id = XMLUtility.getAttribute(el, "consequent");
        if (!itemsets_ht.containsKey(antecedent_id))
          throw new AssociationModelException(
              "Illegal PMML Association Model: document do not contain ref <" +
              antecedent_id + "> for itemsets"
              , null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
        if (!itemsets_ht.containsKey(consequent_id))
          throw new AssociationModelException(
              "Illegal PMML Association Model: document do not contain ref <" +
              consequent_id + "> for itemsets",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

        AssociationRule ar = new AssociationRule((Itemset)
            itemsets_ht.get(antecedent_id),
            (Itemset) itemsets_ht.get(consequent_id),
            support, confidence);
        if (!contains(ris, ar))
          ris.add(ar);
      }
      itemsets_ht = null;
      System.gc();
      return ris;
    }
    catch(Exception e) {
      throw new AssociationModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Returns the list of itemsets.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of Itemsets
   */
  protected TreeSet buildItemsets() throws AssociationModelException {
    try {
      CompareItemSets comp = new CompareItemSets();
      TreeSet ris = new TreeSet(comp);
      NodeList n = association_model.getElementsByTagName("Itemset");
      itemsets_ht = new Hashtable();
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);
        String itemset_id = el.getAttribute("id");
        Double support = null;
        if (el.getAttribute("support") != "")
          support = new Double(el.getAttribute("support"));
        // scorre i riferimenti per creare gli itemsets...
        NodeList n2 = el.getElementsByTagName("ItemRef");
        Itemset itemset = new Itemset();
        for (int j = 0; j < n2.getLength(); j++) {
          Element el2 = (Element) n2.item(j);
          String ref = el2.getAttribute("itemRef");
          if (!items_ht.containsKey(ref))
            throw new AssociationModelException(
                "Illegal PMML Association Model: document do not contain ref <" +
                ref + "> for itemset <" + itemset_id + ">.",
                null,
                KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

          Item item = (Item) items_ht.get(ref);
          itemset.addItem(item);
        }
        if (itemsets_ht.containsKey(itemset))
          throw new AssociationModelException(
              "Illegal PMML Association Model: document already contains itemset <" +
              itemset.toString() + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
        if (support != null)
          itemset.setSupport(support.doubleValue());
        itemsets_ht.put(itemset_id, itemset);
        if (!contains(ris, itemset))
          ris.add(itemset);
      }
      items_ht = null;
      System.gc();
      return ris;
    }
    catch(Exception e) {
      throw new AssociationModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Returns the list of items.
   *
   * @throws AssociationModelException
   * @return TreeSet an ordered set of Item
   */
  protected TreeSet buildItems() throws AssociationModelException {
    try {
      CompareItems comp = new CompareItems();
      TreeSet ris = new TreeSet(comp);
      NodeList n = association_model.getElementsByTagName("Item");
      // tabella hash contenente l'association chiave --> item
      items_ht = new Hashtable();
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);
        Item item = new Item(XMLUtility.getAttribute(el, "value"));
        item.mapped_value = XMLUtility.getAttribute(el, "mappedValue");
        if (!el.getAttribute("weight").equals(""))
          item.weight = new Double(el.getAttribute("weight"));

        if (items_ht.containsKey(item))
          throw new AssociationModelException(
              "Illegal PMML Association Model: document already contains item <" +
              item.value + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
        items_ht.put(el.getAttribute("id"), item);
        if (!contains(ris, item))
          ris.add(item);
      }
      return ris;
    }
    catch(Exception e) {
      throw new AssociationModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  private static boolean contains(TreeSet set, Item item) {
    Iterator iter = set.iterator();
    while (iter.hasNext()) {
      Item i = (Item) iter.next();
      if (i.equals((Item) item))
        return true;
    }
    return false;
  }

  private boolean contains(TreeSet itemsets, Itemset itemset) {
    Iterator iter = itemsets.iterator();
    while (iter.hasNext()) {
      Itemset i = (Itemset) iter.next();
      if (i.equals((Itemset) itemset))
        return true;
    }
    return false;
  }

  private boolean contains(TreeSet association_rules, AssociationRule rule) {
    Iterator iter = association_rules.iterator();
    while (iter.hasNext()) {
      AssociationRule i = (AssociationRule) iter.next();
      if (i.equals((AssociationRule) rule))
        return true;
    }
    return false;
  }

  private void readAttributes() {
    min_support = association_model.getAttribute("minimumSupport");
    min_confidence = association_model.getAttribute("minimumConfidence");
    number_of_transactions = association_model.getAttribute("numberOfTransactions");
    model_name = XMLUtility.getAttribute(association_model, "modelName");
    algorithm_name = XMLUtility.getAttribute(association_model, "algorithmName");
    model_name = association_model.getAttribute("modelName");
    algorithm_name = association_model.getAttribute("algorithmName");
    max_number_of_items = association_model.getAttribute("maxNumberOfItemsPerTA");
    avg_number_of_items = association_model.getAttribute("avgNumberOfItemsPerTA");
  }
}
