/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.TreeSet;
import java.util.*;
import java.lang.*;

/**
 * Factory class that builds an association model given a set of association rules.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class RulesAssociationFactory extends AssociationProprietaryFactory {

  private int max_itemsets;

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param min_support double the minimum support.
   * @param min_confidence double the minimum confidence.
   * @param number_of_transactions int the number of transactions.
   * @param association_rules AssociationRule[] the list of association rules. Ordering is
   * mantained by the class.
   * @throws KDDMLCoreException
   */
  public RulesAssociationFactory(java.io.File f,
                                 DataDictionary dd,
                                 MiningSchema ms,
                                 double min_support,
                                 double min_confidence,
                                 int number_of_transactions,
                                 AssociationRule[] association_rules)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);

    Comparator comp = new CompareAssociationRules();
    TreeSet set = new TreeSet(comp);
    for (int i=0; i<association_rules.length; i++)
      set.add(association_rules[i]);
    this.association_rules = set;

    max_itemsets = AssociationModel.MAX_NUMBER_OF_ITEMSETS;
  }

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param min_support double the minimum support.
   * @param min_confidence double the minimum confidence.
   * @param number_of_transactions int the number of transactions.
   * @param association_rules TreeSet the list of association rules.
   * @throws KDDMLCoreException
   */
  public RulesAssociationFactory(java.io.File f,
                                 DataDictionary dd,
                                 MiningSchema ms,
                                 double min_support,
                                 double min_confidence,
                                 int number_of_transactions,
                                 TreeSet association_rules)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);

    this.association_rules = association_rules;

    max_itemsets = AssociationModel.MAX_NUMBER_OF_ITEMSETS;
  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    AssociationModel model = (AssociationModel) super.newInstance();

    model.MAX_NUMBER_OF_ITEMSETS = max_itemsets;

    // calcola gli items e gli itemsets....
    this.itemsets = RdA2ItemsetUtility.compute(association_rules, max_itemsets);
    this.items = Itemset2ItemUtility.compute(itemsets);

    model.items = items;
    model.itemsets = itemsets;
    model.association_rules = association_rules;
    return model;
  }

  /**
   * Sets the maximum number of itemsets admitted in the model.
   * This parameter is optional. If omitted, all itemsets will be computed.
   *
   * @param max_itemsets int
   */
  public void setMaxNumberOfItemsets(int max_itemsets) {
    this.max_itemsets = max_itemsets;
  }

  public static void main(String args[]) {
  }
}
