/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.AssociationRules;

import java.util.HashSet;
import java.util.Iterator;
import weka.core.*;

/**
 * This class represents a single data transaction (i.e. a set of items).
 * Transactions can be boolean (if no value is specified) or multivalue.
 * E.g.: {"milk", "apple", "bread"} vs {"outlook=sunny", "temperature=12"}.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class Transaction implements TransactionManager {

  /**
   * List of items belonging to the transaction.
   */
  protected HashSet items;

  // the number of missing value of instances (not included in items)
  //protected int num_missing;

  /**
   * Constructor used to build a multivalue transactions from the input instance.
   * Each single instance field is an item. The name of the attribute corresponds
   * to the name of the item. The value of the attribute corresponds to the value
   * of item. If the value is missing, no item is included in the transaction.
   * No other items features are allowed.
   *
   * @param instances Instances
   */
  public Transaction(Instance instance) {
    items = new HashSet();
    for (int i = 0; i < instance.numAttributes(); i++) {
      if (!instance.isMissing(i)) {
        String name = instance.attribute(i).name();
        String value = instance.toString(i);
        Item item = new Item(name + "=" + value);
        items.add(item);
      }
    }
  }

  /**
   * Constructor given the list of items.
   * The instances contains an attribute transaction (the identifier of the transaction)
   * and an attribute item (the item value).
   * The identifier of transaction is equal for each instance.
   *
   * @param instances Instances
   */
  public Transaction(Item[] list_of_items) {
    items = new HashSet();
    for (int i =0; i< list_of_items.length; i++) {
      Item item = list_of_items[i];
      items.add(item);
    }
  }

  /**
   * Returns an enumeration of items belonging to the transaction.
   *
   * @return Iterator
   */
  public java.util.Iterator getItems() {
    return items.iterator();
  }

  /**
   * Test if the transaction contain an item.
   *
   * @param item String
   * @return boolean
   */
  public boolean contains(Item item) {
    return items.contains(item);
  }

  /**
   * Returns the number of items.
   *
   * @return int
   */
  public int getNumberOfItems() {
    return items.size();
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    Iterator iter = items.iterator();
    String s= "";
    while(iter.hasNext()) {
      s = s + ((Item) iter.next()).toString() + ",";
    }
    if (s.length() > 0)
      return s.substring(0, s.length()-1);
    else
      return "";
  }
}
