/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

import java.text.*;
import org.w3c.dom.*;
import kddml.Core.XMLUtility;


/**
 * A CategoryMatrix represents a matrix whose axes are categories.
 * The CategoryMatrix can be used to represent confusion matrix in classification trees
 * or similarity matrix in clustering.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CategoryMatrix implements CategoryMatrixManager {

  protected String[] categories;

  protected double[][] matrix;

  /**
   * Build an empty matrix with default values 0.
   *
   * @param categories String[]
   */
  public CategoryMatrix(String[] categories) {

    this.categories = categories;
    int l = categories.length;
    matrix = new double[l][l];

    for (int i=0; i<l; i++)
      for (int j=0; j<l; j++)
        matrix[i][j] = 0;
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param category_matrix Element
   */
  protected CategoryMatrix(Element category_matrix) {

    Element array = (Element) XMLUtility.getFirstChild(category_matrix);
    int n = new Integer(array.getAttribute("n")).intValue();
    String values = array.getChildNodes().item(0).getNodeValue();
    categories = new String[n];
    matrix = new double[n][n];
    java.util.StringTokenizer tokeniz = new java.util.StringTokenizer(values);
    int cnt=0;
    while(tokeniz.hasMoreElements()) {
      String s = tokeniz.nextToken();
      categories[cnt] = s.replaceAll("&lt;", "<");
      cnt++;
    }

    Element matrix = (Element) XMLUtility.getLastChild(category_matrix);
    java.util.Vector v = XMLUtility.getChildren(matrix);
    for (int i=0; i<v.size(); i++) {
      array = (Element) v.get(i);
      String matrix_values = array.getChildNodes().item(0).getNodeValue();
      tokeniz = new java.util.StringTokenizer(matrix_values);
      int j=0;
      while(tokeniz.hasMoreElements()) {
        int d = new Double(tokeniz.nextToken()).intValue();
        this.matrix[i][j] = d;
        j++;
      }
    }
  }


  /**
   * Returns a collection of category values used to represent the axes of the matrix.
   *
   * @return String[]
   */
  public String[] getCategories() {
    return categories;
  }

  /**
   * Returns the value of the matrix for a given row and column.
   *
   * @param row_name String the row name of the matrix.
   * @param column_name String the column name of the matrix.
   * @throws MiningModelException if the row_name or the column_name has not been found.
   * @return double
   */
  public double getValue(String row_name, String column_name)
      throws MiningModelException {

    int index1 = getIndex(row_name);
    int index2 = getIndex(column_name);

    if (index1 == -1) {
      String msg = "Category < "+row_name+" > do not exist in category matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (index2 == -1) {
      String msg = "Category < "+column_name+" > do not exist in category matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    return matrix[index1][index2];
  }

  /**
   * Sets the value for a given row and a given column.
   *
   * @param row_name String the row name of the matrix.
   * @param column_name String the column name of the matrix.
   * @param value double
   * @throws MiningModelException if the row_name or the column_name has not been found.
   */
  public void setValue(String row_name, String column_name, double value)
      throws MiningModelException {

    int index1 = getIndex(row_name);
    int index2 = getIndex(column_name);

    if (index1 == -1) {
      String msg = "Category < "+row_name+" > do not exist in category matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (index2 == -1) {
      String msg = "Category < "+column_name+" > do not exist in category matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    matrix[index1][index2] = value;
  }

  /**
   * Returns the index of the categoric. Returns -1 if not found.
   *
   * @param category String
   * @return int
   */
  int getIndex(String category) {
    for (int i=0; i<categories.length; i++) {
      if (categories[i].equalsIgnoreCase(category))
        return i;
    }
    return -1;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    //display precision, number of decimal places
    int d = 5;
    //display precision, total width of floating value
    int w = 5;

    String ris = "";
    // Some definitions for handling output formating
    NumberFormat myFormat = NumberFormat.getNumberInstance();
    FieldPosition fp = new FieldPosition(NumberFormat.INTEGER_FIELD);
    myFormat.setMaximumIntegerDigits(d);
    myFormat.setMaximumFractionDigits(d);
    myFormat.setMinimumFractionDigits(d);
    for (int i=0; i<categories.length; i++)
    {
      // Print each row, elements separated by spaces
      for (int j=0; j<categories.length; j++)
      // Following unfortunately doesn't format at all
      //                  System.out.print(m[i][j] + "  ");
      {
        String valString = myFormat.format(matrix[i][j], new StringBuffer(), fp).toString();
        valString = getSpaces(w - fp.getEndIndex()) + valString;
        ris = ris + valString;
      }
      // Start a new line at the end of a row
      ris = ris + "\n";
    }
    // Leave a gap after the entire matrix
    ris = ris + "\n";
    return ris;
  }

  private static String getSpaces(int n) {

    StringBuffer sb = new StringBuffer(n);
    for (int i = 0; i < n; i++) sb.append(' ');
    return sb.toString();
  }
}
