/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

/**
 * This enumeration specifies the comparison operator that associates two or more
 * Predicate. The association is boolean: and, or, xor, or surrogate. <br>
 * The operators and, or, and xor are associative binary operators, having their usual
 * semantics. The order of evaluation is irrelevant for all the predicates within
 * one compound predicate.<br>
 * The operator 'surrogate' allows for specifing surrogate predicates.
 * They are used for cases where a missing value appears in the evaluation of the
 * parent predicate such that an alternative predicate is available.
 * An expression 'surrogate(a,b)' is equivalent to 'if not unknown(a) then a else b'.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class BooleanOperator implements java.io.Serializable {

  /**
   * And boolean operator.<br>
   * <br>
   * true AND true --> true.<br>
   * true AND false --> false.<br>
   * false AND false --> false.<br>
   * unknown AND true --> unknown.<br>
   * unknown AND false --> false.<br>
   * unknown AND unknown --> unknown.<br>
   */
  public final static BooleanOperator AND = new BooleanOperator();

  /**
   * Or boolean operator.<br>
   * <br>
   * true OR true --> true.<br>
   * true OR false --> true.<br>
   * false OR false --> false.<br>
   * unknown OR true --> true.<br>
   * unknown OR false --> unknown.<br>
   * unknown OR unknown --> unknown.<br>
   */
  public final static BooleanOperator OR = new BooleanOperator();

  /**
   * The operator 'xor' is only used with two arguments only, it indicates an evaluation to
   * TRUE if the predicates evaluate to different values (one TRUE and one FALSE).<br>
   * <br>
   * true XOR true --> true.<br>
   * true XOR false --> false.<br>
   * false XOR false --> true.<br>
   * true XOR unknown --> unknown.<br>
   * false XOR unknown --> unknown.<br>
   * unknown XOR unknown --> unknown.<br>
   */
  public final static BooleanOperator XOR = new BooleanOperator();

  /**
   * The surrogate operator.
   * The operator surrogate provides a special means to handle logical expressions with
   * missing values. It is applied to a sequence of predicates. The order of the predicates
   * matters, the first predicate is the primary, the next predicates are the surrogates.
   * Evaluation order is left-to-right.
   * The cascaded predicates are applied when the primary predicate evaluates to 'Unknown'.
   * Therefore, the surrogate predicates can provide a resolution to an undetermined predicate.
   */
  public final static BooleanOperator SURROGATE = new BooleanOperator();
}
