/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import kddml.Core.DataMining.*;
import kddml.Core.*;

/**
 * A factory class that creates a classification model from an external resource (e.g. text file,
 * PMML model).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class ClassificationExternalFactory extends ModelExternalFactory {

  /**
   * Constructor given the physical file containing the object.
   *
   * @param f File
   * @throws KDDMLCoreException
   */
  public ClassificationExternalFactory(java.io.File f)
      throws KDDMLCoreException {

    super(f);
    object = new TreeModel(f);
  }

  /**
   * Returns a new instance of this object.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject newInstance() throws KDDMLCoreException {

    super.newInstance();

    Node n = buildTree();
    ConfusionMatrix m1 = buildTrainingConfusionMatrix();
    ConfusionMatrix m2 = buildTestConfusionMatrix();

    ((TreeModel) object).root = n;
    if (m1 != null)
      ((TreeModel) object).setTrainingConfusionMatrix(m1);

    if (m2 != null)
      ((TreeModel) object).setTestConfusionMatrix(m2);

    ((TreeModel) object).binary_split = isBinarySplitting();
    return object;
  }

  /**
   * Returns the node root of the tree. Cannot be null.
   *
   * @throws TreeModelException
   * @return Node
   */
  protected abstract Node buildTree() throws TreeModelException;

  /**
   * Returns the confusion training matrix of the model.
   * Can be null if no confusion training matrix is provided with the model.
   *
   * @throws TreeModelException
   * @return ConfusionMatrix
   */
  protected abstract ConfusionMatrix buildTrainingConfusionMatrix()
      throws TreeModelException;

  /**
   * Returns the test confusion matrix of the model. Can be null.
   * Can be null if no text confusion matrix is provided with the model.
   *
   * @throws TreeModelException
   * @return ConfusionMatrix
   */
  protected abstract ConfusionMatrix buildTestConfusionMatrix()
      throws TreeModelException;

  /**
   * Returns true if the model uses the binary splitting (every node with exactly 2
   * children). Returns false otherwise.
   *
   * @return boolean
   */
  protected abstract boolean isBinarySplitting();
}
