/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.io.File;

/**
 * A factory class that creates a generic model from a proprietary (i.e., not external)
 * resource.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ClassificationProprietaryFactory extends ModelProprietaryFactory {

  protected Node root;

  protected ConfusionMatrix training_matrix;

  protected ConfusionMatrix test_matrix;

  protected boolean binary_split;

  /**
   * Constructor.
   *
   * @param f File the file containing the model.
   * @param dd DataDictionary the data dictionary.
   * @param ms MiningSchema the mining schema.
   * @param root Node the node root of the tree. Cannot be null.
   * @param training_matrix ConfusionMatrix the training confusion matrix. Can be null.
   * @param test_matrix ConfusionMatrix the test confusion matrix. Can be null.
   * @throws KDDMLCoreException
   */
  public ClassificationProprietaryFactory(java.io.File f,
                                          DataDictionary dd,
                                          MiningSchema ms,
                                          Node root,
                                          ConfusionMatrix training_matrix,
                                          ConfusionMatrix test_matrix)
      throws KDDMLCoreException {

    super(f, dd, ms);
    this.root = root;
    this.training_matrix = training_matrix;
    this.test_matrix = test_matrix;
    binary_split = false;
  }

  /**
   * Returns a new instance of this object.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject newInstance() throws KDDMLCoreException {

    TreeModel model = new TreeModel(physical_file);

    model = (TreeModel) setMiningSchema((MiningModel) model);
    model = (TreeModel) setDataDictionary((MiningModel) model);

    if (algorithm_name != null)
      model = (TreeModel) setAlgorithmName((MiningModel) model);
    if (model_name != null)
      model = (TreeModel) setModelName((MiningModel) model);

    // set the tree root
    model.root = this.root;
    if (training_matrix != null)
      model.training_matrix = training_matrix;
    if (test_matrix != null)
      model.test_matrix = test_matrix;
    model.binary_split = this.binary_split;
    return model;
  }

  /**
   * Sets the splitting value of the tree. If binary_split is true, then the tree
   * has exactly two children for each node. By default, binary_split is false.
   *
   * @param binary_split boolean
   */
  public void setBinaryNodeSplitting(boolean binary_split) {
    this.binary_split = binary_split;
  }

  /**
   * Only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    try {

      String[] categories = {"yes", "no"};

      // data dictionary
      DataDictionary dd = new DataDictionary();
      dd.addField(new DataField("humidity", DataField.CONTINUOUS));
      DataField df2 = new DataField("play", DataField.CATEGORICAL);
      df2.addEnumeratedValue("yes");
      df2.addEnumeratedValue("no");
      dd.addField(df2);

      // mining schema
      MiningSchema ms = new MiningSchema();
      ms.addField(new MiningField("humidity"));
      ms.addField(new MiningField("play", MiningField.PREDICTED));

      // root node
      double[] d1 = {8.0, 12.0};
      ScoreDistribution d = new ScoreDistribution(categories, d1);
      SimpleNode root = new SimpleNode("", d);

      // left child
      double[] d2 = {1.0, 8.0};
      ScoreDistribution d_l = new ScoreDistribution(categories, d2);
      SimpleComparisonPredicate p_l =
          new SimpleComparisonPredicate("humidity", 80, ComparisonOperator.LESS_OR_EQUAL);
      SimpleNode left = new SimpleNode("yes", p_l , d_l, root);

      // right child
      double[] d3 = {7.0, 4.0};
      ScoreDistribution d_r = new ScoreDistribution(categories, d3);
      SimpleComparisonPredicate p_r =
          new SimpleComparisonPredicate("humidity", 80, ComparisonOperator.GREATER_THAN);
      SimpleNode right = new SimpleNode("no", p_r , d_r, root);

      // adds the left child and the right child to the root
      root.addChild(left);
      root.addChild(right);

      // confusion matrix on training set
      ConfusionMatrix matrix = new ConfusionMatrix(categories, true);
      matrix.setValue("yes", "yes", 6.0);
      matrix.setValue("yes", "no", 2.0);
      matrix.setValue("no", "yes", 3.0);
      matrix.setValue("no", "no", 9.0);

      // creates the model
      File dest = new File("C:/tree.xml");
      ClassificationProprietaryFactory factory =
          new ClassificationProprietaryFactory(dest, dd, ms, root, matrix, null);

      // sets the binary spit node to true
      factory.setBinaryNodeSplitting(true);
      // sets the model name
      factory.setModelName("my_tree");
      // sets the overwriting to true
      factory.setOverwrite(true);

      // create a new instance, print out the tree and save it in the specified destination
      TreeModel my_tree = (TreeModel) factory.newInstance();
      System.out.println(my_tree.toString());
      my_tree.saveToRepository();

    }
    catch(kddml.Core.KDDMLCoreException e2) {
      System.out.println("Invalid input document: "+e2.getErrorMessage());
    }
  }
}
