/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import java.util.Vector;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * A CompoundBaggingsTree implements a combination between two or more classification trees
 * with a voting operator.<br>
 * More precisely, given n distinct classifiers C1, . . . ,Cn, a voting classifier assigns
 * to a tuple the class mostly assigned by C1, . . . ,Cn.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CompoundBaggingsTree extends CompoundTree {

  /**
   * Empty constructor.
   */
  public CompoundBaggingsTree() {
    super();
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param el Element
   * @throws TreeModelException
   */
  protected CompoundBaggingsTree(Element el)
      throws TreeModelException {

    super(el);
    java.util.Vector list = XMLUtility.getChildren(el);
    Element child;

    // scorre i figli del nodo e crea ricorsivamente la classe
    for (int i=0; i<list.size(); i++) {
      child = (Element) list.get(i);
      if (child.getTagName().equals("Node")) {
        if (! ( (Element) XMLUtility.getFirstChild(child)).getTagName().equals("Extension")) {
          // simple node
          trees.add(new SimpleNode(child));
        }
        else {
          // extension
          Element tmp = (Element) kddml.Core.XMLUtility.getFirstChild(child);
          Element ext = (Element) kddml.Core.XMLUtility.getFirstChild(tmp);
          String combination = ext.getAttribute("combination_type");
          if ( (combination.equalsIgnoreCase("and")) ||
              (combination.equalsIgnoreCase("or"))) {
            trees.add(new CompoundBooleanTree(child));
          }
          else
            trees.add(new CompoundBaggingsTree(child));
        }
      }
    }
  }

  /**
   * Method used to classify an input instance.
   * Given n distinct classifiers C1, . . . ,Cn, a voting classifier assigns
   * to a tuple the class mostly assigned by C1, . . . ,Cn.
   * If two or more classification values have the same score, then the method returns
   * the first one.
   *
   * @param values String[] classification values returned by single trees.
   * @return String the classification value.
   */
  public String classify(String[] values) {

    // -----------------------------------------------
    // TO DO:
    // Implementare una valutazione probabilistica delle classificazioni parimerito....
    // -----------------------------------------------

    // vettore di stringhe contenente i valori di classificazione...
    Vector v1 = new Vector();
    // vettore di interi contenente le votazioni...
    Vector v2 = new Vector();
    for (int i=0; i<trees.size(); i++) {
      if (!v1.contains(values[i])) {
        v1.add(values[i]);
        v2.add(new Integer(1));
      }
      else {
        int index = v1.indexOf(values[i]);
        v2.setElementAt(new Integer(((Integer) v2.get(index)).intValue() + 1), index);
      }
    }
    String ris = (String) v1.get(maxValue(v2));
    return ris;
  }

  /**
   * Metodo che restituisce l'indice del valore di classificazione scelto da un comitato.
   *
   * @param vettore Vector
   * @return int
   */
  private int maxValue(Vector vettore) {
    int max=-1;
    int posizione = -1;
    for(int i=0; i<vettore.size();i++) {
      if (((Integer) vettore.get(i)).intValue() > max) {
        max = ((Integer) vettore.get(i)).intValue();
        posizione = i;
      }
      else if (((Integer) vettore.get(i)).intValue()==max)
        posizione = -1;
    }
    return posizione;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "COMMITTEE between "+super.trees.size()+" trees:\n";
    for (int i=0; i<trees.size(); i++) {
      ris = ris + "Tree "+i+":\n";
      Node n = (Node) trees.get(i);
      ris = ris + n.toString() + "\n";
    }
    return ris;
  }

  /**
   * Returns a representation of this object as XML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("Node");
    Element ext = XMLUtility.doc.createElement("Extension");
    Element extension = XMLUtility.doc.createElement("X-VotingTree");
    ext.appendChild(extension);
    extension.setAttribute("combination_type", "committee");
    extension.setAttribute("number_of_trees", new Integer(trees.size()).toString());

    ris.appendChild(ext);
    // predicate is required
    Element predicate = XMLUtility.doc.createElement("True");
    ris.appendChild(predicate);
    for (int i=0; i<trees.size(); i++) {
      Node n = (Node) trees.get(i);
      ris.appendChild(n.toXML());
    }
    ris.setAttribute("score", "");
    return ris;
  }
}
