/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import java.util.Vector;
import weka.core.Instance;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * A CompundPredicate is a set of predicates connected by logical or relational operators.
 * For example, "age < 30 AND salary > 50K" is a compound predicate which is connected by
 * a boolean and operator.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CompoundBooleanPredicate extends CompoundPredicate {

  /**
   * Vector of Predicate.
   */
  protected Vector predicates;

  /**
   * The boolean operator.
   */
  protected BooleanOperator operator;

  /**
   * Constructor given a boolean operator and an array of predicates.
   *
   * @param predicates Predicate[]
   * @param operator BooleanOperator
   */
  public CompoundBooleanPredicate(Predicate[] predicates, BooleanOperator operator) {

    this.predicates = new Vector();
    for (int i=0; i<predicates.length; i++) {
      this.predicates.add(predicates[i]);
    }
    this.operator = operator;
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param elem Element
   * @throws TreeModelException if the XML element is not valid.
   */
  public CompoundBooleanPredicate(Element elem) throws TreeModelException {
    predicates = new Vector();
    if (!elem.getTagName().equalsIgnoreCase("CompoundPredicate")) {
      String msg = "Invalid XML element. No CompoundPredicate found!";
      throw new TreeModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    String op = elem.getAttribute("booleanOperator");
    if (op.equals("and"))
      operator = BooleanOperator.AND;
    if (op.equals("or"))
      operator = BooleanOperator.OR;
    if (op.equals("xor"))
      operator = BooleanOperator.XOR;
    if (op.equals("surrogate"))
      operator = BooleanOperator.SURROGATE;
    Vector childes = XMLUtility.getChildren(elem);
    for (int i=0; i< childes.size(); i++) {
      Element leaf = (Element) childes.get(i);
      if (leaf.getTagName().equals("SimplePredicate")) {
        SimpleComparisonPredicate p = new SimpleComparisonPredicate(leaf);
        predicates.add(p);
      }
      else
      if (leaf.getTagName().equals("CompoundPredicate")) {
        CompoundBooleanPredicate p = new CompoundBooleanPredicate(leaf);
        predicates.add(p);
      }
      else
      if (leaf.getTagName().equals("SimpleSetPredicate")) {
        SimpleSetPredicate p = new SimpleSetPredicate(leaf);
        predicates.add(p);
      }
      else
      if ((leaf.getTagName().equals("True")) || (leaf.getTagName().equals("False"))) {
        SimpleBooleanPredicate p = new SimpleBooleanPredicate(leaf);
        predicates.add(p);
      }
    }
  }

  /**
   * Evaluates this predicate on a single input instance.
   *
   * @param instance Object the input instance.
   * @throws TreeModelException
   * @return PredicateEvaluation the evaluation result of this predicate.
   */
  public PredicateEvaluation evaluate(Object instance)
      throws TreeModelException {
    Instance tupla = (Instance) instance;
    if (operator.equals(BooleanOperator.AND))
      return evaluateAND(tupla);
    if (operator.equals(BooleanOperator.OR))
      return evaluateOR(tupla);
    if (operator.equals(BooleanOperator.XOR))
      return evaluateXOR(tupla);
    return evaluateSURROGATE(tupla);
  }

  /**
   * Evaluate AND operator.
   * <br>
   * true AND true --> true.<br>
   * true AND false --> false.<br>
   * false AND false --> false.<br>
   * unknown AND true --> unknown.<br>
   * unknown AND false --> false.<br>
   * unknown AND unknown --> unknown.<br>
   *
   * @param instance Instance
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  protected PredicateEvaluation evaluateAND(Instance instance)
      throws TreeModelException {
    for (int i=0 ; i<getSize(); i++) {
      Predicate p = (Predicate) predicates.get(i);
      PredicateEvaluation pe = evaluate(p);
      if ((pe.equals(PredicateEvaluation.FALSE)) ||
          (pe.equals(PredicateEvaluation.UNKNOWN)))
        return pe;
    }
    return PredicateEvaluation.TRUE;
  }

  /**
   * Evaluate OR operator.
   * <br>
   * true OR true --> true.<br>
   * true OR false --> true.<br>
   * false OR false --> false.<br>
   * unknown OR true --> true.<br>
   * unknown OR false --> unknown.<br>
   * unknown OR unknown --> unknown.<br>
   *
   * @param instance Instance
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  protected PredicateEvaluation evaluateOR(Instance instance)
      throws TreeModelException {
    for (int i=0 ; i<getSize(); i++) {
      Predicate p = (Predicate) predicates.get(i);
      PredicateEvaluation pe = evaluate(p);
      if ((pe.equals(PredicateEvaluation.TRUE)) ||
          (pe.equals(PredicateEvaluation.UNKNOWN)))
        return pe;
    }
    return PredicateEvaluation.FALSE;
  }

  /**
   * Evaluate XOR operator.
   * <br>
   * true XOR true --> true.<br>
   * true XOR false --> false.<br>
   * false XOR false --> true.<br>
   * true XOR unknown --> unknown.<br>
   * false XOR unknown --> unknown.<br>
   * unknown XOR unknown --> unknown.<br>
   *
   * @param instance Instance
   * @throws TreeModelException if the operator is applied to a non-binary argument.
   * @return PredicateEvaluation
   */
  protected PredicateEvaluation evaluateXOR(Instance instance)
      throws TreeModelException {
    if (getSize() != 2) {
      String msg = "XOR operator can be applied only at two arguments!";
      throw new TreeModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    Predicate p1 = (Predicate) predicates.get(0);
    PredicateEvaluation pe1 = p1.evaluate(instance);
    Predicate p2 = (Predicate) predicates.get(1);
    PredicateEvaluation pe2 = p2.evaluate(instance);
    if ((pe1.equals(PredicateEvaluation.UNKNOWN)) ||
        (pe2.equals(PredicateEvaluation.UNKNOWN)))
      return PredicateEvaluation.UNKNOWN;
    if (pe1.equals(pe2))
      return PredicateEvaluation.TRUE;
    else
      return PredicateEvaluation.FALSE;
  }

  /**
   * The operator surrogate provides a special means to handle logical expressions with
   * missing values. It is applied to a sequence of predicates. The order of the predicates
   * matters, the first predicate is the primary, the next predicates are the surrogates.
   * Evaluation order is left-to-right.<br>
   * The cascaded predicates are applied when the primary predicate evaluates to 'Unknown'.
   * Therefore, the surrogate predicates can provide a resolution to an undetermined predicate.
   *
   * @param instance Instance
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  public PredicateEvaluation evaluateSURROGATE(Instance instance)
      throws TreeModelException {
    for (int i=0 ; i<getSize(); i++) {
      Predicate p = (Predicate) predicates.get(i);
      PredicateEvaluation pe = evaluate(p);
      if ((pe.equals(PredicateEvaluation.TRUE)) ||
          (pe.equals(PredicateEvaluation.FALSE)))
        return pe;
    }
    return PredicateEvaluation.UNKNOWN;
  }

  /**
   * Returns the size of the predicate (i.e. the number of compounding predicates).
   *
   * @return int
   */
  public int getSize() {
    return predicates.size();
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = operator.toString() + "(";
    for (int i=0; i<predicates.size(); i++)
      ris = ris + ((Predicate) predicates.get(i)).toString() + ", ";
    return ris+")";
  }

  /**
   * Returns a representation of this object as XML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("CompoundPredicate");
    for (int i=0; i< predicates.size(); i++) {
      Predicate p = (Predicate) predicates.get(i);
      Element elem = p.toXML();
      ris.appendChild(elem);
    }
    if (operator.equals(BooleanOperator.AND))
      ris.setAttribute("booleanOperator", "and");
    if (operator.equals(BooleanOperator.OR))
      ris.setAttribute("booleanOperator", "or");
    if (operator.equals(BooleanOperator.XOR))
      ris.setAttribute("booleanOperator", "xor");
    if (operator.equals(BooleanOperator.SURROGATE))
      ris.setAttribute("booleanOperator", "surrogate");
    return ris;
  }
}
