/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import java.util.Vector;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * A CompoundBooleanTree implements a combination between two or more classification trees
 * with a boolean operator AND, OR. AND and OR evaluations can be applied only at decision trees,
 * in witch the target attribute is binary with a positive class (e.g. true, yes) and a
 * negative class (e.g. false, no) as values.<br>
 * More precisely, given n distinct decision trees D1, . . . ,Dn, the AND tree (OR tree) assigns
 * to a tuple the positive class (negative class) if all the trees D1, . . . ,Dn classify the
 * instance with the positive class (negative class); otherwise, the negative class (positive class)
 * is returned.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CompoundBooleanTree extends CompoundTree {

  /**
   * AND tree combination.
   */
  public final static byte AND_TREE = 0;

  /**
   * OR tree combination
   */
  public final static byte OR_TREE = 1;

  /**
   * The combination type used for this trees. Can be AND, OR.
   */
  protected byte combination_type;

  /**
   * The positive class relaterd to the target attribute (eg. "yes", "true", ...)
   */
  protected String positive_class;

  /**
   * The negative class related to the target attribute (eg. "no", "false", ...)
   */
  protected String negative_class;

  private final static String ATTRIBUTE_NEGATIVE_CLASS = "negative_class";
  private final static String ATTRIBUTE_POSITIVE_CLASS = "positive_class";
  private final static String ATTRIBUTE_COMBINATION_TYPE = "combination_type";

  /**
   * Constructor.
   *
   * @param positive_class String the positive class relaterd to the target
   * attribute (eg. "yes", "true", ...)
   * @param negative_class String the negative class related to the target
   * attribute (eg. "no", "false", ...)
   * @param combination_type byte the combination type used for this trees.
   * Can be AND, OR.
   */
  public CompoundBooleanTree(String positive_class, String negative_class, byte combination_type) {

    super();
    this.positive_class = positive_class;
    this.negative_class = negative_class;
    this.combination_type = combination_type;
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param el Element
   * @throws TreeModelException
   */
  protected CompoundBooleanTree(Element el)
      throws TreeModelException {

    super(el);
    java.util.Vector list = XMLUtility.getChildren(el);
    Element tmp = (Element) XMLUtility.getFirstChild(el);
    Element extension = (Element) XMLUtility.getFirstChild(tmp);
    positive_class = XMLUtility.getAttribute(extension, ATTRIBUTE_POSITIVE_CLASS);
    negative_class = XMLUtility.getAttribute(extension, ATTRIBUTE_NEGATIVE_CLASS);

    String combination = extension.getAttribute(ATTRIBUTE_COMBINATION_TYPE);
    if (combination.equalsIgnoreCase("and"))
      combination_type = this.AND_TREE;
    else
      combination_type = this.OR_TREE;


    Element child;
    // scorre i figli del nodo e crea ricorsivamente la classe
    for (int i=0; i<list.size(); i++) {
      child = (Element) list.get(i);
      if (child.getTagName().equals("Node")) {
        if (! ( (Element) XMLUtility.getFirstChild(child)).getTagName().equals(
            "Extension"))

          // simple node
          trees.add(new SimpleNode(child));
        else {
          // extension
          tmp = (Element) kddml.Core.XMLUtility.getFirstChild(child);
          Element ext = (Element) kddml.Core.XMLUtility.getFirstChild(tmp);
          combination = ext.getAttribute(ATTRIBUTE_COMBINATION_TYPE);
          if ( (combination.equalsIgnoreCase("and")) ||
              (combination.equalsIgnoreCase("or"))) {
            trees.add(new CompoundBooleanTree(child));
          }
          else
            trees.add(new CompoundBaggingsTree(child));
        }
      }
    }
  }

  /**
   * Method used to classify an input instance.
   * Given n distinct decision trees D1, . . . ,Dn, the AND tree (OR tree) assigns
   * to a tuple the positive class (negative class) if all the trees D1, . . . ,Dn
   * classify the instance with the positive class (negative class);
   * otherwise, the negative class (positive class) is returned.
   *
   * @param values String[] classification values returned by single trees.
   * @return String the classification value.
   */
  public String classify(String[] values) {

    switch(combination_type) {
      case OR_TREE:
        for (int i = 0; i < values.length; i++) {
          if (values[i].equalsIgnoreCase(positive_class)) {
            return positive_class;
          }
        }
        return negative_class;
      case AND_TREE:
        for (int i=0; i<values.length; i++) {
          if (!values[i].equalsIgnoreCase(positive_class)) {
            return negative_class;
          }
        }
        return positive_class;
      default:
        return null;
    }
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "";
    if (combination_type == AND_TREE)
      ris = "AND combination between "+super.trees.size()+" trees:\n";
    if (combination_type == OR_TREE)
      ris = "OR combination between "+super.trees.size()+" trees:\n";
    for (int i=0; i<trees.size(); i++) {
      ris = ris + "Tree "+i+":\n";
      Node n = (Node) trees.get(i);
      ris = ris + n.toString() + "\n";
    }
    return ris;
  }

  /**
   * Returns a representation of this object as XML element.
   *
   * @return Element
   */
  protected Element toXML() {

    Element ris = XMLUtility.doc.createElement("Node");
    Element ext = XMLUtility.doc.createElement("Extension");
    Element extension = XMLUtility.doc.createElement("X-VotingTree");
    ext.appendChild(extension);
    String t = "";
    if (combination_type == AND_TREE)
      t = "and";
    if (combination_type == OR_TREE)
      t = "or";
    extension.setAttribute(ATTRIBUTE_COMBINATION_TYPE, t);
    extension.setAttribute("number_of_trees", new Integer(trees.size()).toString());
    XMLUtility.setAttribute(extension, ATTRIBUTE_POSITIVE_CLASS, positive_class);
    XMLUtility.setAttribute(extension, ATTRIBUTE_NEGATIVE_CLASS, negative_class);

    ris.appendChild(ext);
    // predicate is required
    Element predicate = XMLUtility.doc.createElement("True");
    ris.appendChild(predicate);
    for (int i=0; i<trees.size(); i++) {
      Node n = (Node) trees.get(i);
      ris.appendChild(n.toXML());
    }
    ris.setAttribute("score", "");
    return ris;
  }
}
