/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;

/**
 * A ConfusionMatrix specifies the statistics of the correct predictions and mispredictions.
 * The rows represent the predicted classes whereas the columns indicate the correct class values.
 * The value at entry (j, k) is the number of predictions for class j when k is the correct
 * prediction.<br>
 * On the rows we assume the actual values.
 * On the columns we assume the predicted values.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ConfusionMatrix extends kddml.Core.DataMining.CategoryMatrix
    implements ConfusionMatrixManager {

  /**
   * The number of incorrectly predicions.
   */
  protected int incorrectly_predictions;

  /**
   * The number of total predictions.
   */
  protected int total_predictions;

  /**
   * Indicates if this is a test or a training confusion matrix.
   */
  protected boolean test_matrix;

  /**
   * Constructor given the list of fields related to the target attribute.
   *
   * @param categories String[] the categories of target attribute (e.g. "sunny", "overcast", "rainy")
   * @param training_set boolean if true, the matrix is a training confusion matrix. Otherwise,
   * the matrix is a test confusion matrix.
   */
  public ConfusionMatrix(String[] categories, boolean training_set) {

    super(categories);
    test_matrix = !training_set;
    total_predictions = 0;
    incorrectly_predictions = 0;
  }

  /**
   * Proprietary constructor given the PMML element.
   *
   * @param confusion_matrix Element
   */
  protected ConfusionMatrix(Element confusion_matrix) {

    super(confusion_matrix);
    if (confusion_matrix.getTagName().equalsIgnoreCase("X-ConfusionMatrixTraining"))
      test_matrix = false;
    else
      test_matrix = true;
    total_predictions = new Double(confusion_matrix.getAttribute("x-totalInst")).intValue();
    incorrectly_predictions = new Double(confusion_matrix.getAttribute("x-incorrectlyInst")).intValue();
  }

  /**
   * Returns a representation of this object as XML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris;
    if (test_matrix)
      ris = XMLUtility.doc.createElement("X-ConfusionMatrixTest");
    else
      ris = XMLUtility.doc.createElement("X-ConfusionMatrixTraining");
    ris.setAttribute("x-incorrectlyInst",new Integer(incorrectly_predictions).toString());
    ris.setAttribute("x-totalInst",new Integer(total_predictions).toString());
    String s = new Double(getPercentageError() * 100).toString();
    if (s.length() > 6)
      s = s.substring(0, 5);
    ris.setAttribute("x-incorrectlyInstPerc", s);
    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n", new Integer(categories.length).toString());
    array.setAttribute("type","string");
    String list="";
    for(int i=0; i<categories.length;i++) {
      String str = (String) categories[i];
      str = str.replaceAll("<", "&lt;");
      list = list + str +" ";
    }
    // rimuove l'ultimo spazio dalla stringa
    list = list.substring(0, list.length()-1);
    Node n = XMLUtility.doc.createTextNode(list);
    array.appendChild(n);
    Element matrix = XMLUtility.doc.createElement("Matrix");
    matrix.setAttribute("kind","any");
    for (int i=0; i<categories.length; i++) {
      String value="";
      for (int j=0; j<categories.length; j++) {
        value = value + (new Double(this.matrix[i][j]).toString()) + " ";
      }
      // rimuove l'ultimo spazio dalla stringa
      value = value.substring(0, value.length()-1);
      Element elem = XMLUtility.doc.createElement("Array");
      elem.setAttribute("n",new Integer(categories.length).toString());
      elem.setAttribute("type","real");
      Node n1 = XMLUtility.doc.createTextNode(value);
      elem.appendChild(n1);
      matrix.appendChild(elem);
    }
    ris.appendChild(array);
    ris.appendChild(matrix);
    return ris;
  }

  /**
   * Returns the total number of predications.
   *
   * @return int
   */
  public int getNumberOfPredictions() {
   return total_predictions;
  }

  /**
   * Returns the categories of the target attribute.
   *
   * @return String[]
   */
  public String[] getCategories() {
    return categories;
  }

  /**
   * Returns the ratio of the sum of the diagonal entries to the sum of all entries.
   * The output value is a number between 0 and 1.
   *
   * @return double
   */
  public double getPercentageAccuracy() {
    int correctly_predictions = total_predictions - incorrectly_predictions;
    return ((double) correctly_predictions / (double) total_predictions);
  }

  /**
   * Returns the sum of the non-diagonal sums where the actual value does not equal to
   * predicted value, divided by the sum of all cells.
   * The output value is a number between 0 and 1.
   *
   * @return double
   */
  public double getPercentageError() {
    return ((double) incorrectly_predictions / (double) total_predictions);
  }

  /**
   * Sets the number of prediction when the actual prediction is
   * actual_value and the predicted value is predicted_value.
   *
   * @param actual_value String the actual category value for the target attribute.
   * @param predicted_value String the predicated category value for the target attribute.
   * @param predictions double the value to set in the matrix.
   * @throws MiningModelException
   */
  public void setValue(String actual_value, String predicted_value, double predictions)
      throws MiningModelException {

    super.setValue(actual_value, predicted_value, predictions);

    total_predictions = total_predictions + (int) predictions;
    if (!actual_value.equals(predicted_value))
      incorrectly_predictions = incorrectly_predictions + (int) predictions;
  }

  /**
   * Returns the number of predictions for the given class when the actual prediction
   * is actual_value and the predicted value is predicated_value.
   *
   * @param actual_value String
   * @param predicted_value String
   * @throws MiningModelException
   * @return double
   */
  public double getValue(String actual_value, String predicted_value)
      throws MiningModelException {

    return super.getValue(actual_value, predicted_value);
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "";
    if (test_matrix)
      s = s + "========= Confusion Matrix on Test Set ==========\n";
    else
      s = s + "========= Confusion Matrix on Training Set ==========\n";
    s = s + "Total number of predictions: "+total_predictions+"\n";
    s = s + "Total number of incorrectly predictions: "+incorrectly_predictions+"\n";
    s = s + "Percentage of incorrectly predictions: "+getPercentageError()+"\n";
    s = s + "\nValues of target attribute: ";
    for (int i=0; i<categories.length; i++) {
      s = s + categories[i] + "   ";
    }
    s = s + "\n";
    for (int i=0; i<categories.length; i++) {
      for (int j=0; j<categories.length; j++) {
        s = s + matrix[i][j] + "   ";
      }
      s = s + "\n";
    }
    return s;
  }
}
