/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import org.w3c.dom.*;
import java.io.*;
import weka.core.*;

/**
 * Factory class that builds a classification model from a PMML model.<br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class PMMLClassificationFactory extends ClassificationExternalFactory {

  private String pmml_version;

  private Element header;

  private Element data_dictionary;

  private Element transformation_dictionary;

  private Element mining_schema;

  private Element tree_model;

  private String algorithm_name;

  private String model_name;

  private boolean binary_split;

  private Element tree_root;

  private Element training_matrix;

  private Element test_matrix;

  /**
   * Constructor given the pmml_path.
   *
   * @param f File the file in the system repository that will contain the generated model.
   * @param pmml_path String the URI identifying the pmml document.
   * @throws KDDMLCoreException if the source cann't be found or the parser returns
   * errors.
   */
  public PMMLClassificationFactory(java.io.File f, String pmml_path)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    // effettua il parsing del documento XML contenente il modello...
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(pmml_path),
                              new DTDResolver(kddml.Core.KDDMLObjectType.PMML_CLASSIFICATION_MODEL));
    }
    catch(java.io.IOException e) {
      String msg = "Invalid PMML reference: "+pmml_path;
      throw new TreeModelException(msg, e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }

    Element root = (Element) doc.getDocumentElement();

    if (!root.getTagName().equalsIgnoreCase("pmml"))
      throw new TreeModelException(pmml_path+ " no contains a PMML document.",
                                             null,
                                             KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the pmml version
    pmml_version = root.getAttribute("version");

    // reading the copyright
    header = (Element) root.getElementsByTagName("Header").item(0);

     // reading the data dictionary element
    data_dictionary = (Element) root.getElementsByTagName("DataDictionary").item(0);

    // reading the transformation dictionary element
    NodeList l = root.getElementsByTagName("TransformationDictionary");
    if (l.getLength() != 0)
      transformation_dictionary = (Element) l.item(0);
    else
      transformation_dictionary = null;

    tree_model = (Element) doc.getElementsByTagName("TreeModel").item(0);
    if (tree_model == null)
      throw new TreeModelException(pmml_path+ " no contains a PMML tree model document.",
                                             null,
                                             KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the mining schema element...
    mining_schema = (Element) tree_model.getElementsByTagName("MiningSchema").item(0);

    // reading attributes...
    model_name = XMLUtility.getAttribute(tree_model, "modelName");
    algorithm_name = XMLUtility.getAttribute(tree_model, "algorithmName");
    String split = tree_model.getAttribute("splitCharacteristic");
    if (split.equals("binarySplit"))
      binary_split = true;
    else
      binary_split = false;

    // tree node root (the first child is the mining schema)
    tree_root = (Element) XMLUtility.getChild(tree_model, 1);

    // get the training/test confusion matrix elements
    training_matrix = (Element) tree_model.getElementsByTagName("X-ConfusionMatrixTraining").item(0);
    test_matrix = (Element) tree_model.getElementsByTagName("X-ConfusionMatrixTest").item(0);
  }

  /**
   * Returns the data dictionary of the output model.
   *
   * @throws MiningModelException
   * @return DataDictionary
   */
  protected DataDictionary buildDataDictionary() throws MiningModelException {

    if (data_dictionary == null)
      return new DataDictionary();
    else
      return PMMLUtility.buildDataDictionary(data_dictionary);
  }

  /**
   * Returns the mining schema of the output model.
   *
   * @throws MiningModelException
   * @return MiningSchema
   */
  protected MiningSchema buildMiningSchema() throws MiningModelException {

    if (mining_schema == null)
      return new MiningSchema();
    else
      return PMMLUtility.buildMiningSchema(mining_schema);
  }

  /**
   * Returns the name of the algorithm.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildAlgorithmName() throws MiningModelException {

    return algorithm_name;
  }

  /**
   * Returns the name of the model.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildModelName() throws MiningModelException {

   return model_name;
  }

  /**
   * Returns the PMML version supported.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildPMMLVersion() throws MiningModelException {

    return pmml_version;
  }

  /**
   * Returns the copyright.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildCopyright() throws MiningModelException {

    if (header != null)
      return header.getAttribute("copyright");
    else
      return null;
  }

  /**
   * Returns the application name.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationName() throws MiningModelException {

    if (header != null) {
      Element el = (Element) header.getElementsByTagName("Application").item(0);
      if (el != null)
        return XMLUtility.getAttribute(el, "name");
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Returns the application version.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationVersion() throws MiningModelException {

    if (header != null) {
      Element el = (Element)header.getElementsByTagName("Application").item(0);
      if (el != null)
        return el.getAttribute("version");
      else
        return null;
    }
    else
      return null;
  }

  /**
   * Returns the node root of the model.
   *
   * @throws TreeModelException
   * @return Node
   */
  protected Node buildTree() throws TreeModelException {

    Element f = (Element) XMLUtility.getFirstChild(tree_root);
    if (!f.getTagName().equals("Extension"))
      // simple node
      return new SimpleNode(tree_root);
    else {
      // extension
      Element tmp = (Element) XMLUtility.getFirstChild(tree_root);
      Element el = (Element) XMLUtility.getFirstChild(tmp);
      if (el.getTagName().equals("X-VotingTree")) {
        // compound node
        String combination = el.getAttribute("combination_type");
        if (combination.equalsIgnoreCase("committee"))
          return new CompoundBaggingsTree(tree_root);
        else {
            return new CompoundBooleanTree(tree_root);
        }
      }
      else
        throw new TreeModelException("Input model don't contains a valid PMML tree model document.",
                                     null,
                                     KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Returns the confusion matrix. Can be null.
   *
   * @throws TreeModelException
   * @return ConfusionMatrix
   */
  protected ConfusionMatrix buildTrainingConfusionMatrix()
      throws TreeModelException {

    if (training_matrix != null)
      return new ConfusionMatrix(training_matrix);
    else
      return null;
  }

  /**
   * Returns the text matrix. Can be null.
   *
   * @throws TreeModelException
   * @return ConfusionMatrix
   */
  protected ConfusionMatrix buildTestConfusionMatrix()
      throws TreeModelException {

    if (test_matrix != null)
      return new ConfusionMatrix(test_matrix);
    else
      return null;
  }

  /**
   * Returns true if the model use the binary splitting (every node with 2 children).
   * Returns false otherwise.
   *
   * @return boolean
   */
  protected boolean isBinarySplitting() {
    return binary_split;
  }

  /**
   * Only for testing...
   *
   * @param args String
   */
  public static void main(String args) {

    File dest = new File("D:/my_tree.xml");
    String model = "http://www.dmg.org/pmml_examples/weather.xml";

    try {
      FileReader instances = new FileReader("D:/DataSet/weather.arff");
      PMMLClassificationFactory factory = new PMMLClassificationFactory(dest, model);
      factory.setOverwrite(true);
      TreeModel my_tree = (TreeModel) factory.newInstance();
      Instances dataset = new Instances(instances);

      // loop on instances
      Instance inst;
      for (int i=0; i<dataset.numInstances(); i++) {
        inst = dataset.instance(i);
        System.out.println(my_tree.classify(inst));
      }
    }
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
    catch (Exception ex1) {
    }
  }
}
