/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * The score distribution is a method to list predicted values in a classification trees
 * structure. If the node holds an enumeration, each entry of the enumeration is stored in
 * an hashtable structure where:
 * <ol>
 * <li> the key is the category of the target attribute; </li>
 * <li> the value is the size (i.e. the number of records) associated with the
 * category </li>
 * </ol>
 * When a SimpleNode is selected as the final node and if this node has no 'score' attribute,
 * then the highest recordCount the ScoreDistribution determines which value is selected
 * as predicted class.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ScoreDistribution implements ScoreDistributionManager {

  /**
   * This hashtable contains a single score distribution:
   * <ol>
   * <li> the key is the category of the target attribute; </li>
   * <li> the value is the size (i.e. the number of records) associated with the
   * category </li>
   * </ol>
   */
  protected Hashtable ht;

  /**
   * The total sum of record counts.
   */
  protected double total;

  /**
   * Constructor.
   *
   * @param categories String[] the list of categories for the target attribute.
   * @param number_of_instances double[] the number of records associated with the categories.
   * @throws TreeModelException if the size of categories and the size of
   * number_of_instances differs.
   */
  public ScoreDistribution(String[] categories, double[] number_of_instances)
      throws TreeModelException {

    ht = new Hashtable();
    total = 0;
    if (categories.length != number_of_instances.length)
      throw new TreeModelException("Categories and values do not match!",
                                   null,
                                   kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    for (int i=0; i<categories.length; i++) {
      ht.put(categories[i], new Double(number_of_instances[i]));
      total = total + number_of_instances[i];
    }
  }

  /**
   * Constructor given the categories values for the target attribute.
   * The number of records related to the category is 0.
   *
   * @param categories String[] the list of categories for the target attribute.
   */
  public ScoreDistribution(String[] categories) {

    ht = new Hashtable();
    total = 0;
    for (int i=0; i<categories.length; i++) {
      ht.put(categories[i], new Double(0));
    }
  }

  /**
   * Proprietary constructor given a list of PMML elements.
   *
   * @param elements Vector
   */
  protected ScoreDistribution(Vector elements) {
    ht = new Hashtable();
    total = 0;
    for (int i=0; i< elements.size(); i++) {
      Element elem = (Element) elements.get(i);
      String value = XMLUtility.getAttribute(elem, "value");
      Double count = new Double(elem.getAttribute("recordCount"));
      ht.put(value, count);
      total = total + count.doubleValue();
    }
  }

  /**
   * Returns the record count of a given category.
   * Returns -1 if no category can be found in the target attribute.
   *
   * @param name String the category name
   * @return double
   */
  public double getCount(String name) {
    if (!ht.containsKey(name))
      return -1;
    else
      return ((Double) ht.get(name)).doubleValue();
  }

  /**
   * Returns the record count as percentage of a given category.
   * Returns -1 if no category can be found in the target attribute.
   *
   * @param name String the category name
   * @return double a value between 0 and 1 or -1 if the category is not found.
   */
  public double getPercCount(String name) {
    if (!ht.containsKey(name))
      return -1;
    else {
      int value = ((Double) ht.get(name)).intValue();
      return ((double) total / (double) value);
    }
  }

  /**
   * Returns the highest category value of this score distribution.
   * If two or more values have the highest distribution, then the method returns
   * the first category value.
   *
   * @return String
   */
  public String getHighestCategory() {
    Enumeration values = ht.elements();
    Enumeration keys = ht.keys();
    double max = -1;
    String ris = "";
    while (values.hasMoreElements()) {
      String value = (String) keys.nextElement();
      double count = ((Double) values.nextElement()).doubleValue();
      if (count > max) {
        max = count;
        ris = value;
      }
    }
    return ris;
  }

  /**
   * Returns the highest record count of this score distribution.
   *
   * @return double
   */
  public double getHighestValue() {
    Enumeration values = ht.elements();
    double max = -1;
    while (values.hasMoreElements()) {
      double count = ((Double) values.nextElement()).doubleValue();
      if (count > max) {
        max = count;
      }
    }
    return max;
  }

  /**
   * Returns the lowest value of this score distribution.
   */
 /* public int getLowestValue() {
    Enumeration values = ht.elements();
    Enumeration keys = ht.keys();
    int min = Integer.MAX_VALUE;
    while (values.hasMoreElements()) {
      String value = (String) values.nextElement();
      int count = ((Double) keys.nextElement()).intValue();
      if (count < min) {
        min = count;
      }
    }
    return min;
  }*/

  /**
   * Returns the total sum of record counts.
   *
   * @return double a positive double
   */
  public double getTotal() {
    return total;
  }

  /**
   * Returns the number of categories for this score distribution.
   *
   * @return int a positive integer
   */
  public int getNumberOfValues() {
    return ht.size();
  }

  /**
   * Updates a distribution for a given category.
   * Returns false if the value do not exist in the hashtable.
   * Returns true if updating has success.
   *
   * @param value String
   * @param record_count double
   * @return boolean
   */
  protected boolean updateDistribution(String value, double record_count) {
    if (!ht.containsKey(value))
      return false;
    double old = ((Double) ht.get(value)).doubleValue();
    ht.put(value, new Double(old + record_count));
    total = total + record_count;
    return true;
  }

  /**
   * Returns the list categories as enumeration.
   *
   * @return Enumeration list of strings.
   */
  public Enumeration getCategories() {
    return ht.keys();
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    return ht.toString();
  }

  /**
   * Returns a representation of this object as list of PMML elements.
   *
   * @return Element[]
   */
  protected Element[] toXML() {

    Enumeration values = ht.elements();
    Enumeration keys = ht.keys();
    Element[] ris = new Element[ht.size()];
    int i=0;
    while (values.hasMoreElements()) {
      String value = (String) keys.nextElement();
      Double count = (Double) values.nextElement();
      Element elem = XMLUtility.doc.createElement("ScoreDistribution");
      XMLUtility.setAttribute(elem, "value", value);
      elem.setAttribute("recordCount",count.toString());
      ris[i] = elem;
      i++;
    }
    return ris;
  }
}
