/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import weka.core.Instance;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A SimpleComparisonPredicate consists of a single comparison between a logical attribute
 * value and a constant. The constant can be either numeric or categorical.
 * A SimpleComparisonPredicate consists of three major parts:
 * <ol>
 * <li> a comparison operator (equal, notEqual, lessThan, lessOrEqual,
 * greaterThan, greaterOrEqual). </li>
 * <li> the name entry of one of the mining field elements at the mining schema. </li>
 * <li> the information to evaluate / compare against. </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SimpleComparisonPredicate extends SimplePredicate {

  /**
   * The mining field.
   */
  protected String string_value;

  /**
   * The constant to compare.
   */
  protected Double numeric_value;

  /**
   * The comparison operator.
   */
  protected ComparisonOperator op_code;

  /**
   * Constructor for a numeric comparison.
   *
   * @param field String
   * @param value String
   * @param op_code ComparisonOperator
   */
  public SimpleComparisonPredicate(String field, String value, ComparisonOperator op_code) {
    this.field = field;
    string_value = value;
    numeric_value = null;
    this.op_code = op_code;
  }

  /**
   * Constructor for a categorical comparison.
   *
   * @param field String
   * @param value double
   * @param op_code ComparisonOperator
   */
  public SimpleComparisonPredicate(String field, double value, ComparisonOperator op_code) {
    this.field = field;
    string_value = null;
    numeric_value = new Double(value);
    this.op_code = op_code;
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param elem Element
   * @throws TreeModelException
   */
  protected SimpleComparisonPredicate(Element elem)
      throws TreeModelException {

    if (!elem.getTagName().equalsIgnoreCase("SimplePredicate")) {
      String msg= "Invalid XML element. No SimplePredicate input!";
      throw new TreeModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    field = XMLUtility.getAttribute(elem, "field");
    String p = XMLUtility.getAttribute(elem, "value");
    try {
      Double d = new Double(p);
      // in input un valore di tipo numerico...
      string_value = null;
      numeric_value = d;
    }
    catch(NumberFormatException e) {
      // in input una valore di tipo stringa...
      string_value = p;
      numeric_value = null;
    }
    String code = elem.getAttribute("operator");
    if (code.equals("equal"))
      op_code = ComparisonOperator.EQUAL;
    if (code.equals("notEqual"))
      op_code = ComparisonOperator.NOT_EQUAL;
    if (code.equals("lessThan"))
      op_code = ComparisonOperator.LESS_THAN;
    if (code.equals("lessOrEqual"))
      op_code = ComparisonOperator.LESS_OR_EQUAL;
    if (code.equals("greaterThan"))
      op_code = ComparisonOperator.GREATER_THAN;
    if (code.equals("greaterOrEqual"))
      op_code = ComparisonOperator.GREATER_OR_EQUAL;
  }

  /**
   * Evaluates this predicate on a single instance.
   *
   * @param instance Object
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  public PredicateEvaluation evaluate(Object instance)
      throws TreeModelException {

    Instance tupla = (Instance) instance;
    // pos indica la posizione del campo field all'interno dell'istanza...
    int pos = -1;
    for(int i=0; i<tupla.numAttributes(); i++) {
      if (tupla.attribute(i).name().equalsIgnoreCase(field))
        pos = i;
    }
    if (pos == -1) {
      String msg = "Attribute < " + field + " > do not exist!";
      throw new TreeModelException(msg, null, KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    // check for missing attribute...
    if (tupla.isMissing(pos))
      // RETURN UNKNOWN!!!
      return PredicateEvaluation.UNKNOWN;

    // controllo la compatibilita' con i tipi...
    if (tupla.attribute(pos).isNumeric() && numeric_value == null) {
      String msg = "Incompatible type: attribute < " + field +" > required numeric value!";
      throw new TreeModelException(msg, null, KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (!tupla.attribute(pos).isNumeric() && string_value == null) {
      String msg = "Incompatible type: attribute < " + field +" > required a string value!";
      throw new TreeModelException(msg, null, KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (tupla.attribute(pos).isNumeric())
      return evaluateNumericField(tupla.value(pos), numeric_value.doubleValue());
    else
      return evaluateStringField(tupla.stringValue(pos), string_value);
  }

  public String toString() {
    if (numeric_value != null)
      return super.field + " "+ op_code.toString() + " " + numeric_value;
    else
      return super.field + " "+ op_code.toString() + " " + string_value;
  }

  /**
   * Evaluate two numeric fields.
   *
   * @param value1 double
   * @param value2 double
   * @return PredicateEvaluation
   */
  private PredicateEvaluation evaluateNumericField(double value1, double value2) {

    if (op_code.equals(ComparisonOperator.EQUAL))
      return ret(value1 == value2);
    if (op_code.equals(ComparisonOperator.NOT_EQUAL))
      return ret(value1 != value2);
    if (op_code.equals(ComparisonOperator.LESS_THAN))
      return ret(value1 < value2);
    if (op_code.equals(ComparisonOperator.LESS_OR_EQUAL))
      return ret(value1 <= value2);
    if (op_code.equals(ComparisonOperator.GREATER_THAN))
      return ret(value1 > value2);
    if (op_code.equals(ComparisonOperator.GREATER_OR_EQUAL))
      return ret(value1 >= value2);
    return PredicateEvaluation.UNKNOWN;
  }

  /**
   * Evaluate two string fields.
   *
   * @param value1 String
   * @param value2 String
   * @return PredicateEvaluation
   */
  private PredicateEvaluation evaluateStringField(String value1, String value2) {

    int compare_result = value1.compareTo(value2);
    if (op_code.equals(ComparisonOperator.EQUAL))
      return ret(compare_result == 0);
    if (op_code.equals(ComparisonOperator.NOT_EQUAL))
      return ret(compare_result != 0);
    if (op_code.equals(ComparisonOperator.LESS_THAN))
      return ret(compare_result < 0);
    if (op_code.equals(ComparisonOperator.LESS_OR_EQUAL))
      return ret(compare_result <= 0);
    if (op_code.equals(ComparisonOperator.GREATER_THAN))
      return ret(compare_result > 0);
    if (op_code.equals(ComparisonOperator.GREATER_OR_EQUAL))
      return ret(compare_result >= 0);
    return PredicateEvaluation.UNKNOWN;
  }

  private PredicateEvaluation ret(boolean b) {
    if (b)
      return PredicateEvaluation.TRUE;
    else
      return PredicateEvaluation.FALSE;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("SimplePredicate");
    XMLUtility.setAttribute(ris, "field",field);
    if (string_value != null)
      XMLUtility.setAttribute(ris, "value",string_value);
    else
      ris.setAttribute("value",numeric_value.toString());
    if (op_code.equals(ComparisonOperator.EQUAL))
      ris.setAttribute("operator", "equal");
    if (op_code.equals(ComparisonOperator.NOT_EQUAL))
      ris.setAttribute("operator", "notEqual");
    if (op_code.equals(ComparisonOperator.LESS_THAN))
      ris.setAttribute("operator", "lessThan");
    if (op_code.equals(ComparisonOperator.LESS_OR_EQUAL))
      ris.setAttribute("operator", "lessOrEqual");
    if (op_code.equals(ComparisonOperator.GREATER_THAN))
      ris.setAttribute("operator", "greaterThan");
    if (op_code.equals(ComparisonOperator.GREATER_OR_EQUAL))
      ris.setAttribute("operator", "greaterOrEqual");
    return ris;
  }
}
