/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import java.util.Vector;
import org.w3c.dom.Element;
import kddml.Core.*;

/**
 * A SimpleNode represents a node in a decision tree. Every SimpleNode contains a predicate
 * that identifies a rule for choosing itself or any of its siblings. A predicate may be an
 * expression composed of other nested predicates.
 * In addition, a SimpleNode may contain a score distribution which characterizes the
 * distribution of data at that node with respect to a reference dataset.
 * Summarizing, a SimpleNode consists of five major parts:
 * <ol>
 * <li> the list of children if the node is an internal node; </li>
 * <li> the reference to the father node if the node is not the root; </li>
 * <li> a not null predicate; </li>
 * <li> the score distribution of data; </li>
 * <li> the category for the target attribute assigned by algorithm and, optionally,
 *  the number of cases assigned to this tree node.</li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SimpleNode extends Node implements SimpleNodeManager {

  /**
   * The predicate related to this node.
   */
  protected Predicate predicate;

  /**
   * The list of children as vector. Can be empty.
   */
  protected Vector children;

  /**
   * The category of the target attribute related to this node.
   */
  protected String score;

  /**
   * The number of cases assigned by the algorithm to this node.
   */
  protected double record_count;

  /**
   * The score distribution.
   */
  protected ScoreDistribution distribution;

  /**
   * The reference to the parent node. Can be null if the node is the root.
   */
  protected SimpleNode parent;

  /**
   * Constructor used for internal node.
   *
   * @param score String the category of the target attribute related to this node.
   * @param p Predicate the predicate. Cannot be null.
   * @param distribution ScoreDistribution the score distribution related to this node.
   * Can be null if no score distribution is provided.
   * @param parent SimpleNode the parent node. Cannot be null.
   */
  public SimpleNode(String score, Predicate p, ScoreDistribution distribution, SimpleNode parent) {
    children = new Vector();
    this.distribution = distribution;
    this.score = score;
    this.predicate = p;
    this.parent = parent;
  }

  /**
   * Constructor used for the root node.
   *
   * @param score String the category of the target attribute related to this node.
   * @param distribution ScoreDistribution the score distribution related to this node.
   */
  public SimpleNode(String score, ScoreDistribution distribution) {
    children = new Vector();
    this.distribution = distribution;
    this.score = score;
    predicate = new SimpleBooleanPredicate(true);
    this.parent = null;
  }

  /**
   * Proprietary constructor given the PMML element.
   *
   * @param elem Element
   * @throws TreeModelException
   */
  protected SimpleNode(Element elem) throws TreeModelException {

    if (!elem.getTagName().equals("Node"))
      throw new TreeModelException("Invalid PMML document. No <Node> in input!", null,
                                   KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    score = XMLUtility.getAttribute(elem, "score");
    if (!elem.getAttribute("recordCount").equals(""))
      record_count = new Double(elem.getAttribute("recordCount")).doubleValue();
    else
      record_count = -1;

    Element predicate_el = (Element) XMLUtility.getFirstChild(elem);
    String tag = predicate_el.getTagName();
    // set the predicate...
    if (tag.equals("SimplePredicate"))
      predicate = new SimpleComparisonPredicate(predicate_el);
    else {
      if (tag.equals("CompoundPredicate"))
        predicate = new CompoundBooleanPredicate(predicate_el);
      else {
        if (tag.equals("SimpleSetPredicate"))
          predicate = new SimpleSetPredicate(predicate_el);
        else {
          if (tag.equals("True"))
            predicate = new SimpleBooleanPredicate(true);
          else {
            if (tag.equals("False"))
              predicate = new SimpleBooleanPredicate(false);
            else
              throw new TreeModelException("Invalid PMML document. No predicate supported!", null,
                                           KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
          }
        }
      }
    }

    // create the score distribution element
    Vector distributions = new Vector();

    // create and add the node children to the vector...
    children = new Vector();
    Vector vect = XMLUtility.getChildren(elem);
    for (int i=0; i<vect.size(); i++) {
      Element el = (Element) vect.get(i);
      if (el.getTagName().equals("Node")) {
        SimpleNode child = new SimpleNode(el);
        // setta il nodo padre del figlio
        child.parent = this;
        children.add(child);
      }
      if (el.getTagName().equals("ScoreDistribution")) {
        distributions.add(el);
      }
    }
    distribution = new ScoreDistribution(distributions);
  }

  /**
   * Returns the direct children of this node or null if the node is a leaf node.
   * If null is returned, then isLeaf() must return true for this node and getNumChildren()
   * must return 0.
   *
   * @return Enumeration a list of SimpleNodeManager or null.
   */
  public java.util.Enumeration getChildren() {
    return children.elements();
  }

  /**
   * Returns the i-esim child of this node.
   * Throws an exception if index is out of bound or the node do not have children.
   *
   * @param index int
   * @throws TreeModelException
   * @return SimpleNodeManager
   */
  public SimpleNodeManager getChild(int index) throws TreeModelException {
    if ((index<0) || (index>children.size()-1))
      throw new TreeModelException("Index out of bound!", null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    return (SimpleNodeManager) children.get(index);
  }

  /**
   * Returns the number of cases assigned to this tree node or -1 if the value is not
   * specified.
   *
   * @return double
   */
  public double getRecordCount() {
    return record_count;
  }

  /**
   * Return the category of the node assigned by the algorithm.
   *
   * @return String
   */
  public String getScore() {
    return score;
  }

  /**
   * Returns the level of this node where level(root)=1.
   * If 1 is returned, then getParent() must return null.
   *
   * @return int
   */
  public int getLevel() {
    if (isRoot())
      return 0;

    return this.parent.getLevel() + 1;
  }

  /**
   * Returns the depth of this node where the leaves have depth equal to 0.
   *
   * @return int a positive value.
   */
  public int getDepth() {
    if (this.isLeaf())
      return 0;
    int ris = 0;
    for (int i=0; i<children.size(); i++) {
      ris = Math.max(ris, ((SimpleNode) children.get(i)).getDepth());
    }
    return ris + 1;
  }

  /**
   * Returns the size of the tree, i.e. the number of total nodes of the tree.
   *
   * @return int
   */
  public int getSize() {
    int ris = 0;
    for (int i=0; i<children.size(); i++) {
      ris = ris + ((SimpleNode) children.get(i)).getSize();
    }
    return ris+1;
  }

  /**
   * Returns the number of leaves for this node.
   *
   * @return int
   */
  public int getNumberOfLeaves() {
    if (this.isLeaf())
      return 1;
    int ris = 0;
    for (int i=0; i<children.size(); i++) {
      ris = ris + ((SimpleNode) children.get(i)).getNumberOfLeaves();
    }
    return ris;
  }

  /**
   * Returns the list of leaves for this node.
   * Returns an empty vector if the node is a leaf.
   *
   * @return SimpleNodeManager[]
   */
  public Vector getLeaves() {
    Vector ris = new Vector();
    if (this.isLeaf())
      ris.add(this);
    else {
      for (int i=0; i<children.size(); i++) {
        // foglie per il ramo i
        Vector leaves = ((SimpleNode) children.get(i)).getLeaves();
        for (int j=0; j<leaves.size(); j++) {
          ris.add(leaves.get(j));
        }
      }
    }
    return ris;
  }

  /**
   * Returns parent of this node or null if the node has no parent.
   * If null is returned, then getLevel() of this node must return 1.
   *
   * @return SimpleNodeManager
   */
  public SimpleNodeManager getParent() {
    return parent;
  }

  /**
   * Returns the number of children available for this tree node.
   * If 0 is returned, then isLeaf() must return true for this node and getChildren()
   * must return null.
   *
   * @return int
   */
  public int getNumOfChildren() {
    return children.size();
  }

  /**
   * Returns the predicate assigned to this node.
   *
   * @return Predicate
   */
  public Predicate getPredicate() {
    return predicate;
  }

  /**
   * Returns the score distribution of this node.
   * Returns null if no score distribution is created.
   *
   * @return ScoreDistributionManager
   */
  public ScoreDistributionManager getDistribution() {
    return distribution;
  }

  /**
   * Sets the score distribution to this node
   *
   * @param distr ScoreDistributionManager
   */
  public void setDistribution(ScoreDistributionManager distr) {
    this.distribution = (ScoreDistribution) distr;
  }

  /**
   * Returns true if this node is a leaf.
   *
   * @return boolean
   */
  public boolean isLeaf() {
    return (getNumOfChildren() == 0);
  }

  /**
   * Returns true if the node is a root
   *
   * @return boolean
   */
  public boolean isRoot() {
    return (parent == null);
  }

  /**
   * Adds a new child to this node. Adds the child to the end.
   *
   * @param node SimpleNodeManager
   */
  public void addChild(SimpleNodeManager node) {
    // clone the input node and add it this as parent node...
    children.add(node);
  }

  /**
   * Returns true if the tree is a compound tree, i.e., a combination with and,
   * or, committee of classification trees. Returns false if the node is part of a
   * simple tree.
   *
   * @return boolean false
   */
  public boolean isCompoundTree() {
    return false;
  }

  /**
   * Sets the recond count value to this node.
   *
   * @param record_count double
   */
  public void setRecordCount(double record_count) {
    this.record_count = record_count;
  }

  /**
   * Cut a branch from this node.
   *
   * @param index int the index of the children to be pruned
   * @return boolean true if the brach has beem cut
   */
  /*public void pruneChild(int index) throws TreeModelException {
    try {
      children.remove(index);
    }
    catch(ArrayIndexOutOfBoundsException e) {
      String msg = "Can't remove node. Index is out of bound!";
      throw new TreeModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }
  }*/

  /*  public void removeAllChildren() {
    children.removeAllElements();
  }*/

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "";
    for (int j=0; j< getLevel(); j++) {
      s = s + " | ";
    }
    s = s + predicate.toString() + " --> " + score +" "+distribution.toString()+"\n";
    for (int i=0 ; i<children.size(); i++) {
      SimpleNode n = (SimpleNode) children.get(i);
      s = s + n.toString();
    }
    return s;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("Node");

    // set the attributes
    ris.setAttribute("recordCount", new Double(record_count).toString());
    XMLUtility.setAttribute(ris, "score", score);

    // append the predicate
    Element predicate = this.predicate.toXML();
    ris.appendChild(predicate);

    // append the distribution elements
    Element[] distribution = this.distribution.toXML();
    for (int i=0; i<distribution.length; i++)
      ris.appendChild(distribution[i]);

    // recursive appending all children...
    for (int j=0; j<children.size(); j++) {
      Element el = ((SimpleNode) children.get(j)).toXML();
      ris.appendChild(el);
    }
    return ris;
  }
}
