/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

/**
 * A manager interface for SimpleNode.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface SimpleNodeManager extends NodeManager {

  /**
   * Returns the direct children of this node or null if the node is a leaf node.
   * If null is returned, then isLeaf() must return true for this node and getNumChildren()
   * must return 0.
   *
   * @return Enumeration a list of SimpleNodeManager or null.
   */
  public java.util.Enumeration getChildren();

  /**
   * Returns the number of cases assigned to this tree node or -1 if the value is not
   * specified.
   *
   * @return double
   */
  public double getRecordCount();

  /**
   * Return the category of the node assigned by the algorithm.
   *
   * @return String
   */
  public String getScore();

  /**
   * Returns the level of this node where level(root)=1.
   * If 1 is returned, then getParent() must return null.
   *
   * @return int
   */
  public int getLevel();

  /**
   * Returns the number of children available for this tree node.
   * If 0 is returned, then isLeaf() must return true for this node and getChildren()
   * must return null.
   *
   * @return int
   */
  public int getNumOfChildren();

  /**
   * Returns parent of this node or null if the node has no parent.
   * If null is returned, then getLevel() of this node must return 1.
   *
   * @return SimpleNodeManager
   */
  public SimpleNodeManager getParent();

  /**
   * Returns the predicate assigned to this node.
   *
   * @return Predicate
   */
  public Predicate getPredicate();

  /**
   * Sets the score distribution to this node
   *
   * @param score_distribution ScoreDistributionManager
   */
  public void setDistribution(ScoreDistributionManager score_distribution);

  /**
   * Returns the score distribution of this node.
   * Returns null if no score distribution is created.
   *
   * @return ScoreDistributionManager
   */
  public ScoreDistributionManager getDistribution();

  /**
   * Returns true if this node is a leaf.
   *
   * @return boolean
   */
  public boolean isLeaf();

  /**
   * Returns true if the node is the root
   *
   * @return boolean
   */
  public boolean isRoot();

  /**
   * Adds a new child to this node. Adds the child to the end.
   *
   * @param node SimpleNodeManager
   */
  public void addChild(SimpleNodeManager node);

  /**
   * Cut a branch from this node.
   *
   * @param child_index int
   * @throws TreeModelException if the index is out of bound
   */
 // public void pruneChild(int child_index) throws TreeModelException;

  /**
   * Remove all chindren from this node.
   */
 // public void removeAllChildren();

  /**
   * Returns the i-esim child of this node.
   * Throws an exception if index is out of bound or the node do not have children.
   *
   * @param index int
   * @throws TreeModelException
   * @return SimpleNodeManager
   */
  public SimpleNodeManager getChild(int index) throws TreeModelException;

  /**
   * Sets the recond count value to this node.
   *
   * @param record_count double
   */
  public void setRecordCount(double record_count);

  /**
   * Returns the depth of this node where the leaves have depth equal to 0.
   *
   * @return int a positive value.
   */
  public int getDepth();

  /**
   * Returns the number of leaves for this node.
   *
   * @return int
   */
  public int getNumberOfLeaves();

  /**
   * Returns the list of leaves for this node.
   * Returns an empty vector if the node is a leaf.
   *
   * @return SimpleNodeManager[]
   */
  public java.util.Vector getLeaves();
}
