/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import weka.core.Instance;
import java.util.Vector;
import java.util.StringTokenizer;
import kddml.Core.XMLUtility;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * A set operator checks whether a field value is element of a set.
 * The attribute associated with this object can take one of following boolean operators:
 * <ol>
 * <li> 'Is In' which becomes true if the value of the attribute is
 * one of the specified set. (e.g., color IN { red, green })</li>
 * <li> 'Is Not In' which becomes true if the value of the attribute is
 * not one of the specified set. (e.g., color NOT IN { white, black })</li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SimpleSetPredicate extends SimplePredicate {

  /**
   * The values belonging to the set.
   */
  protected Vector values;

  /**
   * The set operator.
   */
  protected SetOperator op_code;

  /**
   * Constructor.
   *
   * @param field String the field to check.
   * @param val String[] the list of values belonging to the set.
   * @param op_code SetOperator the operator to use for checking.
   */
  public SimpleSetPredicate(String field, String[] val, SetOperator op_code) {
    this.field = field;
    values = new Vector();
    for (int i=0; i<val.length; i++) {
      values.add(val[i]);
    }
    this.op_code = op_code;
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param elem Element
   * @throws TreeModelException
   */
  protected SimpleSetPredicate(Element elem)
      throws TreeModelException {

    if (!elem.getTagName().equalsIgnoreCase("SimpleSetPredicate")) {
      String msg = "Invalid XML element. No SimpleSetPredicate input!";
      throw new TreeModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }

    Element array = (Element) XMLUtility.getFirstChild(elem);
    int n = new Integer(array.getAttribute("n")).intValue();
    String v = array.getChildNodes().item(0).getNodeValue();
    values = new Vector();
    StringTokenizer tokeniz = new StringTokenizer(v);
    while(tokeniz.hasMoreElements())
      values.add(tokeniz.nextToken());
    field = XMLUtility.getAttribute(elem, "field");
    String op = elem.getAttribute("booleanOperator");
    if (op.equals("isIn"))
      op_code = SetOperator.IN;
    if (op.equals("isNotIn"))
      op_code = SetOperator.NOT_IN;
  }

  /**
   * Evaluates this predicate on a single instance.
   *
   * @param instance Object
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  public PredicateEvaluation evaluate(Object instance)
      throws TreeModelException {
    Instance tupla = (Instance) instance;

    // pos indica la posizione del campo field all'interno dell'istanza...
    int pos = -1;
    for(int i=0; i<tupla.numAttributes(); i++) {
      if (tupla.attribute(i).name().equalsIgnoreCase("field"))
        pos = i;
    }
    if (pos == -1) {
      String msg = "Attribute < " + field + " > do not exist!";
      throw new TreeModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }


    // check for missing attribute...
    if (tupla.isMissing(pos))
      return PredicateEvaluation.UNKNOWN;

    if (op_code.equals(SetOperator.IN))
      return evaluateIN(tupla.toString(pos));
    return evaluateNOT_IN(tupla.toString(pos));
  }

  /**
   * Evaluates IS IN operator.
   *
   * @param value String
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  private PredicateEvaluation evaluateIN(String value)
      throws TreeModelException {
    if (values.contains(value))
      return PredicateEvaluation.TRUE;
    else
      return PredicateEvaluation.FALSE;
  }

  /**
   * Evaluates IS NOT IN operator.
   *
   * @param value String
   * @throws TreeModelException
   * @return PredicateEvaluation
   */
  private PredicateEvaluation evaluateNOT_IN(String value)
      throws TreeModelException {
    if (!values.contains(value))
      return PredicateEvaluation.TRUE;
    else
      return PredicateEvaluation.FALSE;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    return op_code.toString() + " " + values.toString();
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {

    Element ris = XMLUtility.doc.createElement("SimpleSetPredicate");
    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n",new Integer(values.size()).toString());
    array.setAttribute("type","string");
    String list="";
    for(int i=0; i<values.size();i++) {
      list = list+ "\""+(String) values.get(i)+"\""+" ";
    }
    Node n = XMLUtility.doc.createTextNode(list);
    array.appendChild(n);
    ris.appendChild(array);
    XMLUtility.setAttribute(ris, "field", field);
    if (op_code.equals(SetOperator.IN))
      ris.setAttribute("booleanOperator", "isIn");
    if (op_code.equals(SetOperator.NOT_IN))
      ris.setAttribute("booleanOperator", "isNotIn");
    return ris;
  }
}
