/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;
import kddml.Core.*;
import weka.core.Instance;
import java.io.File;

/**
 * A tree model in data mining is used to predict the class of a case with unknown
 * categorical label.
 * A classification tree is a flow-chart-like structure, where each internal node denotes
 * a test on an attribute, each branch represents an outcome of the test, and the
 * leaf nodes represent classes or class distributions. The top-most node in a tree is the
 * root node.<br>
 * In order to classify an unknown sample, the attribute values of the sample are
 * tested against the decision tree. A path is traced from the root to a leaf node that
 * holds the class prediction for that sample.<br>
 * Given a pre-determined set of classes in the target attribute, classification analyzes
 * the build data to determine to which class a given case belongs.<br>
 * A decision tree is a classification tree in witch the target attribute is binary.
 * A tree model consists of a reference to the Node root. Each Node holds a logical predicate
 * expression that defines the rule for choosing the Node or any of the branching Nodes.<br>
 * Optionally, a classification model can have two confusion matrixes related to the training
 * and test instances resplectively. It depends to the algorithm used to build the model.
 * A confusion matrix is a two-dimensional table that indicates the number of correct and
 * incorrect predictions a tree model made on specific data.
 * It provides a measure of accuracy of the model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class TreeModel extends kddml.Core.DataMining.MiningModel
    implements TreeModelManager  {

  /**
   * The training confusion matrix. Can be null.
   */
  protected ConfusionMatrix training_matrix;

  /**
   * The test confusion matrix. Can be null.
   */
  protected ConfusionMatrix test_matrix;

  /**
   * The root of the tree.
   */
  protected Node root;

  /**
   * Indicates whether the tree model has exactly two children per node, or multiple
   * childrens per node.
   * In the case of binary_split = false, it means that each node may have 2 or more
   * child nodes.
   */
  protected boolean binary_split;

  /**
   * Package proprietary constructor. It creates an empty classification model.
   *
   * @param physical_path File the repository physical path containing the model.
   */
  protected TreeModel(java.io.File physical_path) {

    super(physical_path);
    root = null;
    training_matrix = null;
    test_matrix = null;
    binary_split = false;
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return kddml.Core.KDDMLObjectType.PMML_CLASSIFICATION_MODEL;
  }

  /**
   * Returns true if the model is empty.
   * For a tree model, this method returns false.
   *
   * @return boolean
   */
  public boolean isEmpty() {
    return false;
  }

  /**
   * Save the object in the system repository. The destination path is provided by
   * the object_path variable.
   *
   * @throws TreeModelException
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws kddml.Core.KDDMLCoreException {

    Document root = initializeDocument();
    Element tree_model;

    tree_model = XMLUtility.doc.createElement("TreeModel");
    if (mining_schema != null)
      tree_model = appendMiningSchema((MiningSchema) mining_schema, tree_model);

    // attribute settings...
    tree_model.setAttribute("functionName", "classification");
    if (algorithm_name != null)
      XMLUtility.setAttribute(tree_model, "algorithmName", algorithm_name);
    if (model_name != null)
      XMLUtility.setAttribute(tree_model, "modelName", model_name);

    if (binary_split)
      tree_model.setAttribute("splitCharacteristic","binarySplit");
    else
      tree_model.setAttribute("splitCharacteristic","multiSplit");

    root.getDocumentElement().appendChild(tree_model);

    // appende l'albero
    tree_model.appendChild(this.root.toXML());

    // appende la matrice di confusione se esiste
    Element extension = null;
    if (training_matrix != null) {
      extension = XMLUtility.doc.createElement("Extension");
      tree_model.appendChild(extension);
      extension.appendChild(training_matrix.toXML());
    }
    // appende la matrice di confusione se esiste
    if (test_matrix != null) {
      tree_model.appendChild(extension);
      extension.appendChild(test_matrix.toXML());
    }

    if (extension != null)
      tree_model.appendChild(extension);

    XMLUtility.toFile(object_path.getAbsolutePath(), root);
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = super.toString();
    s = s + "---------------------------------\n";
    s = s + "Tree Model:\n";
    s = s + "---------------------------------\n";
    s = s + root.toString() + "\n";
    s = s + "---------------------------------\n";
    if (training_matrix != null)
      s = s + training_matrix.toString();
    if (test_matrix != null)
      s = s + test_matrix.toString();
    return s;
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML() throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.PMML_CLASSIFICATION_MODEL);

    String s1 = getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    try {
      XMLUtility.formatting(s1, s2, s3, null, null);
    }
    catch(Exception e) {
      String msg = "Exception found during the HTML code generation.";
      msg = e.toString();
      throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
    }
  }

  /**
   * Returns true if the target attribute is binary (i.e. with exactly two category
   * values). Throws an exception if the target attribute connot be found in
   * the mining schema and data dictionary of the model.
   *
   * @throws KDDMLCoreException
   * @return boolean
   */
  public boolean isDecisionTree() throws KDDMLCoreException {
    String[] s = getTargetAttributeValues();
    return (s.length == 2);
  }

  /**
   * Returns the root node for this classification tree.
   *
   * @return NodeManager
   */
  public NodeManager getRoot() {
    return root;
  }

  /**
   * Returns the training confusion matrix related to the model.
   * Returns null if the model do not have a confusion matrix associated.
   *
   * @return ConfusionMatrixManager the training confusion matrix or null.
   */
  public ConfusionMatrixManager getTrainingConfusionMatrix() {
    return training_matrix;
  }

  /**
   * Returns the test confusion matrix related to the model.
   * Returns null if the model do not have a confusion matrix associated.
   *
   * @return ConfusionMatrixManager the test confusion matrix or null.
   */
  public ConfusionMatrixManager getTestConfusionMatrix() {
    return test_matrix;
  }

  /**
   * Sets the training confusion matrix to the model.
   *
   * @param cm ConfusionMatrixManager
   */
  public void setTrainingConfusionMatrix(ConfusionMatrixManager cm) {
    training_matrix = (ConfusionMatrix) cm;
  }

  /**
   * Sets the test confusion matrix to the model.
   *
   * @param cm ConfusionMatrixManager
   */
  public void setTestConfusionMatrix(ConfusionMatrixManager cm) {
    test_matrix = (ConfusionMatrix) cm;
  }

  /**
   * Indicates whether the tree model has exactly two children per node, or multiple
   * childrens per node.
   * In the case of binary_split = false, it means that each node may have 2 or more
   * child nodes.
   *
   * @return boolean true if the model has exactly two clindren per node. False otherwise.
   */
  public boolean isBinarySplitting() {
    return binary_split;
  }

  /**
   * Returns the name of the target attribute.
   * Throws an exception if the target attribute connot be found in the mining schema.
   *
   * @throws KDDMLCoreException
   * @return String
   */
  public String getTargetAttributeName() throws KDDMLCoreException {
    MiningField f;
    java.util.Iterator iter = mining_schema.getFields();
    // cerca l'attributo con type = PREDICTED
    while(iter.hasNext()) {
      f = (MiningField) iter.next();
      if (f.getType() == MiningField.PREDICTED)
        return f.getName();
    }
    throw new KDDMLCoreException("Target attribute cannot be found in mining schema!",
                                 null,
                                 KDDMLCoreErrorCodes.INVALID_MODEL);
  }

  /**
   * Returns the categories associated to the target attributes.
   * Throws an exception if the target attribute connot be found in the mining schema.
   *
   * @throws KDDMLCoreException
   * @return String[]
   */
  public String[] getTargetAttributeValues() throws KDDMLCoreException {

    String name = getTargetAttributeName();
    // cerca nel DataDictionary l'attributo di classificazione
    DataField d;
    java.util.Iterator iter = data_dictionary.getFields();
    while(iter.hasNext()) {
      d = (DataField) iter.next();
      if (d.getName().equalsIgnoreCase(name))
        return d.getEnumeratedValues();
    }
    throw new KDDMLCoreException("Target attribute cannot be found in data dictionary!",
                                 null,
                                 KDDMLCoreErrorCodes.INVALID_MODEL);

  }

  /**
   * Method used to classify an input instance.
   * Returns the classification value. Cannot be null.
   * Throws an exception is an error occours.
   *
   * @param instance Object the instance as in weka.Core.Instance.
   * @throws KDDMLCoreException
   * @return String the classification balue. Cannot be null.
   */
  public String classify(Object instance) throws KDDMLCoreException {
    return classify(root, (Instance) instance);
  }

  /**
   * Returns the prediction target value as string for the input node.
   *
   * @param node NodeManager
   * @param instance Instance
   * @throws KDDMLCoreException
   * @return String
   */
  protected String classify(NodeManager node, Instance instance) throws KDDMLCoreException {

    if (!node.isCompoundTree()) {
      // simple node...
      ScoreDistribution distr = getDistribution((SimpleNode) node, (Instance) instance);
      //System.out.println(instance.toString()+ " "+ distr.toString());
      return distr.getHighestCategory();
    }
    else {
      // compound tree...
      NodeManager[] compounds = ((CompoundTree) node).getTrees();
      // valori di classificazione per ogni singolo figlio
      String[] values = new String[compounds.length];
      for (int i=0; i<compounds.length; i++) {

        // recursion
        values[i] = classify(compounds[i], instance);
      }

      return ((CompoundTree) node).classify(values);
    }
  }

  /**
   * Returns the distribution related to a node for a given input instance.
   *
   * @param node SimpleNode
   * @param instance Instance
   * @throws KDDMLCoreException
   * @return ScoreDistribution
   */
  protected ScoreDistribution getDistribution(SimpleNode node, Instance instance)
      throws KDDMLCoreException {

    if (node.isLeaf()) {

      return node.distribution;
    }

    // iterate on children...

    // evaluation contains the result of evaluation predicate for each child
    PredicateEvaluation[] evaluation = new PredicateEvaluation[node.children.size()];

    SimpleNode tn = null;
    int num_false = 0;
    for (int i=0; i<node.children.size(); i++) {

      tn = (SimpleNode) node.children.get(i);
      Predicate p = tn.getPredicate();

      evaluation[i] = p.evaluate(instance);

      //System.out.println(p.toString()+" --> "+evaluation[i]);

      if (evaluation[i].equals(PredicateEvaluation.TRUE)) {
        // if exist at least a predicate returning true...
        // STOP EVALUATION: recursive evaluation of the i-esim branch
        return getDistribution(tn, instance);
      }

      if (evaluation[i].equals(PredicateEvaluation.FALSE))
        num_false++;
    }

    String[] categories = getTargetAttributeValues();
    ScoreDistribution cloned = new ScoreDistribution(categories);

    if (num_false == evaluation.length) {
      // ALL THE PREDICATES RETURN FALSE --> evaluation of all predicates
      for (int j=0; j<evaluation.length; j++) {

        // evaluate the j-esim child
        ScoreDistribution distr2 = getDistribution((SimpleNode) node.children.get(j), instance);

        // update the output distribution
        for (int k=0; k<categories.length;k++) {
          cloned.updateDistribution(categories[k], distr2.getCount(categories[k]));
        }
      }
    }
    else {
      // SOME PREDICATES RETURN UNKNOWN --> evaluate only this predicates
      for (int j = 0; j < evaluation.length; j++) {

        if (evaluation[j].equals(PredicateEvaluation.UNKNOWN)) {

          // evaluate the j-esim child
          ScoreDistribution distr2 = getDistribution( (SimpleNode) node.
              children.get(j), instance);

          // update the output distribution
          for (int k = 0; k < categories.length; k++) {
            cloned.updateDistribution(categories[k], distr2.getCount(categories[k]));
          }
        }
      }
    }

    return cloned;
  }
}
