/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.ClassificationTrees;

import kddml.Core.KDDMLCoreException;

/**
 * A manager interface for classification trees.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface TreeModelManager extends kddml.Core.DataMining.MiningModelManager {

  /**
   * Method used to classify an input instance.
   * Returns the classification value. Cannot be null.
   * Throws an exception is an error occours.
   *
   * @param instance Object a generic input instance.
   * @throws KDDMLCoreException
   * @return String the classification balue. Cannot be null.
   */
  public String classify(Object instance) throws KDDMLCoreException;

  /**
   * Returns true if the target attribute is binary (i.e. with exactly two category
   * values). Throws an exception if the target attribute connot be found in
   * the mining schema and data dictionary of the model.
   *
   * @throws KDDMLCoreException
   * @return boolean
   */
  public boolean isDecisionTree() throws KDDMLCoreException;

  /**
   * Returns the root node for this classification tree.
   *
   * @return NodeManager
   */
  public NodeManager getRoot();

  /**
   * Returns the training confusion matrix related to the model.
   * Returns null if the model do not have a confusion matrix associated.
   *
   * @return ConfusionMatrixManager the training confusion matrix or null.
   */
  public ConfusionMatrixManager getTrainingConfusionMatrix();

  /**
   * Returns the test confusion matrix related to the model.
   * Returns null if the model do not have a confusion matrix associated.
   *
   * @return ConfusionMatrixManager the test confusion matrix or null.
   */
  public ConfusionMatrixManager getTestConfusionMatrix();

  /**
   * Sets the training confusion matrix to the model.
   *
   * @param cm ConfusionMatrixManager
   */
  public void setTrainingConfusionMatrix(ConfusionMatrixManager cm);

  /**
   * Sets the test confusion matrix to the model.
   *
   * @param cm ConfusionMatrixManager
   */
  public void setTestConfusionMatrix(ConfusionMatrixManager cm);

  /**
   * Indicates whether the tree model has exactly two children per node, or multiple
   * childrens per node.
   * In the case of binary_split = false, it means that each node may have 2 or more
   * child nodes.
   *
   * @return boolean true if the model has exactly two clindren per node. False otherwise.
   */
  public boolean isBinarySplitting();

  /**
   * Returns the name of the target attribute.
   * Throws an exception if the target attribute connot be found in the mining schema.
   *
   * @throws KDDMLCoreException
   * @return String
   */
  public String getTargetAttributeName() throws KDDMLCoreException;

  /**
   * Returns the categories associated to the target attributes.
   * Throws an exception if the target attribute connot be found in the mining schema.
   *
   * @throws KDDMLCoreException
   * @return String[]
   */
  public String[] getTargetAttributeValues() throws KDDMLCoreException;
}
