/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.KDDMLCoreErrorCodes;

/**
 * Given two field values x and y, the absolute difference is defined as: <br>
 * <p align="center">
 * c(x,y)=|x-y|.
 * </p><br>
 * The absolute difference is an inner function, used to compare two single field
 * values.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ABS_DIFF extends AttributeComparisonMeasure {

  /**
   * Computes the comparison measure for a numeric attribute.
   *
   * @param value1 double the centroid coordinate for this attribute.
   * Cannot be null.
   * @param value2 Double the attribute value to compare. Can be null for missing
   * attributes.
   * @return double the absolute difference between value1 and value2.
   */
  protected double computeNumeric(double value1, Double value2) {
    if (value2 == null) {
      // value missing
      if (value1 < 0.5) {
        return 1.0 - value1;
      }
      return value1;
    }

    // no missing value
    return Math.abs(value1 - value2.doubleValue());
  }

  /**
   * Throws an exception: absolute difference is applied to numeric attributes only.
   *
   * @param value1 String
   * @param value2 String
   * @throws ClusteringModelException
   * @return double
   */
  protected double computeDiscrete(String value1, String value2)
      throws ClusteringModelException {

    String msg = "Can't compute absolute difference for discrete attribute!";
    throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_MODEL);
  }

  /**
   * Returns the function name.
   *
   * @return String
   */
  public String getFunctionName() {
    return "absDiff";
  }
}
