/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

/**
 * The tanimoto function is an outer function, used to compare two binary or
 * categorical records X, Y where:
 * <ul>
 * <li> a11 is the number of times where Xi=1 and Yi=1</li>
 * <li> a10 is the number of times where Xi=1 and Yi=0</li>
 * <li> a01 is the number of times where Xi=0 and Yi=1</li>
 * <li> a00 is the number of times where Xi=0 and Yi=0</li>
 * <li> cij and dij parameters > 0</li>
 * </ul>
 * The tanimoto function is defined as:
 * <br>
 * <p align="left">
 *  &nbsp;&nbsp;&nbsp;    c11*a11 + c10*a10 + c01*a01 + c00*a00 <br>
 *  D = -------------------------------------------------------- <br>
 *  &nbsp;&nbsp;&nbsp;    d11*a11 + d10*a10 + d01*a01 + d00*a00 <br>
 * </p><br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */

public class BINARY_SIMILARITY extends CentroidBasedComparisonMeasure {

  private double c00;
  private double c01;
  private double c10;
  private double c11;

  private double d00;
  private double d01;
  private double d10;
  private double d11;

  /**
   * Constructor given the list of comparison measures. Used for centroid-based
   * clustering only.
   *
   * @param attribute_comparison AttributeComparisonMeasure[] the comparison measure
   * for each attribute.
   * @param weights double[] the field weight for each attribute. Can be null.
   * @param c00_param double
   * @param c01_param double
   * @param c10_param double
   * @param c11_param double
   * @param d00_param double
   * @param d01_param double
   * @param d10_param double
   * @param d11_param double
   */
  public BINARY_SIMILARITY(AttributeComparisonMeasure[] attribute_comparison, double[] weights,
                           double c00_param, double c01_param, double c10_param, double c11_param,
                           double d00_param, double d01_param, double d10_param, double d11_param) {

    super(attribute_comparison, weights);
    c00 = c00_param;
    c01 = c01_param;
    c10 = c10_param;
    c11 = c11_param;

    d00 = d00_param;
    d01 = d01_param;
    d10 = d10_param;
    d11 = d11_param;

  }

  /**
   * Constructor given the centroid description. Used both for centroid-based clustering
   * and distribution-based clustering.
   *
   * @param cluster_description ClusterDescription
   * @param c00_param double
   * @param c01_param double
   * @param c10_param double
   * @param c11_param double
   * @param d00_param double
   * @param d01_param double
   * @param d10_param double
   * @param d11_param double
   * @throws ClusteringModelException
   */
  public BINARY_SIMILARITY(ClusterDescription cluster_description, double c00_param, double c01_param,
                           double c10_param, double c11_param, double d00_param, double d01_param,
                           double d10_param, double d11_param)
      throws ClusteringModelException {

    super(cluster_description);
    c00 = c00_param;
    c01 = c01_param;
    c10 = c10_param;
    c11 = c11_param;

    d00 = d00_param;
    d01 = d01_param;
    d10 = d10_param;
    d11 = d11_param;

  }

  /**
   * Returns true if the comparison measure is a similarity function, in which the
   * value returned by the compare method is optimal for greater values.
   * Returns false if the comparison measure is a distance measure, in which the
   * value returned by the compare method is optimal if it is 0.
   *
   * @return boolean true
   */
  public boolean isSimilarityMeasure() {
    return true;
  }

  /**
   * Evaluates the comparison measure given the comparison values for each single
   * attribute (i.e., the values returned by each single inner function).
   *
   * @param x double[]
   * @return double 0
   */
  protected double evaluate(double[] x) {
    return 0;
  }

  public double compare(Cluster cluster, weka.core.Instance instance)
      throws ClusteringModelException {

    int a11 = 0;
    int a10 = 0;
    int a01 = 0;
    int a00 = 0;

    // get the centroid from the cluster...
    double[] x = cluster.getSeedCoordinate();

    // get the coordinate instance...
    double[] y = new double[x.length];
    for (int i=0; i<instance.numAttributes(); i++) {
      if (instance.attribute(i).isNumeric()) {
        // numeric attribute --> error
        String msg = "Can't compute absolute binary similarity for numeric attribute "+instance.attribute(i).name();
        throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
      }

      String value;
      if (instance.isMissing(i))
        // get array of 0
        value = "?";
      else
        value = instance.stringValue(i);
      double[] d = cluster.normalizeFieldValue(instance.attribute(i).name(), value);
      for (int j=0; j<d.length; j++)
        y[i+j] = d[j];
    }

    for (int i=0; i<x.length; i++) {

      // discrete attribute
      if (!instance.isMissing(i)) {
        if ((x[i] == 1) && (y[i] == 1))
          a11++;
        if ((x[i] == 1) && (y[i] == 0))
          a10++;
        if ((x[i] == 0) && (y[i] == 1))
          a01++;
        if ((x[i] == 0) && (y[i] == 0))
          a00++;

      }
    }
    double d1 = c11*a11 + c10*a10 + c01*a01 + c00*a00;
    double d2 = d11*a11 + d10*a10 + d01*a01 + d00*a00;

    if (d2 == 0)
      return 0;
    else
      return d1 / d2;

  }

  /**
   * Returns the function name as in PMML.
   *
   * @return String
   */
  public String getFunctionName() {
    return "binarySimilarity";
  }
}


