/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

/**
 * The chebychev function is an outer function, used to compare two records X, Y,
 * where:
 * <ul>
 * <li> Xi is the value for the i-esim field of X</li>
 * <li> Yi is the value for the i-esim field of X</li>
 * <li> Wi is the weight for the i-esim field of X (default 1)</li>
 * <li> a p-parameter > 0</li>
 * </ul>
 * The chebychev distance is defined as:
 * <br>
 * <p align="center">
 * D = max (Wi*c(Xi,Yi)) over all i.
 * </p><br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */

public class CHEBYCHEV extends CentroidBasedComparisonMeasure {

  /**
   * Constructor given the list of comparison measures. Used for centroid-based
   * clustering only.
   *
   * @param attribute_comparison AttributeComparisonMeasure[] the comparison measure
   * for each attribute.
   * @param weights double[] the field weight for each attribute. Can be null.
   */
  public CHEBYCHEV(AttributeComparisonMeasure[] attribute_comparison, double[] weights) {
    super(attribute_comparison, weights);
  }

  /**
   * Constructor given the centroid description. Used both for centroid-based clustering
   * and distribution-based clustering.
   *
   * @param cluster_description ClusterDescription
   */
  public CHEBYCHEV(ClusterDescription cluster_description) {
    super(cluster_description);
  }

  /**
   * Returns true if the comparison measure is a similarity function, in which the
   * value returned by the compare method is optimal for greater values.
   * Returns false if the comparison measure is a distance measure, in which the
   * value returned by the compare method is optimal if it is 0.
   *
   * @return boolean false
   */
  public boolean isSimilarityMeasure() {
    return false;
  }

  /**
   * Returns the function name as in PMML.
   *
   * @return String
   */
  public String getFunctionName() {
    return "chebychev";
  }

  /**
   * Evaluates the comparison measure given the comparison values for each single
   * attribute (i.e., the values returned by each single inner function).
   *
   * @param x double[]
   * @return double
   */
  protected double evaluate(double[] x) {
    double ris = x[0];
    for (int i=1; i<x.length; i++) {

      double d1 = x[i];
      double d2 = d1 * w[i];

      ris = Math.max(ris, d2);
    }
    return ris;

  }
}
