/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import java.util.Enumeration;
import org.w3c.dom.*;
import kddml.Core.XMLUtility;
import weka.core.*;

/**
 * Center-based (or distance-based) cluster need a distance or similarity measurement
 * based on which they try to group those most similar objects into one cluster.
 * Example: K-means, CLARANS.
 * A center-based cluster is always defined by their centroid. Statistical information
 * are optional.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CentroidBasedCluster extends Cluster implements CentroidBasedClusterManager {

  /**
   * The vector of center coordinates in respect on the ClusteringDescription.
   */
  protected double[] centroid;

  /**
   * Constructor given the cluster name and the centroid coordinate.
   *
   * @param name String
   * @param centroid double[] the center coordinate in respect on the ClusteringDescription
   * fields.
   */
  public CentroidBasedCluster(String name, double[] centroid) {
    super(name);
    this.centroid = centroid;
    cluster_distribution = null;
  }

  /**
   * Returns true if the clustering is distribution-based.
   *
   * @return boolean false
   */
  public boolean isDistributionBased() {
    return false;
  }

  /**
   * Returns true if the clustering is center-based.
   *
   * @return boolean true
   */
  public boolean isCentroidBased() {
    return true;
  }

  /**
   * Returns the representative cluster seed as instance.
   * Returns the centroid as instance.
   *
   * @throws ClusteringModelException
   * @return Object a single instance as in weka.core.Instance.
   */
  public Object getSeedAsInstance() throws ClusteringModelException {

    // get the schema...
    Instances schema = (Instances) centroid_description.toInstances();

    // get the centroid as instance
    Instance instance = new Instance(centroid_description.getNumberOfFields());
    //instance.setDataset(ris);

    instance.setDataset(schema);
    Enumeration enum2 = centroid_description.getFields();
    ClusteringField field;
    int centroid_index = 0;
    int instance_index = 0;
    while(enum2.hasMoreElements()) {
      field = (ClusteringField) enum2.nextElement();
      if (field.isContinuous()) {
        // numeric field - denormalize the probability value
        double d = ((ContinuousClusteringField) field).denormalize(centroid[centroid_index]);
        instance.setValue(instance_index, d);
        centroid_index++;
        instance_index++;
      }
      else {
        // discrete field - get the probabily category
        String[] categories = ((DiscreteClusteringField) field).getCategories();
        double[] values = new double[categories.length];
        for (int j=0; j<values.length; j++) {
          values[j] = centroid[centroid_index + j];
        }
        // get the max index of the distribution
        int max = Utils.getMaxIndex(values);
        instance.setValue(instance_index, categories[max]);
        centroid_index = centroid_index + categories.length;
        instance_index++;
      }
    }

    return instance;
  }

  /**
   * Returns the centroid coordinate as array of normalized doubles in respect on the
   * ClusterDescription fields.
   *
   * @throws ClusteringModelException
   * @return double[]
   */
  public double[] getSeedCoordinate() {
    return centroid;
  }

  /**
   * Returns the centroid point of the specified attribute for the cluster.
   * Returns null if the cluster has missing values for all cases in the cluster,
   * or if this model does not support centroids.
   * If invoked on a categorical attribute, throws an exception.
   */
  //public Double getCentroidCoordinate(String numericalAttributeName)
  //    throws ClusteringModelException {

  //  return null;
  //}

  /**
   * Returns the centroid point of the specified attribute for the cluster.
   * Returns null if the cluster has missing values for all cases in the cluster,
   * or if this model does not support centroids.
   * If a numerical attribute is provided, throws an exception.
   * Result is greater than or equal to 0, or null if all values in attribute are null.
   */
 // public Double getCentroidCoordinate(String categoricalAttributeName, String category)
  //    throws ClusteringModelException {

  ///  return null;
  //}

  /**
   * Returns a representation of this cluster as string.
   *
   * @return String
   */
  public String toString() {
    String ris = super.toString();
    try {
      ris = ris + "Centroid As Instance: " + ( (Instance) getSeedAsInstance()).toString() + "\n";
    }
    catch (ClusteringModelException ex) {
    }
    ris = ris + "Centroid As Normalization Vector: ";
    for (int i=0; i<centroid.length; i++)
      ris = ris + Utils.doubleToString(centroid[i], 3) +", ";
    ris = ris+"\n";
    if (cluster_distribution != null)
      ris = ris + cluster_distribution.toString();
    return ris;
  }

  /**
   * Returns a representation of this cluster as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {

    Element cluster = super.toXML();

    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n", new Integer(centroid.length).toString());
    array.setAttribute("type","real");
    String list="";
    for(int i=0; i<centroid.length;i++) {
      list = list + centroid[i] + " ";
    }
    Node n = XMLUtility.doc.createTextNode(list);

    array.appendChild(n);
    cluster.appendChild(array);
    if (cluster_distribution != null) {
      Element statistic = cluster_distribution.toXML();
      cluster.appendChild(statistic);
    }

    return cluster;
  }
}
