/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import weka.core.Instance;
import java.util.Vector;
import kddml.Core.XMLUtility;
import org.w3c.dom.Element;

/**
 * A comparison measure centroid-based.
 * A centroid-based comparison measure can be used either for a centroid-based clustering and
 * a distribution-based clustering.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class CentroidBasedComparisonMeasure extends ComparisonMeasure {

  /**
   * The comparison measure for each attribute.
   */
  AttributeComparisonMeasure[] attribute_comparison;

  /**
   * The field weight for each attribute.
   */
  double[] w;

  /**
   * Constructor given the list of comparison measures. Used for centroid-based
   * clustering only.
   *
   * @param attribute_comparison AttributeComparisonMeasure[] the comparison measure
   * for each attribute.
   * @param weights double[] the field weight for each attribute. Can be null.
   */
  public CentroidBasedComparisonMeasure(AttributeComparisonMeasure[] attribute_comparison,
                                        double[] weights) {
    this.attribute_comparison = attribute_comparison;
    if (weights != null)
      this.w = weights;
    else {
      weights = new double[attribute_comparison.length];
      for (int i=0; i<weights.length; i++)
        weights[i] = 1.0;
    }
  }

  /**
   * Constructor given the centroid description. Used both for centroid-based clustering
   * and distribution-based clustering.
   *
   * @param cluster_description ClusterDescription
   */
  public CentroidBasedComparisonMeasure(ClusterDescription cluster_description) {

    Vector v = cluster_description.fields;
    attribute_comparison = new AttributeComparisonMeasure[v.size()];
    w = new double[v.size()];
    for (int i=0; i<v.size(); i++) {
      ClusteringField field = (ClusteringField) v.get(i);
      attribute_comparison[i] = field.function;
      if (field.getFieldWeight() != null)
        w[i] = field.getFieldWeight().doubleValue();
      else
        w[i] = 1.0;
    }
  }

  /**
   * Returns true if the comparison measure is a similarity function, in which the
   * value returned by the compare method is optimal for greater values.
   * Returns false if the comparison measure is a distance measure, in which the
   * value returned by the compare method is optimal if it is 0.
   *
   * @return boolean
   */
  public abstract boolean isSimilarityMeasure();

  /**
   * Compares an input istance with the seed of the cluster.
   * The seed depends on the type of clustering. If the clustering is center-based,
   * the seed is the centroid as instance.
   * If the clustering is distribution-based, the seed is calculated on the statistics
   * associated to the cluster. In particular, for numeric attributes, the mean of
   * cluster instances is used as seed. For discrete attribute, the most probable
   * category is reported as seed.
   *
   * @param cluster ClusterManager
   * @param instance Instance
   * @throws ClusteringModelException
   * @return double the optimal values depends on the isSimilarityMeasure method.
   */
  public double compare(ClusterManager cluster, Instance instance)
      throws ClusteringModelException {

    // get the centroid from the cluster
    Instance seed = (Instance) cluster.getSeedAsInstance();
    //System.out.println("Normalized seed Instance: "+seed.toString());

    double[] d = new double[seed.numAttributes()];
    for (int i=0; i<seed.numAttributes(); i++) {

      // evaluate the comparison attribute functions
      if (seed.attribute(i).isNumeric()) {
        // numeric attribute
        double seed_value = ((Cluster) cluster).normalizeFieldValue(seed.attribute(i).name(), seed.value(i));
        if (instance.isMissing(i))
          // attribute missing

          // NORMALIZZARE SEED E INSTANCE

          d[i] = attribute_comparison[i].computeNumeric(seed_value, null);
        else {
          double inst_value = ((Cluster) cluster).normalizeFieldValue(seed.attribute(i).name(), instance.value(i));
          d[i] = attribute_comparison[i].computeNumeric(seed_value, new Double(inst_value));
        }
      }

      else {
        // discrete attribute
        if (instance.isMissing(i))
          // attribute missing
          d[i] = attribute_comparison[i].computeDiscrete(seed.stringValue(i), null);
        else
          d[i] = attribute_comparison[i].computeDiscrete(seed.stringValue(i), instance.stringValue(i));
      }
    }
    return evaluate(d);
  }

  /**
   * Evaluates the comparison measure given the comparison values for each single
   * attribute (i.e., the values returned by each single inner function).
   * This method is implemented in sub-classes.
   *
   * @param x double[]
   * @return double
   */
  protected abstract double evaluate(double[] x);

  /**
   * Returns true if the function is related to a centroid-based clustering.
   * Returns true if the function is related to a distribution-based clustering.
   *
   * @return boolean true
   */
  public boolean isCentroidBased() {
    return true;
  }

  /**
   * Returns the comparison measure for each attribute.
   *
   * @return AttributeComparisonMeasure[]
   */
  public AttributeComparisonMeasure[] getAttributeComparisonMeasure() {
    return attribute_comparison;
  }

  /**
   * Returns a representation of this function as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("ComparisonMeasure");

    if (isSimilarityMeasure())
      ris.setAttribute("kind", "similarity");
    else
      ris.setAttribute("kind", "distance");

    Element elem = XMLUtility.doc.createElement(getFunctionName());
    ris.appendChild(elem);
    return ris;
  }
}
