/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.DataMining.*;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * A Cluster object holds the metadata about a cluster discovered by running a clustering
 * algorithm. Clusters are always associated with a ClusteringModel.
 * A cluster is defined by its center vector or by statistics. This information is contained
 * in a ClusterDescription.
 * The definition of a cluster may contain a center vector as well as statistics and the
 * cluster is defined by the centroid (for center-based) od by statistics (for
 * distribution-based).
 * Each cluster is identified by an identifier, that is automatically generated by the
 * constructor.
 * A covariace matrix can be given for this cluster that contain information on overall
 * data distribution,. A covariance matrix stores coordinate-by-coordinate variances
 * (diagonal cells) and covariances (non-diagonal cells).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class Cluster implements ClusterManager {

  /**
   * The centroid description for this cluster. Setted in ClusteringModel.
   * Equal for each cluster.
   */
  ClusterDescription centroid_description;

  /**
   * The unique identifier of the cluster.
   */
  int identifier;

  /**
   * The cluster name.
   */
  String name;

  /**
   * The number of elements. Can be null.
   */
  Integer size;

  /**
   * The cluster statistic. This field is optional.
   */
  ClusterStatistic cluster_distribution;

  /**
   * The covariances matrix used to store variances (diagonal cells) and covariances
   * (non-diagonal cells).
   */
  CategoryMatrix covariances_matrix;

  /**
   * Constructor given the cluster name.
   *
   * @param name String
   */
  public Cluster(String name) {
    this.name = name;
    size = null;
    covariances_matrix = null;
    identifier = ClusteringModel.CLUSTER_IDENTIFIER_COUNTER;
    ClusteringModel.CLUSTER_IDENTIFIER_COUNTER++;
  }

  /**
   * Returns the cluster id.
   *
   * @return int
   */
  public int getClusterId() {
    return identifier;
  }

  /**
   * Returns the name of the cluster designated by the clustering algorithm.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the number of cases in the portion of the training data assigned to the cluster
   * during the model build. This is a non-negative value.
   * It can be null.
   *
   * @return Integer
   */
  public Integer getSize() {
    return size;
  }

  /**
   * Sets the number of cases in the portion of the training data assigned to the cluster
   * during the model build. This must be a non-negative value.
   *
   * @param num_elements int
   */
  public void setSize(int num_elements) {
    size = new Integer(num_elements);
  }

  /**
   * Sets the covariance matrix.
   *
   * @param matrix CategoryMatrix
   */
  public void setCovariancesMatrix(CategoryMatrixManager matrix) {
    this.covariances_matrix = (CategoryMatrix) matrix;
  }

  /**
   * Returns the covariances matrix provided for this cluster.
   * Returns null if no covariances matrix is avaible.
   *
   * @return CategoryMatrixManager the convariance matrix. Can be null.
   */
  public CategoryMatrixManager getCovariancesMatrix() {
    return (CategoryMatrixManager) covariances_matrix;
  }

  /**
   * Returns the cluster distribution. Can be null, for example for center-based
   * clustering.
   *
   * @return ClusterStatisticManager
   */
  public ClusterStatisticManager getClusterStatistic() {
    return cluster_distribution;
  }

  /**
   * Normalizes the input value for a numeric field.
   *
   * @param field_name String
   * @param value double
   * @throws ClusteringModelException
   * @return double
   */
  protected double normalizeFieldValue(String field_name, double value)
      throws ClusteringModelException {

    ClusteringField f = (ClusteringField) centroid_description.getField(field_name);
    if (!f.isContinuous())
      throw new ClusteringModelException("Can't normalize a discrete field!", null,
                                         kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    // numeric field - normalize the value
    return ((ContinuousClusteringField) f).normalize(value);
  }

  /**
   * Normalizes the input value for a discrete field.
   *
   * @param field_name String
   * @param category String
   * @throws ClusteringModelException
   * @return double[]
   */
  protected double[] normalizeFieldValue(String field_name, String category)
      throws ClusteringModelException {

    ClusteringField f = (ClusteringField) centroid_description.getField(field_name);
    if (f.isContinuous())
      throw new ClusteringModelException("Can't normalize a numeric field!", null,
                                         kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    // discrete field - normalize the value
    return ((DiscreteClusteringField) f).getDistribution(category);
  }

  /**
   * Returns a representation of this cluster as string.
   *
   * @return String
   */
  public String toString() {
    String ris;
    ris = "Cluster number: "+identifier + "\n";
    ris += "Name: "+name + "\n";
    if (size != null)
      ris += "Size: "+size.toString();
    ris += "\n";
    return ris;
  }

  /**
   * Returns a representation of this cluster as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("Cluster");
    XMLUtility.setAttribute(ris, "name", name);

    return ris;

  }

  /**
   * Sets the cluster statistics.
   * This field is optional for center-based clustering and it is required for
   * distribution-based clustering.
   *
   * @param statistics ClusterStatisticManager
   */
  public void setClusterStatistic(ClusterStatisticManager statistics) {
    this.cluster_distribution = (ClusterStatistic) statistics;
  }
}
