/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.DataSources.*;
import java.util.Enumeration;
import java.util.Vector;
import kddml.Core.KDDMLCoreErrorCodes;
import weka.core.*;
import org.w3c.dom.*;
import kddml.Core.XMLUtility;

/**
 * A cluster model basically consists of a set of clusters. For each cluster a center
 * vector can be given. In center-based models a cluster is defined by a vector of center
 * coordinates. Some distance measure is used to determine the nearest center, that is
 * the nearest cluster for a given input record.
 * For distribution-based models, the clusters are defined by their statistics.
 * Some similarity measure is used to determine the best matching
 * cluster for a given record. The center vectors then only approximate the clusters.
 * The ClusterDescription contains the fields as used in center vectors.
 * The fields which are used in the center vectors are normalized, in particular this
 * allows to map categorical input fields to numeric values in center vectors.
 * Other features depend on the type of the cluster field.
 * For each discrete field, the list of categories is reported.
 * For each continuous field, the normalization matrix is given.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ClusterDescription implements ClusterDescriptionManager {

  /**
   * Center-based or distribution-based clustering.
   */
  boolean is_centroid_based;

  /**
   * The list of fields ClusteringField.
   */
  protected Vector fields;

  /**
   * Constructor.
   *
   * @param stat DataStatisticsManager data statistics on which build the clustering.
   * @param func AttributeComparisonMeasure[] the attribute comparison measures for each
   * clustering fields. Can be null only for a distribution-based clustering that use
   * a distribution-based comparison measure.
   * @param is_centroid_based boolean
   * @throws ClusteringModelException if the attribute comparison measures and the
   * data statistics do not match.
   */
  public ClusterDescription(DataStatisticsManager stat,
                            AttributeComparisonMeasure[] func,
                            boolean is_centroid_based)
      throws ClusteringModelException {

    fields = new Vector();
    AttributeStatisticManager a;
    this.is_centroid_based = is_centroid_based;

    if ((func != null) && (stat.getNumAttributes() != func.length)) {
      // sollevare errore
      String msg = "Incompatible input comparison functions!";
      throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    AttributeStatisticManager[] attr = stat.getStatistics();
    for (int i=0; i<attr.length; i++) {
      a = attr[i];
      if (a.getType().equals(AttributeType.NOMINAL)) {
        // discrete field
        Enumeration enum2 = ((NominalStatisticManager) a).getValues();
        // create the field...
        DiscreteClusteringField f;
        if (func != null)
          f = new DiscreteClusteringField(a.getName(), func[i], enumeration2Array(enum2));
        else
          f = new DiscreteClusteringField(a.getName(), null, enumeration2Array(enum2));
        fields.add(f);
      }

      if (a.getType().equals(AttributeType.NUMERIC)) {

          ContinuousClusteringField f;
          if (func != null)
            f = new ContinuousClusteringField(a.getName(), func[i], (NumericalStatisticManager) a);
          else
            f = new ContinuousClusteringField(a.getName(), null, (NumericalStatisticManager) a);
          fields.add(f);
      }
      // string fields are ignored
    }
  }

  /**
   * Empty constructor that build a ClusterDescription with no fields.
   *
   * @param is_centroid_based boolean true if the clustering is centroid-based.
   */
  public ClusterDescription(boolean is_centroid_based) {

    fields = new Vector();
    this.is_centroid_based = is_centroid_based;
  }

  /**
   * Returns an enumeration of all fields.
   *
   * @return Enumeration return an enumeration of ClusteringField objects.
   */
  public Enumeration getFields() {
    return fields.elements();
  }

  /**
   * Returns a clustering fields with a specified name.
   * Returns null if the field do not exist.
   *
   * @param name String
   * @return ClusteringFieldManager the clustering field. Can be null.
   */
  public ClusteringFieldManager getField(String name) {
    for (int i=0; i<fields.size(); i++) {
      if (((ClusteringFieldManager) fields.get(i)).getFieldName().equalsIgnoreCase(name))
        return (ClusteringFieldManager) fields.get(i);
    }
    // not found!!!
    return null;
  }

  /**
   * Returns the number of fields.
   *
   * @return int
   */
  public int getNumberOfFields() {
    return fields.size();
  }

  /**
   * Adds a new fields to this enumeration.
   *
   * @param field ClusteringFieldManager
   */
  public void addField(ClusteringFieldManager field) {
    fields.add(field);
  }

  /**
   * Returns true if the clustering is centroid-based.
   * Returns false otherwise.
   *
   * @return boolean
   */
  public boolean isCentroidBased() {
    return is_centroid_based;
  }

  private static String[] enumeration2Array(Enumeration enum2) {
    Vector v = new Vector();
    while(enum2.hasMoreElements()) {
      v.add( (String) enum2.nextElement());
    }
    String[] ris = new String[v.size()];
    for (int i=0; i<v.size(); i++) {
      ris[i] = (String) v.get(i);
    }
    return ris;
  }

  /**
   * Returns a representation of this object as instances.
   *
   * @return Object the schema of instances as in weka.core.Instances.
   */
  protected Object toInstances() {

    FastVector fv = new FastVector();
    ClusteringFieldManager field;
    Attribute attr;
    for (int i=0; i<fields.size(); i++) {
      field = (ClusteringFieldManager) fields.get(i);
      if (field.isContinuous()) {
        // numeric field
        attr = new Attribute(field.getFieldName());
      }
      else {
        // discrete field
        String[] str = ((DiscreteClusteringField) field).getCategories();
        FastVector categories = new FastVector();
        for (int j=0; j<str.length; j++)
          categories.addElement(str[j]);
        attr = new Attribute(field.getFieldName(), categories);
      }
      fv.addElement(attr);
    }
    return new Instances("", fv, 0);
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("CenterFields");
    ClusteringField f;
    for (int i=0; i<fields.size(); i++) {
      f = (ClusteringField) fields.get(i);
      if (f.isContinuous())
        ris.appendChild( ( (ContinuousClusteringField) f).toXML());
      else {
        Element[] elem = ( (DiscreteClusteringField) f).toXML();
        for (int j=0; j<elem.length; j++)
          ris.appendChild(elem[j]);
      }
    }
    return ris;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "Clustering Description\n";
    if (is_centroid_based)
      ris = ris + "Model Class: centroid-based\n";
    else
      ris = ris + "Model Class: distribution-based\n";
    ris = ris + "Fields Description:\n";
    for (int i=0; i<fields.size(); i++) {
      ris = ris + ( (ClusteringField) fields.get(i)).toString() + "\n";
      ris = ris + "--\n";
    }
    return ris;
  }
}
