/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.DataMining.CategoryMatrixManager;

/**
 * A manager interface for Cluster.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface ClusterManager {

  /**
   * Returns the cluster id.
   *
   * @return int
   */
  public int getClusterId();

  /**
   * Returns the name of the cluster designated by the clustering algorithm.
   *
   * @return String
   */
  public String getName();

  /**
   * Returns the number of cases in the portion of the training data assigned to the cluster
   * during the model build. This is a non-negative value.
   * It can be null.
   *
   * @return Integer
   */
  public Integer getSize();

  /**
   * Sets the number of cases in the portion of the training data assigned to the cluster
   * during the model build. This must be a non-negative value.
   *
   * @param num_elements int
   */
  public void setSize(int num_elements);

  /**
   * Sets the covariance matrix.
   *
   * @param matrix CategoryMatrix
   */
  public void setCovariancesMatrix(CategoryMatrixManager matrix);

  /**
   * Returns the covariances matrix provided for this cluster.
   * Returns null if no covariances matrix is avaible.
   *
   * @return CategoryMatrixManager the convariance matrix. Can be null.
   */
  public CategoryMatrixManager getCovariancesMatrix();

  /**
   * Returns the cluster distribution. Can be null, for example for center-based
   * clustering.
   *
   * @return ClusterStatisticManager
   */
  public ClusterStatisticManager getClusterStatistic();

  /**
   * Returns the representative cluster seed as instance.
   * This depends on the type of clustering.
   *
   * @throws ClusteringModelException
   * @return Object
   */
  public abstract Object getSeedAsInstance() throws ClusteringModelException;

  /**
   * Returns the seed coordinate as array of normalized doubles in respect on the
   * ClusterDescription fields.
   *
   * @throws ClusteringModelException
   * @return double[]
   */
  public abstract double[] getSeedCoordinate() throws ClusteringModelException;

  /**
   * Returns true if the clustering is distribution-based.
   *
   * @return boolean
   */
  public boolean isDistributionBased();

  /**
   * Returns true if the clustering is center-based.
   *
   * @return boolean
   */
  public boolean isCentroidBased();

  /**
   * Sets the cluster statistics.
   * This field is optional for center-based clustering and it is required for
   * distribution-based clustering.
   *
   * @param statistics ClusterStatisticManager
   */
  public void setClusterStatistic(ClusterStatisticManager statistics);
}
