/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import java.util.Vector;
import java.util.Iterator;
import org.w3c.dom.Element;
import kddml.Core.*;

/**
 * In center-based models a cluster is defined by a vector of center coordinates.
 * Some distance measure is used to determine the nearest center, that is
 * the nearest cluster for a given input record.
 * For distribution-based models, the clusters are defined by their statistics.
 * However, the definition of a cluster may contain a center vector as well as
 * statistics.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ClusterStatistic implements ClusterStatisticManager {

  /**
   * The cluster name.
   */
  protected String name;

  /**
   * The list of fields as StatisticField.
   */
  protected Vector field_list;

  /**
   * The cluster probability as a real between 0 and 1. This field is implied.
   */
  protected Double cluster_priors;

  /**
   * Empty constructor with no statistic fields.
   *
   * @param cluster_name String the cluster name.
   */
  public ClusterStatistic(String cluster_name) {
    name = cluster_name;
    field_list = new Vector();
    this.cluster_priors = null;
  }

  /**
   * Returns the cluster name.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the prior cluster probablity. Can be null.
   *
   * @return Double a value between 0 and 1 or null.
   */
  public Double getClusterPriors() {
    return cluster_priors;
  }

  /**
   * Sets the cluster priors.
   *
   * @param cluster_priors double a value between 0 and 1.
   */
  public void setClusterPriors(double cluster_priors) {
    this.cluster_priors = new Double(cluster_priors);
  }

  /**
   * Adds a new numeric statistic field to this ClusterDescription.
   *
   * @param field StatisticFieldManager
   */
  public void addStatisticField(StatisticFieldManager field) {
    field_list.add(field);
  }

  /**
   * Returns an iterator on statistic fields.
   *
   * @return Iterator an iterator of StatisticField
   */
  public Iterator getStatisticFields() {
    return field_list.iterator();
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String result = "Distribution. ";
    if (cluster_priors != null)
      result += "Cluster priors: "+Utils.doubleToString(cluster_priors.doubleValue(), 4)+"\n";
    for (int i=0; i<field_list.size(); i++)
      result += ((StatisticField) field_list.get(i)).toString();
    return result;
  }

  /**
   * Returns the number of fields.
   *
   * @return int
   */
  public int getNumberOfFields() {
    return field_list.size();
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("Partition");
    for (int i = 0; i< field_list.size(); i++) {
      Element data_field = ((StatisticField) field_list.get(i)).toXML();
      ris.appendChild(data_field);
    }
    XMLUtility.setAttribute(ris, "name", name);
    //if (cluster_priors != null)
      //ris.setAttribute("x-probability", cluster_priors.toString());
    return ris;

  }

  /**
   * Get the centroid as empty instances.
   *
   * @return Object
   */
 /* protected Object toInstances() {

    FastVector fv = new FastVector();
    StatisticField field;
    Attribute attr;
    for (int i=0; i<field_list.size(); i++) {
      field = (StatisticField) field_list.get(i);
      if (field.isNumeric()) {
        // numeric field
        attr = new Attribute(field.name);
      }
      else {
        // discrete field
        String[] str = ((DiscreteStatisticField) field).values;
        FastVector categories = new FastVector();
        for (int j=0; j<str.length; j++)
          categories.addElement(str[j]);
        attr = new Attribute(field.name, categories);
      }
      fv.addElement(attr);
    }
    return new Instances("", fv, 0);
  }*/

}
