/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

/**
 * This class represents a single clustering field, as used in the ClusterDescription.
 * A ClusteringField consists of four major parts:
 * <ol>
 * <li> the name that refere a mining schema field; </li>
 * <li> a field weight that is the importance factor for the field. This field weight is used
 * in the comparison functions in order to compute the comparison measure.
 * The default value is 1.0; </li>
 * <li> the similarity scale that is the distance such that similarity becomes 0.5. </li>
 * <li> the attribute comparison measure that is a function of taking two field values and
 * a similarityScale to define similarity/distance. It can be null, for example when
 * a distribution-based function is used.</li>
 * </ol>
 * Other features depend on the type of clustering field. For continuous clustering fields
 * a normalization matrix is given. For discrete clustering fields, a list of categories
 * is given.
 *
 * @author Andrea Romei.
 */
public abstract class ClusteringField implements ClusteringFieldManager {

  /**
   * The attribute comparison measure.
   */
  protected AttributeComparisonMeasure function;

  /**
   * The field name as in MiningSchema.
   */
  protected String field_name;

  /**
   * The importance factory of this field.
   */
  protected Double weight;

  /**
   * The similarity scale.
   */
  protected Double similarity_scale;

  /**
   * Constructor with the field name and the attribute comparison measure.
   *
   * @param name String the name of the field as in MiningSchema.
   * @param func AttributeComparisonMeasure the attribute comparison measure. Can be null,
   * for example when a distribution-based function is used.
   */
  public ClusteringField(String name,
                         AttributeComparisonMeasure func) {

    field_name = name;
    function = func;
    weight = null;
    similarity_scale = null;
  }

  /**
   * Returns the field name as in MiningSchema.
   *
   * @return String
   */
  public String getFieldName() {
    return field_name;
  }

  /**
   * Sets the field weight.
   *
   * @param weight double
   */
  public void setFieldWeight(double weight) {
    this.weight = new Double(weight);
  }

  /**
   * Returns the field weight or null if no field weight is used.
   *
   * @return Double can be null.
   */
  public Double getFieldWeight() {
    return weight;
  }

  /**
   * Sets the similarity scale.
   *
   * @param sim double a value between 0 and 1.
   */
  public void setSimilarityScale(double sim) {
    this.similarity_scale = new Double(sim);
  }

  /**
   * Returns the similarity scale of null if no similarity scale is used.
   *
   * @return Double a value between 0 and 1 or null.
   */
  public Double getSimilarityScale() {
    return similarity_scale;
  }

  /**
   * Returns a representation of this field as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "Clustering Field Name: "+field_name+"\n";
    if (function == null)
      ris = ris + "Attribute Comparison Measure: Proprietary function\n";
    else
      ris = ris + "Attribute Comparison Measure: "+function.getFunctionName()+"\n";
    if (weight != null)
      ris = ris + "Field Weight: "+weight.toString()+"\n";
    if (similarity_scale != null)
      ris = ris + "Similarity Scale: "+similarity_scale.toString()+"\n";
    return ris;
  }

  /**
   * Returns the attribute comparison measure or null if no attribute comparison
   * measure is used.
   *
   * @return AttributeComparisonMeasure can be null.
   */
  public AttributeComparisonMeasure getFieldComparisonFunction() {
    return function;
  }
}
