/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.Vector;
import java.util.Iterator;
import weka.core.*;

/**
 * The process of grouping a set of physical object into classes of similar objects is
 * called clustering. A cluster is a collection of data object that are similar to one
 * another within the same cluster and are dissimilar to the objects in other
 * cluster.<br>
 * Clustering methods may be classified into three groups: distance-based,
 * distribution-based (or model-based), density-based methods.<br> Distance-based clustering
 * needs a distance or dissimilarity measurement based on which they try to group those
 * most similar objects into one cluster. K-Means is distance-based partitioning method.<br>
 * Model-based or distribution-based clustering methods assume the data of each cluster
 * conforms to a specific statistical distribution (e.g. the Gaussian distribution)
 * and the whole dataset is a mixture of several distribution models. EM is an example
 * of distribution-based partitioning clustering that do not require the specification
 * of distance measures.<br>
 * Density-based approaches regard a cluster as a dense region of data objects.<br>
 * This class manages both center-based clustering and distribution-based clustering.
 * A ClusteringModel basically consists of a set of clusters. For each cluster a center
 * vector can be given. In center-based models a cluster is defined by a vector of center coordinates.
 * Some distance measure is used to determine the nearest center, that is the nearest cluster for
 * a given input record.
 * For distribution-based models the clusters are defined by their statistics.
 * Some similarity measure is used to determine the best matching cluster for a given record.
 * The center vectors then only approximate the clusters. <br>
 * The model must contain information on the distance or similarity measure used for clustering.
 * It may also contain information on overall data distribution, such as covariance matrix,
 * or other statistics.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ClusteringModel extends kddml.Core.DataMining.MiningModel
    implements ClusteringModelManager {

  protected static int CLUSTER_IDENTIFIER_COUNTER;

  protected final static int MAX_NUMBER_OF_CLUSTERS = 1000;

  /**
   * Center-based or distribution-based clustering.
   */
  boolean is_center_based;

  /**
   * The set of clusters.
   */
  Vector clusters;

  /**
   * The comparison measure.
   */
  ComparisonMeasure function;

  /**
   * The cluster description.
   */
  ClusterDescription cluster_description;

  /**
   * Package proprietary constructor. Create an empty set of clusters.
   *
   * @param physical_path File the repository physical path containing the model.
   */
  protected ClusteringModel(java.io.File physical_path) {
    super(physical_path);
    clusters = new Vector();
    CLUSTER_IDENTIFIER_COUNTER = 0;
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return kddml.Core.KDDMLObjectType.PMML_CLUSTERING_MODEL;
  }

  /**
   * Returns ture if the model do not contain clusters.
   *
   * @return boolean
   */
  public boolean isEmpty() {
    return (clusters.size() == 0);
  }

  /**
   * Save the object in the system repository.
   * The destination path is provided by the object_path variable.
   *
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws kddml.Core.KDDMLCoreException {

    Document root = initializeDocument();

    // attribute settings...
    Element clustering_model;

    clustering_model = XMLUtility.doc.createElement("ClusteringModel");
    if (mining_schema != null)
      clustering_model = appendMiningSchema((MiningSchema) mining_schema, clustering_model);

    clustering_model = setAttributes(clustering_model);
    root.getDocumentElement().appendChild(clustering_model);

    // generate the XML for the comparison measure
    Element function = this.function.toXML();
    clustering_model.appendChild(function);

    // generate the clustering fields...
    for (int i=0; i<cluster_description.getNumberOfFields(); i++) {

      ClusteringField field = (ClusteringField) cluster_description.fields.get(i);
      Element ris = XMLUtility.doc.createElement("ClusteringField");
      XMLUtility.setAttribute(ris, "field", field.field_name);
      if (field.function != null)
        ris.setAttribute("compareFunction", field.getFieldComparisonFunction().getFunctionName());
      if (field.weight != null)
        ris.setAttribute("fieldWeight", field.getFieldWeight().toString());
      if (field.similarity_scale != null)
        ris.setAttribute("similarityScale", field.getSimilarityScale().toString());

      clustering_model.appendChild(ris);
    }

    // generate the XML for the centroid element
    //if (cluster_description.is_centroid_based)
    clustering_model.appendChild(cluster_description.toXML());

    // generate the XML element for each clusters
    for (int i=0; i<clusters.size(); i++) {
      Element cluster = ((Cluster) clusters.get(i)).toXML();
      clustering_model.appendChild(cluster);
    }

    XMLUtility.toFile(object_path.getAbsolutePath(), root);
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = super.toString();
    s = s + "---------------------------------\n";
    s = s + "Clustering Model:\n";
    s = s + "---------------------------------\n";
    s = s + "Comparison Measure: "+function.getFunctionName()+"\n";
    s = s + "Number of clusters: "+clusters.size()+"\n";

    s = s + "----\n";
    s = s + cluster_description.toString();

    for (int i=0; i<clusters.size(); i++) {
      s = s + "------\n";
      s = s + "Clusters:\n";
      String cluster = ((Cluster) clusters.get(i)).toString();
      s = s + cluster;
    }

    return s;
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML()
      throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.PMML_CLUSTERING_MODEL);

    String s1 = getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    try {
      XMLUtility.formatting(s1, s2, s3, null, null);
    }
    catch(Exception e) {
      String msg = "Exception found during the HTML code generation.";
      msg = e.toString();
      throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
    }
  }

  /**
   * Adds a new cluster to the model, both for centroid-based and distribution-based
   * clustering.
   * Throws an exception if the number of clusters exceededs the maximum number
   * of clusters allowed.
   *
   * @param cluster ClusterManager
   * @throws ClusteringModelException
   */
  public void addCluster(ClusterManager cluster) throws ClusteringModelException {

    if (clusters.size() > this.MAX_NUMBER_OF_CLUSTERS) {
      String msg = "Exceeded the maximum number of clusters allowed!";
      throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    // set the centroid description
    ((Cluster) cluster).centroid_description = this.cluster_description;

    clusters.add(cluster);
  }

  /**
   * Returns the aggregate function used to compare two objects.
   * This depends on the type of clustering. Cannot return null.
   *
   * @return ComparisonMeasure
   */
  public ComparisonMeasure getComparisonMeasure() {
    return function;
  }

  /**
   * Returns the cluster description for this model. Cannor return null.
   *
   * @return ClusterDescriptionManager
   */
  public ClusterDescriptionManager getClusterDescription() {
    return (ClusterDescriptionManager) cluster_description;
  }

  /**
   * Returns the Cluster object in the model with the specified identifier.
   * Throws an exception if the specified index do not exist in the clustering model.
   *
   * @param identifier int a positive value representing the cluster index.
   * @throws ClusteringModelException
   * @return ClusterManager
   */
  public ClusterManager getCluster(int identifier) throws ClusteringModelException {
    if (clusters.size() > identifier)
      throw new ClusteringModelException("Cluster number "+identifier+" do not exist!",
                                         null,
                                         kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    return (ClusterManager) clusters.get(identifier);
  }

  /**
   * Returns an iterator of cluster objects.
   *
   * @return Iterator the set of cluster as ClusterManager.
   */
  public java.util.Iterator getClusters() {
    return clusters.iterator();
  }

  /**
   * Returns the number of clusters in the model.
   *
   * @return int
   */
  public int getNumberOfClusters() {
    return clusters.size();
  }

  /**
   * Returns true if the clustering is center-based.
   *
   * @return boolean
   */
  public boolean isCentroidBased() {
    return this.is_center_based;
  }

  /**
   * Returns true if the clustering is distribution-based.
   *
   * @return boolean
   */
  public boolean isDistributionBased() {
    return !this.is_center_based;
  }

  /**
   * Returns the maximum number of clusters that the model can contain.
   *
   * @return int
   */
  public int getMaxNumberOfClusters() {
    return this.MAX_NUMBER_OF_CLUSTERS;
  }

  /**
   * Returns the cluster containing the input instance.
   * This depends on the type of clustering and the comparison measure used.
   *
   * @param instance Object
   * @throws ClusteringModelException
   * @return ClusterManager
   */
  public ClusterManager getCluster(Object instance) throws ClusteringModelException {

    Cluster[] c = new Cluster[clusters.size()];
    for (int i=0; i<c.length; i++) {
      c[i] = (Cluster) clusters.get(i);
    }

    if (function.isCentroidBased()) {

      // CENTROID-BASED function
      double[] d = new double[clusters.size()];
      for (int i = 0; i < clusters.size(); i++) {
        d[i] = ((CentroidBasedComparisonMeasure) function).compare( (Cluster) clusters.get(i),
            (weka.core.Instance) instance);
      }

      int index;
      if (((CentroidBasedComparisonMeasure) function).isSimilarityMeasure())
        index = Utils.getMaxIndex(d);
      else
        index = Utils.getMinIndex(d);
      return (ClusterManager) clusters.get(index);
    }
    else {

      // DISTRIBUTION-BASED function
      double[] prob = ((DistributionBasedComparisonMeasure) function).getDistributionProbability(c,
          (weka.core.Instance) instance);
      int index = Utils.getMaxIndex(prob);

      return (ClusterManager) clusters.get(index);
    }
  }

  /**
   * Returns a representation of each single cluster as instance.
   * This depends on the type of clustering. For center-based clustering, each cluster
   * is represented by the centroid. In this case, it returns the centroid poind as
   * instance. For distribution-based clustering, each cluster is represented by the
   * statistics. In this case, the instace values depend on the type of attribute.
   * For numeric attribute, the mean containing in the statistics is reported.
   * For discrete attribute, the most probable category value is reported.
   *
   * @throws ClusteringModelException if an error occours.
   * @return Object the set of instances as in weka.core.Instances. Each record
   * represent a single cluster.
   */
  public Object toInstances() throws ClusteringModelException {

    Instance tupla;
    Instances ris = null;

    ris = (Instances) cluster_description.toInstances();

    for (int i=0; i<clusters.size(); i++) {
      tupla = (Instance) ((Cluster) clusters.get(i)).getSeedAsInstance();
      tupla.setDataset(ris);
      ris.add(tupla);
    }
    return ris;
  }

  /**
   * Sets the PMML attribute to the input element.
   *
   * @param clustering_model Element
   * @throws ClusteringModelException
   * @return Element
   */
  protected Element setAttributes(Element clustering_model)
      throws ClusteringModelException {

    clustering_model.setAttribute("functionName", "clustering");
    XMLUtility.setAttribute(clustering_model, "algorithmName", algorithm_name);
    XMLUtility.setAttribute(clustering_model, "modelName", model_name);
    if (cluster_description.isCentroidBased())
      clustering_model.setAttribute("modelClass", "centerBased");
    else
      clustering_model.setAttribute("modelClass", "distributionBased");
    clustering_model.setAttribute("numberOfClusters", new Integer(clusters.size()).toString());
    return clustering_model;
  }
}
