/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

/**
 * A manager interface for ClusteringModel.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface ClusteringModelManager extends kddml.Core.DataMining.MiningModelManager {

  /**
   * Adds a new cluster to the model, both for centroid-based and distribution-based
   * clustering.
   * Throws an exception if the number of clusters exceededs the maximum number
   * of clusters allowed.
   *
   * @param cluster ClusterManager
   * @throws ClusteringModelException
   */
  public void addCluster(ClusterManager cluster) throws ClusteringModelException;

  /**
   * Returns the aggregate function used to compare two objects.
   * This depends on the type of clustering. Cannot return null.
   *
   * @return ComparisonMeasure
   */
  public ComparisonMeasure getComparisonMeasure();

  /**
   * Returns the cluster description for this model. Cannor return null.
   *
   * @return ClusterDescriptionManager
   */
  public ClusterDescriptionManager getClusterDescription();

  /**
   * Returns the Cluster object in the model with the specified identifier.
   * Throws an exception if the specified index do not exist in the clustering model.
   *
   * @param identifier int a positive value representing the cluster index.
   * @throws ClusteringModelException
   * @return ClusterManager
   */
  public ClusterManager getCluster(int identifier) throws ClusteringModelException;

  /**
   * Returns an iterator of cluster objects.
   *
   * @return Iterator the set of cluster as ClusterManager.
   */
  public java.util.Iterator getClusters();

  /**
   * Returns the number of clusters in the model.
   *
   * @return int
   */
  public int getNumberOfClusters();

  /**
   * Returns true if the clustering is center-based.
   *
   * @return boolean
   */
  public boolean isCentroidBased();

  /**
   * Returns true if the clustering is distribution-based.
   *
   * @return boolean
   */
  public boolean isDistributionBased();

  /**
   * Returns the maximum number of clusters that the model can contain.
   *
   * @return int
   */
  public int getMaxNumberOfClusters();

  /**
   * Returns the cluster containing the input instance.
   * This depends on the type of clustering and the comparison measure used.
   *
   * @param instance Object
   * @throws ClusteringModelException
   * @return ClusterManager
   */
  public ClusterManager getCluster(Object instance) throws ClusteringModelException;

  /**
   * Returns a representation of each single cluster as instance.
   * This depends on the type of clustering. For center-based clustering, each cluster
   * is represented by the centroid. In this case, it returns the centroid poind as
   * instance. For distribution-based clustering, each cluster is represented by the
   * statistics. In this case, the instace values depend on the type of attribute.
   * For numeric attribute, the mean containing in the statistics is reported.
   * For discrete attribute, the most probable category value is reported.
   *
   * @throws ClusteringModelException if an error occours.
   * @return Object the set of instances. Each record represent a single cluster.
   */
  public Object toInstances() throws ClusteringModelException;


}
