/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.DataSources.*;
import java.util.Comparator;
import java.util.TreeSet;
import java.util.Iterator;
import org.w3c.dom.*;
import kddml.Core.XMLUtility;

/**
 * A continuous clustering field. This field is normalized.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ContinuousClusteringField extends ClusteringField
    implements ContinuousClusteringFieldManager {

  /**
   * This matrix represents the linear norm for this numeric field.
   * The matrix is 2 X n. The first row contains the origin values.
   * The second row contains the normalized values.
   */
  protected TreeSet normalization_matrix;

  /**
   * Constructor given the attribute statistics.
   *
   * @param name String the field name as in MiningSchema.
   * @param func AttributeComparisonMeasure the attribute comparison measure. Can be null,
   * for example when a distribution-based function is used.
   * @param stat NumericalStatisticManager
   */
  public ContinuousClusteringField(String name,
                                   AttributeComparisonMeasure func,
                                   NumericalStatisticManager stat) {

    super(name, func);

    normalization_matrix = new TreeSet(new ComparePoints());

    Point p1 = new Point(stat.getMinValue().doubleValue(), 0);
    Point p2 = new Point(stat.getMean(), 0.5);
    Point p3 = new Point(stat.getMaxValue().doubleValue(), 1);

    normalization_matrix.add(p1);
    normalization_matrix.add(p2);
    normalization_matrix.add(p3);
  }

  /**
   * Empty constructor. Build a new continuous clustering field with an empty
   * normalization matrix.
   *
   * @param name String
   * @param func AttributeComparisonMeasure
   */
  public ContinuousClusteringField(String name,
                                   AttributeComparisonMeasure func) {

    super(name, func);

    normalization_matrix = new TreeSet(new ComparePoints());
  }

  /**
   * Adds a new point to the normalization matrix.
   * Ordered is preserved.
   *
   * @param origin double
   * @param norm double
   */
  protected void addPoint(double origin, double norm) {
    // aggiunge il nuovo punto mantenendo l'ordinamento

    Point p1 = new Point(origin, norm);
    normalization_matrix.add(p1);
  }

  /**
   * Normalizes the input value using the normalization matrix.
   *
   * @param value double
   * @throws ClusteringModelException if the normalization matrix is empty.
   * @return double a value between 0 and 1.
   */
  public double normalize(double value) throws ClusteringModelException {
    if (normalization_matrix.size() < 2) {
      // return error!!!
      String msg = "Two or more points needed per normalization. Can't normalize value!";
      throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    // ottiene il punto precedente ed il punto successivo al valore da normalizzare
    Iterator iter = normalization_matrix.iterator();
    // the previous point
    Point p1 = (Point) iter.next();
    // the next point
    Point p2 = (Point) iter.next();
    if (p1.origin > value)
      return p1.norm;
    while(iter.hasNext()) {
      if ((p1.origin <= value) && (p2.origin >= value))
        break;
      p1 = p2;
      p2 = (Point) iter.next();
    }
    if (p2.origin < value)
      return p2.norm;

    double d1 = value - p1.origin;
    double d2 = p2.origin - p1.origin;
    double d3 = p2.norm - p1.norm;

    return p1.norm + (d1 / d2 * d3);

  }

  /**
   * Denormalizes the input value using the normalization matrix.
   *
   * @param value double a value between 0 and 1.
   * @throws ClusteringModelException if the normalization matrix is empty.
   * @return double
   */
  public double denormalize(double value) throws ClusteringModelException {

    if (normalization_matrix.size() < 2) {
      // return error!!!
      String msg = "Two or more points needed per normalization. Can't normalize value!";
      throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    // ottiene il punto precedente ed il punto successivo al valore da denormalizzare
    Iterator iter = normalization_matrix.iterator();
    // the previous point
    Point p1 = (Point) iter.next();
    // the next point
    Point p2 = (Point) iter.next();
    if (p1.norm > value)
      return p1.origin;
    while(iter.hasNext()) {
      if ((p1.norm <= value) && (p2.norm >= value))
        break;
      p1 = p2;
      p2 = (Point) iter.next();
    }
    if (p2.norm < value)
      return p2.origin;

    double d1 = value - p1.norm;
    double d2 = p2.norm - p1.norm;
    double d3 = p2.origin - p1.origin;

    return p1.origin + (d1 / d2 * d3);

  }

  /**
   * Returns the number of points used for the normalization.
   *
   * @return int
   */
  public int getNumberOfPoints() {
    return normalization_matrix.size();
  }

  /**
   * Returns true if the field is continuous. Returns false if it is discrete.
   *
   * @return boolean true.
   */
  public boolean isContinuous() {
    return true;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {

    Element ris = XMLUtility.doc.createElement("DerivedField");

    Element el = XMLUtility.doc.createElement("NormContinuous");
    XMLUtility.setAttribute(el, "field", super.field_name);

    Iterator iter = normalization_matrix.iterator();
    Point p;
    while(iter.hasNext()) {
      Element norm = XMLUtility.doc.createElement("LinearNorm");
      el.appendChild(norm);
      p = (Point) iter.next();
      norm.setAttribute("orig", new Double(p.origin).toString());
      norm.setAttribute("norm", new Double(p.norm).toString());
    }
    ris.appendChild(el);
    return ris;
  }

  public static void main(String args[]) {
    ContinuousClusteringField f = new ContinuousClusteringField("pippo", null);
    f.addPoint(74, 0.5);
    f.addPoint(64, 0);
    f.addPoint(85, 1);

    try {
      System.out.println(f.normalize(72));
      System.out.println(f.denormalize(0.4));
    }
    catch(ClusteringModelException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {

    String ris = super.toString();
    ris = ris + "Normalization Matrix: ";
    Iterator iter = normalization_matrix.iterator();
    while (iter.hasNext())
      ris = ris + ((Point) iter.next()).toString() + " ";
    ris = ris + "\n";
    return ris;
  }
}

/**
 * This class represents a point used for numeric normalization.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
class Point {

  protected double origin;

  protected double norm;

  protected Point(double origin, double norm) {
    this.origin = origin;
    this.norm = norm;
  }

  public String toString() {
    return "("+Utils.doubleToString(origin, 3)+", "+Utils.doubleToString(norm, 3)+")";
  }
}

/**
 * This class is used to compare two points.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
class ComparePoints implements Comparator {

  /**
   * Ascending origin ordered.
   *
   * @param i1 Object
   * @param i2 Object
   * @return int
   */
  public int compare(Object i1, Object i2) {
    return Double.compare(((Point) i1).origin, ((Point) i2).origin);
  }
}

