/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * This class represents a statistic field used in ClusterStatistic and contains statistical
 * information about a single continuous mining field over the overall population.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class ContinuousStatisticField extends StatisticField
    implements ContinuousStatisticFieldManager {

  /**
   * The mean of population instances.
   */
  double mean;

  /**
   * The standard deviation of population instances.
   */
  double std_dev;

  /**
   * Constructor.
   *
   * @param field_name String the field name as in MiningSchema.
   * @param mean double the mean of population.
   * @param std_dev double the standard deviation of population.
   */
  public ContinuousStatisticField(String field_name, double mean, double std_dev) {

    super(field_name);
    this.mean = mean;
    this.std_dev = std_dev;
  }

  /**
   * Returns true if the field is continuous.
   *
   * @return boolean true
   */
  public boolean isContinuous() {
    return true;
  }

  /**
   * Returns true if the field is discrete.
   *
   * @return boolean true
   */
  public boolean isDiscrete() {
    return false;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String result = "Numeric Estimator on Attribute "+name+".";
    result = result + " Mean = "+Utils.doubleToString(mean, 5);
    result = result + " Standard Deviation = "+Utils.doubleToString(std_dev, 5)+"\n";
    return result;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("PartitionFieldStats");
    XMLUtility.setAttribute(ris, "field", super.name);

    Element el = XMLUtility.doc.createElement("NumericInfo");
    el.setAttribute("mean",new Double(mean).toString());
    el.setAttribute("standardDeviation",new Double(std_dev).toString());

    ris.appendChild(el);

    return ris;
  }

  /**
   * Returns the mean of the population.
   *
   * @return double
   */
  public double getMean() {
    return mean;
  }

  /**
   * Returns the standard deviation of the population.
   *
   * @return double
   */
  public double getStandardDeviation() {
    return std_dev;
  }
}
