/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import org.w3c.dom.*;
import kddml.Core.XMLUtility;

/**
 * This class represents a discrete centroid field.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DiscreteClusteringField extends ClusteringField
    implements DiscreteClusteringFieldManager {

  /**
   * The set of categories of this field.
   */
  protected String categories[];

  /**
   * Constructor.
   *
   * @param name String the field name as in MiningSchema.
   * @param func AttributeComparisonMeasure the attribute comparison measure. Can be null,
   * for example when a distribution-based function is used.
   * @param categories String[] the list of categories for this field.
   */
  public DiscreteClusteringField(String name,
                                 AttributeComparisonMeasure func,
                                 String categories[]) {
    super(name, func);
    this.categories = categories;
  }

  /**
   * Returns the categories of this clustering field.
   *
   * @return String[]
   */
  public String[] getCategories() {
    return categories;
  }

  /**
   * Returns the number of categories of this clustering field.
   *
   * @return int
   */
  public int getNumberOfCategories() {
    return categories.length;
  }

  /**
   * Returns true if the field is continuous. Returns false if it is discrete.
   *
   * @return boolean false.
   */
  public boolean isContinuous() {
    return false;
  }

  /**
   * Returns the distribution probability for a given category.
   * In particular, an element (f, v) defines that a category has value 1.0
   * if the value of input field category is v, otherwise it is 0.
   *
   * @param category String
   * @throws ClusteringModelException if the category do not exist.
   * @return double[] a probability vector.
   */
  public double[] getDistribution(String category) throws ClusteringModelException {

    int index = -1;
    for (int i=0; i<categories.length; i++) {
      if (categories[i].equalsIgnoreCase(category)) {
        index = i;
        break;
      }
    }
    if (index == -1)
      throw new ClusteringModelException("Category "+category+" not found!",
                                         null,
                                         kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    double[] ris = new double[categories.length];
    for (int j=0; j<ris.length; j++)
      if (j == index)
        ris[j] = 1;
      else
        ris[j] = 0;
    return ris;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element[]
   */
  protected Element[] toXML() {

    Element[] ris = new Element[categories.length];
    for (int i=0; i<categories.length; i++) {
      Element elem = XMLUtility.doc.createElement("DerivedField");
      Element el = XMLUtility.doc.createElement("NormDiscrete");
      XMLUtility.setAttribute(el, "field", super.field_name);
      String s = categories[i];
      s.replaceAll("<", "&lt;");
      XMLUtility.setAttribute(el, "value", s);
      elem.appendChild(el);
      ris[i] = elem;
    }
    return ris;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {

    String ris = super.toString();
    ris = ris + "Discrete Categories: ";
    for (int i=0; i< categories.length; i++)
      ris = ris + categories[i] +",";
    ris = ris + "\n";
    return ris;
  }
}
