/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.XMLUtility;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import java.util.Hashtable;
import java.util.Enumeration;

/**
 * This class represents a statistic field used in ClusterStatistic and contains statistical
 * information about a single discrete mining field over the overall population.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DiscreteStatisticField extends StatisticField
    implements DiscreteStatisticFieldManager {

  /**
   * Lists the categories of the field with frequency counts.
   * As key, the category name.
   * As value, the frequency that do not include missing values.
   */
  Hashtable frequencies;

  double total;

  /**
   * Constructor given the field name.
   * Builds an empty field, with no categories.
   *
   * @param field_name String the field name as in MiningSchema.
   */
  public DiscreteStatisticField(String field_name) {

    super(field_name);
    frequencies = new Hashtable();
    total = 0;
  }

  /**
   * Adds a new category with cluster frequency.
   * Return true if the category is added to the field.
   * Returns false if the field already contain the category name.
   *
   * @param value String the field category value.
   * @param frequency double the cluster frequency. Do not include missing values.
   * @return boolean
   */
  public boolean addValue(String value, double frequency) {

    if (frequencies.containsKey(value))
      return false;
    frequencies.put(value, new Double(frequency));
    total += frequency;
    return true;
  }

  /**
   * Returns true if the field is continuous.
   *
   * @return boolean false
   */
  public boolean isContinuous() {
    return false;
  }

  /**
   * Returns true if the field is discrete.
   *
   * @return boolean true
   */
  public boolean isDiscrete() {
    return true;
  }

  /**
   * Returns the probability value for the input category.
   *
   * @param value String the input category.
   * @throws ClusteringModelException if the category is not found.
   * @return double a value between 0 and 1.
   */
  public double getProbability(String value) throws ClusteringModelException {

    if (!frequencies.containsKey(value)) {
      String msg = value+" not found in discrete field "+name;
      throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (total == 0) {
      return 0;
    }
    double d = ((Double) frequencies.get(value)).doubleValue();
    return d / total;
  }

  /**
   * Returns the distribution probability of all categories.
   * Ordering of the output object respects the ordering of the method getCategories().
   *
   * @throws ClusteringModelException
   * @return double[] a normalization vector.
   */
  public double[] getProbability() throws ClusteringModelException {

    double[] ris = new double[frequencies.size()];

    Enumeration enum2 = frequencies.elements();
    int i=0;
    while(enum2.hasMoreElements()) {

      double d = ((Double) enum2.nextElement()).doubleValue();
      ris[i] = d / total;
      i++;
    }

    return ris;
  }

  /**
   * Returns the must probable category value.
   *
   * @throws ClusteringModelException
   * @return String
   */
  public String getProbableValue() throws ClusteringModelException {

    int p = Utils.getMaxIndex(getProbability());

    Enumeration enum2 = frequencies.keys();
    int i=0;
    while(enum2.hasMoreElements()) {

      String ris = (String) enum2.nextElement();
      if (i == p)
        return ris;
    }

    return null;
  }

  /**
   * Returns the categories of this field.
   * Ordering of the output object respects the ordering of the method getProbability().
   *
   * @return String[]
   */
  public String[] getCategories() {

    String[] ris = new String[frequencies.size()];

    Enumeration enum2 = frequencies.keys();
    int i=0;
    while(enum2.hasMoreElements()) {

      ris[i] = (String) enum2.nextElement();
      i++;
    }

    return ris;
  }

  /**
   * Returns the frequency of all categories.
   * Missing value are not included.
   * Ordering of the output object respects the ordering of the method getCategories().
   *
   * @throws ClusteringModelException
   * @return double[]
   */
  public double[] getFrequencies() {
    double[] ris = new double[frequencies.size()];

    Enumeration enum2 = frequencies.elements();
    int i=0;
    while(enum2.hasMoreElements()) {

      ris[i] = ((Double) enum2.nextElement()).doubleValue();
      i++;
    }

    return ris;

  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {

    String result = "Discrete Estimator on Attribute "+name+".";

    Enumeration enum2 = frequencies.keys();
    int i=0;
    while(enum2.hasMoreElements()) {

      String key = (String) enum2.nextElement();
      result += "  "+key+": " + Utils.doubleToString(((Double) frequencies.get(key)).doubleValue(), 2);
      i++;
    }
    result += "  (Total = " + Utils.doubleToString(total, 2) + ")\n";
    return result;
  }

  /**
   * Returns a representation of this object as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("PartitionFieldStats");
    XMLUtility.setAttribute(ris, "field", super.name);

    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n",new Integer(frequencies.size()).toString());
    array.setAttribute("type","real");
    String list="";

    Enumeration enum2 = frequencies.keys();
    int i=0;
    while(enum2.hasMoreElements()) {

      String key = (String) enum2.nextElement();
      list = list + ((Double) frequencies.get(key)).toString()+" ";
      i++;
    }
    Node n = XMLUtility.doc.createTextNode(list);
    array.appendChild(n);
    ris.appendChild(array);

    return ris;

  }
}
