/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import org.w3c.dom.*;
import java.util.Vector;
import weka.core.*;
import kddml.Core.XMLUtility;

/**
 * Model-based or distribution-based clustering methods assume the data of each cluster
 * conforms to a specific statistical distribution (e.g. the Gaussian distribution)
 * and the whole dataset is a mixture of several distribution models. EM is an example
 * of distribution-based partitioning clustering that do not require the specification
 * of distance measures.<br>
 * A distribution-based cluster is always defined by their statistical information.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DistributionBasedCluster extends Cluster
    implements DistributionBasedClusterManager {

  /**
   * Constructor given the cluster name and the statistic distribution.
   *
   * @param name String
   * @param distribution ClusterStatistic
   */
  public DistributionBasedCluster(String name, ClusterStatistic distribution) {

    super(name);
    cluster_distribution = distribution;
  }

  /**
   * Returns true if the clustering is distribution-based.
   *
   * @return boolean true
   */
  public boolean isDistributionBased() {
    return true;
  }

  /**
   * Returns true if the clustering is center-based.
   *
   * @return boolean false
   */
  public boolean isCentroidBased() {
    return false;
  }

  /**
   * Returns the representative cluster seed as instance.
   * Returns the mean value of the distribution for a continuous field.
   * Returns the most probable category of the distribution for a discrete field.
   *
   * @throws ClusteringModelException
   * @return Object a single instance as in weka.core.Instance.
   */
  public Object getSeedAsInstance() throws ClusteringModelException {

    // get the schema...
    Instances schema = (Instances) centroid_description.toInstances();

    // get the centroid as instance
    Instance instance = new Instance(centroid_description.getNumberOfFields());
    instance.setDataset(schema);

    double ris[] = new double[cluster_distribution.getNumberOfFields()];
    Vector v = cluster_distribution.field_list;
    for(int i=0; i<v.size(); i++) {
      StatisticField field = (StatisticField) v.get(i);
      if (field.isContinuous()) {
        double d = ((ContinuousStatisticField) field).mean;
        instance.setValue(i, d);
      }
      else {
        String s = ((DiscreteStatisticField) field).getProbableValue();
        instance.setValue(i, s);
      }
    }

    return instance;

  }

  /**
   * Returns the seed coordinate as array of normalized doubles in respect on the
   * ClusterDescription fields.
   * Returns the normalized mean value of the distribution for a continuous field.
   * Returns the normalized distribution for a discrete field.
   *
   * @throws ClusteringModelException
   * @return double[]
   */
  public double[] getSeedCoordinate() throws ClusteringModelException {
    Vector v = cluster_distribution.field_list;
    Vector ris = new Vector();
    for(int i=0; i<v.size(); i++) {
      StatisticField field = (StatisticField) v.get(i);
      if (field.isContinuous()) {
        // get the mean...
        double mean = ((ContinuousStatisticField) field).mean;
        // normalize the mean using the cluster description...
        double mean_norm = super.normalizeFieldValue(field.name, mean);
        ris.add(new Double(mean_norm));
      }
      else {
        double[] freq = ((DiscreteStatisticField) field).getProbability();
        for (int j=0; j<freq.length; j++)
          ris.add(new Double(freq[j]));
      }
    }
    double[] ris2 = new double[ris.size()];
    for (int h=0; h<ris.size(); h++)
      ris2[h] = ((Double) ris.get(h)).doubleValue();
    return ris2;
  }

  /**
   * Returns a representation of this cluster as string.
   *
   * @return String
   */
  public String toString() {
    String ris = super.toString();
    ris += cluster_distribution.toString();
    return ris;
  }

  /**
   * Returns a representation of this cluster as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element root = super.toXML();

    // append the centroid...

    double[] centroid = null;
    try {
      centroid = getSeedCoordinate();
    }
    catch (ClusteringModelException ex) {
      System.out.println(ex.getDetailErrorMessage());
      return null;
    }
    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n", new Integer(centroid.length).toString());
    array.setAttribute("type","real");
    String list="";
    for(int i=0; i<centroid.length;i++) {
      list = list + centroid[i] + " ";
    }
    Node n = XMLUtility.doc.createTextNode(list);
    array.appendChild(n);
    root.appendChild(array);

    // append the distribution...
    Element el = cluster_distribution.toXML();
    root.appendChild(el);
    return root;
  }
}
