/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import weka.core.Instance;
import java.util.Vector;
import java.util.Iterator;
import weka.core.*;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import kddml.Core.XMLUtility;

/**
 * Proprietary comparison function used by EM algorithm.
 * EM assigns a probability distribution to each instance which
 * indicates the probability of it belonging to each of the clusters.
 * For each cluster, a prior probability is given.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class EM_DISTANCE extends DistributionBasedComparisonMeasure {

  /**
   * The natural logarithm of 2.
   */
  final static double log2 = Math.log(2);

  /**
   * Constant for normal distribution.
   */
  final static double m_normConst = Math.log(Math.sqrt(2*Math.PI));

  /**
   * Priors probability.
   */
  double[] priors;

  /**
   * Constructor given the prior probability for each cluster.
   *
   * @param cluster_priors double[]
   */
  public EM_DISTANCE(double[] cluster_priors) {
    super();
    priors = cluster_priors;
  }

  /**
   * Given an input instance, returns a distribution probability in which the i-esim
   * array location contains the probability that the instance belongs to the cluster
   * i.
   *
   * @param clusters Cluster[]
   * @param instance Instance
   * @throws ClusteringModelException
   * @return double[] a normalized array.
   */
  public double[] getDistributionProbability(Cluster[] clusters, Instance instance)
      throws ClusteringModelException {

    double[] ris = new double[clusters.length];
    for (int i=0; i<clusters.length; i++) {
      ris[i] = compare(clusters[i], instance);
    }
    return logs2probs(ris);
  }

  /**
   * Returns the function name as in PMML.
   *
   * @return String
   */
  public String getFunctionName() {
    return "X-EMDistance";
  }

  private double compare(Cluster cluster, Instance instance)
      throws ClusteringModelException {

    double ris;

    // get the cluster distribution
    ClusterStatistic distribution = ((DistributionBasedCluster) cluster).cluster_distribution;
    Iterator iter = distribution.getStatisticFields();
    int i = 0;
    // a vector of double[]
    Vector discrete_estimator = new Vector();
    double[] mean = new double[instance.numAttributes()];
    double[] std_dev = new double[instance.numAttributes()];

    // iterate on fields
    while (iter.hasNext()) {

      StatisticField f = (StatisticField) iter.next();

      if (f.isContinuous()) {
        // numeric field
        mean[i] = ( (ContinuousStatisticField) f).mean;
        std_dev[i] = ( (ContinuousStatisticField) f).std_dev;
        discrete_estimator.add(new double[0]);
      }
      if (f.isDiscrete()) {
        // discrete field - calculate the distribution for each value
        double[] freq = ( (DiscreteStatisticField) f).getProbability();
        mean[i] = 0;
        std_dev[i] = 0;
        discrete_estimator.add(freq);
      }
      i++;
    }

    try {
      // calculate the density for cluster
      // get the prior probability for this cluster...
      int p = cluster.getClusterId();
      ris = logJointDensitiesForInstance(instance, mean, std_dev, discrete_estimator, priors[p]);
      return ris;
    }
    catch(Exception e) {
      String s = "Can't compute distance function for instance: "+instance.toString();
      throw new ClusteringModelException(s, e.toString(),
                                         kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
    }
  }

  /**
   * Returns the logs of the joint densities for a given instance.
   *
   * @param inst Instance
   * @param mean double[]
   * @param std_dev double[]
   * @param d_e Vector
   * @param cluster_prior double
   * @throws Exception
   * @return double
   */
  private static double logJointDensitiesForInstance(Instance inst,
                                                     double[] mean,
                                                     double[] std_dev,
                                                     Vector d_e,
                                                     double cluster_prior)

      throws Exception {

    double weights = logDensityForInstance(inst, mean, std_dev, d_e);

    if (cluster_prior > 0) {
      weights += Math.log(cluster_prior);
    } else {
      throw new IllegalArgumentException("Cluster empty!");
    }
    return weights;
  }

  /**
   * Converts an array containing the natural logarithms of
   * probabilities stored in a vector back into probabilities.
   * The probabilities are assumed to sum to one.
   *
   * @param a double[]
   * @return double[]
   */
  private static double[] logs2probs(double[] a) {

    double max = a[maxIndex(a)];
    double sum = 0.0;

    double[] result = new double[a.length];
    for(int i = 0; i < a.length; i++) {
      result[i] = Math.exp(a[i] - max);
      sum += result[i];
    }

    Utils.normalize(result, sum);

    return result;
  }

  /**
   * Returns index of maximum element in a given
   * array of doubles. First maximum is returned.
   *
   * @param doubles double[]
   * @return int
   */
  private static int maxIndex(double[] doubles) {

    double maximum = 0;
    int maxIndex = 0;

    for (int i = 0; i < doubles.length; i++) {
      if ((i == 0) || (doubles[i] > maximum)) {
        maxIndex = i;
        maximum = doubles[i];
      }
    }

    return maxIndex;
  }

  /**
   * Computes the log of the conditional density for a given instance.
   *
   * @param inst Instance
   * @param mean double[]
   * @param std_dev double[]
   * @param d_e Vector
   * @throws Exception
   * @return double
   */
  private static double logDensityForInstance(Instance inst,
                                              double[] mean,
                                              double[] std_dev,
                                              Vector d_e) throws Exception {

    int i, j;
    double logprob;

    logprob = 0.0;

    for (j = 0; j < inst.numAttributes(); j++) {
      if (!inst.isMissing(j)) {
        if (inst.attribute(j).isNominal()) {
          // discrete attribute
          double[] distribution = (double[]) d_e.get(j);
          int value = (int) inst.value(j);
          // add 1 to the distribution
          logprob += Math.log(distribution[value]);
        }
        else {
          // numeric attribute
          logprob += logNormalDens(inst.value(j), mean[j], std_dev[j]);
        }
      }
    }

    return logprob;
  }

  /**
   * Density function of normal distribution.
   *
   * @param x double
   * @param mean double
   * @param stdDev double
   * @return double
   */
  private static double logNormalDens (double x, double mean, double stdDev) {

    double diff = x - mean;
    return - (diff * diff / (2 * stdDev * stdDev))  - m_normConst - Math.log(stdDev);
  }

  /**
   * Returns a representation of this function as PMML element.
   *
   * @return Element
   */
  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("ComparisonMeasure");

    ris.setAttribute("kind", "similarity");

    Element extension =  XMLUtility.doc.createElement("Extension");
    ris.appendChild(extension);
    Element dist = XMLUtility.doc.createElement("X-EMDistance");
    extension.appendChild(dist);

    // create the array of priors...
    Element array = XMLUtility.doc.createElement("Array");
    array.setAttribute("n", new Integer(priors.length).toString());
    array.setAttribute("type","real");
    String list="";
    for(int i=0; i<priors.length;i++) {
      list = list + priors[i] + " ";
    }
    Node n = XMLUtility.doc.createTextNode(list);
    array.appendChild(n);
    dist.appendChild(array);

    Element k = XMLUtility.doc.createElement("cityBlock");
    ris.appendChild(k);

    return ris;
  }
}
