/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

/**
 * The minkowcki function is an outer function, used to compare two records X, Y,
 * where:
 * <ul>
 * <li> Xi is the value for the i-esim field of X</li>
 * <li> Yi is the value for the i-esim field of X</li>
 * <li> Wi is the weight for the i-esim field of X (default 1)</li>
 * <li> a p-parameter > 0</li>
 * </ul>
 * The minkowski distance is defined as:
 * <br>
 * <p align="center">
 * D = (sum |Xi-Yi|^p)^(1/p).
 * </p><br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */

public class MINKOWSKI extends CentroidBasedComparisonMeasure {

  private int p_parameter;

  /**
   * Constructor given the list of comparison measures. Used for centroid-based
   * clustering only.
   *
   * @param attribute_comparison AttributeComparisonMeasure[] the comparison measure
   * for each attribute.
   * @param weights double[] the field weight for each attribute. Can be null.
   * @param p_param int the positive p-parameter.
   */
  public MINKOWSKI(AttributeComparisonMeasure[] attribute_comparison, double[] weights, int p_param) {
    super(attribute_comparison, weights);
    this.p_parameter = p_param;
  }

  /**
   * Constructor given the centroid description. Used both for centroid-based clustering
   * and distribution-based clustering.
   *
   * @param cluster_description ClusterDescription
   * @param p_param int the positive p-parameter.
   */
  public MINKOWSKI(ClusterDescription cluster_description, int p_param) {
    super(cluster_description);
    this.p_parameter = p_param;
  }

  /**
   * Returns true if the comparison measure is a similarity function, in which the
   * value returned by the compare method is optimal for greater values.
   * Returns false if the comparison measure is a distance measure, in which the
   * value returned by the compare method is optimal if it is 0.
   *
   * @return boolean false
   */
  public boolean isSimilarityMeasure() {
    return false;
  }

  /**
   * Evaluates the comparison measure given the comparison values for each single
   * attribute (i.e., the values returned by each single inner function).
   *
   * @param x double[]
   * @return double
   */
  protected double evaluate(double[] x) {
    double ris = 0;
    for (int i=0; i<x.length; i++) {

      double d1 = Math.abs(x[i]);
      double d2 = Math.pow(d1, p_parameter);

      ris = ris + d2;
    }
    return Math.pow(ris, 1/p_parameter);
  }

  /**
   * Returns the function name as in PMML.
   *
   * @return String
   */
  public String getFunctionName() {
    return "minkowski";
  }
}
