/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataMining.Clustering;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import org.w3c.dom.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Vector;

public final class PMMLClusteringFactory extends ClusteringExternalFactory {

  private String pmml_version;

  private Element header;

  private Element data_dictionary;

  private Element transformation_dictionary;

  private Element mining_schema;

  private Element clustering_model;

  private String algorithm_name;

  private String model_name;

  private boolean is_center_based;

  public PMMLClusteringFactory(java.io.File f, String pmml_path)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    // effettua il parsing del documento XML contenente il modello...
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(pmml_path),
                              new DTDResolver(kddml.Core.KDDMLObjectType.PMML_CLUSTERING_MODEL));
    }
    catch(java.io.IOException e) {
      String msg = "Invalid PMML reference: "+pmml_path;
      throw new ClusteringModelException(msg, e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }

    Element root = (Element) doc.getDocumentElement();

    if (!root.getTagName().equalsIgnoreCase("pmml"))
      throw new ClusteringModelException(pmml_path+ " no contains a PMML document.",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the pmml version
    pmml_version = root.getAttribute("version");

    // reading the copyright
    header = (Element) root.getElementsByTagName("Header").item(0);

     // reading the data dictionary element
    data_dictionary = (Element) root.getElementsByTagName("DataDictionary").item(0);

    // reading the transformation dictionary element
    NodeList l = root.getElementsByTagName("TransformationDictionary");
    if (l.getLength() != 0)
      transformation_dictionary = (Element) l.item(0);
    else
      transformation_dictionary = null;

    clustering_model = (Element) doc.getElementsByTagName("ClusteringModel").item(0);
    if (clustering_model == null)
      throw new ClusteringModelException(pmml_path+ " no contains a PMML clustering model document.",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the mining schema element...
    mining_schema = (Element) clustering_model.getElementsByTagName("MiningSchema").item(0);

    // reading attributes...
    if (clustering_model.getAttribute("modelClass").equalsIgnoreCase("centerBased"))
      is_center_based = true;
    else
      is_center_based = false;

    model_name = XMLUtility.getAttribute(clustering_model, "modelName");
    algorithm_name = XMLUtility.getAttribute(clustering_model, "algorithmName");
  }

  protected DataDictionary buildDataDictionary() throws MiningModelException {

    if (data_dictionary == null)
      return new DataDictionary();
    else
      return PMMLUtility.buildDataDictionary(data_dictionary);
  }

  protected MiningSchema buildMiningSchema() throws MiningModelException {

    if (mining_schema == null)
      return new MiningSchema();
    else
      return PMMLUtility.buildMiningSchema(mining_schema);
  }

  protected String buildAlgorithmName() throws MiningModelException {

    return algorithm_name;
  }

  protected String buildModelName() throws MiningModelException {

   return model_name;
  }

  protected String buildPMMLVersion() throws MiningModelException {

    return pmml_version;
  }

  protected String buildCopyright() throws MiningModelException {

    if (header != null)
      return header.getAttribute("copyright");
    else
      return null;
  }

  protected String buildApplicationName() throws MiningModelException {

    if (header != null) {
      Element el = (Element) header.getElementsByTagName("Application").item(0);
      if (el != null)
        return XMLUtility.getAttribute(el, "name");
      else
        return null;
    }
    else
      return null;
  }


  protected String buildApplicationVersion() throws MiningModelException {

    if (header != null) {
      Element el = (Element)header.getElementsByTagName("Application").item(0);
      if (el != null)
        return el.getAttribute("version");
      else
        return null;
    }
    else
      return null;
  }


 protected ComparisonMeasure buildComparisonMeasure(ClusterDescription d)
     throws ClusteringModelException {

   Element comparison_measure = (Element) clustering_model.getElementsByTagName("ComparisonMeasure").item(0);
   boolean is_distance;
   if (comparison_measure.getAttribute("kind").equals("distance"))
     is_distance = true;
   else
     is_distance = false;

   Element function = (Element) XMLUtility.getFirstChild(comparison_measure);

   // EM distance
   if (function.getTagName().equals("Extension")) {
     Element function2 = (Element) XMLUtility.getFirstChild(function);

     if (function2.getTagName().equals("X-EMDistance")) {
       Element array = (Element) function.getElementsByTagName("Array").item(0);
       String n = array.getAttribute("n");

       String arrayValues = array.getChildNodes().item(0).getNodeValue();
       // scorre l'array per creare il centroide...
       java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(
           arrayValues.trim());

       if (!n.equals("") && Double.parseDouble(n) != tokenizer.countTokens())
         throw new ClusteringModelException(
             "Illegal PMML Clustering Model: predicate set length mismatch <" +
             n + "> for cluster.", null,
             KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

       double[] values = new double[ (new Integer(n)).intValue()];

       int i = 0;
       while (tokenizer.hasMoreTokens()) {
         String itemStr = tokenizer.nextToken();
         values[i] = (new Double(itemStr)).doubleValue();
         i++;
       }

       return new EM_DISTANCE(values);
     }
     else {
       // no em-distance found on extension element
       String msg = "Incompatible PMML document: unsupported comparison measure " +
           function.getTagName();
       throw new ClusteringModelException(msg, null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
     }
   }

   // euclidean
   if (function.getTagName().equals("euclidean"))
     return new EUCLIDEAN(d);

   // squared euclidean
   if (function.getTagName().equals("squaredEuclidean"))
     return new SQUARED_EUCLIDEAN(d);

   // city block
   if (function.getTagName().equals("cityBlock"))
     return new CITY_BLOCK(d);

   // chebychev
   if (function.getTagName().equals("chebychev"))
     return new CHEBYCHEV(d);

   // minkowski
   if (function.getTagName().equals("minkowski")) {
     int p_param = new Integer(function.getAttribute("p-parameter")).intValue();
     return new MINKOWSKI(d, p_param);
   }

   // simple matching
   if (function.getTagName().equals("simpleMatching"))
     return new SIMPLE_MATCHING(d);

   // jaccard
   if (function.getTagName().equals("jaccard"))
     return new JACCARD(d);

   // tanimoto
   if (function.getTagName().equals("tanimoto"))
     return new TANIMOTO(d);

   // binary similarity
   if (function.getTagName().equals("binarySimilarity")) {
     double c00 = new Double(function.getAttribute("c00-parameter")).doubleValue();
     double c01 = new Double(function.getAttribute("c01-parameter")).doubleValue();
     double c10 = new Double(function.getAttribute("c10-parameter")).doubleValue();
     double c11 = new Double(function.getAttribute("c11-parameter")).doubleValue();
     double d00 = new Double(function.getAttribute("d00-parameter")).doubleValue();
     double d01 = new Double(function.getAttribute("d01-parameter")).doubleValue();
     double d10 = new Double(function.getAttribute("d10-parameter")).doubleValue();
     double d11 = new Double(function.getAttribute("d11-parameter")).doubleValue();
     return new BINARY_SIMILARITY(d, c00, c01, c10, c11, d00, d01, d10, d11);
   }

   // no comparison measure supported
   String msg = "Incompatible PMML document: unsupported comparison measure "+function.getTagName();
   throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
 }

 protected ClusterDescription buildClusterDescription() throws ClusteringModelException {

   NodeList clustering_fields = clustering_model.getElementsByTagName("ClusteringField");
   if (clustering_fields.getLength() == 0) {
     // error: clustering fields missing
     String msg = "Incompatible PMML document: no ClusteringFields found!";
     throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
   }

   ClusterDescription ris = new ClusterDescription(is_center_based);

   //if (is_center_based) {
     // centroid-based clustering
     NodeList n = clustering_model.getElementsByTagName("CenterFields");
     if (n.getLength() == 0) {
       String msg = "Incompatible PMML document: no CenterField found for centroid-based clustering!";
       throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
     }
     Element center_fields = (Element) n.item(0);
     NodeList derived_fields = center_fields.getElementsByTagName("DerivedField");

     Element clustering_field;
     // scorre ogni singolo ClusteringField del modello ...
     for (int i=0; i<clustering_fields.getLength(); i++) {
       clustering_field = (Element) clustering_fields.item(i);
       // ... individua i corrispondenti derivedField
       Vector index = new Vector();
       Element derived_field = null;
       for (int j = 0; j < derived_fields.getLength(); j++) {
         derived_field = (Element) derived_fields.item(j);

         String str = XMLUtility.getAttribute((Element) XMLUtility.getFirstChild(derived_field), "field");

         if (XMLUtility.getAttribute(clustering_field, "field").equalsIgnoreCase(str)) {
           // field found
           index.add(new Integer(j));
         }
       }
       if (index.size() == 0) {
         String msg = "Incompatible PMML document: no CenterField match a ClusteringField!";
         throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
       }
       Element[] d_f = new Element[index.size()];
       for (int k=0; k<index.size(); k++)
         d_f[k] = (Element) derived_fields.item(((Integer) index.get(k)).intValue());
       ClusteringField field = buildClusteringField(clustering_field, d_f);
       ris.addField(field);
     }
   //}
   /*else {
     // distribution-based clustering
     Element clustering_field;
     // scorre ogni singolo ClusteringField del modello ...
     for (int i=0; i<clustering_fields.getLength(); i++) {
       // get the attribute comparison function
       clustering_field = (Element) clustering_fields.item(i);
       String function_name = clustering_field.getAttribute("compareFunction");
       AttributeComparisonMeasure f = null;
       if (!function_name.equalsIgnoreCase(""))
         f = getComparisonFunction(function_name, clustering_field);

       clustering_field = (Element) clustering_fields.item(i);
       String name = matchFieldName(clustering_field.getAttribute("field"));
       Element data_field = getFieldFromDataDictionary(name);
       if (data_field == null) {
         String msg = "Incompatible PMML document: no DataField match the ClusteringField "+name;
         throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
       }

       if (data_field.getAttribute("optype").equals("continuous"))
         ris.addField(new ContinuousClusteringField(name, f));
       if (data_field.getAttribute("optype").equals("categorical")) {
         NodeList n = data_field.getElementsByTagName("Value");
         String[] categories = new String[n.getLength()];
         for (int m=0; m<n.getLength(); m++)
           categories[m] = ((Element) n.item(m)).getAttribute("value");
         ris.addField(new DiscreteClusteringField(name, f, categories));
       }
     }
   }*/

   return ris;
 }

 /**
  * Build the clustering field.
  *
  * @param clustering_field Element
  * @param derived_field Element
  * @throws ClusteringModelException
  * @return ClusteringField
  */
 private ClusteringField buildClusteringField(Element clustering_field, Element[] derived_field)
     throws ClusteringModelException {

   Element el = (Element) XMLUtility.getFirstChild(derived_field[0]);
   if (el.getTagName().equals("NormDiscrete"))
     return buildDiscreteClusteringField(derived_field, clustering_field);
   if (el.getTagName().equals("NormContinuous"))
     return buildContinuousClusteringField(el, clustering_field);
   String msg = "Incompatible PMML document: no CenterField match a ClusteringField!";
   throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
 }

 /**
  * Build the clustering field for a discrete field.
  *
  * @param norm_discrete Element
  * @param cluster_field Element
  * @throws MiningModelException
  * @return DiscreteClusteringField
  */
 private DiscreteClusteringField buildDiscreteClusteringField(Element[] norm_discrete, Element cluster_field)
     throws ClusteringModelException {

   // controlla che il nome on cluster_field sia matchato con il mining schema
   String name = matchFieldName(XMLUtility.getAttribute(cluster_field, "field"));

   String[] categories = new String[norm_discrete.length];
   for (int i=0; i<norm_discrete.length; i++) {

     String value = XMLUtility.getAttribute((Element) XMLUtility.getFirstChild(norm_discrete[i]), "value");
     categories[i] = value;
     if (value.equals("")) {
       String s = "Invalid discrete field. No value specified for field "+name;
       throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
     }
   }

   // get the attribute comparison function
   String function_name = cluster_field.getAttribute("compareFunction");
   AttributeComparisonMeasure f = null;
   if (!function_name.equalsIgnoreCase(""))
     f = getComparisonFunction(function_name, cluster_field);

   // get other attributes
   String weight = cluster_field.getAttribute("fieldWeight");
   String similarity = cluster_field.getAttribute("similarityScale");

   DiscreteClusteringField ris = new DiscreteClusteringField(name, f, categories);
   if (!weight.equals(""))
     ris.setFieldWeight(new Double(weight).doubleValue());
   if (!similarity.equals(""))
     ris.setSimilarityScale(new Double(similarity).doubleValue());

   return ris;
 }

 /**
  * Build the clustering field for a continuous field.
  *
  * @param norm_continue Element
  * @param cluster_field Element
  * @throws MiningModelException
  * @return ContinuousClusteringField
  */
 private ContinuousClusteringField buildContinuousClusteringField(Element norm_continue, Element cluster_field)
     throws ClusteringModelException {

   // controlla che il nome on cluster_field sia matchato con il mining schema
   String name = matchFieldName(XMLUtility.getAttribute(cluster_field, "field"));

   // get the attribute comparison function
   String function_name = cluster_field.getAttribute("compareFunction");
   AttributeComparisonMeasure f = null;
   if (!function_name.equalsIgnoreCase(""))
     f = getComparisonFunction(function_name, cluster_field);

   // get other attributes
   String weight = cluster_field.getAttribute("fieldWeight");
   String similarity = cluster_field.getAttribute("similarityScale");

   ContinuousClusteringField ris = new ContinuousClusteringField(name, f);
   if (!weight.equals(""))
     ris.setFieldWeight(new Double(weight).doubleValue());
   if (!similarity.equals(""))
     ris.setSimilarityScale(new Double(similarity).doubleValue());

   NodeList list = norm_continue.getElementsByTagName("LinearNorm");
   for (int i=0; i<list.getLength(); i++) {
     double d1 = new Double(((Element) list.item(i)).getAttribute("orig")).doubleValue();
     double d2 = new Double(((Element) list.item(i)).getAttribute("norm")).doubleValue();
     ris.addPoint(d1, d2);
   }
   return ris;

 }


 protected Cluster[] buildClusters() throws ClusteringModelException {
   NodeList clusters = clustering_model.getElementsByTagName("Cluster");
   Cluster[] ris = new Cluster[clusters.getLength()];
   for (int i=0; i<clusters.getLength(); i++) {
     if (is_center_based)
       ris[i] = buildCentroidCluster((Element) clusters.item(i));
     else
       ris[i] = buildDistributionCluster((Element) clusters.item(i));

   }
   return ris;
 }

 private CentroidBasedCluster buildCentroidCluster(Element cluster) throws ClusteringModelException {

   String name = XMLUtility.getAttribute(cluster, "name");
   Element array = (Element) cluster.getElementsByTagName("Array").item(0);

   String n = array.getAttribute("n");

   String arrayValues = array.getChildNodes().item(0).getNodeValue();
   // scorre l'array per creare il centroide...
   java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(arrayValues.trim());

   if( !n.equals("") && Double.parseDouble(n) != tokenizer.countTokens() )
     throw new ClusteringModelException(
        "Illegal PMML Clustering Model: array length mismatch <" +
        n + "> for cluster.", null,
        KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

   double[] values = new double[(new Integer(n)).intValue()];

   int i=0;
   while (tokenizer.hasMoreTokens()) {
     String itemStr = tokenizer.nextToken();
     values[i] = (new Double(itemStr)).doubleValue();
     i++;
   }
   return new CentroidBasedCluster(name, values);
 }

 /**
  * Build the cluster element for a distribution-based clustering.
  *
  * @param cluster Element
  * @return DistributionBasedCluster
  */
 private DistributionBasedCluster buildDistributionCluster(Element cluster)
     throws ClusteringModelException {

   String name = XMLUtility.getAttribute(cluster, "name");

   Element partition = (Element) cluster.getElementsByTagName("Partition").item(0);
   if (partition == null) {
     String msg = "Incompatible PMML document: no Partition element found for Distribution-Based Clustering!";
     throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
   }

   NodeList fields = partition.getElementsByTagName("PartitionFieldStats");

   ClusterStatistic stat = new ClusterStatistic(name);

   for (int i=0; i<fields.getLength(); i++) {

     String field_name = XMLUtility.getAttribute((Element) fields.item(i), "field");
     Element el = (Element) XMLUtility.getFirstChild((Element) fields.item(i));
     // array for discrete field or numeric info for numeric field...
     if (el.getTagName().equals("Array")) {
       // discrete distribution
       String n = el.getAttribute("n");

       String arrayValues = el.getFirstChild().getNodeValue();
       // scorre l'array per creare il centroide...
       java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(arrayValues.trim());

       int k=0;
       DiscreteStatisticField f = new DiscreteStatisticField(field_name);

       // get the category value from the data dictionary
       Element data_field = getFieldFromDataDictionary(field_name);
       if (data_field == null) {
         String msg = "Incompatible PMML document: no DataField match the ClusteringField "+name;
         throw new ClusteringModelException(msg, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
       }

       NodeList list = data_field.getElementsByTagName("Value");
       String[] categories = new String[list.getLength()];
       for (int m=0; m<list.getLength(); m++)
         categories[m] = XMLUtility.getAttribute((Element) list.item(m), "value");

       while (tokenizer.hasMoreTokens()) {
         String itemStr = tokenizer.nextToken();
         double value = (new Double(itemStr)).doubleValue();
         // add the category and the distribution value to the field
         f.addValue(categories[k], value);

         k++;
       }
       stat.addStatisticField(f);
     }
     if (el.getTagName().equals("NumericInfo")) {
       // discrete distribution
       double m = (new Double(el.getAttribute("mean"))).doubleValue();
       double s = (new Double(el.getAttribute("standardDeviation"))).doubleValue();
       stat.addStatisticField(new ContinuousStatisticField(field_name, m, s));
     }
   }

   DistributionBasedCluster ris = new DistributionBasedCluster(name, stat);
   return ris;
 }


 /**
  * Get the attribute comparison function given the function name.
  * No reflection is applied.
  *
  * @param function_name String
  * @throws ClusteringModelException
  * @return AttributeComparisonMeasure
  */
 private AttributeComparisonMeasure getComparisonFunction(String function_name, Element cluster_field)
     throws ClusteringModelException {

   if (function_name.equalsIgnoreCase("absDiff"))
     return new ABS_DIFF();

   if (function_name.equalsIgnoreCase("gaussSimm")) {
     // get the similarity scale
     String sim = cluster_field.getAttribute("similarityScale");
     if (sim.equals("")) {
       String s = "Attribute similarityScale is required for gaussSimm distance measure.";
       throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
     }
     return new GAUSS_SIM((new Double(sim)).doubleValue());
   }

   if (function_name.equalsIgnoreCase("delta"))
     return new DELTA();

   if (function_name.equalsIgnoreCase("equal"))
     return new EQUAL();

   if (function_name.equalsIgnoreCase("table")) {
     // get the similarity matrix in the cluster field element
     SymmetricMatrix matrix = null;
     try {
       matrix = getSimilarityMatrix(cluster_field);
     }
     catch(MiningModelException e) {
       throw new ClusteringModelException(e.getErrorMessage(), e.getDetailErrorMessage(), e.getErrorCode());
     }

     return new SIMILARITY_MATRIX(matrix);
   }

   String s = "Unsupported attribute comparison function: "+function_name;
   throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
 }


 private String matchFieldName(String name) throws ClusteringModelException {

   // viene controllato che name sia in mining_schema oppure in transformation_dictionary
   int index = -1;
   NodeList list = mining_schema.getElementsByTagName("MiningField");
   for (int j=0; j<list.getLength(); j++) {
     String name2 = XMLUtility.getAttribute((Element) list.item(j), "name");
     if (name.equals(name2)) {
       index = j;
       break;
     }
   }
   if (index == -1) {
     // clustering field not found in mining schema --> viene controllato il transformation dictionary
     if (transformation_dictionary != null) {
       index = -1;
       list = mining_schema.getElementsByTagName("DerivedField");
       for (int j=0; j<list.getLength(); j++) {
         String name2 = XMLUtility.getAttribute((Element) list.item(j), "name");
         if (name.equals(name2)) {
           index = j;
           break;
         }
       }
       if (index != -1) {
         // field found
         Element el = (Element) XMLUtility.getFirstChild((Element) list.item(index));
         name = el.getAttribute("field");
         name = XMLUtility.getAttribute(el, "field");
       }
       else {
         // fiend do not match
         String msg ="Illegal PMML document: no mining field match the center field " +name;
         throw new ClusteringModelException(msg, null,
                                            KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
       }
     }
     else {
       // fiend do not match
       String msg ="Illegal PMML document: no mining field match the center field " +name;
       throw new ClusteringModelException(msg, null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
     }
   }

   return name;

 }

 private Element getFieldFromDataDictionary(String name) {
   NodeList list = data_dictionary.getElementsByTagName("DataField");
   for (int j=0; j<list.getLength(); j++) {
     String name2 = XMLUtility.getAttribute((Element) list.item(j), "name");
     if (name.equals(name2))
       return (Element) list.item(j);
   }
   return null;
 }

 /**
  * Get the similarity matrix for the input clustering field.
  *
  * @param cluster_field Element
  * @return SymmetricMatrix
  */
 private SymmetricMatrix getSimilarityMatrix(Element cluster_field) throws MiningModelException {

   NodeList l = cluster_field.getElementsByTagName("Comparisons");
   if (l.getLength() == 0) {
     String s = "Similarity Matrix is required for table distance measure.";
     throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
   }
   Element el = (Element) l.item(0);
   Element matrix = (Element) el.getElementsByTagName("Matrix").item(0);

   if (!matrix.getAttribute("kind").equalsIgnoreCase("symmetric")) {
     String s = "Similarity Matrix must be symmetric.";
     throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
   }
   int num_rows = new Integer(matrix.getAttribute("numberOfRows")).intValue();

   // get the attribute values from the data field element
   int index = -1;
   String name = XMLUtility.getAttribute(cluster_field, "field");
   NodeList list = data_dictionary.getElementsByTagName("DataField");
   for (int j=0; j<list.getLength(); j++) {
     String name2 = XMLUtility.getAttribute((Element) list.item(j), "name");
     if (name.equals(name2)) {
       index = j;
       break;
     }
   }
   if (index == -1 ) {
     // fiend do not match
     String msg ="Illegal PMML document: no mining field match the center field " +name;
     throw new ClusteringModelException(msg, null,
                                        KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

   }
   Element data_field = (Element) list.item(index);
   if (!data_field.getAttribute("optype").equalsIgnoreCase("categorical")) {
     String s = "Similarity Matrix only for categorical attributes!";
     throw new ClusteringModelException(s, null, KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
   }
   NodeList values = data_field.getElementsByTagName("Value");

   // get the list of categories for the enumerated attribute
   String[] categories = new String[values.getLength()];
   for (int i=0; i<values.getLength(); i++) {
     categories[i] = XMLUtility.getAttribute((Element) values.item(i), "value");
   }

   SymmetricMatrix ris = new SymmetricMatrix(categories);

   Element array;
   StringTokenizer tokeniz;
   // set the values of the matrix
   for (int j=0; j<categories.length; j++) {
     array = (Element) matrix.getElementsByTagName("Array").item(j);
     String matrix_values = array.getChildNodes().item(0).getNodeValue();
     tokeniz = new StringTokenizer(matrix_values);

     for (int k=0; k<j; k++) {
       ris.setValue(categories[j], categories[k], new Double(tokeniz.nextToken()).doubleValue());
     }
   }
   return ris;
 }
}
