/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import kddml.Core.DataMining.SymmetricMatrix;
import kddml.Core.DataMining.*;

/**
 * Given two field values x and y, the similarity matrix stores coordinate-by-coordinate
 * the similarity values.
 * In this case comparison is a symmetric matrix which contains the similarity values
 * for each discrete value.
 * Similarity matrix can be used on discrete attributes only.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SIMILARITY_MATRIX extends AttributeComparisonMeasure {

  /**
   * The similarity matrix.
   */
  protected SymmetricMatrixManager sim_matrix;

  /**
   * Constructor given the similarity matrix.
   *
   * @param similarity_matrix SymmetricMatrixManager
   */
  public SIMILARITY_MATRIX(SymmetricMatrixManager similarity_matrix) {
    sim_matrix = similarity_matrix;
  }

  /**
   * Throws an exception: similarity matrix is applied to numeric attributes only.
   *
   * @param value1 String
   * @param value2 String
   * @throws ClusteringModelException
   * @return double
   */
  protected double computeNumeric(double value1, Double value2) throws ClusteringModelException {

    String msg = "Can't use similarity matrix for numeric attribute!";
    throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
  }

  /**
   * Computes the similarity matrix for a discrete attribute.
   *
   * @param value1 String the first category to compare. Cannot be null.
   * @param value2 String the second category to compare. Cannot be null.
   * @throws ClusteringModelException
   * @return double.
   */
  protected double computeDiscrete(String value1, String value2)
      throws ClusteringModelException {

    if (value2 == null) {
      String msg = "Can't compute similarity matrix for missing attribute values!";
      throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
    }

    try {
      double ris = sim_matrix.getValue(value1, value2);
      return ris;
    }
    catch (MiningModelException ex) {
      String msg = "Can't compute similarity matrix for value "+value1+" and "+value2;
      throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
    }
  }

  /**
   * Returns the function name.
   *
   * @return String
   */
  public String getFunctionName() {
    return "table";
  }
}
