/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Clustering;

import weka.core.Instance;

/**
 * The simple matching function is an outer function, used to compare two binary or
 * categorical records X, Y where:
 * <ul>
 * <li> a11 is the number of times where Xi=1 and Yi=1</li>
 * <li> a10 is the number of times where Xi=1 and Yi=0</li>
 * <li> a01 is the number of times where Xi=0 and Yi=1</li>
 * <li> a00 is the number of times where Xi=0 and Yi=0</li>
 * </ul>
 * The simple matching function is defined as:
 * <br>
 * <p align="center">
 * D = ( a11 + a00 ) / ( a11 + a10 + a01 + a00 ).
 * </p><br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SIMPLE_MATCHING extends CentroidBasedComparisonMeasure {

  /**
   * Constructor given the list of comparison measures. Used for centroid-based
   * clustering only.
   *
   * @param attribute_comparison AttributeComparisonMeasure[] the comparison measure
   * for each attribute.
   * @param weights double[] the field weight for each attribute. Can be null.
   */
  public SIMPLE_MATCHING(AttributeComparisonMeasure[] attribute_comparison, double[] weights) {
    super(attribute_comparison, weights);
  }

  /**
   * Constructor given the centroid description. Used both for centroid-based clustering
   * and distribution-based clustering.
   *
   * @param cluster_description ClusterDescription
   */
  public SIMPLE_MATCHING(ClusterDescription cluster_description) {
    super(cluster_description);
  }

  /**
   * Returns true if the comparison measure is a similarity function, in which the
   * value returned by the compare method is optimal for greater values.
   * Returns false if the comparison measure is a distance measure, in which the
   * value returned by the compare method is optimal if it is 0.
   *
   * @return boolean true
   */
  public boolean isSimilarityMeasure() {
    return true;
  }

  public double compare(ClusterManager cluster, Instance instance)
      throws ClusteringModelException {

    int a11 = 0;
    int a10 = 0;
    int a01 = 0;
    int a00 = 0;

    // get the centroid from the cluster...
    double[] x = cluster.getSeedCoordinate();

    // get the coordinate instance...
    double[] y = new double[x.length];
    for (int i=0; i<instance.numAttributes(); i++) {
      if (instance.attribute(i).isNumeric()) {
        // numeric attribute --> error
        String msg = "Can't compute absolute simple matching for numeric attribute "+instance.attribute(i).name();
        throw new ClusteringModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_MODEL);
      }

      String value;
      if (instance.isMissing(i))
        // get array of 0
        value = "?";
      else
        value = instance.stringValue(i);
      double[] d = ((Cluster) cluster).normalizeFieldValue(instance.attribute(i).name(), value);
      for (int j=0; j<d.length; j++)
        y[i+j] = d[j];
    }

    for (int i=0; i<x.length; i++) {

      // discrete attribute
      if (!instance.isMissing(i)) {
        if ((x[i] == 1) && (y[i] == 1))
          a11++;
        if ((x[i] == 1) && (y[i] == 0))
          a10++;
        if ((x[i] == 0) && (y[i] == 1))
          a01++;
        if ((x[i] == 0) && (y[i] == 0))
          a00++;

      }
    }
    if (a11 + a10 + a01 + a00 == 0)
      return 0;
    else
      return (a11 + a00) / (a11 + a10 + a01 + a00);
  }

  /**
   * Evaluates the comparison measure given the comparison values for each single
   * attribute (i.e., the values returned by each single inner function).
   *
   * @param x double[]
   * @return double 0
   */
  protected double evaluate(double[] x) {
    return 0;
  }

  /**
   * Returns the function name as in PMML.
   *
   * @return String
   */
  public String getFunctionName() {
    return "simpleMatching";
  }
}
