/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */



package kddml.Core.DataMining;

import java.util.Vector;
import java.util.Iterator;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;
import kddml.Core.DataSources.*;
import java.util.Enumeration;

/**
 * The data dictionary defines the fields that are the inputs to models and
 * specifies the type and value range for each field.
 * These definitions are assumed to be independent of specific data sets as used for
 * training or scoring a specific model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DataDictionary implements DataDictionaryManager {

  /**
   * List of fields.
   */
  protected Vector fields;

  /**
   * Empty constructor.
   */
  public DataDictionary() {
    fields = new Vector();
  }

  /**
   * Constructor given a set of data statistics.
   *
   * @param statistic DataStatisticsManager
   */
  public DataDictionary(DataStatisticsManager statistic) {

    String name;
    AttributeStatisticManager attr;
    fields = new Vector();
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      name = attr.getName();
      if (attr.getType().equals(AttributeType.NUMERIC)) {
        addField(new DataField(name, DataField.CONTINUOUS));
      }
      else {
        if (attr.getType().equals(AttributeType.NOMINAL)) {
          // attributo di tipo enumerato
          DataField mf = new DataField(name, DataField.CATEGORICAL);

          Enumeration enum2 = ((NominalStatistic) attr).getValues();
          while(enum2.hasMoreElements()) {
            String s = (String) enum2.nextElement();
            try {
              mf.addEnumeratedValue(s);
            } catch (Exception e) {
            }
          }
          addField(mf);
        }
        else {
          // attributo di tipo stringa
          addField(new DataField(name));
        }
      }
    }
  }

  /**
   * Adds a field to this data dictionary.
   *
   * @param f DataFieldManager
   * @return boolean
   */
  public boolean addField(DataFieldManager f) {
    if (!fields.contains(f)) {
      fields.add(f);
      return true;
    }
    return false;
  }

  /**
   * Returns the number of fields of the data dictionary.
   *
   * @return int
   */
  public int getNumberOfFields() {
    return fields.size();
  }

  /**
   * Returns the list of fields as iterator.
   *
   * @return Iterator
   */
  public Iterator getFields() {
    return fields.iterator();
  }

  /**
   * Return the data field naming name or null if it do not exist.
   *
   * @param name String
   * @return DataFieldManager
   */
  public DataFieldManager getField(String name) {
    for (int i=0; i<fields.size(); i++) {
      if (((DataField) fields.get(i)).getName().equalsIgnoreCase(name))
        return (DataField) fields.get(i);
    }
    return null;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "Data Dictionary. List of fields: \n";
    for (int i = 0; i< fields.size(); i++) {
      ris = ris + ((DataField) fields.get(i)).toString()+"\n";
    }
    return ris;
  }


  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("DataDictionary");
    for (int i = 0; i< fields.size(); i++) {
      Element data_field = ((DataField) fields.get(i)).toXML();
      ris.appendChild(data_field);
    }
    ris.setAttribute("numberOfFields",new Integer(fields.size()).toString());
    return ris;
  }

  /**
   * Compares two data dictionary
   *
   * @param data_dictionary DataDictionaryManager
   * @return boolean
   */
  public boolean equals(DataDictionaryManager data_dictionary) {
    DataDictionary dd = (DataDictionary) data_dictionary;
    Iterator iter1 = this.getFields();

    if (this.getNumberOfFields() != dd.getNumberOfFields())
      return false;
    while(iter1.hasNext()) {
      DataField df1 = (DataField) iter1.next();

      if (!dd.contains(df1))
        return false;
    }
    return true;
  }

  /**
   * Clone this object.
   *
   * @return Object
   */
  public java.lang.Object clone() {
    DataDictionary output = new DataDictionary();
    Iterator iter = this.getFields();
    while(iter.hasNext()) {
      output.addField((DataField) ((DataField) iter.next()).clone());
    }
    return output;
  }

  /**
   * Tests if the data dictionary contains a given data field.
   *
   * @param m DataFieldManager
   * @return boolean
   */
  public boolean contains(DataFieldManager m) {

    DataField f;
    for (int i=0; i<fields.size(); i++) {
      f = (DataField) fields.get(i);
      if (f.equals(m))
        return true;
    }
    return false;
  }
}
