/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

import org.w3c.dom.Element;
import kddml.Core.XMLUtility;
import java.util.Vector;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * The field used in DataDictionary.
 * The name of a data field must be unique in the data dictionary.
 * The displayName is a string which may be used by applications to refer to that field.
 * Within the XML document only the value of name is significant.
 * If displayName is not given, then name is the default value.
 * <br>
 * Field type can be:
 * <ol>
 * <li> categorical for discrete attributes;</li>
 * <li> ordinal for string attributes;</li>
 * <li> continuous for numeric attributes;</li>
 * </ol>
 * <br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DataField implements DataFieldManager {

  /**
   * Categorical attribute.
   */
  public final static byte CATEGORICAL = 0;

  /**
   * Ordinal attribute.
   */
  public final static byte ORDINAL = 1;

  /**
   * Continuous attribute.
   */
  public final static byte CONTINUOUS = 2;

  /**
   * Name of the field.
   */
  protected String name;

  /**
   * Display name of the field.
   * The displayName is a string which may be used by applications to refer to that field.
   * If displayName is not given, then name is the default value.
   */
  protected String display_name;

  /**
   * Type of the field.
   * Can be categorical, ordinal, continuous.
   * Categorical fields have the operator "=", ordinal fields have an additional "<",
   * and continuous fields also have arithmetic operators.
   */
  protected byte type;

  /**
   * The list of values for an enumerated attribute.
   */
  protected Vector enumerated_values;

  /**
   * Constructor.
   * Default type: ordinal attribute.
   *
   * @param the name of the field.
   */
  /**
   * Constructor give the field name. It constructs an ordinal attribute.
   *
   * @param name String
   */
  public DataField(String name) {
    this.name = name;
    type = ORDINAL;
    display_name = null;
    enumerated_values = new Vector();
  }

  /**
   * Constructor given the field name and the type.
   *
   * @param name String
   * @param type byte
   */
  public DataField(String name, byte type) {
    this.name = name;
    this.type = type;
    display_name = null;
    enumerated_values = new Vector();
  }

  /**
   * Adds a new category to a discrete field.
   * Throws an exception if the field is not discrete.
   *
   * @param value String
   * @throws MiningModelException
   */
  public void addEnumeratedValue(String value) throws MiningModelException {
    if (this.getType() == CATEGORICAL) {
      if (!enumerated_values.contains(value))
        enumerated_values.add(value);
    }
    else
      throw new MiningModelException("Field "+name+" is not discrete!", null,
                                     KDDMLCoreErrorCodes.INVALID_SETTINGS);
  }

  /**
   * Returns enumerated values for a nominal attribute as array of strings.
   * Return an empty array if the attribute is not discrete.
   *
   * @return String[]
   */
  public String[] getEnumeratedValues() {
    Object[] e = enumerated_values.toArray();
    String[] out = new String[enumerated_values.size()];
    for (int i=0; i<out.length; i++)
      out[i] = (String) e[i];
    return out;
  }

  /**
   * Compares two data fields. Returns true two data fields are equal.
   *
   * @param data_field DataFieldManager
   * @return boolean
   */
  public boolean equals(DataFieldManager data_field) {

    if (!this.name.equals(data_field.getName()))
      return false;

    if (this.type != data_field.getType())
      return false;

    if (type == DataField.CATEGORICAL) {
      String[] s = data_field.getEnumeratedValues();
      for (int i=0; i<s.length; i++) {
        if (!this.enumerated_values.contains(s[i]))
          return false;
      }
    }

    return true;
  }

  /**
  * Sets the displayed name of the field.
   *
   * @param display_name String
   */
  public void setDisplayedName(String display_name) {
    this.display_name = display_name;
  }

  /**
   * Sets the type of the field.
   *
   * @param type byte
   */
  public void setType(byte type) {
    this.type = type;
  }

  /**
   * Returns the name of the data field.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Return the displayed name of the field.
   *
   * @return String
   */
  public String getDisplayedName() {
    return display_name;
  }

  /**
   * Return the type of the field.
   *
   * @return byte
   */
  public byte getType() {
    return type;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    return name + " (" + getStringType() + ")";
  }

  /**
   * Returns a string representation of the type.
   *
   * @return String
   */
  public String getStringType() {
    switch(type) {
      case CATEGORICAL:
        return "categorical";
      case CONTINUOUS:
        return "continuous";
      default:
        return "ordinal";
    }
  }

  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("DataField");
    XMLUtility.setAttribute(ris, "name", name);
    ris.setAttribute("optype", getStringType());
    if (display_name != null)
      XMLUtility.setAttribute(ris, "displayName", display_name);
    // appende i valori dell'attributo se di tipo nominale
    if (type == CATEGORICAL) {
      for (int j=0; j<enumerated_values.size(); j++) {
        Element el = XMLUtility.doc.createElement("Value");
        String s = (String) enumerated_values.get(j);
        XMLUtility.setAttribute(el, "value", s);
        ris.appendChild(el);
      }
    }
    return ris;
  }

  public java.lang.Object clone() {
    DataField output = new DataField(this.getName());
    output.setDisplayedName(this.display_name);
    output.setType(this.getType());
    if (this.getType() == CATEGORICAL) {
      String[] s = this.getEnumeratedValues();
      for (int i=0; i<s.length; i++)
        try {
          output.addEnumeratedValue(s[i]);
        }
        catch (MiningModelException ex) {
        }
    }
    return output;
  }
}
