/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * A mining field is a single attribute used in MiningSchema.
 * Each mining field is composed by a name and an usage type. <br>
 * The name of a field is used as an identifier within the PMML document.
 * When a application uses a PMML model it binds the actual parameter values to
 * the input fields in the MiningSchema.<br>
 * The usage type can be:
 * <ol>
 * <li> active: field used as input (independent field).</li>
 * <li> predicted: field whose value is predicted by the model.</li>
 * <li> supplementary: field holding additional descriptive information. Supplementary
 * fields are not required to apply a model. They are provided as additional information
 * for explanatory purpose, though. When some field has gone through preprocessing
 * transformations before a model is built, then an additional supplementary field
 * is typically used to describe the statistics for the original field values.</li>
 * <li>group: field similar to the SQL "group by". For example, this is used by the association
 * and sequence models to group items into transactions by customerID or by transactionID. </li>
 * </ol>
 * <br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class MiningField implements MiningFieldManager {

  public final static byte ACTIVE = 0;

  public final static byte PREDICTED = 1;

  public final static byte SUPPLEMENTARY = 2;

  public final static byte GROUP = 3;

  /**
   * Name of the field.
   */
  protected String name;

  /**
   * The usage Type. Can be: active, predicted, supplementary, group.
   */
  protected byte type;

  /**
   * Constructor given the name of the field. Its type become active.
   *
   * @param name String
   */
  public MiningField(String name) {
    this.name = name;
    this.type = ACTIVE;
  }

  /**
   * Constructor given the name and the usage type.
   *
   * @param name String
   * @param usage byte
   */
  public MiningField(String name, byte usage) {
    this.name = name;
    this.type = usage;
  }

  /**
   * Returns true if two mining fields are equal.
   * Two mining fields are equal if they have the same name.
   *
   * @param mining_field MiningFieldManager
   * @return boolean
   */
  public boolean equals(MiningFieldManager mining_field) {

    return (this.name.equalsIgnoreCase(mining_field.getName()));
  }

  /**
   * Returns the name of the mining field.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Returns a number identifying the usage type of the mining field.
   *
   * @return byte
   */
  public byte getType() {
    return type;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    return name + " (" + getStringType() + ")";
  }

  private String getStringType() {
    switch(type) {
      case 0:
        return "active";
      case 1:
        return "predicted";
      case 2:
        return "supplementary";
      default:
        return "group";
    }
  }

  public Element toXML() {
    Element el = XMLUtility.doc.createElement("MiningField");
    el.setAttribute("name", name);
    XMLUtility.setAttribute(el, "name", name);
    el.setAttribute("usageType", getStringType());
    return el;
  }

  public java.lang.Object clone() {
    MiningField output = new MiningField(this.getName(), this.getType());
    return output;
  }
}
