/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import java.io.File;
import kddml.Core.*;
import kddml.Core.DataSources.*;

/**
 * KDDML uses PMML to represent mining models. The structure of the models is described by
 * a DTD which is called the PMML DTD.
 * This abstract class is used to represent a generic PMML document. In general, a mining
 * model is composed by:
 * <ol>
 * <li> an header containing information on application name and version;</li>
 * <li> a data dictionary containing definitions for fields as used in mining models.</li>
 * <li> a mining schema containing the list of fields used in that model.</li>
 * </ol>
 * <br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class MiningModel extends kddml.Core.KDDMLObject implements MiningModelManager {

  protected String PMML_version = "2.0";

  protected String copyright =
      "Copyright (c) 2004-2005 - University of Pisa, Department of Computer Science - All Rights Reserved.";

  protected String application_name = "KDDML (Knowledge Discovery in Databases Markup Language)";

  protected String application_version = "2.0";

  /**
   * Name of algorithm used to extract the model.
   */
  protected String algorithm_name = "Unspecified";

  /**
   * Name of the model.
   */
  protected String model_name = "Unspecified";

  /**
   * The DataDictionary Element.
   */
  protected DataDictionary data_dictionary;

  /**
   * The mining schema element.
   */
  protected MiningSchema mining_schema;

  protected MiningModel(File physical_path) {
    super(physical_path);
    data_dictionary = new DataDictionary();
    mining_schema = new MiningSchema();
  }

  /**
   * Returns the data dictionary of the model.
   *
   * @return DataDictionaryManager
   */
  public DataDictionaryManager getDataDictionary() {
    return data_dictionary;
  }

  /**
   * Sets the data dictionary.
   *
   * @param data_dictionary DataDictionaryManager
   */
  public void setDataDictionary(DataDictionaryManager data_dictionary) {
    this.data_dictionary = (DataDictionary) data_dictionary;
  }

  /**
   * Returns the mining schema of the model.
   *
   * @return MiningSchemaManager
   */
  public MiningSchemaManager getMiningSchema() {
    return mining_schema;
  }

  /**
   * Sets the mining schema.
   *
   * @param mining_schema MiningSchemaManager
   */
  public void setMiningSchema(MiningSchemaManager mining_schema) {
    this.mining_schema = (MiningSchema) mining_schema;
  }

  /**
   * Returns the model name.
   *
   * @return String
   */
  public String getModelName() {
    return model_name;
  }

  /**
   * Sets the model name.
   *
   * @param model_name String
   */
  public void setModelName(String model_name) {
    this.model_name = model_name;
  }

  /**
   * Returns the algorithm name.
   *
   * @return String
   */
  public String getAlgorithmName() {
    return algorithm_name;
  }

  /**
   * Sets the algorithm name.
   *
   * @param algorithm_name String
   */
  public void setAlgorithmName(String algorithm_name) {
    this.algorithm_name = algorithm_name;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "------------------------------------------"+"\n";
    s = s + "PMMLModel - Name: "+model_name + "\n";
    s = s + "Algorithm: "+algorithm_name + "\n";
    s = s + "------------------------------------------"+"\n";
    s = s + data_dictionary.toString();
    s = s + "\n";
    s = s + mining_schema.toString();
    return s;
  }

  /**
   * Clone this model and return a new mining model as output.
   *
   * @return Object
   */
  //public abstract MiningModel clone(File output_path);

  /**
   * Initialize the XML document to be saved into the repository.
   *
   * @throws KDDMLCoreException
   * @return Document the DOM output PMML document
   */
  protected Document initializeDocument() throws KDDMLCoreException {

    Document pmml_model = XMLUtility.initializeDocument("PMML");

    pmml_model.getDocumentElement().setAttribute("version", PMML_version);
    Element header = XMLUtility.doc.createElement("Header");
    pmml_model.getDocumentElement().appendChild(header);
    XMLUtility.setAttribute(header, "copyright", copyright);
    Element application = XMLUtility.doc.createElement("Application");
    header.appendChild(application);
    XMLUtility.setAttribute(application, "name", application_name);
    application.setAttribute("version", application_version);

    if (data_dictionary != null) {
      Element el = data_dictionary.toXML();
      pmml_model.getDocumentElement().appendChild(el);
    }

    return pmml_model;
  }

  /**
   * Tests if the model is compatible with an input data schema.
   * A model is compatible with a data schema if for each active mining field there is
   * an attribute in the data in witch the signature (i.e. name and type) conincides with
   * the mining field.
   *
   * @param instances DataStatistics
   * @return boolean
   */
  public boolean isCompatible(DataStatisticsManager instances) {

    java.util.Iterator iter = mining_schema.getFields();
    while (iter.hasNext()) {
      MiningField m_f = (MiningField) iter.next();
      DataField d_f = (DataField) data_dictionary.getField(m_f.getName());
      if (m_f.getType() == MiningField.ACTIVE) {
        AttributeStatisticManager stat = instances.getStatistic(m_f.getName());
        if (stat == null)
          return false;
        if ((stat.getType().equals(AttributeType.NUMERIC)) &&
            (d_f.getType() != d_f.CONTINUOUS))
          return false;
        if ((stat.getType().equals(AttributeType.NOMINAL)) &&
            (d_f.getType() != d_f.CATEGORICAL))
          return false;
        // controllare ogni singola categoria per l'attributo nominale?
      }
    }
    return true;
  }

  protected static Element appendMiningSchema(MiningSchema schema, Element model) {
    Element el = schema.toXML();
    model.appendChild(el);
    return model;
  }
}
