/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

import java.util.Vector;
import java.util.Iterator;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * Each model contains one mining schema, which lists the fields used in
 * the model. These fields are a subset of the fields in the Data Dictionary. The mining
 * schema contains information that is specific to a certain model, while the data dictionary
 * contains data definitions that do not vary with the model. For example, the Mining
 * Schema specifies the usage type of an attribute, which may be active (an input of the
 * model), predicted (an output of the model), or supplementary (holding descriptive
 * information and ignored by the model).
 * The mining schema contains the list of fields used in that model. This is a subset
 * of the fields as defined in the data dictionary. The main purpose of the mining
 * schema is to list the fields which a user has to provide in order to apply the model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class MiningSchema implements MiningSchemaManager {

  /**
   * List of mining fields.
   */
  protected Vector fields;

  /**
   * Creates an empty mining schema with no fields.
   */
  public MiningSchema() {
    fields = new Vector();
  }

  /**
   * Adds a new mining field to this mining shema.
   *
   * @param mf MiningFieldManager
   * @return boolean
   */
  public boolean addField(MiningFieldManager mf) {
    if (!fields.contains(mf)) {
      fields.add(mf);
      return true;
    }
    return false;
  }

  /**
   * Returns the number of fields belonging to this mining schema.
   *
   * @return int
   */
  public int getNumberOfFields() {
    return fields.size();
  }

  /**
   * Returns the list of mining fields as iterator.
   *
   * @return Iterator
   */
  public Iterator getFields() {
    return fields.iterator();
  }

  /**
   * Returns the field identified by name, or null if it do not exist.
   *
   * @param name String
   * @return MiningFieldManager
   */
  public MiningFieldManager getField(String name) {
    for (int i=0; i<fields.size(); i++) {
      if (((MiningField) fields.get(i)).getName().equalsIgnoreCase(name))
        return (MiningField) fields.get(i);
    }
    return null;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "Mining schema. List of fields: \n";
    for (int i = 0; i< fields.size(); i++) {
      ris = ris + ((MiningField) fields.get(i)).toString()+"\n";
    }
    return ris;
  }

  protected Element toXML() {
    Element ris = XMLUtility.doc.createElement("MiningSchema");
    for (int i = 0; i< fields.size(); i++) {
      Element mining_field = ((MiningField) fields.get(i)).toXML();
      ris.appendChild(mining_field);
    }
    return ris;
  }

  /**
   * Compares two mining schema.
   * Two mining schema are equal if they share the same mining fields.
   *
   * @param mining_schema MiningSchemaManager
   * @return boolean
   */
  public boolean equals(MiningSchemaManager mining_schema) {

    MiningSchema dd = (MiningSchema) mining_schema;
    Iterator iter1 = this.getFields();

    if (this.getNumberOfFields() != dd.getNumberOfFields())
      return false;
    while(iter1.hasNext()) {
      MiningField df1 = (MiningField) iter1.next();

      if (!dd.contains(df1))
        return false;
    }
    return true;
  }

  public Object clone() {
    MiningSchema output = new MiningSchema();
    Iterator iter = this.getFields();
    while(iter.hasNext()) {
      output.addField((MiningField) ((MiningField) iter.next()).clone());
    }
    return output;
  }

  /**
   * Tests if this mining schema contains an input mining field.
   * Checking is performed by field name.
   *
   * @param m MiningFieldManager
   * @return boolean
   */
  public boolean contains(MiningFieldManager m) {

    MiningField f;
    for (int i=0; i<fields.size(); i++) {
      f = (MiningField) fields.get(i);
      if (f.equals(m))
        return true;
    }
    return false;
  }
}
