/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.Comparator;

/**
 * This class is used to establish an order between the sequence rules.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class CompareSequenceRules implements Comparator {

  /**
   * This method is used to compare two sequence rules. Comparison task
   * follows this order :
   * -Comparison is done using the confidence of the rules.
   * -If rules have the same confidence, then comparison is done using the support.
   * -If rules have the same support, then comparison is done using the number
   * of sequences that compose the rules (short rules are preferred).
   * -A total order is obtained comparing the string representation of the
   * rules.
   *
   * @param seq_rule1 rule to be compared
   * @param seq_rule2 rule to be compared
   * @return This method returns 1, 0 or -1 if seq_rule1 is greater than,
   *   equal to or less than seq_rule2
   */
  public int compare(Object seq_rule1, Object seq_rule2) {

    SequenceRuleManager rule1 = (SequenceRuleManager) seq_rule1;
    SequenceRuleManager rule2 = (SequenceRuleManager) seq_rule2;

    if (rule1.equals(rule2))
      return 0;

    // ordiniamo le regole in base alla confidenza...
    if (rule1.getConfidence().doubleValue() > rule2.getConfidence().doubleValue())
      return -1;
    if (rule1.getConfidence().doubleValue() < rule2.getConfidence().doubleValue())
      return 1;

    // ...a parita' di confidenza consideriamo il supporto...
    if (rule1.getSupport().doubleValue() > rule2.getSupport().doubleValue())
      return -1;
    if (rule1.getSupport().doubleValue() < rule2.getSupport().doubleValue())
      return 1;

    // ...a parita' di supporto preferiamo regole corte
    if (rule1.getNumberOfSequenceElements() < rule2.getNumberOfSequenceElements())
      return -1;
    if (rule1.getNumberOfSequenceElements() > rule2.getNumberOfSequenceElements())
      return 1;

    // total ordering needed...
    return rule1.toString().compareTo(rule2.toString());
  }
}
