/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemManager;
import kddml.Core.DataMining.AssociationRules.ItemsetManager;
import kddml.Core.DataMining.AssociationRules.Transaction;
import kddml.Core.DataMining.AssociationRules.TransactionManager;

import java.util.TreeSet;
import java.util.Iterator;

/**
 * This class represents a specialized sequence element implemented as an itemset.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class ElementItemSet extends SequenceElement implements ItemsetManager{

  /**
   * Support of this element (Optional)
   */
  protected Double support;

  /**
   * Build an empty sequence element
   */
  public ElementItemSet() {
    super();
    support = null;
  }

  /**
   * Build a new sequence element with given items.
   * No checking on structure are done.
   *
   * @param items a TreeSet containing Item object
   */
  public ElementItemSet(TreeSet items) {
    super(items);
    support = null;
  }

  /**
   * Set support for this sequence element.
   * Support must be expressed as probabilistic number between
   * 0 and 1 inclusive.
   *
   * @param support support of this sequence element
   */
  public void setSupport(double support) {
    this.support = new Double(support);
  }

  /**
   * Return the support of this sequence element, or null if not specified.
   *
   * @return element support
   */
  public Double getSupport() {
    return support;
  }

  /**
   * Return absolute support, i.e. the number of occurrence of this element
   * in input data.
   *
   * @param num_transactions int
   * @return absolute support
   */
  public int getAbsoluteSupport(int num_transactions) {
    double d = support.doubleValue() * num_transactions;
    return (int) d;
  }

  /**
   * An alias for getNumberOfItems().
   *
   * @return number of items
   */
  public int getSize() {
    return super.getNumberOfItems();
  }

  /**
   * Return true if given object is equal to this one.
   * If given object is a generic element, this method siply recall
   * SequenceElement.equals().
   * If given object is an itemset then Itemset.equals(this_object) is called.
   *
   * @param sequence_element Object
   * @return boolean
   */
  public boolean equals(Object sequence_element){
    if( sequence_element instanceof ItemsetManager ) {

      if (this.getSize() != ((ItemsetManager) sequence_element).getSize())
        return false;
      Iterator it1 = this.getItems();
      Iterator it2 = ((ItemsetManager) sequence_element).getItems();
      boolean b;
      while (it1.hasNext()) {
        Item item1 = (Item) it1.next();
        Item item2 = (Item) it2.next();
        b = item1.equals(item2);
        if (!b)
          return false;
      }
      return true;

      //return ((ItemsetManager)sequence_element).equals(this);
    }
    return super.equals(sequence_element);
  }

  /**
   * Returns the list of items as transaction.
   *
   * @return a transaction
   */
  public TransactionManager toTransaction() {
    Iterator iter = getItems();
    Item[] s = new Item[this.getNumberOfItems()];
    int i=0;
    while(iter.hasNext()) {
      Item item = (Item) iter.next();
      s[i] = item;
      i++;
    }
    return (TransactionManager) new Transaction(s);
  }

  /**
   * Return a string representation of this object
   *
   * @return this object as string
   */
  public String toString() {
    if (items.size() == 0)
      return "Empty set!!!";
    else {
      String s = "";
      Iterator iter = items.iterator();
      while (iter.hasNext())
        s = s + ((ItemManager) iter.next()).toString() + ", ";
      if (support != null)
        s = s + "(supp="+support+")";
      else
        s = s + "(supp=n/a)";
      return s;
    }
  }

}
