/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.AssociationRules.*;
import kddml.Core.DataMining.*;
import kddml.Core.*;
import org.w3c.dom.*;
import java.io.*;
import java.util.*;

/**
 * Factory class that builds asequence  model from a PMML model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public final class PMMLSequenceFactory extends SequenceExternalFactory {

  private String pmml_version;

  private Element header;

  private Element data_dictionary;

  private Element mining_schema;

  private Element sequence_model;

  private String min_support;

  private String min_confidence;

  private String number_of_transactions;

  private String algorithm_name;

  private String model_name;

  private String max_number_of_items;

  private String avg_number_of_items;

  private String length_limit;

  private String time_window_width;

  private String min_time;

  private String max_time;

  private Hashtable items_ht;

  private Hashtable elements_ht;

  private Hashtable sequences_ht;

  /**
   * Build a new sequence model parsing a given PMML document.
   *
   * @param f File the file in the system repository that will contain the generated model.
   * @param pmml_path String the URI identifying the pmml document.
   * @throws KDDMLCoreException if the source cann't be found or the parser returns
   * errors.
   */
  public PMMLSequenceFactory(java.io.File f, String pmml_path)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    // effettua il parsing del documento XML contenente il modello...
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(pmml_path),
                              new DTDResolver(kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL));
    }
    catch(java.io.IOException e) {
      String msg = "Invalid PMML reference: "+pmml_path;
      throw new SequenceModelException(msg, e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }

    Element root = (Element) doc.getDocumentElement();

    if (!root.getTagName().equalsIgnoreCase("pmml"))
      throw new SequenceModelException(pmml_path+ " no contains a PMML document.",
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the pmml version
    pmml_version = root.getAttribute("version");

    // reading the copyright
    header = (Element) root.getElementsByTagName("Header").item(0);

     // reading the data dictionary element
    data_dictionary = (Element) root.getElementsByTagName("DataDictionary").item(0);

    sequence_model = (Element) doc.getElementsByTagName("SequenceModel").item(0);
    if (sequence_model == null)
      throw new SequenceModelException(pmml_path+ " no contains a PMML sequence model document.",
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the mining schema element...
    mining_schema = (Element) sequence_model.getElementsByTagName("MiningSchema").item(0);

    // reading attributes...
    readAttributes();
  }

  /**
   * Returns the data dictionary of the output model.
   *
   * @throws MiningModelException
   * @return DataDictionary
   */
  protected DataDictionary buildDataDictionary() throws MiningModelException {
    if (data_dictionary == null)
      return new DataDictionary();
    else
      return PMMLUtility.buildDataDictionary(data_dictionary);
  }

  /**
   * Returns the mining schema of the output model.
   *
   * @throws MiningModelException
   * @return MiningSchema
   */
  protected MiningSchema buildMiningSchema() throws MiningModelException {
    if (mining_schema == null)
      return new MiningSchema();
    else
      return PMMLUtility.buildMiningSchema(mining_schema);
  }

  /**
   * Returns the name of the algorithm.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildAlgorithmName() throws MiningModelException {
    return algorithm_name;
  }

  /**
   * Returns the name of the model.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildModelName() throws MiningModelException {
    return model_name;
  }

  /**
   * Returns the PMML version supported.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildPMMLVersion() throws MiningModelException {
    return pmml_version;
  }

  /**
   * Returns the copyright.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildCopyright() throws MiningModelException {
    if (header != null)
      return header.getAttribute("copyright");
    else
      return null;
  }

  /**
   * Returns the application name.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationName() throws MiningModelException {
    if (header != null) {
      Element el = (Element)header.getElementsByTagName("Application").item(0);
      if( el == null ) return null;
      return el.getAttribute("name");
    }
    else
      return null;
  }

  /**
   * Returns the application version.
   *
   * @throws MiningModelException
   * @return String
   */
  protected String buildApplicationVersion() throws MiningModelException {
    if (header != null) {
      Element el = (Element)header.getElementsByTagName("Application").item(0);
      if( el == null ) return null;
      return el.getAttribute("version");
    }
    else
      return null;
  }

  /**
   * Returns the minimum support of the sequences.
   * Mininmum support is a real in (0,1]
   *
   * @throws SequenceModelException
   * @return Double
   */
  protected Double buildMinSupport() throws SequenceModelException {
    Double ris;
    try {
      ris = new Double(min_support);
    }
    catch(Exception e) {
      String msg = "Minimum support must be a real between 0 and 1.";
      throw new SequenceModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the minimum confidence of the sequences.
   * Mininmum confidence is a real in (0,1].
   *
   * @throws SequenceModelException
   * @return Double
   */
  protected Double buildMinConfidence() throws SequenceModelException {
    Double ris;
    try {
      ris = new Double(min_confidence);
    }
    catch(Exception e) {
      String msg = "Minimum confidence must be a real between 0 and 1.";
      throw new SequenceModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the number of transaction used to build the sequences.
   * Returns a positive integer.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildNumberOfTransactions() throws SequenceModelException {
    Integer ris;
    try {
      ris = new Integer(number_of_transactions);
    }
    catch(Exception e) {
      String msg = "Number of transactions must be a positive integer.";
      throw new SequenceModelException(msg,
                                          null,
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    return ris;
  }

  /**
   * Returns the maximum number of iterms per transaction.
   * Returns a positive integer.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildMaxNumberOfItemsPerTA() throws SequenceModelException {
    if (max_number_of_items.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(max_number_of_items);
      }
      catch (Exception e) {
        String msg =
            "The max number of items per transction must be a positive integer.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Returns the average number of iterms per transaction.
   * Returns a positive integer.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Double buildAvgNumberOfItemsPerTA() throws SequenceModelException {
    if (avg_number_of_items.equals(""))
      return null;
    else {
      Double ris;
      try {
        ris = new Double(avg_number_of_items);
      }
      catch (Exception e) {
        String msg =
            "The average number of items per transaction must be a positive real.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Return the maximum length of a sequence to be discovered.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildLengthLimit() throws SequenceModelException {
    if (length_limit.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(length_limit);
      }
      catch (Exception e) {
        String msg =
            "The length limit must be a positive real.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Return the time window width:
   * this may be used to separate items associated with an object into discrete
   * events, but only if no clear key already exists for the separate events.
   * Two consecutive items must have a time gap of less than this value to be
   * considered as being part of the same event.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildTimeWindow() throws SequenceModelException {
    if (time_window_width.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(time_window_width);
      }
      catch (Exception e) {
        String msg =
            "The time window width must be a positive real.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Return the minimum time between items (see buildTimeWindow())
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildMinTime() throws SequenceModelException {
    if (min_time.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(min_time);
      }
      catch (Exception e) {
        String msg =
            "The min time must be a positive real.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Return the maximum time between items (see buildTimeWindow())
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected Integer buildMaxTime() throws SequenceModelException {
    if (max_time.equals(""))
      return null;
    else {
      Integer ris;
      try {
        ris = new Integer(max_time);
      }
      catch (Exception e) {
        String msg =
            "The max time must be a positive real.";
        throw new SequenceModelException(msg,
                                            null,
                                            KDDMLCoreErrorCodes.
                                            INVALID_PMML_DOCUMENT);
      }
      return ris;
    }
  }

  /**
   * Return the list of sequence rules.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected TreeSet buildSequenceRules() throws SequenceModelException {
    try {
      CompareSequenceRules comp = new CompareSequenceRules();
      TreeSet ris = new TreeSet(comp);
      NodeList n = sequence_model.getElementsByTagName("SequenceRule");
      //Get sequence rules element...
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);

        SequenceRuleManager sr = UPMMLParser.parseRule(el, sequences_ht);

        if (!contains(ris, sr))
          ris.add(sr);
      }
      sequences_ht = null;
      System.gc();
      return ris;
    }
    catch(Exception e) {
      throw new SequenceModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Return the list of sequences.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected TreeSet buildSequences() throws SequenceModelException{
    try {
      CompareSequences comp = new CompareSequences();
      TreeSet ris = new TreeSet(comp);
      this.sequences_ht = new Hashtable();
      NodeList n = sequence_model.getElementsByTagName("Sequence");

      for (int i = 0; i < n.getLength(); i++) {

        //Get sequence element...
        Element el = (Element) n.item(i);

        String seq_id = el.getAttribute("id");

        SequenceManager sequence = UPMMLParser.parseSequence(el, elements_ht);

        if (sequences_ht.containsKey(sequence))
          throw new SequenceModelException(
              "Illegal PMML Sequence Model: document already contains sequence <" +
              sequence.toString() + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

        sequences_ht.put(seq_id, sequence);

        if (!contains(ris, sequence))
          ris.add(sequence);

      }
      elements_ht = null;
      System.gc();
      return ris;
    }
    catch(Exception e) {
      throw new SequenceModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Return the list of sequence elements.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected TreeSet buildSets() throws SequenceModelException{
    if( this.items_ht.isEmpty() )
      return this.buildSetPredicates();
    return this.buildItemsets();
  }

  private TreeSet buildSetPredicates() throws SequenceModelException {
    throw new SequenceModelException("Unsupported PMML Document!",
                                     "PMML SetPredicate element not yet supported",
                                     KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    /*try {
      CompareSequenceElements comp = new CompareSequenceElements();
      TreeSet ris = new TreeSet(comp);
      NodeList n = sequence_model.getElementsByTagName("SetPredicate");
      this.elements_ht = new Hashtable();
      for (int i = 0; i < n.getLength(); i++) {
        //SetPredicate Element...
        Element el = (Element) n.item(i);

        String setPredicate_id = el.getAttribute("id");

        SequenceElementManager seqElement = UPMMLParser.parseSetPredicate(el);

        if (elements_ht.containsKey(seqElement))
          throw new SequenceModelException(
              "Illegal PMML Sequence Model: document already contains predicate set <" +
              seqElement.toString() + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

        this.elements_ht.put(setPredicate_id, seqElement);
        if (!contains(ris, seqElement))
          ris.add(seqElement);
      }
      this.items_ht = null;

      return ris;
    }
    catch(Exception e) {
      throw new SequenceModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }*/
  }

  private TreeSet buildItemsets() throws SequenceModelException {
    try {
      CompareSequenceElements comp = new CompareSequenceElements();
      TreeSet ris = new TreeSet(comp);
      NodeList n = sequence_model.getElementsByTagName("Itemset");
      this.elements_ht = new Hashtable();
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);

        String itemset_id = el.getAttribute("id");

        SequenceElementManager seqElement = UPMMLParser.parseItemSet(el, items_ht);

        if (elements_ht.containsKey(seqElement))
          throw new SequenceModelException(
              "Illegal PMML Sequence Model: document already contains itemset <" +
              seqElement.toString() + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

        this.elements_ht.put(itemset_id, seqElement);

        if (!contains(ris, seqElement))
          ris.add(seqElement);
      }
      this.items_ht = null;
      System.gc();
      return ris;
    }
    catch(Exception e) {
      throw new SequenceModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  /**
   * Return the list of items.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected TreeSet buildItems() throws SequenceModelException {
    try {
      CompareItems comp = new CompareItems();
      TreeSet ris = new TreeSet(comp);
      NodeList n = sequence_model.getElementsByTagName("Item");
      // tabella hash contenente l'associazione chiave --> item
      this.items_ht = new Hashtable();
      for (int i = 0; i < n.getLength(); i++) {
        Element el = (Element) n.item(i);
        String itemId = el.getAttribute("id");

        ItemManager item = UPMMLParser.parseItem(el);

        if (items_ht.containsKey(item))
          throw new SequenceModelException(
              "Illegal PMML Sequence Model: document already contains item <" +
              item.getValue() + ">",
              null,
              KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

        this.items_ht.put(itemId, item);
        if (!contains(ris, item))
          ris.add(item);
      }
      return ris;
    }
    catch(Exception e) {
      throw new SequenceModelException("Invalid PMML Document!",
                                          e.toString(),
                                          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
  }

  private static boolean contains(TreeSet set, ItemManager item) {
    Iterator iter = set.iterator();
    while (iter.hasNext()) {
      Item i = (Item) iter.next();
      if (i.equals((Item) item))
        return true;
    }
    return false;
  }

  private static boolean contains(TreeSet elements, SequenceElementManager element) {
    Iterator iter = elements.iterator();
    while (iter.hasNext()) {
      SequenceElement i = (SequenceElement) iter.next();
      if (i.equals(element))
        return true;
    }
    return false;
  }

  private static boolean contains(TreeSet sequences, SequenceManager sequence) {
    Iterator iter = sequences.iterator();
    while (iter.hasNext()) {
      SequenceManager i = (SequenceManager) iter.next();
      if (i.equals(sequence))
        return true;
    }
    return false;
  }

  private static boolean contains(TreeSet rules, SequenceRuleManager rule) {
    Iterator iter = rules.iterator();
    while (iter.hasNext()) {
      SequenceRuleManager i = (SequenceRuleManager) iter.next();
      if (i.equals(rule))
        return true;
    }
    return false;
  }

  private void readAttributes() {
    model_name = sequence_model.getAttribute("modelName");
    algorithm_name = sequence_model.getAttribute("algorithmName");

    number_of_transactions = sequence_model.getAttribute("numberOfTransactions");
    max_number_of_items = sequence_model.getAttribute("maxNumberOfItemsPerTransaction");
    avg_number_of_items = sequence_model.getAttribute("avgNumberOfItemsPerTransaction");

    min_support = sequence_model.getAttribute("minimumSupport");
    min_confidence = sequence_model.getAttribute("minimumConfidence");

    length_limit = sequence_model.getAttribute("lengthLimit");
    time_window_width = sequence_model.getAttribute("timeWindowWidth");
    min_time = sequence_model.getAttribute("minTime");
    max_time = sequence_model.getAttribute("maxTime");
  }

  /**
   * Only for testing
   */
  public static void main(String args[]) {
    try{
      String path = "G:/WINDOWS/jbprojects/kddml/repository/models/sequences/sequential_rules.xml";
      File file = new File(path);
      PMMLSequenceFactory f = new PMMLSequenceFactory(file, path);
      System.out.println("Parsed...");
      /*TreeSet t = f.buildItems();
      System.out.println(t.toString());

      t = f.buildSets();
      System.out.println(t.toString());

      t = f.buildSequences();
      System.out.println(t.toString());*/

      SequenceModel sm = (SequenceModel)f.newInstance();
      //System.out.println(o.toString());


      //sm.clone(new File(path+"-result.xml")).saveToRepository();
      sm.saveHTML();

    }catch(KDDMLCoreException e){
      System.out.println("###"+e.getMessage()+"###");
      System.out.println("###"+e.getDetailErrorMessage()+"###");
      e.printStackTrace(); }
  }
}
