/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.TreeSet;
import java.util.*;
import java.lang.*;

/**
 * This class is used to build a sequence model from a given set of rules.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public final class RulesSequenceFactory extends SequenceProprietaryFactory {

  private int max_sequences;
  private int max_rules;

  /**
   * Constructor.
   *
   * @param f the file containing the model
   * @param dd the data dictionary
   * @param ms the mining schema
   * @param min_support the minimum support
   * @param min_confidence the minimum confidence
   * @param number_of_transactions the number of transactions
   * @param sequences the entire set of sequences, from which rules was built
   * @param sequence_rules the ordered set of rules
   * @throws KDDMLCoreException
   */
  public RulesSequenceFactory(java.io.File f,
                                 DataDictionary dd,
                                 MiningSchema ms,
                                 double min_support,
                                 double min_confidence,
                                 int number_of_transactions,
                                 TreeSet sequences,
                                 TreeSet sequence_rules)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);
    this.sequences = sequences;
    this.rules = sequence_rules;
    max_sequences = SequenceModel.MAX_NUMBER_OF_SEQUENCES;
    max_rules = SequenceModel.MAX_NUMBER_OF_RULES;
  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    SequenceModel model = (SequenceModel) super.newInstance();

    SequenceModel.MAX_NUMBER_OF_SEQUENCES = max_sequences;
    SequenceModel.MAX_NUMBER_OF_RULES = max_rules;

    this.elements = USequenceElementBuilder.compute(this.sequences);
    //if elements instances are Itemsets build items...
    this.items = UItemsBuilder.compute(this.elements);
    if(this.items == null )
      this.items = new TreeSet(new kddml.Core.DataMining.AssociationRules.CompareItems());

    model.items = items;
    model.elements = elements;
    model.sequences = sequences;
    model.rules = new TreeSet(new CompareSequenceRules());

    //add rules...
    SequenceRuleManager[] array_rules = (SequenceRuleManager[])
        rules.toArray(new SequenceRuleManager[]{});
    model.addRules(array_rules);

    return model;
  }

  /**
   * Set maximum number of sequences allowed by this model
   *
   * @param max_sequences max number of sequences
   */
  public void setMaxNumberOfSequences(int max_sequences) {
    this.max_sequences = max_sequences;
  }

  /**
   * Set maximum number of rules allowed by this model
   *
   * @param max_rules max number of rules
   */
  public void setMaxNumberOfRules(int max_rules) {
    this.max_rules = max_rules;
  }

  public static void main(String args[]) {
    /*try {
       String s =
           "G:/WINDOWS/jbprojects/kddml/daniele_sequence.xml";
       SequenceRepositoryFactory f = new SequenceRepositoryFactory(new java.io.File(s));

       SequenceModel m = (SequenceModel) f.newInstance();
       //System.out.println(m.toString());

       //m.clone(new java.io.File(s+"-result.xml")).saveToRepository();
       //m.saveHTML();

       //TreeSet rules = m.rules;
       TreeSet rules = URuleBuilder.compute(m.sequences,0.1,m.getNumberOfTransactions().intValue(),5000);
       String s2 = "G:/WINDOWS/jbprojects/kddml/daniele_rules.xml";
       DataDictionary d = (DataDictionary)m.getDataDictionary();
       MiningSchema ms = (MiningSchema)m.getMiningSchema();
       RulesSequenceFactory f2 = new RulesSequenceFactory(
           new java.io.File(s2), d, ms, m.getMinimumSupport().doubleValue(), 0.1, m.getNumberOfTransactions().intValue(), rules);
       f2.setAlgorithmName("Internal");
       SequenceModel model = (SequenceModel) f2.newInstance();
       //System.out.println(model.toString());
       model.saveToRepository();
       model.saveHTML();

     }
     catch(KDDMLCoreException e) {
       System.out.println("########################");
       e.printStackTrace();
       System.out.println("########################");
       System.out.println(e.getDetailErrorMessage());
     }*/

  }

}
