/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemManager;

/**
 * A sequence consists of:
 * <ol>
 * <li> Optional attributes (support and occurrence); </li>
 * <li> A list of sets composing the sequences (SequenceElementManager); </li>
 * </ol>
 * A sequence is structured as a set of consecutive elements
 * (SequenceElementManager).
 * Each of this elements is separated by the one that follows with a delimiter
 * (DelimiterManager).
 *
 * A sequence must contain at least one element. To force this concept, basic
 * constructor require a SequenceElementManager as parameter.
 * Each following element, must be preceded by a delimiter, that specify the
 * kind of relation between the elements. Again, here, to force this concept,
 * the addElement() method, require both delimiter and element.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class Sequence implements SequenceManager{

  /** A list containing elements in sequence */
  protected ArrayList elements;

  /** Sequence occurrence */
  protected Integer occurrence;

  /** Sequence support */
  protected Double support;

  /**
   * Build a sequence with one element.
   *
   * @param element sequence element that forms this sequence
   */
  public Sequence(SequenceElementManager element) {
    this.elements = new ArrayList();
    this.elements.add(element);
  }

  /**
   * Build a new sequence composed by elements present in given sequence.
   *
   * @param sm input sequence
   */
  public Sequence(SequenceManager sm){
    this.elements = new ArrayList();

    Iterator iterator = sm.getElements();
    while( iterator.hasNext() ){
      this.elements.add( (SequenceElementManager) iterator.next() );
      if( iterator.hasNext() )
        this.elements.add( (DelimiterManager)iterator.next() );
    }
  }

  /**
   * Internal constructor
   *
   * @param elements List
   */
  private Sequence(List elements) {
    this.elements = new ArrayList(elements);
  }
  /**
   * Add an element to this sequence.
   *
   * @param d delimiter between last element and the new one
   * @param el new element to be added
   * @return true if element is added, false otherwise
   */
  public boolean addElement(DelimiterManager d, SequenceElementManager el) {
    return (this.elements.add(d) && this.elements.add(el));
  }

  /**
   * Check if given data sequence support this sequence.
   *
   * For each element e in this sequence, there is a transaction t such that
   * e.isSupported(t) holds true, and is respected element order.
   * E.g. s = (e1)...(em); d = (t1)...(tn)
   * 1) m <= n
   * 2) e(1).isSupported(ti) => e(2).isSupported(tj) , j > i
   *
   * @param dseq data sequence to be checked
   * @return true if given data sequence support this sequence, false otherwise
   */
  public boolean isSupported(DataSequenceManager dseq){
    if( this.getNumberOfElements() > dseq.getNumberOfTransactions() )
      return false;

    Iterator elements_iterator = this.getElements();
    Iterator transactions_iterator = dseq.getTransactions();
    boolean transactions_ended = false;
    boolean transaction_supported = false;

    // get all elements in this sequence
    while ( elements_iterator.hasNext() && !transactions_ended ) {

      SequenceElementManager element = (SequenceElementManager)
          elements_iterator.next();
      // skip delimiter
      if (elements_iterator.hasNext()) { elements_iterator.next(); }

      transaction_supported = false;
      // check elements in this sequence, until one is supported by
      // current transaction
      while( transactions_iterator.hasNext() && !transaction_supported ) {
        TimestampedTransactionManager transaction = (
            TimestampedTransactionManager) transactions_iterator.next();
        transaction_supported = element.isSupported(transaction);
      }

      if( !transactions_iterator.hasNext() ) {
        transactions_ended = true;
      }

    }

    // If there are no more sequence elements, but there are still
    // some transctions, then given data sequence support this sequence
    if( !transactions_ended ) {
      return true;
    }
    // If there are no more transactions, this sequence is supported only if
    // last element is supported by last transaction
    if( !elements_iterator.hasNext() && transaction_supported ) {
      return true;
    }

    return false;
  }

  /**
   * This method return only those transactions (and relative items) in given
   * data sequence, that support the sequence.
   * If sequence is not support by given data sequence, this method return null.
   *
   * This method works exactly as isSupported(DataSequenceManager dseq), except
   * that instead of only checking which transactions support sequence elements,
   * it return those transactions.
   *
   * @param dseq data sequence to be checked
   * @return the set of transactions that support this sequence, or null if
   *   sequence is not supported
   */
  public TimestampedTransactionManager[] coveringTransactions(
      DataSequenceManager dseq) {
    if (this.getNumberOfElements() > dseq.getNumberOfTransactions())
      return null;

    TimestampedTransactionManager[] cover_set =
        new TimestampedTransactionManager[this.getNumberOfElements()];

    int index = 0;

    Iterator elements_iterator = this.getElements();
    Iterator transactions_iterator = dseq.getTransactions();
    boolean transactions_ended = false;
    boolean transaction_supported = false;

    while (elements_iterator.hasNext() && !transactions_ended) {

      SequenceElementManager element = (SequenceElementManager)
          elements_iterator.next();

      // skip delimiter
      if (elements_iterator.hasNext()) { elements_iterator.next(); }

      transaction_supported = false;

      while (transactions_iterator.hasNext() && !transaction_supported) {
        TimestampedTransactionManager transaction =
            (TimestampedTransactionManager) transactions_iterator.next();

        ItemManager[] items = element.coveringItems(transaction);

        if (items != null) {
          Item[] items_t = new Item[items.length];
          for (int cc = 0; cc < items.length; cc++)
            { items_t[cc] = (Item) items[cc]; }
          cover_set[index++] = new TimestampedTransaction(items_t,
              transaction.getTimestamp());
          transaction_supported = true;
        }

      }

      if (!transactions_iterator.hasNext()) transactions_ended = true;

    }

    boolean supported = false;

    if (!transactions_ended) supported = true;

    if (!elements_iterator.hasNext() && transaction_supported)
      supported = true;

    if( supported ) {
      return cover_set;
    }

    return null;
  }

  /**
   * Check if given sequence is contained in this one, i.e. if given sequence
   * is a subset of this one.
   * This method, also return true, if given sequence is equals to this one.
   *
   * @param sm sequence to be checked
   * @return true if sequence is contained, false otherwise
   */
  public boolean contains(SequenceManager sm){
    if( this.equals(sm) ) return true;

    Iterator sequence_iterator = sm.getElements();
    Iterator this_iterator = this.getElements();

    boolean isContained = false;
    // get all elements in given sequence
    while( sequence_iterator.hasNext() ){
      SequenceElementManager seq_element = (SequenceElementManager)
          sequence_iterator.next();
      isContained = false;

      // check elements in this sequence, until one contain current element
      while( !isContained && this_iterator.hasNext() ){
        SequenceElementManager this_element = (SequenceElementManager)
            this_iterator.next();

        if( this_element.contain(seq_element) ) {
          isContained = true;
        }

        // skip delimiter
        if( this_iterator.hasNext() ) { this_iterator.next(); }
      }

      // if threre are no more sequence elements, but given sequence
      // has more ones, given sequence is not contained
      if( !this_iterator.hasNext() && sequence_iterator.hasNext() ) {
        return false;
      }

      // skip delimiter
      if( sequence_iterator.hasNext() ) { sequence_iterator.next(); }
    }

    // if threre are no more sequence elements in both sequences, given
    // sequence is contained only if last it's element is too
    if( !sequence_iterator.hasNext() && !this_iterator.hasNext() ) {
      return isContained;
    }

    return true;
  }

  /**
   * Check if given sequence element belong to this sequence.
   * This method return only if sequence element is equal to one
   * in this sequence.
   *
   * @param element sequence element to be checked
   * @return true if sequence element belong to this sequene, false otherwise
   */
  public boolean contains(SequenceElementManager element) {
    Iterator iterator = this.getElements();
    while( iterator.hasNext() ){
      SequenceElementManager s = (SequenceElementManager) iterator.next();
      if( s.contain(element)) {
        return true;
      }
      //skip delimiter;
      if( iterator.hasNext() ) { iterator.next(); }
    }
    return false;
  }

  /**
   * Test if two sequence are equals.
   * To be equals, sequences must have the same elements in the same order.
   *
   * @param sequence sequence to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence){
    SequenceManager sm = (SequenceManager)sequence;

    if( this.getNumberOfElements() != sm.getNumberOfElements() )
      return false;

    Iterator iter = sm.getElements();
    Iterator iter2 = this.getElements();

    while( iter.hasNext() && iter2.hasNext() ){
      SequenceElementManager e = (SequenceElementManager) iter.next();
      SequenceElementManager e1 = (SequenceElementManager) iter2.next();

       if( !e.equals(e1) ) return false;

      if( iter.hasNext() && iter2.hasNext() ){
        iter.next();//skip delimiter
        iter2.next();//skip delimiter
      }

    }

    return true;
  }

  /**
   * Return the numbers of element in this sequence.
   *
   * @return number of element
   */
  public int getNumberOfElements() {
    //skips delimiters
    return ((this.elements.size() - 1) / 2)+1;
  }

  /**
   * Return the numbers of items in all element of this sequence.
   *
   * @return total umber of items
   */
  public int getNumberOfItems() {
    int num_items = 0;

    Iterator i = this.getElements();
    while( i.hasNext() ){
      SequenceElementManager e = (SequenceElementManager) i.next();
      num_items += e.getNumberOfItems();
      if( i.hasNext() ) i.next();//skips delimiters
    }

    return num_items;
  }

  /**
   * Return the numbers of objects in the data for which this sequence holds
   * true.
   *
   * @return number of occurrence
   */
  public Integer getOccurrence() {
    return this.occurrence;
  }

  /**
   * Return an iterator over all object in sequence.
   * This include both elements and delimiters.
   *
   * @return iterator over all objects in sequence
   */
  public Iterator getElements() {
    return this.elements.iterator();
  }

  /**
   * Return an iterator over all elements in sequence.
   * Delimiters object excluded.
   *
   * @return iterator over all elements in sequence
   */
  public Iterator getSequenceElements() {
    // instantiate a new iterator
    Iterator i = new Iterator(){
      // new iterator is backed by
      // iterator returned in getElements()
      Iterator it = getElements();

      // does nothing special
      public boolean hasNext(){ return it.hasNext(); }

      // redefined next() method to filter
      // delimiter object
      public Object next(){
        Object o = it.next();
        // skip delimiter
        if( it.hasNext() ) { it.next(); }
        return o;
      }

      // does nothing special
      public void remove(){ it.remove(); }
    };

    return i;
  }

  /**
    * Return the ratio of the number of objects in the data for which this
    * sequence holds true, to the total number of objects in the data.
    *
    * @return sequence support
    */
  public Double getSupport() {
    return this.support;
  }

  /**
   * Update number of occurrence
   *
   * @param occurrence new occurence
   */
  public void setOccurrence(Integer occurrence) {
    this.occurrence = occurrence;
  }

  /**
   * Update sequence support.
   * Support must between 0 and 1 inclusive.
   *
   * @param support new support
   */
  public void setSupport(Double support) {
    this.support = support;
  }

  /**
   * Return a string representation of this object.
   *
   * @return this object as string
   */
  public String toString() {
    if (elements.size() == 0)
      return "Empty Sequence!!!";
    else {
      String s = "";
      Iterator iter = elements.iterator();
      while (iter.hasNext()){
        SequenceElementManager set = (SequenceElementManager)iter.next();
        s += "{" + set.toString() + "}";
        //delimiter
        if( iter.hasNext()){
          s += iter.next().toString();
        }
      }
      if (this.support != null)
        s = s + "(supp="+support+")";
      else
        s = s + "(supp=n/a)";
      if (this.occurrence != null)
        s = s + "(occ="+occurrence+")";
      return s;
    }
  }

  /**
   * Return a subsequence starting at index "from" inclusive and ending at
   * index "to" excluded.
   * Returned sequence doesn't have neither support nor occurence.
   * Index range is [0 - number of elements].
   * Calling sequence.subSequence(0, sequence.getNumberOfElements()) return
   * a copy of sequence.
   *
   * @param from start index (inclusive)
   * @param to end index (excluded)
   * @return subsequence
   */
  public SequenceManager subSequence(int from, int to) {
    List sub_sequence = this.elements.subList( from*2, (to*2)-1 );
    return new Sequence(sub_sequence);
  }

  /**
   * Return delimiter object that precedes element with specified index.
   * This method could be useful in conjunction with suSequence() method.
   * Index range is [0, number of elements].
   *
   * @param index element index
   * @return delimiter
   */
  public DelimiterManager getPrecedingDelimiter(int index){
    DelimiterManager d = (DelimiterManager) this.elements.get( (index*2)-1 );
    DelimiterManager new_delimiter = new Delimiter(d);
    return new_delimiter;
  }

  /**
   * Append a sequence to the end of this one.
   * Delimiter is used between old and new sequence elements.
   *
   * @param delimiter a delimiter between old and new eleements
   * @param sequence sequence to be appended
   */
  public void append(DelimiterManager delimiter, SequenceManager sequence){
    Iterator iterator = sequence.getElements();
    if (iterator.hasNext())
      this.addElement(delimiter, (SequenceElementManager) iterator.next());
    while (iterator.hasNext())
      this.addElement(
          (DelimiterManager) iterator.next(),
          (SequenceElementManager) iterator.next());
  }

  /**
   * Return the hash code of this sequence.
   *
   * @return hash code
   */
  public int hashCode(){
    int hcode = 0;
    Iterator iter = elements.iterator();
    while (iter.hasNext()){
      SequenceElementManager set = (SequenceElementManager)iter.next();
      hcode += ((SequenceElement)set).toString().hashCode();
      //delimiter
      if( iter.hasNext())
        iter.next();
    }

    return hcode;
  }

  /**
   * Return a representation of this sequence as an Instance.
   * The instance have 6 fields representing:
   * <ol>
   * <li> the sequence as string -- obtained by toInternalFormat() method
   * (attribute sequence_elements); </li>
   * <li> the support of sequence (attribute sequence_support). </li>
   * <li> the number of transactions (attribute transactions_cardinality); </li>
   * <li> the number of distinct items (attribute distinct_items_cardinality).
   * </li>
   * <li> the maximal number of items per transactions
   * (attribute maximal_items_per_transactions). </li>
   *
   * This method does not check if the instance is compatible with the dataset.
   * Valid dataset must have at least following attribute:
   *
   * Attribute attr1 = new Attribute("sequence_elements", null);
   * Attribute attr2 = new Attribute("sequence_support");
   * Attribute attr3 = new Attribute("transactions_cardinality");
   * Attribute attr4 = new Attribute("distinct_items_cardinality");
   * Attribute attr5 = new Attribute("maximal_items_per_transactions");
   *
   * FastVector attributes = new FastVector();
   * attributes.addElement(attr1);
   * attributes.addElement(attr2);
   * attributes.addElement(attr3);
   * attributes.addElement(attr4);
   * attributes.addElement(attr5);
   *
   * Instances ris = new Instances("instances_name", attributes, 0);
   *
   * Attributes must exactly have the order showed, other attributes may be
   * added at the end.
   *
   * @param dataset dataset
   * @return this sequence as Instance object
   */
  public weka.core.Instance toInstance(weka.core.Instances dataset) {
    weka.core.Instance instance = new weka.core.Instance(5);

    String s = new String("");
    List distinct_items = new ArrayList();
    int max_items_per_ta = 0;

    Iterator iter = elements.iterator();
    while (iter.hasNext()){
      SequenceElementManager set = (SequenceElementManager)iter.next();

      if( set.getNumberOfItems() > max_items_per_ta )
        max_items_per_ta = set.getNumberOfItems();

      s += "<";

      Iterator set_iterator = set.getItems();
      while(set_iterator.hasNext()){
        ItemManager item = (ItemManager)set_iterator.next();
        if( !distinct_items.contains(item) ) distinct_items.add(item);
        s +=  encodeItem(item.toString()) + ",";
      }

      s = s.substring(0, s.length()-1);
      s += ">";

      //delimiter
      if( iter.hasNext()){
        DelimiterManager d = (DelimiterManager) iter.next();
        s += d.toString();
      }
    }

    instance.setDataset(dataset);

    instance.setValue(0, s);
    if( this.getSupport() != null )
      instance.setValue(1, this.getSupport().doubleValue());
    else
      instance.setMissing(1);
    instance.setValue(2, this.getNumberOfElements());
    instance.setValue(3, distinct_items.size());
    instance.setValue(4, max_items_per_ta);

    return instance;
  }

  /**
   * Transform an instance into a sequence object. Given instance must have
   * 2 fields, representing:
   * <ol>
   * <li> the sequence as string -- conform toInternalFormat() method --
   * (attribute sequence_elements: required); </li>
   * <li> the support of sequence (attribute sequence_support: may be missing).
   * </li> Other
   * field are allowed, but ignored. See toInstance() method for other details.
   *
   * @return sequence object
   * @param inst sequence as instance
   */
  public static SequenceManager instance2Sequence(weka.core.Instance inst){
    String seq_str = inst.toString(0);

    SequenceManager seq = fromInternalFormat(seq_str);

    if( !inst.isMissing(1) )
      seq.setSupport( new Double(inst.value(1) ) );

    return seq;
  }

  /**
   * Return an internal string representation of this sequence.
   * Internal format has the form:
   * <encoded_item,...,encoded_item>(delimiter_string)...<encoded_item,...,encoded_item>.
   * Item value is encoded to avoid problems related string manipulation.
   * Item encoding is related to characters ('<','>',',','&') used for internal
   * format representation.
   * To reverse this process, use decode() method.
   *
   * @return internal string format
   */
  public String toInternalFormat() {
    String s = new String("");

    Iterator iter = elements.iterator();
    while (iter.hasNext()){
      SequenceElementManager set = (SequenceElementManager)iter.next();

      s += "<";

      Iterator set_iterator = set.getItems();
      while(set_iterator.hasNext()){
        ItemManager item = (ItemManager)set_iterator.next();
        s +=  encodeItem(item.toString()) + ",";
      }

      s = s.substring(0, s.length()-1);
      s += ">";

      //delimiter
      if( iter.hasNext()){
        DelimiterManager d = (DelimiterManager) iter.next();
        s += "(";

        if( d.getGAP().equals(GapType.FALSE) )
          s +=  "-";
        else if( d.getGAP().equals(GapType.TRUE) )
          s +=  "+";
        else
          s +=  "*";

        if( d.getDELIMITER().equals(DelimiterType.ACROSS_TIME_WINDOWS) )
          s += "A";
        else
          s += "S";

        s += ")";

      }
    }

    return s;
  }

  /**
   * Transform an internal string representation of a sequence into sequence
   * object.
   * For more details about internal representation see toInternalFormat()
   * method.
   *
   * @param internal string representation of sequence
   * @return sequence object
   */
  public static SequenceManager fromInternalFormat(String internal){
    //remove '' from string if any
    if( internal.charAt(0) == '\'' )
      internal = internal.substring(1);
    if( internal.charAt(internal.length()-1) == '\'' )
      internal = internal.substring(0, internal.length()-1);

    int start_index, end_index;

    start_index = internal.indexOf('<');
    end_index = internal.indexOf('>', start_index);

    String set = internal.substring(start_index, end_index+1);
    String del;

    SequenceManager seq = new Sequence( toSet(set) );

    while( internal.length() > end_index + 1 ){
      // end_index + 5: delimiter string has length 5...
      del = internal.substring(end_index + 1, end_index + 5);

      start_index = internal.indexOf('<', end_index);
      end_index = internal.indexOf('>', start_index);

      set = internal.substring(start_index, end_index+1);

      seq.addElement( toDelimiter(del), toSet(set) );
    }

    return seq;
  }

  // This method is used by fromInternalFormat(String sequence)
  // to parse a delimiter string
  private static DelimiterManager toDelimiter(String del){
    // remove ( )
    String delimiter = del.substring(1, del.length()-1);

    DelimiterType dtype;
    GapType gtype;

    char window = delimiter.charAt(1);
    if( window == 'A')
      dtype = DelimiterType.ACROSS_TIME_WINDOWS;
    else
      dtype = DelimiterType.SAME_TIME_WINDOW;

    char contiguos = delimiter.charAt(0);
    if( contiguos == '*' ) gtype = GapType.UNKNOWN;
    else if( contiguos == '+' ) gtype = GapType.TRUE;
    else gtype = GapType.FALSE;

    return new Delimiter(dtype, gtype);
  }

  // This method is used by fromInternalFormat(String sequence)
  // to parse a set string
  private static SequenceElementManager toSet(String list_of_items){
    ElementItemSet set = new ElementItemSet();

    // remove < >
    list_of_items = list_of_items.substring(1, list_of_items.length()-1);
    StringTokenizer tokenizer = new StringTokenizer(list_of_items, ",");

    while( tokenizer.hasMoreTokens() ){
      String item = tokenizer.nextToken();
      item = decodeItem(item);
      set.addItem( new Item(item) );
    }

    return set;
  }

  // This method is used by toInternalFormat() to encode an item
  private static String encodeItem(String item){
    String eItem = new String(item);
    eItem = item.replaceAll("&", "&amp");
    eItem = item.replaceAll("<", "&lt");
    eItem = item.replaceAll(">", "&gt");
    eItem = item.replaceAll(",", "&comma");
    return eItem;
  }

  // This method is used by fromInternalFormat(String sequence) to decode
  // an encoded item
  private static String decodeItem(String item){
    String eItem = new String(item);
    eItem = item.replaceAll("&lt", "<");
    eItem = item.replaceAll("&gt", ">");
    eItem = item.replaceAll("&comma", ",");
    eItem = item.replaceAll("&amp", "&");
    return eItem;
  }

public static void main(String[] args) {
  ElementItemSet i = new ElementItemSet();
  i.addItem(new Item("item"));
  ElementItemSet i2 = new ElementItemSet();
  i2.addItem(new Item("item2"));
  ElementItemSet i3 = new ElementItemSet();
  i3.addItem(new Item("item3"));
  ElementItemSet i4 = new ElementItemSet();
  i4.addItem(new Item("item4"));

  Sequence s = new Sequence(i);
  s.addElement(Delimiter.DEFAULT,i2);
  s.addElement(new Delimiter(DelimiterType.SAME_TIME_WINDOW, GapType.TRUE),i3);
  s.addElement(Delimiter.DEFAULT,i4);

  System.out.println(s.toString());

  Sequence sub =  (Sequence)s.subSequence(0,2);
  System.out.println(sub.getNumberOfElements());
  System.out.println(sub.elements.size());
  System.out.println(sub.toString());

  Delimiter d = (Delimiter)s.getPrecedingDelimiter(2);

  Sequence sub2 =  (Sequence)s.subSequence(2,s.getNumberOfElements());
  System.out.println(sub2.getNumberOfElements());
  System.out.println(sub2.elements.size());
  System.out.println(sub2.toString());

  Sequence copy = new Sequence(sub);
  copy.append(d, sub2);

  System.out.println(copy.toString());

  System.out.println(s.equals(copy));

  System.out.println(s.equals(s.subSequence(0,s.getNumberOfElements())));

  Iterator it = s.getElements();
  while( it.hasNext() ){
    it.next();
    it.remove();
  }
}
}
