/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.TreeSet;
import java.util.Iterator;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemManager;
import kddml.Core.DataMining.AssociationRules.CompareItems;
import kddml.Core.DataMining.AssociationRules.TransactionManager;

/**
 * This abstract class represents a generic sequence element (i.e. a generic set of items).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public abstract class SequenceElement implements SequenceElementManager {

  /**
   * The set of items is ordered using an item comparator.
   * See kddml.Core.DataMining.AssociationRules.CompareItems for more details.
   */
  protected TreeSet items;

  /**
   * Build an empty sequence element.
   */
  protected SequenceElement() {
    this.items = new TreeSet( new CompareItems() );
  }

  /**
   * Build a new sequence element with given items.
   * No checking on structure are done.
   *
   * @param items a TreeSet containing Item object
   */
  protected SequenceElement(TreeSet items) {
    this();
    this.items.addAll(items);
  }

  /**
   * Add a new item to this element.
   * If specified item is already present will not be added.
   *
   * @param item new item to be added
   * @return true if item is added, false otherwise
   */
  public boolean addItem(ItemManager item) {
    return this.items.add(item);
  }

  /**
   * Check if gievn item is already present in this element.
   *
   * @param item item to be checked
   * @return true if item is contained, false otherwise
   */
  public boolean contain(ItemManager item) {
    Iterator iter = this.getItems();
    while( iter.hasNext() ){
      ItemManager i = (ItemManager) iter.next();
      if( i.equals(item) ) return true;
    }
    return false;
  }

  /**
   * Check if given element is contained in this one, i.e. if given element is a
   * subset of this one.
   * This method, also return true, if given element is equals to this one.
   *
   * @param element element to be checked
   * @return true if element is contained, false otherwise
   */
  public boolean contain(SequenceElementManager element) {
    if( element.getNumberOfItems() > this.getNumberOfItems() )
      return false;

    Iterator iter = element.getItems();

    while( iter.hasNext() ){
      if( !this.contain((ItemManager)iter.next()) )
        return false;
    }

    return true;
  }

  /**
   * Check if given object is equals to this one.
   * This method perform an item-by-item checking, i.e. items in given
   * must be ordered as described in
   * kddml.Core.DataMining.AssociationRules.CompareItems class.
   *
   * @param sequence_element object to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence_element){
    SequenceElementManager sm = (SequenceElementManager) sequence_element;

    if( this.getNumberOfItems() != sm.getNumberOfItems() )
      return false;

    Iterator sm_iter = sm.getItems();
    Iterator this_iter = this.getItems();

    while( sm_iter.hasNext() && this_iter.hasNext()){
      ItemManager sm_item = (ItemManager)sm_iter.next();
      ItemManager this_item = (ItemManager)this_iter.next();
      if( !sm_item.equals(this_item) )
        return false;
    }

    return true;

  }

  /**
   * Check if this element is supported by given transaction, i.e. if items in
   * this element are contained in given transaction.
   *
   * @param transaction transaction to be evaluated
   * @return true if all items in this element are contained in given
   *   transaction, false otherwise
   */
  public boolean isSupported(TransactionManager transaction){
    if( transaction.getNumberOfItems() < this.getNumberOfItems() )
      return false;

    Iterator e_iterator = this.getItems();
    while( e_iterator.hasNext() ) {
      ItemManager e_item = (ItemManager) e_iterator.next();
      if( !transaction.contains((Item)e_item) )
        return false;
    }

    return true;
  }

  /**
   * If this element is supported by given transaction,
   * this method return all items in transaction, that are contained in this
   * element; if not, return null.
   *
   * The set of items returned is identical to one in this element; difference
   * is made from optional features that may be associated with items.
   *
   *
   * @param transaction transaction to be checked
   * @return covering items, or null if this element is not supported
   */
  public ItemManager[] coveringItems(TransactionManager transaction){
    if( transaction.getNumberOfItems() < this.getNumberOfItems() )
      return null;

    ItemManager[] cover_set = new ItemManager[ this.getNumberOfItems() ];
    int index = 0;

    Iterator el_iterator = this.getItems();
    while( el_iterator.hasNext() ) {
      ItemManager el_item = (ItemManager) el_iterator.next();
      ItemManager tr_item = contain(transaction, el_item);
      if( tr_item == null )
        return null;
      cover_set[index++] = tr_item;
    }

    return cover_set;
  }

  /**
   * If given transaction contain specified item, return the equivalent item
   * in transaction; if not return null.
   *
   * @param transaction TransactionManager
   * @param item ItemManager
   * @return ItemManager
   */
  private ItemManager contain(TransactionManager transaction, ItemManager item){
    Iterator iterator = transaction.getItems();
    while( iterator.hasNext() ) {
      ItemManager i = (ItemManager) iterator.next();
      if( i.equals(item) )
        return i;
    }
    return null;
  }

  /**
   * Return an iterator over items in this element.
   * Remember that items are ordered using
   * kddml.Core.DataMining.AssociationRules.CompareItems class.
   *
   * @return an iterator over items in this element
   */
  public Iterator getItems() {
    return this.items.iterator();
  }

  /**
   * Return the number of items in this element.
   *
   * @return number of items
   */
  public int getNumberOfItems() {
    return this.items.size();
  }

  /**
   * Return a string representation of this object
   *
   * @return this object as string
   */
  public String toString() {
    if (items.size() == 0)
      return "Empty Sequence Element!!!";
    else {
      String s="(";
      Iterator iter = items.iterator();
      while (iter.hasNext())
        s = s + ((Item) iter.next()).toString() + ", ";
      s = s + ")";
      return s;
    }

  }

}
