/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.Iterator;
import kddml.Core.DataMining.AssociationRules.ItemManager;
import kddml.Core.DataMining.AssociationRules.TransactionManager;

/**
 * This abstract class represents a generic sequence element (i.e. a generic set of items).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public interface SequenceElementManager {

  /**
   * Add a new item to this element.
   * If specified item is already present will not be added.
   *
   * @param item new item to be added
   * @return true if item is added, false otherwise
   */
  public boolean addItem(ItemManager item);

  /**
   * Check if gievn item is already present in this element.
   *
   * @param item item to be checked
   * @return true if item is contained, false otherwise
   */
  public boolean contain(ItemManager item);

  /**
   * Check if given element is contained in this one, i.e. if given element is a
   * subset of this one.
   * This method, also return true, if given element is equals to this one.
   *
   * @param element element to be checked
   * @return true if element is contained, false otherwise
   */
  public boolean contain(SequenceElementManager element);

  /**
   * Check if given object is equals to this one.
   * This method perform an item-by-item checking, i.e. items in given
   * must be ordered as described in
   * kddml.Core.DataMining.AssociationRules.CompareItems class.
   *
   * @param sequence_element object to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence_element);
  /**
   * Check if this element is supported by given transaction, i.e. if items in
   * this element are contained in given transaction.
   *
   * @param transaction transaction to be evaluated
   * @return true if all items in this element are contained in given
   *   transaction, false otherwise
   */
  public boolean isSupported(TransactionManager transaction);

  /**
   * If this element is supported by given transaction,
   * this method return all items in transaction, that are contained in this
   * element; if not, return null.
   *
   * The set of items returned is identical to one in this element; difference
   * is made from optional features that may be associated with items.
   *
   *
   * @param transaction transaction to be checked
   * @return covering items, or null if this element is not supported
   */
  public ItemManager[] coveringItems(TransactionManager transaction);

  /**
   * Return an iterator over items in this element.
   * Remember that items are ordered using
   * kddml.Core.DataMining.AssociationRules.CompareItems class.
   *
   * @return an iterator over items in this element
   */
  public Iterator getItems();

  /**
   * Return the number of items in this element.
   *
   * @return number of items
   */
  public int getNumberOfItems();

  /**
   * Return a string representation of this object
   *
   * @return this object as string
   */
  public String toString();

}
