/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.*;
import kddml.Core.*;

/**
 * Abstract class that create instances of a SequenceModel from an external
 * source (e.g. pmml or JDM connection).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public abstract class SequenceExternalFactory extends ModelExternalFactory {

  /**
   * Constructor given the physical file containing the object.
   *
   * @param f File
   * @throws KDDMLCoreException
   */
  public SequenceExternalFactory(java.io.File f)
      throws KDDMLCoreException {
    super(f);
    object = new SequenceModel(f);
  }


  public KDDMLObject newInstance() throws KDDMLCoreException {
    super.newInstance();

    SequenceModel sequenceModel = (SequenceModel)object;

    Double min_support = buildMinSupport();
    UChecker.checkMinSupport(min_support);
    sequenceModel.minimumSupport = min_support;

    Double min_confidence = buildMinConfidence();
    UChecker.checkMinConfidence(min_confidence);
    sequenceModel.minimumConfidence = min_confidence;

    Integer number_of_transactions = buildNumberOfTransactions();
    UChecker.checkNumTransactions(number_of_transactions);
    sequenceModel.numberOfTransactions = number_of_transactions;

    Integer max_items = buildMaxNumberOfItemsPerTA();
    if (max_items != null) {
      UChecker.checkMaxItems(max_items);
      sequenceModel.maxNumberOfItemsPerTA = max_items;
    }

    Double avg_items = buildAvgNumberOfItemsPerTA();
    if (avg_items != null) {
      UChecker.checkAvgItems(avg_items);
      sequenceModel.avgNumberOfItemsPerTA = avg_items;
    }

    Integer length_limit = buildLengthLimit();
    if (length_limit != null) {
      UChecker.checkLengthLimit(length_limit);
      sequenceModel.lengthLimit = length_limit;
    }

    Integer time_window = buildTimeWindow();
    if (time_window != null) {
      UChecker.checkTimeWindow(time_window);
      sequenceModel.timeWindowWidth = time_window;
    }

    Integer min_time = buildMinTime();
    if (min_time != null) {
      UChecker.checkMinTime(min_time);
      sequenceModel.minimumTime = min_time;
    }

    Integer max_time = buildMaxTime();
    if (max_time != null) {
      UChecker.checkMaxTime(max_time);
      sequenceModel.maximumTime = max_time;
    }

    sequenceModel.items = buildItems();
    sequenceModel.elements = buildSets();
    sequenceModel.sequences = buildSequences();
    sequenceModel.rules = buildSequenceRules();

    return object;
  }

  /**
   * Returns the minimum support of the sequences.
   * Mininmum support is a real in (0,1]
   *
   * @throws SequenceModelException
   * @return Double
   */
  protected abstract Double buildMinSupport() throws SequenceModelException;

  /**
   * Returns the minimum confidence of the sequence rules.
   * Mininmum confidence is a real in (0,1].
   *
   * @throws SequenceModelException
   * @return Double
   */
  protected abstract Double buildMinConfidence() throws SequenceModelException;

  /**
   * Returns the number of transaction used to build the sequences.
   * Returns a positive integer.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildNumberOfTransactions() throws SequenceModelException;

  /**
   * Returns the maximum number of iterms per transaction.
   * Returns a positive integer.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildMaxNumberOfItemsPerTA() throws SequenceModelException;

  /**
   * Returns the average number of items per transaction.
   * Returns a positive double.
   *
   * @throws SequenceModelException
   * @return Double
   */
  protected abstract Double buildAvgNumberOfItemsPerTA() throws SequenceModelException;

  /**
   * Return the maximum length of a sequence to be discovered.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildLengthLimit() throws SequenceModelException;

  /**
   * Return the time window width:
   * this may be used to separate items associated with an object into discrete
   * events, but only if no clear key already exists for the separate events.
   * Two consecutive items must have a time gap of less than this value to be
   * considered as being part of the same event.
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildTimeWindow() throws SequenceModelException;

  /**
   * Return the minimum time between items (see buildTimeWindow())
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildMinTime() throws SequenceModelException;

  /**
   * Return the maximum time between items (see buildTimeWindow())
   *
   * @throws SequenceModelException
   * @return Integer
   */
  protected abstract Integer buildMaxTime() throws SequenceModelException;

  /**
   * Return the list of sequence rules.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected abstract java.util.TreeSet buildSequenceRules() throws SequenceModelException;

  /**
   * Return the list of sequences.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected abstract java.util.TreeSet buildSequences() throws SequenceModelException;

  /**
   * Return the list of sequence elements.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected abstract java.util.TreeSet buildSets() throws SequenceModelException;

  /**
   * Return the list of items.
   *
   * @throws SequenceModelException
   * @return TreeSet
   */
  protected abstract java.util.TreeSet buildItems() throws SequenceModelException;
}
