/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.*;

/**
 * Factory used to build a sequence model given a set of sequences and
 * a specified minimum support and confidence.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public final class SequenceFactory extends SequenceProprietaryFactory {

  private int max_sequences;

  /**
   * Constructor.
   *
   * @param f the file containing the model
   * @param dd the data dictionary
   * @param ms the mining schema
   * @param min_support the minimum support
   * @param min_confidence the minimum confidence
   * @param number_of_transactions the number of transactions
   * @param sequences the ordered set of sequences
   * @throws KDDMLCoreException
   */
  public SequenceFactory(java.io.File f,
                                 DataDictionary dd,
                                 MiningSchema ms,
                                 double min_support,
                                 double min_confidence,
                                 int number_of_transactions,
                                 TreeSet sequences)
      throws KDDMLCoreException {

    super(f, dd, ms, min_support, min_confidence, number_of_transactions);
    this.sequences = sequences;
    max_sequences = SequenceModel.MAX_NUMBER_OF_SEQUENCES;
  }

  /**
   * Build a new instance.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject newInstance() throws KDDMLCoreException {

    SequenceModel model = (SequenceModel) super.newInstance();

    SequenceModel.MAX_NUMBER_OF_SEQUENCES = max_sequences;

    this.elements = USequenceElementBuilder.compute(sequences);
    //if elements instances are Itemsets build items...
    this.items = UItemsBuilder.compute(elements);
    if(this.items == null ) this.items = new TreeSet(new kddml.Core.DataMining.AssociationRules.CompareItems());

    model.items = items;
    model.elements = elements;
    model.sequences = sequences;
    model.rules = new TreeSet(new CompareSequenceRules());

    return model;
  }

  /**
   * Set the maximum number of sequences allowed in the model.
   *
   * @param max_sequences maximum sequences
   */
  public void setMaxNumberOfSequences(int max_sequences) {
    this.max_sequences = max_sequences;
  }

  public static void main(String args[]) {
    /*try {
       String s =
           "C:\\Documents and Settings\\Amministratore\\Desktop\\sequence.xml";
       SequenceRepositoryFactory f = new SequenceRepositoryFactory(new java.io.File(s));

       SequenceModel m = (SequenceModel) f.newInstance();
       //System.out.println(m.toString());

       //m.clone(new java.io.File(s+"-result.xml")).saveToRepository();
       //m.saveHTML();

       String s2 = "C:/Documents and Settings/Amministratore/Desktop/rules.xml";
       DataDictionary d = new DataDictionary();
       MiningSchema m2 = new MiningSchema();
       RulesSequenceFactory f2 = new RulesSequenceFactory(
           new java.io.File(s2), d, m2, 0.5, 0.5, 10, m.rules);
       f2.setAlgorithmName("Internal");
       SequenceModel model = (SequenceModel) f2.newInstance();
       //System.out.println(model.toString());
       model.saveToRepository();
       model.saveHTML();

     }
     catch(KDDMLCoreException e) {
       System.out.println("########################");
       e.printStackTrace();
       System.out.println("########################");
       System.out.println(e.getDetailErrorMessage());
     }*/

  }

}
