/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.Iterator;

/**
 * A sequence consists of:
 * <ol>
 * <li> Optional attributes (support and occurrence); </li>
 * <li> A list of sets composing the sequences (SequenceElementManager); </li>
 * </ol>
 * A sequence is structured as a set of consecutive elements
 * (SequenceElementManager).
 * Each of this elements is separated by the one that follows with a delimiter
 * (DelimiterManager).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public interface SequenceManager {

  /**
   * Add an element to this sequence.
   *
   * @param d delimiter between last element and the new one
   * @param el new element to be added
   * @return true if element is added, false otherwise
   */
  public boolean addElement(DelimiterManager d, SequenceElementManager el);

  /**
   * Return the numbers of element in this sequence.
   *
   * @return number of element
   */
  public int getNumberOfElements();

  /**
   * Return the numbers of items in all element of this sequence.
   *
   * @return total umber of items
   */
  public int getNumberOfItems();

  /**
   * Return the numbers of objects in the data for which this sequence holds
   * true.
   *
   * @return number of occurrence
   */
  public Integer getOccurrence();

  /**
   * Return the ratio of the number of objects in the data for which this
   * sequence holds true, to the total number of objects in the data.
   *
   * @return sequence support
   */
  public Double getSupport();

  /**
   * Return an iterator over all object in sequence.
   * This include both elements and delimiters.
   *
   * @return iterator over all objects in sequence
   */
  public Iterator getElements();

  /**
   * Return an iterator over all elements in sequence.
   * Delimiters object excluded.
   *
   * @return iterator over all elements in sequence
   */
  public Iterator getSequenceElements();

  /**
   * Test if two sequence are equals.
   * To be equals, sequences must have the same elements in the same order.
   *
   * @param sequence sequence to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence);

  /**
   * Check if given sequence is contained in this one, i.e. if given sequence
   * is a subset of this one.
   * This method, also return true, if given sequence is equals to this one.
   *
   * @param sm sequence to be checked
   * @return true if sequence is contained, false otherwise
   */
  public boolean contains(SequenceManager sm);

  /**
   * Check if given sequence element belong to this sequence.
   * This method return only if sequence element is equal to one
   * in this sequence.
   *
   * @param element sequence element to be checked
   * @return true if sequence element belong to this sequene, false otherwise
   */
  public boolean contains(SequenceElementManager element);

  /**
   * Check if given data sequence support this sequence.
   *
   * For each element e in this sequence, there is a transaction t such that
   * e.isSupported(t) holds true, and is respected element order.
   * E.g. s = (e1)...(em); d = (t1)...(tn)
   * 1) m <= n
   * 2) e(1).isSupported(ti) => e(2).isSupported(tj) , j > i
   *
   * @param dseq data sequence to be checked
   * @return true if given data sequence support this sequence, false otherwise
   */
  public boolean isSupported(DataSequenceManager dseq);

  /**
   * This method return only those transactions (and relative items) in given
   * data sequence, that support the sequence.
   * If sequence is not support by given data sequence, this method return null.
   *
   * This method works exactly as isSupported(DataSequenceManager dseq), except
   * that instead of only checking which transactions support sequence elements,
   * it return those transactions.
   *
   * @param dseq data sequence to be checked
   * @return the set of transactions that support this sequence, or null if
   *   sequence is not supported
   */
  public TimestampedTransactionManager[] coveringTransactions(
      DataSequenceManager dseq);

  /**
   * Return a subsequence starting at index "from" inclusive and ending at
   * index "to" excluded.
   * Returned sequence doesn't have neither support nor occurence.
   * Index range is [0 - number of elements].
   * Calling sequence.subSequence(0, sequence.getNumberOfElements()) return
   * a copy of sequence.
   *
   * @param from start index (inclusive)
   * @param to end index (excluded)
   * @return subsequence
   */
  public SequenceManager subSequence(int from, int to);

  /**
   * Append a sequence to the end of this one.
   * Delimiter is used between old and new sequence elements.
   *
   * @param delimiter a delimiter between old and new eleements
   * @param sequence sequence to be appended
   */
  public void append(DelimiterManager delimiter, SequenceManager sequence);

  /**
   * Return delimiter object that precedes element with specified index.
   * This method could be useful in conjunction with suSequence() method.
   * Index range is [0, number of elements].
   *
   * @param index element index
   * @return delimiter
   */
  public DelimiterManager getPrecedingDelimiter(int index);

  /**
   * Return a string representation of this object.
   *
   * @return this object as string
   */
  public String toString();

  /**
   * Return the hash code of this sequence.
   *
   * @return hash code
   */
  public int hashCode();

  /**
   * Update sequence support
   * Support must between 0 and 1 inclusive.
   *
   * @param support new support
   */
  public void setSupport(Double support);

  /**
   * Update number of occurrence
   *
   * @param occurrence new occurence
   */
  public void setOccurrence(Integer occurrence);

}
