/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;
import kddml.Core.*;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemManager;
import java.util.TreeSet;
import java.util.Iterator;
import java.util.Vector;
import weka.core.*;

/**
 * A sequence model consists of four major parts:
 * <ol>
 * <li> Model attributes (such as minimum support, minimum confidence, etc.); </li>
 * <li> A list of items; </li>
 * <li> A list of sets composing the sequences (either Itemsets or SetPredicates); </li>
 * <li> A list of frequent sequences; </li>
 * <li> A list of sequence rules. </li>
 * </ol>
 * Items, sets, sequences and sequence rules are ordered using a precise function and this
 * objects are mantained in a TreeSet structure.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SequenceModel extends kddml.Core.DataMining.MiningModel
    implements SequenceModelManager {

  /**
   * Max number of sequences that the model can contains.
   */
  public static int MAX_NUMBER_OF_SEQUENCES = 5000;

  /**
   * Max number of sequence rules that the model can contains.
   */
  public static int MAX_NUMBER_OF_RULES = 5000;

  /**
   * The set of ordered sequence rules.
   */
  protected TreeSet rules;

  /**
   * The set of ordered sequences.
   */
  protected TreeSet sequences;

  /**
   * The set of ordered sequence elements.
   */
  protected TreeSet elements;

  /**
   * The set of ordered items.
   */
  protected TreeSet items;

  /**
   * The number of objects contained in the input data (e.g. the number of unique
   * customers).
   */
  protected Integer numberOfTransactions;

  /**
   * Minimum support.
   * Support is calculate as: (#supporting transactions / #total transactions).
   */
  protected Double minimumSupport;

  /**
   * Minimum confidence.
   * Confidence is calculate as: (support(rule) / support(antecedent)).
   */
  protected Double minimumConfidence;

  /**
   * The maximum number of events (e.g. items) per object.
   */
  protected Integer maxNumberOfItemsPerTA;

  /**
   * The average number of events that make up the object.
   */
  protected Double avgNumberOfItemsPerTA;

  /**
   * The maximum length of a sequence to be discovered.
   */
  protected Integer lengthLimit;

  /**
   * This may be used to separate items associated with an object into discrete
   * events, but only if no clear key already exists for the separate events.
   * Two consecutive items must have a time gap of less than this value to be
   * considered as being part of the same event.
   */
  protected Integer timeWindowWidth;

  /**
   * Minimum time between items (see timeWindowWidth).
   */
  protected Integer minimumTime;

  /**
   * Maximum time between items (see timeWindowWidth).
   */
  protected Integer maximumTime;

  /**
   * Package proprietary constructor.
   *
   * @param physical_path File
   */
  protected SequenceModel(java.io.File physical_path) {
    super(physical_path);
  }

  public kddml.Core.KDDMLObjectType getType() {
    return kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL;
  }

  public boolean isEmpty() {
    return (this.sequences.size() == 0 && this.rules.size() == 0);
  }

  /**
   * Save the object in the system repository.
   *
   * @throws SequenceModelException
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws SequenceModelException, kddml.Core.KDDMLCoreException {

    if( this.isEmpty() )
      throw new SequenceModelException(
        "Illegal PMML Sequence Model: document contains no data.",
        null,
        KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    Document root = initializeDocument();

   // attribute settings...
   Element sequence_model = UPMMLBuilder.buildModel(this);

   if (mining_schema != null)
      sequence_model = appendMiningSchema((MiningSchema) mining_schema, sequence_model);

   // costruisce gli elementi PMML che rappresentano gli items...
   Vector itemsRef = appendPMMLItems(sequence_model);
   // costruisce gli elementi PMML che rappresentano gli itemsets...
   Vector elementsRef = appendPMMLSets(sequence_model, itemsRef);

   itemsRef = null;
   System.gc();

   // costruisce gli elementi PMML che rappresentano le sequenze...
   Vector sequenceRef = appendPMMLSequences(sequence_model, elementsRef);

   elementsRef = null;
   System.gc();

   // costruisce gli elementi PMML che rappresentano le regole sequenziali...
   appendPMMLRules(sequence_model, sequenceRef);

   sequenceRef = null;
   System.gc();

   root.getDocumentElement().appendChild(sequence_model);

   XMLUtility.toFile(object_path.getAbsolutePath(), root);

  }

  /**
   * Return a string representation of this object.
   *
   * @return this object as string
   */
  public String toString() {
    String s = super.toString();
    s = s + "---------------------------------\n";
    s = s + "Sequence Model:\n";
    s = s + "---------------------------------\n";
    s = s + "Number of Items: "+items.size()+"\n";
    s = s + "Number of Elements: "+elements.size()+"\n";
    s = s + "Number of Sequences: "+sequences.size()+"\n";
    s = s + "Number of Rules: "+rules.size()+"\n";
    Iterator r = items.iterator();
    s = s + "---------\n";
    s = s + "List of items: ";
    while(r.hasNext()) {
      ItemManager i = (ItemManager) r.next();
      s = s + i.toString()+", ";
    }
    s = s + "\n";
    s = s + "---------\n";
    r = elements.iterator();
    s = s + "---------\n";
    s = s + "List of elements: ";
    while(r.hasNext()) {
      SequenceElementManager i = (SequenceElementManager) r.next();
      s = s + i.toString()+", ";
    }
    s = s + "\n";
    s = s + "---------\n";
    s = s + "List of frequent sequences: \n";
    r = sequences.iterator();
    while(r.hasNext()) {
      SequenceManager i = (SequenceManager) r.next();
      s = s + i.toString()+"\n";
    }
    s = s + "---------\n";
    s = s + "List of sequence rules: \n";
    r = rules.iterator();
    while(r.hasNext()) {
      SequenceRuleManager i = (SequenceRuleManager) r.next();
      s = s + i.toString()+"\n";
    }
    return s;
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML()
      throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.PMML_SEQUENCE_MODEL);

    String s1 = getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    try {
      XMLUtility.formatting(s1, s2, s3, null, null);
    }
    catch(Exception e) {
      String msg = "Exception found during the HTML code generation.";
      msg = e.toString();
      throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
    }

  }

  /**
   * Clone this model returning a new sequence model.
   *
   * @param output_path File the path in the model system repository containing the cloned
   * model
   * @return cloned model
   */
  public SequenceModel clone(java.io.File output_path) {

    SequenceModel output = new SequenceModel(output_path);
    output.data_dictionary = (DataDictionary) this.data_dictionary.clone();
    output.mining_schema = (MiningSchema) this.mining_schema.clone();
    // set other proprieties here...

    output.model_name = (this.getModelName() == null) ? (null) : (new String(this.getModelName()));
    output.algorithm_name = (this.getAlgorithmName() == null) ? (null) : (new String(this.getAlgorithmName()));
    output.maxNumberOfItemsPerTA = this.getMaxNumberOfItemsPerTransaction();
    output.avgNumberOfItemsPerTA = this.getAvgNumberOfItemsPerTransaction();

    output.lengthLimit = this.getLengthLimit();
    output.timeWindowWidth = this.getTimeWindowWidth();
    output.minimumTime = this.getMinimumTime();
    output.maximumTime = this.getMaximumTime();

    output.minimumSupport = this.getMinimumSupport();
    output.minimumConfidence = this.getMinimumConfidence();
    output.numberOfTransactions = this.getNumberOfTransactions();

    output.items = new TreeSet(this.items);
    output.elements = new TreeSet(this.elements);
    output.sequences = new TreeSet(this.sequences);
    output.rules = new TreeSet(this.rules);

    return output;
  }

  /**
   * Remove unused elements and items.
   */
  public void cleanData(){
    this.elements = USequenceElementBuilder.compute(this.sequences);
    this.items = UItemsBuilder.compute(this.elements);
  }

  /**
   * Add a rule to this model.
   * If rule is already contained in this model, or confidence of rule is less
   * than minimum confidence, or support of rule is less than minimum support,
   * rule is not added.
   *
   * @param rule rule to be added
   * @return true if rule is added
   */
  public boolean addRule(SequenceRuleManager rule) {
    if (this.contains(rule))
      return false;
    if (rule.getSupport().doubleValue() < this.getMinimumSupport().doubleValue())
      return false;
    if (rule.getConfidence().doubleValue() < this.getMinimumConfidence().doubleValue())
      return false;
    this.rules.add(rule);

    //Antecedent sequence
    SequenceManager ant = rule.getAntecedent();
    //Consequent sequence
    SequenceManager cons = rule.getConsequent();
    DelimiterManager d = rule.getDelimiter();
    //Union sequence
    SequenceManager union = new Sequence(ant);
    union.append( d, cons);
    union.setSupport( rule.getSupport() );

    //possibili nuove sequenze da aggiungere...
    TreeSet new_sequences = new TreeSet( new CompareSequences() );
    new_sequences.add(ant);
    new_sequences.add(cons);
    new_sequences.add(union);

    // scorre le sequenze per calcolare l'insieme degli elementi che le compongono...
    TreeSet new_elements = USequenceElementBuilder.compute(new_sequences);
    // scorre gli elementi delle sequenze per calcolare gli items...
    TreeSet new_items = UItemsBuilder.compute(new_elements);

    this.sequences.addAll( new_sequences );
    this.elements.addAll( new_elements );
    if( new_items != null ) this.items.addAll( new_items );

    return true;
  }

  /**
   * Add a set of rules to this model.
   * If a rule is already contained in this model, or confidence of  a rule is
   * less than minimum confidence, or support of a rule is less than minimum
   * support, rule is not added.
   *
   * @param rules rules to be added
   */
  public void addRules(SequenceRuleManager[] rules) {
    TreeSet rules_added = new TreeSet( new CompareSequenceRules() );

    for (int i=0; i<rules.length; i++) {
      SequenceRuleManager rule = (SequenceRuleManager) rules[i];
      if (!this.contains(rule))
        if (rule.getSupport().doubleValue() >= this.getMinimumSupport().doubleValue() &&
            rule.getConfidence().doubleValue() >= this.getMinimumConfidence().doubleValue())
          if( this.rules.add(rules[i]) ) rules_added.add(rules[i]);
    }

    // scorre le regole aggiunte per calcolare le sequenze...
    TreeSet new_sequences = USequenceBuilder.compute(rules_added, MAX_NUMBER_OF_SEQUENCES);
    // scorre le sequenze per calcolare l'insieme degli elementi che le compongono...
    TreeSet new_elements = USequenceElementBuilder.compute(new_sequences);
    // scorre gli elementi delle sequenze per calcolare gli items...
    TreeSet new_items = UItemsBuilder.compute(new_elements);

    this.sequences.addAll( new_sequences );
    this.elements.addAll( new_elements );
    if( new_items != null ) this.items.addAll( new_items );
  }

  /**
   * Remove a rule from model.
   *
   * @param rule rule to be removed
   * @return true if rule is removed
   */
  public boolean removeRule(SequenceRuleManager rule) {
   return this.rules.remove(rule);
  }

  /**
   * Remove a set of rules from model.
   *
   * @param rules rules to be removed
   */
  public void removeRules(SequenceRuleManager[] rules) {
    for (int i = 0; i < rules.length; i++) {
      this.rules.remove(rules[i]);
    }
  }

  /**
   * Remove a sequence from model.
   * All referred rules are removed too.
   *
   * @param sequence sequence to be removed
   * @return true if sequence is removed
   */
  public boolean removeSequence(SequenceManager sequence) {
     boolean ris = this.sequences.remove(sequence);

     if (ris) {
       // remove sequence rules refered...
       Iterator iter = this.rules.iterator();
       Vector v = new Vector();
       while (iter.hasNext()) {
         SequenceRuleManager rule = (SequenceRuleManager) iter.next();
         if (rule.contains(sequence)){
           v.add(rule);
           break;
         }
       }

       for (int j=0; j<v.size();j++)
         this.rules.remove(v.get(j));

       //rimuove items e sequence elements non utilizzzati
       this.cleanData();

     }

     return ris;
  }

  /**
   * Remove a set of sequences from model.
   * All referred rules are removed too.
   *
   * @param sequences sequences to be removed
   */
  public void removeSequences(SequenceManager[] sequences) {
    for (int i = 0; i < sequences.length; i++) {
      this.removeSequence(sequences[i]);
    }

    // remove sequence rules refered...
    Iterator iter = this.rules.iterator();
    Vector v = new Vector();
    while (iter.hasNext()) {
      SequenceRuleManager rule = (SequenceRuleManager) iter.next();
      for (int i = 0; i < sequences.length; i++) {
        if (rule.contains(sequences[i])) {
          v.add(rule);
          break;
        }
      }
    }

    for (int j=0; j<v.size();j++)
      this.rules.remove(v.get(j));

    //rimuove items e sequence elements non utilizzzati
    this.cleanData();
  }

  /**
   * Check if this model contain given rule.
   *
   * @param rule rule to be cheked
   * @return true if rule is contained
   */
  public boolean contains(SequenceRuleManager rule) {
    Iterator iter = this.rules.iterator();
     while (iter.hasNext()) {
       SequenceRuleManager i = (SequenceRuleManager) iter.next();
       if (i.equals(rule))
         return true;
     }
     return false;
  }

  /**
   * Check if this model contain given sequence
   *
   * @param sequence Ssequence to be cheked
   * @return true if sequence is contained
   */
  public boolean contains(SequenceManager sequence) {
    Iterator iter = this.sequences.iterator();
     while (iter.hasNext()) {
       SequenceManager i = (SequenceManager) iter.next();
       if (i.equals(sequence))
         return true;
     }
     return false;
  }

  /**
   * Check if this model contain given element
   *
   * @param element element to be checked
   * @return true if element is contained
   */
  public boolean contains(SequenceElementManager element) {
    Iterator iter = this.elements.iterator();
     while (iter.hasNext()) {
       SequenceElementManager i = (SequenceElementManager) iter.next();
       if (i.equals(element))
         return true;
     }
     return false;
  }

  /**
   * Check if this model contain givenitem
   *
   * @param item item to be checked
   * @return true if item is contained
   */
  public boolean contains(ItemManager item) {
    Iterator iter = this.items.iterator();
     while (iter.hasNext()) {
       ItemManager i = (ItemManager) iter.next();
       if (i.equals(item))
         return true;
     }
     return false;
  }

  /**
   * Return an iterator over all items in this model.
   *
   * @return an items iterator
   */
  public Iterator getItems() {
    return this.items.iterator();
  }

  /**
   * Return an iterator over all sequence elements in this model.
   *
   * @return a sequence elements iterator
   */
  public Iterator getSequenceElements() {
    return this.elements.iterator();
  }

  /**
   * Return an iterator over all sequences in this model.
   *
   * @return a sequences iterator
   */
  public Iterator getSequences() {
    return this.sequences.iterator();
  }

  /**
   * Return an iterator over all sequence rules in this model.
   *
   * @return a sequence rules iterator
   */
  public Iterator getSequenceRules() {
    return this.rules.iterator();
  }

  /**
   * Return the number of transactions of input data used to build this model.
   *
   * @return number of transactions
   */
  public Integer getNumberOfTransactions() {
    return this.numberOfTransactions;
  }

  /**
   * Return the maximum number of items per transactions of input data
   * used to build this model.
   *
   * @return max number of items per transactions
   */
  public Integer getMaxNumberOfItemsPerTransaction() {
    return this.maxNumberOfItemsPerTA;
  }

  /**
   * Return the average number of items per transactions of input data
   * used to build this model.
   *
   * @return average number of items per transactions
   */
  public Double getAvgNumberOfItemsPerTransaction() {
    return this.avgNumberOfItemsPerTA;
  }

  /**
   * Return minimum support of sequences.
   *
   * @return minimum support
   */
  public Double getMinimumSupport() {
    return this.minimumSupport;
  }

  /**
   * Return minimum confidence of sequence rules.
   *
   * @return minimum confidence
   */
  public Double getMinimumConfidence() {
    return this.minimumConfidence;
  }

  /**
   * Return the maximum length of a sequence to be discovered.
   *
   * @return lenght limit
   */
  public Integer getLengthLimit() {
    return this.lengthLimit;
  }

  /**
   * Return the total number of items in this model.
   *
   * @return total number of items
   */
  public int getNumberOfItems() {
    if( this.items.size() == 0 ){
      int length = 0;
      Iterator i = this.getSequenceElements();
      while(i.hasNext())
        length += ((SequenceElementManager)i.next()).getNumberOfItems();
      return length;
    }
    return this.items.size();
  }

  /**
   * Return the total number of sequene elements in this model.
   *
   * @return total number of sequence elements
   */
  public int getNumberOfElements() {
    return this.elements.size();
  }

  /**
   * Return the total number of sequenes in this model.
   *
   * @return total number of sequences
   */
  public int getNumberOfSequences() {
    return this.sequences.size();
  }

  /**
   * Return the total number of sequene rules in this model.
   *
   * @return total number of sequence rules
   */
  public int getNumberOfRules() {
    return this.rules.size();
  }

  /**
   * Return the time window width:
   * this may be used to separate items associated with an object into discrete
   * events, but only if no clear key already exists for the separate events.
   * Two consecutive items must have a time gap of less than this value to be
   * considered as being part of the same event.
   *
   * @return time window width
   */
  public Integer getTimeWindowWidth() {
    return this.timeWindowWidth;
  }

  /**
   * Return the minimum time between items (see getTimeWindowWidth())
   *
   * @return minimum time
   */
  public Integer getMinimumTime() {
    return this.minimumTime;
  }

  /**
   * Return the maximum time between items (see getTimeWindowWidth())
   *
   * @return miximum time
   */
  public Integer getMaximumTime() {
    return this.maximumTime;
  }

  /**
   * Return a representation of the sequences contained in the model as a set of instances.
   * The table have a total number of instances given by the number of sequences,
   * and 5 fields representing:
   * <ol>
   * <li> the sequence as string (attribute sequence_elements); </li>
   * <li> the support of sequence (attribute sequence_support). </li>
   * <li> the number of transactions of sequence (attribute transactions_cardinality); </li>
   * <li> the number of distinct items (attribute distinct_items_cardinality). </li>
   * <li> the maximal number of items per transactions (attribute maximal_items_per_transactions). </li>
   *
   * @return the set of instances representing all sequences present in model.
   */
  public Object sequencesToInstances() {
    FastVector fv = null; // per indicare che l'attributo e' una stringa, viene passato un FastVector nullo.

    Attribute attr1 = new Attribute("sequence", fv);
    Attribute attr2 = new Attribute("sequence_support");
    Attribute attr3 = new Attribute("sequence_cardinality");
    Attribute attr4 = new Attribute("sequence_distinct_items_cardinality");
    Attribute attr5 = new Attribute("max_number_of_items_per_set");


    FastVector attributes = new FastVector();
    attributes.addElement(attr1);
    attributes.addElement(attr2);
    attributes.addElement(attr3);
    attributes.addElement(attr4);
    attributes.addElement(attr5);

    Instances ris = new Instances(this.model_name+"_sequences", attributes, this.getNumberOfSequences());

    // scorre le sequenze...
    Iterator iter = this.getSequences();
    while(iter.hasNext()) {
      Sequence sequence = (Sequence) iter.next();
      //crea l'instanza edlla sequenza...
      Instance instance = sequence.toInstance(ris);

      ris.add(instance);
    }

    ris.compactify();

    return ris;
  }

  /**
   * Transform items in model to PMML elements, and add it to
   * given sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @return items added
   */
  private Vector appendPMMLItems(Element sequence_model) {

      Vector ht_item = new Vector();
      Iterator iter = items.iterator();
      int id = 0;
      while(iter.hasNext()) {
        id++;
        Item item = (Item) iter.next();
        Element pmml = UPMMLBuilder.buildItem(item, id);
        sequence_model.appendChild(pmml);

        if (!ht_item.contains(item))
          ht_item.add(item);
      }
      return ht_item;

    }

  /**
   * Transform sequence elements in model to PMML elements,
   * and add it to given sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @param ht_items items composing sequence elements
   * @return sequence elements added
   */
  private Vector appendPMMLSets(Element sequence_model, Vector ht_items){
    //if( ht_items.isEmpty() )
      //return this.appendPMMLSetPredicates(sequence_model);
    return this.appendPMMLItemSets(sequence_model, ht_items);
  }

  /**
   * Transform itemset elements in model to PMML elements,
   * and add it to given sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @param ht_items items composing sequence elements
   * @return sequence elements added
   */
  private Vector appendPMMLItemSets(Element sequence_model, Vector ht_items) {

    Vector ht_elements = new Vector();
    Iterator iter = elements.iterator();
    int id = 0;
    ElementItemSet element;
    while (iter.hasNext()) {
      id++;
      element = (ElementItemSet) iter.next();
      Element pmml = UPMMLBuilder.buildItemSet(element, ht_items, id);
      sequence_model.appendChild(pmml);

      if (!ht_elements.contains(element))
        ht_elements.add(element);
    }
    return ht_elements;

  }

  /**
   * Transform set predicate elements in model to PMML elements,
   * and add it to given sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @return sequence elements added
   */
  /*protected Vector appendPMMLSetPredicates(Element sequence_model) {

    Vector ht_elements = new Vector();
    Iterator iter = elements.iterator();
    int id = 0;
    ElementSetPredicate element;
    while (iter.hasNext()) {
      id++;
      element = (ElementSetPredicate) iter.next();
      Element pmml = UPMMLBuilder.buildSetPredicate(element, id);
      sequence_model.appendChild(pmml);

      if (!ht_elements.contains(element))
        ht_elements.add(element);
    }
    return ht_elements;

  }*/

  /**
   * Transform sequences in model to PMML elements,
   * and add it to gievn sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @param ht_elements elements composing sequence
   * @return sequences added
   */
  private Vector appendPMMLSequences(Element sequence_model,
                                       Vector ht_elements) {

    Vector ht_sequences = new Vector();
    Iterator iter = sequences.iterator();
    int id = 0;
    Sequence sequence;
    while (iter.hasNext()) {
      id++;
      sequence = (Sequence) iter.next();
      Element pmml = UPMMLBuilder.buildSequence(sequence, ht_elements, id);
      sequence_model.appendChild(pmml);

      if (!ht_sequences.contains(sequence))
        ht_sequences.add(sequence);
    }
    return ht_sequences;

  }

  /**
   * Transform sequence rules in model to PMML elements,
   * and add it to gievn sequence model element.
   *
   * @param sequence_model sequence model PMML element
   * @param ht_sequences sequences composing rule
   */
  private void appendPMMLRules(Element sequence_model, Vector ht_sequences) {

    Iterator iter = rules.iterator();
    int id = 0;
    while (iter.hasNext()) {
      id++;
      SequenceRule rule = (SequenceRule) iter.next();
      Element pmml = UPMMLBuilder.buildRule(rule, ht_sequences, id);
      sequence_model.appendChild(pmml);
    }

  }

}
