/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.Iterator;
import kddml.Core.DataMining.AssociationRules.ItemManager;

/**
 * The interface for sequential patterns operations.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public interface SequenceModelManager extends kddml.Core.DataMining.MiningModelManager {

  /**
   * Add a rule to this model.
   * If rule is already contained in this model, or confidence of rule is less
   * than minimum confidence, or support of rule is less than minimum support,
   * rule is not added.
   *
   * @param rule rule to be added
   * @return true if rule is added
   */
  public boolean addRule(SequenceRuleManager rule);

  /**
   * Add a set of rules to this model.
   * If a rule is already contained in this model, or confidence of  a rule is
   * less than minimum confidence, or support of a rule is less than minimum
   * support, rule is not added.
   *
   * @param rules rules to be added
   */
  public void addRules(SequenceRuleManager[] rules);

  /**
   * Remove a rule from model.
   *
   * @param rule rule to be removed
   * @return true if rule is removed
   */
  public boolean removeRule(SequenceRuleManager rule);

  /**
   * Remove a set of rules from model.
   *
   * @param rules rules to be removed
   */
  public void removeRules(SequenceRuleManager[] rules);

  /**
   * Remove a sequence from model.
   * All referred rules are removed too.
   *
   * @param sequence sequence to be removed
   * @return true if sequence is removed
   */
  public boolean removeSequence(SequenceManager sequence);

  /**
   * Remove a set of sequences from model.
   * All referred rules are removed too.
   *
   * @param sequences sequences to be removed
   */
  public void removeSequences(SequenceManager[] sequences);

  /**
   * Check if this model contain given rule.
   *
   * @param rule rule to be cheked
   * @return true if rule is contained
   */
  public boolean contains(SequenceRuleManager rule);

  /**
   * Check if this model contain given sequence
   *
   * @param sequence Ssequence to be cheked
   * @return true if sequence is contained
   */
  public boolean contains(SequenceManager sequence);

  /**
   * Check if this model contain given element
   *
   * @param element element to be checked
   * @return true if element is contained
   */
  public boolean contains(SequenceElementManager element);

  /**
   * Check if this model contain givenitem
   *
   * @param item item to be checked
   * @return true if item is contained
   */
  public boolean contains(ItemManager item);

  /**
   * Return an iterator over all items in this model.
   *
   * @return an items iterator
   */
  public Iterator getItems();

  /**
   * Return an iterator over all sequence elements in this model.
   *
   * @return a sequence elements iterator
   */
  public Iterator getSequenceElements();

  /**
   * Return an iterator over all sequences in this model.
   *
   * @return a sequences iterator
   */
  public Iterator getSequences();

  /**
   * Return an iterator over all sequence rules in this model.
   *
   * @return a sequence rules iterator
   */
  public Iterator getSequenceRules();

  /**
   * Return the number of transactions of input data used to build this model.
   *
   * @return number of transactions
   */
  public Integer getNumberOfTransactions();

  /**
   * Return the maximum number of items per transactions of input data
   * used to build this model.
   *
   * @return max number of items per transactions
   */
  public Integer getMaxNumberOfItemsPerTransaction();

  /**
   * Return the average number of items per transactions of input data
   * used to build this model.
   *
   * @return average number of items per transactions
   */
  public Double getAvgNumberOfItemsPerTransaction();

  /**
   * Return minimum support of sequences.
   *
   * @return minimum support
   */
  public Double getMinimumSupport();

  /**
   * Return minimum confidence of sequence rules.
   *
   * @return minimum confidence
   */
  public Double getMinimumConfidence();

  /**
   * Return the maximum length of a sequence to be discovered.
   *
   * @return length limit
   */
  public Integer getLengthLimit();

  /**
   * Return the total number of items in this model.
   *
   * @return total number of items
   */
  public int getNumberOfItems();

  /**
   * Return the total number of sequene elements in this model.
   *
   * @return total number of sequence elements
   */
  public int getNumberOfElements();

  /**
   * Return the total number of sequenes in this model.
   *
   * @return total number of sequences
   */
  public int getNumberOfSequences();

  /**
   * Return the total number of sequene rules in this model.
   *
   * @return total number of sequence rules
   */
  public int getNumberOfRules();

  /**
   * Return the time window width:
   * this may be used to separate items associated with an object into discrete
   * events, but only if no clear key already exists for the separate events.
   * Two consecutive items must have a time gap of less than this value to be
   * considered as being part of the same event.
   *
   * @return time window width
   */
  public Integer getTimeWindowWidth();

  /**
   * Return the minimum time between items (see getTimeWindowWidth())
   *
   * @return minimum time
   */
  public Integer getMinimumTime();

  /**
   * Return the maximum time between items (see getTimeWindowWidth())
   *
   * @return maximum time
   */
  public Integer getMaximumTime();

  /**
   * Return a representation of the sequences contained in the model as a set of instances.
   * The table have a total number of instances given by the number of sequences,
   * and 5 fields representing:
   * <ol>
   * <li> the sequence as string (attribute sequence_elements); </li>
   * <li> the support of sequence (attribute sequence_support). </li>
   * <li> the number of transactions of sequence (attribute transactions_cardinality); </li>
   * <li> the number of distinct items (attribute distinct_items_cardinality). </li>
   * <li> the maximal number of items per transactions (attribute maximal_items_per_transactions). </li>
   *
   * @return the set of instances representing all sequences present in model.
   */
  public Object sequencesToInstances();

  /**
   * Clone this model returning a new sequence model.
   *
   * @param output_path File the path in the model system repository containing the cloned
   * model
   * @return cloned model
   */
  public SequenceModel clone(java.io.File output_path);

}
