/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.util.TreeSet;

/**
 * A factory class that create a generic sequence model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public abstract class SequenceProprietaryFactory extends ModelProprietaryFactory {

  protected Double minSupport;

  protected Double minConfidence;

  protected Integer numberOfTransactions;

  protected Integer maxNumberOfItemsPerTA;

  protected Double avgNumberOfItemsPerTA;

  protected Integer lengthLimit;

  protected Integer timeWindowWidth;

  protected Integer minimumTime;

  protected Integer maximumTime;

  protected TreeSet items;

  protected TreeSet elements;

  protected TreeSet sequences;

  protected TreeSet rules;

  /**
   * Constructor.
   *
   * @param f the file containing the model
   * @param dd the data dictionary
   * @param ms the mining schema
   * @param min_support the minimum support
   * @param min_confidence the minimum confidence
   * @param number_of_transactions the number of transactions
   * @throws KDDMLCoreException
   */
  public SequenceProprietaryFactory(java.io.File f,
                                         DataDictionary dd,
                                         MiningSchema ms,
                                         double min_support,
                                         double min_confidence,
                                         int number_of_transactions) throws
      KDDMLCoreException {

    super(f, dd, ms);
    // controlli sui parametri immessi
    UChecker.checkMinSupport(new Double(min_support));
    this.minSupport = new Double(min_support);

    UChecker.checkMinConfidence(new Double(min_confidence));
    this.minConfidence = new Double(min_confidence);

    UChecker.checkNumTransactions(new Integer(number_of_transactions));
    this.numberOfTransactions = new Integer(number_of_transactions);

  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    SequenceModel model = new SequenceModel(physical_file);

    model = (SequenceModel) setMiningSchema( (MiningModel) model);
    model = (SequenceModel) setDataDictionary( (MiningModel) model);

    if (algorithm_name != null)
      model = (SequenceModel) setAlgorithmName( (MiningModel) model);
    if (model_name != null)
      model = (SequenceModel) setModelName( (MiningModel) model);

    model.minimumSupport = minSupport;
    model.minimumConfidence = minConfidence;
    model.numberOfTransactions = numberOfTransactions;

    if( maxNumberOfItemsPerTA != null )
      model.maxNumberOfItemsPerTA = maxNumberOfItemsPerTA;

    if( avgNumberOfItemsPerTA != null )
      model.avgNumberOfItemsPerTA = avgNumberOfItemsPerTA;

    return model;
  }

  /**
   * Set the maximum number of items per transaction.
   * Here transaction is intended as the entirely set of transactions
   * with same identifier.
   *
   * @param i maximum items per transaction.
   * @throws SequenceModelException
   */
  public void setMaxNumberOfItemsPerTA(int i) throws SequenceModelException {
    UChecker.checkMaxItems(new Integer(i));
    maxNumberOfItemsPerTA = new Integer(i);
  }

  /**
   * Set the average number of items per transaction.
   * Here transaction is intended as the entirely set of transactions
   * with same identifier.
   *
   * @param i average items per transaction.
   * @throws SequenceModelException
   */
  public void setAvgNumberOfItemsPerTA(double i) throws
      SequenceModelException {
    UChecker.checkAvgItems(new Double(i));
    avgNumberOfItemsPerTA = new Double(i);
  }
}

