/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.ArrayList;
import java.util.Iterator;

/**
 * Basically a rule tells the relation between the elements of a sequence, i.e.
 * probability that an ordered group of elements in sequence follow another one.
 *
 * A sequence rule consists of an antecedent and a consequent
 * sequence, separated by a delimiter and, possibly, time.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SequenceRule implements SequenceRuleManager{

  /**
   * antecent sequence
   */
  protected SequenceManager antecent;

  /**
   * times between elements in antecedent sequence
   */
  protected ArrayList antTimes;

  /**
   * delimiter between antecedent and consequent sequence
   */
  protected DelimiterManager delimiter;

  /**
   * times between antecedent and consequent sequence
   */
  protected ArrayList times;

  /**
   * consequent sequence
   */
  protected SequenceManager consequent;

  /**
   * times between elements in consequent sequence
   */
  protected ArrayList consTimes;

  /**
   * The ratio of the number of objects in the data for which the
   * antecedent and consequent Sequences hold true, to the total
   * number of objects in the data.
   */
  protected Double support;

  /**
   * Probability of the consequent following the antecedent.
   * Calculated as the number of occurrences of a sequence divided
   * by the number of occurrences of the antecedent.
   */
  protected Double confidence;

  /**
   * The number of objects in the data for which the antecedent
   * and consequent sequences hold true.
   */
  protected Integer occurrence;

  /**
   * Build a new rule with given arguments.
   *
   * @param antecedent antecedent sequence
   * @param delimiter delimiter between antecedent and consequent sequence
   * @param consequent consequence sequence
   * @param confidence rule confidence - between 0 and 1 inclusive
   * @param support rule support - between 0 and 1 inclusive
   * @param occurrence number of occurrence of rule - greater than 0
   * @throws IllegalArgumentException if confidence, support or occurence don't
   *   respect specified range values
   */
  public SequenceRule(SequenceManager antecedent, DelimiterManager delimiter,
                      SequenceManager consequent, double confidence,
                      double support, int occurrence) throws
      IllegalArgumentException {
    this.antecent = antecedent;
    this.antTimes = new ArrayList();
    this.delimiter = delimiter;
    this.times = new ArrayList();
    this.consequent = consequent;
    this.consTimes = new ArrayList();
    this.support = new Double(support);
    this.confidence = new Double(confidence);
    this.occurrence = new Integer(occurrence);

    if( confidence <= 0 || confidence > 1 ){
      throw new IllegalArgumentException();
    }
    if( support <= 0 || support > 1 ){
      throw new IllegalArgumentException();
    }
    if( occurrence <= 0 ){
      throw new IllegalArgumentException();
    }
  }

  /**
   * Set times between antecedent and consequent sequences.
   *
   * @param times times between sequences
   */
  public void setTimes(TimeManager[] times){
    if( times != null ){
      for (int i = 0; i < times.length; i++) {
        this.times.add(times[i]);
      }
    }
  }

  /**
   * Set times between elements in antecedent sequence.
   * Each TimeManager object represent time information between
   * two following elements in sequence.
   * E.g. let be s = <e1, e2, e3> the antecedent sequence, setting times
   * for s, mean to give two time information, respectively the time between
   * e1 and e2, and the time between e2 and e3.
   *
   * @param times times between elements
   * @throws IllegalArgumentException times mismatch elements in antecedent
   *   sequence
   */
  public void setAntecedentTimes(TimeManager[] times)
      throws IllegalArgumentException{
    if( times != null ){
      if( this.antecent.getNumberOfElements()-1 != times.length ){
        throw new IllegalArgumentException();
      }
      for (int i = 0; i < times.length; i++) {
        this.antTimes.add(times[i]);
      }
    }
  }

  /**
   * Set times between elements in consequent sequence.
   * Each TimeManager object represent time information between
   * two following elements in sequence.
   * E.g. let be s = <e1, e2, e3> the consequent sequence, setting times
   * for s, mean to give two time information, respectively the time between
   * e1 and e2, and the time between e2 and e3.
   *
   * @param times times between elements
   * @throws IllegalArgumentException times mismatch elements in consequent
   *   sequence
   */
  public void setConsequentTimes(TimeManager[] times)
      throws IllegalArgumentException{
    if( times != null ){
      if( this.consequent.getNumberOfElements()-1 != times.length ){
        throw new IllegalArgumentException();
      }
      for (int i = 0; i < times.length; i++) {
        this.consTimes.add(times[i]);
      }
    }
  }

  /**
   * Return antecedent sequence composing this rule.
   *
   * @return antecedent sequence
   */
  public SequenceManager getAntecedent() {
    return this.antecent;
  }

  /**
   * Return an iterator over times between elements in antecedent sequence.
   *
   * @return antecedent sequence times iterator
   */
  public Iterator getAntecedentTimes() {
    return this.antTimes.iterator();
  }

  /**
   * Return confidence of this rule:
   * probability of the consequent following the antecedent.
   * Calculated as the number of occurrences of a sequence divided
   * by the number of occurrences of the antecedent.
   *
   * Confidence is expressed as probabilistic number between 0 and 1.
   *
   * @return confidence
   */
  public Double getConfidence() {
    return this.confidence;
  }

  /**
   * Return consequent sequence composing this rule.
   *
   * @return consequent sequence
   */
  public SequenceManager getConsequent() {
    return this.consequent;
  }

  /**
   * Return an iterator over times between elements in consequent sequence.
   *
   * @return consequent sequence times iterator
   */
  public Iterator getConsequentTimes() {
    return this.consTimes.iterator();
  }

  /**
   * Return delimiter between antecedent and consequent sequence.
   *
   * @return delimiter
   */
  public DelimiterManager getDelimiter() {
    return this.delimiter;
  }

  /**
   * Return the total number of sequence elements in both the antecedent
   * and consequent sequences.
   *
   * @return number of sequence elements
   */
  public int getNumberOfSequenceElements() {
    return (this.antecent.getNumberOfElements() +
            this.consequent.getNumberOfElements());
  }

  /**
   * Return number of occurrence of this rule:
   * the number of objects in the data for which the antecedent
   * and consequent sequences hold true.
   *
   * @return occurrence
   */
  public Integer getOccurrence() {
    return this.occurrence;
  }

  /**
   * Return support of this rule:
   * the ratio of the number of objects in the data for which the
   * antecedent and consequent Sequences hold true, to the total
   * number of objects in the data.
   *
   * Support is expressed as probabilistic number between 0 and 1.
   *
   * @return support
   */
  public Double getSupport() {
    return this.support;
  }

  /**
   * Return an iterator over times between antecedent and consequent sequence.
   *
   * @return rule times iterator
   */
  public Iterator getTimes() {
    return this.times.iterator();
  }

  /**
   * Check if antecedent or consequent sequences contain given sequence.
   *
   * @param sequence sequence to be checked
   * @return true if sequence belog to this rule
   */
  public boolean contains(SequenceManager sequence) {
    return (this.antecent.contains(sequence) ||
            this.consequent.contains(sequence));
  }

  /**
   * Check when two rules are equals.
   * To be equal, two rules must have the same antecedent and consequent
   * sequences.
   *
   * @param sequence_rule object to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence_rule){
    SequenceRuleManager rule = (SequenceRuleManager) sequence_rule;

    boolean ant = this.getAntecedent().equals( rule.getAntecedent() );
    boolean cons = this.getConsequent().equals( rule.getConsequent() );

    return (ant && cons);
  }

  /**
   * Return a string representation of rule.
   *
   * @return this object as string
   */
  public String toString(){
    String s = "";
    s = s + this.getAntecedent().toString();
    s = s + this.getDelimiter().toString();
    s = s + this.getConsequent().toString();
    s = s + "(supp=" + this.getSupport() + ")";
    s = s + "(conf=" + this.getConfidence() + ")";
    s = s + "(occ=" + this.getOccurrence() + ")";

    return s;
  }

}
