/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import java.util.Iterator;

/**
 * Basically a rule tells the relation between the elements of a sequence, i.e.
 * probability that an ordered group of elements in sequence follow another one.
 *
 * A sequence rule consists of an antecedent sequence and a consequent
 * sequence, separated by a delimiter and, possibly, time.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public interface SequenceRuleManager {

  /**
   * Return the total number of sequence elements in both the antecedent
   * and consequent sequences.
   *
   * @return number of sequence elements
   */
  public int getNumberOfSequenceElements();

  /**
   * Return number of occurrence of this rule:
   * the number of objects in the data for which the antecedent
   * and consequent sequences hold true.
   *
   * @return occurrence
   */
  public Integer getOccurrence();

  /**
   * Return support of this rule:
   * the ratio of the number of objects in the data for which the
   * antecedent and consequent Sequences hold true, to the total
   * number of objects in the data.
   *
   * Support is expressed as probabilistic number between 0 and 1.
   *
   * @return support
   */
  public Double getSupport();

  /**
   * Return confidence of this rule:
   * probability of the consequent following the antecedent.
   * Calculated as the number of occurrences of a sequence divided
   * by the number of occurrences of the antecedent.
   *
   * Confidence is expressed as probabilistic number between 0 and 1.
   *
   * @return confidence
   */
  public Double getConfidence();

  /**
   * Return antecedent sequence composing this rule.
   *
   * @return antecedent sequence
   */
  public SequenceManager getAntecedent();

  /**
   * Return an iterator over times between elements in antecedent sequence.
   *
   * @return antecedent sequence times iterator
   */
  public Iterator getAntecedentTimes();

  /**
   * Return delimiter between antecedent and consequent sequence.
   *
   * @return delimiter
   */
  public DelimiterManager getDelimiter();

  /**
   * Return an iterator over times between antecedent and consequent sequence.
   *
   * @return rule times iterator
   */
  public Iterator getTimes();

  /**
   * Return consequent sequence composing this rule.
   *
   * @return consequent sequence
   */
  public SequenceManager getConsequent();

  /**
   * Return an iterator over times between elements in consequent sequence.
   *
   * @return consequent sequence times iterator
   */
  public Iterator getConsequentTimes();

  /**
   * Set times between antecedent and consequent sequences.
   *
   * @param times times between sequences
   */
  public void setTimes(TimeManager[] times);

  /**
   * Set times between elements in antecedent sequence.
   * Each TimeManager object represent time information between
   * two following elements in sequence.
   * E.g. let be s = <e1, e2, e3> the antecedent sequence, setting times
   * for s, mean to give two time information, respectively the time between
   * e1 and e2, and the time between e2 and e3.
   *
   * @param times times between elements
   * @throws IllegalArgumentException times mismatch elements in antecedent
   *   sequence
   */
  public void setAntecedentTimes(TimeManager[] times)
      throws IllegalArgumentException;

  /**
   * Set times between elements in consequent sequence.
   * Each TimeManager object represent time information between
   * two following elements in sequence.
   * E.g. let be s = <e1, e2, e3> the consequent sequence, setting times
   * for s, mean to give two time information, respectively the time between
   * e1 and e2, and the time between e2 and e3.
   *
   * @param times times between elements
   * @throws IllegalArgumentException times mismatch elements in consequent
   *   sequence
   */
  public void setConsequentTimes(TimeManager[] times)
      throws IllegalArgumentException;

  /**
   * Check if antecedent or consequent sequences contain given sequence.
   *
   * @param sequence sequence to be checked
   * @return true if sequence belog to this rule
   */
  public boolean contains(SequenceManager sequence);

  /**
   * Check when two rules are equals.
   * To be equal, two rules must have the same antecedent and consequent
   * sequences.
   *
   * @param sequence_rule object to be checked
   * @return true if objects are equals, false otherwise
   */
  public boolean equals(Object sequence_rule);

  /**
   * Return a string representation of rule.
   *
   * @return this object as string
   */
  public String toString();

}
