/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.AssociationRules.Item;

import kddml.Core.XMLUtility;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import java.util.Iterator;
import java.util.Vector;

/**
 * This utility class contains methods used inside this package
 * to build PMML representation of sequence model.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
final class UPMMLBuilder {

  /**
   * Utility to build a PMML SequencModel element.
   *
   * PMML v2.0 SequenceModel
   *
   * <!ELEMENT SequenceModel (Extension*, MiningSchema, Item*, Itemset*,
   *                         SetPredicate*, Sequence+,
   *                         SequenceRule*, Extension*)>
   *
   * <!ATTLIST SequenceModel
   *  modelName                       CDATA          #IMPLIED
   *  functionName                    %MINING-FUNCTION; #REQUIRED
   *  algorithmName                   CDATA          #IMPLIED
   *  numberOfTransactions            %INT-NUMBER;   #REQUIRED
   *  maxNumberOfItemsPerTransaction  %INT-NUMBER;   #IMPLIED
   *  avgNumberOfItemsPerTransaction  %REAL-NUMBER;  #IMPLIED
   *  minimumSupport                  %REAL-NUMBER;  #REQUIRED
   *  minimumConfidence               %REAL-NUMBER;  #REQUIRED
   *  lengthLimit                     %INT-NUMBER;   #IMPLIED
   *  numberOfItems                   %INT-NUMBER;   #REQUIRED
   *  numberOfSets                    %INT-NUMBER;   #REQUIRED
   *  numberOfSequences               %INT-NUMBER;   #REQUIRED
   *  numberOfRules                   %INT-NUMBER;   #REQUIRED
   *  timeWindowWidth                 %INT-NUMBER;   #IMPLIED
   *  minimumTime                     %INT-NUMBER;   #IMPLIED
   *  maximumTime                     %INT-NUMBER;   #IMPLIED
   * >
   *
   * @param model SequenceModel
   * @return Element
   */
  protected static Element buildModel(SequenceModel model){

    Element pmmlModel = XMLUtility.doc.createElement("SequenceModel");

    //Set SequenceModel attributes...
    if( model.getModelName() != null && ! model.getModelName().equals("") )
      pmmlModel.setAttribute("modelName", model.getModelName());

    pmmlModel.setAttribute("functionName", "sequences");

    if( model.getAlgorithmName() != null && !model.getAlgorithmName().equals("")   )
      pmmlModel.setAttribute("algorithmName", model.getAlgorithmName());

    pmmlModel.setAttribute("numberOfTransactions", model.getNumberOfTransactions().toString());

    if (model.getMaxNumberOfItemsPerTransaction() != null)
      pmmlModel.setAttribute("maxNumberOfItemsPerTransaction", model.getMaxNumberOfItemsPerTransaction().toString());

    if (model.getAvgNumberOfItemsPerTransaction() != null)
      pmmlModel.setAttribute("avgNumberOfItemsPerTransaction", model.getAvgNumberOfItemsPerTransaction().toString());

    pmmlModel.setAttribute("minimumSupport", model.getMinimumSupport().toString());
    pmmlModel.setAttribute("minimumConfidence", model.getMinimumConfidence().toString());

    if (model.lengthLimit != null)
      pmmlModel.setAttribute("lengthLimit", model.getLengthLimit().toString());

    pmmlModel.setAttribute("numberOfItems", new Integer(model.getNumberOfItems()).toString());

    pmmlModel.setAttribute("numberOfSets", new Integer(model.getNumberOfElements()).toString());

    pmmlModel.setAttribute("numberOfSequences", new Integer(model.getNumberOfSequences()).toString());

    pmmlModel.setAttribute("numberOfRules", new Integer(model.getNumberOfRules()).toString());

    if (model.getTimeWindowWidth() != null)
      pmmlModel.setAttribute("timeWindowWidth", model.getTimeWindowWidth().toString());

    if (model.getMinimumTime() != null)
      pmmlModel.setAttribute("minimumTime", model.getMinimumTime().toString());

    if (model.getMaximumTime() != null)
      pmmlModel.setAttribute("maximumTime", model.getMaximumTime().toString());

    return pmmlModel;
  }

  /**
   * Utility to build a PMML SequenceRule element.
   *
   * PMML v2.0 SequenceRule
   *
   * <!ELEMENT SequenceRule ( Extension*, AntecedentSequence, Delimiter,
   *       Time*, ConsequentSequence )>
   *
   * <!ATTLIST SequenceRule
   *  id                %ELEMENT-ID;  #REQUIRED
   *  numberOfSets      %INT-NUMBER;  #REQUIRED
   *  occurrence        %INT-NUMBER;  #REQUIRED
   *  support           %REAL-NUMBER; #REQUIRED
   *  confidence        %REAL-NUMBER; #REQUIRED
   * >
   *
   * <!ENTITY % SEQUENCE  "( SequenceReference, Time* )" >
   *
   * <!ELEMENT SequenceReference EMPTY >
   * <!ATTLIST SequenceReference
   *  seqId     %ELEMENT-ID; #REQUIRED
   * >
   *
   * <!ELEMENT AntecedentSequence (%SEQUENCE;) >
   * <!ELEMENT ConsequentSequence (%SEQUENCE;) >
   *
   * @param rule SequenceRule
   * @param sequenceReferences Vector
   * @param ruleId int
   * @return Element
   */
  protected static Element buildRule(SequenceRule rule, Vector sequenceReferences, int ruleId){

    Element pmmlRule = XMLUtility.doc.createElement("SequenceRule");

    //Set Rule attributes...
    pmmlRule.setAttribute("id", new Integer(ruleId).toString());
    pmmlRule.setAttribute("numberOfSets", new Integer(rule.getNumberOfSequenceElements()).toString());
    pmmlRule.setAttribute("occurrence", rule.getOccurrence().toString());
    pmmlRule.setAttribute("support", rule.getSupport().toString());
    pmmlRule.setAttribute("confidence", rule.getConfidence().toString());

    //Set AntecedentSequence element...
    Sequence antecedentSequence = (Sequence)rule.getAntecedent();
    Integer antecedentSequenceId = new Integer(sequenceReferences.indexOf(antecedentSequence) + 1);

    Element pmmlAntecedentSequenceReference = XMLUtility.doc.createElement("SequenceReference");
    pmmlAntecedentSequenceReference.setAttribute("seqId", antecedentSequenceId.toString());

    Element pmmlAntecedentSequence = XMLUtility.doc.createElement("AntecedentSequence");
    pmmlAntecedentSequence.appendChild(pmmlAntecedentSequenceReference);

    //Set AntecedentSequence Time(s)...
    Iterator iterator = rule.getAntecedentTimes();
    Time time;
    while( iterator.hasNext() ){
      time = (Time) iterator.next();
      pmmlAntecedentSequence.appendChild( buildTime(time) );
    }

    //Append AntecedentSequence element to Rule element...
    pmmlRule.appendChild( pmmlAntecedentSequence );

    //Set Rule Delimiter and Time(s)...
    pmmlRule.appendChild( buildDelimiter((Delimiter)rule.getDelimiter()) );
    iterator = rule.getTimes();
    while( iterator.hasNext() ){
      time = (Time) iterator.next();
      pmmlRule.appendChild( buildTime(time) );
    }

    //Set ConsequentSequence element...
    Sequence consequentSequence = (Sequence)rule.getConsequent();
    Integer consequentSequenceId = new Integer(sequenceReferences.indexOf(consequentSequence) + 1);

    Element pmmlConsequentSequenceReference = XMLUtility.doc.createElement("SequenceReference");
    pmmlConsequentSequenceReference.setAttribute("seqId", consequentSequenceId.toString());

    Element pmmlConsequentSequence = XMLUtility.doc.createElement("ConsequentSequence");
    pmmlConsequentSequence.appendChild(pmmlConsequentSequenceReference);

    //Set AntecedentSequence Time(s)...
    iterator = rule.getConsequentTimes();
    while( iterator.hasNext() ){
      time = (Time) iterator.next();
      pmmlConsequentSequence.appendChild( buildTime(time) );
    }

    //Append ConsequentSequence element to Rule element...
    pmmlRule.appendChild( pmmlConsequentSequence );

    return pmmlRule;

  }

  /**
   * Utility to build a PMML Sequence element.
   *
   * PMML v2.0 Sequence
   *
   * <!ENTITY % FOLLOW_SET  "(Delimiter, SetReference )" >
   *
   * <!ELEMENT Sequence ( SetReference, (%FOLLOW_SET;)* )>
   * <!ATTLIST Sequence
   *  id                %ELEMENT-ID;  #REQUIRED
   *  numberOfSets      %INT-NUMBER;  #IMPLIED
   *  occurrence        %INT-NUMBER;  #IMPLIED
   *  support           %REAL-NUMBER; #IMPLIED
   * >
   *
   * @param sequence Sequence
   * @param elementReferences Vector
   * @param sequenceId int
   * @return Element
   */
  protected static Element buildSequence(Sequence sequence, Vector elementReferences, int sequenceId){

    Element pmmlSequence = XMLUtility.doc.createElement("Sequence");

    //Set Sequence attributes...
    pmmlSequence.setAttribute("id", new Integer(sequenceId).toString());
    pmmlSequence.setAttribute("numberOfSets", new Integer(sequence.getNumberOfElements()).toString());
    if (sequence.getOccurrence() != null )
      pmmlSequence.setAttribute("occurrence", sequence.getOccurrence().toString());
    if (sequence.getSupport() != null)
      pmmlSequence.setAttribute("support", sequence.getSupport().toString() );

      //Set SetReference(s)...
    Iterator iterator = sequence.getElements();
    SequenceElement element;
    while (iterator.hasNext()) {
      element = (SequenceElement) iterator.next();
      Integer setId = new Integer( elementReferences.indexOf(element) + 1);

      Element pmmlSetReference = XMLUtility.doc.createElement("SetReference");
      pmmlSetReference.setAttribute("setId", setId.toString());
      pmmlSequence.appendChild(pmmlSetReference);
      //Set Delimiter...
      if( iterator.hasNext() )
        pmmlSequence.appendChild( buildDelimiter((Delimiter)iterator.next()) );
    }

    return pmmlSequence;

  }

  /**
   * Utility to build a PMML SetPredicate element.
   *
   * PMML v2.0 SetPredicate
   *
   * <!ENTITY % ELEMENT-ID    "CDATA">
   *
   * <!ELEMENT SetPredicate ( %STRING-ARRAY; )>
   * <!ATTLIST SetPredicate
   *  id               %ELEMENT-ID;              #IMPLIED
   *  field            %FIELD-NAME;              #REQUIRED
   *  operator         CDATA                     #FIXED "supersetOf"
   * >
   *
   * <!ELEMENT ItemSetReference EMPTY>
   * <!ATTLIST ItemSetReference
   *  itemSetId        %ELEMENT-ID;              #REQUIRED
   * >
   *
   * @param setPredicate ElementSetPredicate
   * @param setId int
   * @return Element
   */
  /*protected static Element buildSetPredicate(ElementSetPredicate setPredicate, int setId){

      Element pmmlSetPredicate = XMLUtility.doc.createElement("SetPredicate");

      //Set SetPredicate attributes...
      pmmlSetPredicate.setAttribute("id", new Integer(setId).toString());
      pmmlSetPredicate.setAttribute("field", setPredicate.getField());
      pmmlSetPredicate.setAttribute("operator", "supersetOf");

      //Set Array element...
      Element pmmlArray = XMLUtility.doc.createElement("Array");
      pmmlArray.setAttribute("n", new Integer(setPredicate.getNumberOfItems()).toString());

      //Set Array values...
      Iterator iterator = setPredicate.getItems();
      String itemsValues = "";
      while (iterator.hasNext())
        itemsValues += " " + ((Item) iterator.next()).toString();

      Node text = XMLUtility.doc.createTextNode(itemsValues);
      pmmlArray.appendChild(text);

      pmmlSetPredicate.appendChild(pmmlArray);

      return pmmlSetPredicate;

  }*/

  /**
   * Utility to build a PMML Itemset element.
   *
   * PMML v2.0 Itemset
   *
   * @param itemset ElementItemSet
   * @param itemReferences Vector
   * @param setId int
   * @return Element
   */
  protected static Element buildItemSet(ElementItemSet itemset, Vector itemReferences, int setId){

    Element pmmlItemset = XMLUtility.doc.createElement("Itemset");

    //Set Itemset Attributes...
    pmmlItemset.setAttribute("id", new Integer(setId).toString());
    if (itemset.getSupport() != null)
      pmmlItemset.setAttribute("support", itemset.getSupport().toString());
    pmmlItemset.setAttribute("numberOfItems", new Integer(itemset.getNumberOfItems()).toString());

    //Set Item(s)...
    Iterator iterator = itemset.getItems();
    Item item;
    while (iterator.hasNext()) {
      item = (Item) iterator.next();
      Integer itemId = new Integer(itemReferences.indexOf(item) + 1);
      Element pmmlItem = XMLUtility.doc.createElement("ItemRef");
      pmmlItem.setAttribute("itemRef", itemId.toString());
      pmmlItemset.appendChild(pmmlItem);
    }

    return pmmlItemset;

  }

  /**
   * Utility to build a PMML Item element.
   *
   * PMML v2.0 Item
   *
   * @param item Item
   * @param itemId int
   * @return Element
   */
  protected static Element buildItem(Item item, int itemId){

    Element pmmlItem = XMLUtility.doc.createElement("Item");

    //Set Item attributes...
    pmmlItem.setAttribute("id", new Integer(itemId).toString());
    pmmlItem.setAttribute("value", item.toString());
    if (item.getMappedValue() != null)
      pmmlItem.setAttribute("mappedValue", item.getMappedValue());
    if (item.getWeight() != null)
      pmmlItem.setAttribute("weight", item.getWeight().toString());

    return pmmlItem;

  }

  /**
   * Utility to build a PMML Time element.
   *
   * PMML v2.0 Time
   *
   * @param time Time
   * @return Element
   */
  protected static Element buildTime(Time time) {
    Element pmmlTime = XMLUtility.doc.createElement("Time");

    pmmlTime.setAttribute("min", time.getMin().toString());
    pmmlTime.setAttribute("max", time.getMax().toString());
    if( time.getMean() != null )
      pmmlTime.setAttribute("mean", time.getMean().toString());

    return pmmlTime;
  }

  /**
   * Utility to build a PMML Delimiter element.
   *
   * PMML v2.0 Delimiter
   *
   * @param delimiter Delimiter
   * @return Element
   */
  protected static Element buildDelimiter(Delimiter delimiter) {
    Element pmmlDelimiter = XMLUtility.doc.createElement("Delimiter");

    pmmlDelimiter.setAttribute("delimiter", delimiter.getDELIMITER().toString());
    pmmlDelimiter.setAttribute("gap", delimiter.getGAP().toString());

    return pmmlDelimiter;
  }

}
