/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataMining.SequentialPatterns;

import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemManager;

import kddml.Core.KDDMLCoreErrorCodes;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import java.util.Hashtable;

/**
 * This utility class contains methods used inside this package
 * to build a sequence model from its PMML representation.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
final class UPMMLParser {

  protected static SequenceRuleManager parseRule(Element rule, Hashtable sequenceReferences) throws Exception{
    //Get rule attributes...
    int numberOfSets = new Integer(rule.getAttribute("numberOfSets")).intValue();
    Integer occurrence = new Integer(rule.getAttribute("occurrence"));
    Double support = new Double(rule.getAttribute("support"));
    try{
      UChecker.checkMinSupport(support);
    }catch(SequenceModelException sme){
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: " + sme.getErrorMessage()
          , sme.getDetailErrorMessage(),
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }
    Double confidence = new Double(rule.getAttribute("confidence"));
    try{
      UChecker.checkMinConfidence(confidence);
    }catch(SequenceModelException sme){
      throw new SequenceModelException(
      "Illegal PMML Sequence Model: " + sme.getErrorMessage()
      , sme.getDetailErrorMessage(),
      KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    }

    //Get AntecentSequence element...
    Element pmmlAntecedent = (Element)rule.getElementsByTagName("AntecedentSequence").item(0);
    Element pmmlAntecedentReference = (Element)pmmlAntecedent.getElementsByTagName("SequenceReference").item(0);
    String antecedentSeqId = pmmlAntecedentReference.getAttribute("seqId");
    //Get AntecedentSequence Time(s)...
    NodeList pmmlAntecedentTimes = pmmlAntecedent.getElementsByTagName("Time");
    TimeManager[] antecedentSequenceTimes = new TimeManager[pmmlAntecedentTimes.getLength()];
    for (int j = 0; j < pmmlAntecedentTimes.getLength(); j++) {
      Element pmmlTime = (Element) pmmlAntecedentTimes.item(j);
      TimeManager tm = parseTime(pmmlTime);
      antecedentSequenceTimes[j] = tm;
    }

    //Get Delimiter element...
    Element pmmlDelimiter = (Element)rule.getElementsByTagName("Delimiter").item(0);
    DelimiterManager delimiter = parseDelimiter(pmmlDelimiter);

    //Get the Time element(s)
    NodeList timeList = rule.getElementsByTagName("Time");
    TimeManager[] ruleTimes = new TimeManager[timeList.getLength()];

    for (int j = 0; j < timeList.getLength(); j++) {
      Element pmmlTime = (Element) timeList.item(j);
      if( pmmlTime.getParentNode().getNodeName().equals("SequenceRule") ){
        TimeManager tm = parseTime(pmmlTime);
        ruleTimes[j] = tm;
      }
    }

    //Get ConsequentSequence element...
    Element pmmlConsequent = (Element)rule.getElementsByTagName("ConsequentSequence").item(0);
    Element pmmlConsequentReference = (Element)pmmlConsequent.getElementsByTagName("SequenceReference").item(0);
    String consequentSeqId = pmmlConsequentReference.getAttribute("seqId");
    //Get ConsequentSequence Time(s)...
    NodeList pmmlConsequentTimes = pmmlConsequent.getElementsByTagName("Time");
    TimeManager[] consequentSequenceTimes = new TimeManager[pmmlConsequentTimes.getLength()];
    for (int j = 0; j < pmmlConsequentTimes.getLength(); j++) {
      Element pmmlTime = (Element) pmmlConsequentTimes.item(j);
      TimeManager tm = parseTime(pmmlTime);
      consequentSequenceTimes[j] = tm;
    }

    if (!sequenceReferences.containsKey(antecedentSeqId))
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: document do not contain ref <" +
          antecedentSeqId + "> for sequences"
          , null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
    if (!sequenceReferences.containsKey(consequentSeqId))
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: document do not contain ref <" +
          consequentSeqId + "> for sequences",
          null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    SequenceRuleManager ruleObj = new SequenceRule(
        (SequenceManager) sequenceReferences.get(antecedentSeqId),
        delimiter,
        (SequenceManager) sequenceReferences.get(consequentSeqId),
        confidence.doubleValue(),
        support.doubleValue(),
        occurrence.intValue()
        );

    ruleObj.setTimes(ruleTimes);
    ruleObj.setAntecedentTimes(antecedentSequenceTimes);
    ruleObj.setConsequentTimes(consequentSequenceTimes);

    return ruleObj;
  }

  protected static SequenceManager parseSequence(Element sequence, Hashtable elementReferences) throws Exception{
    //Get sequence attributes...
    String sequenceId = sequence.getAttribute("id");
    Integer numberOfSets = null;
    if( !sequence.getAttribute("numberOfSets").equals("") )
      numberOfSets = new Integer(sequence.getAttribute("numberOfSets"));
    Integer occurrence = null;
    if( !sequence.getAttribute("occurrence").equals("") )
      occurrence = new Integer(sequence.getAttribute("occurrence"));
    Double support = null;
    if( !sequence.getAttribute("support").equals("") ) {
      support = new Double(sequence.getAttribute("support"));
      try{
        UChecker.checkMinSupport(support);
      }catch(SequenceModelException sme){
        throw new SequenceModelException(
        "Illegal PMML Sequence Model: " + sme.getErrorMessage()
        , sme.getDetailErrorMessage(),
        KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
      }

    }

    //Get sequence set references...
    NodeList pmmlSetReferenceList = sequence.getElementsByTagName("SetReference");
    //Get the first SetReference...
    Element pmmlSetReference = (Element)pmmlSetReferenceList.item(0);
    String setId = pmmlSetReference.getAttribute("setId");
    //Check reference...
    if( !elementReferences.containsKey(setId) )
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: document do not contain ref <" +
          setId + "> for set <" + sequenceId + ">.",
          null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    //Check the FOLLOW_SET(s) length...
    NodeList pmmlDelimiterList = sequence.getElementsByTagName("Delimiter");
    if( pmmlDelimiterList.getLength() != (pmmlSetReferenceList.getLength()-1) )
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: document do not contain valid FOLLOW_SET(s)" +
          "for sequence <" + sequenceId + ">.",
          null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    //Instance a new Sequence...
    SequenceManager sequenceObj = new Sequence((SequenceElementManager) elementReferences.get(setId));

    //For each FOLLOW_SET get the pair (Delimiter,SetReference)
    for( int j = 0; j < pmmlDelimiterList.getLength(); j++ ) {

      Element pmmlDelimiter = (Element)pmmlDelimiterList.item(j);
      pmmlSetReference = (Element)pmmlSetReferenceList.item(j+1);

      //Check reference
      setId = pmmlSetReference.getAttribute("setId");
      if( !elementReferences.containsKey(setId) )
        throw new SequenceModelException(
            "Illegal PMML Sequence Model: document do not contain ref <" +
            setId + "> for set <" + sequenceId + ">.",
            null,
            KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

      DelimiterManager d = parseDelimiter(pmmlDelimiter);
      SequenceElementManager s = (SequenceElementManager) elementReferences.get(setId);

      sequenceObj.addElement(d, s);
    }

    if (support != null)
      sequenceObj.setSupport(support);
    if ( occurrence != null )
      sequenceObj.setOccurrence(occurrence);

    return sequenceObj;
  }

  /*protected static SequenceElementManager parseSetPredicate(Element setPredicate) throws Exception{
    //Get SetPredicate attributes...
    String setPredicateId = setPredicate.getAttribute("id");
    String field = setPredicate.getAttribute("field");

    //Get SetPredicate Array...
    NodeList pmmlArrayList = setPredicate.getElementsByTagName("Array");
    Element pmmlArray = (Element) pmmlArrayList.item(0);
    //SetPredicate Array attributes...
    String length = pmmlArray.getAttribute("n");
    String type = pmmlArray.getAttribute("type");

    //SequenceElement Object
    SequenceElementManager setPredicateObj = new ElementSetPredicate(field);

    if( !type.equals("") && !type.toLowerCase().equals("string") )
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: predicate set type mismatch <" +
          type + "> for predicate set <" + setPredicateId + ">.",
          null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    String arrayValues = pmmlArray.getFirstChild().getNodeValue();
    // scorre l'array per creare gli items...
    java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(arrayValues.trim());

    if( !length.equals("") && Integer.parseInt(length) != tokenizer.countTokens() )
      throw new SequenceModelException(
          "Illegal PMML Sequence Model: predicate set length mismatch <" +
          length + "> for predicate set <" + setPredicateId + ">.",
          null,
          KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    while (tokenizer.hasMoreTokens()) {
       String itemStr = tokenizer.nextToken();
       Item item = new Item(itemStr);
       if (setPredicateObj.contain(item))
         throw new SequenceModelException(
             "Illegal PMML Sequence Model: document already contains predicate set value <" +
             item.toString() + "> for predicate set <" + setPredicateId + ">.",
             null,
             KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
       setPredicateObj.addItem(item);
     }

     return setPredicateObj;
  }*/

  protected static SequenceElementManager parseItemSet(Element itemset, Hashtable itemReferences) throws Exception{
    String itemsetId = itemset.getAttribute("id");
    Double support = null;
    if (itemset.getAttribute("support") != "") {
      support = new Double(itemset.getAttribute("support"));
      try {
        UChecker.checkMinSupport(support);
      }
      catch (SequenceModelException sme) {
        throw new SequenceModelException(
            "Illegal PMML Sequence Model: " + sme.getErrorMessage()
            , sme.getDetailErrorMessage(),
            KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);
      }
    }
    // scorre i riferimenti per creare gli items...
    NodeList pmmlItemsList = itemset.getElementsByTagName("ItemRef");

    SequenceElementManager itemsetObj = new ElementItemSet();

    for (int j = 0; j < pmmlItemsList.getLength(); j++) {
      Element pmmlItem = (Element) pmmlItemsList.item(j);
      String itemId = pmmlItem.getAttribute("itemRef");
      if (!itemReferences.containsKey(itemId))
        throw new SequenceModelException(
            "Illegal PMML Sequence Model: document do not contain ref <" +
            itemId + "> for itemset <" + itemsetId + ">.",
            null,
            KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

      Item item = (Item) itemReferences.get(itemId);
      itemsetObj.addItem(item);
    }

    if (support != null)
      ((ElementItemSet)itemsetObj).setSupport(support.doubleValue());

    return itemsetObj;
  }

  protected static ItemManager parseItem(Element item) throws Exception{
    Item itemObj = new Item(item.getAttribute("value"));
    String mappedValue = item.getAttribute("mappedValue");
    if( !mappedValue.trim().equals("") )
      itemObj.setMappedValue( mappedValue );
    if (!item.getAttribute("weight").equals(""))
      itemObj.setWeight( new Double(item.getAttribute("weight")).doubleValue() );

    return itemObj;
  }

  protected static TimeManager parseTime(Element pmmlTime) throws SequenceModelException{
    String minAttr = pmmlTime.getAttribute("min");
    String maxAttr = pmmlTime.getAttribute("max");
    String meanAttr = pmmlTime.getAttribute("mean");

    Double min = null;
    Double max = null;
    Double mean = null;

    try {
      min = new Double(minAttr);
    }
    catch (Exception e) {
      String msg =
          "The min time must be a positive real.";
      throw new SequenceModelException(msg,
                                       null,
                                       KDDMLCoreErrorCodes.
                                       INVALID_PMML_DOCUMENT);
    }
    try {
      max = new Double(maxAttr);
    }
    catch (Exception e) {
      String msg =
          "The max time must be a positive real.";
      throw new SequenceModelException(msg,
                                       null,
                                       KDDMLCoreErrorCodes.
                                       INVALID_PMML_DOCUMENT);
    }
    if( meanAttr.equals("") ) return new Time(min,max);
    try {
      mean = new Double(meanAttr);
    }
    catch (Exception e) {
      String msg =
          "The mean time must be a positive real.";
      throw new SequenceModelException(msg,
                                       null,
                                       KDDMLCoreErrorCodes.
                                       INVALID_PMML_DOCUMENT);
    }

    return new Time(min,max,mean);
  }

  protected static DelimiterManager parseDelimiter(Element pmmlDelimiter) throws SequenceModelException{
    String delimiterAttr = pmmlDelimiter.getAttribute("delimiter");
    String gapAttr = pmmlDelimiter.getAttribute("gap");

    DelimiterType delimiter = null;
    if( delimiterAttr.equals(DelimiterType.SAME_TIME_WINDOW_VALUE))
      delimiter = DelimiterType.SAME_TIME_WINDOW;
    else if( delimiterAttr.equals(DelimiterType.ACROSS_TIME_WINDOWS_VALUE))
      delimiter = DelimiterType.ACROSS_TIME_WINDOWS;
    else throw new SequenceModelException("The delimiter must have one of the allowed value.",
                                       null,
                                       KDDMLCoreErrorCodes.
                                       INVALID_PMML_DOCUMENT);

    GapType gap = null;
    if( gapAttr.equals(GapType.UNKNOWN_VALUE))
      gap = GapType.UNKNOWN;
    else if( gapAttr.equals(GapType.TRUE_VALUE))
      gap = GapType.TRUE;
    else if( gapAttr.equals(GapType.FALSE_VALUE))
      gap = GapType.FALSE;
    else throw new SequenceModelException("The gap must have one of the allowed value.",
                                       null,
                                       KDDMLCoreErrorCodes.
                                       INVALID_PMML_DOCUMENT);

    return new Delimiter(delimiter, gap);
  }

}
