/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining;

/**
 * A symmetric matrix, i.e., the value at (j, k) is the same as the value at (k, j).
 * If nothing is set, getValue returns 1 in the diagonal and 0 in the off-diagonal.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class SymmetricMatrix extends CategoryMatrix implements SymmetricMatrixManager {

  /**
   * Build an empty matrix with default values 0.
   *
   * @param categories String[]
   */
  public SymmetricMatrix(String[] categories) {

    super(categories);

    int l = categories.length;
    matrix = new double[l][l];

    for (int i=0; i<l; i++)
      for (int j=0; j<l; j++)
        if (i==j)
          matrix[i][j] = 1;
        else
          matrix[i][j] = 0;
  }

  /**
   * Sets the value for a given row and a given column.
   *
   * @param row_name String the row name of the matrix.
   * @param column_name String the column name of the matrix.
   * @param value double
   * @throws MiningModelException if the row_name or the column_name has not been found.
   */
  public void setValue(String row_name, String column_name, double value)
      throws MiningModelException {

    int index1 = getIndex(row_name);
    int index2 = getIndex(column_name);

    if (index1 == -1) {
      String msg = "Category < "+row_name+" > do not exist in similarity matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    if (index2 == -1) {
      String msg = "Category < "+column_name+" > do not exist in similarity matrix.";
      throw new MiningModelException(msg, null, kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }

    // the matrix is symmetric
    matrix[index1][index2] = value;
    matrix[index2][index1] = value;
  }
}
