/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Taxonomy;

import org.w3c.dom.Element;
import org.w3c.dom.Document;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Iterator;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.*;
import kddml.Core.*;

/**
 * The values of a categorical field can be organized in a hierarchy.
 * The representation of hierarchies in PMML is based on parent/child relationships.
 * A tabular format is used to provide the data for these relationships.
 * A taxonomy is constructed from a sequence of one or more parent/child tables.
 * In KDDML, the actual values are stored in the hierarchy object. So, the tabular data
 * can also be part of the PMML document itself. The table would be recursive in the sense
 * that a value in the parent column can also appear in the child column.
 * An hierarchy model consists of two major parts:
 * <ol>
 * <li> the hierarhcy name; </li>
 * <li> the pointer to the hierarchy root. </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class HierarchyModel extends kddml.Core.DataMining.MiningModel
    implements HierarchyModelManager {

  protected String hierarchy_name;

  protected HierarchyNode root;

  /**
   * Package proprietary constructor.
   *
   * @param physical_path File
   */
  protected HierarchyModel(java.io.File physical_path) {
    super(physical_path);
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return kddml.Core.KDDMLObjectType.PMML_HIERARCHY;
  }

  /**
   * Returns true if the hierarhcy do not contain nodes.
   *
   * @return boolean
   */
  public boolean isEmpty() {
    return (root == null);
  }

  /**
   * Returns the node root of this hierarchy.
   * Returns null if the hierarchy is empty.
   *
   * @return HierarchyNodeManager
   */
  public HierarchyNodeManager getNodeRoot() {
    return root;
  }

  /**
   * Returns the hierarchy name.
   *
   * @return String
   */
  public String getTaxonomyName() {
    return hierarchy_name;
  }

  /**
   * Save the object in the system repository.
   *
   * @throws AssociationModelException
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws kddml.Core.KDDMLCoreException {

    Document root = initializeDocument();
    Element taxonomy = XMLUtility.doc.createElement("Taxonomy");

    Element data_dictionary = (Element)
        root.getDocumentElement().getElementsByTagName("DataDictionary").item(0);

    Element child = XMLUtility.doc.createElement("DataField");
    Element parent = XMLUtility.doc.createElement("DataField");
    child.setAttribute("name", "child");
    child.setAttribute("optype","ordinal");
    parent.setAttribute("name", "parent");
    parent.setAttribute("optype","ordinal");
    data_dictionary.appendChild(child);
    data_dictionary.appendChild(parent);
    data_dictionary.setAttribute("numberOfFields","2");

    data_dictionary.appendChild(taxonomy);
    XMLUtility.setAttribute(taxonomy, "name", hierarchy_name);
    Element child_parent = XMLUtility.doc.createElement("ChildParent");
    child_parent.setAttribute("childField", "member");
    child_parent.setAttribute("parentField", "group");
    child_parent.setAttribute("isRecursive", "yes");
    XMLUtility.setAttribute(child_parent, "x-RootName", this.root.getName());
    taxonomy.appendChild(child_parent);

    Element inline_table = XMLUtility.doc.createElement("InlineTable");
    child_parent.appendChild(inline_table);

    Element extension = XMLUtility.doc.createElement("Extension");
    if (root != null) {
      Vector v = this.root.toXML();
      for(int i=0; i<v.size(); i++) {
        extension.appendChild((Element) v.get(i));
      }
    }
    inline_table.appendChild(extension);

    System.out.println(XMLUtility.toString(data_dictionary));
    XMLUtility.toFile(object_path.getAbsolutePath(), root);

  }

  /**
   * Clones this model returning a new hierarchy model as output.
   *
   * @param output_path File the path in the model system repository containing the cloned
   * model
   * @return AssociationModel
   */
  public HierarchyModel clone(java.io.File output_path) {
    HierarchyModel output = new HierarchyModel(output_path);
    output.data_dictionary = (DataDictionary) this.data_dictionary.clone();
    output.mining_schema = new MiningSchema();

    output.hierarchy_name = new String(this.hierarchy_name);
    output.root = (HierarchyNode) this.root.clone(null);

    return output;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "Hierarchy Name: "+hierarchy_name+"\n";
    if (root != null)
      s = s + root.toString();
    return s;
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML()
      throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.PMML_HIERARCHY);

    String s1 = getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    try {
      XMLUtility.formatting(s1, s2, s3, null, null);
    }
    catch(Exception e) {
      String msg = "Exception found during the HTML code generation.";
      msg = e.toString();
      throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
    }
  }

  /**
   * Returns the list of leaves as iterator of HierarchyNodeManager.
   *
   * @return Iterator
   */
  public Iterator getLeaves() {
    return root.getLeaves();
  }

  /**
   * Removes an input node from the hierarchy.
   * Returns false if the node do not exist or cannot be removed.
   *
   * @param node_name String
   * @return boolean
   */
  public boolean removeNode(String node_name) {
    HierarchyNodeManager n = getNode(node_name);
    // node not found...
    if (n == null)
      return false;
    // the node is the root...
    if (n.isRoot()) {
      root = null;
      return true;
    }
    // otherwise...
    HierarchyNode parent = (HierarchyNode) n.getParent();
    for (int i=0; i<parent.children.size(); i++) {
      if (((HierarchyNode) parent.children.get(i)).getName().equals(node_name)) {
        parent.children.remove(i);
        return true;
      }
    }
    return false;
  }

  /**
   * Returns the number of nodes of the hierarchy.
   *
   * @return int
   */
  public int getDimension() {
    return root.getDimension();
  }

  /**
   * Returns the number of level of the hierarchy.
   *
   * @return int
   */
  public int getDepth() {
    int depth = 0;
    Iterator iter = getLeaves();
    while(iter.hasNext()) {
      HierarchyNode node = (HierarchyNode) iter.next();
      int level = node.getLevel();
      if (level > depth)
        depth = level;
    }
    return depth;
  }

  /**
   * Returns the hierarchy node given the node name.
   * Returns null if the node do not exist.
   *
   * @param node_name String
   * @return HierarchyNodeManager
   */
  public HierarchyNodeManager getNode(String node_name) {
    return root.getNode(node_name);
  }

  public static void main(String args[]) {
    Hashtable ht = new Hashtable();
    ht.put("San Jose", "CA");
    ht.put("Girloy","CA");
    ht.put("Long Beach","CA");
    ht.put("Oaklown","IL");
    ht.put("Everett","MA");
    ht.put("CA","USA");
    ht.put("IL","USA");
    ht.put("MA","USA");
    ht.put("Everett Nord","Everett");
    ht.put("Everett Sud","Everett");
    HierarchyModel m = null;
    try {
      m = new HierarchyModel(null);
      m.root = (HierarchyNode) HierarchyHashtableFactory.buildHierarchy(ht, "USA");
      m.hierarchy_name = "cities";
      System.out.println(m.getDimension());
      HierarchyModel cloned = (HierarchyModel) m.clone(null);
      cloned.removeNode("MA");
      System.out.println("Original:");
      System.out.println(m.toString());
      System.out.println("Cloned:");
      System.out.println(cloned.toString());
      System.out.println("Original Dimension: "+m.getDimension());
      System.out.println("Pruned Dimension: "+cloned.getDimension());
    }
    catch (HierarchyModelException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
  }
}
