/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Taxonomy;

import java.util.Vector;
import java.util.Iterator;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * This class represents a single node hierarchy.
 * It is composed by three major parts:
 * <ol>
 * <li> the node name; </li>
 * <li> the pointer to the children list. </li>
 * <li> the pointer to the parent node. </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class HierarchyNode implements HierarchyNodeManager {

  /**
   * The node name.
   */
  protected String node_name;

  /**
   * The list of children.
   */
  protected Vector children;

  /**
   * The pointer to the father node.
   */
  protected HierarchyNode parent;

  /**
   * Proprietary constructor.
   *
   * @param node_name String
   * @param father HierarchyNodeManager
   */
  protected HierarchyNode(String node_name, HierarchyNodeManager father) {
    this.node_name = node_name;
    children = new Vector();
    this.parent = (HierarchyNode) father;
  }

  /**
   * Adds a new child to this hierarchy node.
   *
   * @param node HierarchyNodeManager
   */
  public void addChild(HierarchyNodeManager node) {
    children.add(node);
  }

  /**
   * Returns the level of this node where level(root)=0.
   *
   * @return int
   */
  public int getLevel() {
    if (parent == null)
      return 0;
    else
      return parent.getLevel() + 1;
  }

  /**
   * Returns the dimension of the sub-nodes of this node.
   *
   * @return int
   */
  public int getDimension() {
    int ris = 1;
    for (int i=0; i<children.size(); i++) {
      HierarchyNodeManager n = (HierarchyNodeManager) children.get(i);
      ris = ris + n.getDimension();
    }
    return ris;
  }

  /**
   * Returns the parent node.
   *
   * @return HierarchyNodeManager
   */
  public HierarchyNodeManager getParent() {
    return parent;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "";
    for (int j=0; j< getLevel(); j++) {
      s = s + " | ";
    }
    s = s+node_name+" (level "+getLevel()+")"+"\n";
    for (int i=0 ; i<children.size(); i++) {
      HierarchyNode n = (HierarchyNode) children.get(i);
      s = s + n.toString();
    }
    return s;
  }

  public String getName() {
    return node_name;
  }

  /**
   * Returns boolean indicating whether this node is a leaf (terminal node).
   *
   * @return boolean
   */
  public boolean isLeaf() {
    return (children.size() == 0);
  }

  /**
   * Returns boolean indicating whether this node is the node root of the hierarchy.
   *
   * @return boolean
   */
  public boolean isRoot() {
    return (parent == null);
  }

  public int getNumberOfChildren() {
    return children.size();
  }

  /**
   * Returns the direct children of this node or null if the node is a leaf node.
   *
   * @return HierarchyNodeManager[]
   */
  public HierarchyNodeManager[] getChildren() {
    HierarchyNodeManager ris[] = new HierarchyNodeManager[children.size()];
    for (int i=0; i<children.size(); i++) {
      ris[i] = (HierarchyNodeManager) children.get(i);
    }
    return ris;
  }

  /**
   * Return all the leaves of the current node as iterator.
   *
   * @return Iterator
   */
  public Iterator getLeaves() {
    Vector v = new Vector();
    if (this.isLeaf())
      v.add(this);
    else {
      for (int i = 0; i < children.size(); i++) {
        HierarchyNodeManager n = (HierarchyNodeManager) children.get(i);
        if (n.isLeaf())
          // the current child is a leaf
          v.add(n);
        else {
          // the current child is not a leaf: recursion
          Iterator iter = n.getLeaves();
          // add to the vector the elements of the iterator
          while (iter.hasNext()) {
            HierarchyNodeManager n2 = (HierarchyNodeManager) iter.next();
            v.add(n2);
          }
        }
      }
    }
    return v.iterator();
  }

  protected Vector toXML() {
    Element ris;
    // the vector of elements
    Vector v = new Vector();
    if (this.isLeaf()) {
      // the node is a leaf
      ris = XMLUtility.doc.createElement("row");
      XMLUtility.setAttribute(ris, "member", node_name);
      XMLUtility.setAttribute(ris, "group", this.parent.getName());
      v.add(ris);
    }
    else {
      if (!isRoot()) {
        // the node is an internal node
        ris = XMLUtility.doc.createElement("row");
        XMLUtility.setAttribute(ris, "member", node_name);
        XMLUtility.setAttribute(ris, "group", this.parent.getName());
        v.add(ris);
      }

      for (int i = 0; i < children.size(); i++) {
        Vector v2 = ( (HierarchyNode) children.get(i)).toXML();
        for (int j = 0; j < v2.size(); j++) {
          v.add((Element) v2.get(j));
        }
      }
    }
    return v;
  }

  protected HierarchyNodeManager getNode(String node_name) {
    if (this.node_name.equals(node_name))
      return this;
    else {
      if (this.isLeaf())
        return null;
      for (int i = 0; i < children.size(); i++) {
        HierarchyNodeManager n = ((HierarchyNode) children.get(i)).getNode(node_name);
        if (n != null)
          return n;
      }
      return null;
    }
  }

  protected HierarchyNode clone(HierarchyNode father) {
    HierarchyNode ris;
    ris = new HierarchyNode(this.getName(), father);

    for (int i=0; i<children.size(); i++) {
      // cloning dei figli
      HierarchyNode n = ((HierarchyNode) children.get(i)).clone(ris);
      ris.addChild(n);
    }
    return ris;
  }
}
