/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataMining.Taxonomy;

import kddml.Core.DataMining.*;
import kddml.Core.*;
import java.io.*;
import org.w3c.dom.*;
import java.util.Hashtable;
import java.util.Vector;

/**
 * A factory class that creates an hierarchy model from the system repository.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class HierarchyRepositoryFactory extends ModelRepositoryFactory {

  private String pmml_version;

  private Element header;

  private Element data_dictionary;

  private Element taxonomy;

  private String root_name;

  private HierarchyNodeManager root;

  /**
   * Constructor given the physical file containing the object.
   *
   * @param f File the file containing the reference to the physical object.
   * @throws KDDMLCoreException an exception if the file cannot be found.
   */
  public HierarchyRepositoryFactory(File f)
      throws KDDMLCoreException {

    super(f);
    // effettua il parsing del documento XML contenente il modello...
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(f.toString()),
                              new DTDResolver(kddml.Core.KDDMLObjectType.PMML_HIERARCHY));
    }
    catch(java.io.IOException e) {
      String msg = "Invalid PMML reference: "+f.toString();
      throw new HierarchyModelException(msg, e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }

    Element root = (Element) doc.getDocumentElement();

    if (!root.getTagName().equalsIgnoreCase("pmml"))
      throw new HierarchyModelException(f.toString()+ " no contains a PMML document.",
                                        null,
                                        KDDMLCoreErrorCodes.INVALID_PMML_DOCUMENT);

    // reading the pmml version
    pmml_version = root.getAttribute("version");

    // reading the copyright
    header = (Element) root.getElementsByTagName("Header").item(0);

    // reading the data dictionary element
    data_dictionary = (Element) root.getElementsByTagName("DataDictionary").item(0);

    // reading the element taxonomy
    taxonomy = (Element) data_dictionary.getElementsByTagName("Taxonomy").item(0);

    Element el = (Element) taxonomy.getElementsByTagName("ChildParent").item(0);
    root_name = XMLUtility.getAttribute(el, "x-RootName");

  }

  public KDDMLObject newInstance() throws KDDMLCoreException {

    HierarchyModel ris = new HierarchyModel(physical_file);
    ris.hierarchy_name = buildHierarchyName();
    Hashtable ht = buildHashtable();
    ris.root = HierarchyHashtableFactory.buildHierarchy(ht, root_name);

    return ris;
  }

  private Hashtable buildHashtable() throws MiningModelException {
    Element el = (Element) taxonomy.getElementsByTagName("InlineTable").item(0);
    if (el == null)
      return null;
    Element extension = (Element) el.getElementsByTagName("Extension").item(0);

    Hashtable ris = new Hashtable();

    // scorre gli elementi xml per costruire le istanze
    Vector v = XMLUtility.getChildren(extension);

    for (int i=0; i<v.size(); i++) {

      Element row = (Element) v.get(i);
      ris.put(XMLUtility.getAttribute(row, "member"), XMLUtility.getAttribute(row, "group"));
    }
    return ris;
  }

  private String buildHierarchyName() throws MiningModelException {
    return XMLUtility.getAttribute(taxonomy, "name");
  }

  /**
   * Only for testing
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    // effettua il parsing del documento XML contenente il modello...
    Document doc;

    try {
      File f = new File(kddml.Core.KDDMLObjectType.PMML_HIERARCHY.getRepositoryPath()+"cities_hierarchy.xml");
      HierarchyRepositoryFactory fact = new HierarchyRepositoryFactory(f);

      HierarchyModel m = (HierarchyModel) fact.newInstance();
      System.out.println(m.toString());
    }
    catch(KDDMLCoreException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }
}
