/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import weka.core.*;
import java.io.StreamTokenizer;
import java.io.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A factory that creates a new internal table from an ARFF source.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class ARFFFactory extends InternalTableExternalFactory {

  private final static String FILE_EXTENSION = ".arff";
  private final static String ARFF_RELATION = "@relation";
  private final static String ARFF_DATA = "@data";
  private final static String ARFF_ATTRIBUTE = "@attribute";
  private final static String ARFF_ATTRIBUTE_INTEGER = "integer";
  private final static String ARFF_ATTRIBUTE_REAL = "real";
  private final static String ARFF_ATTRIBUTE_NUMERIC = "numeric";
  private final static String ARFF_ATTRIBUTE_STRING = "string";
  private final static String ARFF_ATTRIBUTE_DATE = "date";

  private StreamTokenizer streamtokenizer;
  private String relation_name;
  private FastVector m_Attributes;

  private BufferedReader reader;

  // pointer to the instances
  private int pointer;


  /**
   * Constructor given the physical file containing the object and the string
   * containing the ARFF path.
   *
   * @param f File
   * @param arff_path String the string containing the ARFF path. ARFF files can be
   * gatered via URI reference.
   * @throws KDDMLCoreException if an I/O error occours.
   */
  public ARFFFactory(java.io.File f, String arff_path)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    try {
      reader = kddml.Core.IOUtil.getReader(arff_path);
      initializeTokenizer();
      // legge lo schema del file ARFF ed inizializza le istanze
      initializeHeader();
    }
    catch(IOException e) {
      throw new InternalTableException("I/O Error", e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }
    // pointer viene posizionato sulla prima riga della parte dati
    pointer = 0;
  }

  /**
   * Returns the next instance from the input resource.
   * Returns null if no more instances are available.
   *
   * @param schema Instances the next instance from the input resource.
   * @throws InternalTableException if an error occours.
   * @return Instance
   */
  protected Instance nextInstance(Instances schema) throws InternalTableException {
    Instance ris = null;
    try {
      getFirstToken();
      // end of file
      if(streamtokenizer.ttype == -1)
      {
        return null;
      }
      ris = getInstanceFull(true, schema);
      pointer++;
    }
    catch (IOException ex) {
      throw new InternalTableException("I/O Error", ex.toString(),
                                       KDDMLCoreErrorCodes.IO_ERROR);
    }
    return ris;
  }

  /**
   * Returns the instances header as fast vector of attributes.
   *
   * @throws InternalTableException
   * @return FastVector
   */
  protected FastVector getHeader() throws InternalTableException {
    if(m_Attributes.size() == 0) {
      throw new InternalTableException("No Header information avaible!", null,
                                       KDDMLCoreErrorCodes.IO_ERROR);
    }
    return m_Attributes;
  }

  /**
   * Returns the relation name of this table.
   *
   * @return String
   */
  protected String getRelationName() {
    return relation_name;
  }

  private void initializeTokenizer() {
    streamtokenizer = new StreamTokenizer(reader);
    streamtokenizer.resetSyntax();
    streamtokenizer.whitespaceChars(0, 32);
    streamtokenizer.wordChars(33, 255);
    streamtokenizer.whitespaceChars(44, 44);
    streamtokenizer.commentChar(37);
    streamtokenizer.quoteChar(34);
    streamtokenizer.quoteChar(39);
    streamtokenizer.ordinaryChar(123);
    streamtokenizer.ordinaryChar(125);
    streamtokenizer.eolIsSignificant(true);
  }

  private void initializeHeader() throws IOException {
    getFirstToken();
    m_Attributes = new FastVector();
    if (streamtokenizer.ttype == -1) {
      errms("premature end of file");
    }
    if (ARFF_RELATION.equalsIgnoreCase(streamtokenizer.sval)) {
      getNextToken();
      relation_name = streamtokenizer.sval;
      getLastToken(false);
    }
    else {
      errms("keyword " + ARFF_RELATION + " expected");
    }

    getFirstToken();
    if (streamtokenizer.ttype == -1) {
      errms("premature end of file");
    }
    do {
      if (!ARFF_ATTRIBUTE.equalsIgnoreCase(streamtokenizer.sval)) {
        break;
      }
      getNextToken();
      String s = streamtokenizer.sval;
      getNextToken();
      if (streamtokenizer.ttype == -3) {
        if (streamtokenizer.sval.equalsIgnoreCase(ARFF_ATTRIBUTE_REAL) ||
            streamtokenizer.sval.equalsIgnoreCase(ARFF_ATTRIBUTE_INTEGER) ||
            streamtokenizer.sval.equalsIgnoreCase(ARFF_ATTRIBUTE_NUMERIC)) {
          m_Attributes.addElement(new Attribute(s));
          readTillEOL(streamtokenizer);
        }
        else
        if (streamtokenizer.sval.equalsIgnoreCase(ARFF_ATTRIBUTE_STRING)) {
          m_Attributes.addElement(new Attribute(s, (FastVector) null));
          readTillEOL(streamtokenizer);
        }
        else
        if (streamtokenizer.sval.equalsIgnoreCase(ARFF_ATTRIBUTE_DATE)) {
          String s1 = null;
          if (streamtokenizer.nextToken() != 10) {
            if (streamtokenizer.ttype != -3 && streamtokenizer.ttype != 39 &&
                streamtokenizer.ttype != 34) {
              errms("not a valid date format");
            }
            s1 = streamtokenizer.sval;
            readTillEOL(streamtokenizer);
          }
          else {
            streamtokenizer.pushBack();
          }
          m_Attributes.addElement(new Attribute(s, s1));
        }
        else {
          errms("no valid attribute type or invalid enumeration");
        }
      }
      else {
        FastVector fastvector = new FastVector();
        streamtokenizer.pushBack();
        if (streamtokenizer.nextToken() != 123) {
          errms("{ expected at beginning of enumeration");
        }
        while (streamtokenizer.nextToken() != 125) {
          if (streamtokenizer.ttype == 10) {
            errms("} expected at end of enumeration");
          }
          else {
            fastvector.addElement(streamtokenizer.sval);
          }
        }
        if (fastvector.size() == 0) {
          errms("no nominal values found");
        }
        m_Attributes.addElement(new Attribute(s, fastvector));
      }
      getLastToken(false);
      getFirstToken();
      if (streamtokenizer.ttype == -1) {
        errms("premature end of file");
      }
    }
    while (true);
    if (!ARFF_DATA.equalsIgnoreCase(streamtokenizer.sval)) {
      errms("keyword " + ARFF_DATA + " expected");
    }
    if (m_Attributes.size() == 0) {
      errms("no attributes declared");
    }
    //m_ValueBuffer = new double[m_Attributes.size()];
    //m_IndicesBuffer = new int[m_Attributes.size()];
  }

  private void readTillEOL(StreamTokenizer streamtokenizer)
      throws IOException
  {
    while(streamtokenizer.nextToken() != 10) ;
    streamtokenizer.pushBack();
  }


  private Instance getInstanceFull(boolean flag, Instances schema)
      throws IOException
  {
    Instance inst = new Instance(m_Attributes.size());
    // ------------------------------
    // Richiesto da weka...
    // ------------------------------
    inst.setDataset(schema);
    inst.setWeight(1.0D);
    for(int j = 0; j < m_Attributes.size(); j++)
    {
      if(j > 0)
      {
        getNextToken();
      }
      if(streamtokenizer.ttype == 63)
      {
        inst.setMissing(j);
        continue;
      }
      if(streamtokenizer.ttype != -3)
      {
        errms("not a valid value");
      }
      switch(attribute(j).type())
      {
        case 1: // '\001'
          int i = attribute(j).indexOfValue(streamtokenizer.sval);
          if(i == -1)
          {
            errms("nominal value not declared in header");
          }
          inst.setValue(j, streamtokenizer.sval);
          break;

        case 0: // '\0'
          try
          {
            inst.setValue(j, Double.valueOf(streamtokenizer.sval).doubleValue());
          }
          catch(NumberFormatException numberformatexception)
          {
            errms("number expected");
          }
          break;

        case 2: // '\002'
          inst.setValue(j, streamtokenizer.sval);
          break;

        case 3: // '\003'
          try
          {
            attribute(j).parseDate(streamtokenizer.sval);
            inst.setValue(j, streamtokenizer.sval);
          }
          catch(Exception parseexception)
          {
            errms("unparseable date: " + streamtokenizer.sval);
          }
          break;

        default:
          errms("unknown attribute type in column " + j);
          break;
      }
    }

    if(flag)
    {
      getLastToken(true);
    }
    return inst;
  }

  private void errms(String s)
      throws IOException
  {
    throw new IOException(s + ", read " + streamtokenizer.toString());
  }

  private void getFirstToken()
      throws IOException
  {
    while(streamtokenizer.nextToken() == 10) ;
    if(streamtokenizer.ttype == 39 || streamtokenizer.ttype == 34)
    {
      streamtokenizer.ttype = -3;
    } else
    if(streamtokenizer.ttype == -3 && streamtokenizer.sval.equals("?"))
    {
      streamtokenizer.ttype = 63;
    }
  }

  private void getNextToken()
      throws IOException
  {
    if(streamtokenizer.nextToken() == 10)
    {
      errms("premature end of line");
    }
    if(streamtokenizer.ttype == -1)
    {
      errms("premature end of file");
    } else
    if(streamtokenizer.ttype == 39 || streamtokenizer.ttype == 34)
    {
      streamtokenizer.ttype = -3;
    } else
    if(streamtokenizer.ttype == -3 && streamtokenizer.sval.equals("?"))
    {
      streamtokenizer.ttype = 63;
    }
  }

  private void getLastToken(boolean flag)
      throws IOException
  {
    if(streamtokenizer.nextToken() != 10 && (streamtokenizer.ttype != -1 || !flag))
    {
      errms("end of line expected");
    }
  }

  private final Attribute attribute(int i)
  {
    return (Attribute)m_Attributes.elementAt(i);
  }

  /**
   * This is only for testing....
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    try {
      File f = new File("C:/kddml/Repository/DataSources/KDDMLTable/weather.xml");
      String s = "C:/kddml/Repository/DataSources/ARFF/weather.arff";
      ARFFFactory fact = new ARFFFactory(f, s);

      // settaggio delle proprieta' del factory...
      fact.setLogicalName("weather");
      fact.setOverwrite(true);
      fact.setMaxNumberOfLoadedInstAtIteration(250);
      // crea l'istanza di InternalTable...
      InternalTable t = (InternalTable) fact.newInstance();
      // salvataggio della tabella nel repository...
      t.saveToRepository();
      // import di un insieme di istanze della tabella...
      Instances inst = (Instances) t.load(5);
      System.out.println(inst.toString());
      // work on instances here...
    }
    catch(kddml.Core.KDDMLCoreException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }
}
