/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataSources;

import weka.core.Instances;
import org.w3c.dom.*;
import kddml.Core.DataMining.Taxonomy.*;
import kddml.Core.XMLUtility;
import kddml.Core.DataMining.DataDictionary;
import kddml.Core.DataMining.DataField;
import java.util.Vector;
import java.util.Hashtable;


/**
 * This class contains statistical information about a single data attribute.
 * Statistic elements have a field name and a field type (for example, discrete or
 * continuous). The statistics depend on the attribute type and can include the most
 * frequent values or the mean of values. <br>
 * An hierarchy object can be related to a non-numeric attributes.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class AttributeStatistic implements AttributeStatisticManager {

  protected String name;

  protected AttributeType type;

  protected int missed;

  protected HierarchyModel taxonomy;

  /**
   * The total number of instances of the attribute (other than missed values)
   */
  protected int num_instances;

  /**
   * Proprietary constructor given the attribute name.
   *
   * @param name String
   */
  protected AttributeStatistic(String name) {
    this.name = name;
    missed = 0;
    num_instances = 0;
    taxonomy = null;
  }

  /**
   * Proprietary constructor give the XML element and the total number of instances.
   *
   * @param elem Element
   * @param total_inst int
   * @throws KDDMLCoreException
   */
  protected AttributeStatistic(Element elem, int total_inst)
      throws kddml.Core.KDDMLCoreException {

    // ottiene il numero di valori mancanti per l'attributo
    missed = new Integer(elem.getAttribute(SpaceNames.ATTRIBUTE_MISSED)).intValue();
    name = kddml.Core.XMLUtility.getAttribute(elem, SpaceNames.ATTRIBUTE_NAME);
    num_instances = total_inst - missed;

    NodeList list = elem.getElementsByTagName("TAXONOMY");
    Element tax = (Element) list.item(0);
    if (tax != null) {
      // taxonomy element
      Element el = (Element) tax.getElementsByTagName("INLINE_TABLE").item(0);
      Element el2 = (Element) tax.getElementsByTagName("CHILD_PARENT").item(0);
      String root_name = el2.getAttribute("root_name");
      String child_field = el2.getAttribute("child_field");
      String parent_field = el2.getAttribute("parent_field");
      DataDictionary dd = buildDataDictionary(child_field, parent_field);
      // tabella hash contenente le associazioni member --> group
      Hashtable ris = new Hashtable();
      // scorre gli elementi xml per costruire le istanze
      Vector v = XMLUtility.getChildren(el);

      for (int i=0; i<v.size(); i++) {

        Element row = (Element) v.get(i);
        ris.put(XMLUtility.getAttribute(row, "member"), XMLUtility.getAttribute(row, "group"));
      }
      String name = tax.getAttribute("name");
      java.io.File f = new java.io.File(kddml.Core.KDDMLObjectType.PMML_HIERARCHY.getRepositoryPath()+name+".xml");
      HierarchyHashtableFactory fact = new HierarchyHashtableFactory(f, name, dd, root_name, ris);
      fact.setOverwrite(false);

      taxonomy = (HierarchyModel) fact.newInstance();
    }
    else
      // no taxonomy
      this.taxonomy = null;
  }

  /**
   * Returns the attribute name.
   *
   * @return String
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the attribute type.
   *
   * @return AttributeType
   */
  public AttributeType getType() {
    return type;
  }

  /**
   * Returns the number of missing instances for that attribute.
   *
   * @return int
   */
  public int getMissingCount() {
    return missed;
  }

  /**
   * Returns the number of missing instances for that attribute as percentage
   * number.
   *
   * @return double a value between 0 and 1.
   */
  public double getMissingCountPerc() {
    return (double) missed / ((double) missed + num_instances);
  }

  protected void updateStatistic(Instances instances) {
    int pos = instances.attribute(name).index();
    for (int i=0; i<instances.numInstances(); i++)
      if (instances.instance(i).isMissing(pos))
        missed++;
      else
        num_instances++;
  }

  /**
   * Get a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "Attribute < "+name+" >\n";
    s = s + "Number of missed instances: "+missed+"\n";
    if (taxonomy != null)
      s = s + "Taxonomy: \n" + taxonomy.toString() + "\n";
    return s;
  }

  /**
   * Returns the taxonomy related to this attribute.
   * Return null if the attribute hasn't a taxonomy.
   *
   * @return HierarchyModelManager
   */
  public HierarchyModelManager getTaxonomy() {
    return taxonomy;
  }

  /**
   * Sets the taxonomy to this attribute.
   *
   * @param taxonomy HierarchyModelManager
   */
  public void setTaxonomy(HierarchyModelManager taxonomy) {
    this.taxonomy = (HierarchyModel) taxonomy;
  }

  protected Element toXML() {
    Element attributo = kddml.Core.XMLUtility.doc.createElement(SpaceNames.ELEMENT_ATTRIBUTE);
    kddml.Core.XMLUtility.setAttribute(attributo, SpaceNames.ATTRIBUTE_NAME, name);

    // ottiene il numero di valori mancanti
    String s = new Integer(getMissingCount()).toString();
    String s1;
    if (num_instances != 0)
      s1 = java.text.NumberFormat.getPercentInstance().format(getMissingCountPerc()).toString();
    else
      s1 = "0%";
    kddml.Core.XMLUtility.setAttribute(attributo, SpaceNames.ATTRIBUTE_MISSED,s);
    kddml.Core.XMLUtility.setAttribute(attributo, SpaceNames.ATTRIBUTE_MISSED_PERC,s1);
    return attributo;
  }

  /**
   * Build the XML for the hierarchy element.
   *
   * @return Element
   */
  protected Element getTaxonomyElement() {

    Element taxonomy = XMLUtility.doc.createElement("TAXONOMY");
    taxonomy.setAttribute("name", this.taxonomy.getTaxonomyName());

    Element child_parent = XMLUtility.doc.createElement("CHILD_PARENT");
    child_parent.setAttribute("child_field", "member");
    child_parent.setAttribute("parent_field", "group");
    child_parent.setAttribute("is_recursive", "yes");
    XMLUtility.setAttribute(child_parent, "root_name", this.taxonomy.getNodeRoot().getName());
    taxonomy.appendChild(child_parent);

    Element inline_table = XMLUtility.doc.createElement("INLINE_TABLE");
    child_parent.appendChild(inline_table);

    if (this.taxonomy.getNodeRoot() != null) {
      Vector v = toXML(this.taxonomy.getNodeRoot());
      for(int i=0; i<v.size(); i++) {
        inline_table.appendChild((Element) v.get(i));
      }
    }
    return taxonomy;
  }

  /**
   * Build the XML for a hierarchy node.
   *
   * @param node HierarchyNodeManager
   * @return Vector
   */
  private Vector toXML(HierarchyNodeManager node) {
    Element ris;
    // the vector of elements
    Vector v = new Vector();
    if (node.isLeaf()) {
      // the node is a leaf
      ris = XMLUtility.doc.createElement("ROW");
      XMLUtility.setAttribute(ris, "member", node.getName());
      XMLUtility.setAttribute(ris, "group", node.getParent().getName());
      v.add(ris);
    }
    else {
      if (!node.isRoot()) {
        // the node is an internal node
        ris = XMLUtility.doc.createElement("ROW");
        XMLUtility.setAttribute(ris, "member", node.getName());
        XMLUtility.setAttribute(ris, "group", node.getParent().getName());
        v.add(ris);
      }

      // RECURSION...
      for (int i = 0; i < node.getChildren().length; i++) {
        Vector v2 = toXML(node.getChildren()[i]);
        for (int j = 0; j < v2.size(); j++) {
          v.add((Element) v2.get(j));
        }
      }
    }
    return v;
  }

  private DataDictionary buildDataDictionary(String attr1, String attr2) {
    DataField f1 = new DataField(attr1);
    f1.setType(DataField.ORDINAL);
    DataField f2 = new DataField(attr2);
    f2.setType(DataField.ORDINAL);

    DataDictionary dd = new DataDictionary();
    dd.addField(f1);
    dd.addField(f2);
    return dd;
  }
}
