/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import java.io.*;
import java.sql.*;
import weka.core.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A factory class that creates a table from a database.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class DataBaseFactory extends InternalTableExternalFactory {

  private final static String[] NUMERIC_SQL_TYPE = {
      "int", "integer", "smallint", "binary", "tinyint",
      "bigint", "numeric", "real", "float", "decimal", "REAL"};

  private ResultSet result_set = null;

  private ResultSetMetaData meta_data = null;

  private int num_colonne = 0;

  private String sql_query = null;

  private boolean empty_table_admitted = false;

  private JDBCConnection connection;

  private boolean empty_result;

  private boolean exit = false;

  /**
   * Constructor given the physical file containing the object, the sql SELECT query
   * as string and the DBMS connection parameters.
   *
   * @param f File the file containing the reference to the physical object.
   * @param sql_query String the SQL SELECT query
   * @param connection JDBCConnection the parameters of the connection.
   * @throws KDDMLCoreException if an error occours.
   */
  public DataBaseFactory(java.io.File f, String sql_query, JDBCConnection connection)
      throws kddml.Core.KDDMLCoreException {


    // TO BE COMPLETED
    // effettuare controllo su password di connessione
    // if (pwd == "*") then l'accesso e' umano e si utilizzano i parametri di connession
    // altrimenti l'accesso e' effettuato da una macchiana e si usa direttamente la stringa fornita
    // con la query per accedere al database

    super(f);
    this.sql_query = sql_query;
    this.connection = connection;

    // istanzia una connessione ad un DBMS
    if (connection == null)
      throw new InternalTableException(
        "No JDBC connection specified.", null, KDDMLCoreErrorCodes.INVALID_SETTINGS);

    // esegue la query SQL
    if (sql_query == null)
      throw new InternalTableException(
        "No Query SQL specified.", null, KDDMLCoreErrorCodes.INVALID_SETTINGS);

    empty_result = !executeSQL(connection, sql_query);
  }

  /**
   * Constructor given the physical file containing the object, the sql SELECT query
   * as file and the DBMS connection parameters.
   *
   * @param f File the file containing the reference to the physical object.
   * @param sql_query_path File the SQL SELECT query as file
   * @param connection JDBCConnection the parameters of the connection.
   * @throws KDDMLCoreException if an error occours.
   */
  public DataBaseFactory(java.io.File f, File sql_query_path, JDBCConnection connection)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    this.connection = connection;
    BufferedReader reader =  null;
    try {
      reader = kddml.Core.IOUtil.getReader(sql_query_path.getPath());
      CharArrayWriter buffer = new CharArrayWriter();
      int c;
      while ((c = reader.read()) != -1) {
        buffer.write((char) c);
      }
      this.sql_query = buffer.toString();
    }
    catch (Exception e) {
      String msg = "I/O error reading the file: "+sql_query_path.getPath();
      throw new InternalTableException(msg, null, KDDMLCoreErrorCodes.IO_ERROR);
    }
    // istanzia una connessione ad un DBMS
    if (connection == null)
      throw new InternalTableException(
        "No JDBC connection specified.", null, KDDMLCoreErrorCodes.INVALID_SETTINGS);

    // esegue la query SQL
    if (sql_query == null)
      throw new InternalTableException(
        "No Query SQL specified.", null, KDDMLCoreErrorCodes.INVALID_SETTINGS);

    empty_result = !executeSQL(connection, sql_query);
  }

  /**
   * Validate empty results. Empty tables can be obtained from not-SELECT queries,
   * for example using UPDATE SQL statements.
   *
   * @param empty_table_admitted boolean if false, no empty results are admitted.
   */
  public void setEmptyResultAdmitted(boolean empty_table_admitted) {
    this.empty_table_admitted = empty_table_admitted;
  }

  /**
   * Returns the next instance from the input resource.
   * Returns null if no more instances are available.
   *
   * @param schema Instances the data schema of the instances.
   * @throws InternalTableException if an error occours.
   * @return Instance
   */
  protected Instance nextInstance(Instances schema)
      throws InternalTableException {

    if (empty_result & !empty_table_admitted)
      throw new InternalTableException("No output table for query: "+sql_query, null,
                                       KDDMLCoreErrorCodes.EMPTY_OBJECT);

    if (empty_result) {
      // viene restituito in output una tabella con una sola istanza contenente
      // le informazioni sulla query SQL eseguita...
      exit = true;
      empty_result = false;
      return buildEmptyInstance(schema);
    }

    if (exit)
      return null;

    try {
      if (!result_set.next()) {
        return null;
      }
    }
    catch (SQLException ex) {
      String msg1 = "JDBC Connection Failure.";
      String msg2 = "JDBC Connection Failure: "+ex.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }

    // ciclo sulle colonne...
    Instance tupla = new Instance(num_colonne);
    for (int i = 1; i <= num_colonne; i++) {
      // ottiene il valore della cella
      String valore = null;
      try {
        valore = result_set.getString(i);
      }
      catch (SQLException ex1) {
        String msg1 = "JDBC Connection Failure.";
        String msg2 = "JDBC Connection Failure: "+ex1.toString();
        throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
      }
      Attribute a = (Attribute) schema.attribute(i-1);
      if ((valore == null) || (valore.equals(""))) {
        tupla.setMissing(a);
      }
      else {
        valore = valore.trim();
        if (a.isNumeric()) {
          // attributo di tipo numerico
          Double d = new Double(valore);
          // WEKA ARROTONDA UN DOUBLE A FLOAT
          tupla.setValue(a, d.doubleValue());
        }
        else {
          // attributo di tipo enumerato o stringa
          tupla.setValue(a, valore);
        }
      }
    }
    return tupla;
  }

  /**
   * Returns the instances header as fast vector of attributes.
   *
   * @throws InternalTableException
   * @return FastVector
   */
  protected FastVector getHeader() throws InternalTableException {

    if (empty_result) {
      // viene restituito in output una tabella con una sola istanza contenente
      // le informazioni sulla query SQL eseguita...
      FastVector fv = new FastVector();
      fv = null;
      Attribute attr2 = new Attribute("User", fv);
      Attribute attr3 = new Attribute("Database", fv);
      Attribute attr4 = new Attribute("Date", fv);
      FastVector attributes = new FastVector();
      attributes.addElement(attr2);
      attributes.addElement(attr3);
      attributes.addElement(attr4);
      return attributes;
    }

    try {
      num_colonne = meta_data.getColumnCount();

      // crea il vettore contenente gli attributi della tabella
      FastVector lista_attributi = new FastVector();

      for (int i = 1; i <= num_colonne; i++) {
        String tipo = meta_data.getColumnTypeName(i);
        String name = meta_data.getColumnName(i);
        boolean numeric = false;
        for (int j=0; j<NUMERIC_SQL_TYPE.length; j++) {
          if (tipo.equals(NUMERIC_SQL_TYPE[j])) {
            numeric = true;
            j = NUMERIC_SQL_TYPE.length;
          }
        }

        if (numeric) {

          // ATTRIBUTO DI TIPO NUMERICO
          Attribute attr = new Attribute(meta_data.getColumnName(i));
          lista_attributi.addElement(attr);
        }
        else {
          // ATTRIBUTO DI TIPO STRINGA
          FastVector fv = new FastVector(0);
          fv = null;
          Attribute attr = new Attribute(name, fv);
          lista_attributi.addElement(attr);
        }
      }
      return lista_attributi;
    }
    catch(Exception e) {
      String msg1 = "Error reading database.";
      String msg2 = e.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
  }


  /**
   * Returns the relation name of this table.
   *
   * @return String
   */
  protected String getRelationName() {
    return "NO_RELATION_NAME";
  }

  private boolean executeSQL(JDBCConnection connection, String sql_query)
      throws InternalTableException {
    Statement statement = connection.getConnection();

    try {
      if (statement.execute(sql_query)) {
        // sono stati prodotti dei risultati
        result_set = statement.getResultSet();
        meta_data = result_set.getMetaData();
        return true;
      }
      else
        return false;
    }
    catch(SQLException sql_exc) {
      String msg1 = "JDBC Connection Failure.";
      String msg2 = "JDBC Connection Failure: "+sql_exc.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
  }

  private Instance buildEmptyInstance(Instances schema) {
    Instance instance = new Instance(3);
    instance.setDataset(schema);
    instance.setValue(0, connection.user);
    instance.setValue(1, connection.database_url);
    java.util.GregorianCalendar c = new java.util.GregorianCalendar();
    instance.setValue(2, new java.util.Date().toString());
    return instance;
  }

  /**
   * Only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    // connection istantiation...
    try {
      File f = new File("D:/anagrafica.xml");
      String sql = "select * from anagrafica";
      JDBCConnection conn = new JDBCConnection("romei", "xxxx", "jdbc:jtds:sqlserver://localhost/ubs");

      DataBaseFactory fact = new DataBaseFactory(f, sql, conn);
      // settings
      // set overwrite true if the table exist
      fact.setOverwrite(true);
      // admitted empty results
      fact.setEmptyResultAdmitted(true);
      // set the iteration lenght
      fact.setMaxNumberOfLoadedInstAtIteration(10000);
      InternalTable t = (InternalTable) fact.newInstance();
      // save the table to repository
      t.saveToRepository();
      // import the first instance from the table...
      Instances inst = (Instances) t.load(1);
      // work on instances here...
      System.out.println(inst.toString());
    }
    catch(kddml.Core.KDDMLCoreException e) {
      System.out.println("Exception: "+e.getDetailErrorMessage());
    }
  }
}
