/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataSources;

import org.w3c.dom.*;
import weka.core.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * This class for statistics provides a basic framework for representing univariate
 * statistics. The statistics for a model is made of the collection of the statistics
 * for single fields. <br>
 * The statistics depends on type of attributes:
 * <ol>
 * <li> mean, variance, min, max for numerical attributes;</li>
 * <li> cardinality of each value for nominal attributes.</li>
 * </ol>
 * <br>
 * This class cannot be istantied outside of the package.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class DataStatistics implements DataStatisticsManager {

  protected AttributeStatisticManager[] attribute_list;

  protected int num_instances;

  protected String relation_name;

  protected int num_numeric_attributes;

  protected int num_string_attributes;

  protected int num_enumerated_attributes;

  /**
   * Proprietary constructor given a set of weka instances.
   *
   * @param instances Instances
   * @throws InternalTableException
   */
  protected DataStatistics(Instances instances)
      throws InternalTableException {

    attribute_list = new AttributeStatistic[instances.numAttributes()];

    relation_name = instances.relationName();
    num_instances = 0;
    num_numeric_attributes = 0;
    num_string_attributes = 0;
    num_enumerated_attributes = 0;

    // vengono inizializzate le statistiche dei singoli attributi...
    for (int i = 0; i < instances.numAttributes(); i++) {
      Attribute attr = instances.attribute(i);
      if (attr.isNumeric()) {
        num_numeric_attributes++;
        NumericalStatistic ns = new NumericalStatistic(attr.name());
        attribute_list[i] = ns;
      }
      else
      if (attr.isNominal()) {
        num_enumerated_attributes++;
        NominalStatistic noms = new NominalStatistic(attr.name(), attr.enumerateValues());
        attribute_list[i] = noms;
      }
      else {
        num_string_attributes++;
        StringStatistic ss = new StringStatistic(attr.name());
        attribute_list[i] = ss;
      }
    }
    updateStatistic(instances);
  }


  /**
   * Proprietary constructor given the XML schema.
   *
   * @param schema Element
   * @throws KDDMLCoreException
   * @throws InternalTableException
   */
  protected DataStatistics(Element schema)
      throws kddml.Core.KDDMLCoreException, InternalTableException {

    num_numeric_attributes = 0;
    num_string_attributes = 0;
    num_enumerated_attributes = 0;

    if (schema == null)
      throw new InternalTableException("Invalid XML input document.", null,
                                       KDDMLCoreErrorCodes.NULL_ARGUMENT);
    // Estrae il nome logico del database contenente le tuple
    relation_name = kddml.Core.XMLUtility.getAttribute(schema, SpaceNames.ATTRIBUTE_RELATION_NAME);
    // setta il numero delle istanze
    num_instances = new Integer(schema.getAttribute(SpaceNames.ATTRIBUTE_NUM_INSTANCES)).intValue();
    // crea il vettore contenente gli attributi della tabella
    int num_attr = new Integer(schema.getAttribute(SpaceNames.ATTRIBUTE_NUM_ATTRIBUTES)).intValue();

    attribute_list = new AttributeStatisticManager[num_attr];

    // scorre gli attributi dello schema
    int i=0;
    for(Element attributo = (Element) schema.getElementsByTagName(SpaceNames.ELEMENT_ATTRIBUTE).item(0);
        attributo!=null;
        attributo = (Element) schema.getElementsByTagName(SpaceNames.ELEMENT_ATTRIBUTE).item(i)) {
      // Controlla il tipo dell'attributo corrente
      String type = attributo.getAttribute(SpaceNames.ATTRIBUTE_TYPE);
      AttributeType attr_type = new AttributeType(type);
      if (attr_type.equals(AttributeType.NOMINAL)) {

        // ATTRIBUTO DI TIPO NOMINALE...
        num_enumerated_attributes++;
        NominalStatistic noms = new NominalStatistic(attributo, num_instances);
        attribute_list[i] = noms;
      }
      else {
        if (attr_type.equals(AttributeType.NUMERIC)) {

          // ATTRIBUTO DI TIPO NUMERICO...
          num_numeric_attributes++;
          NumericalStatistic num = new NumericalStatistic(attributo, num_instances);
          attribute_list[i] = num;
        }
        else {
          if (attr_type.equals(AttributeType.STRING)) {

            // ATTRIBUTO DI TIPO STRINGA...
            num_string_attributes++;
            StringStatistic str = new StringStatistic(attributo, num_instances);
            attribute_list[i] = str;
          }
        }
      }
      i++;
    }
  }

  /**
   * Updates the statistical information on attributes given a set of weka instances
   * as input.
   *
   * @param instances Instances
   * @throws InternalTableException
   */
  protected void updateStatistic(Instances instances)
      throws InternalTableException {

    if (instances.numInstances() == 0)
      return;
    num_instances = num_instances + instances.numInstances();
    // vengono aggiornate le singole statistiche...
    for (int i = 0; i < instances.numAttributes(); i++) {
      Attribute a = instances.attribute(i);
      if (a.isNumeric()) {
        try {
          NumericalStatistic ns = (NumericalStatistic) attribute_list[i];
          ns.updateStatistic(instances);
        }
        catch(java.lang.ClassCastException e) {
          throw new InternalTableException(
              "Statistics building exception: Incompatible input instances on attribute <"+a.name()+">.",
            null, KDDMLCoreErrorCodes.STATISTIC_TYPE_CASTING_ERROR);
        }
      }
      else
      if (a.isNominal()) {
        try {
          NominalStatistic ns = (NominalStatistic) attribute_list[i];
          ns.updateStatistic(instances);
        }
        catch(java.lang.ClassCastException e) {
          throw new InternalTableException(
              "Statistics building exception: Incompatible input instances on attribute <"+a.name()+">.",
            null, KDDMLCoreErrorCodes.STATISTIC_TYPE_CASTING_ERROR);
        }
      }
      else {
        try {
          StringStatistic ss = (StringStatistic) attribute_list[i];
          ss.updateStatistic(instances);
        }
        catch(java.lang.ClassCastException e) {
          throw new InternalTableException(
              "Statistics building exception: Incompatible input instances on attribute <"+a.name()+">.",
            null, KDDMLCoreErrorCodes.STATISTIC_TYPE_CASTING_ERROR);
        }
      }
    }
  }

  /**
   * Returns the number of instances.
   *
   * @return int
   */
  public int getNumInstances() {
    return num_instances;
  }

  /**
   * Returns the number of attributes.
   *
   * @return int
   */
  public int getNumAttributes() {
    return attribute_list.length;
  }

  /**
   * Returns the attributes statistic.
   *
   * @return AttributeStatistic[] a vector of attribute statistic.
   */
  public AttributeStatisticManager[] getStatistics() {
    return attribute_list;
  }

  /**
   * Returns the statistics of a given attribute.
   * Returns null if the attribute do not exist.
   *
   * @param attribute_name String the attribute name.
   * @return AttributeStatistic
   */
  public AttributeStatisticManager getStatistic(String attribute_name) {
    for (int i=0; i<attribute_list.length; i++) {
      AttributeStatistic at = (AttributeStatistic) attribute_list[i];
      if (at.getName().equalsIgnoreCase(attribute_name))
        return (AttributeStatisticManager) at;
    }
    return null;
  }

  /**
   * Returns the attribute index.
   * Returns -1 if the attribute do not exist.
   *
   * @param attribute_name String the attribute name.
   * @return int
   */
  public int getAttributeIndex(String attribute_name) {

    for (int i=0; i < attribute_list.length; i++) {
      AttributeStatisticManager attr = attribute_list[i];
      if (attr.getName().equals(attribute_name))
        return i;
    }
    return -1;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
   String s = "------------------------------------------------------\n";
   s = s + "Relation Name: "+relation_name + "\n";
   s= s + "Number of Instances: "+num_instances + "\n";
   s= s + "Number of Attributes: "+attribute_list.length + "\n";
   s = s + "-----------\n";
   s = s + "Attribute Description:\n";
   s = s + "-----------\n";
   for (int i=0; i<attribute_list.length; i++) {
     AttributeStatistic at = (AttributeStatistic) attribute_list[i];
     s = s + at.toString();
   }
   return s;
  }

  /**
   * Returns the logical relation name.
   *
   * @return String
   */
  public String getRelationName() {
    return relation_name;
  }

  /**
   * Returns the number of numerical attributes.
   *
   * @return int
   */
  public int getNumberOfNumericAttributes() {
    return num_numeric_attributes;
  }

  /**
   * Returns the number of nominal attributes.
   *
   * @return int
   */
  public int getNumberOfNominalAttributes() {
    return num_enumerated_attributes;
  }

  /**
   * Returns the number of string attributes.
   *
   * @return int
   */
  public int getNumberOfStringAttributes() {
    return num_string_attributes;
  }

  protected Element toXML(boolean isdata) {
    String name;
    // costruisce l'albero DOM...
    if (isdata){
      name=SpaceNames.ELEMENT_SCHEMA;
    }
    else{
      name=SpaceNames.ELEMENT_PPSCHEMA;
    }
    Element schema = kddml.Core.XMLUtility.doc.createElement(name);
    kddml.Core.XMLUtility.setAttribute(schema, SpaceNames.ATTRIBUTE_RELATION_NAME,relation_name);
    //schema.setAttribute("file_name", relation_name + PhysicalDataRepositoryFactory.DATA_EXTENSION);
    // settaggio del numero delle istanze
    kddml.Core.XMLUtility.setAttribute(schema,
                                        SpaceNames.ATTRIBUTE_NUM_ATTRIBUTES,
                                        new Integer(getNumAttributes()).toString());
    kddml.Core.XMLUtility.setAttribute(schema,
                                        SpaceNames.ATTRIBUTE_NUM_INSTANCES,
                                        new Integer(num_instances).toString());
    // ciclo che scorre gli attributi
    for (int i=0; i<attribute_list.length; i++) {
      AttributeStatistic at = (AttributeStatistic) attribute_list[i];
      schema.appendChild(at.toXML());
    }
    return schema;
  }
}
