/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import java.io.*;
import java.util.*;
import org.w3c.dom.Element;


/**
 * This utility class contains methods and variables used inside this package
 * for HTML table generation.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
final class HTMLUtility {

  // stabilisce il numero massimo di istanze visualizzabili nel documento HTML
  public final static int INSTANCES_VISUALIZATION_LIMIT = 10000;

  // stabilisce il numero massimo di attributi visualizzabili nel documento HTML
  public final static int ATTRIBUTES_VISUALIZATION_LIMIT = 100;

  /**
   * Produce i risultati HTML per un documento XML di tipo TABELLA.
   * Data la presenza di dati in formato testuale, si utilizza direttamente Java per la creazione
   * del file HTML richiesto
   *
   * @param html_path String
   * @param data_path String
   * @param stat DataStatistics
   * @throws FileNotFoundException
   * @throws IOException
   * @return boolean
   */
  protected static boolean writeHTMLInstances(String html_path,
                                              String data_path,
                                              DataStatistics stat)
      throws java.io.FileNotFoundException, java.io.IOException {

    if ((stat.getNumInstances() > INSTANCES_VISUALIZATION_LIMIT) ||
        (stat.getNumAttributes() > ATTRIBUTES_VISUALIZATION_LIMIT))
      return false;

    OutputStream os = new FileOutputStream(html_path);
    PrintWriter pw = new PrintWriter(os,true);
    HTMLUtility.openHTMLDataTable(pw);

    HTMLUtility.insertHTMLAttribute(pw, stat);

    // lettura delle istanze dal file .data
    FileReader inputFile = new FileReader(data_path);

    BufferedReader in = new BufferedReader(inputFile);

    // line viene posizionato alla prima riga del file
    String line = in.readLine();
    while (line != null) {
      HTMLUtility.insertHTMLRow(pw, line);
      line = in.readLine();
    }
    HTMLUtility.closeHTMLDataTable(pw);
    return true;
  }

  /**
   * Genera il codice HTML per la visualizzazione della tabella.
   *
   * @param pw PrintWriter
   */
  protected static void openHTMLDataTable(PrintWriter pw) {

    // creazione dell'intestazione
    pw.println("<html>");
    pw.println("<body bgcolor=\"#E7F3F7\">");
    pw.println("<h2>");
    pw.println("<left>");
    pw.println("<font color=\"#808080\"><b>");
    pw.println("<TABLE border=\"2\" cellpadding=\"10\" cellspacing=\"2\" width=\"100%\">");
    pw.println("<TR>");
    pw.println("<TH width=\"100%\" align=\"left\"><FONT color=\"black\" size=\"+1\">");
    pw.println("<B>Table: Instances</B></FONT></TH>");
    pw.println("</TR>");
    pw.println("</TABLE>");
    pw.println("<BR d=\"\">");
    pw.println("<BR d=\"\">");
    pw.println("</b></font>");
    pw.println("</left>");
    pw.println("</h2>");
  }

  /**
   * Genera il codice HTML per la visualizzazione della tabella.
   *
   * @param pw PrintWriter
   */
  protected static void closeHTMLDataTable(PrintWriter pw) {
    pw.println("</table>");
    pw.println("</p>");
    pw.println("</body>");
    pw.println("</html>");
  }

  /**
   * Genera il codice HTML per la colonna degli attributi.
   *
   * @param pw PrintWriter the print writer object.
   * @param stat DataStatistics
   */
  protected static void insertHTMLAttribute(PrintWriter pw, DataStatistics stat) {
    int num_attr = stat.getNumAttributes();
    // prima riga della tabella
    //String num_attr = kddml_table.getAttribute("numberOfAttributes");
    pw.println("<table border=\"1\" width=\"100%\">");
    pw.println("<tr>");
    pw.println("<th bgcolor=\"#FF9933\" colspan=\""+num_attr+"\" height=\"18\">Output Data</th>");
    pw.println("</tr>");
    pw.println("<tr>");
    // scorre gli attributi dello schema
    AttributeStatisticManager[] statistics = stat.getStatistics();
    for (int i = 0; i< statistics.length; i++) {
      pw.println("<th bgcolor=\"#FFCC66\" height=\"19\">" + statistics[i].getName() + "</th>");
    }
    pw.println("</tr>");
  }

  /**
   * Inserimento di una riga nella tabella.
   *
   * @param pw PrintWriter
   * @param instance String
   */
  protected static void insertHTMLRow(PrintWriter pw, String instance) {
    StringTokenizer st = new StringTokenizer(instance, ",");
    String word;
    pw.println("<tr>");
    while (st.hasMoreElements()) {
      word = st.nextToken();
      pw.println("<td bgcolor=\"#FFFFCC\" align=\"center\">");
      pw.println(word);
      pw.println("</td>");
    }
    pw.println("</tr>");
  }
}
