/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataSources;

import org.w3c.dom.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;

/**
 * This class represent the history related to a PPInternalTable.
 * The history contains the list of preprocessing operations (with description)
 * performed on the table.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class History implements java.io.Serializable, HistoryManager {

  protected Vector task;

  /**
   * Empty constructor used to start the preprocessing description.
   */
  public History() {
    task = new Vector();
    HistoryTask t = new HistoryTask("TABLE_2_PPTABLE", "Start preprocessing phase");
    task.add(t);
  }

  /**
   * Constructor given a new history.
   * It clones the input object returning a new object.
   *
   * @param hist History
   */
  public History(History hist) {
    task = new Vector();
    for (int i=0;i<hist.task.size();i++){
      task.add((HistoryTask)hist.task.get(i));
    }
  }

  /**
   * Proprietary constructor given the XML element.
   *
   * @param history Element
   * @throws InternalTableException
   */
  protected History(Element history) throws InternalTableException {
    task = new Vector();
    if (history == null)
      throw new InternalTableException("Invalid XML input document.", null,
                                       KDDMLCoreErrorCodes.NULL_ARGUMENT);

    // scorre gli attributi della history
    int i = 0;
    for (Element task_el =  (Element)history.getElementsByTagName(SpaceNames.
        ELEMENT_PREPROCESSING_TASK).item(0);

         task_el != null;
         task_el = (Element) history.getElementsByTagName(SpaceNames.
        ELEMENT_PREPROCESSING_TASK).item(i)) {
      task.add( new HistoryTask (task_el));

      i++;

    }
  }

  protected Element toXML() {
    String name;

    name = SpaceNames.ELEMENT_HISTORY;

    Element hist = kddml.Core.XMLUtility.doc.createElement(name);

    // ciclo che scorre gli attributi
    for (int i = 0; i < task.size(); i++) {
      HistoryTask tsk = (HistoryTask) task.get(i);
      hist.appendChild(tsk.toXML());
    }
    return hist;
  }

  public String toString() {
    String s = "------------------------------------------------------\n";
    s = s + "Preprocessing history: " + "\n";

    for (int i = 0; i < task.size(); i++) {
      HistoryTask at = (HistoryTask) task.get(i);
      s = s + at.toString();
    }
    return s;
  }

  /**
   * Adds an history description to the history sequence.
   *
   * @param op String the preprocessing operator name.
   * @param desc String the description of operation.
   * @throws InternalTableException
   */
  public void addHistory(String op,String desc) {

    task.add(new HistoryTask(op,desc));
  }
}

/**
 * This class represents a single history task (i.e. a single history row).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */

class HistoryTask {

  /**
   * The preprocessing operator name.
   */
  protected String operation;

  /**
   * The description of preprocessing operation.
   */
  protected String description;

  /**
   * Proprietary constructor given the XML element representing the history task.
   *
   * @param task_element Element
   */
  protected HistoryTask(Element task_element) {
    operation = kddml.Core.XMLUtility.getAttribute(task_element, SpaceNames.ELEMENT_OPERATION);
    description = kddml.Core.XMLUtility.getAttribute(task_element, SpaceNames.ELEMENT_DESCRIPTION);
  }

  /**
   * Proprietary constructor given the operator name and operator description.
   *
   * @param op String
   * @param descr String
   */
  protected HistoryTask(String op,String descr) {
    operation = op;
    description = descr;
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s = "Attribute < "+operation+" >\n";
    s = s + "description: "+description+"\n";
    return s;
  }

  /**
   * Returns the XML element representing this object.
   *
   * @return Element
   */
  protected Element toXML() {
    Element elemento = kddml.Core.XMLUtility.doc.createElement(SpaceNames.
        ELEMENT_PREPROCESSING_TASK);
    kddml.Core.XMLUtility.setAttribute(elemento, SpaceNames.ELEMENT_OPERATION, operation);
    kddml.Core.XMLUtility.setAttribute(elemento, SpaceNames.ELEMENT_DESCRIPTION, description);

    return elemento;
  }
}
