/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import weka.core.*;
import java.io.*;
import kddml.Core.*;
import weka.core.FastVector;
import java.util.StringTokenizer;
import kddml.Core.DataSources.AttributeType;

/**
 * This utility class contains methods and variables used inside this package
 * for accessing to data repository.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
final class IOUtility {

  /**
   * Save a set of weka instances table in the data repository.
   *
   * @param xml_file String
   * @param table Instances
   * @param metadata Instances
   * @param statistics DataStatistics
   * @throws InternalTableException
   * @return boolean
   */
  protected static boolean saveTable(String xml_file,
                                     Instances table,
                                     Instances metadata,
                                     DataStatistics statistics)
      throws InternalTableException {

    String extension = xml_file.substring(xml_file.length()-3, xml_file.length());
    if (!extension.toLowerCase().equals("xml")) {
      String msg = "Illegal input path for: " + xml_file +". No XML extension found!";
      throw new InternalTableException(msg, null, KDDMLCoreErrorCodes.IO_ERROR);
    }

    String xml_dest = xml_file;
    String data_dest = xml_file.substring(0, xml_file.length()-4) + SpaceNames.DATA_EXTENSION;

    // salvataggio dei metadati
    if (statistics != null) {
      //System.out.println(xml_dest);
      //Element schema = statistics.toXML();
      Document kddml_object = XMLUtility.initializeDocument("KDDML_OBJECT");

      //kddml_table.appendChild(schema);
      Element table_el = XMLUtility.doc.createElement("KDDML_TABLE");
      kddml_object.getDocumentElement().appendChild(table_el);
      //System.out.println(XMLUtility.toString(kddml_object));
      XMLUtility.toFile(xml_dest, kddml_object);
    }

    // salvataggio dei dati...
    if (table != null) {
      try {
        File f = new File(data_dest);
        FileWriter fw = new FileWriter(f, true);
        insertData(table, fw);

        if (metadata!=null){
          String metadata_dest = xml_file.substring(0, xml_file.length() - 4)+ ".metadata"+ SpaceNames.DATA_EXTENSION
            ;
           try {
             File f2 = new File(metadata_dest);
             FileWriter fw2 = new FileWriter(f2, true);
             insertData(metadata, fw2);
           } catch (java.io.IOException e) {
        String msg = "I/O error saving data source in "+metadata_dest+".";
        throw new InternalTableException(msg,msg+e.toString(),KDDMLCoreErrorCodes.IO_ERROR);
      }


  }

      }
      catch (java.io.IOException e) {
        String msg = "I/O error saving data source in "+data_dest+".";
        throw new InternalTableException(msg,msg+e.toString(),KDDMLCoreErrorCodes.IO_ERROR);
      }
    }
    return true;
  }

  protected static void refreshDataFile(String xml_file) throws InternalTableException {
    String data_dest = xml_file.substring(0, xml_file.length()-4) +
        SpaceNames.DATA_EXTENSION;
    File f = new File(data_dest);
    try {
      (new FileWriter(f)).write("");
    }
    catch(Exception e) {
      throw new InternalTableException("I/O Error.", e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
    }
  }
  protected static void refreshMetaDataFile(String xml_file) throws InternalTableException {
   String data_dest = xml_file.substring(0, xml_file.length()-4) +".metadata"+
       SpaceNames.DATA_EXTENSION;
   File f = new File(data_dest);
   try {
     (new FileWriter(f)).write("");
   }
   catch(Exception e) {
     throw new InternalTableException("I/O Error.", e.toString(), KDDMLCoreErrorCodes.IO_ERROR);
   }
 }



  /**
   * Insert the physical tuples in the output text file.
   *
   * @param istanze Input instances
   * @param out the output FileWriter
   * @throws InternalTableException
   */
  private static void insertData(Instances istanze, FileWriter out) throws
      InternalTableException {

    String valore;
    Instance tupla;
    try {
      // effettua un ciclo sulle istanze
      for (int k = 0; k < istanze.numInstances(); k++) {
        // ottiene la singola tuple
        tupla = istanze.instance(k);
        if (istanze.numAttributes() != 0) {
          // inserisce il valore per il primo attributo della tupla
          if (tupla.isMissing(0))
            valore = "?";
          else {
            if (tupla.attribute(0).isNumeric())
              valore = new Double(tupla.value(0)).toString();
            else
              valore = tupla.stringValue(0);
          }
          out.write(valore);
        }
        // effettua un ciclo sui valori della singola tupla
        for (int h = 1; h < istanze.numAttributes(); h++) {
          // ottiene il valore (una stringa) in posizione h della tupla k
          if (tupla.isMissing(h))
            valore = "?";
          else {
            if (tupla.attribute(h).isNumeric())
              valore = new Double(tupla.value(h)).toString();
            else
              valore = tupla.stringValue(h);
          }
          // vengono rimpiazzate le virgole con una rappresentazione di default
          valore = valore.replaceAll(",", "_COMMA_");
          //valore = valore.replaceAll("'", "");
          out.write("," + valore);
        }
        out.write("\n");
      }
      out.flush();
      out.close();
    }
    catch (java.io.IOException ecc) {
      String msg1 = "I/O error writing in "+out.toString();
      String msg2 = ecc.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  /**
   * Import the object from the physical data repository.
   * In the repository must exist the .xml file and the .csv file.
   *
   * @param xml_path File
   * @param data_path File
   * @param begin Long
   * @param end Long
   * @throws InternalTableException
   * @throws DOMParserException
   * @throws KDDMLCoreException
   * @return Instances
   */
  protected static Instances loadTable(File xml_path, File data_path, Long begin, Long end)
      throws InternalTableException, kddml.Core.DOMParserException, KDDMLCoreException {

    if (!xml_path.exists()) {
      String msg1 = "Table not exist in repository: " + xml_path.toString();
      throw new InternalTableException(msg1,
                                       null,
                                       KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);
    }
    // effettua il parsing del documento XML contenente lo schema della tabella
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(xml_path.getAbsolutePath()),
                              new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_TABLE));
    }
    catch(IOException e) {
      throw new InternalTableException("I/O Error on file: "+xml_path.toString(),
                                       null,
                                       KDDMLCoreErrorCodes.IO_ERROR);
    }
    Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
    String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

    Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_SCHEMA).item(0);

    DataStatistics stat = new DataStatistics(schema);

    Instances inst = readSchema(schema);
    return readData(inst, data_path.getAbsolutePath(), begin, end);
  }

  /**
   * Import the object from the physical data repository.
   * In the repository must exist the .xml file and the .csv file.
   *
   * @param xml_path File
   * @param data_path File
   * @param index int[]
   * @throws InternalTableException
   * @throws DOMParserException
   * @throws KDDMLCoreException
   * @return Instances
   */
  protected static Instances loadTable(File xml_path, File data_path, int[] index)
      throws InternalTableException, kddml.Core.DOMParserException, KDDMLCoreException {

    if (!xml_path.exists()) {
      String msg1 = "Table not exist in repository: " + xml_path.toString();
      throw new InternalTableException(msg1,
                                       null,
                                       KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);
    }
    // effettua il parsing del documento XML contenente lo schema della tabella
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(xml_path.getAbsolutePath()),
                              new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_TABLE));
    }
    catch(IOException e) {
      throw new InternalTableException("I/O Error on file: "+xml_path.toString(),
                                       null,
                                       KDDMLCoreErrorCodes.IO_ERROR);
    }
    Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
    String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

    Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_SCHEMA).item(0);

    DataStatistics stat = new DataStatistics(schema);

    Instances inst = readSchema(schema);
    return readData(inst, data_path.getAbsolutePath(), index);
  }

   /**
    * Import the object from the physical data repository.
    * In the repository must exist the .xml file and the .csv file.
    *
    * @param xml_path File
    * @param data_path File
    * @param begin Long
    * @param end Long
    * @throws InternalTableException
    * @throws DOMParserException
    * @throws KDDMLCoreException
    * @return Instances
    */
   protected static Instances loadMetaTable(File xml_path, File data_path, Long begin, Long end)
       throws InternalTableException, kddml.Core.DOMParserException, KDDMLCoreException {

     if (!xml_path.exists()) {
       String msg1 = "Table not exist in repository: " + xml_path.toString();
       throw new InternalTableException(msg1,
                                        null,
                                        KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);
     }
     // effettua il parsing del documento XML contenente lo schema della tabella
     Document doc;
     try {
       doc = XMLUtility.parse(IOUtil.getInputStream(xml_path.getAbsolutePath()),
                               new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE));
     }
     catch(IOException e) {
       throw new InternalTableException("I/O Error on file: "+xml_path.toString(),
                                        null,
                                        KDDMLCoreErrorCodes.IO_ERROR);
     }
     Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
     String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

     Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_PPSCHEMA).item(0);

     DataStatistics stat = new DataStatistics(schema);

     Instances inst = readSchema(schema);
     return readData(inst, data_path.getAbsolutePath(), begin, end);
   }


   protected static Instances loadMetaTable(File xml_path, File data_path, int[] index)
       throws InternalTableException, kddml.Core.DOMParserException, KDDMLCoreException {

     if (!xml_path.exists()) {
       String msg1 = "Table not exist in repository: " + xml_path.toString();
       throw new InternalTableException(msg1,
                                        null,
                                        KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);
     }
     // effettua il parsing del documento XML contenente lo schema della tabella
     Document doc;
     try {
       doc = XMLUtility.parse(IOUtil.getInputStream(xml_path.getAbsolutePath()),
                               new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE));
     }
     catch(IOException e) {
       throw new InternalTableException("I/O Error on file: "+xml_path.toString(),
                                        null,
                                        KDDMLCoreErrorCodes.IO_ERROR);
     }
     Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
     String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

     Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_PPSCHEMA).item(0);

     DataStatistics stat = new DataStatistics(schema);

     Instances inst = readSchema(schema);
     return readData(inst, data_path.getAbsolutePath(), index);
   }

  /**
   * Importazione delle tuple della tabella.
   *
   * @param istanze Instances
   * @param path String
   * @param begin Long
   * @param end Long
   * @throws InternalTableException
   * @return Instances
   */
  private static Instances readData(Instances istanze, String path,
                                    Long begin, Long end)
      throws InternalTableException {
    try {
      int num_attributi = istanze.numAttributes();
      FileReader inputFile = new FileReader(path);
      BufferedReader in = new BufferedReader(inputFile);

      // line viene posizionato alla prima riga del file
      String line = in.readLine();
      long pointer = 1;
      if (begin != null) {
        while (line != null) {
          if (pointer == begin.longValue())
            break;
          pointer++;
          line = in.readLine();
        }
        // raggiunta la fine del file...
        if (line == null) {
          String msg1 = "I/O error reading "+path+".\n";
          String msg2 = "Index of instance out of range.";
          throw new InternalTableException(msg1+msg2, null, KDDMLCoreErrorCodes.IO_ERROR);
        }
      }

      while (line != null) {
        if (end != null) {
          if (pointer > end.longValue())
            // esce dal ciclo perche' raggiunto il limite.
            break;
        }
        pointer++;

        Instance tupla = getInstance(istanze, line);
        tupla.setDataset(istanze);
        istanze.add(tupla);
        tupla = new Instance(num_attributi);
        line = in.readLine();
      }
      in.close();
      return istanze;
    }
    catch (java.io.IOException e) {
      String msg1 = "I/O error reading "+path;
      String msg2 = e.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  /**
   * Importazione dello schema della tabella.
   *
   * @param schema Element
   * @throws InternalTableException
   * @return Instances
   */
  protected static Instances readSchema(Element schema)
      throws InternalTableException {

    if (schema == null)
      throw new InternalTableException("Invalid XML input document.", null,
                                       KDDMLCoreErrorCodes.NULL_ARGUMENT);

    // Estrae il nome logico del database contenente le tuple
    String nomeDatabase = XMLUtility.getAttribute(schema, SpaceNames.ATTRIBUTE_RELATION_NAME);
    // crea il vettore contenente gli attributi della tabella
    FastVector lista_attributi = new FastVector();

    // scorre gli attributi dello schema
    int i=0;
    for(Element attributo = (Element) schema.getElementsByTagName(SpaceNames.ELEMENT_ATTRIBUTE).item(0);
        attributo!=null;
        attributo = (Element) schema.getElementsByTagName(SpaceNames.ELEMENT_ATTRIBUTE).item(i)) {
      i++;
      // Controlla il tipo dell'attributo corrente
      String type = attributo.getAttribute(SpaceNames.ATTRIBUTE_TYPE);
      AttributeType type2 = new AttributeType(type);
      if (type2.equals(AttributeType.NOMINAL)) {

        // ATTRIBUTO DI TIPO NOMINALE...
        FastVector fv = new FastVector();
        int j=0;
        // Inserisce i valori dell'attributo all'interno del vettore creato
        for(Element enumerato = (Element) attributo.getElementsByTagName(SpaceNames.ELEMENT_VALUE).item(j);
            enumerato!=null;
            enumerato = (Element) attributo.getElementsByTagName(SpaceNames.ELEMENT_VALUE).item(j)) {
          j++;
          fv.addElement(XMLUtility.getAttribute(enumerato, SpaceNames.ATTRIBUTE_VALUE).trim());
        }
        Attribute attr = new Attribute(attributo.getAttribute(SpaceNames.ATTRIBUTE_NAME),fv);
        lista_attributi.addElement(attr);
      }
      else {
        if (type2.equals(AttributeType.NUMERIC)) {
          // ATTRIBUTO DI TIPO NUMERICO...
          Attribute attr = new Attribute(attributo.getAttribute(SpaceNames.ATTRIBUTE_NAME));
          lista_attributi.addElement(attr);
        }
        else {
          // ATTRIBUTO DI TIPO STRINGA...
          FastVector fv = new FastVector(0);
          fv = null;
          Attribute attr = new Attribute(attributo.getAttribute(SpaceNames.ATTRIBUTE_NAME), fv);
          lista_attributi.addElement(attr);
        }
      }
    }
    // CREAZIONE DELL'OGGETTO InternalTable
    Instances inst = new Instances(nomeDatabase,lista_attributi,0);
    return inst;
  }

  /**
   * Importazione delle tuple della tabella.
   *
   * @param inst un insieme di istanze vuote.
   * @param path il riferimento al file di testo contenente i dati.
   * @param begin l'indice della prima istanza da importare.
   * @param end l'indice dell'ultima istanza da importare.
   *
   * @return le istanze finali.
   */
  /**
   * Importazione di un insieme di tuple identificate da indici
   *
   * @param istanze Instances
   * @param path String
   * @param index int[] vettore ordinato di indici. The instances start from 0.
   * @throws InternalTableException
   * @return Instances
   */
  private static Instances readData(Instances istanze, String path, int[] index)
      throws InternalTableException {
    try {
      FileReader inputFile = new FileReader(path);
      BufferedReader in = new BufferedReader(inputFile);
      // line viene posizionato alla prima riga del file
      String line = in.readLine();
      long pointer = 0;
      int pointer_array = 0;

      while (line != null) {

        if (pointer == index[pointer_array]) {
          // index instance found
          Instance tupla = getInstance(istanze, line);
          tupla.setDataset(istanze);
          istanze.add(tupla);

          if (pointer_array == (index.length-1))
            // exit
            break;
          else
            pointer_array++;
        }
        pointer++;
        line = in.readLine();

      }
      in.close();
      return istanze;
    }
    catch (java.io.IOException e) {
      String msg1 = "I/O error reading "+path;
      String msg2 = e.toString();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  private static Instance getInstance(Instances schema,
                                      String line) throws InternalTableException {

    StringTokenizer st = new StringTokenizer(line, SpaceNames.SEPARATOR);
    String word;
    int j=0;
    Instance tupla = new Instance(schema.numAttributes());

    while (st.hasMoreElements()) {
      word = st.nextToken();
      // ripristina le virgole...
      word = word.replaceAll("_COMMA_", SpaceNames.SEPARATOR);
      // SETTAGGIO DEL VALORE PER L'ATTRIBUTO CORRENTE
      // rimuove ogni eventuale spazio bianco presente nella stringa
      word = word.trim();
      if (!word.equals("")) {
        Attribute a = (Attribute) schema.attribute(j);
        j++;
        // controlla che il valore sia presente
        if (word.equals(SpaceNames.MISSING))
          tupla.setMissing(a);
        else
        if (a.isNumeric()) {
          // attributo di tipo numerico
          Double d = new Double(word);
          // WEKA ARROTONDA UN DOUBLE A FLOAT
          tupla.setValue(a, d.doubleValue());
        }
        else {
          // attributo di tipo enumerato o stringa
          tupla.setValue(a, word);
        }
      }
      else {
        String msg1 = "I/O error reading input file.\n";
        String msg2 = "Not value for instance: "+line.toString();
        throw new InternalTableException(msg1+msg2, null, KDDMLCoreErrorCodes.IO_ERROR);
      }
    }
    return tupla;
  }

  protected static void copyFile(File inputFile, File outputFile) throws IOException {

    FileReader in = new FileReader(inputFile);
    FileWriter out = new FileWriter(outputFile);
    int c;

    while ((c = in.read()) != -1)
      out.write(c);

    in.close();
    out.close();
  }

  public static void main(String args[]) {
    try {
      File f1 = new File("C:/kddml/repository/data/tables/weather.xml");
      File f2 = new File("C:/kddml/repository/data/tables/weather.csv");
      int[] index = {1,3,5,7};
      Instances i = loadTable(f1, f2, index);
      System.out.println(i.numInstances());
      System.out.println(i.toString());
    }
    catch (Exception ex) {
      System.out.println(ex.toString());
    }
  }
}
