/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 */

package kddml.Core.DataSources;

import weka.core.Instances;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import weka.core.Instances;
import java.io.*;
import kddml.Core.*;

/**
 * This object is used to represent relational data sources.
 * An internal table is composed by
 *
 * <ol>
 * <li>the data schema, that includes attributes types and some simple statistics
 * on attribute values; </li>
 * <li>the physical data as a text file in comma separated format.</li>
 * </ol>
 * <br>
 * Physical data can occur in several forms: relational table, transactional table, and timestamp table.
 * In the relational format, each column of the data corresponds to a logical attribute, e.g., temperature,
 * play tennis. Each row of the data corresponds to an individual case (transaction)
 * to be considered during mining. This data format is also known as single-record case table.
 * Sparse data are more effectively stored in a transactional format. Here, data that have a variable
 * number of entries (or items) from among many possible ones can be stored more compactly, since
 * only the items present are stored in the table. A transactional table (also known as multi-record
 * case table) has an attribute transaction identifying the transaction and an attribute event
 * containing the single item. Transactions are ordered with respect to the attribute transaction.
 * Other columns are allowed in the table (e.g., the price or quantity of items for each transaction),
 * but they can be ignored by the operator, depending on the context. This representation is typically
 * used for association rules.
 * Timestamp format is similar to a transactional table, but with an extra attribute timestamp.
 * This attribute defines a partial time order between transactions and items. Typically, this format is
 * used for sequential patterns analysis.
 * <br>
 * InternalTable class is used to represent relational tables.
 * <br>
 * This class can be istantied only using the factory classes of the package. It uses the weka.Core.* package
 * as representation of single instances.
 *
 * <p>Title: KDDML-MQL</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class InternalTable extends KDDMLObject implements InternalTableManager {

  /**
   * The physical data object.
   */
  protected File object_data_path;

  /**
   * All instaces loaded in RAM.
   */
  public final static int ALL = -1;

  /**
   * This element represents the data statistics.
   */
  protected DataStatistics statistic;

  /**
   * The physical instances pointer.
   */
  protected long pointer = 1;

  /**
   * Proprietary constructor given the file containing the schema and the data statistics.
   *
   * @param physical_path File
   * @param stat DataStatistics
   */
  protected InternalTable(File physical_path, DataStatistics stat) {
    super(physical_path);
    String s = physical_path.getName();
    s = s.substring(0, s.length()-4) + SpaceNames.DATA_EXTENSION;
    object_data_path = new File(physical_path.getParent() + File.separator + s);
    statistic = stat;
  }

  /**
   * Returns the statistics related to this table.
   *
   * @return InternalTableManager.
   */
  public DataStatisticsManager getStatistic() {
    return statistic;
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return KDDMLObjectType.INTERNAL_TABLE;
  }

  /**
   * Returns the table dimension in KBytes.
   *
   * @return int
   */
  public int getDimension() {
    return (int) (object_data_path.length() / 1000);
  }

  /**
   * Returns a representation of this table as string.
   *
   * @return String
   */
  public String toString() {
    String s;
    s = "Data Source: "+object_path.toString()+"\n";
    s = s + "----------------------------------\n";
    s = s + "Number of Instances: "+statistic.getNumInstances()+"\n";
    s = s + "Number of Attributes: "+statistic.getNumAttributes()+"\n";
    return s;
  }

  /**
   * Saves the object in the system repository.
   * The data schema is representend as XML document. The physical instances as
   * text file in comma separated format.
   * The destination path is provided by the object_path variable for the data schema
   * and by the object_data_path variable for the row data.
   * By definition, row data is already stored in the repository.
   *
   * @throws KDDMLCoreException
   */
  public void saveToRepository() throws kddml.Core.KDDMLCoreException {

    // salvataggio dei metadati
    if (statistic != null) {
      Document kddml_object = XMLUtility.initializeDocument(SpaceNames.ELEMENT_KDDML_OBJECT);
      Element table_el = XMLUtility.doc.createElement(SpaceNames.ELEMENT_KDDML_TABLE);

      XMLUtility.setAttribute(table_el, SpaceNames.ATTRIBUTE_DATA_FILE, object_data_path.getName());
      Element root = kddml_object.getDocumentElement();
      root.appendChild(table_el);
      Element schema = statistic.toXML(true);
      table_el.appendChild(schema);

      XMLUtility.toFile(object_path.getAbsolutePath(), kddml_object);
    }
    // data is in the repository
  }

  /**
   * It tests if the table contains instances.
   *
   * @return boolean
   */
  public boolean isEmpty() {
    return statistic.getNumInstances() == 0;
  }

  /**
   * Loads i instances from the position specified by the pointer and returns they
   * as generic object. If the parameter is greater than the total number of instances,
   * it returns all instances. If i is equal to 0, it returns only the schema of
   * tuples.
   *
   * @param i long
   * @throws KDDMLCoreException if no instances has been saved in the repository.
   * @return Object an set of instances as weka.core.Instances.
   */
  public java.lang.Object load(long i)
      throws KDDMLCoreException {

    if ((!object_path.exists()) || (!object_data_path.exists()))
      throw new KDDMLCoreException("Can't import instances: save object in the repository before loading.",
                                   null,
                                   KDDMLCoreErrorCodes.IO_ERROR);

    if (i == ALL) {
      // importazione di tutte le istanze del file
      pointer = statistic.getNumInstances() +1;
      return IOUtility.loadTable(object_path, object_data_path, null, null);
    }
    else {
      // importazione di un blocco delle istanze
      long begin, end;
      begin = pointer;
      end = java.lang.Math.min(pointer + i - 1, statistic.getNumInstances());
      // posiziona pointer all'istanza successiva
      pointer = end + 1;
      return IOUtility.loadTable(object_path, object_data_path,
                                 new Long(begin), new Long(end));
    }
  }

  /**
   * Returns only the instances whose indexes belong to the input vector.
   *
   * @param indexes int[] the array of indexes.
   * @throws KDDMLCoreException if indexes are out of range.
   * @return Object the output instances as weka.core.Instances.
   */
  public java.lang.Object loadInstances(int[] indexes) throws KDDMLCoreException {

    // ordina il vettore degli indici...
    java.util.Arrays.sort(indexes);

    if((indexes[indexes.length-1] > getNumOfInstances()-1) ||
       (indexes[0] < 0))
      throw new KDDMLCoreException("Can't import instances: instances index out of range.",
                                   null,
                                   KDDMLCoreErrorCodes.IO_ERROR);

    return IOUtility.loadTable(object_path, object_data_path, indexes);

  }

  /**
   * Returns true if the table is a transactional table with attributes transaction
   * and event. Returns false if the table is a relational table.
   *
   * @return boolean
   */
  public boolean isTransactionalTable() {
    return false;
  }

  /**
   * Returns true if the table is a timestamp table with attributes transaction,
   * event and timestamp.
   *
   * @return boolean
   */
  public boolean isTimestampTable() {
    return false;
  }

  /**
   * Adds the input instances to the table.
   * Throws an exception if the data schema of the input table and the data schema of
   * the input instances are not compatible.
   *
   * @param inst Object the set of instances as weka.core.Instances.
   * @throws InternalTableException if the schemas are not compatibles.
   */
  public void addInstances(Object inst)
      throws InternalTableException {

    // aggiorna le statistiche
    statistic.updateStatistic((Instances) inst);

    // aggiunge al file fisico l'insieme di istanze.
    IOUtility.saveTable(object_path.getAbsolutePath(), (Instances) inst, null, null);
  }

  /**
   * Tests if there are instances to load.
   *
   * @return boolean returns true if there are instances to load. Returns false if
   * the pointer is at the end of the data file.
   */
  public boolean hasNext() {
    return pointer <= statistic.getNumInstances();
  }

  /**
   * Resets the pointer to the begin of data file.
   */
  public void reset() {
    pointer = 1;
  }

  /**
   * Increase the pointer of the specified quantity.
   *
   * @param i int
   */
  public void increasePointer(int i) {
    pointer = pointer + i;
  }

  /**
   * Computes the optimal number of instances to load.
   * It can depend from the total number of instances and attributes and the quantity
   * of available system RAM.
   *
   * @return long
   */
  public long optimize() {
    // _______________________OTTIMIZZARE_____________________
    // dimensione massima di (num_inst * num_attr) che si desidera importare in RAM
    // [50.000 istanze X 10 attributi]
    final int MAX = 500000;

    // dimenensione minima di (num_inst * num_attr) che si desidera importare in RAM
    // [10.000 istanze X 10 attributi]
    final int MIN = 100000;

    // dimensione complessiva della tabella in termini di (num_inst * num_attr)
    long weight = statistic.getNumInstances() * statistic.getNumAttributes();
    if (weight < MIN)
      return this.ALL;
    if (weight > MAX)
      return MAX / statistic.getNumAttributes();

    return MIN / statistic.getNumAttributes();
  }

  /**
   * Save the instances in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML()
      throws KDDMLCoreException {

    XSLResolver resolver = new XSLResolver(KDDMLObjectType.INTERNAL_TABLE);

    String s1 = this.getPhysicalObject().getAbsolutePath();
    String s2 = resolver.getXSLPath();
    String s3 = s1 + ".html";

    // effettua il parsing del documento XML contenente lo schema della tabella
    Document doc;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream(this.getPhysicalObject().getAbsolutePath()),
                              new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_TABLE));
    }
    catch(IOException e) {
      throw new InternalTableException("I/O Error on file: "+this.getPhysicalObject().toString(),
                                       null,
                                       KDDMLCoreErrorCodes.IO_ERROR);
    }
    Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
    String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

    Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_SCHEMA).item(0);

    DataStatistics stat = new DataStatistics(schema);

    try {
      String data_ref = this.getPhysicalObject().getParent() + File.separator + data_file_name;
      // generazione del .html che rappresenta lo schema
      String[] p1 = {"file_xml", "data_file"};
      String[] p2 = {this.getPhysicalObject().getName(), data_file_name+".html"};
      XMLUtility.formatting(s1, s2, s3, p1, p2);

      // generazione del .html che contiene i dati
      String html = data_ref + ".html";
      if (!HTMLUtility.writeHTMLInstances(html, data_ref, stat)) {
        // le istanze non sono visualizzabili
        OutputStream os = new FileOutputStream(html);
        PrintWriter pw = new PrintWriter(os,true);
        HTMLUtility.openHTMLDataTable(pw);
        pw.print("No HTML visualization is available: too many instances or attributes.\n");
        pw.print("Please open the .csv file");
        HTMLUtility.closeHTMLDataTable(pw);
      }

    }
    catch(Exception e) {
      String msg1 = "Exception found during the HTML code generation:\n";
      String msg2 = msg1 + e.getMessage();
      throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.HTML_TRANSFORMATION_ERROR);
    }
  }

  /**
   * Returns the total number of instances.
   *
   * @return int
   */
  public int getNumOfInstances() {
    return statistic.getNumInstances();
  }

  /**
   * Returns the total number of attributes.
   *
   * @return int
   */
  public int getNumOfAttributes() {
    return statistic.getNumAttributes();
  }
}
