/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import weka.core.*;
import java.io.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A generic factory class that create a table from an external source
 * (e.g., ARFF, Database, etc).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class InternalTableExternalFactory extends InternalTableResourceFactory {

  protected String logical_name;

  protected Integer max_number_of_inst_loaded_at_iteration;

  /**
   * Constructor given the physical file containing the table.
   * The file refers the XML table schema.
   *
   * @param f File the file containing the reference to the physical object.
   * @throws KDDMLCoreException an exception if the file cannot be found.
   */
  public InternalTableExternalFactory(java.io.File f)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    max_number_of_inst_loaded_at_iteration = null;
    logical_name = null;
  }

  /**
   * Create a new instance of this object.
   *
   * @throws KDDMLCoreException if the object can be overwrite.
   * @return KDDMLObject the KDDML internal table.
   */
  public kddml.Core.KDDMLObject newInstance() throws kddml.Core.KDDMLCoreException {
    // genera eccezione se il file di dest esiste e non puo' essere sovrascritto
    if (existObject() & !overwrite)
      throw new InternalTableException(
          "Destination object < " + physical_file.toString() + " > already exist in data repository.",
          null,
          KDDMLCoreErrorCodes.OBJECT_EXIST_IN_REPOSITORY);

    if (existObject() & overwrite)
      IOUtility.refreshDataFile(physical_file.getAbsolutePath());

    DataStatistics stat = null;
    int n;
    if (max_number_of_inst_loaded_at_iteration == null)
      n = Integer.MAX_VALUE;
    else
      n = max_number_of_inst_loaded_at_iteration.intValue();

    boolean b = true;

    FastVector attributes = getHeader();

    String relation_name = null;
    if (logical_name != null)
      relation_name = logical_name;
    else
      relation_name = getRelationName();

    Instances instances = null;
    Instances schema = null;
    Instance istance = null;

    int total_rows = 0;
    do {
      instances = new Instances(relation_name, attributes, 0);
      schema = instances;
      for (int i=0; i<n; i++) {
        // ciclo che inserisce le istanze nel dataset
        if (max_number_of_inst_loaded_at_iteration != null)
          if (total_rows % max_number_of_inst_loaded_at_iteration.intValue() == 0) {
            System.out.println("Scanning " + total_rows + " rows from data source..");
          }

        istance = nextInstance(schema);
        if (istance == null)
          break;
        // aggiunge l'istanza al dataset
        istance.setDataset(instances);
        instances.add(istance);
        total_rows++;
      }
      // aggiorna le statistiche e scrive le istanze su disco
      if (stat == null)
        stat = new DataStatistics(instances);
      else
        stat.updateStatistic(instances);
      IOUtility.saveTable(physical_file.getAbsolutePath(), instances,null, null);

      // call the garbage collector
      System.gc();
    }
    while (istance != null);
    if (InternalTableResourceFactory.isVertical(instances)) {

      Attribute id_trans = instances.attribute(TimestampInternalTable.TRANSACTION_ATTRIBUTE);
      Attribute item = instances.attribute(TransactionalInternalTable.EVENT_ATTRIBUTE);

      // controlla che non ci siano valori missing per l'attributo transaction
      AttributeStatisticManager a_trans = stat.getStatistic(TimestampInternalTable.TRANSACTION_ATTRIBUTE);
      if (a_trans.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for transaction attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);

      // controlla che non ci siano valori missing per l'attributo event
      AttributeStatisticManager a_event = stat.getStatistic(TimestampInternalTable.EVENT_ATTRIBUTE);
      if (a_event.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for event attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);


      if (InternalTableResourceFactory.hasTimestamAttribute(instances)) {
        // timestamp table
        Attribute timestamp = instances.attribute(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);

        AttributeStatisticManager a_timestamp = stat.getStatistic(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);
        if (a_timestamp.getMissingCount() > 0)
          throw new InternalTableException("Timestamp table do not admit missing values for timestamp attribute!",
                                           null,
                                           KDDMLCoreErrorCodes.INVALID_SETTINGS);

        return new TimestampInternalTable(physical_file, stat, id_trans.index(), item.index(), timestamp.index());
      }
      else {
        // tabella di tipo transazionale
        return new TransactionalInternalTable(physical_file, stat, id_trans.index(), item.index());
      }
    }
    else
      return new InternalTable(physical_file, stat);
  }

  /**
   * Set the logical name of this table.
   *
   * @param logical_name String
   */
  public void setLogicalName(String logical_name) {
    this.logical_name = logical_name;
  }

  /**
   * Set the maximum number of instances imported ad once from the input resource.
   * If null, all instances will be imported in a single iteration.
   *
   * @param number int
   */
  public void setMaxNumberOfLoadedInstAtIteration(int number) {
    max_number_of_inst_loaded_at_iteration = new Integer(number);
  }

  /**
   * Returns the next instance from the input resource.
   * Returns null if no more instances are available.
   *
   * @param schema Instances the data schema of the instances.
   * @throws KDDMLCoreException if an error occours.
   * @return Instance
   */
  abstract Instance nextInstance(Instances schema)
      throws kddml.Core.KDDMLCoreException;

  /**
   * Returns the relation name of this table.
   *
   * @return String
   */
  abstract String getRelationName();

  /**
   * Returns the instances header as fast vector of attributes.
   *
   * @throws KDDMLCoreException
   * @return FastVector
   */
  abstract FastVector getHeader()
      throws kddml.Core.KDDMLCoreException;
}
