/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.DataSources;

import kddml.Core.KDDMLCoreException;

/**
 * An interface for internal table.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public interface InternalTableManager {

  /**
   * Returns the statistics related to this table.
   *
   * @return InternalTableManager.
   */
  public DataStatisticsManager getStatistic();

  /**
   * Returns the table dimension in KBytes.
   *
   * @return int
   */
  public int getDimension();

  /**
   * Returns only the instances whose indexes belong to the input vector.
   *
   * @param index int[] the array of indexes.
   * @throws KDDMLCoreException if an error occours.
   * @return Object the output instances as generic object.
   */
  public java.lang.Object loadInstances(int[] index) throws KDDMLCoreException;

  /**
   * Returns true if the table is a transactional table with attributes transaction
   * and event. Returns false if the table is a relational table.
   *
   * @return boolean
   */
  public boolean isTransactionalTable();

  /**
   * Adds the input instances to the table.
   * Throws an exception if the data schema of the input table and the data schema of
   * the input instances do not coincide.
   *
   * @param instances Object
   * @throws InternalTableException if the schema do not coincide.
   */
  public void addInstances(Object instances)
      throws InternalTableException;

  /**
   * Tests if there are instances to load.
   *
   * @return boolean returns true if there are instances to load. Returns false if
   * the pointer is at the end of the data file.
   */
  public boolean hasNext();

  /**
   * Resets the pointer to the begin of data file.
   */
  public void reset();

  /**
   * Increase the pointer of the specified quantity.
   *
   * @param i int
   */
  public void increasePointer(int i);

  /**
   * Computes the optimal number of instances to load.
   * It can depend from the total number of instances and attributes and the quantity
   * of available system RAM.
   *
   * @return long
   */
  public long optimize();

  /**
   * Returns the total number of instances.
   *
   * @return int
   */
  public int getNumOfInstances();

  /**
   * Returns the total number of attributes.
   *
   * @return int
   */
  public int getNumOfAttributes();

  /**
   * Returns true if the table is a timestamp table with attributes transaction,
   * event and timestamp.
   *
   * @return boolean
   */
  public boolean isTimestampTable();

  /**
   * Loads i instances from the position specified by the pointer and returns they
   * as generic object. If the parameter is greater than the total number of instances,
   * it returns all instances. If i is equal to 0, it returns only the schema of
   * tuples.
   *
   * @param i long
   * @throws KDDMLCoreException
   * @return Object
   */
   public java.lang.Object load(long i)
      throws KDDMLCoreException;
}
