/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import java.io.*;
import kddml.Core.*;

/**
 * A factory class that import InternalTable objects from the physical data system
 * repository.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class InternalTableRepositoryFactory extends KDDMLObjectRepositoryFactory {

  protected File physical_data_file;

  protected Element kddml_table;

  /**
   * Constructor given the physical file refering the object.
   *
   * @param object File
   * @throws KDDMLCoreException
   */
  public InternalTableRepositoryFactory(File object)
      throws KDDMLCoreException {

    super(object);

    // effettua il parsing del documento XML contenente lo schema della tabella
    Document doc = null;
    try {
      doc = XMLUtility.parse(IOUtil.getInputStream
                              (physical_file.getAbsolutePath()),
                              new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_TABLE));
    }
    catch (IOException ex) {
      throw new InternalTableException("I/O Error on file: "+physical_file.toString(),
                                 null,
                                 KDDMLCoreErrorCodes.IO_ERROR);
    }
    Element kddml_object = (Element) doc.getDocumentElement();
    kddml_table = (Element) XMLUtility.getFirstChild(kddml_object);

    String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);

    // memorizza il riferimento al file contenente i dati
    physical_data_file = new File(physical_file.getParent() + File.separator + data_file_name);

    // controlla se esiste il file contenente i dati
    if (!physical_data_file.exists())
      throw new KDDMLCoreException(
        "Source file < "+physical_data_file.getAbsolutePath()+" > do not exist in data repository.",
        null,
        KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);
  }


  public kddml.Core.KDDMLObject newInstance()
      throws kddml.Core.KDDMLCoreException {

    Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_SCHEMA).item(0);
    DataStatistics stat = new DataStatistics(schema);

    // controlla se e' una tabella di tipo transazionale
    int id_trans = stat.getAttributeIndex(TransactionalInternalTable.TRANSACTION_ATTRIBUTE);
    int item = stat.getAttributeIndex(TransactionalInternalTable.EVENT_ATTRIBUTE);
    if ((id_trans != -1) && (item != -1)) {

      int id_timestamp = stat.getAttributeIndex(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);
      // controlla che non ci siano valori missing per l'attributo transaction
      AttributeStatisticManager a_trans = stat.getStatistic(TimestampInternalTable.TRANSACTION_ATTRIBUTE);
      if (a_trans.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for transaction attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);

      // controlla che non ci siano valori missing per l'attributo event
      AttributeStatisticManager a_event = stat.getStatistic(TimestampInternalTable.EVENT_ATTRIBUTE);
      if (a_event.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for event attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);

      if (id_timestamp != -1) {
        // timestamp table
        AttributeStatisticManager a_timestamp = stat.getStatistic(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);
        if (a_timestamp.getMissingCount() > 0)
          throw new InternalTableException(
              "Timestamp table do not admit missing values for timestamp attribute!",
              null,
              KDDMLCoreErrorCodes.INVALID_SETTINGS);

        return new TimestampInternalTable(physical_file, stat, id_trans, item,
                                          id_timestamp);
      }
      else
        // transactional table
        return new TransactionalInternalTable(physical_file, stat, id_trans, item);
    }
    else
      return new InternalTable(physical_file, stat);

  }

  /**
   * Rename an object
   *
   * @param new_name String the new name without xml extension
   * @return boolean
   */
  public boolean renameObject(String new_name) {

    boolean b = super.renameObject(new_name);

    if (!b)
      return false;

    File new_data_file = new File(physical_data_file.getParent() +
                                  File.separator +
                                  new_name +
                                  SpaceNames.DATA_EXTENSION);
    boolean b2 = physical_data_file.renameTo(new_data_file);
    if (b2) {
      physical_data_file = new_data_file;
    }
    return b2;
  }


  public boolean removeObject() {
    super.removeObject();
    return physical_data_file.delete();
  }

  public static void removeAll(File repository_dir) {
    KDDMLObjectRepositoryFactory.removeAll(repository_dir);
    File[] list = repository_dir.listFiles(new CSVFileFilter());
    for (int i=0;i<list.length; i++) {
      list[i].delete();
    }
  }

 /**
  *
  *
  * @param args String[]
  */
 public static void main(String args[]) {
    try {
      File f = new File
          ("C:/kddml/Repository/DataSources/KDDMLTable/weather.xml");
      InternalTableRepositoryFactory rep = new InternalTableRepositoryFactory(f);
      InternalTable t = (InternalTable) rep.newInstance();
      // importa le istanze
      weka.core.Instances inst = null;
      int i=0;
      while (t.hasNext()) {
        inst = (weka.core.Instances) t.load(100);
        i++;
        // work on instances here...
        System.out.println("Iteration "+i+". Imported: "+inst.numInstances()+" instances");
      }
      System.out.println("Total iterations: "+i);
      System.out.println("Total instances: "+t.getNumOfInstances());
    }
    catch(kddml.Core.KDDMLCoreException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }
}

class CSVFileFilter implements FilenameFilter {

  public boolean accept(File dir, String name) {
    String extension = name.substring(name.length()-3, name.length());
    if (extension.toLowerCase().equals("csv"))
      return true;
    else
      return false;
  }
}

