/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.Statement;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * This class represent a connection to a generic DBMS.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class JDBCConnection {

  protected String user;

  protected String password;

  protected String database_url;

  protected String driver;

  /**
   * Constructor given the user name, the password and the JDBC connection URI.
   *
   * @param user String
   * @param password String
   * @param database_url String
   * @throws InternalTableException
   */
  public JDBCConnection(String user, String password, String database_url)
      throws InternalTableException {
    this.user = user;
    this.password = password;
    this.database_url = database_url;
    if (!database_url.startsWith("jdbc:")) {
      String msg = "Can't connect to database: no input driver JDBC!!!";
      throw new InternalTableException(msg, null, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
    driver = JDBCDriver.getDriver(database_url);
    if (driver == null) {
      String msg = "Can't connect to database: no input driver avalaible!!!";
      throw new InternalTableException(msg, null, KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
  }

  /**
   * Establish a connection to DBMS.
   *
   * @throws InternalTableException
   * @return Statement
   */
  public Statement getConnection() throws InternalTableException {
    checkParameters();
    try {
      // load datasource driver and establish connection
      Class.forName(driver);
    }
    catch(ClassNotFoundException e) {
      String msg1 = "Connection with database is not avaible: problem with access driver.";
      String msg2 = "No valid driver: "+e.toString();
      throw new InternalTableException(msg1, msg2,
                                       kddml.Core.KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
    try {
      Connection connection = DriverManager.getConnection(database_url, user, password);
      Statement statement = connection.createStatement();
      return statement;
    }
    catch(Exception e) {
      String msg1 = "Connection with database is not avaible: \n"+e.toString();
      String msg2 = e.toString();
      throw new InternalTableException(msg1, msg2,
                                       kddml.Core.KDDMLCoreErrorCodes.JDBC_CONNECTION_FAILURE);
    }
  }

  /**
   * Checks the connection parameters.
   *
   * @throws InternalTableException
   */
  public void checkParameters() throws InternalTableException {
    // controllo dei parametri
    if (user == null)
      throw new InternalTableException(
          "Invalid Connection parameters: user name is null!", null,
          kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    if (password == null)
      throw new InternalTableException(
          "Invalid Connection parameters: password name is null!", null,
          kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    if (database_url == null)
      throw new InternalTableException(
          "Invalid Connection parameters: database url is null!", null,
          kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
  }
}
