/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import java.util.Hashtable;
import java.util.Enumeration;
import weka.core.Instances;
import org.w3c.dom.Element;

/**
 * This class contains statistical information about a single discrete attribute.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class NominalStatistic extends AttributeStatistic
  implements NominalStatisticManager {

  /**
   * Hashtable with the enumerated value and the cardinality.
   */
  protected Hashtable enumerated_values;

  /**
   * Proprietary constructor given the attribute name and an enumeration of
   * categories.
   *
   * @param name String
   * @param values Enumeration
   */
  protected NominalStatistic(String name, Enumeration values) {
    super(name);
    enumerated_values = new Hashtable();
    while(values.hasMoreElements()) {
      String s = (String) values.nextElement();
      enumerated_values.put(s, new Integer(0));
    }
    type = AttributeType.NOMINAL;
  }

  /**
   * Proprietary constructory given the XML element e the number of instances.
   *
   * @param elem Element
   * @param num_inst int
   * @throws KDDMLCoreException
   */
  protected NominalStatistic(Element elem, int num_inst)
      throws kddml.Core.KDDMLCoreException {
    super(elem, num_inst);
    enumerated_values = new Hashtable();
    type = AttributeType.NOMINAL;
    Element el = (Element) kddml.Core.XMLUtility.getFirstChild(elem);
    int j=0;
    // Inserisce i valori dell'attributo all'interno del vettore creato
    for(Element enumerato = (Element) el.getElementsByTagName(SpaceNames.ELEMENT_VALUE).item(j);
        enumerato!=null;
        enumerato = (Element) el.getElementsByTagName(SpaceNames.ELEMENT_VALUE).item(j)) {
      j++;
      String s = kddml.Core.XMLUtility.getAttribute(enumerato, SpaceNames.ATTRIBUTE_VALUE);
      enumerated_values.put(s, new Integer(enumerato.getAttribute(SpaceNames.ATTRIBUTE_CARD)));
    }
  }

  /**
   * Update the attribute statistics given a set of instances. The schema of instances
   * must be compatible with the attribute statistics.
   *
   * @param instances Instances
   */
  protected void updateStatistic(Instances instances) {
    super.updateStatistic(instances);
    int pos = instances.attribute(name).index();
    for (int i=0; i<instances.numInstances(); i++) {
      if (!instances.instance(i).isMissing(pos)) {
        String value = instances.instance(i).stringValue(pos);
        if (enumerated_values.containsKey(value)) {
          // the hashtable contains the input value as key...
          Integer cardinality = (Integer) enumerated_values.get(value);
          Integer new_cardinality = new Integer(cardinality.intValue() + 1);
          enumerated_values.remove(value);
          enumerated_values.put(value, new_cardinality);
        }
        else {
          // the hashtable do not contains the input value as key...
          enumerated_values.put(value, new Integer(1));
        }
      }
    }
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return Enumeration
   */
  public String toString() {
    String s1 = super.toString();
    String s = s1 + "Type: nominal - ";
    s = s + "Values = cardinality : "+enumerated_values.toString()+"\n";
    return s;
  }

  /**
   * Returns an enumeration of category values as strings.
   *
   * @return Enumeration
   */
  public java.util.Enumeration getValues() {
    return enumerated_values.keys();
  }

  /**
   * Returns the number of categories.
   *
   * @return int
   */
  public int getNumberOfValues() {
    return enumerated_values.size();
  }

  /**
   * Returns the frequency of a given category value.
   * Returns -1 if the input category do not exist.
   *
   * @param value String
   * @return int
   */
  public int getCardinality(String value) {
    if (enumerated_values.containsKey(value))
      return ((Integer) enumerated_values.get(value)).intValue();
    else
      return -1;
  }

  /**
   * Returns the frequency of a given category value as percentage.
   * Returns -1 if the input category do not exist.
   *
   * @param value String
   * @return double a number between 0 and 1.
   */
  public double getCardinalityPerc(String value) {
    if (enumerated_values.containsKey(value)) {
      int v = ((Integer) enumerated_values.get(value)).intValue();
      return (double) v / (double) num_instances;
    }
    else
      return -1;

  }

  protected Element toXML() {

    Element attr = super.toXML();
    kddml.Core.XMLUtility.setAttribute(attr, SpaceNames.ATTRIBUTE_TYPE, "nominal");
    Element nominal = kddml.Core.XMLUtility.doc.createElement(SpaceNames.ELEMENT_NOMINAL_DESCRIPTION);
    kddml.Core.XMLUtility.setAttribute(nominal,
                                        SpaceNames.ATTRIBUTE_NUM_VALUES,
                                        new Integer(enumerated_values.size()).toString());
    attr.appendChild(nominal);
    java.util.Enumeration attValues = getValues();
    while (attValues.hasMoreElements()) {
      String string = (String) attValues.nextElement();

      Element Valore = kddml.Core.XMLUtility.doc.createElement(SpaceNames.ELEMENT_VALUE);
      int cardinality = getCardinality(string);
      double cardinality_perc = getCardinalityPerc(string);
      kddml.Core.XMLUtility.setAttribute(Valore, SpaceNames.ATTRIBUTE_VALUE, string);
      kddml.Core.XMLUtility.setAttribute(Valore, SpaceNames.ATTRIBUTE_CARD, new Integer(cardinality).toString());
      String s2 = java.text.NumberFormat.getPercentInstance().format(cardinality_perc).toString();
      kddml.Core.XMLUtility.setAttribute(Valore, SpaceNames.ATTRIBUTE_CARD_PERC, s2);
      nominal.appendChild(Valore);
    }
    if (taxonomy != null) {
      // set the taxonomy element
      Element tax = getTaxonomyElement();
      attr.appendChild(tax);
    }

    return attr;
  }
}
