/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import weka.core.Instances;
import org.w3c.dom.Element;
import kddml.Core.XMLUtility;

/**
 * This class is used to represent statistics on numerical attributes.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class NumericalStatistic extends AttributeStatistic
    implements NumericalStatisticManager {

  /**
   * The sum of numeric values.
   */
  protected double total;

  /**
   * Used for the compute of variance.
   */
  protected double squared_value;

  /**
   * The minimum value of the attribute.
   */
  protected Double min_value;

  /**
   * The maximum value of the attribute
   */
  protected Double max_value;

  /**
   * Proprietary constructor given the attribute name.
   *
   * @param name String
   */
  protected NumericalStatistic(String name) {
    super(name);
    total = squared_value = 0;
    min_value = new Double(Double.MAX_VALUE);
    max_value = new Double(Double.MIN_VALUE);
    type = AttributeType.NUMERIC;
  }

  /**
   * Proprietary constructor given the XML element and the total number of instances.
   *
   * @param elem Element
   * @param num_instances int
   * @throws KDDMLCoreException
   */
  protected NumericalStatistic(Element elem, int num_instances)
      throws kddml.Core.KDDMLCoreException {
    super(elem, num_instances);
    type = AttributeType.NUMERIC;
    Element el = (Element) kddml.Core.XMLUtility.getFirstChild(elem);
    String s_sum = el.getAttribute(SpaceNames.ATTRIBUTE_SUM);
    String s_sumq = el.getAttribute(SpaceNames.ATTRIBUTE_SUMQ);
    String s_min = el.getAttribute(SpaceNames.ATTRIBUTE_MIN);
    String s_max = el.getAttribute(SpaceNames.ATTRIBUTE_MAX);
    total = squared_value = 0;
    min_value = new Double(Double.MAX_VALUE);
    max_value = new Double(Double.MIN_VALUE);
    if (s_sum != "")
      total = (new Double(s_sum)).doubleValue();
    if (s_sumq != "")
      squared_value = (new Double(s_sumq)).doubleValue();
    if (s_min != "")
      min_value = new Double(s_min);
    if (s_max != "")
      max_value = new Double(s_max);
  }

  /**
   * Proprietary constructor given the name of the attribute, the sum total,
   * the squared sum total, the minimum and the maximum values.
   *
   * @param name String
   * @param total double
   * @param total_sq double
   * @param min double
   * @param max double
   */
  protected NumericalStatistic(String name, double total, double total_sq,
                               double min, double max) {
    super(name);
    this.total = total;
    type = AttributeType.NUMERIC;
    this.squared_value = total_sq;
    this.min_value = new Double(min);
    this.max_value = new Double(max);
  }

  /**
   * Update the attribute statistics given a set of instances. The schema of instances
   * must be compatible with the attribute statistics.
   *
   * @param instances Instances
   */
  protected void updateStatistic(Instances instances) {
    super.updateStatistic(instances);
    int pos = instances.attribute(name).index();
    for (int i=0; i<instances.numInstances(); i++) {
      if (!instances.instance(i).isMissing(pos)) {
        double d = instances.instance(i).value(pos);
        total = total + d;
        squared_value = squared_value + (d * d);
      }
    }
    double min = instances.attributeStats(pos).numericStats.min;
    double max = instances.attributeStats(pos).numericStats.max;
    min_value = new Double(Math.min(min_value.doubleValue(), min));
    max_value = new Double(Math.max(max_value.doubleValue(), max));
  }

  /**
   * Returns the mean of the values.
   *
   * @return double
   */
  public double getMean() {
    double d1 = num_instances;
    double d = total;
    if(d1 <= 0.0D)
      return 0.0D;
    else
      return d / d1;
  }

  /**
   * Returns the variance of the values.
   *
   * @return double
   */
  public double getVariance() {
    double d = total;
    double d1 = squared_value;
    double d2 = num_instances;

    if(d2 <= 1.0D) {
      return 0.0D;
    }
    double d3 = (d1 - (d * d) / d2) / (d2 - 1.0D);
    if(d3 < 0.0D) {
      return 0.0D;
    } else {
      return d3;
    }
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s1 = super.toString();
    String s = s1 + "Type: numeric - Mean: "+getMean()+" - Variance: "+getVariance()+"\n";
    return s;
  }

  /**
   * Returns the minimum value.
   * Returns null if the number of instances is 0.
   *
   * @return Double
   */
  public Double getMinValue() {
    if (num_instances == 0)
      return null;
    return min_value;
  }

  /**
   * Returns the maximum value.
   * Returns null if the number of instances is 0.
   *
   * @return Double
   */
  public Double getMaxValue() {
    if (num_instances == 0)
      return null;
    return max_value;
  }

  protected Element toXML() {
    Element attr = super.toXML();
    XMLUtility.setAttribute(attr, SpaceNames.ATTRIBUTE_TYPE, "numeric");
    Element numeric = kddml.Core.XMLUtility.doc.createElement(SpaceNames.ELEMENT_NUMERIC_DESCRIPTION);
    attr.appendChild(numeric);
    XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_MEAN, new Double(getMean()).toString());
    XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_VARIANCE, new Double(getVariance()).toString());
    XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_SUM, new Double(total).toString());
    XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_SUMQ, new Double(squared_value).toString());
    if (getMinValue() != null)
      XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_MIN, getMinValue().toString());
    else
      XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_MIN, new Double(Double.MIN_VALUE).toString());

    if (getMaxValue() != null)
      XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_MAX, getMaxValue().toString());
    else
      XMLUtility.setAttribute(numeric, SpaceNames.ATTRIBUTE_MAX, new Double(Double.MAX_VALUE).toString());
    return attr;
  }
}
