/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import weka.core.Instances;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import weka.core.Instances;
import java.io.*;
import kddml.Core.*;

/**
 * This class contains a preprocessing table, i.e. a table used in the preprocessing
 * task. It is composed by:
 * <ol>
 * <li>the data schema, that includes attributes types and some simple statistics
 * on attribute values; </li>
 * <li>the physical data as a text file in comma separated format;</li>
 * <li>the physical metadata as a text file in comma separated format.
 * Preprocessing data including preprocessing informations, such as marks associated
 * to a physical instance value;</li>
 * <li>an history containing the list of preprocessing steps performed on table.</li>
 * </ol>
 * <br>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */
public class PPInternalTable extends InternalTable implements PPInternalTableManager {

  /**
   * The reference to the preprocessing file.
   */
  protected File object_PPdata_path;

  /**
   * The statistics on preprocessing informations.
   */
  protected DataStatistics PPstatistic;

  /**
   * The history.
   */
  protected History history;

  /**
   * Proprietary constructor.
   *
   * @param physical_path File
   * @param stat DataStatistics
   * @param PPstat DataStatistics
   * @param hist History
   */
  protected PPInternalTable(File physical_path,
                            DataStatistics stat,
                            DataStatistics PPstat,
                            History hist) {

    super(physical_path, stat);
    this.PPstatistic = PPstat;
    String s = physical_path.getName();
    s = s.substring(0, s.length()-4) + ".metadata"+SpaceNames.DATA_EXTENSION;
    object_PPdata_path = new File(physical_path.getParent() + File.separator + s);
    history = new History (hist);
  }

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType
   */
  public kddml.Core.KDDMLObjectType getType() {
    return KDDMLObjectType.INTERNAL_PPTABLE;
  }

  /**
   * Saves the object in the system repository.
   * The data schema is representend as XML document. The physical instances as
   * text file in comma separated format. The preprocessing instances as text
   * file in comma separated format.
   * The destination path is provided by the object_path variable for the data schema,
   * by the object_data_path variable for the row data and by the object_PPdata_path
   * variable for the preprocessing instances.
   * By definition, row data and preprocessing information are already stored
   * in the repository.
   *
   * @throws KDDMLCoreException
   */
  public void saveToRepository() throws kddml.Core.KDDMLCoreException {

    // salvataggio dei metadati
    if (statistic != null) {
      Document kddml_object = XMLUtility.initializeDocument(SpaceNames.
          ELEMENT_KDDML_OBJECT);
      Element table_el = XMLUtility.doc.createElement(SpaceNames.
          ELEMENT_KDDML_TABLE);
      XMLUtility.setAttribute(table_el, SpaceNames.ATTRIBUTE_DATA_FILE, object_data_path.getName());
      XMLUtility.setAttribute(table_el, SpaceNames.ATTRIBUTE_PPDATA_FILE, object_PPdata_path.getName());
      Element root = kddml_object.getDocumentElement();
      root.appendChild(table_el);
      Element schema = statistic.toXML(true);
      Element PPschema = PPstatistic.toXML(false);
      Element history_el = history.toXML();
      table_el.appendChild(schema);
      table_el.appendChild(PPschema);

      PPschema.appendChild(history_el);
      System.out.println(XMLUtility.toString(kddml_object));
      XMLUtility.toFile(object_path.getAbsolutePath(), kddml_object);
    }

    // data is in the repository
  }


  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String s;
    s = "Data Source: "+object_path.toString()+"\n";
    s = s + "----------------------------------\n";
    s = s + "Number of Instances: "+statistic.getNumInstances()+"\n";
    s = s + "Number of Attributes: "+statistic.getNumAttributes()+"\n";
    return s;
  }

  /**
   * Loads i instances from the position specified by the pointer and returns they
   * as generic object. If the parameter is greater than the total number of instances,
   * it returns all instances. If i is equal to 0, it returns only the schema of
   * tuples. It returns row data and preprocessing data also.
   *
   * @param i long
   * @throws DOMParserException
   * @throws KDDMLCoreException if no instances has been saved in the repository.
   * @return Object a two-dimensional array of weka.core.Instances. The first element
   * contains the row data; the second element contains the preprocessing data.
   */
  public java.lang.Object load(long i)
      throws DOMParserException, KDDMLCoreException {

    Instances[] ris = new Instances[2];

    if ((!object_path.exists()) || (!object_PPdata_path.exists()))
      throw new KDDMLCoreException("Can't import instances: save object in the repository before loading.",
                                   null,
                                   KDDMLCoreErrorCodes.IO_ERROR);

    if (i == ALL) {
      // importazione di tutte le istanze del file
      pointer = statistic.getNumInstances() +1;
      ris[0] = IOUtility.loadTable(object_path, object_data_path, null, null);
      ris[1] = IOUtility.loadMetaTable(object_path, object_PPdata_path, null, null);
    }
    else {
      // importazione di un blocco delle istanze
      long begin, end;
      begin = pointer;
      end = java.lang.Math.min(pointer + i - 1, statistic.getNumInstances());
      // posiziona pointer all'istanza successiva
      pointer = end + 1;
      ris[0] = IOUtility.loadTable(object_path, object_data_path,
                                   new Long(begin), new Long(end));
      ris[1] = IOUtility.loadMetaTable(object_path, object_PPdata_path,
                                       new Long(begin), new Long(end));
    }
    return ris;
  }

  /**
   * Returns only the instances whose indexes belong to the input vector.
   *
   * @param indexes int[] the array of indexes.
   * @throws KDDMLCoreException if indexes are out of range.
   * @return Object a two-dimensional array of weka.core.Instances. The first element
   * contains the row data; the second element contains the preprocessing data.
   */
   public java.lang.Object loadInstances(int[] indexes) throws KDDMLCoreException {

     Instances[] ris = new Instances[2];
     // ordina il vettore degli indici...
     java.util.Arrays.sort(indexes);

     if((indexes[indexes.length-1] > getNumOfInstances()-1) ||
        (indexes[0] < 0))
       throw new KDDMLCoreException("Can't import instances: instances index out of range.",
                                    null,
                                    KDDMLCoreErrorCodes.IO_ERROR);

     ris[0] = IOUtility.loadTable(object_path, object_data_path, indexes);
     ris[1] = IOUtility.loadMetaTable(object_path, object_PPdata_path, indexes);

     return ris;
   }

   /**
    * Adds the input instances to the table.
    * Throws an exception if the data schema of the input table and the data schema of
    * the input instances are not compatible.
    *
    * @param inst1 Object the set of instances representing the data as weka.core.Instances.
    * @param inst2 Object the set of instances representing the preprocessing data
    * as weka.core.Instances.
    * @throws InternalTableException if the schemas are not compatible.
    */
   public void addInstances(Object inst1, Object inst2)
       throws InternalTableException {

     // aggiorna le statistiche
     statistic.updateStatistic((Instances) inst1);
     PPstatistic.updateStatistic((Instances) inst2);

     // aggiunge al file fisico l'insieme di istanze.
     IOUtility.saveTable(object_path.getAbsolutePath(), (Instances) inst1, (Instances) inst2, null);

   }

   /**
    * Computes the optimal number of instances to load.
    * It can depend from the total number of instances and attributes and the quantity
    * of available system RAM.
    *
    * @return long
    */
   public long optimize() {

     long l = super.optimize();
     if (l == ALL)
       return ALL;
     // divido per due per considerare anche la tabella dei metadati
     return l / 2;
   }

   /**
    * Returns the history related to this preprocessing table.
    *
    * @return HistoryManager
    */
   public HistoryManager getHistory() {

     return history;
   }

   /**
    * Sets the history to this preprocessing table.
    *
    * @param elem HistoryManager
    * @throws KDDMLCoreException
    */
   public void setHistory (HistoryManager elem)
       throws kddml.Core.KDDMLCoreException {

     history = (History) elem;
   }

   /**
    * Memorizza l'insieme di istanze come documento HTML.
    *
    * @throws KDDMLCoreException
    */
   public void saveHTML() throws KDDMLCoreException {

     XSLResolver resolver = new XSLResolver(KDDMLObjectType.INTERNAL_TABLE);

     String s1 = this.getPhysicalObject().getAbsolutePath();
     String s2 = resolver.getXSLPath();
     String s3 = s1 + ".html";

     // effettua il parsing del documento XML contenente lo schema della tabella
     Document doc;
     try {
       doc = XMLUtility.parse(IOUtil.getInputStream(this.getPhysicalObject().getAbsolutePath()),
                               new DTDResolver(kddml.Core.KDDMLObjectType.INTERNAL_TABLE));
     }
     catch(IOException e) {
       throw new InternalTableException("I/O Error on file: "+this.getPhysicalObject().toString(),
                                        null,
                                        KDDMLCoreErrorCodes.IO_ERROR);
     }
     Element kddml_table = (Element) doc.getDocumentElement().getElementsByTagName(SpaceNames.ELEMENT_KDDML_TABLE).item(0);
     String data_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_DATA_FILE);
     String PPdata_file_name = kddml_table.getAttribute(SpaceNames.ATTRIBUTE_PPDATA_FILE);

     Element schema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_SCHEMA).item(0);
     Element PPschema = (Element) kddml_table.getElementsByTagName(SpaceNames.ELEMENT_PPSCHEMA).item(0);

     DataStatistics stat = new DataStatistics(schema);
     DataStatistics PPstat = new DataStatistics(PPschema);

     try {
       String data_ref = this.getPhysicalObject().getParent() + File.separator + data_file_name;
       String PPdata_ref = this.getPhysicalObject().getParent() + File.separator + PPdata_file_name;

       // generazione del .html che rappresenta lo schema
       String[] p1 = {"file_xml", "data_file","pp_data_file"};
       String[] p2 = {this.getPhysicalObject().getName(), data_file_name+".html",PPdata_file_name+".html"};
       XMLUtility.formatting(s1, s2, s3, p1, p2);

       // generazione del .html che contiene i dati
       String html = data_ref + ".html";
       if (!HTMLUtility.writeHTMLInstances(html, data_ref, stat)) {
         // le istanze non sono visualizzabili
         OutputStream os = new FileOutputStream(html);
         PrintWriter pw = new PrintWriter(os,true);
         HTMLUtility.openHTMLDataTable(pw);
         pw.print("No HTML visualization is available: too many instances or attributes.\n");
         pw.print("Please open the .csv file");
         HTMLUtility.closeHTMLDataTable(pw);
       }
       // generazione del .html che contiene i metadati
       String PPhtml = PPdata_ref + ".html";
       if (!HTMLUtility.writeHTMLInstances(PPhtml, PPdata_ref, PPstat)) {
         // le istanze non sono visualizzabili
         OutputStream os2 = new FileOutputStream(PPhtml);
         PrintWriter pw2 = new PrintWriter(os2,true);
         HTMLUtility.openHTMLDataTable(pw2);
         pw2.print("No HTML visualization is available: too many instances or attributes.\n");
         pw2.print("Please open the .csv file");
         HTMLUtility.closeHTMLDataTable(pw2);
       }
     }
     catch(Exception e) {
       String msg1 = "Exception found during the HTML code generation:\n";
       String msg2 = msg1 + e.getMessage();
       throw new InternalTableException(msg1, msg2, KDDMLCoreErrorCodes.HTML_TRANSFORMATION_ERROR);
     }
   }
 }
