/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataSources;

import kddml.Core.DataMining.SequentialPatterns.DataSequence;
import kddml.Core.DataMining.SequentialPatterns.DataSequenceManager;
import kddml.Core.DataMining.SequentialPatterns.TimestampedTransaction;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.AssociationRules.ItemFeature;
import kddml.Core.KDDMLCoreException;
import weka.core.*;
import java.util.Vector;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.io.*;

/**
 * This class is used to represent a table in a timestamp format.
 * Timestamp format is similar to a transactional table, but with an extra attribute timestamp.
 * This attribute defines a partial time order between transactions and items.
 * Typically, this format is used for sequential patterns analysis.
 * <br>
 * <p> A timestamp table contains at least three fields: </p>
 * <ol>
 * <li> the identifier of transaction (field transaction) </li>
 * <li> the identifier of timestamp (field timestamp) </li>
 * <li> the item value (field event) </li>
 * </ol>
 *
 * <p>Title: KDDML-MQL</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class TimestampInternalTable extends TransactionalInternalTable
    implements TimestampInternalTableManager {

  /**
   * The timestamp attribute name.
   */
  public final static String TIMESTAMP_ATTRIBUTE = "timestamp";

  /**
   * The position of transaction field in the data schema.
   */
  protected int transaction_pos;

  /**
   * The position of timestamp field in the data schema.
   */
  protected int timestamp_pos;

  /**
   * The position of event field in the data schema.
   */
  protected int event_pos;

  private BufferedReader reader;

  private String inst;

  private int transaction_pointer;

  /**
   * Attributes to be loaded as item features.
   */
  protected Vector item_features;

  /**
   * Proprietary constructor used only from factory.
   *
   * @param physical_path File
   * @param stat DataStatistics
   * @param attribute_transaction_pos int
   * @param attribute_event_pos int
   * @param attribute_timestamp_pos int
   */
  protected TimestampInternalTable(java.io.File physical_path,
                                   DataStatistics stat,
                                   int attribute_transaction_pos,
                                   int attribute_event_pos,
                                   int attribute_timestamp_pos) {

    super(physical_path, stat, attribute_transaction_pos, attribute_event_pos);

    this.transaction_pos = attribute_transaction_pos;
    this.timestamp_pos = attribute_timestamp_pos;
    this.event_pos = attribute_event_pos;

    // posiziona il puntatore alle transazione alla prima transazione
    transaction_pointer = 1;

    FileReader inputFile = null;
    try {
      inputFile = new FileReader(object_data_path.getAbsolutePath());
    }
    catch (FileNotFoundException ex) {
    }
    reader = new BufferedReader(inputFile);

    this.item_features = null;
  }

  /**
   * Returns true if the table is a timestamp table with attributes transaction,
   * event and timestamp.
   *
   * @return true if table is in timestmap format
   */
  public boolean isTimestampTable() {
    return true;
  }

  /**
    * Returns the timestamp field position in the data schema.
    *
    * @return timestamp position
    */
  public int getTimestampPosition() {
    return timestamp_pos;
  }

  /**
   * Return inited features, or null if features are not inited.
   * Each feature is returned as Weka Attribute object "as is" in table schema.
   *
   * @return inited features
   */
  public Enumeration getInitedFeatures(){
    if( this.item_features == null ){
      return null;
    }
    return this.item_features.elements();
  }

  /**
   * Returns the next data sequence of the timestamp table.
   *
   * @throws InternalTableException
   * @return a data sequence
   */
  public DataSequenceManager loadNextDataSequence()
      throws InternalTableException {

    try {

      if ( (!object_path.exists()) || (!object_data_path.exists())){
        throw new KDDMLCoreException(
            "Can't import transactions: save object in the repository before loading.",
            null,
            kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
      }

      if ( !hasNextDataSequence() ) {
        return null;
      }

      if (inst == null){
        // legge la prima riga del file
        inst = readNextLine();
      }

      String current_sequence_id = getSequenceId(inst);

      String first_sequence_id = current_sequence_id;

      Vector datasequence_transactions = new Vector();

      while ( (current_sequence_id.equals(first_sequence_id)) &&
              (inst != null)) {

        Vector transaction_items = new Vector();
        Item item = null;
        String item_value = null;

        String current_timestamp = getTimestamp(inst);
        String first_timestamp = current_timestamp;

        while( (current_sequence_id.equals(first_sequence_id)) &&
               (current_timestamp.equals(first_timestamp)) &&
               (inst != null) ) {

          item_value = getItem(inst);
          item = new Item(item_value);
          transaction_items.add(item);

          inst = readNextLine();
          if (inst != null){
            current_sequence_id = getSequenceId(inst);
            current_timestamp = getTimestamp(inst);
          }

        }

        Item[] items = (Item[]) transaction_items.toArray(new Item[] {});
        TimestampedTransaction transaction = new TimestampedTransaction(items,
            first_timestamp);
        datasequence_transactions.add(transaction);

      }

      if (inst == null){
        reader.close();
      }

      TimestampedTransaction[] transactions =
          (TimestampedTransaction[])
          datasequence_transactions.toArray(new TimestampedTransaction[] {});

      return new DataSequence( transactions );

    }
    catch(Exception e) {
      throw new InternalTableException("IO Error!", e.toString(),
                                       kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  /**
   * Returns the next data sequence of the timestamp table, including
   * optional attributes as item features.
   *
   * Before calling this method, some of initXxxFeatures() may be called;
   * otherwise this method call by default initAllFeatures() method, which
   * include all optional attributes in table.
   *
   * If no optional attribute are present in input table, objects returned
   * by this method are identically to those returned by loadNextDataSequence().
   *
   * @throws InternalTableException
   * @return a data sequence
   */
  public DataSequenceManager loadNextFeaturedDataSequence()
      throws InternalTableException {

    try {

      if ( (!object_path.exists()) || (!object_data_path.exists())){
        throw new KDDMLCoreException(
            "Can't import transactions: save object in the repository before loading.",
            null,
            kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
      }

      // if one of the initXXXFeatures() has not called
      // all optional attributes are loaded
      if (this.item_features == null){
        this.initAllFeatures();
      }

      if ( !hasNextDataSequence() ) {
        return null;
      }

      if (inst == null){
        // legge la prima riga del file
        inst = readNextLine();
      }

      String current_sequence_id = getSequenceId(inst);

      String first_sequence_id = current_sequence_id;

      Vector datasequence_transactions = new Vector();

      while ( (current_sequence_id.equals(first_sequence_id)) &&
              (inst != null)) {

        Vector transaction_items = new Vector();
        Item item = null;
        String item_value = null;
        String[] features_value = null;

        String current_timestamp = getTimestamp(inst);
        String first_timestamp = current_timestamp;

        while( (current_sequence_id.equals(first_sequence_id)) &&
               (current_timestamp.equals(first_timestamp)) &&
               (inst != null) ) {

          item_value = getItem(inst);
          item = new Item(item_value);
          // get all attributes value from an instance
          features_value = split_instance(inst);
          // get item features
          Enumeration enum2 = this.item_features.elements();
          while( enum2.hasMoreElements() ) {
            Attribute a = (Attribute) enum2.nextElement();
            // create a new feature
            ItemFeature ifeat = new ItemFeature(a.name(), a.isNumeric());
            // get feature value from current instance
            String ifeat_value = features_value[a.index()];
            // set value
            if( !ifeat_value.equals(SpaceNames.MISSING) ){
              if( a.isNumeric() )
                ifeat.setValue(Double.parseDouble(ifeat_value));
              else
                ifeat.setValue(ifeat_value);
            }
            // add the new feature
            item.addNewFeature(ifeat);
          }

          transaction_items.add(item);

          inst = readNextLine();
          if (inst != null){
            current_sequence_id = getSequenceId(inst);
            current_timestamp = getTimestamp(inst);
          }

        }

        Item[] items = (Item[]) transaction_items.toArray(new Item[] {});
        TimestampedTransaction transaction = new TimestampedTransaction(items,
            first_timestamp);
        datasequence_transactions.add(transaction);

      }

      if (inst == null){
        reader.close();
      }

      TimestampedTransaction[] transactions =
          (TimestampedTransaction[])
          datasequence_transactions.toArray( new TimestampedTransaction[]{});
      return new DataSequence( transactions );

    }
    catch(Exception e) {
      throw new InternalTableException("IO Error!", e.toString(),
                                       kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  /**
   * Returns true if the timestamp table has more data sequences.
   *
   * @return true if there are more transactions
   */
  public boolean hasNextDataSequence() {
    return transaction_pointer <= statistic.getNumInstances();
  }

  /**
   * This method is automatically called inside loadNextFeaturedDataSequence()
   * if no other init features method are called.
   *
   * Collect all optional attributes from input table.
   *
   * @throws KDDMLCoreException
   */
  public void initAllFeatures() throws KDDMLCoreException{
    this.initFeatures(null, ALL_FEATURES);
  }

  /**
    * Calling this method once, before first call to
    * loadNextFeaturedDatasequence(), affect optional attributes that will
    * be included as item feature; only numeric optional attributes are
    * included.
    *
    * @throws KDDMLCoreException
    */
  public void initNumericFeatures() throws KDDMLCoreException{
        this.initFeatures(null, NUMERIC_FEATURES);
  }

  /**
   * Specify wich optional attributes must be added as item features.
   * If "names" is null, then all optional attributes are added, else
   * only attributes in "names" are added.
   * "feature_type" allow to filter attributes by their type
   * (e.g. ALL_FEATURES, NUMERIC_FEATURES, STRING_FEATURES).
   *
   * @param names name of attributes to be added, or null for all attributes
   * @param feature_type type of attributes to be added
   * @throws KDDMLCoreException
   */
  public void initFeatures(String[] names, int feature_type) throws
      KDDMLCoreException {
    // load table schema
    Instances schema = (Instances)this.load(0);
    // init features
    this.item_features = new Vector();

    Enumeration enum2 = schema.enumerateAttributes();
    while (enum2.hasMoreElements()) {

      Attribute a = (Attribute) enum2.nextElement();
      // skip timestamp format attributes
      if (
          !a.name().equalsIgnoreCase(TimestampInternalTable.
                                     TRANSACTION_ATTRIBUTE) &&
          !a.name().equalsIgnoreCase(TimestampInternalTable.
                                     TIMESTAMP_ATTRIBUTE) &&
          !a.name().equalsIgnoreCase(TimestampInternalTable.
                                     EVENT_ATTRIBUTE)
          ) {

        boolean names_check = true;
        // filter attribute by name
        if (names != null) {
          names_check = false;
          for (int i = 0; i < names.length; i++) {
            if (names[i].equalsIgnoreCase(a.name())) {
              names_check = true;
              break;
            }
          }
        }

        // if attribute name is selected
        if (names_check) {
          // filter attribute by type
          switch (feature_type) {
            case NUMERIC_FEATURES:
              if (a.isNumeric())
                item_features.add(a);
              break;
            case STRING_FEATURES:
              if (!a.isNumeric())
                item_features.add(a);
              break;
            case ALL_FEATURES:
              item_features.add(a);
              break;
            default:
              throw new KDDMLCoreException(
                  "Invalid feature type <" + feature_type + ">",
                  null,
                  kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
          }

        }
      }

    }

  }

  // Read a line from input table file (an instance)
  private String readNextLine() throws IOException {
    String s = reader.readLine();
    transaction_pointer++;
    return s;
  }

  private String getSequenceId(String instance) throws InternalTableException{
    // posiziona il puntatore all'indice della transazione
    return getAttribute(instance, transaction_pos);
  }

  private String getTimestamp(String instance) throws InternalTableException{
    // posiziona il puntatore all'indice del timestamp
    return getAttribute(instance, timestamp_pos);
  }

  private String getItem(String instance) throws InternalTableException{
    // posiziona il puntatore all'indice della transazione
    return getAttribute(instance, event_pos);
  }

  // Get a generic attribute from an instance, by the given position
  private String getAttribute(String instance, int attribute_position) throws
      InternalTableException {
    StringTokenizer st = new StringTokenizer(instance, SpaceNames.SEPARATOR);
    // posiziona il puntatore all'indice dell'attributo
    if (st.countTokens() < attribute_position)
      throw new InternalTableException(
          "Instance (" + instance + ") does not contains attribute {" +
          attribute_position + "}",
          "", kddml.Core.KDDMLCoreErrorCodes.NULL_ARGUMENT);
    for (int i = 0; i < attribute_position; i++)
      st.nextToken();
    return st.nextToken();
  }

  // get all attribute from an instance as array
  // each array position contain the i-th attribute
  private String[] split_instance(String instance) throws
      InternalTableException {
    StringTokenizer st = new StringTokenizer(instance, SpaceNames.SEPARATOR);
    String[] attributes = new String[st.countTokens()];
    for( int i=0; st.hasMoreTokens(); i++ )
      attributes[i] = st.nextToken();
    return attributes;

  }

  /**
   * Only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    String arff = "C:/arff/80M_1feat.arff";
    //String arff = "C:/arff/web_log_timestamp.arff";
    //String arff = "C:/arff/coop_timestamp.arff";
    //String arff = "C:/arff/market_timestamp.arff";
    java.io.File f = new java.io.File("C:/arff/80M_1feat.xml");
    //java.io.File f = new java.io.File("C:/arff/web_log.xml");
    //java.io.File f = new java.io.File("C:/arff/coop.xml");
    //java.io.File f = new java.io.File("C:/arff/market.xml");
    ARFFFactory fact = null;

    //try {
      /*fact = new ARFFFactory(f, arff);
      fact.setOverwrite(true);
      InternalTable t = (InternalTable) fact.newInstance();
      t.saveToRepository();
      TimestampInternalTable table = (TimestampInternalTable) t;
      int i=0;
      while(table.hasNextDataSequence()) {
        DataSequenceManager ds = table.loadNextFeaturedDataSequence();
        java.util.Iterator it = ds.getTransactions();
        while( it.hasNext() ){
          TimestampedTransaction tt = (TimestampedTransaction)it.next();
          java.util.Iterator it2 = tt.getItems();
          while( it2.hasNext() ){
            Item item = (Item)it2.next();
            System.out.print((i+1)+", "+tt.getTimestamp()+", "+item.toString()+", ");
            java.util.Iterator it3 = item.getFeatures();
            while( it3.hasNext() ){
              ItemFeature itemf = (ItemFeature)it3.next();
              System.out.print(itemf.value()+", ");
            }
            System.out.println("");
          }

        }
        //System.out.println(ds.toString());
        i++;
      }
      System.out.println("Number of data-sequences: "+i);*/

  int x=0,j=0,i=0;
  try{
  for( x=1;x<2;x++){
    if( x == 0 )
      arff = "C:/arff/80M_1feat.arff";
    if( x == 1 )
      arff = "C:/arff/24M_4feat.arff";
    f = new java.io.File(arff + ".xml");

    for( j = 1; j < 2; j++) {
      InternalTable t = (InternalTable) new InternalTableRepositoryFactory(f).newInstance();
      TimestampInternalTable table = (TimestampInternalTable) t;

      System.out.println("Featured "+f);i=0;
      System.out.println(new java.util.Date(System.currentTimeMillis()));
      table.initFeatures(null, ALL_FEATURES);
      while (table.hasNextDataSequence()) {
        DataSequenceManager ds = table.loadNextFeaturedDataSequence();
        i++;
      }
      System.out.println(new java.util.Date(System.currentTimeMillis()));

      t = (InternalTable) new InternalTableRepositoryFactory(f).newInstance();
      table = (TimestampInternalTable) t;

      System.out.println("Non-Featured "+f);i = 0;
      System.out.println(new java.util.Date(System.currentTimeMillis()));
      while (table.hasNextDataSequence()) {
        DataSequenceManager ds = table.loadNextDataSequence();
        i++;
      }
      System.out.println(new java.util.Date(System.currentTimeMillis()));
    }
  }
    }
    catch (KDDMLCoreException ex) {
      System.out.println(x+" "+j+" "+i);
      System.out.println(ex.getDetailErrorMessage());
    }
  }
}
