/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.DataSources;

import kddml.Core.*;
import kddml.Core.DataMining.SequentialPatterns.DataSequenceManager;
import java.util.Enumeration;

/**
 * This class is used to represent a table in a timestamp format.
 * Timestamp format is similar to a transactional table, but with an extra attribute timestamp.
 * This attribute defines a partial time order between transactions and items.
 * Typically, this format is used for sequential patterns analysis.
 * <br>
 * <p> A timestamp table contains at least three fields: </p>
 * <ol>
 * <li> the identifier of transaction (field transaction) </li>
 * <li> the identifier of timestamp (field timestamp) </li>
 * <li> the item value (field event) </li>
 * </ol>
 *
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public interface TimestampInternalTableManager extends TransactionalInternalTableManager {

  /**
   * Load all optional attributes in table as item features
   */
  public final static int ALL_FEATURES = 0;

  /**
   * Load all numeric attributes in table as item features
   */
  public final static int NUMERIC_FEATURES = 1;

  /**
   * Load all non numeric attributes in table as item features
   */
  public final static int STRING_FEATURES = 2;

  /**
   * Returns the next data sequence of the timestamp table.
   *
   * @throws InternalTableException
   * @return a data sequence
   */
  public DataSequenceManager loadNextDataSequence()
      throws InternalTableException;

  /**
   * Returns the next data sequence of the timestamp table, including
   * optional attributes as item features.
   *
   * Before calling this method, some of initXxxFeatures() may be called to
   * filter optional attributes;
   * otherwise this method call by default initAllFeatures() method, which
   * include all optional attributes present in table.
   *
   * If no optional attribute are present in input table, objects returned
   * by this method are identically to those returned by loadNextDataSequence().
   *
   * @throws InternalTableException
   * @return a data sequence
   */
  public DataSequenceManager loadNextFeaturedDataSequence()
      throws InternalTableException;

  /**
   * Return inited features, or null if features are not inited.
   * Each feature is returned as Weka Attribute object "as is" in table schema.
   *
   * @return inited features
   */
  public Enumeration getInitedFeatures();

  /**
   * This method is automatically called inside loadNextFeaturedDataSequence()
   * if no other init features method are called.
   *
   * Collect all optional attributes from input table.
   *
   * @throws KDDMLCoreException
   */
  public void initAllFeatures() throws KDDMLCoreException;

  /**
    * Calling this method once, before first call to
    * loadNextFeaturedDatasequence(), affect optional attributes that will
    * be included as item feature; only numeric optional attributes are
    * included.
    *
    * @throws KDDMLCoreException
    */
   public void initNumericFeatures() throws KDDMLCoreException;

   /**
    * Specify wich optional attributes must be added as item features.
    * If "names" is null, then all optional attributes are added, else
    * only attributes in "names" are added.
    * "feature_type" allow to filter attributes by their type
    * (e.g. ALL_FEATURES, NUMERIC_FEATURES, STRING_FEATURES).
    *
    * @param names name of attributes to be added, or null for all attributes
    * @param feature_type type of attributes to be added
    * @throws KDDMLCoreException
    */
   public void initFeatures(String[] names, int feature_type) throws
       KDDMLCoreException;

  /**
   * Returns true if the timestamp table has more data sequences.
   *
   * @return true if there are more transactions
   */
  public boolean hasNextDataSequence();

  /**
   * Returns the timestamp field position in the data schema.
   *
   * @return timestamp position
   */
  public int getTimestampPosition();

}
