/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.AssociationRules.Item;
import java.util.Vector;
import java.util.StringTokenizer;
import java.io.*;

/**
 * This class is used to represent a table in a transactional format.
 * Sparse data is more effectively stored in multi-record case format.
 * Here, data that has a variable number of entries (items) from among many possible
 * can be stored more compactly where only the items present are stored in the table.
 * This representation is typically used for association rules.
 * <br>
 * <p> A multi record table contains at least two fields: </p>
 * <ol>
 * <li> the identifier of transaction (field transaction) </li>
 * <li> the item value (field event) </li>
 * </ol>
 *
 * <p>Title: KDDML-MQL</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class TransactionalInternalTable extends InternalTable
    implements TransactionalInternalTableManager {

  /**
   * The transaction attribute name.
   */
  public final static String TRANSACTION_ATTRIBUTE = "transaction";

  /**
   * The event attribute name.
   */
  public final static String EVENT_ATTRIBUTE = "event";

  private int transaction_pos;

  private int event_pos;

  private BufferedReader reader;

  private String inst;

  private int transaction_pointer;

  /**
   * Proprietary constructor used only from factory.
   *
   * @param physical_path File
   * @param stat DataStatistics
   * @param attribute_transaction_pos int
   * @param attribute_event_pos int
   */
  protected TransactionalInternalTable(java.io.File physical_path,
                                       DataStatistics stat,
                                       int attribute_transaction_pos,
                                       int attribute_event_pos) {
    super(physical_path, stat);
    transaction_pos = attribute_transaction_pos;
    event_pos = attribute_event_pos;

    // posiziona il puntatore alle transazione alla prima transazione
    transaction_pointer = 1;

    FileReader inputFile = null;
    try {
      inputFile = new FileReader(object_data_path.getAbsolutePath());
    }
    catch (FileNotFoundException ex) {
    }
    reader = new BufferedReader(inputFile);
  }

  /**
   * Returns true if the table is a transactional table with attributes transaction
   * and event. Returns false if the table is a relational table.
   *
   * @return boolean
   */
  public boolean isTransactionalTable() {
    return true;
  }

  /**
   * Returns the position of the field transaction in the data schema.
   *
   * @return int returns a positive integer.
   */
  public int getTransactionPosition() {
    return transaction_pos;
  }

  /**
   * Returns the position of the field event in the data schema
   *
   * @return int returns a positive integer.
   */
  public int getEventPosition() {
    return event_pos;
  }

  /**
   * Returns the next transaction of the transactional table.
   *
   * @throws InternalTableException
   * @return DataSequenceManager
   */
  public kddml.Core.DataMining.AssociationRules.TransactionManager loadNextTransaction()
      throws InternalTableException {

    try {

      if ( (!object_path.exists()) || (!object_data_path.exists()))
        throw new KDDMLCoreException(
            "Can't import transactions: save object in the repository before loading.",
            null,
            kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);

      if (hasNext()) {
        if (inst == null)
          // legge la prima riga del file
          inst = readNextLine();

        String current_transaction_id = getTransactionId(inst);
        String first_transaction_id = current_transaction_id;
        Vector ris = new Vector();
        String item="";

        while ( (current_transaction_id.equals(first_transaction_id)) &&
                (inst != null)) {
          item = getItem(inst);
          if (!item.equals(SpaceNames.MISSING))
            ris.add(item);
          inst = readNextLine();
          if (inst != null)
            current_transaction_id = getTransactionId(inst);
        }
        if (inst == null)
          reader.close();

        Item[] ris2 = new Item[ris.size()];
        for (int i=0; i<ris2.length; i++)
          ris2[i] = new Item((String) ris.get(i));
        return new kddml.Core.DataMining.AssociationRules.Transaction(ris2);
      }
      else
        return null;
    }
    catch(Exception e) {
      throw new InternalTableException("IO Error!", e.toString(),
                                       kddml.Core.KDDMLCoreErrorCodes.IO_ERROR);
    }
  }

  /**
   * Returns true if the transactional table has more transactions.
   *
   * @return boolean
   */
  public boolean hasNextTransaction() {
    return transaction_pointer <= statistic.getNumInstances();
  }

  private String readNextLine() throws IOException {
    String s = reader.readLine();
    transaction_pointer++;
    return s;
  }

  private String getTransactionId(String instance) {
    StringTokenizer st = new StringTokenizer(instance, SpaceNames.SEPARATOR);
    // posiziona il puntatore all'indice della transazione
    for (int i=0; i<transaction_pos;i++)
      st.nextToken();
    return st.nextToken();
  }

  private String getItem(String instance) {
    StringTokenizer st = new StringTokenizer(instance, SpaceNames.SEPARATOR);
    // posiziona il puntatore all'indice della transazione
    for (int i=0; i<event_pos;i++)
      st.nextToken();
    return st.nextToken();
  }

  /**
   * Only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    java.io.File f = new java.io.File("C:/kddml/pippo.xml");
    String arff = "C:/kddml/Repository/Data/ARFF/coop_transactional.arff";
    ARFFFactory fact = null;
    try {
      fact = new ARFFFactory(f, arff);
      fact.setOverwrite(true);
      InternalTable t = (InternalTable) fact.newInstance();
      t.saveToRepository();
      System.out.println(t.isTransactionalTable());
      int i=0;
      while(t.hasNext()) {
        ((TransactionalInternalTable) t).loadNextTransaction();
        i++;
      }
      System.out.println("Number of transactions: "+i);
    }
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
  }
}
