/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Core.DataSources;

import weka.core.*;
import java.io.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A factory class that create a table given a set of weka instances.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class WekaFactory extends InternalTableProprietaryFactory {

  private Instances instances;

  /**
   * Constructor given the physical file containing the object, and the set of
   * instances as weka.core.Instances.
   *
   * @param f File
   * @param instances Instances
   * @throws KDDMLCoreException
   */
  public WekaFactory(java.io.File f, Instances instances)
      throws kddml.Core.KDDMLCoreException {

    super(f);
    this.instances = instances;
  }

  /**
   * Creates an instance of this object.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public kddml.Core.KDDMLObject newInstance() throws kddml.Core.KDDMLCoreException {

    // genera eccezione se il file di dest esiste e non puo' essere sovrascritto
    if (existObject() & !overwrite)
      throw new InternalTableException(
        "Destination object < " + physical_file.toString() + " > already exist in data repository.",
        null,
        kddml.Core.KDDMLCoreErrorCodes.OBJECT_EXIST_IN_REPOSITORY);

    if (existObject() & overwrite)
      IOUtility.refreshDataFile(physical_file.getAbsolutePath());

    // inizializza le statistiche
    DataStatistics stat = new DataStatistics(instances);
    // salve le istanze importate su disco
    IOUtility.saveTable(physical_file.getAbsolutePath(), instances, null, null);

    if (InternalTableResourceFactory.isVertical(instances)) {
      Attribute id_trans = instances.attribute(TimestampInternalTable.TRANSACTION_ATTRIBUTE);
      Attribute item = instances.attribute(TransactionalInternalTable.EVENT_ATTRIBUTE);

      // controlla che non ci siano valori missing per l'attributo transaction
      AttributeStatisticManager a_trans = stat.getStatistic(TimestampInternalTable.TRANSACTION_ATTRIBUTE);
      if (a_trans.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for transaction attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);

      // controlla che non ci siano valori missing per l'attributo event
      AttributeStatisticManager a_event = stat.getStatistic(TimestampInternalTable.EVENT_ATTRIBUTE);
      if (a_event.getMissingCount() > 0)
        throw new InternalTableException("Transactional table do not admit missing values for event attribute!",
                                         null,
                                         KDDMLCoreErrorCodes.INVALID_SETTINGS);


      if (InternalTableResourceFactory.hasTimestamAttribute(instances)) {
        // timestamp table
        Attribute timestamp = instances.attribute(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);

        AttributeStatisticManager a_timestamp = stat.getStatistic(TimestampInternalTable.TIMESTAMP_ATTRIBUTE);
        if (a_timestamp.getMissingCount() > 0)
          throw new InternalTableException("Timestamp table do not admit missing values for timestamp attribute!",
                                           null,
                                           KDDMLCoreErrorCodes.INVALID_SETTINGS);

        return new TimestampInternalTable(physical_file, stat, id_trans.index(), item.index(), timestamp.index());
      }
      else {
        // tabella di tipo transazionale
        return new TransactionalInternalTable(physical_file, stat, id_trans.index(), item.index());
      }
    }
    else
      return new InternalTable(physical_file, stat);
  }

  /**
   * This is only for testing...
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    try {
      java.io.File f = new java.io.File(
      "C:/Documents and Settings/romei/kddml/repository/data/ARFF/weather.xml");
      String s = "C:/Documents and Settings/romei/kddml/repository/data/ARFF/weather.arff";
      Instances in = null;
      try {
        in = new Instances(new java.io.FileReader(s));
      }
      catch (IOException ex) {
      }

      WekaFactory fact = new WekaFactory(f, in);

      fact.setOverwrite(true);
      // crea l'istanza di InternalTable...
      InternalTable t = (InternalTable) fact.newInstance();
      // salvataggio della tabella nel repository...
      t.saveToRepository();
    }
    catch(kddml.Core.KDDMLCoreException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }
}
