/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import kddml.Core.*;
import org.w3c.dom.*;
import kddml.Core.*;
import java.util.Vector;

/**
 * This object is used to specify a data mining or preprocessing algorithm;
 * it is composed by the algorithm name identifying the algorithm and a list of parameter
 * specifications with parameter name and parameter value
 * (e.g., the minimum support for a rda miner algorithm).
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class KDDMLAlgorithm extends KDDMLObject {

  private final static String ALG_NAME_ATTR = "algorithm_name";

  private final static String PARAM_ELEMENT = "PARAM";

  private final static String PARAM_NAME_ATTR = "name";

  private final static String PARAM_VALUE_ATTR = "value";

  /**
   * The list of formal parameters.
   */
  Vector names;

  /**
   * The list of actual parameters as string.
   */
  Vector values;

  /**
   * The algorithm name.
   */
  String name;

  /**
   * Constructor given the algorthm name and the list of formal and actual parameters.
   *
   * @param alg_name String
   * @param formal_params Vector
   * @param actual_params Vector
   */
  public KDDMLAlgorithm(String alg_name, Vector formal_params, Vector actual_params) {

    super(null);
    names = formal_params;
    values = actual_params;
    name = alg_name;
  }

  /**
   * Constructor given the XML element.
   *
   * @param element Element
   * @throws KDDMLCoreException
   */
  public KDDMLAlgorithm(Element element) throws kddml.Core.KDDMLCoreException {

    super(null);

    if (element == null)
      throw new KDDMLCoreException("Algorithm element is null!", null,
                                   KDDMLCoreErrorCodes.INVALID_SETTINGS);

    // estraggo dall'elemento XML il nome dell'algoritmo ed i parametri
    names = new Vector();
    values = new Vector();
    name = element.getAttribute(ALG_NAME_ATTR);
    NodeList l = element.getElementsByTagName(PARAM_ELEMENT);
    for (int i=0; i<l.getLength(); i++) {
      Element param = (Element) l.item(i);
      names.add(param.getAttribute(PARAM_NAME_ATTR));
      values.add(XMLUtility.getAttribute(param, PARAM_VALUE_ATTR));
    }
  }

  /**
   * Returns an enumeration of actual parameters.
   * Actual parameters and formal parameters match.
   *
   * @return Enumeration
   */
  public java.util.Enumeration getActualParameters() {
    return values.elements();
  }

  /**
   * Returns an enumeration of formal parameters.
   * Actual parameters and formal parameters match.
   *
   * @return Enumeration
   */
  public java.util.Enumeration getFormalParameters() {
    return names.elements();
  }

  /**
   * Returns the algorithm name.
   *
   * @return String
   */
  public String getAlgorithmName() {
    return name;
  }

  /**
   * Returns the number of algorithm parameters.
   *
   * @return int a non-negative integer.
   */
  public int getNumParameters() {
    return names.size();
  }

  /**
   * Returns the type of the object.
   *
   * @return KDDMLObjectType ALGORITHM
   */
  public kddml.Core.KDDMLObjectType getType() {
    return KDDMLObjectType.ALGORITHM;
  }

  /**
   * Tests if the object is empty.
   *
   * @return boolean false
   */
  public boolean isEmpty() {
    return false;
  }

  public void saveToRepository() throws KDDMLCoreException {

    throw new KDDMLCoreException("Cannot save object in the repository!",
                                 null,
                                 KDDMLCoreErrorCodes.INVALID_SETTINGS);
  }

  /**
   * Returns a representation of this object as string.
   *
   * @return String
   */
  public String toString() {
    String ris = "Algorithm Name: "+name+"\n";
    for (int i=0; i<this.getNumParameters(); i++) {
      ris = ris + "Param "+i+". Formal Value: "+names.get(i)+" - Actual Value: "+values.get(i)+"\n";
    }
    return ris;
  }

  public void saveHTML()
      throws KDDMLCoreException {
  }
}
