/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import java.io.File;

/**
 * The Core package provides to higher levels a data/model access interface to manage
 * tables and mining models.
 * Accessing a table yields a Java object satisfying an interface InternalTableManager,
 * which abstracts sequential and random read/write access to the table rows, and
 * provides metadata (such as column types and preprocessing history) and statistics
 * on table columns.
 * Accessing a model yields a Java object satisfying a specific interface. Such
 * interfaces provide programmatic read/write access to the model contents,
 * e.g. AssociationModelManager includes methods for adding and removing rules and itemsets.
 * All valid KDDML objects extend this abstract class.
 * KDDML includes also algorithm specifications, conditions and expressions as language
 * objects.
 * In other terms, there is a subclass of KDDMLObject for each possible type returned by
 * a language operator.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class KDDMLObject
    implements java.io.Serializable, HTMLTranslator {

  /**
   * The physical path containing the object
   */
  protected java.io.File object_path;

  /**
   * Proprietary constructor given the file containing the object.
   *
   * @param object_path File the physical file containing the object.
   */
  protected KDDMLObject(java.io.File object_path) {
    this.object_path = object_path;
  }

  /**
   * Tests if the object is empty. This depends on the type of the object. For example,
   * it returns true if a table do not contain instances or if an association model
   * do not contain association rules.
   *
   * @return boolean
   */
  public abstract boolean isEmpty();

  /**
   * Save the object in the system repository.
   * The destination path is provided by the object_path variable.
   * Throws an exception if a I/O error occours.
   *
   * @throws KDDMLCoreException
   */
  public abstract void saveToRepository()
      throws KDDMLCoreException;

  /**
   * It returns the type of the object.
   *
   * @return KDDMLObjectType the type of this object
   */
  public abstract KDDMLObjectType getType();

  /**
   * It returns the physical file containing the object.
   *
   * @return File
   */
  public File getPhysicalObject() {
    return object_path;
  }
}
