/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import java.io.File;

/**
 * The common superclass of factory classes used to build KDDML objects.
 * The core layer includes wrapper modules for accessing and importing tables and
 * models from external sources and to store, retrieve and delete objects in the repositories.
 * Wrappers take care of translating the format of the data into the internal representation
 * (InternalTable or KDDMLModel). Automatic conversions are performed by the system on data and
 * metadata. Additional conversions can be forced by the preprocessing operators of
 * the KDDML language.
 * As far as data is concerned, there are wrappers around ARFF text files, serialized
 * Weka tables and RDBMS tables (accessed via JDBC). Concerning models, there
 * is a wrapper importing from PMML files5 and it is currently being implemented a
 * wrapper around serialized JDM (Java Data Mining) objects.
 * KDDML assumes a psysical repository for each object in which the model is stored.
 * XML is used as psysical representation of the object. <br>
 * A KDDMLObject factory can be related
 * <ol>
 * <li> to proprietary system repository, or </li>
 * <li> to external resource factory. </li>
 * </ol>
 * The first one is used to import an existent object from the system repository.
 * The second one is used to build a new object from an external psysical (or not) resource,
 * such as a text file.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class KDDMLObjectFactory {

  /**
   * The physical repository path.
   */
  protected File physical_file;

  /**
   * Constructor given the physical file containing the object.
   *
   * @param f File the file containing the reference to the physical object.
   * Can be null if the object is abstract.
   * @throws KDDMLCoreException an exception if the file do not refere an XML file.
   */
  public KDDMLObjectFactory(File f)
      throws KDDMLCoreException {

    if (f != null) {
      if (!f.getAbsolutePath().endsWith(".xml"))
        throw new KDDMLCoreException(
            "Input file < " + f.getAbsolutePath() +" > do not refere an XML file.",
            null,
            KDDMLCoreErrorCodes.INVALID_SETTINGS);

    }
    physical_file = f;
  }

  /**
   * Returns a new instance of this object.
   * Throws an exception if an error occours.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public abstract KDDMLObject newInstance() throws KDDMLCoreException;
}
