/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import java.io.*;

/**
 * The common superclass for system repository factory.
 * Loads an existent KDDML object from the system repository.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class KDDMLObjectRepositoryFactory extends KDDMLObjectFactory {

  /**
   * Constructor given the physical file.
   * The object must exist in the system repository.
   *
   * @param f File the XML reference to the object
   * @throws KDDMLCoreException if the XML object do not exist in the system repository.
   */
  public KDDMLObjectRepositoryFactory(File f)
      throws KDDMLCoreException {

    super(f);
    if (!physical_file.exists())
      throw new KDDMLCoreException(
        "Source file < "+physical_file.getAbsolutePath()+" > do not exist in the repository!",
        null,
        KDDMLCoreErrorCodes.OBJECT_NOT_EXIST_IN_REPOSITORY);

  }

  /**
   * Rename this object.
   *
   * @param new_name String the new name of the object.
   * @return boolean false if the renaming fails.
   */
  public boolean renameObject(String new_name) {

    if (new_name.length() > 4) {
      String extension = new_name.substring(new_name.length() - 4, new_name.length());
      if (!extension.equalsIgnoreCase(".xml"))
        new_name = new_name + ".xml";
    }
    File new_file = new File(physical_file.getParent()+ File.separator + new_name);

    boolean b = physical_file.renameTo(new_file);
    if (b) {
      physical_file = new_file;
    }
    return b;
  }

  /**
   * Remove this object from the repository.
   *
   * @return boolean false if the object cannot be removed.
   */
  public boolean removeObject() {
    return physical_file.delete();
  }

  /**
   * Returns the list of object names given the repository path.
   *
   * @param repository_dir File the file directory referring the repository.
   * @return String[]
   */
  public static String[] getObjectNames(File repository_dir) {
    String[] ris = repository_dir.list(new XMLFileFilter());
    return ris;
  }

  /**
   * Returns the list of object files given the repository path.
   *
   * @param repository_dir File the file directory referring the repository.
   * @return String[]
   */
  public static File[] getObjectFiles(File repository_dir) {
    File[] ris = repository_dir.listFiles(new XMLFileFilter());
    return ris;
  }

  /**
   * Remove all objects from the system repository.
   *
   * @param repository_dir File
   */
  public static void removeAll(File repository_dir) {
    File[] list = repository_dir.listFiles(new XMLFileFilter());
    for (int i=0;i<list.length; i++) {
      list[i].delete();
    }
  }
}

/**
 * An XML file filter.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
class XMLFileFilter implements FilenameFilter {

  public boolean accept(File dir, String name) {
    String extension = name.substring(name.length()-3, name.length());
    if (extension.toLowerCase().equals("xml"))
      return true;
    else
      return false;
  }
}

