/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import org.w3c.dom.*;

/**
 * This class represents a simple scalar value as a string or a number.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class KDDMLScalar extends KDDMLObject implements KDDMLScalarManager {

  public final static byte NUMERIC = 0;

  public final static byte STRING = 1;

  private Object result;

  private byte type;

  /**
   * Constructor.
   *
   * @param physical_path File the repository physical path containing the object.
   * @param expression String the scalar value.
   */
  public KDDMLScalar(java.io.File physical_path, String expression) {
    super(physical_path);
    try {
      result = new Double(expression);
      type = NUMERIC;
    }
    catch(NumberFormatException e2) {
      result = expression;
      type = STRING;
    }
  }

  /**
   * Returns the type of the object.
   *
   * @return KDDMLObjectType KDDML_SCALAR
   */
  public KDDMLObjectType getType() {
    return KDDMLObjectType.KDDML_SCALAR;
  }

  /**
   * Tests if the object is empty.
   *
   * @return boolean false.
   */
  public boolean isEmpty() {
    return false;
  }

  /**
   * Save the object in the system repository.
   * The destination path is provided by the object_path variable. If it is null,
   * the method do nothing.
   *
   * @throws KDDMLCoreException
   */
  public void saveToRepository()
      throws kddml.Core.KDDMLCoreException {

    if (this.object_path != null) {
      Document kddml_object = XMLUtility.initializeDocument("KDDML_OBJECT");
      Element el = XMLUtility.doc.createElement("KDDML_SCALAR");
      if (type == NUMERIC) {
        el.setAttribute("type", "numeric");
        el.setAttribute("value", ( (Double) result).toString());
      }
      else {
        el.setAttribute("type", "string");
        el.setAttribute("value", (String) result);
      }
      kddml_object.getDocumentElement().appendChild(el);

      //System.out.println(XMLUtility.toString(kddml_object));
      XMLUtility.toFile(object_path.getAbsolutePath(), kddml_object);
    }
  }

  /**
   * Returns a representation of this scalar as string.
   *
   * @return String
   */
  public String toString() {
    return result.toString();
  }

  /**
   * Save the object in the system repository as HTML document.
   *
   * @throws KDDMLCoreException
   */
  public void saveHTML()
      throws KDDMLCoreException {

    if (this.getPhysicalObject() != null) {
      XSLResolver resolver = new XSLResolver(KDDMLObjectType.KDDML_SCALAR);

      String s1 = getPhysicalObject().getAbsolutePath();
      String s2 = resolver.getXSLPath();
      String s3 = s1 + ".html";

      try {
        XMLUtility.formatting(s1, s2, s3, null, null);
      }
      catch (Exception e) {
        String msg = "Exception found during the HTML code generation.";
        msg = e.toString();
        throw new KDDMLCoreException(msg, msg, KDDMLCoreErrorCodes.XSL_ERROR);
      }
    }
  }

  /**
   * Returns the expression value.
   * Returns a Double if the method getExprType() returns 0.
   * Returns a String if the method getExprType() returns 1.
   *
   * @return Object
   */
  public Object getExprValue() {
    return result;
  }

  /**
   * Returns the expression type.
   * Returns 0 if the object is a number (i.e. an integer or a real);
   * return 1 if the object is a string.
   *
   * @return byte
   */
  public byte getExprType() {
    return type;
  }

  /**
   * Evaluate the object as expression. If the object is a string that contains
   * a math expression (as "3+3"), then the object value become a number ("6" in this case)
   * and the object type become numeric. If expression parsing returns an error,
   * then no actions will be performed.
   */
  public void evaluateAsExpression() {

    // parser per la valutazione dell'expressione
    org.nfunk.jep.JEP parser_expr = new org.nfunk.jep.JEP();
    parser_expr.addStandardFunctions();

    // evaluate the math expression
    parser_expr.parseExpression((String) this.result);
    if (!parser_expr.hasError()) {
      double result = parser_expr.getValue();
      this.result = new Double(result).toString();
      this.type = NUMERIC;
    }
    // parsing error: nothing!
  }

  /**
   * Evaluates the object as real. The method tries to evaluate the object as double
   * (as 4.5, -0.67).
   * If a casting error is found, then the method evaluates the object as string.
   */
  /*public void evaluateAsReal() {

    try {
      result = new Double(expression);
      type = NUMERIC;
    }
    catch(NumberFormatException e2) {
      result = expression;
      type = STRING;
    }
  }*/


  /**
   * Evaluate the object as expression. If the object is a string that contains
   * a math expression (as "3+3"), then the object value become a number ("6" in this case)
   * and the object type become numeric. If expression parsing returns an error,
   * then no actions will be performed.
   */
  /*public void evaluateAsMathExpression() {

    // parser per la valutazione dell'expressione
    org.nfunk.jep.JEP parser_expr = new org.nfunk.jep.JEP();

    parser_expr.addStandardFunctions();

    // evaluate the math expression
    parser_expr.parseExpression(expression);
    if (!parser_expr.hasError()) {
      double result = parser_expr.getValue();
      this.result = new Double(result).toString();
      this.type = NUMERIC;
    }
    // parsing error: do nothing!
  }*/

}
