/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import weka.core.*;
import kddml.Core.ProprietaryElement.*;
import kddml.Core.KDDMLCoreErrorCodes;

/**
 * A primitive condition used to evaluate boolean operators (such as not_equal) on table
 * attributes and/or constants.
 * Here, table attributes stand for both relational (or preprocessing) table columns
 * and model properties (e.g. the support of an association rule).
 * Table attributes can be referred using the special symbol "@" in the value attribute.
 * Each base condition can be expressed using the attribute op type, representing
 * the name of the used boolean operator.
 * Operators can be unary, binary or ternary expressed by the term1, term2, term3
 * fields.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class BaseCondition extends Condition {

  /**
   * The op_type attribute.
   */
  String op_type;

  /**
   * The first term. Can be null.
   */
  String term1 = null;

  /**
   * The second term. Can be null.
   */
  String term2 = null;

  /**
   * The third term. Can be null.
   */
  String term3 = null;

  /**
   * True if the first term is a constant.
   * False if the first term is an attribute table.
   */
  boolean term1_is_constant;

  /**
   * True if the second term is a constant.
   * False if the second term is an attribute table.
   */
  boolean term2_is_constant;

  /**
   * True if the third term is a constant.
   * False if the third term is an attribute table.
   */
  boolean term3_is_constant;

  /**
   * Constructor.
   *
   * @param op_type String the value of op_type attribute
   * @param term1 String the first term. Is null only if the operator is 0-ary.
   * @param term1_is_constant boolean true if the first term is a constant value.
   * False if the first term is a table attribute.
   * @param term2 String the second term. Is null only if the operator is unary.
   * @param term2_is_constant boolean true if the second term is a constant value.
   * False if the second term is a table attribute.
   * @param term3 String the third term. Is null only if the operator is binary.
   * @param term3_is_constant boolean true if the third term is a constant value.
   * False if the third term is a table attribute.
   */
  public BaseCondition(String op_type,
                       String term1, boolean term1_is_constant,
                       String term2, boolean term2_is_constant,
                       String term3, boolean term3_is_constant) {

    super();
    this.op_type = op_type;

    if ((term1 != null) && (!term1.equals("")))
      this.term1 = term1;
    if ((term2 != null) && (!term2.equals("")))
      this.term2 = term2;
    if ((term3 != null) && (!term3.equals("")))
      this.term3 = term3;
    this.term1_is_constant = term1_is_constant;
    this.term2_is_constant = term2_is_constant;
    this.term3_is_constant = term3_is_constant;
  }

  /**
   * Checks the static correctness of this condition.
   * In particular, for each primitive case of the condition, the method checks if
   * the terms refere to a legal type as defined in the language DTDs and the operation
   * type is supported.
   * Throws an exception if some base term do not match the KDDML language DTDs.
   * It checks if:
   * <ol>
   * <li> the operator op_type is supported; </li>
   * <li> the term attributes are not null for binary or ternary operators; </li>
   * <li> every term is compatible. </li>
   * </ol>
   *
   * @throws XMLSpecialElementException
   */
  public void checkConditionValidity() throws XMLSpecialElementException {

    String class_name = null;
    try {
      // istanziazione della classe ..._DESCR
      class_name = this.getClass().getName() + "_DESCR";
      BaseConditionDescr type_class;
      Class c = null;

      c = Class.forName(class_name);
      type_class = (BaseConditionDescr) c.newInstance();


      // controllare che l'operando sia supportato...
      if (!type_class.isSupported(op_type)) {
        String msg = "Unsupported condition operator: " + op_type;
        throw new XMLSpecialElementException(msg,
                                             null,
                                             KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
      }

      // controllare attributo unario, binario, ternario...
      if (type_class.getTermCardinality(op_type) == 1) {
        // unary operator
        if (term1 == null)  {
          String msg = "Operator " + op_type+" is unary. Term1 cannot be null";
          throw new
              XMLSpecialElementException(msg,
                                         null,
                                         kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
        }
      }

      if (type_class.getTermCardinality(op_type) == 2) {
        // unary operator
        if ((term1 == null) || (term2 == null))  {
          String msg = "Operator " + op_type+" is binary. Term1 and Term2 cannot be null";
          throw new
              XMLSpecialElementException(msg,
                                         null,
                                         kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
        }
      }

      if (type_class.getTermCardinality(op_type) == 3) {
        // unary operator
        if ((term1 == null) || (term2 == null) || (term2 == null))  {
          String msg = "Operator " + op_type+" is unary. Term1, Term2 and Term3 cannot be null";
          throw new
              XMLSpecialElementException(msg,
                                         null,
                                         kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
        }
      }

      // controllare singoli terms...
      type_class.checkTerm1(op_type, term1, term1_is_constant);
      type_class.checkTerm2(op_type, term2, term2_is_constant);
      type_class.checkTerm3(op_type, term3, term3_is_constant);

    }
    catch (XMLSpecialElementException ex1) {
      throw ex1;
    }
    catch (Exception ex) {
      String msg = "Class "+class_name+" not implemented!";
      throw new XMLSpecialElementException(msg, null, KDDMLCoreErrorCodes.NOT_YET_IMPLEMENTED);
    }
  }

  /**
   * Returns the class description for this base condition.
   *
   * @throws XMLSpecialElementException
   * @return BaseConditionDescr
   */
  public BaseConditionDescr getConditionDescription() throws XMLSpecialElementException {

    String class_name = null;
    try {
      // istanziazione della classe TYPE
      class_name = this.getClass().getName() + "_DESCR";
      BaseConditionDescr type_class;
      Class c = null;

      c = Class.forName("kddml.Core.ProprietaryElement.Condition." + class_name);
      type_class = (BaseConditionDescr) c.newInstance();

      return type_class;

    }
    catch (Exception ex) {
      String msg = "Class "+class_name+" not implemented!";
      throw new XMLSpecialElementException(msg, null, KDDMLCoreErrorCodes.NOT_YET_IMPLEMENTED);
    }
  }

  /**
   * Evaluates this condition on input istance.
   * Here, instance stand both for a relational tupla and a preprocessing tupla.
   * In the first case, the field preprocessing_instance is null.
   *
   * @param instance Instance the input instance for the relationa table. The input instance
   * is defined in the weka.Core.Instance class. Cannot be null.
   * @param meta_instance Instance Instance the input preprocessing instance. The input instance
   * is defined in the weka.Core.Instance class. Can be null.
   * @throws XMLSpecialElementException
   * @return boolean the evaluation of this condition.
   */
  public boolean evalCond(weka.core.Instance instance, weka.core.Instance meta_instance)
      throws XMLSpecialElementException {

    String value1_s = null;
    String value2_s = null;
    String value3_s = null;

    String s1;
    if (term1 != null) {
      if (!isMetaInstancesOperator())
        s1 = evaluateTerm(term1, term1_is_constant, instance, false);
      else
        s1 = evaluateTerm(term1, term1_is_constant, meta_instance, true);
      if (s1 == null) {
        // caso di attributo missing
        return evalMissingAttribute();
      }
      value1_s = s1;
    }

    String s2;
    if (term2 != null) {
      if (!isMetaInstancesOperator())
        s2 = evaluateTerm(term2, term2_is_constant, instance, false);
      else
        s2 = evaluateTerm(term2, term2_is_constant, meta_instance, true);
      if (s2 == null)

        // caso di attributo missing
        return evalMissingAttribute();
      value2_s = s2;
    }

    if (term3 != null) {
      String s3;
      if (!isMetaInstancesOperator())
        s3 = evaluateTerm(term3, term3_is_constant, instance, false);
      else
        s3 = evaluateTerm(term3, term3_is_constant, meta_instance, true);
      if (s3 == null) {
        // caso di attributo missing
        return evalMissingAttribute();
      }
      value3_s = s3;
    }

    // rimuove eventuali '' dalle stringhe
    if (value1_s != null) {
      if (value1_s.endsWith("'"))
        value1_s = value1_s.substring(0, value1_s.length() - 1);
      if (value1_s.startsWith("'"))
        value1_s = value1_s.substring(1, value1_s.length());
    }

    if (value2_s != null) {
      if (value2_s.endsWith("'"))
        value2_s = value2_s.substring(0, value2_s.length() - 1);
      if (value2_s.startsWith("'"))
        value2_s = value2_s.substring(1, value2_s.length());
    }

    if (value3_s != null) {
      if (value3_s.endsWith("'"))
        value3_s = value3_s.substring(0, value3_s.length() - 1);
      if (value2_s.startsWith("'"))
        value3_s = value3_s.substring(1, value3_s.length());
    }

    String[] sss = {value1_s, value2_s, value3_s};
    return evalSimpleCase(op_type, sss);
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the op_type is evaluated on the preprocessing data, including
   * preprocessing information such as marks associated to a physical instance value.
   *
   * @return boolean
   */
  abstract boolean isMetaInstancesOperator();

  public String toString() {
    return op_type+"("+term1+", "+term2+", "+term3+")";
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  abstract boolean evalMissingAttribute();

  /**
   * Evaluates a simple case expression.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  abstract boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException;


  static String evaluateTerm(String term, boolean term_is_constant, Instance instance, boolean meta)
      throws XMLSpecialElementException {

    if (!term_is_constant) {

      if (meta)
        term = term + "_metadata";

      // il primo termine identifica un attributo
      Attribute a = null;
      for (int i=0; i<instance.numAttributes(); i++) {
        if (instance.attribute(i).name().equalsIgnoreCase(term))
          a = instance.attribute(i);
        if (a != null)
          break;
      }
      // attributo non trovato nello schema --> si solleva errore
      if (a==null)
        throw new XMLSpecialElementException("Attribute <"+term+"> not found!",
                                                   null,
                                                   kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
      // l'attributo e' stato individuato --> ottiene il valore dall'istanza
      if (!instance.isMissing(a))
        // attributo non missing
        if (a.isNumeric())
          return instance.toString(a);
        else
          return instance.stringValue(a);
      else
        // attribute missing
        return null;
    }
    else {
      // il termine identifica una costante
      return term;
    }
  }

  /**
   * Evaluates the equal operator.
   * Returns true if v1 = v2.
   * It can be appled to numeric or string values.
   *
   * @param v1 String
   * @param v2 String
   * @return boolean
   */
  boolean evalEqual(String v1, String v2) {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() == d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      return v1.equalsIgnoreCase(v2);
    }
  }

  /**
   * Evaluates the not_equal operator.
   * Returns true if v1 != v2.
   * It can be appled to numeric or string values.
   *
   * @param v1 String
   * @param v2 String
   * @return boolean
   */
  boolean evalNotEqual(String v1, String v2) {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() != d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      return !v1.equalsIgnoreCase(v2);
    }
  }

  /**
   * Evaluates the greater operator.
   * Returns true if v1 > v2.
   * It can be appled to numeric values only.
   *
   * @param v1 String
   * @param v2 String
   * @throws XMLSpecialElementException
   * @return boolean
   */
  boolean evalGreater(String v1, String v2)
      throws XMLSpecialElementException {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() > d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      String msg = "Operator GREATER can be applied only to numeric fields!";
      throw new XMLSpecialElementException(msg, null,
                                             KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Evaluates the greater_or_equal operator.
   * Returns true if v1 >= v2.
   * It can be appled to numeric values only.
   *
   * @param v1 String
   * @param v2 String
   * @throws XMLSpecialElementException
   * @return boolean
   */
  boolean evalGreaterOrEqual(String v1, String v2)
      throws XMLSpecialElementException {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() >= d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      String msg = "Operator GREATER_OR_EQUAL can be applied only to numeric fields!";
      throw new XMLSpecialElementException(msg, null,
                                             KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Evaluates the less operator.
   * Returns true if v1 < v2.
   * It can be appled to numeric values only.
   *
   * @param v1 String
   * @param v2 String
   * @throws XMLSpecialElementException
   * @return boolean
   */
  boolean evalLess(String v1, String v2)
      throws XMLSpecialElementException {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() < d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      String msg = "Operator LESS can be applied only to numeric fields!";
      throw new XMLSpecialElementException(msg, null,
                                             KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Evaluates the less_or_equal operator.
   * Returns true if v1 <= v2.
   * It can be appled to numeric values only.
   *
   * @param v1 String
   * @param v2 String
   * @throws XMLSpecialElementException
   * @return boolean
   */
  boolean evalLessOrEqual(String v1, String v2)
      throws XMLSpecialElementException {
    try {
      Double d_v1 = new Double(v1);
      Double d_v2 = new Double(v2);
      return (d_v1.doubleValue() <= d_v2.doubleValue());
    }
    catch(NumberFormatException e) {
      String msg = "Operator LESS_OR_EQUAL can be applied only to numeric fields!";
      throw new XMLSpecialElementException(msg, null,
                                             KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
    }
  }
}
