/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * This class contains an enumeration of op_type base condition.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
abstract class BaseConditionDescr {

  /**
   * equal operator.
   */
  protected final static byte EQUAL = 0;

  /**
   * not_equal operator.
   */
  protected final static byte NOT_EQUAL = 1;

  /**
   * greater operator.
   */
  protected final static byte GREATER = 2;

  /**
   * greater_or_equal operator.
   */
  protected final static byte GREATER_OR_EQUAL = 3;

  /**
   * less operator.
   */
  protected final static byte LESS = 4;

  /**
   * less_or_equal operator.
   */
  protected final static byte LESS_OR_EQUAL = 5;

  /**
   * Proprietary constructor.
   */
  protected BaseConditionDescr() {
  }

  /**
   * Returns the list of legal values for the op_type attribute.
   *
   * @return String[]
   */
  public abstract String[] list();

  /**
   * Tests if an operator is supported by this base condition.
   *
   * @param op_type String
   * @return boolean
   */
  public boolean isSupported(String op_type) {

    String[] l = list();
    for (int i=0; i<l.length; i++)
      if (l[i].equalsIgnoreCase(op_type))
        return true;
    return false;
  }

  /**
   * Returns the operator code of the input op_type attribute.
   * Returns -1 if the operator is not supported.
   *
   * @param op_type String
   * @return byte
   */
  byte getOpCode(String op_type) {
    if (op_type.equalsIgnoreCase("equal"))
      return EQUAL;
    if (op_type.equalsIgnoreCase("not_equal"))
      return NOT_EQUAL;
    if (op_type.equalsIgnoreCase("greater"))
      return GREATER;
    if (op_type.equalsIgnoreCase("greater_or_equal"))
      return GREATER_OR_EQUAL;
    if (op_type.equalsIgnoreCase("less"))
      return LESS;
    if (op_type.equalsIgnoreCase("less_or_equal"))
      return LESS_OR_EQUAL;
    return -1;
  }

  /**
   * Checks the first term.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected abstract void checkTerm1(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException;

  /**
   * Checks the second term.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected abstract void checkTerm2(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException;

  /**
   * Checks the third term.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected abstract void checkTerm3(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException;

  /**
   * Returns the op_type cardinality.
   * Returns 0 if the input operator has not arguments.
   * Returns 1 if the input operator is unary.
   * Returns 2 if the input operator is binary.
   * Returns 3 if the input operatro has three arguments.
   *
   * @param op_type String
   * @return byte
   */
  public abstract byte getTermCardinality(String op_type);
}
