/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import kddml.Core.ProprietaryElement.*;

/**
 * This class represents a condition as used in the IF statement.
 * If statement allows for evaluation an XQuery expression, and, on the basis of the
 * boolean result (in non-strict semantics), a then branch or an else branch is
 * applied.
 * This condition, already contains the XQuery evaluation in the term attribute.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class IF_CONDITION extends BaseCondition {

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute. It can be equal, not_equal,
   * less, less_or_equal, greater, greater_or_equal.
   * @param term1 String the first term.
   * @param term1_is_constant boolean true
   * @param term2 String the second term value. It cannot be null.
   * @param term2_is_constant boolean false.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  public IF_CONDITION(String type,
                      String term1, boolean term1_is_constant,
                      String term2, boolean term2_is_constant,
                      String term3, boolean term3_is_constant) {

    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }

  /**
   * Evaluates a simple case expression.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  protected boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    RDA_FILTER_CONDITION_DESCR t = new RDA_FILTER_CONDITION_DESCR();
    byte b = t.getOpCode(op);

    switch (b) {
      case IF_CONDITION_DESCR.EQUAL:
        return evalEqual(values[0], values[1]);
      case IF_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(values[0], values[1]);
      case IF_CONDITION_DESCR.GREATER:
        return evalGreater(values[0], values[1]);
      case IF_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(values[0], values[1]);
      case IF_CONDITION_DESCR.LESS:
        return evalLess(values[0], values[1]);
      case IF_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(values[0], values[1]);

      default:
        String msg = "Unsupported condition operator < " + op +" >";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.
                                                   UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the op_type is evaluated on the preprocessing data, including
   * preprocessing information such as marks associated to a physical instance value.
   *
   * @return boolean
   */
  protected boolean isMetaInstancesOperator() {
    return false;
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  public boolean evalMissingAttribute() {
    return false;
  }
}
