/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * This class contains an enumeration of op_type base condition values for the
 * condition used in the EXPRESSION element.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class IF_EXPRESSION_CONDITION_DESCR extends BaseConditionDescr {

  /**
   * Returns true if the attribute value is missing.
   */
  protected final static byte IS_MISSING = 6;

  /**
   * Proprietary constructor.
   */
  protected IF_EXPRESSION_CONDITION_DESCR() {
    super();
  }

  /**
   * Returns the list of legal values for the op_type attribute.
   *
   * @return String[]
   */
  public String[] list() {

    String ris[] = {
        "equal", "not_equal", "greater", "greater_or_equal", "less", "less_or_equal",
        "is_missing"
    };

    return ris;
  }

  /**
   * Returns the operator code of the input op_type attribute.
   * Returns -1 if the operator is not supported.
   *
   * @param op_type String
   * @return byte
   */
  byte getOpCode(String op_type) {
    byte b = super.getOpCode(op_type);
    if (b != -1)
      return b;
    if (op_type.equalsIgnoreCase("is_missing"))
      return IS_MISSING;
    return -1;
  }

  /**
   * Returns the op_type cardinality.
   * Returns 0 if the input operator has not arguments.
   * Returns 1 if the input operator is unary.
   * Returns 2 if the input operator is binary.
   * Returns 3 if the input operatro has three arguments.
   *
   * @param op_type String
   * @return byte 2 for each legal operator.
   */
  public byte getTermCardinality(String op_type) {

    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        return 2;
      case IS_MISSING:
        return 1;
      default:
        return -1;
    }
  }

  /**
   * Checks the first term.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm1(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
        // the first term is an attribute...
        if (is_constant)
          msg = "Illegal condition operator: in "+op_type+" the first term must refere an attribute!";
        break;
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the first term is a numeric attribute...
        if (is_constant)
          msg = "Illegal condition operator: in "+op_type+" the first term must refere a numeric attribute!";
        break;
      case IS_MISSING:
        // first term attribute
        if (is_constant)
          msg = "Illegal condition operator: in "+op_type+" the first term must refere an attribute!";
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
  }

  /**
   * Checks the second term.
   *
   * @param op_type String
   * @param term2_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm2(String op_type, String term2_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
        break;
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the second term is a numeric attribute or a numeric constant
        if(is_constant) {
          try {
            new Double(term2_value);
          }
          catch(NumberFormatException e) {
            msg = "Illegal condition operator: in "+op_type+
                " the second term must refere a numeric attribute or a numeric constant!";
          }
        }
        break;
      case IS_MISSING:
        break;

    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
  }

  /**
   * Checks the third term.
   * No controls are needed.
   *
   * @param op_type String
   * @param term3_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm3(String op_type, String term3_value, boolean is_constant)
      throws XMLSpecialElementException {

  }
}
