/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import java.util.StringTokenizer;
import kddml.Core.ProprietaryElement.*;

/**
 * This condition can be used to mark remove rows of a preprocessing table.
 * Examples include:
 * <ol>
 * <li> to remove all instances with the attribute temperature less than 80; </li>
 * <li> to remove all instances where the attribute outlook is missing. </li>
 * <li> to remove all instances where the attribute temperature contains the mark
 * out_of_range. </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class PP_REMOVE_ROWS_CONDITION extends BaseCondition {

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute.
   * @param term1 String the first term.
   * @param term1_is_constant boolean.
   * @param term2 String the second term value.
   * @param term2_is_constant boolean.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  public PP_REMOVE_ROWS_CONDITION(String type,
                                  String term1, boolean term1_is_constant,
                                  String term2, boolean term2_is_constant,
                                  String term3, boolean term3_is_constant) {

    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the op_type is evaluated on the preprocessing data, including
   * preprocessing information such as marks associated to a physical instance value.
   *
   * @return boolean
   */
  boolean isMetaInstancesOperator() {

    PP_REMOVE_ROWS_CONDITION_DESCR t = new PP_REMOVE_ROWS_CONDITION_DESCR();
    byte b = t.getOpCode(op_type);

    if ((b == PP_REMOVE_ROWS_CONDITION_DESCR.HAS_MARK) ||
        (b == PP_REMOVE_ROWS_CONDITION_DESCR.HAS_EXCEPTION))
      return true;
    else
      return false;
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  boolean evalMissingAttribute() {
    PP_REMOVE_ROWS_CONDITION_DESCR t = new PP_REMOVE_ROWS_CONDITION_DESCR();
    byte b = t.getOpCode(op_type);
    if (b == PP_REMOVE_ROWS_CONDITION_DESCR.IS_MISSING)
      // condizione vera
      return true;
    else
      // condizione false
      return false;
  }

  /**
   * Evaluates a simple case expression.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    PP_REMOVE_ROWS_CONDITION_DESCR t = new PP_REMOVE_ROWS_CONDITION_DESCR();
    byte b = t.getOpCode(op);

    switch (b) {
      case PP_REMOVE_ROWS_CONDITION_DESCR.EQUAL:
        return evalEqual(values[0], values[1]);
      case PP_REMOVE_ROWS_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(values[0], values[1]);
      case PP_REMOVE_ROWS_CONDITION_DESCR.GREATER:
        return evalGreater(values[0], values[1]);
      case PP_REMOVE_ROWS_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(values[0], values[1]);
      case PP_REMOVE_ROWS_CONDITION_DESCR.LESS:
        return evalLess(values[0], values[1]);
      case PP_REMOVE_ROWS_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(values[0], values[1]);
      // HAS EXCEPTION
      case PP_REMOVE_ROWS_CONDITION_DESCR.HAS_EXCEPTION:
        return evalHasException(values[0], values[1]);
      // HAS MARK
      case PP_REMOVE_ROWS_CONDITION_DESCR.HAS_MARK:
        return evalHasMark(values[0], values[1]);
      // IS MISSING
      case PP_REMOVE_ROWS_CONDITION_DESCR.IS_MISSING:
        return false;
      default:
        String msg = "Unsupported condition operator < " + op +" >";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.MALFORMED_CONDITION);

    }
  }

  private boolean evalHasException(String value1, String value2) {
    // value1 contains the list of items
    // value2 contains the value to compare
    if (value2.indexOf(";") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }

    StringTokenizer st = new StringTokenizer(value1, ";");
    while (st.hasMoreElements()) {
      String s = st.nextToken().trim();
      if (s.equalsIgnoreCase("Exception: "+value2.trim()))
        return true;
    }
    return false;

  }

  private boolean evalHasMark(String value1, String value2) {
    // value1 contains the list of items
    // value2 contains the value to compare
    if (value2.indexOf(";") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }

    StringTokenizer st2 = new StringTokenizer(value1, ";");
    while (st2.hasMoreElements()) {
      String s = st2.nextToken().trim();
      if (s.equalsIgnoreCase(value2.trim())) {
        return true;
      }
    }
    return false;

  }
}
