/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import kddml.Core.ProprietaryElement.*;
import java.util.regex.*;

/**
 * This condition can be used to restricts the rows to be rewrite in the PP_REWRITING
 * operator.
 * In other terms, the condition regards restrictions on the rows to be rewritten and the corresponding
 * rewriting rule is applied only on instances on which the condition is satisfied. By using this strategy,
 * more rewriting algorithm can be used on a single record. Each condition provides a constraint
 * on these algorithms.
 * Examples include:
 * <ol>
 * <li> to rewrite only instances with the attribute temperature less than 80; </li>
 * <li> to rewrite only instances where the attribute outlook is missing. </li>
 * </ol>
 * A special operator group can be used in the condition specification. With the group
 * operator, given an attribute value previously matched with a regular expression, a positive
 * integer N representing the index of a sub-sequence (group) related to pattern matching, and
 * a string constant, it checks if the group N of the attribute value is equal to the input string. As
 * a matter of notation, sub-sequences can be identified in the regular expression by grouping
 * them within round parentheses.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class PP_REWRITING_CONDITION extends BaseCondition {

  /**
   * The regular expression to be used for @group operator.
   */
  private static String regular_expression;

  /**
   * The rewriting attribute to be used for @group operator.
   */
  private static int rewriting_attribute_position;

  /**
   * The value of the rewriting attribute for a given instance.
   */
  private String rewriting_attribute_value;

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute.
   * @param term1 String the first term.
   * @param term1_is_constant boolean.
   * @param term2 String the second term value.
   * @param term2_is_constant boolean.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  protected PP_REWRITING_CONDITION(String type,
                                   String term1, boolean term1_is_constant,
                                   String term2, boolean term2_is_constant,
                                   String term3, boolean term3_is_constant) {
    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }


  public boolean evalCond(weka.core.Instance instance, weka.core.Instance meta_instance)
      throws XMLSpecialElementException {

    rewriting_attribute_value = instance.stringValue(rewriting_attribute_position);
    return super.evalCond(instance, meta_instance);
  }

  /**
   * Sets the regular expression related to this condition.
   * The regular expression is used to evaluate the group condition.
   *
   * @param pattern String the regular expression.
   */
  public static void setRegularExpression(String pattern) {
    regular_expression = pattern;
  }

  /**
   * Sets the rewriting attribute position.
   * This information is used to evaluate the group condition.
   *
   * @param attr_pos int a positive integer representing the rewriting attribute position.
   */
  public static void setRewritingAttributePosition(int attr_pos) {
    rewriting_attribute_position = attr_pos;
  }

  /**
   * Evaluates a simple case expression.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    PP_REWRITING_CONDITION_DESCR t = new PP_REWRITING_CONDITION_DESCR();
    byte b = t.getOpCode(op);
    switch (b) {
      case PP_REWRITING_CONDITION_DESCR.EQUAL:
        return evalEqual(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.GREATER:
        return evalGreater(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.LESS:
        return evalLess(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(values[0], values[1]);
      case PP_REWRITING_CONDITION_DESCR.IS_MISSING:
        return false;
      case PP_REWRITING_CONDITION_DESCR.GROUP:
        return evalGroup(values[0], values[1]);
      default:
        String msg = "Unsupported condition operator < " + op +" >";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.MALFORMED_CONDITION);
    }
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the op_type is evaluated on the preprocessing data, including
   * preprocessing information such as marks associated to a physical instance value.
   *
   * @return boolean
   */
  boolean isMetaInstancesOperator() {
    return false;
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  boolean evalMissingAttribute() {
    PP_REWRITING_CONDITION_DESCR t = new PP_REWRITING_CONDITION_DESCR();
    byte b = t.getOpCode(op_type);
    if (b == PP_REWRITING_CONDITION_DESCR.IS_MISSING)
      // condizione vera
      return true;
    else
      // condizione false
      return false;
  }

  /**
   * Evaluates the group condition.
   * Given an attribute value previously matched with a regular expression, a positive
   * integer N representing the index of a sub-sequence (group) related to pattern matching, and
   * a string constant, it checks if the group N of the attribute value is equal to the input string. As
   * a matter of notation, sub-sequences can be identified in the regular expression by grouping
   * them within round parentheses.
   *
   * @param value1 String
   * @param value2 String
   * @throws XMLSpecialElementException
   * @return boolean
   */
  private boolean evalGroup(String value1, String value2)
      throws XMLSpecialElementException {
    Pattern p = Pattern.compile(regular_expression);
    Matcher m = p.matcher(rewriting_attribute_value);
    String value_group = getString(m, new Double(value1).intValue());
    if (value_group.equalsIgnoreCase(value2))
      return true;
    else
      return false;
  }

  private static String getString(Matcher m, int num_raggr)
      throws XMLSpecialElementException {

    return getGroupVal(m, num_raggr);
  }

  private static String getGroupVal(Matcher m, int num)
      throws IllegalStateException, IndexOutOfBoundsException, XMLSpecialElementException {

    boolean has_matches = m.find();
    String ris = null;
    if (has_matches) {
      if (num <= m.groupCount()) {
        ris = m.group(num);
      }
      else {
        String msg = "Matching group number < " + num + " > does not exist.";
        throw new XMLSpecialElementException(msg, null, KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
      }
      m.reset();
    }
    return ris;
  }
}
