/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import java.util.StringTokenizer;
import kddml.Core.ProprietaryElement.*;

/**
 * This condition can be used to filter a set of association rules.
 * The condition is expressed by using an AND/OR/NOT combination between a set of
 * primitive filters, concerning:
 * <ol>
 * <li> the support or the cardinality of the itemsets; </li>
 * <li> the support, the confidence or the cardinality of body/head of the association rules; </li>
 * <li> the single items contained in the itemset, body or head elements. </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class RDA_FILTER_CONDITION extends BaseCondition {

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute. It can be equal, not_equal,
   * less, less_or_equal, greater, greater_or_equal, is_in, is_not_in.
   * @param term1 String the first term. It can be itemset_support, rda_support,
   * rda_confidence, itemset_cardinality, body_cardinality, head_cardinality,
   * itemset, body, head.
   * @param term1_is_constant boolean true
   * @param term2 String the second term value. It cannot be null.
   * @param term2_is_constant boolean false.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  public RDA_FILTER_CONDITION(String type,
                              String term1, boolean term1_is_constant,
                              String term2, boolean term2_is_constant,
                              String term3, boolean term3_is_constant) {

    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }

  /**
   * Evaluates a simple case expression.
   * We suppose that body and head items are in a comma separated format.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    RDA_FILTER_CONDITION_DESCR t = new RDA_FILTER_CONDITION_DESCR();
    byte b = t.getOpCode(op);

    switch (b) {

      case RDA_FILTER_CONDITION_DESCR.EQUAL:
        return evalEqual(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.GREATER:
        return evalGreater(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.LESS:
        return evalLess(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(values[0], values[1]);

      case RDA_FILTER_CONDITION_DESCR.IS_IN:
        return evalIsIn(values[0], values[1]);
      case RDA_FILTER_CONDITION_DESCR.IS_NOT_IN:
        return evalIsNotIn(values[0], values[1]);
      default:
        String msg = "Unsupported condition operator < " + op +
            " > for body/head/itemset attributes.";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.
                                                   UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the operator is evaluated on a preprocessing instances also.
   * The last one is used for preprocessing operators.
   *
   * @return boolean
   */
  boolean isMetaInstancesOperator() {
    return false;
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  boolean evalMissingAttribute() {
    return true;
  }

  /**
   * Evaluates the is_in operator.
   * It checks if an item belongs to an itemset, or to either the body or the
   * head elements of an association rule (e.g. filter rules with the item milk
   * in the body).
   *
   * @param value1 String
   * @param value2 String
   * @return boolean
   */
  private boolean evalIsIn(String value1, String value2) {
    // value1 deve contenere la lista di items
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf(",") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    StringTokenizer st = new StringTokenizer(value1, ",");
    while (st.hasMoreElements()) {
      String s = st.nextToken().trim();
      if (s.equalsIgnoreCase(value2.trim()))
        return true;
    }
    return false;

  }

  /**
   * Evaluates the is_not_in operator.
   * It checks if an item does not belong to itemset or to either the body or the head
   * elements of an association rule.
   *
   * @param value1 String
   * @param value2 String
   * @return boolean
   */
  private boolean evalIsNotIn(String value1, String value2) {
    // value1 deve contenere la lista di items
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf(",") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    StringTokenizer st = new StringTokenizer(value1, ",");
    while (st.hasMoreElements()) {
      String s = st.nextToken().trim();
      if (s.equalsIgnoreCase(value2.trim()))
        return false;
    }
    return true;
  }
}
