/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * This class contains an enumeration of op_type base condition values for the
 * RDA_FILTER operator.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public final class RDA_FILTER_CONDITION_DESCR extends BaseConditionDescr {

  /**
   * checks if an item belongs to an itemset, or to either the body or the
   * head elements of an association rule (e.g. filter rules with the item milk
   * in the body).
   */
  protected final static byte IS_IN = 6;

  /**
   * checks if an item does not belong to itemset or to either the body or the head
   * elements of an association rule.
   */
  protected final static byte IS_NOT_IN = 7;

  /**
   * Proprietary constructor.
   */
  protected RDA_FILTER_CONDITION_DESCR() {
    super();
  }

  /**
   * Returns the list of legal values for the op_type attribute.
   *
   * @return String[]
   */
  public String[] list() {

    String ris[] = {
        "equal", "not_equal", "greater", "greater_or_equal", "less", "less_or_equal",
        "is_in", "is_not_in"
    };

    return ris;
  }

  /**
   * Returns the operator code of the input op_type attribute.
   * Returns -1 if the operator is not supported.
   *
   * @param op_type String
   * @return byte
   */
  byte getOpCode(String op_type) {
    byte b = super.getOpCode(op_type);
    if (b != -1)
      return b;
    if (op_type.equalsIgnoreCase("is_in"))
      return IS_IN;
    if (op_type.equalsIgnoreCase("is_not_in"))
      return IS_NOT_IN;
    return -1;
  }

  /**
   * Returns the op_type cardinality.
   * Returns 0 if the input operator has not arguments.
   * Returns 1 if the input operator is unary.
   * Returns 2 if the input operator is binary.
   * Returns 3 if the input operatro has three arguments.
   *
   * @param op_type String
   * @return byte 2 for each legal operator.
   */
  public byte getTermCardinality(String op_type) {

    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
      case IS_IN:
      case IS_NOT_IN:
        return 2;
      default:
        return -1;
    }
  }

  /**
   * Checks the first term.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm1(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the first term is a numeric attribute: @support, @confidence, @head_cardinality, @body_cardinality
        if (is_constant)
          msg = "Illegal condition operator: in "+op_type+" the first term must be one of: "
              + "@rda_support, @rda_confidence, @itemset_support, "+
                "@head_cardinality, @body_cardinality, @itemset_cardinality";
        else
        if ((!term1_value.equals("rda_support")) && (!term1_value.equals("rda_confidence")) &&
            (!term1_value.equals("body_cardinality")) && (!term1_value.equals("head_cardinality")) &&
            (!term1_value.equals("itemset_support")) && (!term1_value.equals("itemset_cardinality"))) {

          msg = "Illegal condition operator: in " + op_type +
              " the first term must be one of: "
              + "@rda_support, @rda_confidence, @itemset_support, " +
              "@head_cardinality, @body_cardinality, @itemset_cardinality";
        }

        break;
      case IS_IN:
      case IS_NOT_IN:
        // the first term is a @itemset, @body, @head
        if (is_constant)
          msg = "Illegal condition operator: in "+op_type+" the first term must refere a @body, @head, @itemset attribute: ";
        else
        if ((!term1_value.equals("body")) &&
            (!term1_value.equals("head")) &&
            (!term1_value.equals("itemset")))
          msg = "Illegal condition operator: in "+op_type+" the first term must refere a @body, @head, @itemset attribute: ";

        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);

  }

  /**
   * Checks the second term.
   *
   * @param op_type String
   * @param term2_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm2(String op_type, String term2_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:

        // the second term is a numeric constant
        if (!is_constant)
          msg = "Illegal condition operator: in "+op_type+" the second term must be a numeric constant!";
        else
          try {
            double d = new Double(term2_value).doubleValue();
            if ((term2_value.equals("rda_support")) ||
                (term2_value.equals("rda_confidence")) ||
                (term2_value.equals("itemset_support"))) {
              if ( (d > 1) || (d < 0)) // double in [0,1]
                msg = "Illegal condition operator: in " + op_type +
                    " the second term must be a real in [0,1]";
            }
            else // positive integer
            if (d < 1)
              msg = "Illegal condition operator: in " + op_type +
                  " the second term must be positive integer";
          }
          catch(NumberFormatException e) {
            msg = "Illegal condition operator: in "+op_type+
                " the second term must be a numeric constant!";
          }
        break;
      case IS_IN:
      case IS_NOT_IN:

          // the second term is a string constant
        if (!is_constant)
          msg = "Illegal condition operator: in "+op_type+" the second term must be a constant!";

        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
  }

  /**
   * Checks the third term. No checking is needed.
   *
   * @param op_type String
   * @param term3_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm3(String op_type, String term3_value, boolean is_constant)
      throws XMLSpecialElementException {

  }
}
