/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import kddml.Core.ProprietaryElement.*;

/**
 * This condition can be used to filter a set of sequential patterns according to an
 * aggregate constraint that is a constraint on an aggregate of items in a pattern, where
 * the aggregate function can be sum, average, max, min, standard deviation
 * (e.g., return all sequences in witch the average price of all items is over than 100$).
 * Values for aggregate of items of the sequence (e.g., the price of item in the example
 * above) are contained in the input time-stamp table.
 * An example includes: return all the sequences in witch the sum of quantities of all items
 * is equal to 10.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 1.0
 */
public class SEQUENCE_AGGREGATE_FILTER_CONDITION extends BaseCondition {

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute. It can be equal, not_equal,
   * less, less_or_equal, greater, greater_or_equal, is_in, is_not_in.
   * @param term1 String the first term. It can be itemset_support, rda_support,
   * rda_confidence, itemset_cardinality, body_cardinality, head_cardinality,
   * itemset, body, head.
   * @param term1_is_constant boolean true
   * @param term2 String the second term value. It cannot be null.
   * @param term2_is_constant boolean false.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  public SEQUENCE_AGGREGATE_FILTER_CONDITION(String type,
                                             String term1, boolean term1_is_constant,
                                             String term2, boolean term2_is_constant,
                                             String term3, boolean term3_is_constant) {
    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }


  /**
   * Evaluates a simple case expression.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  protected boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    RDA_FILTER_CONDITION_DESCR t = new RDA_FILTER_CONDITION_DESCR();
    byte b = t.getOpCode(op);

    String d_array = values[1];
    String value = values[2];
    String function = values[0];

    double[] att_values = toArray(d_array);
    double res = f(function.trim(), att_values);
    String result = String.valueOf(res);

    switch (b) {
      //
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.EQUAL:
        return evalEqual(result, value);
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(result, value);
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.GREATER:
        return evalGreater(result, value);
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(result, value);
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.LESS:
        return evalLess(result, value);
      case SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(result, value);
      default:
        String msg = "Unsupported condition operator < " + op +" >.";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.
                                                   UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the operator is evaluated on a preprocessing instances also.
   * The last one is used for preprocessing operators.
   *
   * @return boolean
   */
  protected boolean isMetaInstancesOperator() {
    return false;
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  public boolean evalMissingAttribute() {
    return true;
  }

  /**
   * Transforms a string representation an array of double.
   * String format is csv.
   *
   * @param str String
   * @return double[]
   */
  private double[] toArray(String str){
    java.util.StringTokenizer tokenizer = new java.util.StringTokenizer(str, ",");
    double[] v = new double[tokenizer.countTokens()];

    for( int i=0; tokenizer.hasMoreTokens(); i++ )
      v[i] = Double.parseDouble(tokenizer.nextToken());

    return v;
  }

  /**
   * Calculates aggregate function on given value
   *
   * @param function aggregate function
   * @param values input values
   * @return result
   */
  private double f(String function, double[] values){
    double result = 0;
    if( function.equals(SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.AVG) )
      result = avg(values);
    else if( function.equals(SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.MAX) )
      result = max(values);
    else if( function.equals(SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.MIN) )
      result = min(values);
    else if( function.equals(SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.STDDEV) )
      result = stddev(values);
    else if( function.equals(SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR.SUM) )
      result = sum(values);
    return result;
  }

  private double sum(double[] values){
    double result = 0;

    for( int i=0; i<values.length; i++ )
      result += values[i];

    return result;
  }

  private double avg(double[] values){
    double result = 0;
    double num_items = values.length;

    for( int i=0; i<values.length; i++ )
      result += values[i];

    return (double)(result / num_items);
  }

  private double max(double[] values){
    double max = Double.MIN_VALUE;

    for( int i=0; i<values.length; i++ )
        if( values[i] > max ) max = values[i];

    return max;
  }

  private double min(double[] values){
    double min = Double.MAX_VALUE;

    for( int i=0; i<values.length; i++ )
        if( values[i] < min ) min = values[i];

    return min;
  }

  private double stddev(double[] values){
    double result = 0;
    double mean = avg(values);
    double num_items = values.length;

    for( int i=0; i<values.length; i++ )
      result += Math.pow(values[i] - mean, 2);

    return Math.sqrt( (double)(result / num_items) );
  }
}
