/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * This class contains an enumeration of op_type base condition values for the
 * SEQUENCE_AGGREGARE_FILTER operator.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 1.0
 */
public final class SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR extends BaseConditionDescr {

  protected final static String AVG = "avg";
  protected final static String SUM = "sum";
  protected final static String MAX = "max";
  protected final static String MIN = "min";
  protected final static String STDDEV = "stddev";

  private final static String FUNCTION_VALUES_LIST =
      AVG+", "+SUM+", "+MAX+", "+MIN+", "+STDDEV+", ";


  /**
   * Proprietary constructor.
   */
  protected SEQUENCE_AGGREGATE_FILTER_CONDITION_DESCR() {
    super();
  }

  /**
   * Returns the list of legal values for the op_type attribute.
   *
   * @return String[]
   */
  public String[] list() {

    String ris[] = {
        "equal", "not_equal", "greater", "greater_or_equal", "less", "less_or_equal"
    };

    return ris;
  }

  /**
   * Returns the operator code of the input op_type attribute.
   * Returns -1 if the operator is not supported.
   *
   * @param op_type String
   * @return byte
   */
  byte getOpCode(String op_type) {
    return super.getOpCode(op_type);
  }

  /**
   * Returns the op_type cardinality.
   * Returns 0 if the input operator has not arguments.
   * Returns 1 if the input operator is unary.
   * Returns 2 if the input operator is binary.
   * Returns 3 if the input operatro has three arguments.
   *
   * @param op_type String
   * @return byte 2 for each legal operator.
   */
  public byte getTermCardinality(String op_type) {

    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        return 3;
      default:
        return -1;
    }
  }

  /**
   * Checks the first term.
   * The first term must be an aggregate function constant such as sum, min, max, avg
   * or stddev.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm1(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the first term is a function constant
        if (!is_constant){
          msg = "Illegal condition operator: in " + op_type +
              " the first term must refere an aggregate function: " +
              FUNCTION_VALUES_LIST;
        }
        else {
          if( !checkFunction(term1_value) ){
            msg = "Illegal condition operator: in " + op_type +
                " the first term must refere an aggregate function: " +
                FUNCTION_VALUES_LIST;
          }
        }
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);

  }

  /**
   * Checks the second term.
   * The second term is a numeric attribute.
   *
   * @param op_type String
   * @param term2_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm2(String op_type, String term2_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the second term is a numeric attribute:
        if (is_constant)
          msg = "Illegal condition operator: in " + op_type +
              " the second term must refere a numeric attribute!";
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
  }

  /**
   * Checks the third term.
   * The third term is a numeric constant.
   *
   * @param op_type String
   * @param term3_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm3(String op_type, String term3_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the third term is a numeric constant
        if (!is_constant){
          msg = "Illegal condition operator: in " + op_type +
              " the third term must be a numeric constant!";
          break;
        }
        else {
          try {
            new Double(term3_value);
          }
          catch (NumberFormatException e) {
            msg = "Illegal condition operator: in " + op_type +
                " the third term must be a numeric constant!";
          }
        }
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);
  }

  private boolean checkFunction(String function_name){
    return ( FUNCTION_VALUES_LIST.indexOf(function_name+",") > -1 );
  }
}
