/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.KDDMLCoreErrorCodes;
import kddml.Core.ProprietaryElement.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataMining.AssociationRules.Item;
import java.util.Iterator;
import java.util.StringTokenizer;

/**
 * This condition can be used to filter a set of sequential patterns.
 * The condition is expressed by using an AND/OR/NOT combination between a set of
 * primitive filters, concerning:
 * <ol>
 * <li> support constraint that specifies a requirement on the support of the sequence
 * (e.g. to filter sequences with a support less than 1); </li>
 * <li> item constraint that specifies what are the particular individual or groups of
 * items that should or should not be present in the pattern (e.g. to filter sequences
 * with the item "milk" in the sequence); </li>
 * <li> length constraint used to specifY the requirement on the length of the patterns,
 * where the length can be either the number of distinct items, or the number of sets,
 * or the maximal number of items per set (e.g. to filter sequences with exactly 3 distinct items);</li>
 * <li>  super-pattern constraint finds patterns that contain a particular list of sets
 * as sub-pattern (e.g. to filter sequences that contains first the item "spaghetti" and then
 * the item "milk"). </li>
 * </ol>
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 1.0
 */
public final class SEQUENCE_FILTER_CONDITION extends BaseCondition {

  /**
   * Constructor.
   *
   * @param type String the value of op_type attribute. It can be equal, not_equal,
   * less, less_or_equal, greater, greater_or_equal, is_in, is_not_in.
   * @param term1 String the first term. It can be: sequence_support, sequence_cardinality,
   * sequence_distinct_items_cardinality, max_number_of_items_per_set, sequence.
   * @param term1_is_constant boolean true
   * @param term2 String the second term value. It cannot be null.
   * @param term2_is_constant boolean false.
   * @param term3 String null.
   * @param term3_is_constant boolean false.
   */
  public SEQUENCE_FILTER_CONDITION(String type,
                                   String term1, boolean term1_is_constant,
                                   String term2, boolean term2_is_constant,
                                   String term3, boolean term3_is_constant) {
    super(type, term1, term1_is_constant, term2, term2_is_constant, term3, term3_is_constant);
  }

  /**
   * Evaluates a simple case expression.
   * We suppose that the elements of the sequence are in the following form:
   * <i1,...,in>(d)<i1,...im>.
   *
   * @param op String the operator type.
   * @param values String[] the term values. The size of the array is equal to the cardinality
   * of the op_type. If the value referes to a missing table attribute, then the
   * array value is null.
   * @throws XMLSpecialElementException
   * @return boolean the condition evaluation.
   */
  protected boolean evalSimpleCase(String op, String[] values)
      throws XMLSpecialElementException {

    SEQUENCE_FILTER_CONDITION_DESCR t = new SEQUENCE_FILTER_CONDITION_DESCR();
    byte b = t.getOpCode(op);

    switch (b) {

      case SEQUENCE_FILTER_CONDITION_DESCR.EQUAL:
        return evalEqual(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.NOT_EQUAL:
        return evalNotEqual(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.GREATER:
        return evalGreater(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.GREATER_OR_EQUAL:
        return evalGreaterOrEqual(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.LESS:
        return evalLess(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.LESS_OR_EQUAL:
        return evalLessOrEqual(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.IS_IN:
        return evalIsIn(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.IS_IN_ALL:
        return evalIsInAll(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.IS_NOT_IN:
        return evalIsNotIn(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.IS_NOT_IN_ALL:
        return evalIsNotInAll(values[0], values[1]);
      case SEQUENCE_FILTER_CONDITION_DESCR.SUB_SEQUENCE:
        return evalSubSequence(values[0], values[1]);
      default:
        String msg = "Unsupported condition operator < " + op +
            " >.";
        throw new XMLSpecialElementException(msg, null,
                                                   KDDMLCoreErrorCodes.
                                                   UNSUPPORTED_CONDITION_OPERATOR);
    }
  }

  /**
   * Check if an item belong to one elements of given sequence
   *
   * @param value1 a sequence in internal string format
   * @param value2 item to be checked
   * @return boolean
   */
  private boolean evalIsIn(String value1, String value2) {
    // value1 deve contenere la lista di sequence element
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf("<") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    SequenceManager sm = Sequence.fromInternalFormat(value1);
    Iterator iter;

    // exists
    iter = sm.getElements();
    while (iter.hasNext()) {
      SequenceElementManager seq_elem = (SequenceElementManager) iter.next();
      if (seq_elem.contain(new Item(value2)))
        return true;
      if( iter.hasNext() ) iter.next();//skip delimiter
    }
    return false;

  }

  /**
   * Check if an item belong to all elements of given sequence
   *
   * @param value1 a sequence in internal string format
   * @param value2 item to be checked
   * @return boolean
   */
  private boolean evalIsInAll(String value1, String value2) {
    // value1 deve contenere la lista di sequence element
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf("<") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    SequenceManager sm = Sequence.fromInternalFormat(value1);
    Iterator iter;

    // forall
    iter = sm.getElements();
    while (iter.hasNext()) {
      SequenceElementManager seq_elem = (SequenceElementManager) iter.next();
      if (!seq_elem.contain(new Item(value2)))
        return false;
      if (iter.hasNext()) iter.next(); //skip delimiter
    }
    return true;

  }

  /**
   * Check if an item not belong to one elements of given sequence
   *
   * @param value1 a sequence in internal string format
   * @param value2 item to be checked
   * @return boolean
   */
  private boolean evalIsNotIn(String value1, String value2) {
    // value1 deve contenere la lista di sequence element
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf("<") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    SequenceManager sm = Sequence.fromInternalFormat(value1);
    Iterator iter;

    // exists
    iter = sm.getElements();
    while (iter.hasNext()) {
      SequenceElementManager seq_elem = (SequenceElementManager) iter.next();
      if ( !seq_elem.contain(new Item(value2)) )
        return true;
      if( iter.hasNext() ) iter.next();//skip delimiter
    }
    return false;

  }

  /**
   * Check if an item not belong to all elements of given sequence
   *
   * @param value1 a sequence in internal string format
   * @param value2 item to be checked
   * @return boolean
   */
  private boolean evalIsNotInAll(String value1, String value2) {
    // value1 deve contenere la lista di sequence element
    // value2 deve contenere il valore che deve essere confrontato
    if (value2.indexOf("<") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }
    SequenceManager sm = Sequence.fromInternalFormat(value1);
    Iterator iter;

    // forall
    iter = sm.getElements();
    while (iter.hasNext()) {
      SequenceElementManager seq_elem = (SequenceElementManager) iter.next();
      if ( seq_elem.contain(new Item(value2)) )
        return false;
      if( iter.hasNext() ) iter.next();//skip delimiter
    }
    return true;

  }

  /**
   * Check if a sequence is a subsequence of another one.
   * Sequence must have the following format :
   * item1, ... , itemN; ... ;item1, ... , itemN
   *
   * @param value1 a sequence in internal string format
   * @param value2 a sequence in specified format
   * @return boolean
   */
  private boolean evalSubSequence(String value1, String value2)
      throws XMLSpecialElementException{

    // value1 deve contenere la lista di sequence element
    // value2 deve contenere la sotto-sequenza
    if (value2.indexOf("<") != -1) {
      String s = new String(value2);
      value2 = new String(value1);
      value1 = s;
    }

    //parsing value2: item1, ... , itemN; ... ;item1, ... , itemN
    StringTokenizer tokenizer = new StringTokenizer(value2, ";");
    Sequence sequence = null;
    while( tokenizer.hasMoreTokens() ){
      String element = tokenizer.nextToken();
      ElementItemSet iset = new ElementItemSet();

      StringTokenizer token = new StringTokenizer(element, ",");
      while( token.hasMoreTokens() ){
        String item = token.nextToken();
        if( !item.trim().equals("") )
          iset.addItem( new Item(item) );
      }

      if( iset.getNumberOfItems() != 0 ){
        if (sequence == null) sequence = new Sequence(iset);
        else sequence.addElement(Delimiter.DEFAULT, iset);
      }
    }

    if( sequence == null ) return true;

    SequenceManager sm = Sequence.fromInternalFormat(value1);
    return sm.contains(sequence);
  }

  /**
   * Evaluates the condition if the attribute value is missing.
   * Returns true if the condition is true when a missing attribute value is found.
   * Returns false otherwise.
   *
   * @return boolean
   */
  boolean evalMissingAttribute() {
    return true;
  }

  /**
   * Returns false it the condition is evaluates on a table instance.
   * Returns true if the operator is evaluated on a preprocessing instances also.
   * The last one is used for preprocessing operators.
   *
   * @return boolean
   */
  boolean isMetaInstancesOperator() {
    return false;
  }
}
