/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Core.ProprietaryElement.Condition;

import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * This class contains an enumeration of op_type base condition values for the
 * SEQUENCE_FILTER operator.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public final class SEQUENCE_FILTER_CONDITION_DESCR extends BaseConditionDescr {

  protected final static byte IS_IN = 6;

  protected final static byte IS_IN_ALL = 7;

  protected final static byte IS_NOT_IN = 8;

  protected final static byte IS_NOT_IN_ALL = 9;

  protected final static byte SUB_SEQUENCE = 10;

  // sequence attributes
  private final static String SEQUENCE = "sequence";

  // numeric attributes
  private final static String SEQUENCE_SUPPORT = "sequence_support";
  private final static String SEQUENCE_CARDINALITY = "sequence_cardinality";
  private final static String SEQUENCE_DISTINCT_ITEMS_CARDINALITY = "sequence_distinct_items_cardinality";
  private final static String MAX_NUMBER_OF_ITEMS_PER_SET = "max_number_of_items_per_set";

  private final static String NUMERIC_ATTRIBUTES_LIST =
      "@" + SEQUENCE_SUPPORT + ", " +
      "@" + SEQUENCE_CARDINALITY + ", " +
      "@" + SEQUENCE_DISTINCT_ITEMS_CARDINALITY + ", " +
      "@" + MAX_NUMBER_OF_ITEMS_PER_SET + ", " ;

  private final static String SEQUENCE_ATTRIBUTES_LIST =
      "@" + SEQUENCE + ", " ;

  private static boolean checkAttribute(String attribute_name, boolean numeric){
    if( numeric )
      return (NUMERIC_ATTRIBUTES_LIST.indexOf("@"+attribute_name+",") > -1 );
    else
      return (SEQUENCE_ATTRIBUTES_LIST.indexOf("@"+attribute_name+",") > -1 );
  }

  private String term1_value;

  /**
   * Proprietary constructor.
   */
  protected SEQUENCE_FILTER_CONDITION_DESCR() {
    super();
  }

  /**
   * Returns the list of legal values for the op_type attribute.
   *
   * @return String[]
   */
  public String[] list() {

    String ris[] = {
        "equal", "not_equal", "greater", "greater_or_equal", "less", "less_or_equal",
        "is_in", "is_in_all", "is_not_in", "is_not_in_all", "sub_sequence"
    };

    return ris;
  }

  /**
   * Returns the operator code of the input op_type attribute.
   * Returns -1 if the operator is not supported.
   *
   * @param op_type String
   * @return byte
   */
  byte getOpCode(String op_type) {
    byte b = super.getOpCode(op_type);
    if (b != -1)
      return b;
    if (op_type.equalsIgnoreCase("is_in"))
      return IS_IN;
    if (op_type.equalsIgnoreCase("is_in_all"))
      return IS_IN_ALL;
    if (op_type.equalsIgnoreCase("is_not_in"))
      return IS_NOT_IN;
    if (op_type.equalsIgnoreCase("is_not_in_all"))
      return IS_NOT_IN_ALL;
    if (op_type.equalsIgnoreCase("sub_sequence"))
      return SUB_SEQUENCE;
    return -1;
  }

  /**
   * Returns the op_type cardinality.
   * Returns 0 if the input operator has not arguments.
   * Returns 1 if the input operator is unary.
   * Returns 2 if the input operator is binary.
   * Returns 3 if the input operatro has three arguments.
   *
   * @param op_type String
   * @return byte 2 for each legal operator.
   */
  public byte getTermCardinality(String op_type) {

    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
      case IS_IN:
      case IS_NOT_IN:
      case IS_IN_ALL:
      case IS_NOT_IN_ALL:
      case SUB_SEQUENCE:
        return 2;
      default:
        return -1;
    }
  }

  /**
   * Checks the first term.
   * It can be: sequence_support, sequence_cardinality, sequence distinct items cardinality,
   * max number of items per set, sequence.
   *
   * @param op_type String
   * @param term1_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm1(String op_type, String term1_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    this.term1_value = term1_value;
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the first term is a numeric attribute
        if (is_constant)
          msg = "Illegal condition operator: in " + op_type +
              " the first term must refer a sequence table attribute: "
              + NUMERIC_ATTRIBUTES_LIST + ".";
        else
        if ( !checkAttribute(term1_value, true) )
          msg = "Illegal condition operator: in " + op_type +
              " the first term must refer a sequence table attribute: "
              + NUMERIC_ATTRIBUTES_LIST + ".";

        break;
      case IS_IN:
      case IS_IN_ALL:
      case IS_NOT_IN:
      case IS_NOT_IN_ALL:
      case SUB_SEQUENCE:
        // the first term is a sequence attribute
        if (is_constant)
          msg = "Illegal condition operator: in " + op_type +
              " the first term must refer a sequence table attribute: "
              + SEQUENCE_ATTRIBUTES_LIST + ".";
        else
        if ( !checkAttribute(term1_value, false) )
          msg = "Illegal condition operator: in " + op_type +
              " the first term must refer a sequence table attribute: "
              + SEQUENCE_ATTRIBUTES_LIST + ".";
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.MALFORMED_CONDITION);

  }

  /**
   * Checks the second term.
   *
   * @param op_type String
   * @param term2_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm2(String op_type, String term2_value, boolean is_constant)
      throws XMLSpecialElementException {

    String msg = null;
    byte op_type_code = this.getOpCode(op_type);
    switch(op_type_code) {

      case EQUAL:
      case NOT_EQUAL:
      case GREATER:
      case GREATER_OR_EQUAL:
      case LESS:
      case LESS_OR_EQUAL:
        // the second term is a numeric constant
        if (!is_constant)
          msg = "Illegal condition operator: in " + op_type +
              " the second term must be a numeric constant!";
        else {
          try {
            Double v = new Double(term2_value);
            if (term1_value.equals(SEQUENCE_SUPPORT)) {
              if (v.doubleValue() <= 0 || v.doubleValue() > 1) {
                msg = "Illegal condition operator: in " + op_type +
                    " the second term must be a real number between 0 and 1!";
                break;
              }
              Integer i = new Integer(term2_value);
              if (i.intValue() <= 0) {
                msg = "Illegal condition operator: in " + op_type +
                    " the second term must be a positive integer!";
                break;
              }
            }
          }
          catch (NumberFormatException e) {
            msg = "Illegal condition operator: in " + op_type +
                " the second term must be a numeric constant!";
          }
        }
        break;
      case IS_IN:
      case IS_IN_ALL:
      case IS_NOT_IN:
      case IS_NOT_IN_ALL:
      case SUB_SEQUENCE:

          // the second term is a string constant
        if (!is_constant){
          msg = "Illegal condition operator: in " + op_type +
              " the second term must be a constant!";
        }
        break;
    }
    if (msg != null)
      throw new XMLSpecialElementException(msg,
                                           null,
                                           kddml.Core.KDDMLCoreErrorCodes.UNSUPPORTED_CONDITION_OPERATOR);
  }

  /**
   * Checks the third term. No checking is needed.
   *
   * @param op_type String
   * @param term3_value String
   * @param is_constant boolean
   * @throws XMLSpecialElementException
   */
  protected void checkTerm3(String op_type, String term3_value, boolean is_constant)
      throws XMLSpecialElementException {

  }
}
