/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Expression;

import weka.core.Instance;
import kddml.Core.*;
import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * SeqTerm is used for sequential terms on expressions.
 * It includes basic operations (such as sum, multiplication, subtraction, division of
 * numbers and concatenation of strings) on primitive terms of the expression.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @author Sandra Zimei
 * @version 1.0
 */
public class SeqTerm extends Expression {

  /**
   * The list of expressions.
   */
  Expression[] expressions;

  /**
   * The operation type.
   */
  BaseExpressionType op_type;

  /**
   * Constructor.
   *
   * @param arguments Expression[] the set of arguments.
   * @param op_type String the operation type.
   * @throws XMLSpecialElementException
   */
  public SeqTerm(Expression[] arguments, String op_type)
      throws XMLSpecialElementException {

    this.expressions = arguments;
    this.op_type= BaseExpressionType.getExpressionType(op_type);
    if (this.op_type == null) {
      String msg = "Unsupported expression operator: " + op_type;
      throw new XMLSpecialElementException(msg, null, KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
    }
  }

  /**
   * Evaluates the expression on input instance.
   * The tupla is defined in weka.core.Instance class.
   *
   * @param instance Instance the input instance as in weka.core.Instance.
   * @throws KDDMLCoreException
   * @return String the expression evaluation as string or the missing value "?".
   */
  public String evalExpr(Instance instance) throws KDDMLCoreException {

    String[] results = new String[expressions.length];
    String evaluation = null;

    for (int i = 0; i < expressions.length; i++) {

      String ris = ( (Expression) expressions[i]).evalExpr(instance);
      results[i] = ris;
    }
    // ris[] contiene gli operandi come stringhe
    double evaluation_numeric = 0;

    if (op_type.equals(BaseExpressionType.SUM)) {
      if (!containMissing(results)) {
        evaluation_numeric = 0;
        for (int i = 0; i < expressions.length; i++) {
          try {
            evaluation_numeric = evaluation_numeric +
                (new Double(results[i]).doubleValue());
          }

          catch (Exception e) {
            throw new XMLSpecialElementException
                ("< " + results[i] + " >" + " must be a number .",
                 null,
                 KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
          }
        }
        evaluation = new Double(evaluation_numeric).toString();
      }
      else
        evaluation = "?";

    }

    else
    if (op_type.equals(BaseExpressionType.SUBTRACT)) {
      if (!containMissing(results)) {
        evaluation_numeric = new Double(results[0]).doubleValue();
        for (int i = 1; i < expressions.length; i++) {
          try {
            evaluation_numeric = evaluation_numeric -
                (new Double(results[i]).doubleValue());
          }
          catch (Exception e) {
            throw new KDDMLCoreException
                ("< " + results[i] + " >" + " must be a number .",
                 null,
                 KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
          }
        }
        evaluation = new Double(evaluation_numeric).toString();
      }
      else
        evaluation = "?";

    }
    else
    if (op_type.equals(BaseExpressionType.MULTIPLY)) {
      if (!containMissing(results)) {
        evaluation_numeric = 1;
        for (int i = 0; i < expressions.length; i++) {
          try {
            evaluation_numeric = evaluation_numeric *
                (new Double(results[i]).doubleValue());
          }
          catch (Exception e) {
            throw new KDDMLCoreException
                ("< " + results[i] + " >" + " must be a number .",
                 null,
                 KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
          }
        }
        evaluation = new Double(evaluation_numeric).toString();
      }
      else
        evaluation = "?";

    }
    else
    if (op_type.equals(BaseExpressionType.DIVIDE)) {

      if (!containMissing(results)) {
        evaluation_numeric = new Double(results[0]).doubleValue();
        for (int i = 1; i < expressions.length; i++) {

          try {
            evaluation_numeric = evaluation_numeric /
                (new Double(results[i]).doubleValue());

          }
          catch (Exception e) {
            throw new KDDMLCoreException
                ("< " + results[i] + " >" + " must be a number .",
                 null,
                 KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
          }
        }
        evaluation = new Double(evaluation_numeric).toString();
      }
      else
        evaluation = "?";

    }
    else
    if (op_type.equals(BaseExpressionType.CONCAT)) {

      evaluation = "?";
      for (int i = 0; i < expressions.length; i++) {
        if (!results[i].equals("?"))
          evaluation = evaluation + ((results[i]));
      }

    }
    else
    if (op_type.equals(BaseExpressionType.EQUAL)) {
      boolean evaluation_bool = true;
      if (!containMissing(results)) {
        for (int i = 0; i < expressions.length - 1; i++) {
          try {
            double d1 = new Double(results[i]).doubleValue();
            double d2 = new Double(results[i+1]).doubleValue();
            evaluation_bool = evaluation_bool && (d1 == d2);
          }
          catch (Exception e) {
            evaluation_bool = evaluation_bool &&
                ( (results[i]).equals(results[i + 1]));
          }
        }
        evaluation = new Boolean(evaluation_bool).toString();
      }
      else
        evaluation = "?";

    }
    else {
      String msg = "Unsupported  operator < " + op_type + " > for numeric attributes.";
      throw new XMLSpecialElementException(msg, msg,
                                           KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
    }
    return evaluation;

  }

  private boolean containMissing(String[] ris) {
    for (int i=0; i<ris.length; i++)
      if (ris[i].equals("?"))
        return true;
    return false;
  }
}
