/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core.ProprietaryElement.Expression;

import org.w3c.dom.Element;
import kddml.Core.*;
import kddml.Core.ProprietaryElement.Condition.*;
import java.util.Vector;
import kddml.Core.ProprietaryElement.XMLSpecialElementException;

/**
 * A factory that creates an expression from an XML language element.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @author Sandra Zimei
 * @version 1.0
 */
public final class XMLExpressionFactory {

  protected final static String EXPRESSION = "EXPRESSION";
  protected final static String BASE_TERM = "BASE_TERM";
  protected final static String SEQ_TERM = "SEQ_TERM";
  protected final static String IF_TERM = "IF_TERM";
  protected final static String ATTRIBUTE_TERM = "value";
  protected final static String OP_TYPE = "op_type";

  private Element root = null;

  /**
   * Constructor given the XML DOM element.
   *
   * @param element Element the XML element representing the expression.
   * @throws KDDMLCoreException
   */
  public XMLExpressionFactory(Element element) throws kddml.Core.KDDMLCoreException {

    root = element;
  }

  /**
   * Returns a new expression statement.
   *
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public kddml.Core.KDDMLObject newInstance() throws KDDMLCoreException {

    if (root == null)
      throw new KDDMLCoreException("Expression element is null!", null,
                                   KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);
   if (root.getTagName().equals(BASE_TERM))

   {
     // BASE_TERM found!!!
     String ris = XMLUtility.getAttribute(root, ATTRIBUTE_TERM);
     return new BaseTerm(ris);
   }

  if (root.getTagName().equals(SEQ_TERM))

   {
     // SEQ_TERM found!!!
     Vector v = XMLUtility.getChildren(root);
     Expression[] ris = new Expression[v.size()];
     for (int i=0; i<v.size(); i++) {
       Element el = (Element) v.get(i);
       XMLExpressionFactory factory = new XMLExpressionFactory(el);
       Expression cond = (Expression) factory.newInstance();
       ris[i] = cond;
     }
     return new SeqTerm(ris, XMLUtility.getAttribute(root, OP_TYPE));

    }

    if (root.getTagName().equals(IF_TERM)) {
      // IF_TERM found!!!
      // an IF_TERM is composed by a CONDITION, a THEN element and (optionally) an
      // ELSE element
      Vector v = XMLUtility.getChildren(root);
      Element condition = (Element) v.get(0);
      Element then = (Element) v.get(1);
      Element _else = null;
      if (v.size() == 3)
        _else = (Element) v.get(2);

      // CONDITION factory
      ConditionFactory cond_factory = new ConditionFactory(condition, "expression");
      Condition c = (Condition) cond_factory.newInstance();

      // THEN factory
      XMLExpressionFactory factory_then = new XMLExpressionFactory(then);
      Expression t = (Expression) factory_then.newInstance();
      if (_else != null) {
        // ELSE factory
        XMLExpressionFactory factory_else = new XMLExpressionFactory(_else);
        Expression e = (Expression) factory_else.newInstance();
        return new IfTerm(c, t, e);
      }
      else
        return new IfTerm(c, t, null);
    }

    throw new XMLSpecialElementException("Malformed element expression: expression do not match DTDs.",
                                         null,
                                         KDDMLCoreErrorCodes.MALFORMED_EXPRESSION);

  }
}
