/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Core;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.ParserConfigurationException;
import org.w3c.dom.*;
import org.xml.sax.SAXParseException;
import java.io.*;
import java.io.IOException;
import java.util.Vector;
import kddml.Core.ParserErrorListener;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.TransformerException;

/**
 * XML parser utility. It uses the DOM stategy.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public abstract class XMLUtility {

  /**
   * DOM document.
   */
  public static Document doc;

  /**
   * Returns all children of a given node.
   *
   * @param nodo Node
   * @return Vector the list of children as Node.
   */
  public static Vector getChildren(Node nodo) {
    Vector lista_risultato = new Vector();
    NodeList lista_figli = nodo.getChildNodes();
    for (int i = 0; i < lista_figli.getLength(); i++) {
      Node figlio = lista_figli.item(i);
      if (figlio.getNodeType() != figlio.TEXT_NODE)
        lista_risultato.add(figlio);
    }
    return lista_risultato;
  }

  /**
   * Returns true if a node has at least a child.
   *
   * @param padre Node
   * @return boolean
   */
  public static boolean hasChildren(Node padre) {
    Vector lista_figli = getChildren(padre);
    if (lista_figli.size() == 0)
      return false;
    else
      return true;
  }

  /**
   * Returns the number of children of a given node.
   *
   * @param padre Node
   * @return int
   */
  public static int getNumOfChildren(Node padre) {
    Vector lista_figli = getChildren(padre);
    return lista_figli.size();
  }

  /**
   * Returns the i-esim child of a given node.
   * Return null if the child do not exist.
   *
   * @param padre Node
   * @param i int
   * @return Node
   */
  public static Node getChild(Node padre, int i) {
    Vector lista_figli = getChildren(padre);
    if (lista_figli.size() <= i) {
      return null;
    }
    return (Node) lista_figli.get(i);
  }

  /**
   * Returns the first child of a given node.
   * Returns null if the node do not have children.
   *
   * @param padre Node
   * @return Node
   */
  public static Node getFirstChild(Node padre) {
    Vector lista_figli = getChildren(padre);
    if (lista_figli.size() == 0)
      return null;
    return (Node) lista_figli.get(0);
  }

  /**
   * Returns the last child of a given node.
   * Returns null if the node do not have children.
   *
   * @param padre Node
   * @return Node
   */
  public static Node getLastChild(Node padre) {
    Vector lista_figli = getChildren(padre);
    if (lista_figli.size() == 0)
      return null;
    return (Node) lista_figli.get(lista_figli.size() - 1);
  }

  /**
   * Set the attribute to the input XML element.
   * Replace every '<' with '&lt;' in attribute value
   * Replace every '>' with '&gt;' in attribute value
   * Replace every '&' with '&amp;'
   * Replace every ''' with '&apos;'
   * Replace every '"' with '&quot;'
   *
   * @param elem Element
   * @param attribute_name String
   * @param attribute_value String
   */
  public static void setAttribute(Element elem, String attribute_name, String attribute_value) {
    // replace every '<' with '&lt;' in attribute value
    // replace every '>' with '&gt;' in attribute value
    // replace every '&' with '&amp;'
    // replace every ''' with '&apos;'
    // replace every '"' with '&quot;'
    attribute_value = attribute_value.replaceAll("<", "&lt;");
    attribute_value = attribute_value.replaceAll(">", "&gt;");
    attribute_value = attribute_value.replaceAll("&", "&amp;");
    attribute_value = attribute_value.replaceAll("'", "&apos;");
    attribute_value = attribute_value.replaceAll("\"", "&quot;");

    elem.setAttribute(attribute_name, attribute_value);
  }

  /**
   * Returns the attribute value of the input XML elemente.
   * Replace every '&lt;' with '<' in attribute value
   * Replace every '&gt;' with '>' in attribute value
   * Replace every '&amp;' with '&'
   * Replace every '&apos;' with '''
   * Replace every '&quot;' with '"'
   *
   * @param elem Element
   * @param attribute_name String
   * @return String
   */
  public static String getAttribute(Element elem, String attribute_name) {
    String ris = elem.getAttribute(attribute_name);
    ris = ris.replaceAll("&lt;", "<");
    ris = ris.replaceAll("&gt;", ">");
    ris = ris.replaceAll("&amp;", "&");
    ris = ris.replaceAll("&apos;", "'");
    ris = ris.replaceAll("&quot;", "\"");
    return ris;
  }

  /**
   * XML Parser. Returns a DOM object.
   * Throws an exception if an error occours.
   *
   * @param input InputStream
   * @param dtd_resolver DTDResolver
   * @throws KDDMLCoreException
   * @throws DOMParserException
   * @return Document
   */
  public static Document parse(InputStream input, DTDResolver dtd_resolver)
      throws KDDMLCoreException, DOMParserException {

    // crea un file temporaneo di appoggio in cui verra' inserito il documento clonato
    // la clonazione del documento e' necessaria per inserire il link alle DTD richieste dal
    // parser. Il documento sara' successivamente eliminato al termine del parsing.
    // la directory e' quella contenente tutte le DTD.
    // Il temporaneo viene creato utilizzando un generatore di numeri casuale.

    File f = new File(dtd_resolver.getDTDPath());
    String st = f.getParent();
    File file_tmp = null;
    int tmp_counter = (int) Math.random() * 1000;
    // documento temporaneo che verra' parsato
    String TMP_DOCUMENT = "_tmp_"+tmp_counter+".xml";
    file_tmp = new File(st + File.separator + TMP_DOCUMENT);
    boolean exit = false;

    while(!exit) {

      if (file_tmp.exists()) {
        tmp_counter = (int) (Math.random() * 1000);
        TMP_DOCUMENT = "_tmp_" + tmp_counter + ".xml";
        file_tmp = new File(st + File.separator + TMP_DOCUMENT);
      }
      else
        exit = true;
    }

    if (dtd_resolver != null) {

      try {
        StringBuffer newXML = new StringBuffer();
        BufferedReader reader = new BufferedReader( new InputStreamReader(input, "UTF-8") );
        for (String line = reader.readLine(); line != null;
             line = reader.readLine())
          newXML.append(line + "\n");

          // rimuove ogni eventuale <?xml...
        int s = newXML.indexOf("<?xml ");
        if (s != -1)
          newXML.replace(s, newXML.indexOf(">", s) + 1, "");

          // rimuove ogni eventuale <!DOCTYPE...
        s = newXML.indexOf("<!DOCTYPE ");
        if (s != -1)
          newXML.replace(s, newXML.indexOf(">", s) + 1, "");

        // appende al documento la DOCTYPE proprietaria...
        StringBuffer newXML2 = new StringBuffer();
        newXML2.append("<?xml version=\"1.0\" encoding=\"" + "UTF-8" + "\"?>\n");
        newXML2.append(dtd_resolver.getDoctype());
        newXML2.append(newXML.toString());

        reader.close();

        FileWriter fw = new FileWriter(file_tmp, false);
        fw.write(newXML2.toString());
        fw.close();
      }
      catch (IOException e) {
        file_tmp.delete();
        throw new DOMParserException("I/O Error: " + e.toString());
      }
    }

    DocumentBuilderFactory builderfactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder factory = null;

    builderfactory.setValidating(true);
    try {
      //Crea un oggetto di tipo DocumentBuilder
      factory = builderfactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException e) {
      file_tmp.delete();
      throw new DOMParserException(e.toString());
    }
    ParserErrorListener handler = new ParserErrorListener();
    Document xmlDoc = null;
    try {
      factory.setErrorHandler(handler);

      if (dtd_resolver != null)
        xmlDoc = factory.parse(file_tmp);
      else
        xmlDoc = factory.parse(input);

      // controllo se si sono verificati degli errori
      if (handler.error) {
        file_tmp.delete();
        // errore durante il parsing
        throw new DOMParserException(handler.messaggio);
      }

    }
    catch (IOException e) {
      System.out.println("IOException!!!: " + e);
      file_tmp.delete();
      throw new DOMParserException("Input file not found: " + e.getMessage());
    }
    catch (SAXParseException e) {
      file_tmp.delete();
      throw new DOMParserException(handler.messaggio);
    }
    catch (Exception e) {
      file_tmp.delete();
      throw new DOMParserException(handler.messaggio);
    }
    file_tmp.delete();
    return xmlDoc;
  }


  /**
   * XML Parser. Returns a DOM object.
   * Throws an exception if an error occours.
   *
   * @param input File
   * @throws KDDMLCoreException
   * @return Document
   */
  public static Document parse(File input)
      throws KDDMLCoreException {

    DocumentBuilderFactory builderfactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder factory = null;

    builderfactory.setValidating(true);
    try {
      //Crea un oggetto di tipo DocumentBuilder
      factory = builderfactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException e) {
      throw new DOMParserException(e.toString());
    }
    ParserErrorListener handler = new ParserErrorListener();
    Document xmlDoc = null;
    try {
      factory.setErrorHandler(handler);
      xmlDoc = factory.parse(input);

      // controllo se si sono verificati degli errori
      if (handler.error) {
        // errore durante il parsing
        throw new DOMParserException(handler.messaggio);
      }

    }
    catch (IOException e) {
      System.out.println("IOException!!!: " + e);
      throw new DOMParserException("Input file not found: " + e.getMessage());
    }
    catch (SAXParseException e) {
      throw new DOMParserException(handler.messaggio);
    }
    catch (Exception e) {
      throw new DOMParserException(handler.messaggio);
    }
    return xmlDoc;
  }

  /**
   * XML Parser. Returns a DOM object.
   * Throws an exception if an error occours.
   *
   * @param input String
   * @param parse_dtd boolean
   * @throws KDDMLCoreException
   * @return Document
   */
  public static Document parse(String input, boolean parse_dtd)
      throws KDDMLCoreException {

    DocumentBuilderFactory builderfactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder factory = null;

    org.xml.sax.InputSource source = new org.xml.sax.InputSource(new StringReader(input));
    builderfactory.setValidating(true);
    try {
      //Crea un oggetto di tipo DocumentBuilder
      factory = builderfactory.newDocumentBuilder();
    }
    catch (ParserConfigurationException e) {
      throw new DOMParserException(e.toString());
    }
    ParserErrorListener handler = new ParserErrorListener();
    Document xmlDoc = null;
    try {
      factory.setErrorHandler(handler);
      xmlDoc = factory.parse(source);

      // DO NOT CATCH ERRORS.....
      if (handler.error && parse_dtd) {
        // errore durante il parsing
        throw new DOMParserException(handler.messaggio);
      }

    }
    catch (IOException e) {
      System.out.println("IOException!!!: " + e);
      throw new DOMParserException("Input file not found: " + e.getMessage());
    }
    catch (SAXParseException e) {
      throw new DOMParserException(handler.messaggio);
    }
    catch (Exception e) {
      throw new DOMParserException(handler.messaggio);
    }
    return xmlDoc;
  }

  /**
   * Print the document.
   *
   * @return String
   */
  public String toString() {
    if (doc == null)
      return ("No Document Created...");
    else
      return (toString(doc));
  }

  /**
   * Returns a representation of the input node as string.
   *
   * @param node Node
   * @return String
   */
  public static String toString(Node node) {

    if (node == null)
      return "null";

    String s = "";
    final String tab = "   ";
    String indent = "";
    for (short i = 0; i < depth_level; i++)
      indent = indent + tab;
    int type = node.getNodeType();
    switch (type) {
      case Node.DOCUMENT_NODE: {
        s = s + "<?xml version=\"1.0\" encoding=\"" + "UTF-8" + "\"?>\n";
        break;
      } //end of document
      case Node.ELEMENT_NODE: {
        depth_level++;
          s = s + indent + "<" + node.getNodeName();
        NamedNodeMap nnm = node.getAttributes();
        if (nnm != null) {
          int len = nnm.getLength();
          Attr attr;
          for (int i = 0; i < len; i++) {
            attr = (Attr) nnm.item(i);
            s = s + " " + attr.getNodeName() + "=\"" + attr.getNodeValue() +
                '"';
          }
        }
        // stampa la sezione CDATA
        Node n = node.getFirstChild();
        if (n != null) {
          s = s + ">" + "\n";
          if (n.getNodeValue() != null)
            if (!n.getNodeValue().trim().equals(""))
              s = s + n.getNodeValue();
        }
        break;
      } //end of element
      case Node.ENTITY_REFERENCE_NODE: {
        s = s + '&' + node.getNodeName() + ';';
        break;
      } //end of entity
      case Node.CDATA_SECTION_NODE: {
        s = s + "<![CDATA[" + node.getNodeValue() + "]]>";
        break;
      }
      case Node.TEXT_NODE: {
        s = s + node.getNodeValue();
        break;
      }
      // ---------------------------------
      // NO DTD LINK IN OUTPUT DOCUMENT!!!
      // ---------------------------------
      /*case Node.DOCUMENT_TYPE_NODE: {
        if (node.getNodeName().equals("KDDML_OBJECT"))

          // query KDDML
          s = s + "<!DOCTYPE " + node.getNodeName() + " SYSTEM " +
              dtd_link + DTD_KDDML_QUERY + ">\n";
        else
        if (node.getNodeName().equals("KDDML_TABLE"))

          // tabella
          s = s + "<!DOCTYPE " + node.getNodeName() + " SYSTEM " +
              dtd_link + DTD_KDDML_TABLE + ">\n";
        else

          // documento PMML
          s = s + "<!DOCTYPE " + node.getNodeName() + " SYSTEM " +
              dtd_link + DTD_PMML + ">\n";
        break;
      }*/
      case Node.PROCESSING_INSTRUCTION_NODE: {
        s = s + "<?" + node.getNodeName();
        String data = node.getNodeValue();
        if (data != null && data.length() > 0) {
          s = s + ' ' + data;
        }
        s = s + "?>" + "\n";
        break;
      }
    } //end of switch
    //recurse
    Vector v = getChildren(node);
    for (int i = 0; i < v.size(); i++) {
      s = s + toString( (Node) v.get(i));
    }
    //without this the ending tags will miss
    if (type == Node.ELEMENT_NODE) {
      depth_level--;
      indent = "";
      for (short i = 0; i < depth_level; i++)
        indent = indent + tab;
      if (node.getFirstChild() != null) {
        if (node.getNodeName().equals("Array")) {
          // for array no indent....
          s = s + /*indent +*/ "</" + node.getNodeName() + ">" + "\n";
        }
        else
          s = s + indent + "</" + node.getNodeName() + ">" + "\n";
      }
      else
        s = s + "/>" + "\n";
    }
    return s;
  }

  private static short depth_level = 0;

  /**
   * Saves the input document in the specified path.
   * Returns false if the document cannot be saved.
   *
   * @param path String
   * @param doc Document
   * @return boolean
   */
  public static boolean toFile(String path, Document doc) {

    PrintWriter pw;
    try {
      pw = new PrintWriter(new BufferedWriter(new FileWriter(path)));
      pw.print(toString(doc));
      pw.flush();
      pw.close();
      return true;
    }
    catch (Exception e) {
      System.out.println("Error found during the document saving: " + e);
      return false;
    }
  }

  /**
   * Cloning of input node.
   * This method is used to append to a document an XML node that belong to another
   * DOM tree.
   *
   * @param node Node
   * @return Element
   */
  public static Element clone(Node node) {
    System.out.println(XMLUtility.toString(node));
    int type = node.getNodeType();
    Element ris = null;
    //System.out.println("Sono in clone: ");

    switch (type) {
      case Node.ELEMENT_NODE: {
        ris = (Element) doc.createElement(node.getNodeName());
        Node n = node.getFirstChild();
        if (n != null)
          if (n.getNodeValue() != null)
            if (!n.getNodeValue().trim().equals(""))
              ris.appendChild(doc.createTextNode(n.getNodeValue()));
        NamedNodeMap nnm = node.getAttributes();
        if (nnm != null) {
          int len = nnm.getLength();
          Attr attr;
          for (int i = 0; i < len; i++) {
            attr = (Attr) nnm.item(i);
            ris.setAttribute(attr.getNodeName(), attr.getNodeValue());
          }
        }
        break;
      }
    }
    //recurse
    Vector v = getChildren(node);
    for (int i = 0; i < v.size(); i++) {
      ris.appendChild(clone( (Node) v.get(i)));
    }
    return ris;
  }

  /**
   * Method used to initialize the XML document.
   *
   * @param root String
   * @return Document
   */
  public static Document initializeDocument(String root) {
    DocumentBuilderFactory builderfactory = DocumentBuilderFactory.newInstance();
    DocumentBuilder factory = null;
    DOMImplementation dom = null;
    doc = null;
    try {
      //Crea il riferimento ad un oggetto di tipo DocumentBuilder
      factory = builderfactory.newDocumentBuilder();
      //builderfactory.setValidating(true);
    }
    catch (ParserConfigurationException e) {
    }
    dom = factory.getDOMImplementation();
    //Crea un oggetto DOCTYPE
    DocumentType doctype = dom.createDocumentType(root, null, null);

    //Crea l'albero DOM con radice root e nodo DOCTYPE
    doc = dom.createDocument(null, root, doctype);
    return doc;
  }

  /**
   * Formatting of input xml document. XML --> HTML translator.
   *
   * @param file_xml String
   * @param file_xsl String
   * @param file_html String
   * @param formal_params String[]
   * @param actual_params String[]
   * @throws TransformerConfigurationException
   * @throws TransformerException
   * @throws FileNotFoundException
   * @return boolean
   */
  public static boolean formatting(String file_xml,
                                   String file_xsl,
                                   String file_html,
                                   String[] formal_params,
                                   String[] actual_params) throws
      javax.xml.transform.TransformerConfigurationException,
      javax.xml.transform.TransformerException,
      java.io.FileNotFoundException {
    //try {
    // Istanzia un TransformerFactory.
    TransformerFactory tFactory = TransformerFactory.newInstance();

    // Usa il TransformerFactory per elaborare il foglio di stile fornito in input e
    // generare un oggetto di tipo Transformer.
    Transformer transformer = tFactory.newTransformer(new StreamSource(
        file_xsl));

    if ((formal_params != null) && actual_params != null) {
      for (int i=0; i<formal_params.length; i++) {
        transformer.setParameter(formal_params[i], actual_params[i]);
      }
    }
      // Usa l'oggetto Transformer per trasformare un file XML fornito in input e
      // ottenere l'output come file HTML.
    transformer.transform(new StreamSource(file_xml), new StreamResult(new
        java.io.FileOutputStream(file_html)));
    return true;
    // }
    // catch (Exception e) {
    //   System.out.println(e.toString());
    //   return false;
    //  }
  }

  /**
   * Only for testing.
   *
   * @param args String[]
   */
  public static void main(String args[]) {
    String xml_doc = "C:/Documents and Settings/romei/kddml/resources/dtd/kddml_algorithms.xml";
    try {
      DTDResolver resolver = new DTDResolver(KDDMLObjectType.KDD_QUERY);
      XMLUtility.parse(new File(xml_doc));
    }
    catch(KDDMLCoreException e) {
      System.out.println(e.getDetailErrorMessage());
    }
    /*catch(java.io.FileNotFoundException ecc) {
      System.out.println("File not found...");
    }*/
  }
}
