/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Interpreter;

import java.util.*;
import org.w3c.dom.*;
import kddml.Core.*;
import kddml.Core.ProprietaryElement.Condition.*;
import kddml.Core.ProprietaryElement.Expression.*;
import kddml.Core.KDDMLAlgorithm;
import kddml.Core.ProprietaryElement.XML;
import kddml.Operators.*;
import java.io.*;
import net.xfra.qizxopen.xquery.*;

/**
 * Interprete del sistema che esegue una generica KDD_QUERY. </BR>
 * Effettua una visita in profondita' dell'albero DOM della query XML da eseguire richiamando
 * ricorsivamente gli operatori e gli algoritmi opportuni. </BR>
 * L'individuazione della classe corrispondente che gestisce l'operatore avviene confrontando
 * il nome della classe stessa con il nome del TAG XML che rappresenta l'operatore da eseguire.
 * Il risultato restituito dal metodo execute() associato all'operatore viene validato dal
 * query executor stesso che si occupa anche di memorizzare su disco i risultati e di
 * generare il corrispondente file HTML.
 *
 * Assunzioni:
 *
 * - La specifica di un algoritmo e' identificata dal tag <ALGORITHM>
 *
 * - La specifica di una condizione ternina con il tag <_CONDITION>
 *   (es: RDA_FILTER_CONDITION)
 *
 * - La specifica di un'espressione termina con il tag <_EXPRESSION>
 *   (es: PP_NEW_ATTRIBUTE_EXPRESSION)
 *
 * I tre elementi sopra elencati sono risolti automaticamente dal QE.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */

public class QueryExecutor extends Thread
    implements InterpreterRunner, InterpreterChecker, InterpreterController {

  public final static String APPLICATION_NAME = "KDDML (Knowledge Discovery in Database Markup Language)";

  public final static String APPLICATION_VERSION = "2.0.14 alfa";

  // progressivo dei files temporanei creati
  private static int prog_tmp = 0;

  private final static String TMP_STRING = "_____tmp_";

  /**
   * Abilita il controllo e la scrittura dei risultati del QE
   */
  private boolean disable_saving = false;

  /**
   * Abilita/disabilita la scrittura su terminale dei risultati
   */
  private boolean disable_printing = false;

  private boolean disable_log_file_writing = true;

  private OutputWriterManager output_manager;

  /**
   * Root original query.
   */
  private Element query;

  /**
   * Cloned query with temporary xml_dest.
   */
  private Element query_cloned;

  private KDDMLObject result;

  private String final_msg;

  /**
   * Enumeration of all regular language types.
   */
  private static KDDMLObjectType[] types = KDDMLObjectType.list();


  /**
   * Metodo costruttore; crea e inizializza una istanza della classe.
   *
   * @param path percorso completo della KDD_QUERY da eseguire
   *        (es C:/examples/mine_rule.xml).
   */
  public QueryExecutor(String query_path) throws InvalidKDDMLQueryException {

    // legge dal repository la query e ne effettia il parsing...
    Document DOMQuery = null;
    DTDResolver resolver = null;
    try {
      resolver = new DTDResolver(KDDMLObjectType.KDD_QUERY);
      InputStream input = IOUtility.getInputStream(query_path);
      DOMQuery = XMLUtility.parse(input, resolver);

    }
    catch(KDDMLCoreException ex) {
      throw new InvalidKDDMLQueryException(ex.getErrorMessage(), ex.getDetailErrorMessage());
    }
    catch(java.io.FileNotFoundException ex2) {
      String msg = "Input query < "+query_path+" > not found!";
      throw new InvalidKDDMLQueryException(msg, ex2.toString());
    }
    catch(java.io.IOException ex3) {
      String msg = "I/O error on < "+query_path+" >";
      throw new InvalidKDDMLQueryException(msg, ex3.toString());
    }
    System.out.println("The input query "+query_path+" is a valid KDDML_QUERY.");

    Element kddml_object = DOMQuery.getDocumentElement();
    Element kdd_query = (Element) XMLUtility.getFirstChild(kddml_object);
    //query_name = kdd_query.getAttribute("name");
    query = (Element) XMLUtility.getFirstChild(kdd_query);

    // inserisce gli attributi xml_dest che riferiscono dei temporanei quando non specificati
    query_cloned = clone((Element) query.cloneNode(true));

    output_manager = new OutputWriterManager(disable_log_file_writing);
  }


  /**
   * Metodo costruttore; crea e inizializza una istanza della classe.
   *
   * @param doc albero DOM della query da eseguire.
   */
  public QueryExecutor(Document DOMQuery) {

    for (int i=0; i<types.length; i++)
      System.out.println(types[i].getName());

    Element kddml_object = DOMQuery.getDocumentElement();
    Element kdd_query = (Element) XMLUtility.getFirstChild(kddml_object);
    //query_name = kdd_query.getAttribute("name");
    query = (Element) XMLUtility.getFirstChild(kdd_query);

    // inserisce gli attributi xml_dest che riferiscono dei temporanei quando non specificati
    query_cloned = clone((Element) query.cloneNode(true));
    System.out.println("Query: "+XMLUtility.toString(query));
    System.out.println("Query Cloned: "+XMLUtility.toString(query_cloned));

    output_manager = new OutputWriterManager(disable_log_file_writing);
  }

  /**
   * Start the interpreter as separated thread.
   */
  public void run() {
    try {
      resolve();
    }
    catch (KDDMLInterpreterException ex) {
      final_msg = "error: "+ex.getDetailErrorMessage();
    }
  }

  public KDDMLObject getResult() {
    return result;
  }

  public String getMessageResult() {
    return final_msg;
  }

  /**
   * Start the interpreter.
   *
   * @throws KDDMLInterpreterException
   * @return KDDMLObject
   */
  public void resolve() throws KDDMLInterpreterException {

    ///removeTmpResults();
    output_manager.executionQueryStart(XMLUtility.toString(query_cloned));

    // execution start from the root tag...
    result = resolve(query_cloned, KDDMLObjectType.ANY);

    output_manager.exitWithoutError();
    final_msg = "success";
  }

  /**
   * Resolver main.
   *
   * @param query Element the sub-query element.
   * @param type KDDMLObjectType the expected output type.
   * @throws KDDMLInterpreterException
   * @return KDDMLObject
   */
  protected KDDMLObject resolve(Element query, KDDMLObjectType type)
      throws KDDMLInterpreterException {

    // controlla che l'operatore sia stato implementato
    if (NotYetImplementedException.isNotImplemented(query.getTagName()))
      throw new NotYetImplementedException(query.getTagName());

    Vector children = XMLUtility.getChildren(query);
    KDDMLObject result = null;

    if (query.getTagName().equals("IF")) {

      result = resolveIF(query);
    }
    else {
      if (query.getTagName().equals("CALL_QUERY")) {

        // CALL_QUERY statement

        Hashtable params = new Hashtable();
        // memorizza la lista dei parametri con il valore attuale in una tabella hash
        for (int i=0; i<children.size(); i++) {
          Element elem = (Element) children.get(i);
          String param_name = elem.getAttribute("formal_param");
          String param_value = elem.getAttribute("actual_param");
          params.put(param_name, param_value);
        }

        String path = query.getAttribute("query_path");
        if (path.equals(""))
          path = KDDMLObjectType.KDD_QUERY.getRepositoryPath();
        String name = query.getAttribute("query_name");
        CallQueryResolver resolver = new CallQueryResolver(path + name);

        Element calledQuery = resolver.loadQuery(params);
        // set XML dest...
        clone(calledQuery);
        query = calledQuery;
        result = resolve( query, type );
        runTimeChecking(result, type);
      }
      else {

        // RESOLVER CORE
        result = resolveCore(query);
      }

    } // END OF EXECUTION

    if (result == null) {
      // null results
      output_manager.exitWhitError(query.getTagName()+" not yet implemented!");
      throw new NotYetImplementedException(query.getTagName());
    }

    output_manager.executionOperatorEnd(query.getTagName());
    String xml_dest = query.getAttribute("xml_dest");

    KDDMLOperatorResolver opr = null;
    try {
      // istanzia il resolver per richiamare i metodi statici...
      if (!query.getTagName().equals("IF"))
        opr = KDDMLOperatorFactory.newResolverInstance(query.getTagName());
    }
    catch (KDDMLOperatorException ex) {
      output_manager.exitWhitError(ex.getDetailErrorMessage());
      final_msg = "error: "+ex.getDetailErrorMessage();
      throw new ExecutionException(ex.getErrorMessage(), ex.getDetailErrorMessage());
    }

    if ((opr != null) && (opr.runTimeCheckNeeded()))
      // run-time type checking
      runTimeChecking(result, type);

    refreshRAM();

    if ((result.isEmpty()) && (opr.abortIfIsEmpty())) {

      // l'operatore non ha prodotto risultati ed e' richiesta la terminazione della query
      EmptyResultsException ecc2 = new EmptyResultsException(query.getTagName());
      output_manager.exitWhitError(ecc2.getDetailErrorMessage());
      throw ecc2;
    }

    // EXECUTION OK!!!
    if ((xml_dest != null) && (!disable_saving) && (opr != null) && (opr.saveXMLOutput()))
      try {
        output_manager.savingResultsStart(query.getTagName());
        result.saveToRepository();
        output_manager.savingResultsEnd(query.getTagName(), xml_dest);

      }
      catch (KDDMLCoreException ex1) {
        output_manager.exitWhitError(ex1.getDetailErrorMessage());
        final_msg = "error: "+ex1.getDetailErrorMessage();
        throw new ExecutionException(ex1.getErrorMessage(), ex1.getDetailErrorMessage());
      }
    if (!disable_printing)
      output_manager.printResults(result.toString());

    return result;
  }

  protected KDDMLObject resolveCore(Element query)
      throws KDDMLInterpreterException {

    KDDMLOperatorResolver opr = null;
    KDDMLOperatorSettings ops = null;

    try {
      // istanziazione del resolver
      opr = KDDMLOperatorFactory.newResolverInstance(query.getTagName());
      // istanziazione del settings
      ops = KDDMLOperatorFactory.newSettingsInstance(query.getTagName());
    }
    catch (KDDMLOperatorException ex) {
      output_manager.exitWhitError(ex.getDetailErrorMessage());
      final_msg = "error: "+ex.getDetailErrorMessage();
      throw new ExecutionException(ex.getErrorMessage(), ex.getDetailErrorMessage());
    }

    // prepara la lista degli attributi dell'operatore corrente
    NamedNodeMap list = query.getAttributes();
    Hashtable attributes = new Hashtable();
    for (int i = 0; i < list.getLength(); i++) {
      Node n = list.item(i);
      KDDMLScalarManager expr = new KDDMLScalar(null, n.getNodeValue());
      attributes.put(n.getNodeName(), expr);
    }

    // recursive resolving of all children
    Vector params = new Vector();
    Vector children = XMLUtility.getChildren(query);
    int cnt_alg = 0;
    for (int i = 0; i < children.size(); i++) {
      Element elem = (Element) children.get(i);
      if (elem.getTagName().equals("ALGORITHM")) {
        // ALGORITHM tag found!!!
        cnt_alg++;
        KDDMLAlgorithm alg = resolveAlgorithm(elem, attributes, cnt_alg);
        // aggiunge al vettore degli argomenti l'algoritmo
        params.add(alg);
      }
      else
      if (elem.getTagName().equals("CONDITION")) {
        // CONDITION tag found!!!
        try {
          params.add(resolveCond(elem));
        }
        catch (KDDMLCoreException ex2) {
          output_manager.exitWhitError(ex2.getDetailErrorMessage());
          final_msg = "error: "+ex2.getDetailErrorMessage();
          throw new ExecutionException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
        }
      }
      else
      if (elem.getTagName().endsWith("EXPRESSION")) {
        // EXPRESSION tag found!!!
        try {
          params.add(resolveExpr(elem));
        }
        catch (KDDMLCoreException ex1) {
          output_manager.exitWhitError(ex1.getDetailErrorMessage());
          final_msg = "error: "+ex1.getDetailErrorMessage();
          throw new ExecutionException(ex1.getErrorMessage(), ex1.getDetailErrorMessage());
        }
      }
      else {
        if (elem.getTagName().endsWith("GENERIC_ELEMENT")) {
          // generic XML element found
          try {
            params.add(resolveGenericXML(elem));
          }
          catch (KDDMLCoreException ex1) {
            output_manager.exitWhitError(ex1.getDetailErrorMessage());
            final_msg = "error: "+ex1.getDetailErrorMessage();
            throw new ExecutionException(ex1.getErrorMessage(), ex1.getDetailErrorMessage());
          }
        }
        else {

          // recursive resolving of the child
          KDDMLObject obj = resolve(elem, opr.getParamType(i));
          params.add(obj);
        }
      }
    }

    // resolve this operator
    output_manager.executionOperatorStart(query.getTagName());
    KDDMLObject result;
    try {
      ops.applySettings(attributes);

      /*if (query.getTagName().equals("DATABASE_LOADER")) {
        // e' necessario passare i parametri della connessione all'operatore
        ((DATABASE_LOADER_RESOLVER) opr).setConnection(jdbc_conn);
      }*/
      result = opr.execute(ops, params);
    }
    catch(KDDMLOperatorException ecc) {
      output_manager.exitWhitError(ecc.getDetailErrorMessage());
      final_msg = "error: "+ecc.getDetailErrorMessage();
      throw new ExecutionException(ecc.getErrorMessage(), ecc.getDetailErrorMessage());
    }
    catch(KDDMLCoreException ecc2) {
      output_manager.exitWhitError(ecc2.getDetailErrorMessage());
      final_msg = "error: "+ecc2.getDetailErrorMessage();
      throw new ExecutionException(ecc2.getErrorMessage(), ecc2.getDetailErrorMessage());
    }
    return result;
  }


  private KDDMLObject resolveIF(Element query)
      throws KDDMLInterpreterException {

    // IF-THEN-ELSE statement

    KDDMLObject result;
    Vector children = XMLUtility.getChildren(query);
    Element cond = (Element) children.get(0);
    // valuta la xquery expression in ogni termine della condizione
    // sostituendo il valore ottenuto con l'xquery expression

    // evaluate condition object
    KDDMLObject cond_obj = resolve((Element) children.get(1), KDDMLObjectType.ANY);
    File obj = cond_obj.getPhysicalObject();

    cond = evaluateXQueryExpression(cond, obj);

    // istanzia e valuta l'oggetto condition
    Condition condition = null;
    try {
      condition = resolveCond(cond);
    }
    catch (KDDMLCoreException ex2) {
      output_manager.exitWhitError(ex2.getDetailErrorMessage());
      final_msg = "error: "+ex2.getDetailErrorMessage();
      throw new ExecutionException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
    }
    boolean evaluation = false;
    try {
      evaluation = condition.evalCond(null, null);
    }
    catch (KDDMLCoreException ex) {
      throw new InvalidKDDMLQueryException(ex.getErrorMessage(), ex.getDetailErrorMessage());
    }

    if (evaluation) {
      // evaluation then
      Element tmp = (Element) XMLUtility.getFirstChild((Element) children.get(2));
      result = resolve( tmp, KDDMLObjectType.ANY);
    }
    else {
      // evaluation else
      Element tmp = (Element) XMLUtility.getFirstChild((Element) children.get(3));
      result = resolve( tmp, KDDMLObjectType.ANY);
    }

    return result;
  }

  /**
   * Evaluate all XQuery expressions belonging to the condition element.
   * Replace each expression with his value.
   * Every XQuery expression starts with return....
   *
   * @param condition Element
   * @param xml_file File
   * @return Element
   */
  private Element evaluateXQueryExpression(Element condition, File xml_file)
      throws KDDMLInterpreterException {

    try {
      if (!condition.getTagName().equals("BASE_COND")) {

        // recursion...
        Vector list = XMLUtility.getChildren(condition);
        for (int i = 0; i < list.size(); i++) {
          Element child = (Element) list.get(i);
          Element child_cloned = evaluateXQueryExpression(child, xml_file);
          condition.appendChild(child_cloned);
        }
      }
      else {

        // case base...
        XQueryResolver resolver = null;
        if (condition.getAttribute("term1").startsWith("return")) {
          resolver = new XQueryResolver(condition.getAttribute("term1"),
                                        xml_file);
          resolver.compileQuery();
          String evaluation = resolver.executeQuery();
          condition.setAttribute("term1", evaluation);
        }

        if (condition.getAttribute("term2").startsWith("return")) {
          resolver = new XQueryResolver(condition.getAttribute("term2"),
                                        xml_file);
          resolver.compileQuery();
          String evaluation = resolver.executeQuery();
          condition.setAttribute("term2", evaluation);
        }
      }
    }
    catch(java.io.IOException ex1) {
      String msg = ex1.toString();
      throw new KDDMLInterpreterException(msg, msg);
    }
    catch(XQueryException ex2) {
      String msg = ex2.toString();
      throw new KDDMLInterpreterException(msg, msg);
    }
    return condition;
  }

  /**
   * Testa se gli algoritmi richiesti sono supportati dal sistema.
   *
   * @return boolean
   */
  /*public boolean checkAlgorithmsCompatibility()
      throws UnsupportedAlgorithmException {

    NodeList n = query.getElementsByTagName("ALGORITHM");
    for (int i=0;i<n.getLength();i++) {
      Element elem = (Element) n.item(i);
      String alg_name = elem.getAttribute("algorithm_name");

      // richiede al factory del livello operatori se l'alg e' supportato
      DMAlgorithmFactory dm_fact = new DMAlgorithmFactory(alg_name);
      boolean b1 = dm_fact.isSupported();
      if (!b1) {
        PPAlgorithmFactory pp_fact = new PPAlgorithmFactory(alg_name);
        if (!pp_fact.isSupported())
          throw new UnsupportedAlgorithmException(true, alg_name);
      }
    }
    return true;
  }*/

  public void validateQuery() throws ExecutionException,
      UnsupportedAlgorithmException, KDDMLInterpreterException {

    // controllo dei parametri per la connession al DBMS...
    checkDatabaseConnection();
    // controllo dei parametri degli attributi, algorithms and conditions...

    checkParametersCompatibility(query_cloned);
  }

  protected void checkParametersCompatibility(Element query)
      throws InvalidKDDMLQueryException, UnsupportedAlgorithmException {

    KDDMLOperatorSettings ops = null;

    try {
      // istanziazione del settings
      if (query.getTagName().equals("CALL_QUERY"))
        // do nothing
        return;

      if (query.getTagName().equals("IF")) {
        // recursive evaluation of children
        Vector children = XMLUtility.getChildren(query);

        // evaluate condition
        try {
          resolveCond((Element) children.get(0));
        }
        catch (KDDMLCoreException ex2) {
          if (output_manager != null)
            output_manager.exitWhitError(ex2.getDetailErrorMessage());
          final_msg = "error: "+ex2.getDetailErrorMessage();
          throw new InvalidKDDMLQueryException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
        }

        // checks object condition
        checkParametersCompatibility((Element) children.get(1));
        // checks then object
        Element tmp = (Element) XMLUtility.getFirstChild((Element) children.get(2));
        checkParametersCompatibility(tmp);
        // checks else object
        tmp = (Element) XMLUtility.getFirstChild( (Element) children.get(3));
        checkParametersCompatibility(tmp);
        return;
      }
      ops = KDDMLOperatorFactory.newSettingsInstance(query.getTagName());
    }
    catch (KDDMLOperatorException ex) {
      output_manager.exitWhitError(ex.getDetailErrorMessage());
      final_msg = "error: "+ex.getDetailErrorMessage();
      throw new InvalidKDDMLQueryException(ex.getErrorMessage(), ex.getDetailErrorMessage());
    }

    // prepara la lista degli attributi dell'operatore corrente
    NamedNodeMap list = query.getAttributes();
    Hashtable attributes = new Hashtable();
    for (int i = 0; i < list.getLength(); i++) {
      Node n = list.item(i);
      KDDMLScalarManager expr = new KDDMLScalar(null, n.getNodeValue());
      attributes.put(n.getNodeName(), expr);
    }

    // recursive resolving of all children
    Vector children = XMLUtility.getChildren(query);
    int cnt_alg = 0;
    for (int i = 0; i < children.size(); i++) {
      Element elem = (Element) children.get(i);
      if (elem.getTagName().equals("ALGORITHM")) {
        // ALGORITHM tag found!!!
        cnt_alg++;
        resolveAlgorithm(elem, attributes, cnt_alg);
      }
      else
      if (elem.getTagName().equals("CONDITION")) {
        // CONDITION tag found!!!
        try {
          resolveCond(elem);
        }
        catch (KDDMLCoreException ex2) {
          if (output_manager != null)
            output_manager.exitWhitError(ex2.getDetailErrorMessage());
          final_msg = "error: "+ex2.getDetailErrorMessage();
          throw new InvalidKDDMLQueryException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
        }
      }
      else
      if (elem.getTagName().equals("EXPRESSION")) {
        // EXPRESSION tag found!!!
        try {
          resolveExpr(elem);
        }
        catch (KDDMLCoreException ex2) {
          if (output_manager != null)
            output_manager.exitWhitError(ex2.getDetailErrorMessage());
          final_msg = "error: "+ex2.getDetailErrorMessage();
          throw new InvalidKDDMLQueryException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
        }
      }
      else {

        // recursive checking the child
        checkParametersCompatibility(elem);
      }
    }

    // check parameters for this root tag...
    try {
      ops.applySettings(attributes);
    }
    catch(KDDMLOperatorException e) {
      throw new InvalidKDDMLQueryException(e.getErrorMessage(), e.getDetailErrorMessage());
    }
  }

  /**
   * Apre una finestra di dialogo se e' presente un operatore <DATABASE_LOADER>
   * e se non sono stati specificati password e user.
   */
  protected void checkDatabaseConnection()
      throws DBMSConnectionException {

    NodeList n1 = query_cloned.getElementsByTagName("DATABASE_LOADER");
    NodeList n2 = query_cloned.getElementsByTagName("DATABASE_WRITER");
    boolean user_missing = false;

    if (((query_cloned.getTagName().equals("DATABASE_LOADER")) ||
        (query_cloned.getTagName().equals("DATABASE_WRITER"))) &&
        (query_cloned.getAttribute("user").equals("")) &&
        (query_cloned.getAttribute("password").equals("")))

      user_missing = true;

    for (int j=0; j < n1.getLength(); j++) {
      Element el = (Element) n1.item(j);
      if ((el.getAttribute("user").equals("")) &&
          (el.getAttribute("password").equals("")))
        user_missing = true;
    }

    for (int j=0; j < n2.getLength(); j++) {
      Element el = (Element) n2.item(j);
      if ((el.getAttribute("user").equals("")) &&
          (el.getAttribute("password").equals("")))
        user_missing = true;
    }


    String user = null;
    String password = null;
    if (user_missing) {

      // user e password non definiti nella query  --->
      // apertura del pop-up per la definizione di user e password...
      javax.swing.JFrame frame = new javax.swing.JFrame();

      DBMSConnectionFrame DBMSd = new DBMSConnectionFrame(frame);
      DBMSd.pack();
      DBMSd.setVisible(true);
      user = DBMSd.getUser();
      password = DBMSd.getPassword();
      frame.dispose();

      if ( (user == null) || (password == null)) {
        String msg = "Can't initialize user or password to access to database!";
        throw new DBMSConnectionException(msg, null);
      }

      if ( (query_cloned.getTagName().equals("DATABASE_LOADER")) ||
          (query_cloned.getTagName().equals("DATABASE_WRITER"))) {

        if ( (query_cloned.getAttribute("user").equals("")) &&
            (query_cloned.getAttribute("password").equals(""))) {

          query_cloned.setAttribute("user", user);
          query_cloned.setAttribute("password", password);
        }
      }

      for (int j = 0; j < n1.getLength(); j++) {
        Element el = (Element) n1.item(j);
        if ( (el.getAttribute("user").equals("")) &&
            (el.getAttribute("password").equals(""))) {

          el.setAttribute("user", user);
          el.setAttribute("password", password);
        }
      }

      for (int j = 0; j < n2.getLength(); j++) {
        Element el = (Element) n2.item(j);
        if ( (el.getAttribute("user").equals("")) &&
            (el.getAttribute("password").equals(""))) {

          el.setAttribute("user", user);
          el.setAttribute("password", password);
        }
      }
    }
  }


  protected void runTimeChecking(KDDMLObject result, KDDMLObjectType type)
      throws TypeCheckingException {

    if ((!result.getType().equals(KDDMLObjectType.ANY)) &&
        (!type.equals(KDDMLObjectType.ANY)))
      if (!type.equals(result.getType())) {

        TypeCheckingException ecc2 = new TypeCheckingException(query.getTagName(), result.getType());
        output_manager.exitWhitError(ecc2.getDetailErrorMessage());
        throw ecc2;
    }
  }

  protected Element loadQuery(String name, Hashtable params) {
    return null;
  }

  /**
   * Cloning of the input sub-query.
   * When the xml_dest is missing for an input operator, this method adds a new
   * xml_dest attribute labeled with a static temporary counter.
   *
   * @param tag Element
   * @return Element
   */
  public static Element clone(Element tag) {

    boolean repository_loader_op = false;
    for (int i = 0; i < types.length; i++) {
      String tmp = types[i].getOperatorCoreName();
      if ( (tmp != null) &&
           (tag.getTagName().equals(tmp + "_LOADER"))) {
        repository_loader_op = true;
        break;
      }
    }

    // i tag seguenti non ammettono l'attributo xml_dest!!!
    if (repository_loader_op ||
        (tag.getTagName().equals("EXPRESSION")) ||
        (tag.getTagName().equals("CONDITION")) ||
        (tag.getTagName().equals("BASE_COND")) ||
        (tag.getTagName().equals("AND_COND")) ||
        (tag.getTagName().equals("NOT_COND")) ||
        (tag.getTagName().equals("OR_COND")) ||
        (tag.getTagName().equals("TRUE")) ||
        (tag.getTagName().equals("FALSE")) ||
        (tag.getTagName().equals("COND")) ||
        (tag.getTagName().equals("THEN")) ||
        (tag.getTagName().equals("ELSE")) ||
        (tag.getTagName().equals("CALL_QUERY")) ||
        (tag.getTagName().equals("SET")) ||
        (tag.getTagName().equals("ALGORITHM")) ||
        (tag.getTagName().equals("SEQ_QUERY")) ||
        (tag.getTagName().equals("PAR_QUERY")) ||
        (tag.getTagName().equals("PARAM")) ||
        (tag.getTagName().equals("KDD_QUERY")) ||
        (tag.getTagName().equals("KDDML_OBJECT"))) {

      // no xml_dest is necessary...
      // recursion...
      Vector list = XMLUtility.getChildren(tag);
      for (int i = 0; i < list.size(); i++) {
        Element child = (Element) list.get(i);
        Element child_cloned = clone(child);
        tag.appendChild(child_cloned);
      }

      return tag;
    }

    // Nome del file che conterra' il documento XML relativo al risultato prodotto.
    String out = tag.getAttribute("xml_dest");

    if (out == null || out.equals("")) {
      prog_tmp++;
      // l'attributo xml_dest viene generato come file temporaneo...
      out = TMP_STRING + prog_tmp + ".xml";
      tag.setAttribute("xml_dest", out);
    }

    if (out.length() < 4) {
      tag.setAttribute("xml_dest", out + ".xml");
      return tag;
    }

    if (!out.substring(out.length() - 4,
                       out.length()).equalsIgnoreCase(".xml")) {
      // aggiunge l'estensione .xml
      out = out + ".xml";
      tag.setAttribute("xml_dest", out);
    }

    //System.out.println(XMLUtility.toString(tag));

    // recursion...
    Vector list = XMLUtility.getChildren(tag);
    for (int i = 0; i < list.size(); i++) {
      Element child = (Element) list.get(i);
      Element child_cloned = clone(child);
      tag.appendChild(child_cloned);
    }
    return tag;
  }

  private static Condition resolveCond(Element elem)
      throws KDDMLCoreException {

    // ottiene il nodo padre che identifica l'operatore che utilizza la condition
    Element parent = (Element) elem.getParentNode();
    ConditionFactory fact = new ConditionFactory(elem, parent.getNodeName());
    Condition cond = (Condition) fact.newInstance();
    cond.checkConditionValidity();
    return cond;
  }

  private static Expression resolveExpr(Element elem) throws KDDMLCoreException {
    XMLExpressionFactory fact = new  XMLExpressionFactory((Element)elem.getElementsByTagName("*").item(0));
    Expression exp = (Expression) fact.newInstance();
    return exp;
  }


  private static XML resolveGenericXML(Element elem) throws KDDMLCoreException {
    Element child = (Element) elem.getElementsByTagName("*").item(0);
    return new XML(child);
  }

  /**
   * Abilita/disabilita la scrittura del file di log.
   *
   * @param disable boolean
   */
  public void disableLogFileWriting(boolean disable) {
    disable_log_file_writing = disable;
  }

  /**
   * Chiamata alla Garbage Collector per liberare memoria.
   */
  protected static void refreshRAM() {
    System.gc();
  }

  /**
   * Abilita o disabilita il salvataggio dei risultati XML all'interno del repository.
   *
   * @param enable boolean
   */
  public void enableSaving(boolean enable) {
    disable_saving = !enable;
  }

  /**
   * Abilita o disabilita la scrittura su terminale.
   *
   * @param enable boolean
   */
  public void enablePrinting(boolean enable) {
    disable_printing = !enable;
  }

  /**
   * Elimina i file temporanei creati durante l'esecuzione della query.
   */
  public static void removeTmpResults() {
    Runtime.getRuntime().gc();
    System.out.println("Remove all temporaly results....");

    String[] elenco = RepositoryResolver.getRepositoryList();
    for(int i=0; i<elenco.length; i++) {
      File dir = new File(elenco[i]);
      File[] list = dir.listFiles();
      if (list != null)
        for (int j=0;j<list.length;j++) {
          if (list[j].getName().startsWith(QueryExecutor.TMP_STRING))
            System.out.println("Removing "+list[j]+": "+list[j].delete());
        }
    }
  }

  /*public static void setJDBCConnection(JDBCConnection connection) {
    jdbc_conn = connection;
  }*/

  /**
   * Istanziazzione di un generico algoritmo.
   *
   * @param elem Element il tag XML ALGORITHM
   * @param attributes Hashtable tabella hash contenente gli attributi del padre di ALGORITHM
   * @param cnt_alg int numero di algoritmi finora riscontrati
   * @throws KDDMLInterpreterException
   * @return KDDMLAlgorithm l'istanza dell'algoritmo
   */
  private KDDMLAlgorithm resolveAlgorithm(Element elem,
                                          Hashtable attributes,
                                          int cnt_alg)
      throws UnsupportedAlgorithmException {

    cnt_alg++;
    // algorithm tag found
    KDDMLAlgorithm alg = null;
    try {
      // istanzia l'algoritmo attraverso la classe factory del core
      alg = new KDDMLAlgorithm(elem);

      String name = alg.getAlgorithmName();
      Enumeration names = alg.getFormalParameters();

      // richiede al factory del livello operatori se l'alg e' supportato
      AlgorithmFactory fact = new AlgorithmFactory(name);
      try {
        fact.initialize();
      }
      catch (KDDMLOperatorException ex) {
        throw new UnsupportedAlgorithmException(ex.getDetailErrorMessage(), null);
      }
      boolean b1 = fact.isSupported();
      if (!b1) {
        String msg = "Unsupported Algorithm: "+name;
        throw new UnsupportedAlgorithmException(msg, null);
      }

      // si controllano i parametri
      while (names.hasMoreElements()) {
        if (b1) {
          String s = (String) names.nextElement();
          if (!fact.isSupportedParam(s)) {

            // parametro non supportato
            String msg = "Unsupported Algorithm Parameter: "+s;
            throw new UnsupportedAlgorithmException(msg, null);
          }
        }
      }

    }
    catch (KDDMLCoreException ex2) {
      output_manager.exitWhitError(ex2.getDetailErrorMessage());
      final_msg = "error: "+ex2.getDetailErrorMessage();
      throw new UnsupportedAlgorithmException(ex2.getErrorMessage(), ex2.getDetailErrorMessage());
    }

    Enumeration actual_param = alg.getActualParameters();
    Enumeration formal_param = alg.getFormalParameters();

    // le linee di codici seguenti servono per permettere di gestire piu' tag ALGORITHM
    // all'interno di un unico operatore
    String str;
    if (!attributes.containsKey("algorithm_name")) {
      KDDMLScalarManager scalar = new KDDMLScalar(null, alg.getAlgorithmName());
      attributes.put("algorithm_name", scalar);
      str = "algorithm_name";
    }
    else {
      KDDMLScalarManager scalar = new KDDMLScalar(null, alg.getAlgorithmName());
      attributes.put("algorithm_name_" + cnt_alg, scalar);
      str = "algorithm_name_" + cnt_alg;
    }

    // aggiunge ad attributes la lista dei parametri dell'algoritmo
    for (int k=0; k<alg.getNumParameters(); k++) {
      KDDMLScalarManager scal = new KDDMLScalar(null, (String) actual_param.nextElement());
      String key = (String) formal_param.nextElement();
      if (!attributes.containsKey(key))
        attributes.put(key, scal);
      else
        attributes.put(key+"_"+str, scal);
    }
    return alg;
  }

  /**
   * Chiamata al query executor da riga di comando....
   * Come parametro vengono passati il path della direcotory di lavoro e
   * il path contenente la query da eseguire.
   */
  public static void main(String args[]) {
    if (args.length != 1) {
      System.out.println("Usage: ");
      System.out.println("java kddml.Interprete.QueryExecutor query_path");
      System.out.println("Where");
      System.out.println("query_path: path of the KDDML query to execute. ;");
    }
    // percorso della query da eseguire...
    String query = args[0];

    try {
      // parsing della query
      QueryExecutor qe = new QueryExecutor(query);
      // esecuzione della query
      qe.resolve();
    }
    catch(KDDMLInterpreterException ecc1) {
      System.out.println(ecc1.getDetailErrorMessage());
    }
  }
}
