/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators;

import kddml.Operators.*;
import java.util.Collection;
import kddml.Operators.DataMining.AssociationAlgorithms.AssociationAlgorithmFactory;
import kddml.Operators.DataMining.ClusteringAlgorithms.ClusteringAlgorithmFactory;
import kddml.Operators.DataMining.ClassificationAlgorithms.ClassificationAlgorithmFactory;
import kddml.Operators.DataMining.SequenceAlgorithms.SequenceAlgorithmFactory;
import kddml.Operators.Preprocessing.DiscretizationAlgorithms.DiscretizationAlgorithmFactory;
import kddml.Operators.Preprocessing.SamplingAlgorithms.SamplingAlgorithmFactory;
import kddml.Operators.Preprocessing.RewritingAlgorithms.RewritingAlgorithmFactory;
import kddml.Operators.Preprocessing.NormalizationAlgorithms.NormalizationAlgorithmFactory;

/**
 * A factory class for DM and preprocessing algorithms.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class AlgorithmFactory extends AlgorithmEnum {

  private final static String[] DISCRETIZATION_ALGORITHMS =
      {"natural_binning_discretization", "equal_frequency_discretization"};

  private final static String[] NORMALIZATION_ALGORITHMS =
      {"min_max_normalization", "z_score_normalization"};

  private final static String[] REWRITING_ALGORITHMS =
      {"rule_s_rewriting", "rule_t_rewriting", "rule_m_rewriting"};

  private final static String[] SAMPLING_ALGORITHMS =
      {"simple_sampling", "stratified_sampling", "by_cluster_sampling"};

  private final static String[] RDA_ALGORITHMS =
      {"dci"};

  private final static String[] CLASSIFICATION_ALGORITHMS =
      {"yadt", "cage"};

  private final static String[] CLUSTERING_ALGORITHMS =
      {"em", "kmeans"};

  private final static String[] SEQUENCE_ALGORITHMS =
      {"prefix_span"};

  // ADD HERE NEW MINING ALGORITHMS FOR NEW MODELS....
  // modfity then the initialize() method


  protected String alg_name;

  public AlgorithmFactory(String alg_name) {
    this.alg_name = alg_name;
  }

  /**
   * This method contains all KDDML algorithms.
   */
  public static void initialize() throws KDDMLOperatorException {

    enumList.clear();

    // discretization algorithms
    for (int i=0; i<DISCRETIZATION_ALGORITHMS.length; i++) {

      DiscretizationAlgorithmFactory fact = new DiscretizationAlgorithmFactory(DISCRETIZATION_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // normalization algorithms
    for (int i=0; i<NORMALIZATION_ALGORITHMS.length; i++) {

      NormalizationAlgorithmFactory fact = new NormalizationAlgorithmFactory(NORMALIZATION_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // rewriting algorithms
    for (int i=0; i<REWRITING_ALGORITHMS.length; i++) {

      RewritingAlgorithmFactory fact = new RewritingAlgorithmFactory(REWRITING_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // sampling algorithms
    for (int i=0; i<SAMPLING_ALGORITHMS.length; i++) {

      SamplingAlgorithmFactory fact = new SamplingAlgorithmFactory(SAMPLING_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // classification algorithms
    for (int i=0; i<CLASSIFICATION_ALGORITHMS.length; i++) {

      ClassificationAlgorithmFactory fact = new ClassificationAlgorithmFactory(CLASSIFICATION_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // association algorithms
    for (int i=0; i<RDA_ALGORITHMS.length; i++) {

      AssociationAlgorithmFactory fact = new AssociationAlgorithmFactory(RDA_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // clustering algorithms
    for (int i=0; i<CLUSTERING_ALGORITHMS.length; i++) {

      ClusteringAlgorithmFactory fact = new ClusteringAlgorithmFactory(CLUSTERING_ALGORITHMS[i]);
      fact.addAlgorithm();
    }

    // sequences algorithms
    for (int i=0; i<SEQUENCE_ALGORITHMS.length; i++) {

      SequenceAlgorithmFactory fact = new SequenceAlgorithmFactory(SEQUENCE_ALGORITHMS[i]);
      fact.addAlgorithm();
    }


    // ADD HERE NEW MINING ALGORITHMS FOR NEW MODELS....
  }

  /**
   * Returns the list of all implemented algorithms.
   *
   * @return AlgorithmFactory[]
   */
  public static AlgorithmFactory[] list() {

    Collection enums = enumList.values();
    Object objects[] = enums.toArray();
    AlgorithmFactory values2[] = new AlgorithmFactory[enumList.size()];
    System.arraycopy(((Object) (objects)), 0, values2, 0, enumList.size());
    return values2;
  }

  public String getAlgorithmName() {
    return alg_name;
  }

  /**
   * Test if the algorithm is supported by the system.
   * Use the java reflection to found the class.
   *
   * @return boolean
   */
  public boolean isSupported() {
    return super.enumList.containsKey(this.alg_name.toUpperCase());
  }

  /**
   * Test if a parameter is supported by the algorithm.
   *
   * @param param_name String
   * @return boolean
   */
  public boolean isSupportedParam(String param_name) {
    AlgorithmSettingsTask s = newSettingsInstance();
    String[] params = s.getSupportedParameters();
    for (int i=0; i<params.length; i++) {
      if (params[i].equalsIgnoreCase(param_name))
        return true;
    }
    return false;
  }

  public AlgorithmSettingsTask newSettingsInstance() {

    if (!isSupported())
      return null;
    AlgorithmSettingsTask as;
    Class settings_class;
    try {

      AlgorithmFactory alg = (AlgorithmFactory) get(alg_name);

      String package_name = alg.getClass().getPackage().getName()+".";
      String algorithm_name = alg_name.toUpperCase() + "_SETTINGS";

      settings_class = Class.forName(package_name + algorithm_name);

      ///Class param_types[] = new Class[1];
      ///param_types[0] = hashtable.getClass();
      ///Constructor ct = settings_class.getConstructor(param_types);
      ///Object arg_list[] = new Object[1];
      ///arg_list[0] = hashtable;
      ///as = (DMAlgorithmSettings) ct.newInstance(arg_list);

      as = (AlgorithmSettingsTask) settings_class.newInstance();
      return as;

    }
    catch(Exception e) {
      System.out.println(e.toString());
      return null;
    }
  }

  public AlgorithmResolverTask newResolverInstance() {

    if (!isSupported())
      return null;
    AlgorithmResolverTask as;
    Class resolver_class;
    try {

      AlgorithmFactory alg = (AlgorithmFactory) get(alg_name);

      String package_name = alg.getClass().getPackage().getName()+".";
      String algorithm_name = alg_name.toUpperCase() + "_RESOLVER";

      resolver_class = Class.forName(package_name + algorithm_name);
      as = (AlgorithmResolverTask) resolver_class.newInstance();
      return as;

    }
    catch(Exception e) {
      System.out.println(e.toString());
      return null;
    }
  }

  public void addAlgorithm() throws KDDMLOperatorException {
    if (!enumList.contains(alg_name))
      put(this.alg_name, this);
    else
      throw new KDDMLOperatorException(alg_name +" algorithm already exist!",
                                       null,
                                       KDDMLOperatorErrorCodes.ALGORITHM_ALREADY_EXIST);
  }

  public static void main(String args[]) {

    try {
      AlgorithmFactory.initialize();
      AlgorithmFactory fact = new AlgorithmFactory("EM");
      System.out.println(fact.isSupported());
      fact.newSettingsInstance();
      AlgorithmFactory[] list = AlgorithmFactory.list();
      for (int i = 0; i < list.length; i++)
        System.out.println(list[i].alg_name);
    }
    catch(KDDMLOperatorException e) {
      System.out.println(e.getDetailErrorMessage());
    }
  }
}
