/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.DataMining.AssociationAlgorithms;

import weka.core.*;
import java.lang.*;
import java.io.*;
import java.util.*;
import kddml.Core.DataSources.*;
import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.*;
import kddml.Core.DataMining.AssociationRules.TransactionManager;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Operators.*;

/**
 * I metodi di questa classe costruiscono un file di testo a partire dalle tuple
 * contenute nella tabella data in input.
 * Tale file puo' essere usato come input sia per l'algoritmo DCI per l'estrazione di
 * itemsets frequenti sia per l'algoritmo PrefixSpan per l'estrazione di
 * patterns frequenti.
 * L'output di questa classe e' una tabella hash in cui sono contenuti i valori degli items
 * con le rispettive chiavi e un file di testo contenente la chiave corrispondente per
 * ogni singolo attributo di ogni singola istanza.
 * Gli items sono ordinati in modo crescente.
 *
 * Es: per la singola istanza (sunny, false, no) viene generata una tabella hash contenente
 * i seguenti valori:
 * 1 - outlook=sunny
 * 2 - windy=false
 * 3 - play=no.
 * Il file di testo contiene la seguente riga:
 * 1 2 3.
 *
 * Vengono ignorati attributi di tipo numerico.
 *
 * <p>Copyright: Copyright (c) 2001</p>
 * <p>Company: Dipartimento di Informatica - Universit di Pisa</p>
 * @author Elena Ratti (ratti.elena@tiscali.it)
 * @author Daniela Iozzia
 */

final class DCIInputFormatting {

  /**
   * Questa tabella hash contiene le associazioni (item --> id_item) prodotte
   * dalla rienumerazione degli item dove item  il "valore", e id_item  la
   * "chiave".
   */
  private Hashtable items_mapped;

  /**
   * Insieme di istanze in ingresso.
   */
  private InternalTableManager input;

  /**
   * Utilizzato per ordinare gli items.
   */
  private TreeSet ordinaItems;

  /**
   * Numero di items distinti.
   */
  private int num_items;

  /**
   * Numero di transazioni.
   * Nel caso di tabelle orizzantali equivale al numero delle istanze.
   * Nel caso di tabelle verticali e' pari al numero di identificatori di transazione
   * distinti.
   */
  private int num_trans;

  //conterr sia il nome dell'attributo che il suo valore, nella forma "attributo=valore".
  private String nome_item;

  // valore di un item contenuto in una istanza (dove il tipo dell'item  nominale o stringa)
  private String value;

  //file di output per l'algoritmo DCI
  private static OutputStream file;
  private static PrintWriter pw;

  /**
   * Indica se la tabella e' di tipo verticale (transaction, item).
   */
  private boolean is_vertical;

  /**
   * Rappresenta il mining schema della tabella in ingresso.
   */
  private MiningSchema mining_schema;

  /**
   * Rappresenta l'elemento data dictionary.
   */
  private DataDictionary data_dictionary;

  /**
   * Nome del modello.
   */
  private String model_name;

  /**
   * Numero massimo di items in una transazione
   */
  private int max_items;

  /**
   * Numero medio di items in una transazione
   */
  private double avg_items;

  /**
   * Statistiche
   */
  private DataStatistics statistic;

  private String input_file;

  protected DCIInputFormatting(InternalTableManager i, String input_file) {

    // controlla se la tabella e' di tipo transazionale
    is_vertical = i.isTransactionalTable();

    statistic = (DataStatistics) i.getStatistic();
    model_name = statistic.getRelationName();
    max_items = 0;
    avg_items = 0;
    num_items = 0;
    num_trans = 0;
    mining_schema = new MiningSchema();
    data_dictionary = new DataDictionary();
    items_mapped = new Hashtable();
    input = i;
    this.input_file = DCI_RESOLVER.DCI_EXE_DIR + input_file;
  }

  protected Hashtable execute() throws KDDMLOperatorException, KDDMLCoreException {

    // viene controllato che siano presenti almeno due attributi non numerici...
    int s = statistic.getNumberOfNominalAttributes() + statistic.getNumberOfStringAttributes();
    if ((!is_vertical) && (s < 2)) {
        throw new KDDMLOperatorException(
            "Error in DCI: the input table <"+model_name+
            "> must contain al least two non numeric attributes.", null,
            KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    System.out.println("Preprocessing DCI...");
    buildMetadataInformation();

    try {
      file = new FileOutputStream(input_file);
      pw = new PrintWriter(file, true);
    }
    catch (IOException e) {
      String msg1 = "I/O Error calling DCI preprocessing.";
      String msg2 = e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

    // esecuzione della fase di preprocessing
    int num_tuple = input.getNumOfInstances();
    int num_attributi = input.getNumOfAttributes();

    long opt = input.optimize();
    System.out.println("OPT: "+opt);

    if (!is_vertical) {
      System.out.println("Relational Table");
      Instances istanze;
      while (input.hasNext()) {
        // importazione delle istanze
        istanze = (Instances) input.load(opt);
        num_trans = statistic.getNumInstances();
        printInstances(istanze);
      }
    }
    else {
      // caso di tabella verticale
      TransactionManager t;
      System.out.println("Transactional Table");
      long total = 0;
      while(((TransactionalInternalTable) input).hasNextTransaction()) {
        t = ((TransactionalInternalTable) input).loadNextTransaction();
        printTransaction(t);
        num_trans++;

        if (t.getNumberOfItems() > max_items)
          max_items = t.getNumberOfItems();
        total = total + t.getNumberOfItems();
      }
      avg_items = (double) total / (double) num_trans;
    }
    System.out.println("Number of Attributes parsed: "+num_attributi);
    System.out.println("Number of Instances parsed: "+num_tuple);
    System.out.println("Number of Transactions: "+num_trans);
    System.out.println("Number of Items: "+num_items);
    System.out.println("Number of Max Items per Transaction: "+max_items);
    System.out.println("Number of Avg Items per Transaction: "+avg_items);
    System.out.println("Preprocessing output written in: "+input_file);
    System.out.println("-----------------------------");
    try {
      file.flush();
      file.close();
    }
    catch (IOException e) {
      new File(input_file).delete();
      String msg1 = "I/O Error calling DCI preprocessing.";
      String msg2 = "I/O Error calling DCI preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
    return items_mapped;
  }

  protected int getNumberTransactions() {
    return num_trans;
  }

  protected int getMaxNumberOfItemsPerTransaction() {
    return max_items;
  }

  protected double getAvgNumberOfItemsPerTransaction() {
    return avg_items;
  }

  protected MiningSchema getMiningSchema() {
    return mining_schema;
  }

  protected DataDictionary getDataDictionary() {
    return data_dictionary;
  }

  protected String getModelName() {
    return model_name;
  }

  private void printTransaction(TransactionManager trans) throws KDDMLOperatorException {

    // per ogni nuova tupla o itemset ordinaItems conterra' gli item in essa contenuti
    ordinaItems = new TreeSet();

    int total = 1;

    Iterator iter = trans.getItems();
    while(iter.hasNext()) {
      Item i = (Item) iter.next();
      inserisci_TabHash(i.toString());
    }
    // scrive l'istanza nel file
    writeRow();
    ordinaItems.clear();
  }


  /**
   * Questo metodo costruisce un file di testo a partire dalle istanze di una
   * tabella in formato orizzontale.
   * Ogni riga del file contiene gli items di una tupla della tabella di input.
   * Gli items sono rappresentati da numeri sotto forma di stringhe.
   * Gli items di ogni riga sono ordinati in modo crescente e sono distinti.
   * Gli items di ogni riga sono separati da uno spazio bianco.
   *
   * Nel caso di estrazione di pattern sequenziali si suppone che il primo
   * attributo della tabella di input specifichi l'identificatore di cliente.
   *
   * Le associazioni (nome_item_originale --> identificatoreIntero)
   * vengono memorizzate nella tabella hash: "hashInfo_Items" di questa classe.
   */
  private void printInstances(Instances istanze) throws KDDMLOperatorException {

    //nome di un attributo delle istanze
    String attr = "";
    Instance istanza = null;
    try {
      int total = 0;
      for (int j = 0; j < istanze.numInstances(); j++) {
        // Seleziona una istanza
        istanza = istanze.instance(j);
        // la seguente struttura dati serve per ordinare in modo crescente gli items di ogni
        // itemset secondo il formato richiesto dall'algoritmo DCI.
        ordinaItems = new TreeSet();
        int k = 0;
        //ciclo che scandisce gli items di una singola istanza
        int max= 0;
        for (k = 0; k < istanze.numAttributes(); k++) {
          // ignora gli attributi di tipo numerico...
          if (!istanze.attribute(k).isNumeric()) {
            attr = istanze.attribute(k).name();
            if (!istanza.isMissing(istanze.attribute(k))) {
              value = istanza.stringValue(k);
              nome_item = attr + "=" + value;
              inserisci_TabHash(nome_item);
              max++;
            }
          }
        }
        if (max > max_items)
          max_items = max;
        total = total + max;
        writeRow();
      }
      avg_items = (double) total / (double) num_trans;
    }
    catch (Exception e) {
      new File(input_file).delete();
      String msg1 = "I/O Error calling DCI preprocessing.";
      String msg2 = "I/O Error calling DCI preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
  }

  private void writeRow()
      throws KDDMLOperatorException {

    Iterator iter = ordinaItems.iterator();

    //conta il numero di items inseriti nel file
    int c = 0;

    Long itemLong;
    String itemStringa;
    //ciclo while per scrivere sul file gli items dell'itemset
    try {
      while (iter.hasNext()) {

        itemLong = (Long) iter.next();
        itemStringa = String.valueOf(itemLong.longValue());
        pw.print(itemStringa);
        c++;
        //controlla se sono all'ultimo item e quindi non devo inserire lo spazio
        //dopo l'inserimento dell'item
        if (c < ordinaItems.size())
          pw.print(" ");
      }
    }
    catch (Exception e) {
      new File(input_file).delete();
      String msg1 = "I/O Error calling DCI preprocessing.";
      String msg2 = "I/O Error calling DCI preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
    pw.println();
  }

  protected void inserisci_TabHash(String nome_item) {

    if (!items_mapped.containsValue(nome_item)) {
      // non esiste ancora un'associazione fra l'item e un valore numerico
      // cos viene creata all'interno della tabella hash
      String chiave;
      chiave = (String) Integer.toString(num_items);
      num_items++;
      items_mapped.put(chiave, nome_item);
      this.ordinaItems.add(new Long(chiave));
    }
    else {
      Enumeration enum2 = items_mapped.keys();
      //mi serve per scandire la tabella hash hashInfo_Items
      //per recuperare la chiave associata a nome_item
      String chiaveItem;
      boolean trovato = false;
      // trovato diventa true se .....   gi presente nella tabella hash
      //scorro la tabella hash:
      while ( (enum2.hasMoreElements())) {
        String current_key = (String) enum2.nextElement();
        if (items_mapped.get(current_key).equals(nome_item)) {
          this.ordinaItems.add(new Long(current_key));
          // inseriamo nel treeSet come long in modo tale da non doversi preoccupare
          // dell'ordinamento che viene gestito automaticamente dal treeSet
          break;
        }
      }
    }
  }

  protected void buildMetadataInformation() {
    String name;
    AttributeStatisticManager attr;
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      name = attr.getName();
      if (attr.getType().equals(AttributeType.NUMERIC)) {
        data_dictionary.addField(new DataField(name, DataField.CONTINUOUS));
        if (is_vertical && name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
          mining_schema.addField(new MiningField(TransactionalInternalTable.TRANSACTION_ATTRIBUTE,
                                                 MiningField.ACTIVE));
        else
          mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));
      }
      else {
        if (attr.getType().equals(AttributeType.NOMINAL)) {
          // attributo di tipo enumerato
          DataField mf = new DataField(name, DataField.CATEGORICAL);

          Enumeration enum2 = ((NominalStatistic) attr).getValues();
          while(enum2.hasMoreElements()) {
            try {
              mf.addEnumeratedValue( (String) enum2.nextElement());
            }
            catch (MiningModelException ex) {
            }
          }
          data_dictionary.addField(mf);
          if (is_vertical) {
            if (name.equalsIgnoreCase(TransactionalInternalTable.EVENT_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.PREDICTED));
            else
            if (name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
              mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));
          }
          else
            mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
        }
        else {
          // attributo di tipo stringa
          data_dictionary.addField(new DataField(name));
          if (is_vertical) {
            if (name.equalsIgnoreCase(TransactionalInternalTable.EVENT_ATTRIBUTE))
              mining_schema.addField(new MiningField(name,
                  MiningField.PREDICTED));
            else
            if (name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
              mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));
          }
          else
            mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
        }
      }
    }
  }

  public static void main(String args[]) {
    File f = new File("C:/kddml/pippo.xml");
    try {
      ARFFFactory fact = new ARFFFactory(f, "C:/kddml/Repository/Data/ARFF/coop_transactional.arff");
      fact.setOverwrite(true);
      InternalTable t = (InternalTable) fact.newInstance();
      t.saveToRepository();
      DCIInputFormatting dci = new DCIInputFormatting(t, "ppp");
      System.out.println(dci.execute().toString());
    }
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
    catch (KDDMLOperatorException ex2) {
      System.out.println(ex2.getDetailErrorMessage());
    }
  }
}
