/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.DataMining.AssociationAlgorithms;

import java.lang.*;
import java.io.*;
import java.util.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Operators.*;

/**
 * Questa classe estrae gli insiemi frequenti e relativi supporti dal
 * file di output prodotto dall'algoritmo DCI.
 * Vengono successivamente estratte le regole di associazione a partire dagli
 * itemsets estratti.
 */

final class DCIOutputFormatting {

  /**
   * Numero di transazioni scandite dall'algoritmo DCI.
   */
  private int num_transactions;

  /**
   * Numero massimo di itemset desiderati.
   * Parametro fornito in input dall'utente.
   */
  private int num_max_of_itemset;

  /**
   * Insieme di itemsets ordinati in base al supporto.
   */
  private TreeSet itemsets;

  private String output_file;

  /**
   * Costruttore.
   * @param ht hashtable contenente le associazioni chiave --> item. Generata in fase
   * di preprocessing.
   * @param num_trans numero di transazioni.
   * @param num_max_itemset numero massimo di itemsets che devono essere estratti.
   */
  public DCIOutputFormatting(int num_trans, int num_max_itemset, String output_file) {
    if (num_max_itemset < AssociationModel.MAX_NUMBER_OF_ITEMSETS)
      num_max_of_itemset = num_max_itemset;
    else
      num_max_of_itemset = AssociationModel.MAX_NUMBER_OF_ITEMSETS;
    num_transactions = num_trans;
    CompareItemSets comparator = new CompareItemSets();
    itemsets = new TreeSet(comparator);
    this.output_file = DCI_RESOLVER.DCI_EXE_DIR + output_file;
  }

  /**
   * Costruisce gli itemsets parsando il file prodotto in uscita dalla libreria DCI.
   * Ogni riga del file contiene gli itemsets nella forma:
   * 3 5 12 4 2 (2)
   * dove tra parentesi viene specificato il supporto associato all'itemset e i valori
   * iniziali rappresentano le chiavi associate agli item contenuti nella tabella hash
   * relativa.
   *
   * @param ht tabella hash contenente le associazioni chiave --> item.
   */
  public TreeSet buildItemsets(Hashtable ht) throws KDDMLOperatorException {

    String inputLine;
    int i = 1;
    try {
      FileReader reader = new FileReader(output_file);
      BufferedReader in = new BufferedReader(reader);
      inputLine = in.readLine();
      Itemset item_set;
      while (inputLine != null) {
        i++;
        // ignoriamo l'itemset frequente vuoto
        if (inputLine.startsWith("(")) {
          inputLine = in.readLine();
          if (inputLine == null) {
            // caso di itemsets vuoto: exit with empty itemsets
            return itemsets;
          }
        }

        StringTokenizer t1 = new StringTokenizer(inputLine, "()");
        StringTokenizer t2 = new StringTokenizer(t1.nextToken()," ");
        int num_items = t2.countTokens();

        item_set = new Itemset();
        while (t2.hasMoreTokens()) {
          // identificatore dell'item
          String id_item = t2.nextToken();
          // valore dell'item ottenuto attraverso la tabella hash
          String value = (String) ht.get(id_item);

          item_set.addItem(new Item(value));
        }
        String support = t1.nextToken().trim();
        double supp = new Double(support).doubleValue();
        // calcola il supporto in forma percentuale...
        double supp2 = supp / (double) num_transactions;
        item_set.setSupport(supp2);

        // aggiunge l'itemset al tree set mantenendo l'ordinamento stabilito...
        if (itemsets.size() == num_max_of_itemset) {
          itemsets.add(item_set);
          itemsets.remove((Itemset) itemsets.last());
        }
        else {
          itemsets.add(item_set);
        }
        inputLine = in.readLine();
      }
      in.close();
    }
    catch (Exception e) {
      String msg1 = "I/O Error calling DCI postprocessing.";
      String msg2 = e.toString();
      new File(output_file).delete();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
    return itemsets;
  }
}
