/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.DataMining.AssociationAlgorithms;

import java.io.*;
import java.util.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Core.DataSources.*;
import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.*;
import kddml.Operators.*;

/**
 * Classe che invoca l'algoritmo DCI ed estrae gli itemsets frequenti
 * dalle istanze di input.
 */
public class DCI_RESOLVER implements AssociationAlgorithmResolverTask {

  protected final static String DCI_EXE_DIR =
      System.getProperty("user.dir") + File.separator + "resources" +
      File.separator + "lib" + File.separator + "algorithms" +
      File.separator + "DCI" + File.separator;

  /**
   * Input file for DCI: path in DCI home
   */
  protected String input_file;

  /**
   * Output file for DCI: path in DCI home
   */
  protected String output_file;

  String xml_dest;

  /**
   * Minimo supporto in forma percentuale passato dall'utente come parametro.
   */
  private double min_support;

  /**
   * Minima confidenza in forma percentuale passata dall'utente come parametro.
   */
  private double min_confidence;

  /**
   * Numero massimo di itemsets che devono essere estratte.
   * Parametro opzionale fornito in input dall'utente.
   */
  private int num_max_itemsets;

  /**
   * Numero massimo di regole che devono essere estratte.
   * Parametro opzionale fornito in input dall'utente.
   */
  private int num_max_rules;

  /**
   * Numero di transazioni esaminate durante la fase di preprocessing.
   */
  private int num_trans;

  /**
   * Nome del modello
   */
  private String model_name;

  /**
   * Elemento mining schema
   */
  private MiningSchema mining_schema;

  /**
   * Elemento data dictionary
   */
  private DataDictionary data_dictionary;

  /**
   * Numero massimo di items per transazione.
   */
  private int max_items;

  /**
   * Numero medio di items per transazione.
   */
  private double avg_items;

  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLOperatorException, KDDMLCoreException {

    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_ASSOCIATION_MODEL.getRepositoryPath() +
        ((DCI_SETTINGS) settings).getXMLDestination();

    File f = new File(xml_dest);
    input_file = "___dci_input_" /*+ f.getName()*/ + ".tmp";
    output_file = "___dci_output_" /*+ f.getName()*/ + ".tmp";

    min_support = ((DCI_SETTINGS) settings).getMinSupport().doubleValue();
    min_confidence = ((DCI_SETTINGS) settings).getMinConfidence().doubleValue();

    if (((DCI_SETTINGS) settings).getMaxNumberOfItemsets() != null)
      num_max_itemsets = ((DCI_SETTINGS) settings).getMaxNumberOfItemsets().intValue();
    else
      num_max_itemsets = AssociationModel.MAX_NUMBER_OF_ITEMSETS;

    if (((DCI_SETTINGS) settings).getMaxNumberOfRules() != null)
      num_max_rules = ((DCI_SETTINGS) settings).getMaxNumberOfRules().intValue();
    else
      num_max_rules = AssociationModel.MAX_NUMBER_OF_RULES;

    System.out.println("------------------------------------------");
    System.out.println("DCI PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("Min Support - "+min_support);
    System.out.println("Min Confidence - "+min_confidence);
    System.out.println("Max Rules - "+num_max_rules);
    System.out.println("Min Itemsets - "+num_max_itemsets);
    System.out.println("------------------------------------------");
  }

  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    DCIInputFormatting preprocessing = new DCIInputFormatting(instances, input_file);
    Hashtable items_mapped = preprocessing.execute();

    num_trans = preprocessing.getNumberTransactions();
    max_items = preprocessing.getMaxNumberOfItemsPerTransaction();
    avg_items = preprocessing.getAvgNumberOfItemsPerTransaction();
    model_name = preprocessing.getModelName();
    mining_schema = preprocessing.getMiningSchema();
    data_dictionary = preprocessing.getDataDictionary();

    // Tabella hash contenente le associazioni chiave --> item.
    // Generata durante la fase di pre-processing.
    return items_mapped;
  }

  public Object execute(Object hash_table)
      throws KDDMLOperatorException, KDDMLCoreException {

    // chiamata della libreria DCI
    // ottiene il supporto in forma assoluta...
    double s = num_trans * min_support;
    int min_absolute_support = (int) s;
    if (min_absolute_support == 0)
      min_absolute_support = 1;
    System.out.println("Absolute Minimum Support: " + min_absolute_support);

    int result = -1;
    try {
      if (System.getProperty("os.name").indexOf("Windows") != -1) {

        // Windows operative System
        result = callDCIForWindows(new Integer(min_absolute_support).toString());
      }

      if (System.getProperty("os.name").indexOf("Linux") != -1) {

        // Linux operative System
        result = callDCIForLinux(new Integer(min_absolute_support).toString());

      }
    }
    catch (Exception ec) {
      String msg1 = "Error invoking DCI libraries: invalid command!!!";
      String msg2 = ec.toString();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }

    if (result != 0)
      throw new KDDMLOperatorException("Error invoking DCI libraries: invalid command!!!", null,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);

    System.out.println("DCI output created in: "+ output_file);
    return hash_table;
  }

  public AssociationModel outputFormatting(Object items_mapped)
      throws KDDMLOperatorException, KDDMLCoreException  {

    System.out.println("Generating Frequents itemsets from DCI output... please wait!!!");
    DCIOutputFormatting post = new DCIOutputFormatting(num_trans, num_max_itemsets, output_file);
    TreeSet itemsets = post.buildItemsets((Hashtable) items_mapped);

    System.out.println("Number of frequent itemsets founds: "+itemsets.size());

    System.out.println("Generating Association Rules... please wait!!!");
    // construct an associaton model by frequent itemsets...

    File f = new File(xml_dest);
    ItemsetAssociationFactory fact = new ItemsetAssociationFactory(
      f, data_dictionary, mining_schema, min_support, min_confidence, num_trans, itemsets);

    // setta le proprieta' del factory
    fact.setOverwrite(true);
    fact.setAlgorithmName("DCI - Direct Count with Intersection - ISTI-CNR, Pisa, Italy");
    fact.setModelName(model_name+"_rda");
    fact.setMaxNumberOfItemsPerTA(max_items);
    fact.setAvgNumberOfItemsPerTA(avg_items);
    fact.setMaxNumberOfRules(num_max_rules);

    AssociationModel output = (AssociationModel) fact.newInstance();
    System.out.println("Association Rules founds: "+output.getNumberOfRules());
    return output;
  }

  public String[] getTmpFiles() {
    String[] ris = {DCI_EXE_DIR + input_file, DCI_EXE_DIR + output_file, DCI_EXE_DIR + "dataset.tmp"};
    return ris;
  }

  private int callDCIForWindows(String min_absolute_support)
      throws IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();

    String exec = "cmd /c start /MIN /D " + "\"" + DCI_EXE_DIR + "\"" + " /wait"+ " " + "dci" + " "+
        input_file + " " + min_absolute_support + " " + output_file;

    System.out.println("Invoking DCI libraries...");
    //System.out.println(exec);

    Process p = esegui.exec(exec);
    p.waitFor();

    return p.exitValue();

  }

  private int callDCIForLinux(String min_absolute_support)
      throws IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();
    String[] exec = {
        DCI_EXE_DIR + "dci",
        DCI_EXE_DIR + input_file,
        min_absolute_support,
        DCI_EXE_DIR + output_file};

    System.out.println("Invoking DCI libraries...");
    //System.out.println(exec);

    Process p = esegui.exec(exec);
    p.waitFor();

    return p.exitValue();

  }
}
