/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.DataMining.ClassificationAlgorithms;

import kddml.Core.DataSources.*;
import kddml.Core.*;
import kddml.Operators.*;
import weka.core.*;
import kddml.Core.DataSources.*;
import java.io.*;
import kddml.Core.DataMining.ClassificationTrees.*;

/**
 * This class contains the implementation of the CAGE algorithm.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CAGE_RESOLVER implements ClassificationAlgorithmResolverTask {

  protected final static String CAGE_EXE =
      System.getProperty("user.dir") + File.separator + "resources" +
      File.separator + "lib" + File.separator + "algorithms" +
      File.separator + "CAGE" + File.separator;

  /**
   * Input file for CAGE: path in CAGE home
   */
  protected final static String INPUT_DATA = "input_cage.arff";

  /**
   * Output file for CAGE: path in CAGE home
   */
  protected final static String OUTPUT = "output_cage.xml";

  /**
   * The xml destination in the system repository.
   */
  private String xml_dest;

  private int num_processors;

  private int num_iterations;

  private boolean has_boosting;

  private File parameter_file;

  private Double percentage_split;

  /**
   * The target attribute
   */
  private String target_attribute_name;

  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLCoreException, KDDMLOperatorException {

    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_CLASSIFICATION_MODEL.getRepositoryPath() +
        ((CAGE_SETTINGS) settings).getXMLDestination();

    target_attribute_name = ((CAGE_SETTINGS) settings).getTargetAttributeName();

    num_processors = ((CAGE_SETTINGS) settings).getNumProcessors();
    num_iterations = ((CAGE_SETTINGS) settings).getNumIterations();
    has_boosting = ((CAGE_SETTINGS) settings).hasBoosting();
    percentage_split = ((CAGE_SETTINGS) settings).getPercentageSplitting();
    parameter_file = ((CAGE_SETTINGS) settings).getParametersFile();

    System.out.println("------------------------------------------");
    System.out.println("CAGE PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("Target Attribute Name - "+target_attribute_name);
    System.out.println("Number of Processors - "+num_processors);
    System.out.println("Number of Iterations - "+num_iterations);
    System.out.println("Boosting technique - "+has_boosting);
    System.out.println("Percentage Splitting - "+percentage_split);
    if (parameter_file != null)
      System.out.println("Parameter File - "+parameter_file.toString());
    else
      System.out.println("Parameter File - None");
    System.out.println("------------------------------------------");

  }

  /**
   * Input formatting. It takes the input internal table and returns the ARFF file as
   * output. ARFF file is stored in CAGE_EXE + INPUT_DATA.
   *
   * @param instances InternalTableManager the input KDDML table.
   * @throws KDDMLOperatorException if an I/O error occours.
   * @throws KDDMLCoreException
   * @return Object a string containing the input file path.
   */
  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    System.out.println("Total number of attributes: "+instances.getNumOfAttributes());
    System.out.println("Total number of instances: "+instances.getNumOfInstances());

    DataStatisticsManager statistic = instances.getStatistic();

    if (statistic.getNumberOfNumericAttributes() != 0)
      throw new KDDMLOperatorException("Algorithm CAGE works on discrete attributes only",
                                       null,
                                       KDDMLOperatorErrorCodes.INVALID_SETTINGS);

    if (statistic.getNumberOfStringAttributes() != 0)
      throw new KDDMLOperatorException("Algorithm CAGE works on discrete attributes only",
                                       null,
                                       KDDMLOperatorErrorCodes.INVALID_SETTINGS);


    FileWriter output = null;
    int num_tuple = 0;
    try {
      long opt = instances.optimize();
      output = new FileWriter(CAGE_EXE + INPUT_DATA);

      // schema import...
      Instances tuple = (Instances) instances.load(0);
      output.write(tuple.toString());

      while (instances.hasNext()) {

        // instances import...
        tuple = (Instances) instances.load(opt);
        num_tuple += opt;
        for (int i=0; i<tuple.numInstances(); i++)
          output.write(tuple.instance(i).toString()+"\n");
      }
      output.close();
    }
    catch(Exception e) {
      throw new KDDMLOperatorException("I/O Error!", e.toString(), KDDMLOperatorErrorCodes.IO_ERROR);
    }
    System.out.println("Input CAGE generated in: "+CAGE_EXE + INPUT_DATA);
    return CAGE_EXE + INPUT_DATA;
  }

  /**
   * Call the CAGE algorithm and generates the PMML model.
   *
   * @param input Object null
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return Object a string containing the path of PMML file.
   */
  public Object execute(Object input)
      throws KDDMLOperatorException, KDDMLCoreException {

    boolean esito = false;//executeCAGE((String) input);

    try {
      if (System.getProperty("os.name").indexOf("Windows") != -1) {

        // Windows operative System
        esito = callCAGEForWindows();
      }

      if (System.getProperty("os.name").indexOf("Linux") != -1) {

        // Linux operative System
        esito = callCAGEForLinux();

      }
    }
     catch (Exception ec) {
      String msg1 = "Error invoking CAGE  libraries: invalid command!!!";
      String msg2 = ec.toString();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }




    if (!esito) {
      // error!!!
      throw new KDDMLOperatorException(
          "Error invoking CAGE libraries: invalid command!!!", null,
          KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }

    // tutto OK!!!
    // nel file CAGE_EXE + OUTPUT e' contenuto l'albero di classificazione prodotto dall'algoritmo
    System.out.println("CAGE output created in: "+ CAGE_EXE + OUTPUT);

    return CAGE_EXE + OUTPUT;
  }

  /**
   * Takes the PMML files generated by CAGE and returns the model object as output.
   *
   * @param tree Object null
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return TreeModel the KDDML object
   */
  public TreeModel outputFormatting(Object tree)
      throws KDDMLOperatorException, KDDMLCoreException  {

    String pmml_path = (String) tree;
    File f = new File(pmml_path);
    if (!f.exists())
      throw new KDDMLOperatorException("Error invoking CAGE: no output generated.",
                                       null,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    PMMLClassificationFactory fact = new PMMLClassificationFactory(new File(xml_dest), pmml_path);

    // sets the factory overwriting...
    fact.setOverwrite(true);

    TreeModel output = (TreeModel) fact.newInstance();
    return output;
  }

  public String[] getTmpFiles() {
    String[] res = {CAGE_EXE + INPUT_DATA, CAGE_EXE + OUTPUT};
    return res;
  }

 /**
  * Call the CAGE algorithm. Returns false if an error occours.
  * It writes in CAGE_EXE + OUTPUT the PMML file containing the result.
  *
  * @param arff_file String the path containing the input ARFF file.
  * @throws KDDMLOperatorException if an error occours.
  * @return boolean returns true if the output XML file has been computed. Returns
  * false if an error occours.
  */
 /*private boolean executeCAGE(String arff_file) throws KDDMLOperatorException {

    throw new KDDMLOperatorException("Algorithm CAGE is not yet implemented.",
                                      null,
                                     KDDMLOperatorErrorCodes.NOT_YET_IMPLEMENTED);
  }
   */

   private boolean callCAGEForLinux()
      throws java.io.IOException, InterruptedException {




    Runtime esegui = Runtime.getRuntime();


    /*
     private String xml_dest;

  private int num_processors;

  private int num_iterations;

  private boolean has_boosting;

  private File parameter_file;

  private Double percentage_split;
  */
     String[] exec = {
          CAGE_EXE +"cage-run",
          "wead",   // ok CAGE_EXE + INPUT_DATA+INPUT_NAMES
         new Integer(num_processors).toString() ,
          new Integer(1).toString(),   // ignored
	  new Integer(num_iterations).toString(),
	  "none",   // ok parameter_file
	  new Integer(34666).toString(),
	  CAGE_EXE + OUTPUT};

 //cage-run wead 2 1 100 none 453545 prova_tmp.xml
      /*   String[] exec = {
          CAGE_EXE + "dTcmd",
          "-fm",
          CAGE_EXE + INPUT_NAMES,
          "-fd",
          CAGE_EXE + INPUT_DATA,
          "-m",
	   CAGE_EXE + OUTPUT};*/


      for (int i = 0; i < exec.length; i++)
        System.out.println(exec[i]);

      System.out.println("Invoking CAGE  libraries...");
      System.out.print(exec);

      Process p = esegui.exec(exec);
      p.waitFor();
    System.out.println("End of  invocation of CAGE  libraries...");
      return (p.exitValue()>=0);
      }

   private boolean callCAGEForWindows() throws KDDMLOperatorException {
     throw new KDDMLOperatorException("Algorithm CAGE is not yet implemented for Windows Operating System.",
                                      null,
                                     KDDMLOperatorErrorCodes.NOT_YET_IMPLEMENTED);
      }




}
