/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.DataMining.ClassificationAlgorithms;

import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.DataMining.*;
import kddml.Operators.*;
import java.util.Hashtable;
import kddml.Core.KDDMLScalarManager;
import java.io.File;

public class CAGE_SETTINGS extends TREE_MINER_SETTINGS implements AlgorithmSettingsTask {

  /**
   * The number of processors. A positive integer. Default 1.
   */
  private final static String NUM_PROCESSORS = "num_processors";

  /**
   * The maximum number of iteration needed. Default 100.
   */
  private final static String NUM_ITERATIONS = "num_iterations";

  /**
   * The name of a configuration file. It can be optional.
   */
  private final static String PARAMETER_FILE = "parameter_file";

  /**
   * The classification type. Can be boost or no_boost. Default no_boost.
   */
  private final static String CLASSIFICATION_TYPE = "classification_type";

  /**
   * The percentage of input instances used by CAGE. Double in (0,1].
   * It is used only with boosting technique.
   */
  private final static String PERC_DATA = "perc_data";

  private int num_processors;

  private int num_iterations;

  private boolean has_boosting;

  private File parameter_file;

  public CAGE_SETTINGS() {
    super();
    num_processors = 1;
    num_iterations = 100;
    has_boosting = false;
    parameter_file = null;
  }

  /**
   * Returns the list of supported parameters.
   *
   * @return String[]
   */
  public String[] getSupportedParameters() {
    String[] s = {
        NUM_PROCESSORS, NUM_ITERATIONS, PARAMETER_FILE, CLASSIFICATION_TYPE, PERC_DATA};
    return s;
  }

  public void applyAlgorithmSettings(java.util.Hashtable attributes)
      throws KDDMLOperatorException {

    if (System.getProperty("os.name").indexOf("Linux") == -1) {

      // no linux operative System
      throw new KDDMLOperatorException("Algorithm CAGE is available on LINUX platform only!",
                                       null,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);

    }

    super.setTargetAttribute(attributes, "target_attribute", true);

    setNumProcessors(attributes, NUM_PROCESSORS, false);
    setNumIterations(attributes, NUM_ITERATIONS, false);
    setClassificationType(attributes, CLASSIFICATION_TYPE, false);
    super.setPercentageSplitting(attributes, PERC_DATA, false);

    setParametersFile(attributes, PARAMETER_FILE, false);

    // controlla che la pecentuale sia settata solo con il boosting
    if ((!has_boosting) && (percentage_splitting != null))
      throw new KDDMLOperatorException
          ("< "+PERC_DATA+" >"+" is used with boosting technique only!",
           null,
           KDDMLOperatorErrorCodes.INVALID_SETTINGS);
  }

  /**
   * Sets the number of processors.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setNumProcessors(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        num_processors = (new Integer(((Double) s.getExprValue()).intValue())).intValue();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (num_processors < 1)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Set the number of processors.
   *
   * @param num int a positive integer value.
   */
  public void setNumProcessors(int num) {
    num_processors = num;
  }

  /**
   * Returns the number of processors.
   *
   * @return int a positive integer value.
   */
  public int getNumProcessors() {
    return num_processors;
  }

  /**
   * Sets the number of iterations.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setNumIterations(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        num_iterations = (new Integer(((Double) s.getExprValue()).intValue())).intValue();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (num_iterations < 1)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Set the number of iterations.
   *
   * @param num int a positive integer value.
   */
  public void setNumIterations(int num) {
    num_iterations = num;
  }

  /**
   * Returns the number of iterations.
   *
   * @return int a positive integer value.
   */
  public int getNumIterations() {
    return num_iterations;
  }

  /**
   * Sets the classification type: boost, no_boost.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setClassificationType(Hashtable attributes,
                                       String attribute_name,
                                       boolean error_if_not_found) throws
      KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      String str = null;
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        str = ((String) s.getExprValue());
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a string with 'boost' or 'no_boost' values.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (str.equalsIgnoreCase("boost")) {
        has_boosting = true;
        return;
      }
      if (str.equalsIgnoreCase("no_boost")) {
        has_boosting = false;
        return;
      }
      throw new KDDMLOperatorException
          ("< " + attribute_name + " >" + " must be a string with 'boost' or 'no_boost' values.",
           null,
           KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Returns true if the algorithm uses the boosting technique.
   *
   * @return boolean
   */
  public boolean hasBoosting() {
    return has_boosting;
  }

  /**
   * Sets the boosting technique for the algorithm.
   *
   * @param boosting boolean
   */
  public void setBoosting(boolean boosting) {
    has_boosting = boosting;
  }

  /**
   * Sets the parameters file.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setParametersFile(Hashtable attributes,
                                   String attribute_name,
                                   boolean error_if_not_found)
      throws KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      String str = null;
      KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
      str = CAGE_RESOLVER.CAGE_EXE + ((String) s.getExprValue());
      parameter_file = new File(str);
      if (!parameter_file.exists())
        throw new KDDMLOperatorException
            ("The specified file < "+str+" > do not exist!",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Returns the parameter file of null if it is not used.
   *
   * @return File
   */
  public File getParametersFile() {
    return parameter_file;
  }

  /**
   * Sets the parameter file used by algorithm.
   *
   * @param file File
   */
  public void setParameterFile(File file) {
    parameter_file = file;
  }
}
