/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.DataMining.ClassificationAlgorithms;

import weka.core.*;
import java.lang.*;
import java.io.*;
import java.util.*;
import kddml.Core.DataSources.*;
import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.*;
import kddml.Core.DataMining.AssociationRules.Transaction;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Operators.*;

/**
 * Formatting for YADT algorithm.
 */
class YADTInputFormatting {

  /**
   * Insieme di istanze in ingresso.
   */
  private InternalTableManager input;

  // files di input per l'algoritmo YADT
  private static PrintWriter pw_names;
  private static PrintWriter pw_data;

  /**
   * Nome del modello.
   */
  private String model_name;

  /**
   * Statistiche
   */
  private DataStatistics statistic;

  /**
   * The target attribute
   */
  private String target;

  private String input_data;

  private String input_names;


  protected YADTInputFormatting(InternalTableManager i,
                                String target_attribute,
                                String target_file_data,
                                String target_file_names) {

    statistic = (DataStatistics) i.getStatistic();
    model_name = statistic.getRelationName();
    input = i;
    target = target_attribute;
    this.input_data = target_file_data;
    this.input_names = target_file_names;
  }

  protected void execute() throws KDDMLOperatorException, KDDMLCoreException {

    // viene controllato che siano presenti almeno due attributi non numerici...
    int s = statistic.getNumberOfStringAttributes();
    if (s > 0) {
        throw new KDDMLOperatorException(
            "Error in YaDT: the input table <"+model_name+
            "> can't contains string attributes!", null,
            KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    AttributeStatisticManager attr= statistic.getStatistic(target);
    int m = attr.getMissingCount();

    if (m > 0) {
        throw new KDDMLOperatorException(
            "Error in YaDT: the attribute <"+target+
            "> can't contain missing values!", null,
            KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    System.out.println("Preprocessing YaDT...");

    FileOutputStream file_names = null;
    FileOutputStream file_data = null;

    try {
      file_names = new FileOutputStream(YADT_RESOLVER.YADT_EXE + input_names);
      pw_names = new PrintWriter(file_names, true);

      file_data = new FileOutputStream(YADT_RESOLVER.YADT_EXE + input_data);
      pw_data = new PrintWriter(file_data, true);
    }
    catch (IOException e) {
      String msg1 = "I/O Error calling YaDT preprocessing.";
      String msg2 = "I/O Error calling YaDT preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

    // esecuzione della fase di preprocessing
    int num_tuple = input.getNumOfInstances();
    int num_attributi = input.getNumOfAttributes();

    long opt = input.optimize();

    Instances istanze;
    // print the schema in .metadata file
    printSchema((Instances) input.load(0));

    while (input.hasNext()) {
      // importazione delle istanze
      istanze = (Instances) input.load(opt);

      printInstances(istanze);
    }
    System.out.println("Number of Attributes parsed: "+num_attributi);
    System.out.println("Number of Instances parsed: "+num_tuple);
    System.out.println("Preprocessing output written in: "+YADT_RESOLVER.YADT_EXE + input_names);
    System.out.println("-----------------------------");
    try {
      file_names.flush();
      file_names.close();
      file_data.flush();
      file_data.close();
    }
    catch (IOException e) {
      String msg1 = "I/O Error calling YadT preprocessing.";
      String msg2 = e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
  }

  protected MiningSchema getMiningSchema() {
    String name;
    AttributeStatisticManager attr;
    MiningSchema ris = new MiningSchema();
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      name = attr.getName();
      if (name.equalsIgnoreCase(target))
        ris.addField(new MiningField(name, MiningField.PREDICTED));
      else
        ris.addField(new MiningField(name, MiningField.ACTIVE));
    }
    return ris;
  }

  protected DataDictionary getDataDictionary() {
    return new DataDictionary(statistic);
  }

  protected String getModelName() {
    return model_name;
  }

  /**
   * Stampa lo schema dei dati come file .names.
   */
  private void printSchema(Instances istanze) throws KDDMLOperatorException {

    Attribute a;
    String row;
    try {
      for (int i = 0; i < istanze.numAttributes(); i++) {
        a = istanze.attribute(i);
        // controlla se abbiamo l'attributo di classificazione
        if (a.name().equalsIgnoreCase(target))
          row = a.name() + ",string,class";
        else
        if (a.isNumeric())

          // numeric attribute
          row = a.name() + ",float,continuous";
        else

          // nominal attribute
          row = a.name() + ",string,discrete";
        pw_names.print(row);
        pw_names.println();
      }
    }
    catch (Exception e) {
      String msg1 = "I/O Error calling YADT preprocessing.";
      String msg2 = "I/O Error calling YADT preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
  }

  /**
   * Questo metodo costruisce un file di testo .data a partire dalle istanze di una
   * tabella.
   */
  private void printInstances(Instances istanze) throws KDDMLOperatorException {

    Instance istanza = null;
    try {
      for (int i=0; i<istanze.numInstances(); i++) {
        istanza = istanze.instance(i);
        pw_data.print(istanza.toString());
        pw_data.println();
      }
    }
    catch (Exception e) {
      String msg1 = "I/O Error calling YADT preprocessing.";
      String msg2 = "I/O Error calling YADT preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }
  }

  public static void main(String args[]) {
    File f = new File("C:/kddml/pippo.xml");
    try {
      ARFFFactory fact = new ARFFFactory(f, "C:/kddml/Repository/Data/ARFF/weather.arff");
      fact.setOverwrite(true);
      InternalTable t = (InternalTable) fact.newInstance();
      t.saveToRepository();
      YADTInputFormatting yadt = new YADTInputFormatting(t, "play", "ppp1", "ppp2");
      yadt.execute();
    }
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
    catch (KDDMLOperatorException ex2) {
      System.out.println(ex2.getDetailErrorMessage());
    }
  }
}
