/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.DataMining.ClassificationAlgorithms;

import kddml.Core.DataSources.*;
import kddml.Core.*;
import kddml.Core.DataMining.*;
import kddml.Operators.*;
import java.io.File;
import kddml.Core.DataMining.ClassificationTrees.*;

public class YADT_RESOLVER implements ClassificationAlgorithmResolverTask {

  protected final static String YADT_EXE =
      System.getProperty("user.dir") + File.separator + "resources" +
      File.separator + "lib" + File.separator + "algorithms" +
      File.separator + "YaDT" + File.separator;

  /**
   * Input file for YaDT: path in YaDT home
   */
  protected String input_file_data;

  /**
   * Output file for YaDT: path in YaDT home
   */
  protected String input_file_names;

  /**
   * Output file for YaDT: path in YaDT home
   */
  protected String output_file;

  private String xml_dest;

  private String attribute_name;
  /**
   * Nome del modello
   */
  private String model_name;

  /**
   * Elemento mining schema
   */
  private MiningSchema mining_schema;

  /**
   * Elemento data dictionary
   */
  private DataDictionary data_dictionary;

  private double confidence_for_pruning;

  private int num_instances_for_leaf;

  private Double percentage_split;

  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLCoreException, KDDMLOperatorException {

    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_CLASSIFICATION_MODEL.getRepositoryPath() +
        ((YADT_SETTINGS) settings).getXMLDestination();

    File f = new File(xml_dest);
    input_file_data = "___yadt_data_input_" + f.getName() + ".data";
    input_file_names = "___yadt_names_input_" + f.getName() + ".names";
    output_file = "___yadt_output_" + f.getName() + ".xml";

    attribute_name = ((YADT_SETTINGS) settings).getTargetAttributeName();
    confidence_for_pruning = ((YADT_SETTINGS) settings).getConfidenceForPruninig().doubleValue();
    num_instances_for_leaf = ((YADT_SETTINGS) settings).getNumInstancesForLeaf().intValue();
    percentage_split = ((YADT_SETTINGS) settings).getPercentageSplitting();

    System.out.println("------------------------------------------");
    System.out.println("YaDT PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("Target Attribute Name - "+attribute_name);
    System.out.println("Pruning Confidence - "+confidence_for_pruning);
    System.out.println("Percentage Splittings - "+percentage_split);
    System.out.println("Num Instances for Leaf - "+num_instances_for_leaf);
    System.out.println("------------------------------------------");

  }

  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    YADTInputFormatting preprocessing =
        new YADTInputFormatting(instances, attribute_name, input_file_data, input_file_names);

    preprocessing.execute();
    model_name = preprocessing.getModelName();
    mining_schema = preprocessing.getMiningSchema();
    data_dictionary = preprocessing.getDataDictionary();

    return null;
  }

  /**
   * Call the Yadt algorithm and returns the tree model as output.
   * YaDT tree construction parameters:
   * - set the training (option -fm file.names)
   * - set the data (option -fd file.data)
   * - set minimum cases to split a node (option -m <num> where num is > 1,default 2)
   * - set pruning confidence level (option -c <num> where num is in the range (0,1],default 0.25)
   * - randomly split training data in a real training data and an additional test data
   *   (option -h <num> where num is in percentage (in the range [0,100]) if cases to
   *   belong to the real training set,default 100)
   * - build the PMML output tree (option -x output.xml)
   *
   * @param input Object null
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return Object the tree model
   */
  public Object execute(Object input)
      throws KDDMLOperatorException, KDDMLCoreException {

    // chiamata della libreria DCI
    // ottiene la percentuale di splitting in forma assoluta...
    Integer p = null;
    if (percentage_split != null)
      p = new Integer((int) (percentage_split.doubleValue() * 100));
    System.out.println("Percentage Splitting: "+p);

    int result = -1;
    try {
      if (System.getProperty("os.name").indexOf("Windows") != -1) {

        // Windows operative System
        result = callYADTForWindows(p);
      }

      if (System.getProperty("os.name").indexOf("Linux") != -1) {

        // Linux operative System
        result = callYADTForLinux(p);

      }
    }
    catch (Exception ec) {
      String msg1 = "Error invoking YaDT libraries: invalid command!!!";
      String msg2 = ec.toString();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }

    if (result != 0)
      throw new KDDMLOperatorException("Error invoking YaDT libraries: invalid command!!!", null,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);

    // tutto OK
    // nel file OUTPUT e' contenuto l'albero di classificazione prodotto dall'algoritmo
    System.out.println("DCI output created in: "+ YADT_EXE + output_file);

    // get the model via PMML tree factory
    String pmml_path = YADT_EXE + output_file;
    String dest = YADT_EXE + output_file;
    PMMLClassificationFactory fact = new PMMLClassificationFactory(new File(dest), pmml_path);

    return (TreeModel) fact.newInstance();

  }

  public TreeModel outputFormatting(Object tree)
      throws KDDMLOperatorException, KDDMLCoreException  {

    // setta il nuovo data dictionary contenente anche le informazioni sugli
    // attributi nominali omessi da YaDT

    File f = new File(xml_dest);

    Node root = (Node) ((TreeModel) tree).getRoot();
    ConfusionMatrix m1 = (ConfusionMatrix) ((TreeModel) tree).getTrainingConfusionMatrix();
    ConfusionMatrix m2 = (ConfusionMatrix) ((TreeModel) tree).getTestConfusionMatrix();

    // build the model...
    ClassificationProprietaryFactory fact =
        new ClassificationProprietaryFactory(f, data_dictionary, mining_schema, root, m1, m2);

    // setta le proprieta' del factory
    fact.setOverwrite(true);
    fact.setAlgorithmName("YaDT version 1.2.1 (c) S. Ruggieri 2002-2005");
    fact.setModelName(model_name+"_tree");
    if (num_instances_for_leaf == 2)
      fact.setBinaryNodeSplitting(true);
    else
      fact.setBinaryNodeSplitting(false);

    TreeModel output = (TreeModel) fact.newInstance();

    return output;
  }

  public String[] getTmpFiles() {
    String[] ris = {
        YADT_EXE + input_file_data,
        YADT_EXE + input_file_names,
        YADT_EXE + output_file};
    return ris;
  }

  private int callYADTForWindows(Integer percentage)
      throws java.io.IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();

    String names = "-fm "+input_file_names;
    String data = "-fd "+input_file_data;
    String num_leaf = "-m "+num_instances_for_leaf;
    String confidence = "-c "+confidence_for_pruning;
    String split;
    if (percentage != null)
      split = "-h "+percentage;
    else
      split = "";
    String output = "-x "+output_file;

    String exec = "cmd /c start /MIN /D " + "\"" + YADT_EXE + "\"" + " /wait"+ " " + "dtcmd" + " "
        + names + " " +data + " " +num_leaf + " " +confidence +" "+split + " " +output + " -tstd";

    System.out.println("Invoking yadt libraries...");
    System.out.println(exec);

    Process p = esegui.exec(exec);
    p.waitFor();

    return p.exitValue();

  }

  private int callYADTForLinux(Integer percentage)
      throws java.io.IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();
    if (percentage != null) {
      String[] exec = {
          YADT_EXE + "dTcmd",
          "-fm",
          YADT_EXE + input_file_names,
          "-fd",
          YADT_EXE + input_file_data,
          "-m",
          (new Integer(num_instances_for_leaf)).toString(),
          "-c",
          (new Double(confidence_for_pruning)).toString(),
          "-h",
          percentage.toString(),
          "-x",
          YADT_EXE + output_file};

      for (int i = 0; i < exec.length; i++)
        System.out.println(exec[i]);

      System.out.println("Invoking yadt libraries...");
      System.out.print(exec);

      Process p = esegui.exec(exec);
      p.waitFor();

      return p.exitValue();
    }
    else {
      String[] exec = {
          YADT_EXE + "dTcmd",
          "-fm",
          YADT_EXE + input_file_names,
          "-fd",
          YADT_EXE + input_file_data,
          "-m",
          (new Integer(num_instances_for_leaf)).toString(),
          "-c",
          (new Double(confidence_for_pruning)).toString(),
          "-x",
          YADT_EXE + output_file};

      for (int i = 0; i < exec.length; i++)
        System.out.println(exec[i]);

      System.out.println("Invoking yadt libraries...");
      System.out.print(exec);

      Process p = esegui.exec(exec);
      p.waitFor();

      return p.exitValue();
    }


  }
}
