/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.DataMining.ClusteringAlgorithms;

import kddml.Core.DataSources.*;
import kddml.Core.DataMining.Clustering.*;
import kddml.Core.KDDMLObject;
import kddml.Operators.*;
import kddml.Operators.DataMining.*;
import kddml.Core.KDDMLCoreException;
import java.io.*;
import weka.core.*;
import weka.clusterers.*;
import weka.estimators.*;
import java.lang.reflect.*;
import java.util.Iterator;
import java.util.Enumeration;
import kddml.Core.DataMining.*;

public class EM_RESOLVER implements ClusteringAlgorithmResolverTask {

  private String xml_dest;

  private int k;

  private int max_iterations;

  private double min_std_dev;

  private int num_attributes;

  private int num_clusters;

  private String model_name;

  private DataDictionary data_dictionary;

  private MiningSchema mining_schema;

  private DataStatistics statistic;

  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLCoreException, KDDMLOperatorException {

    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_CLUSTERING_MODEL.getRepositoryPath() +
        ((EM_SETTINGS) settings).getXMLDestination();

    k = ((EM_SETTINGS) settings).getNumberOfClusters().intValue();
    max_iterations = ((EM_SETTINGS) settings).getMaxNumberOfIterations().intValue();
    min_std_dev = ((EM_SETTINGS) settings).getMinStdDev().doubleValue();

    System.out.println("------------------------------------------");
    System.out.println("EM PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("k - "+k);
    System.out.println("Max iterations - "+max_iterations);
    System.out.println("Max standard deviation - "+min_std_dev);
    System.out.println("------------------------------------------");
  }

  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    statistic = (DataStatistics) instances.getStatistic();
    model_name = statistic.getRelationName()+"_clusters";

    // viene controllato che non siano presenti attributi stringa
    int s = statistic.getNumberOfStringAttributes();
    if (s != 0) {
      throw new KDDMLOperatorException(
      "Error in EM: no string attributes are allowed!", null,
      KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // build the data dictionary
    data_dictionary = new DataDictionary(statistic);

    // build the mining schema
    mining_schema = new MiningSchema();
    AttributeStatisticManager attr;
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      // all fields are active
      mining_schema.addField(new MiningField(attr.getName(), MiningField.ACTIVE));
    }

    num_attributes = statistic.getNumAttributes();

    // returns all weka instances
    return (Instances) instances.load(InternalTable.ALL);
  }

  public Object execute(Object instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    EM em = null;
    try {
      em = new EM();

      // set em parameters
      em.setNumClusters(k);
      em.setMaxIterations(max_iterations);
      em.setMinStdDev(min_std_dev);

      // build the clustering
      try {
        em.buildClusterer( (Instances) instances);
      }
      catch(java.lang.OutOfMemoryError err) {
        throw new KDDMLOperatorException("Out of Memory Error!!!", null,
                                         KDDMLOperatorErrorCodes.OUT_OF_MEMORY);
      }


      num_clusters = em.numberOfClusters();
      //System.out.println(em.toString());

      /*Instances inst = (Instances) instances;
      for (int i=0; i<inst.numInstances(); i++) {
        System.out.println("Instance: "+inst.instance(i).toString()+" - Cluster: "+em.clusterInstance(inst.instance(i)));
        double[] distr = em.distributionForInstance(inst.instance(i));
        System.out.print("Distribution: ");
        for (int k=0; k<distr.length; k++)
          System.out.print(distr[k]+" ");
        System.out.println();
      }*/

      return em;

    }
    catch (Exception ex) {
      String msg = "Error in EM: can't build clusters";
      String msg2 = ex.toString();
      throw new KDDMLOperatorException(msg, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }
  }

  public ClusteringModel outputFormatting(Object clusters)
      throws KDDMLOperatorException, KDDMLCoreException  {

    File f = new File(xml_dest);

    // get the centroid description
    ClusterDescription descr = getClusteringDescription();
    double[] priors = ((EM) clusters).clusterPriors();
    // get the distance funtion as the euclidean
    EM_DISTANCE distance = new EM_DISTANCE(priors);

    EmptyClusteringFactory fact = new EmptyClusteringFactory(f, data_dictionary, mining_schema, distance, descr);

    // setta le proprieta' del factory
    fact.setOverwrite(true);
    fact.setAlgorithmName("EM - Expectation Maximization (Weka implementation)");
    fact.setModelName(model_name);

    ClusteringModel output = (ClusteringModel) fact.newInstance();

    // add clusters
    ClusterStatistic[] stat = buildClusters((EM) clusters);
    for (int i=0; i< stat.length; i++) {
      String name = "Cluster_"+i;
      output.addCluster(new DistributionBasedCluster(name, stat[i]));
    }

    return output;
  }

  public String[] getTmpFiles() {
    String[] ris = {};
    return ris;
  }


  /**
   * Metodo che trova il cluster di appartenenza di una istanza.
   *
   * @param numClusters il numero di clusters da analizzare
   * @param distrIstance vettore contenente la probabilita' di appartenenza
   * di un'istanza ad ogni cluster
   *
   */
  private static int getCluster(int numClusters, double[] distrInstance)
  {
    int i;
    int j = 0;
    double max = -1;
    for (i = 0; i < numClusters;i++) {
      if (distrInstance[i] >= max) {
        j = i;
        max = distrInstance[i];
      }
    }
    return j;
  }

  /**
   * Set the comparison measure as the euclidean distance.
   * For numeric attributes set the ABS function.
   * For discrete attributes set the DELTA function.
   *
   * @return CentroidBasedComparisonMeasure
   */
  private ClusterDescription getClusteringDescription()
      throws ClusteringModelException {

    return new ClusterDescription(statistic, null, false);
  }

  /**
   * Costruisce il vettore delle coordinate che rappresenta il centroide inserendo la media
   * per attributi numerici e la distribuzione per attributi discreti.
   * Il vettore viene normalizzato.
   *
   * @return double[]
   */
  private ClusterStatistic[] buildClusters(EM clusters)
      throws KDDMLOperatorException {

    ClusterStatistic[] a_stat = new ClusterStatistic[num_clusters];
    ClusterStatistic stat;

    // get the distribution for discrete fields
    DiscreteEstimator[][] discrete_dis = null;
    try {
      discrete_dis = getDiscreteEstimator(clusters);
    }
    catch (Exception ex) {
      throw new KDDMLOperatorException("Error calculating the distance measure",
                                       null,
                                       KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // get the distribution for numeric fields
    double[][][] numeric_dis = clusters.getClusterModelsNumericAtts();

    // ciclo sul numero di clusters
    for (int i=0; i< a_stat.length; i++) {

      stat = new ClusterStatistic("Cluster_"+i);
      stat.setClusterPriors(clusters.clusterPriors()[i]);

      // scorre i campi per costruire le varie statistiche
      AttributeStatisticManager attr;
      StatisticField field = null;

      // ciclo sugli attributi
      for (int j = 0; j < statistic.getNumAttributes(); j++) {
        attr = statistic.getStatistics()[j];
        if (attr.getType().equals(AttributeType.NUMERIC)) {
          // attributo di tipo numerico
          double mean = numeric_dis[i][j][0];
          double std_dev = numeric_dis[i][j][1];
          field = new ContinuousStatisticField(attr.getName(), mean, std_dev);
          stat.addStatisticField((ContinuousStatisticField) field);
        }
        if (attr.getType().equals(AttributeType.NOMINAL)) {
          // attributo di tipo nominale
          field = new DiscreteStatisticField(attr.getName());

          // add values to the field....
          Enumeration enum2 = ((NominalStatistic) attr).getValues();
          int k=0;
          while (enum2.hasMoreElements()) {

            String value = (String) enum2.nextElement();
            double count = discrete_dis[i][j].getCount(k);
            // WEKA ADD 1 TO THE DISCRETE DISTRIBUTION VALUES
            ((DiscreteStatisticField) field).addValue(value, count /*- 1*/);
            k++;
          }
          stat.addStatisticField((DiscreteStatisticField) field);
        }
      }
      a_stat[i] = stat;
    }
    return a_stat;
  }

  /**
   * Use the reflection to get the private field of em clustering.
   *
   * @param em EM
   * @throws IllegalAccessException
   * @throws IllegalArgumentException
   * @throws SecurityException
   * @throws NoSuchFieldException
   * @return DiscreteEstimator[][]
   */
  private static DiscreteEstimator[][] getDiscreteEstimator(EM em) throws
      IllegalAccessException, IllegalArgumentException, SecurityException,
      NoSuchFieldException {

    Class c = em.getClass();

    Field fl= c.getDeclaredField("m_model");
    fl.setAccessible(true);
    DiscreteEstimator[][] discrete_estimator = (DiscreteEstimator[][]) fl.get(em);

    return discrete_estimator;
  }

  public static void main(String args[]) {
    File f = new File("C:/Documents and Settings/romei/kddml/repository/data/ARFF/weather_nominal.arff");
    try {
      Instances inst = new Instances(new FileReader(f));
      System.out.println("Num instances: "+inst.numInstances());
      System.out.println("Input Data:");
      System.out.println(inst.toString());
      EM em = new EM();
      em.setNumClusters(3);
      em.buildClusterer(inst);

      System.out.println("Clusters: ");
      System.out.println(em.toString());

      // Crea il vettore delle distribuzioni dei valori per le istanze.
      int[] distrClusters = new int[inst.numInstances()];
      for (int i=0; i < inst.numInstances(); i++) {
        distrClusters[i] = getCluster(em.numberOfClusters(), em.distributionForInstance(inst.instance(i)));
      }

      // Calcola la prob. di appartenenza dell'istanza a un cluster.

      double[] prob;
      for (int i = 0; i<inst.numInstances(); i++) {
        //for (int j=0;j<prob.length;j++)
        prob = em.distributionForInstance(inst.instance(i));
        System.out.println(inst.instance(i).toString() + " - Cluster 3:" + prob[2]);
      }

      // returns the clusters priors
      /*double[] priors = em.clusterPriors();
             for (int h=0;h<priors.length; h++)
        System.out.println("Cluster "+h+": "+priors[h]);*/

      Class c = em.getClass();

      Field fl= c.getDeclaredField("m_model");
      fl.setAccessible(true);
      DiscreteEstimator[][] discrete_estimator = (DiscreteEstimator[][]) fl.get(em);
      System.out.println(discrete_estimator[0][0].toString());

      System.out.println("Count: "+discrete_estimator[0][0].getCount(0));
      System.out.println("Count: "+discrete_estimator[0][0].getCount(1));
      System.out.println("Count: "+discrete_estimator[0][0].getCount(2));
      System.out.println("Probability: "+discrete_estimator[0][0].getProbability(0));

    }
    catch (Exception ex) {
      System.out.println(ex.toString());
    }
  }
}
