/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.DataMining.ClusteringAlgorithms;

import kddml.Core.DataSources.*;
import kddml.Core.DataMining.Clustering.*;
import kddml.Core.KDDMLObject;
import kddml.Operators.*;
import kddml.Operators.DataMining.*;
import weka.estimators.*;
import kddml.Core.KDDMLCoreException;
import java.io.*;
import weka.core.*;
import weka.clusterers.*;
import java.util.Vector;
import kddml.Core.DataMining.*;
import java.lang.reflect.*;



public class KMEANS_RESOLVER implements ClusteringAlgorithmResolverTask {

  private String xml_dest;

  private int k;

  private int num_attributes;

  private int num_clusters;

  private String model_name;

  private DataDictionary data_dictionary;

  private MiningSchema mining_schema;

  private DataStatistics statistic;

  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLCoreException, KDDMLOperatorException {

    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_CLUSTERING_MODEL.getRepositoryPath() +
        ((KMEANS_SETTINGS) settings).getXMLDestination();

    k = ((KMEANS_SETTINGS) settings).getNumberOfClusters().intValue();

    System.out.println("------------------------------------------");
    System.out.println("KMEANS PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("k - "+k);
    System.out.println("------------------------------------------");
  }

  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    statistic = (DataStatistics) instances.getStatistic();
    model_name = statistic.getRelationName()+"_clusters";

    // viene controllato che non siano presenti attributi stringa
    int s = statistic.getNumberOfStringAttributes();
    if (s != 0) {
      throw new KDDMLOperatorException(
      "Error in EM: no string attributes are allowed!", null,
      KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // build the data dictionary
    data_dictionary = new DataDictionary(statistic);

    // build the mining schema
    mining_schema = new MiningSchema();
    AttributeStatisticManager attr;
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      // all fields are active
      mining_schema.addField(new MiningField(attr.getName(), MiningField.ACTIVE));
    }

    num_attributes = statistic.getNumAttributes();

    // returns all weka instances
    return (Instances) instances.load(InternalTable.ALL);
  }

  public Object execute(Object instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    SimpleKMeans kmeans = null;
    try {
      kmeans = new SimpleKMeans();

      // set em parameters
      kmeans.setNumClusters(k);

      // build the clustering
      try {
        kmeans.buildClusterer((Instances) instances);
      }
      catch(java.lang.OutOfMemoryError err) {
        throw new KDDMLOperatorException("Out of Memory Error!!!", null,
                                         KDDMLOperatorErrorCodes.OUT_OF_MEMORY);
      }

      num_clusters = kmeans.numberOfClusters();
      //System.out.println(kmeans.toString());
      //System.out.println(kmeans.getClusterStandardDevs());
      //System.out.println("Centroids:");
      //System.out.println(kmeans.getClusterCentroids().toString());

      //Instances inst = (Instances) instances;
      //for (int i=0; i<inst.numInstances(); i++) {
      //  System.out.println("Instance: "+inst.instance(i).toString()+" - Cluster: "+kmeans.clusterInstance(inst.instance(i)));
      //}

      return kmeans;

    }
    catch (Exception ex) {
      String msg = "Error in KMeans: can't build clusters";
      String msg2 = ex.toString();
      throw new KDDMLOperatorException(msg, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }
  }

  public ClusteringModel outputFormatting(Object clusters)
      throws KDDMLOperatorException, KDDMLCoreException  {

    File f = new File(xml_dest);
    // create an empty clustering

    // get the centroid description
    ClusterDescription descr = getClusteringDescription();
    // get the distance funtion as the euclidean
    EUCLIDEAN eucl = new EUCLIDEAN(descr);

    EmptyClusteringFactory fact = new EmptyClusteringFactory(f, data_dictionary, mining_schema, eucl, descr);
    // setta le proprieta' del factory
    fact.setOverwrite(true);
    fact.setAlgorithmName("KMeans (Weka implementation)");
    fact.setModelName(model_name);

    ClusteringModel output = (ClusteringModel) fact.newInstance();

    // add clusters to the model
    Instances centroids = ((SimpleKMeans) clusters).getClusterCentroids();

    for (int i=0; i<centroids.numInstances(); i++) {
      double[] centroid = getClusterCoordinate(centroids.instance(i), descr);
      String name = "Cluster_"+i;
      CentroidBasedCluster cluster = new CentroidBasedCluster(name, centroid);

      cluster.setClusterStatistic(buildStatistics((SimpleKMeans) clusters)[i]);
      output.addCluster(cluster);
    }

    return output;
  }

  public String[] getTmpFiles() {
    String[] ris = {};
    return ris;
  }


  /**
   * Costruisce il vettore delle coordinate che rappresenta il centroide inserendo la media
   * per attributi numerici e la distribuzione per attributi discreti.
   * Il vettore viene normalizzato.
   *
   * @return double[]
   */
  private ClusterStatistic[] buildStatistics(SimpleKMeans clusters)
      throws KDDMLOperatorException {

    ClusterStatistic[] a_stat = new ClusterStatistic[num_clusters];
    ClusterStatistic stat;

    // get the distribution for discrete fields
    int[][][] discrete_dis = clusters.getClusterNominalCounts();

    // get the distribution for numeric fields
    double[][][] numeric_dis = getNumericEstimator(clusters.getClusterCentroids(),
                                                   clusters.getClusterStandardDevs());

    // ciclo sul numero di clusters
    for (int i=0; i< a_stat.length; i++) {

      stat = new ClusterStatistic("Cluster_"+i);

      // scorre i campi per costruire le varie statistiche
      AttributeStatisticManager attr;
      StatisticField field = null;

      // ciclo sugli attributi
      for (int j = 0; j < statistic.getNumAttributes(); j++) {
        attr = statistic.getStatistics()[j];
        if (attr.getType().equals(AttributeType.NUMERIC)) {
          // attributo di tipo numerico
          double mean = numeric_dis[i][j][0];
          double std_dev = numeric_dis[i][j][1];
          field = new ContinuousStatisticField(attr.getName(), mean, std_dev);
          stat.addStatisticField((ContinuousStatisticField) field);
        }
        if (attr.getType().equals(AttributeType.NOMINAL)) {
          // attributo di tipo nominale
          field = new DiscreteStatisticField(attr.getName());

          // add values to the field....
          java.util.Enumeration enum2 = ((NominalStatistic) attr).getValues();
          int k=0;
          while (enum2.hasMoreElements()) {

            String value = (String) enum2.nextElement();

            int new_index = getCategoryPosition(clusters.getClusterCentroids(), j, value);
            double count = discrete_dis[i][j][new_index];

            // WEKA ADDS 1 TO THE DISCRETE DISTRIBUTION VALUES
            ((DiscreteStatisticField) field).addValue(value, count /*- 1*/);

            k++;
          }
          stat.addStatisticField((DiscreteStatisticField) field);
        }
      }
      a_stat[i] = stat;
    }
    return a_stat;
  }

  private static double[][][] getNumericEstimator(Instances centroid, Instances variances) {

    double[][][] ris = new double[centroid.numInstances()][centroid.numAttributes()][2];
    // set the mean
    for (int i=0; i<centroid.numInstances(); i++) {
      for (int j=0; j<centroid.numAttributes(); j++) {
        if (centroid.attribute(j).isNumeric())
          // insert the mean of the centroid
          ris[i][j][0] = centroid.instance(i).value(j);
        else
          ris[i][j][0] = 0;
      }
    }

    // set the variances
    for (int i=0; i<variances.numInstances(); i++) {
      for (int j=0; j<variances.numAttributes(); j++) {
        if (centroid.attribute(j).isNumeric())
          // insert the mean of the centroid
          ris[i][j][1] = variances.instance(i).value(j);
        else
          ris[i][j][1] = 0;
      }
    }

    return ris;
  }

  private static int getCategoryPosition(Instances centroid, int attribute_index, String category) {

    Attribute attr = centroid.attribute(attribute_index);
    java.util.Enumeration enum2 = attr.enumerateValues();
    int ris = -1;
    while (enum2.hasMoreElements()) {
      ris++;
      String cat = (String) enum2.nextElement();
      if (cat.equalsIgnoreCase(category))
        return ris;
    }
    return -1;
  }

  /**
   * Set the comparison measure as the euclidean distance.
   * For numeric attributes set the ABS function.
   * For discrete attributes set the DELTA function.
   *
   * @return CentroidBasedComparisonMeasure
   */
  private ClusterDescription getClusteringDescription()
      throws ClusteringModelException {

    AttributeComparisonMeasure[] measures = new AttributeComparisonMeasure[statistic.getNumAttributes()];
    AttributeStatisticManager[] stat = statistic.getStatistics();
    for (int i=0; i<statistic.getNumAttributes(); i++) {
      if (stat[i].getType().equals(AttributeType.NUMERIC)) {
        // attributo di tipo numerico
        measures[i] = new ABS_DIFF();
      }
      if (stat[i].getType().equals(AttributeType.NOMINAL)) {
        // attributo di tipo discreto
        measures[i] = new DELTA();
      }
    }

    return new ClusterDescription(statistic, measures, true);
  }

  private double[] getClusterCoordinate(Instance centroid, ClusterDescription centroid_desc)
      throws ClusteringModelException {

    // vector of Double
    Vector v = new Vector();
    for (int i=0; i<centroid.numAttributes(); i++) {
      if (centroid.attribute(i).isNumeric()) {
        // numeric field
        double d = centroid.value(i);
        ContinuousClusteringField f = (ContinuousClusteringField) centroid_desc.getField(centroid.attribute(i).name());
        v.add(new Double(f.normalize(d)));
      }
      else {
        // discrete field
        String s = centroid.stringValue(i);
        DiscreteClusteringField f = (DiscreteClusteringField) centroid_desc.getField(centroid.attribute(i).name());
        double[] distr = f.getDistribution(s);
        for (int k=0; k<distr.length; k++)
          v.add(new Double(distr[k]));
      }
    }
    double[] ris = new double[v.size()];
    for (int h=0; h<v.size(); h++)
      ris[h] = ((Double) v.get(h)).doubleValue();
    return ris;
  }

  public static void main(String args[]) {
  }
}
