/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.DataMining.SequenceAlgorithms;

import java.io.*;
import java.util.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataSources.*;
import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.*;
import kddml.Operators.*;

/**
 * Classe che invoca l'algoritmo PrefixSpan ed estrae i pattern sequenziali
 * dalle istanze di input.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class PREFIX_SPAN_RESOLVER implements SequenceAlgorithmResolverTask {

  // PrefixSpan library
  protected final static String PREFIX_SPAN_LIB =
      System.getProperty("user.dir") + File.separator + "resources" +
      File.separator + "lib" + File.separator + "algorithms" +
      File.separator + "PrefixSpan" + File.separator;

  private String input_file;
  private String output_file;

  String xml_dest;

  /**
   * Minimo supporto in forma percentuale passato dall'utente come parametro.
   */
  private double min_support;

  /**
   * Minima confidenza in forma percentuale -- 1.0 non vengono genrate regole .
   */
  private double min_confidence;

  /**
   * Numero massimo di sequenze che devono essere estratte.
   * Parametro opzionale fornito in input dall'utente.
   */
  private int num_max_sequences;

  /**
   * Numero di transazioni esaminate durante la fase di preprocessing.
   */
  private int num_trans;

  /**
   * Nome del modello
   */
  private String model_name;

  /**
   * Elemento mining schema
   */
  private MiningSchema mining_schema;

  /**
   * Elemento data dictionary
   */
  private DataDictionary data_dictionary;

  /**
   * Numero massimo di items per transazione.
   */
  private int max_items;

  /**
   * Numero medio di items per transazione.
   */
  private double avg_items;


  public void readSettings(KDDMLOperatorSettings settings)
      throws KDDMLCoreException, KDDMLOperatorException {
    // memorizza i parametri ottenuti dal settings
    xml_dest = kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL.getRepositoryPath() +
        ((PREFIX_SPAN_SETTINGS) settings).getXMLDestination();

    File f = new File(xml_dest);

    input_file = PREFIX_SPAN_LIB + "___ps_input_" + f.getName() + ".txt";
    output_file = PREFIX_SPAN_LIB + "___ps_output_" + f.getName() + ".txt";

    min_support = ((PREFIX_SPAN_SETTINGS) settings).getMinSupport().doubleValue();

    min_confidence = 1.0;

    if (((PREFIX_SPAN_SETTINGS) settings).getMaxNumberOfSequences() != null)
      num_max_sequences = ((PREFIX_SPAN_SETTINGS) settings).getMaxNumberOfSequences().intValue();
    else
      num_max_sequences = SequenceModel.MAX_NUMBER_OF_SEQUENCES;

    System.out.println("------------------------------------------");
    System.out.println("PrefixSpan PARAMETERS: ");
    System.out.println("XML Destination - "+xml_dest);
    System.out.println("Min Support - "+min_support);
    System.out.println("Max Sequences - "+num_max_sequences);
    System.out.println("------------------------------------------");

  }

  public Object inputFormatting(InternalTableManager instances)
      throws KDDMLOperatorException, KDDMLCoreException {

    PrefixSpanInputFormatting preprocessing =
        new PrefixSpanInputFormatting(instances, input_file);
    Hashtable items_mapped = preprocessing.execute();

    num_trans = preprocessing.getNumberOfTransactions();
    max_items = preprocessing.getMaxNumberOfItemsPerTransaction();
    avg_items = preprocessing.getAvgNumberOfItemsPerTransaction();
    model_name = preprocessing.getModelName();
    mining_schema = preprocessing.getMiningSchema();
    data_dictionary = preprocessing.getDataDictionary();

    // Tabella hash contenente le associazioni chiave --> item.
    // Generata durante la fase di pre-processing.
    return items_mapped;

  }

  public Object execute(Object hash_table)
      throws KDDMLOperatorException, KDDMLCoreException {

    // chiamata della libreria PrefixSpan
    // ottiene il supporto in forma assoluta...
    double s = num_trans * min_support;

    int min_absolute_support = (int) s;
    if (min_absolute_support == 0)
      min_absolute_support = 1;

    System.out.println("Absolute Minimum Support: " + min_absolute_support);

    int min_percentage_support = (int)(min_support * 100);

    System.out.println("Percentage Minimum Support: " + min_percentage_support);

    try {
      System.out.println("Invoking PrefixSpan libraries...");

      String[] prefixspan_args =
          {String.valueOf(min_percentage_support), input_file, output_file, null};
      prefixspan_x_sp.PrefixSpan.main(prefixspan_args);
      System.gc();

    }
    catch (OutOfMemoryError oom){
      String msg1 = "Error invoking PrefixSpan libraries: Out of memory!!!";
      String msg2 = oom.toString();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }
    catch (Throwable t) {
      String msg1 = "Error invoking PrefixSpan libraries.";
      String msg2 = t.toString();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }
    System.out.println("PrefixSpan output created in: "+ output_file);
    return hash_table;

  }

  public SequenceModel outputFormatting(Object items_mapped)
      throws KDDMLOperatorException, KDDMLCoreException  {

    System.out.println("Generating Frequent sequences from PrefixSpan output... please wait!!!");
    PrefixSpanOutputFormatting post = new PrefixSpanOutputFormatting(num_trans, num_max_sequences, output_file);
    TreeSet sequences = post.buildSequences((Hashtable) items_mapped);

    System.out.println("Number of frequent sequences found: "+sequences.size());

    File f = new File(xml_dest);
    SequenceFactory fact = new SequenceFactory(
      f, data_dictionary, mining_schema, min_support, min_confidence, num_trans, sequences);

    // setta le proprieta' del factory
    fact.setOverwrite(true);
    fact.setAlgorithmName("PrefixSpan - Computer Science Departement, Pisa, Italy");
    fact.setModelName(model_name+"_sequence");
    fact.setMaxNumberOfItemsPerTA(max_items);
    fact.setAvgNumberOfItemsPerTA(avg_items);

    SequenceModel output = (SequenceModel) fact.newInstance();

    return output;

  }

  public String[] getTmpFiles() {
    String[] ris = {input_file, output_file};
    return ris;
  }
}
