/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.DataMining.SequenceAlgorithms;

import java.lang.*;
import java.io.*;
import java.util.*;
import kddml.Core.DataSources.*;
import kddml.Core.KDDMLCoreException;
import kddml.Core.DataMining.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Operators.*;

/**
 * I metodi di questa classe costruiscono un file di testo a partire dalle data-sequence
 * contenute nella tabella in input.
 * Tale file puo' essere usato come input per l'algoritmo PrefixSpan per l'estrazione
 * di patterns sequenziali frequenti.
 * L'output di questa classe e' una tabella hash in cui sono contenuti i valori degli items
 * con le rispettive chiavi e un file di testo contenente la chiave corrispondente per
 * ogni singolo attributo di ogni singola transazione.
 * Gli items sono ordinati in modo crescente.
 *
 * Es: per la singola data-sequene (client_id, timestamp, pasta, pane, sale) viene generata una
 * tabella hash contenente i seguenti valori:
 * 1 - pasta
 * 2 - pane
 * 3 - sale
 * Il file di testo contiene la seguente riga:
 * 1 1 2 3,
 * dove il primo attributo rappresenta l'id della data-sequence, i restanti attributi rappresentano
 * gli items della transazione.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0.0 alfa
 */

final class PrefixSpanInputFormatting {

  /**
   * Insieme di istanze in ingresso.
   */
  private InternalTableManager input;

  /**
   * Indica se la tabella e' di tipo timestamp (transaction, timestamp, event).
   */
  private boolean is_timestamp;

  /**
   * Statistiche
   */
  private DataStatistics statistic;

  /**
   * Nome del modello.
   */
  private String model_name;

  /**
   * Rappresenta l'elemento data dictionary.
   */
  private DataDictionary data_dictionary;

  /**
   * Rappresenta il mining schema della tabella in ingresso.
   */
  private MiningSchema mining_schema;

  /**
   * Numero di data-sequence: numero di oggetti distinti in input (e.g. il numero di clienti)
   */
  private int num_dseqs;

  /**
   * Numero massimo di items in una data-sequence.
   */
  private int max_dseq_items;

  /**
   * Numero medio di items in una data-sequence.
   */
  private double avg_dseq_items;

  /**
   * Numero di items distinti.
   */
  private int num_items;

  /**
   * Questa tabella hash contiene le associazioni (id_item --> item) prodotte
   * dalla rienumerazione degli item dove item  il "valore", e id_item  la
   * "chiave".
   */
  private Hashtable items_mapped;

  /**
   * Questa tabella hash contiene le associazioni (item --> item_id) prodotte
   * dalla rienumerazione degli item dove item  il "valore", e id_item  la
   * "chiave". Per ogni item mappato in "items_mapped" viene eseguito un
   * mappaggio inverso, per ottimizzare i controlli.
   */
  private Hashtable reverse_items_mapped;

  //file di output per l'algoritmo PrefixSpan
  private OutputStream file;
  private BufferedOutputStream out;
  private PrintWriter pw;
  private String input_file;

  protected PrefixSpanInputFormatting(InternalTableManager i, String input_file) {

    input = i;
    // controlla se la tabella e' di tipo timestamped
    is_timestamp = i.isTimestampTable();

    statistic = (DataStatistics) i.getStatistic();
    model_name = statistic.getRelationName();
    data_dictionary = new DataDictionary();
    mining_schema = new MiningSchema();

    num_dseqs = 0;
    max_dseq_items = 0;
    avg_dseq_items = 0;
    num_items = 0;

    items_mapped = new Hashtable();
    reverse_items_mapped = new Hashtable();

    this.input_file = input_file;
  }

  protected Hashtable execute() throws KDDMLOperatorException, KDDMLCoreException {

    // viene controllato che la tabella sia di tipo timestamp...
    if ( !is_timestamp ) {
        throw new KDDMLOperatorException(
            "Error in PrefixSpan: the input table <"+model_name+
            "> must be a timestamp table.", null,
            KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    System.out.println("Preprocessing PrefixSpan...");
    buildMetadataInformation();

    try {
      file = new FileOutputStream(input_file);
      out = new BufferedOutputStream(file);
      pw = new PrintWriter(out, true);
    }
    catch (IOException e) {
      String msg1 = "I/O Error calling PrefixSpan preprocessing.";
      String msg2 = "I/O Error calling PrefixSpan preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

    // esecuzione della fase di preprocessing
    int num_tuple = input.getNumOfInstances();
    int num_attributi = input.getNumOfAttributes();

    if ( is_timestamp ) {
      //tabella timestamped
      DataSequenceManager d_sequence;
      System.out.println("TimeStamped Table");

      while(((TimestampInternalTableManager) input).hasNextDataSequence()) {
        d_sequence = ((TimestampInternalTableManager) input).loadNextDataSequence();
        printDataSequence(d_sequence, num_dseqs);
        num_dseqs++;
      }

      avg_dseq_items = avg_dseq_items / (double) num_dseqs;
    }

    System.out.println("Number of Attributes parsed: "+num_attributi);
    System.out.println("Number of Instances parsed: "+num_tuple);
    System.out.println("Number of Transactions: "+getNumberOfTransactions());
    System.out.println("Number of Max Items per Transaction: "+getMaxNumberOfItemsPerTransaction());
    System.out.println("Number of Avg Items per Transaction: "+getAvgNumberOfItemsPerTransaction());
    System.out.println("Number of Items: "+getNumberOfItems());
    System.out.println("Preprocessing output written in: "+input_file);
    System.out.println("-----------------------------");

    try {
      pw.flush();
      out.flush();
      file.flush();
      file.close();
    }
    catch (IOException e) {
      String msg1 = "I/O Error calling PrefixSpan preprocessing.";
      String msg2 = "I/O Error calling PrefixSpan preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

    return items_mapped;
  }

  protected int getNumberOfTransactions() {
    return num_dseqs;
  }

  protected int getMaxNumberOfItemsPerTransaction() {
    return max_dseq_items;
  }

  protected double getAvgNumberOfItemsPerTransaction() {
    return avg_dseq_items;
  }

  protected int getNumberOfItems(){
    return num_items;
  }

  protected MiningSchema getMiningSchema() {
    return mining_schema;
  }

  protected DataDictionary getDataDictionary() {
    return data_dictionary;
  }

  protected String getModelName() {
    return model_name;
  }

  /**
   * Questo metodo costruisce un file di testo a partire dalle istanze
   * di una tabella timestamped.
   * Ogni riga del file contiene gli items di una transazione della tabella di input.
   * Gli items sono rappresentati da numeri sotto forma di stringhe.
   * Gli items di ogni riga sono ordinati in modo crescente e sono distinti.
   * Gli items di ogni riga sono separati da uno spazio bianco.
   *
   * Nel caso di estrazione di pattern sequenziali si suppone che il primo
   * attributo della tabella di input specifichi l'identificatore di cliente.
   *
   * Le associazioni (nome_item_originale --> identificatoreIntero)
   * vengono memorizzate nella tabella hash: "items_mapped" di questa classe.
   */

  private void printDataSequence(DataSequenceManager d_sequence, int seq_id) throws KDDMLOperatorException {

    Iterator iterator = d_sequence.getTransactions();

    int number_of_items = 0;

    while( iterator.hasNext() ) {
      TimestampedTransactionManager trans = (TimestampedTransactionManager) iterator.next();
      number_of_items += trans.getNumberOfItems();

      // per ogni nuova transazione conterra' gli item in essa contenuti
      TreeSet items = new TreeSet();

      Iterator iter = trans.getItems();
      while (iter.hasNext()) {
        Item i = (Item) iter.next();
        String item_str = new String(i.toString().trim());
        String key = mapItem(item_str);
        items.add(new Long(key));
      }

      // scrive l'istanza nel file
      writeRow(seq_id, items);

    }

    if( number_of_items > max_dseq_items )
      max_dseq_items = number_of_items;
    avg_dseq_items += (double) number_of_items;

  }



  private void writeRow(int seq_id, TreeSet orderedItems) throws KDDMLOperatorException {

    Iterator iterator = orderedItems.iterator();

    //conta il numero di items inseriti nel file
    int count_items = 0;

    Long item;

    //ciclo while per scrivere sul file gli items dell'itemset
    try {
      pw.print(seq_id + " ");

      while (iterator.hasNext()) {

        item = (Long) iterator.next();
        pw.print(item);
        count_items++;
        //controlla se sono all'ultimo item e quindi non devo inserire lo spazio
        //dopo l'inserimento dell'item
        if (count_items < orderedItems.size())
          pw.print(" ");
      }

      pw.println();
    }
    catch (Exception e) {
      String msg1 = "I/O Error calling PrefixSpan preprocessing.";
      String msg2 = "I/O Error calling PrefixSpan preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

  }

  private String mapItem(String nome_item) {

    //if (!items_mapped.containsValue(nome_item.trim())) {
    if (!reverse_items_mapped.containsKey(nome_item.trim())) {
      // non esiste ancora un'associazione fra l'item e un valore numerico
      // cos viene creata all'interno della tabella hash
      String key = (String) Integer.toString(num_items);
      num_items++;
      items_mapped.put(key, nome_item.trim());
      reverse_items_mapped.put(nome_item.trim(), key);
      return key;
    }
    else {
      //recuperaro la chiave associata a nome_item
      return (String) reverse_items_mapped.get(nome_item.trim());
      /*Enumeration enum2 = items_mapped.keys();
      //scorro la tabella hash:
      while ( (enum.hasMoreElements())) {
        String current_key = (String) enum.nextElement();
        if (items_mapped.get(current_key).equals(nome_item.trim()))
          return current_key;
      }*/
    }

    //return null;
  }

  protected void buildMetadataInformation() {
    String name;
    AttributeStatisticManager attr;
    for (int j = 0; j < statistic.getNumAttributes(); j++) {
      attr = statistic.getStatistics()[j];
      name = attr.getName();

      if (attr.getType().equals(AttributeType.NUMERIC)) {
        data_dictionary.addField(new DataField(name, DataField.CONTINUOUS));
        if (is_timestamp && name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
          mining_schema.addField(new MiningField(TransactionalInternalTable.TRANSACTION_ATTRIBUTE,
                                                 MiningField.ACTIVE));
        else if(is_timestamp && name.equalsIgnoreCase(TimestampInternalTable.TIMESTAMP_ATTRIBUTE))
          mining_schema.addField(new MiningField(TimestampInternalTable.TIMESTAMP_ATTRIBUTE,
                                                 MiningField.ACTIVE));
        else
          mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));
      }
      else {
        if (attr.getType().equals(AttributeType.NOMINAL)) {
          // attributo di tipo enumerato
          DataField mf = new DataField(name, DataField.CATEGORICAL);

          Enumeration enum2 = ((NominalStatistic) attr).getValues();
          while(enum2.hasMoreElements()) {
            try {
              mf.addEnumeratedValue( (String) enum2.nextElement());
            }
            catch (MiningModelException ex) {
            }
          }
          data_dictionary.addField(mf);
          if (is_timestamp) {

            if (name.equalsIgnoreCase(TransactionalInternalTable.EVENT_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.PREDICTED));
            else
            if (name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
            if (is_timestamp && name.equalsIgnoreCase(TimestampInternalTable.TIMESTAMP_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
              mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));

          }
          else
            mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
        }
        else {
          // attributo di tipo stringa
          data_dictionary.addField(new DataField(name));
          if (is_timestamp) {

            if (name.equalsIgnoreCase(TransactionalInternalTable.EVENT_ATTRIBUTE))
              mining_schema.addField(new MiningField(name,
                  MiningField.PREDICTED));
            else
            if (name.equalsIgnoreCase(TransactionalInternalTable.TRANSACTION_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
            if (is_timestamp && name.equalsIgnoreCase(TimestampInternalTable.TIMESTAMP_ATTRIBUTE))
              mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
            else
              mining_schema.addField(new MiningField(name, MiningField.SUPPLEMENTARY));

          }
          else
            mining_schema.addField(new MiningField(name, MiningField.ACTIVE));
        }
      }

    }
  }

  public static void main(String args[]) {
    /*File f = new File("G:/WINDOWS/jbprojects/kddml/coop_ts.xml");
    try {

      try {
        System.setOut(new PrintStream(new FileOutputStream("e.log")));
      }
      catch (Exception ex) {

      }

      ARFFFactory fact = new ARFFFactory(f, "G:/WINDOWS/jbprojects/kddml/repository/data/ARFF/coop_timestamp.arff");
      fact.setOverwrite(true);
      InternalTable t = (InternalTable) fact.newInstance();
      t.saveToRepository();

      PrefixSpanInputFormatting preprocessing = new PrefixSpanInputFormatting(t);
      Hashtable ht = preprocessing.execute();

      Enumeration enum2 = ht.keys();
      Collection c = ht.values();
      Iterator i = c.iterator();
      while(enum.hasMoreElements())
        System.out.println(enum.nextElement()+"\t"+i.next());
       //System.out.println(ht);

      try{
        BufferedReader br = new BufferedReader(new FileReader(PREFIX_SPAN_RESOLVER.INPUT));
        FileWriter fw = new FileWriter("transf0.txt");

        String seq_id="";
        int trans_id=0;

        String line;
        while( (line=br.readLine()) != null ){
          StringTokenizer st = new StringTokenizer(line);
          seq_id = st.nextToken();
          while( st.hasMoreTokens() ){
            String item_id = st.nextToken();
            String item = (String)ht.get(item_id);
            fw.write("client_" + seq_id + ",id_" + trans_id + "," + item+"\n");
          }
          trans_id++;
        }

        br.close();
        fw.close();
      }catch(Exception e){ e.printStackTrace(); }

      Runtime runtime = Runtime.getRuntime();

      String cmd =
          "java -Xmx256M -classpath resources/lib/PrefixSpan/prefixspan.jar prefixspan_x_sp.PrefixSpan 90 "+PREFIX_SPAN_RESOLVER.INPUT+" "+PREFIX_SPAN_RESOLVER.OUTPUT+" null";

      try {

        System.out.println( cmd );
        Process p = runtime.exec( cmd );
        int exit_val = p.waitFor();

      }
      catch (Exception ec) {
      }

      PrefixSpanOutputFormatting psof = new PrefixSpanOutputFormatting(preprocessing.getNumberOfTransactions(), SequenceModel.max_number_of_sequences);
      TreeSet sequences = psof.buildSequences(ht);
      System.out.println(sequences.size());
      System.out.println(sequences);

      new File(PREFIX_SPAN_RESOLVER.INPUT).renameTo(new File(PREFIX_SPAN_RESOLVER.INPUT+"0"));
      new File(PREFIX_SPAN_RESOLVER.OUTPUT).renameTo(new File(PREFIX_SPAN_RESOLVER.OUTPUT+"0"));
      //new File("ds.txt").renameTo(new File("ds0.txt"));

      try {
        Thread.sleep(5000);
      }
      catch (Exception ex) {

      }

      preprocessing=null;
      ht=null;
      System.gc();

      fact = new ARFFFactory(f, "G:/WINDOWS/jbprojects/kddml/repository/data/ARFF/coop_timestamp.arff");
      fact.setOverwrite(true);
       t = (InternalTable) fact.newInstance();
      t.saveToRepository();

       preprocessing = new PrefixSpanInputFormatting(t);
      ht = preprocessing.execute();

      enum2 = ht.keys();
      c = ht.values();
      i = c.iterator();
      while(enum.hasMoreElements())
        System.out.println(enum.nextElement()+"\t"+i.next());

      //System.out.println(ht);

      try{
        BufferedReader br = new BufferedReader(new FileReader(PREFIX_SPAN_RESOLVER.INPUT));
        FileWriter fw = new FileWriter("transf2.txt");

        String seq_id="";
        int trans_id=0;

        String line;
        while( (line=br.readLine()) != null ){
          StringTokenizer st = new StringTokenizer(line);
          seq_id = st.nextToken();
          while( st.hasMoreTokens() ){
            String item_id = st.nextToken();
            String item = (String)ht.get(item_id);
            fw.write("client_" + seq_id + ",id_" + trans_id + "," + item+"\n");
          }
          trans_id++;
        }

        br.close();
        fw.close();
      }catch(Exception e){ e.printStackTrace(); }

      runtime = Runtime.getRuntime();

     cmd =
          "java -Xmx256M -classpath resources/lib/PrefixSpan/prefixspan.jar prefixspan_x_sp.PrefixSpan 90 "+PREFIX_SPAN_RESOLVER.INPUT+" "+PREFIX_SPAN_RESOLVER.OUTPUT+" null";

      try {

        System.out.println( cmd );
        Process p = runtime.exec( cmd );
        int exit_val = p.waitFor();

      }
      catch (Exception ec) {
      }

      psof = new PrefixSpanOutputFormatting(preprocessing.getNumberOfTransactions(), SequenceModel.max_number_of_sequences);
      sequences = psof.buildSequences(ht);
      System.out.println(sequences.size());
      System.out.println(sequences);

      //System.out.println(ht.toString());

      //PrefixSpanOutputFormatting psof = new PrefixSpanOutputFormatting(preprocessing.getNumberOfTransactions(), SequenceModel.max_number_of_sequences);
      //TreeSet sequences = psof.buildSequences(ht);

      /*int num_trans = preprocessing.getNumberOfTransactions();
      int max_items = preprocessing.getMaxNumberOfItemsPerTransaction();
      double avg_items = preprocessing.getAvgNumberOfItemsPerTransaction();
      String model_name = preprocessing.getModelName();
      MiningSchema mining_schema = preprocessing.getMiningSchema();
      DataDictionary data_dictionary = preprocessing.getDataDictionary();


      SequenceFactory seq_fact = new SequenceFactory(
        new File("C:/timestamp-alg-pref.xml"), data_dictionary, mining_schema, 0.9, 1.0, num_trans, sequences);

      // setta le proprieta' del factory
      seq_fact.setOverwrite(true);
      seq_fact.setAlgorithmName("PrefixSpan - Computer Science Departement, Pisa, Italy");
      seq_fact.setModelName(model_name+"_sequence");
      seq_fact.setMaxNumberOfItemsPerTA(max_items);
      seq_fact.setAvgNumberOfItemsPerTA(avg_items);

      SequenceModel output = (SequenceModel) seq_fact.newInstance();
      output.saveToRepository();*/

    /*}
    catch (KDDMLCoreException ex) {
      System.out.println(ex.getDetailErrorMessage());
    }
    catch (KDDMLOperatorException ex2) {
      System.out.println(ex2.getDetailErrorMessage());
    }*/
  }
}
