/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.DataMining.SequenceAlgorithms;

import java.io.*;
import java.util.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Operators.*;

/**
 * Questa classe estrae le sequenze frequenti e relativi supporti dal
 * file di output prodotto dall'algoritmo PrefixSpan.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
final class PrefixSpanOutputFormatting {

  /**
   * Numero di transazioni scandite dall'algoritmo PrefixSpan.
   */
  private int num_transactions;

  /**
   * Numero massimo di sequenze desiderate.
   * Parametro fornito in input dall'utente.
   */
  private int num_max_of_sequence;

  /**
   * Insieme di sequenze ordinate in base al supporto.
   */
  private TreeSet sequences;

  private String output_file;

  /**
   * Costruttore.
   * @param ht hashtable contenente le associazioni chiave --> item. Generata in fase
   * di preprocessing.
   * @param num_trans numero di transazioni.
   * @param num_max_itemset numero massimo di sequenze che devono essere estratti.
   */
  public PrefixSpanOutputFormatting(int num_trans,
                                    int num_max_sequence,
                                    String output_file) {

    if (num_max_sequence < SequenceModel.MAX_NUMBER_OF_SEQUENCES)
      num_max_of_sequence = num_max_sequence;
    else
      num_max_of_sequence = SequenceModel.MAX_NUMBER_OF_SEQUENCES;
    num_transactions = num_trans;
    CompareSequences comparator = new CompareSequences();
    sequences = new TreeSet(comparator);

    this.output_file = output_file;
  }

  /**
   * Costruisce le sequenze parsando il file prodotto in uscita dalla libreria PrefixSpan.
   * La prima riga del file contiene gli items frequenti nella forma:
   * 3 5 9 ...
   * Ogni riga pari del file contiene il pattern sequenziale massimo nella forma:
   * 14 (42,47);94.0%
   * dove i valori numerici rappresentano le chiavi associate agli item contenuti nella tabella
   * hash relativa. I valori tra parentesi rappresentano intra-transaction items.
   * Il supporto del pattern e' espresso in percentuale e segue il ";".
   *
   * Ogni riga dispari del file contiene tutti i sub-pattern del pattern sequenziale massimale nella forma:
   * 14 42;100.0%14 47;100.0%14;100.0%42;100.0%(42,47);94.0%47;100.0%
   *
   * @param ht tabella hash contenente le associazioni chiave --> item.
   */
  public TreeSet buildSequences(Hashtable ht) throws KDDMLOperatorException {

    String inputLine;

    try {

      FileReader reader = new FileReader(output_file);
      BufferedReader in = new BufferedReader(reader);

      String frequent_items = in.readLine();

      inputLine = in.readLine();

      while (inputLine != null) {

        StringTokenizer tokenizer = new StringTokenizer(inputLine, "%");

        while ( tokenizer.hasMoreTokens() ) {
          // pattern
          String pattern = tokenizer.nextToken();
          // modello rappresentante il pattern
          SequenceManager seq = buildSequence(ht, pattern);
          // aggiunge la sequenza al tree set mantenendo l'ordinamento stabilito...
          sequences.add(seq);

          //i risultati non sono ordinati in base al supporto...
          //Rischio di eliminare sequenze con un supporto piu' alto...
          if( sequences.size() == num_max_of_sequence ) {
            sequences.remove( (SequenceManager) sequences.last() );
          }
        }

        inputLine = in.readLine();
      }
      in.close();

    }
    catch (Exception e) {
      String msg1 = "I/O Error calling PrefixSpan preprocessing.";
      String msg2 = "I/O Error calling PrefixSpan preprocessing: "+e.toString();
      throw new KDDMLOperatorException(msg1, msg2, KDDMLOperatorErrorCodes.IO_ERROR);
    }

    return sequences;
  }

  /*
     Build a sequence given the string representation: 3 (10,13) 67;30.0
   */
  private SequenceManager buildSequence(Hashtable ht, String pattern){
    SequenceManager sequence;
    int support_delimiter_index = pattern.indexOf(';');

    String support_string = pattern.substring(support_delimiter_index + 1).trim();
    Double perc_support = Double.valueOf(support_string);
    double prob_support = perc_support.doubleValue() / 100.0;
    Double support = new Double(prob_support);

    String seq_elements = pattern.substring(0, support_delimiter_index);

    StringTokenizer tokenizer = new StringTokenizer(seq_elements);
    //Get the first sequence element to build the sequence...
    String element = tokenizer.nextToken();
    SequenceElementManager seq_element = buildSequenceElement(ht, element);
    //build the sequence...
    sequence = new Sequence(seq_element);

    while( tokenizer.hasMoreTokens() ){
      element = tokenizer.nextToken();
      seq_element = buildSequenceElement(ht, element);
      sequence.addElement(Delimiter.DEFAULT, seq_element);
    }

    sequence.setSupport(support);
    sequence.setOccurrence(new Integer( (int)(support.doubleValue()*this.num_transactions)) );//controllare

    return sequence;
  }

  // param element is "trimmed"
  private SequenceElementManager buildSequenceElement(Hashtable ht, String element){
    ElementItemSet seq_element = new ElementItemSet();

    if( element.startsWith("(") ) {
      String seq_element_items = element.substring(1, element.length() - 1 );
      StringTokenizer tokenizer = new StringTokenizer(seq_element_items, ",");
      while( tokenizer.hasMoreTokens() ) {
        String item_id = tokenizer.nextToken();
        String value = (String) ht.get(item_id);
        seq_element.addItem( new Item(value) );
      }
    }
    else {
      // identificatore dell'item == element
      // valore dell'item ottenuto attraverso la tabella hash
      String value = (String) ht.get(element);
      seq_element.addItem( new Item(value) );
    }

    return seq_element;
  }

}
