/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.DataMining;

import java.util.Hashtable;
import kddml.Operators.*;
import kddml.Operators.DataMining.*;
import kddml.Core.KDDMLScalarManager;

public class TREE_MINER_SETTINGS extends kddml.Operators.DataMining.DMAlgorithmSettings {

  protected String target_attribute;

  protected Integer max_depth;

  protected Double max_pvalue;

  protected Double confidence_for_pruning;

  protected Integer num_instances_for_leaf;

  protected Double percentage_splitting;

  public TREE_MINER_SETTINGS() {
    super();
    target_attribute = null;
    max_depth = null;
    max_pvalue = null;
    confidence_for_pruning = null;
    num_instances_for_leaf = null;
    percentage_splitting = null;

  }

  public void applySettings(java.util.Hashtable attributes)
      throws KDDMLOperatorException {

    super.applySettings(attributes);
    setTargetAttribute(attributes, "target_attribute", true);
  }


  /**
   * The target attribute name.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setTargetAttribute(Hashtable attributes,
                                    String attribute_name,
                                    boolean error_if_not_found) throws
      KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        target_attribute = (String) s.getExprValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a string .",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  /**
   * Sets the maximum depth of the tree model to be built.
   * This is a termination criterion. The build process halts the search for extensions of
   * any node at this depth.
   * The maximum depth must be a positive number and less than or equal to the maximum depth
   * allowed.
   *
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorSettingsException
   */
  protected void setMaxDepth(Hashtable attributes,
                             String attribute_name,
                             boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        max_depth = new Integer(((Double) s.getExprValue()).intValue());
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (max_depth.intValue() < 2)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Sets the largest acceptable probability of any target value to split a node.
   * A node with a target value probability exceeding the maximum value is considered to be
   * homogenous, i.e., no further splitting is required.
   * This node becomes terminal (a leaf node).
   * The maximum P value must be between 0 and 1.
   *
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorSettingsException
   */
  protected void setMaxPValue(Hashtable attributes,
                              String attribute_name,
                              boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        max_pvalue = ((Double) s.getExprValue());
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if ((max_pvalue.doubleValue() < 0) || (max_pvalue.doubleValue() >= 1))
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Sets the number of children at any interior node.
   * Choices are binary or k-ary where k >= 2.
   *
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorSettingsException
   */
  protected void setNumInstancesForLeaf(Hashtable attributes,
                                        String attribute_name,
                                        boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        num_instances_for_leaf = new Integer(((Double) s.getExprValue()).intValue());
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must a positive integer value > 2.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (num_instances_for_leaf.intValue() < 2)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value > 2.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * The confidence for pruning. A real between 0 and 1.
   *
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorSettingsException
   */
  protected void setConfidenceForPruning(Hashtable attributes,
                                         String attribute_name,
                                         boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        confidence_for_pruning = ((Double) s.getExprValue());
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if ((confidence_for_pruning.doubleValue() <= 0) || (confidence_for_pruning.doubleValue() > 1))
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * Randomly split training data in a real training data and an additional test data.
   * This is a real between 0 and 1.
   *
   * @param attributes Hashtable
   * @param attribute_name String
   * @param error_if_not_found boolean
   * @throws KDDMLOperatorException
   */
  protected void setPercentageSplitting(Hashtable attributes,
                                        String attribute_name,
                                        boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        percentage_splitting = ((Double) s.getExprValue());
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if ((percentage_splitting.doubleValue() <= 0) || (percentage_splitting.doubleValue() > 1))
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive double value between 0 and 1.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  public void setConfidenceForPrunining(double confidence) {
    this.confidence_for_pruning = new Double(confidence);
  }

  public Double getConfidenceForPruninig() {
    return this.confidence_for_pruning;
  }

  public void setMaxDepth(int max) {
    this.max_depth = new Integer(max);
  }

  public Integer getMaxDepth() {
    return this.max_depth;
  }

  public void setMaxPValue(double p_value) {
    this.max_pvalue = new Double(p_value);
  }

  public Double getMaxPValue() {
    return this.max_pvalue;
  }

  public void setNumInstancesForLeaf(int num) {
    this.num_instances_for_leaf = new Integer(num);
  }

  public Integer getNumInstancesForLeaf() {
    return this.num_instances_for_leaf;
  }

  public void setPercentageSplitting(double perc) {
    this.percentage_splitting = new Double(perc);
  }

  public Double getPercentageSplitting() {
    return this.percentage_splitting;
  }

  public void setTargetAttributeName(String name) {
    this.target_attribute = name;
  }

  public String getTargetAttributeName() {
    return target_attribute;
  }
}
