/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators;

import java.util.Hashtable;
import kddml.Core.KDDMLScalarManager;
import kddml.Core.KDDMLScalar;
import java.util.*;

/**
 * Interfaccia base per un gestire i settings delle proprieta' di un operatore.
 * I metodi di tale classe vengono richiamati in automatico dall'interprete del sistema.
 */

public abstract class KDDMLOperatorSettings {

  String operator_name;

  protected String xml_dest;

  /**
   * Constructor given the hashtalbe and the proprietary statement.
   *
   * @param attributes Hashtable
   * @param statement XMLProprietaryStatement the XML proprietary element for this
   * operator. It can be null if no proprietary element exist in operator's DTD.
   */
  public KDDMLOperatorSettings() {
    operator_name = getOperatorName();
    xml_dest = null;
  }

  /**
   * Apply the input settings to this class.
   *
   * @throws KDDMLOperatorException
   */
  public void applySettings(Hashtable attributes)
      throws KDDMLOperatorException {

    setXMLDestination(attributes, "xml_dest", true);
  }

  public void setXMLDestination(String xml_dest) {
    this.xml_dest = xml_dest;
  }

  protected void setXMLDestination(Hashtable attributes,
                                   String attribute_name,
                                   boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        xml_dest = (String) s.toString();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a string.",
             null,
             KDDMLOperatorErrorCodes.ILLEGAL_CASTING);
      }
    }
  }

  public String getXMLDestination() {
    return xml_dest;
  }

  protected boolean existAttribute(Hashtable attributes,
                                   String attribute_name,
                                   boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (!attributes.containsKey(attribute_name))
      if (error_if_not_found)
        throw new KDDMLOperatorException(
          operator_name+" : specify a value for the < "+attribute_name+" > attribute.",
          null,
          KDDMLOperatorErrorCodes.ILLEGAL_CASTING);
      else
        return false;
    return true;
  }

  public String getOperatorName() {
    String operator_name = this.getClass().getName();
    String op1 = operator_name.substring(operator_name.lastIndexOf(".")+1);
    return op1.substring(0,op1.lastIndexOf("_"));
  }

  /**
   * Checks if the input attribute is an integer and returns its value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Integer the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Integer the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Integer checkAsInteger(Hashtable parameters,
                                   String parameter_name,
                                   boolean parameter_is_required,
                                   Integer default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be an integer value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          int value = d.intValue();
          return new Integer(value);
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/

  /**
   * Checks if the input attribute is a positive integer greater than 0 and returns its
   * value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Integer the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Integer the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Integer checkAsPositiveInteger(Hashtable parameters,
                                 String parameter_name,
                                 boolean parameter_is_required,
                                 Integer default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a positive integer value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          int value = d.intValue();
          if (value < 1)
            throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
          else
            return new Integer(value);
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/


  /**
   * Checks if the input attribute is a non_negative integer greater or equal than 0 and
   * returns its value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Integer the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Integer the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Integer checkAsNonNegativeInteger(Hashtable parameters,
                                              String parameter_name,
                                              boolean parameter_is_required,
                                              Integer default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a non-negative integer value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          int value = d.intValue();
          if (value < 0)
            throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
          else
            return new Integer(value);
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/


  /**
   * Checks if the input attribute is an real and returns its value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Double the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Double the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Double checkAsReal(Hashtable parameters,
                               String parameter_name,
                               boolean parameter_is_required,
                               Double default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a real value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          return d;
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/

  /**
   * Checks if the input attribute is a positive real greater than 0 and returns its
   * value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Double the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Double the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Double checkAsPositiveReal(Hashtable parameters,
                                       String parameter_name,
                                       boolean parameter_is_required,
                                       Double default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a positive real value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          if (d.doubleValue() < 1)
            throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
          else
            return d;
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/


  /**
   * Checks if the input attribute is a non-negative real greater or equal than 0 and returns
   * its value as output.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then the default_value is returned. The default_value
   * can be null, so, the method can return null. <br>
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param default_value Double the default value to be used if the attribute is not
   * found in the hashtable. Can be null, if the attribute do not admit a default
   * value.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return Double the parameter value. Can be null if the attribute is not required
   * and a default value is not provided.
   */
  /*protected Double checkAsNonNeagativeReal(Hashtable parameters,
                                           String parameter_name,
                                           boolean parameter_is_required,
                                           Double default_value)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a non-negative real value.";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        s.evaluateAsMathExpression();
        if (s.getExprType() != KDDMLScalar.NUMERIC) {
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
        else {
          Double d = new Double(s.toString());
          if (d.doubleValue() < 0)
            throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
          else
            return d;
        }
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return default_value;
    }
  }*/


  /**
   * Checks if the input attribute is a list of numbers comma separated (e.g.
   * 2, 5, -23, 0, 8.344) and returns a vector of doubles.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then null is returned.
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return double[] the parameter value as vector of doubles. Can be null if the attribute
   * is not required and it do not exist in the hashtable.
   */
  protected double[] checkAsNumberList(Hashtable parameters,
                                       String parameter_name,
                                       boolean parameter_is_required)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a list of numbers in " +
          "a comma separated format (e.g. 3, 0, 9.8, -12).";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        String list = s.toString();
        StringTokenizer st = new StringTokenizer(list, ",");
        String word;
        Vector word_as_string = new Vector();

        while (st.hasMoreElements()) {
          word = st.nextToken();
          word_as_string.add(word.trim());
        }

        // converte le stringhe in reali
        double[] ris = new double[word_as_string.size()];
        for (int i=0; i<ris.length; i++)
          ris[i] = new Double((String) word_as_string.get(i)).doubleValue();
        return ris;
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return null;
    }
  }

  /**
   * Checks if the input attribute is a list of strings comma separated (e.g.
   * temperature, outlook, play) and returns a vector of strings.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then null is returned.
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the appropriate type. If the value has a wrong type, then an exception
   * is rased by the method. Otherwise, the value is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return double[] the parameter value as vector of strings. Can be null if the attribute
   * is not required and it do not exist in the hashtable.
   */
  protected String[] checkAsStringList(Hashtable parameters,
                                       String parameter_name,
                                       boolean parameter_is_required)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a list of strings in " +
          "a comma separated format (e.g. outlook, temperature, play).";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        String list = s.toString();
        StringTokenizer st = new StringTokenizer(list, ",");
        String word;
        Vector word_as_string = new Vector();

        while (st.hasMoreElements()) {
          word = st.nextToken();
          word_as_string.add(word.trim());
        }

        String[] ris = new String[word_as_string.size()];
        for (int i=0; i<ris.length; i++)
          ris[i] = (String) word_as_string.get(i);
        return ris;
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return null;
    }
  }

  /**
   * Checks if the input attribute is a value belonging to a set of categories and returns
   * the category value as string.
   * The attribute is identified by the attribute_name field in the input hashtable.<br>
   * If the parameter name is not found in the hashtable and the field parameter_is_required
   * is true (i.e. the attribute is required in the XML specification), then an
   * exception is raised by the method. <br>
   * Otherwise, it the parameter name is not found in the hashtable, but the
   * parameter_is_required field is false (i.e. the attribute is optional in the
   * XML specification), then null is returned.
   * Finally, if the attribute is found in the hashtable, then his value is compared
   * with the list of categories provided by the input field. If the value do not belongs
   * to the set of categories, then an exception is rased by the method.
   * Otherwise, the matching category is returned as output.
   *
   * @param attributes Hashtable the hashtable containing the list of parameters about this
   * operator as XML attributes. It contains a set of pairs with the parameter name
   * as key and the parameter value as value. The hashtable is generated by the system
   * interpreter.
   * @param parameter_name String the attribute to check. It is the key in the hashtable.
   * Cannot be null.
   * @param parameter_is_required boolean true if the attribute is required in the
   * operator specification; false otherwise.
   * @param categories String[] the set of categories accepted by the parameter.
   * @throws KDDMLOperatorException if the parameter value has a wrong type or if it
   * do not exist in the hashtable and his value is required.
   * @return String the category value that match the input parameter value. Can be null if
   * the attribute is not required and it do not exist in the hashtable.
   */
  protected String checkAsEnumerated(Hashtable parameters,
                                     String parameter_name,
                                     boolean parameter_is_required,
                                     String[] categories)
      throws KDDMLOperatorException {

    if (existAttribute(parameters, parameter_name, parameter_is_required)) {

      String msg = "< " + parameter_name + " >" + " must be a value in {";
      for (int i=0; i<categories.length; i++)
        msg = msg + categories[i] + ", ";
      msg = msg + "}";
      // the parameter exists in the hashtable
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) parameters.get(parameter_name);
        String value = s.toString();

        for (int i=0; i<categories.length; i++)
          if (value.equalsIgnoreCase(categories[i]))
            return value;

        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      catch (Exception e) {
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
    else {
      // the parameter do not exist in the hashtable
      // and his value is not required...
      return null;
    }
  }


  protected Integer checkAsIntegerRange(Hashtable parameters,
                                        String parameter_name,
                                        boolean parameter_is_required,
                                        Integer lower_bound,
                                        boolean lower_bound_is_closed,
                                        Integer upper_bound,
                                        boolean upper_bound_is_closed)
      throws KDDMLOperatorException {

    return null;
  }

  protected Double checkAsIntegerRange(Hashtable attributes,
                                       String attribute_name,
                                       boolean error_if_not_found,
                                       double lower_bound,
                                       double upper_bound)
      throws KDDMLOperatorException {

    return null;
  }
}

class Interval {

  private Integer lower_bound = null;

  private Integer upper_bound = null;

  protected byte closure;

  protected byte OPEN_OPEN = 0;

  protected byte CLOSED_CLOSED = 1;

  protected byte OPEN_CLOSED = 2;

  protected byte CLOSED_OPEN = 3;

  protected Interval(Integer lower_bound, Integer upper_bound, byte closure)
      throws KDDMLOperatorException {

    if( lower_bound != null && upper_bound != null ) {

      //open_open
      if(closure == OPEN_OPEN){
        // upper >= lower
        if (upper_bound.intValue() < lower_bound.intValue())
          throw new KDDMLOperatorException
              ("Upper bound must be greater or equal than lower bound in interval!",
               null,
               kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
      }
      //open_closed, closed_open, closed_closed
      else {
        // upper > lower
        if (upper_bound.intValue() <= lower_bound.intValue())
          throw new KDDMLOperatorException
              ("Upper bound must be greater than lower bound in interval!",
               null,
               kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
      }

      this.lower_bound = lower_bound;
      this.upper_bound = upper_bound;

    }
    this.closure = closure;
  }

  protected Interval(Double lower_bound, Double upper_bound, byte closure) {

  }

  public boolean isInRange(int value) {
    return ((checkLowerBound(value)) && (checkUpperBound(value)));
  }

  public String toString() {
    String l_s, u_s;
    String lb, ub;
    if ((closure == CLOSED_CLOSED) || (closure == CLOSED_OPEN))
      l_s = "[";
    else
      l_s = "(";
    if ((closure == CLOSED_CLOSED) || (closure == OPEN_CLOSED))
      u_s = "]";
    else
      u_s = ")";

    if (lower_bound == null)
      lb = "-inf";
    else
      lb = lower_bound.toString();

    if (upper_bound == null)
      ub = "-inf";
    else
      ub = upper_bound.toString();

    return l_s + lb + ", " + ub + u_s;
  }

  private boolean checkLowerBound(int value) {
    if (lower_bound == null)
      // -infinite is assumed
      return true;
    if ((closure == CLOSED_CLOSED) || (closure == CLOSED_OPEN))
      return (lower_bound.intValue() <= value);
    else
      return (lower_bound.intValue() < value);
  }

  private boolean checkUpperBound(int value) {
    if (upper_bound == null)
      // -infinite is assumed
      return true;
    if ((closure == CLOSED_CLOSED) || (closure == OPEN_CLOSED))
      return (upper_bound.intValue() >= value);
    else
      return (upper_bound.intValue() > value);
  }
}
