/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.Clustering.*;
import kddml.Core.DataSources.*;
import weka.core.*;

public class CLUSTER_NUMBER_RESOLVER extends PostProcessingResolver {

  ClusteringModelManager model;

  int cluster_number;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    model = (ClusteringModelManager) arguments.get(0);
    InternalTableManager input = (InternalTableManager) arguments.get(1);

    // controlla la compatibilita' del modello con le istanze...
    if (!model.isCompatible((DataStatistics) input.getStatistic())) {
      String msg = "The input clustering model is incompatible with the data source!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_TABLE.getRepositoryPath() +
                                         settings.getXMLDestination());

    WekaFactory fact = new WekaFactory(dest, (Instances) input.load(0));
    // set overwrite to be true
    fact.setOverwrite(true);
    // create an internal table from input internal table with no instances
    InternalTable output = (InternalTable) fact.newInstance();

    // get the cluster number
    cluster_number = ((CLUSTER_NUMBER_SETTINGS) settings).getClusterNumber();
    Instances tuple;

    if (cluster_number == -1) {
      // cluster max required...
      output.addInstances(getMaxClusterInstances(input));
      return output;
    }

    // identifier of cluster specified by the user
    if (cluster_number >= model.getNumberOfClusters()) {
      String msg = "Cluster number "+cluster_number+" do not exist in the model!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    long opt = input.optimize();
    // ripristina il puntatore alle istanze
    input.reset();
    while (input.hasNext()) {
      tuple = (Instances) input.load(opt);
      tuple = loadClusterNumber(tuple);
      output.addInstances(tuple);
    }

    return output;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==1)
      return KDDMLObjectType.INTERNAL_TABLE;
    if (i==0)
      return KDDMLObjectType.PMML_CLUSTERING_MODEL;
    return null;
  }

  /**
   * Return true if the operator performe an high number of access to the hard disk.
   *
   * @return boolean
   */
  public boolean highDiskAccessIsNeeded() {
    return true;
  }


  private Instances loadClusterNumber(Instances instances) throws KDDMLCoreException {

    int removed = 0;
    int num_instances = instances.numInstances();
    for (int i=0; i<num_instances; i++) {
      int cluster_id = ((Cluster) model.getCluster(instances.instance(i-removed))).getClusterId();
      System.out.println("Instance "+instances.instance(i-removed)+ ": "+cluster_id);
      if (cluster_id != cluster_number) {
        instances.delete(i - removed);
        removed++;
      }
    }
    return instances;
  }

  private Instances getMaxClusterInstances(InternalTableManager table) throws KDDMLCoreException {

    // cluster_index[] contiene l'indice del cluster per ogni istanza
    int[] cluster_index = new int[table.getNumOfInstances()];
    int counter = 0;

    // counts[] contiene il numero di istance in ogni cluster
    int[] counts = new int[model.getNumberOfClusters()];
    for (int i=0; i<counts.length; i++)
      counts[i] = 0;

    long opt = table.optimize();
    Instances tuple;

    while (table.hasNext()) {
      tuple = (Instances) table.load(opt);
      for (int i=0; i<tuple.numInstances(); i++) {
        int cluster_id = ( (Cluster) model.getCluster(tuple.instance(i))).getClusterId();
        System.out.println(tuple.instance(i)+" --> "+cluster_id);
        cluster_index[counter] = cluster_id;
        counter++;
        counts[cluster_id]++;
      }
    }

    // indice del cluster di massima cardinalita'
    int cluster_max = getMaxValue(counts);
    // numero di istanze per il cluster di massima cardinalita'
    int num_inst_cluster_max = counts[cluster_max];

    // instances_index[] contiene gli indici delle istanze che appartengono al cluster
    int[] instances_index = new int[num_inst_cluster_max];
    counter = 0;
    for (int j=0; j<cluster_index.length; j++) {
      if (cluster_index[j] == cluster_max) {
        instances_index[counter] = j;
        counter++;
      }
    }
    return (Instances) table.loadInstances(instances_index);
  }

  /**
   * Get the index of the maximum element of an array.
   *
   * @param array int[]
   * @return int
   */
  private static int getMaxValue(int[] array) {
    double max = Integer.MIN_VALUE;
    int index = -1;
    for(int i = array.length - 1; i >= 0; i--)
      if(array[i] > max) {
        max = array[i];
        index = i;
      }

    return index;
  }
}
