
/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import java.io.*;
import kddml.Core.DataSources.*;
import weka.core.*;

public class CLUSTER_PARTITION_SPARROW_RESOLVER extends PostProcessingResolver {

  private final static String SPARROW_EXE =
      System.getProperty("user.dir") + File.separator + "resources" +
      File.separator + "lib" + File.separator + "algorithms" +
      File.separator + "Sparrow" + File.separator;

  protected final static String JVM_OPTIONS = "-Xmx256M";

  protected final static String INPUT_DATA = "input_sparrow.arff";

  protected final static String OUTPUT_DATA = "output_sparrow.arff";


  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {


    // la tabella in ingresso ha due attributi di tipo numerico che rappresentano
    // le coordinate spaziali (x,y).
    // Non sono ammessi valori missing.

    // la catella di lavoro e' SPARROW

    InternalTableManager instances = (InternalTableManager) arguments.get(0);
    DataStatisticsManager schema = instances.getStatistic();
    checkLogicalData(schema);

    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_TABLE.getRepositoryPath() +
                                         settings.getXMLDestination());

    // read parameters
    int agents = ((CLUSTER_PARTITION_SPARROW_SETTINGS) settings).getNumAgents();
    int iterations = ((CLUSTER_PARTITION_SPARROW_SETTINGS) settings).getNumIterations();
    int density = ((CLUSTER_PARTITION_SPARROW_SETTINGS) settings).getDensityThreshold();
    int radius = ((CLUSTER_PARTITION_SPARROW_SETTINGS) settings).getVisibilityRadius();

    System.out.println("------------------------------------------");
    System.out.println("SPARROW PARAMETERS: ");
    System.out.println("XML Destination - "+dest);
    System.out.println("Number of agents - "+agents);
    System.out.println("Number of Iterations - "+iterations);
    System.out.println("Density threshold - "+density);
    System.out.println("Visibility radius - "+radius);
    System.out.println("------------------------------------------");

    inputSparrowFormatting(instances);

    runSparrow(INPUT_DATA, OUTPUT_DATA, agents, iterations, density, radius);
    InternalTable output = outputSparrowFormatting(OUTPUT_DATA, dest);

    removeTmpFiles();
    return output;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return KDDMLObjectType.INTERNAL_TABLE;
    return null;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  /**
   * Checks the input logica data schema.
   * The input table has only two numeric attributes.
   * The first one is the x-coodinate in the space. The second one is the y-coordinate in the space.
   * The table do not admit missing values.
   *
   * @param stat DataStatisticsManager
   */
  private void checkLogicalData(DataStatisticsManager stat) throws KDDMLOperatorException {

    String msg1 = "Input table must have only two numeric attributes for the SPARROW algorithm";
    String msg2 = "No missing values are allowed in input table for the SPARROW algorithm";
    if (stat.getNumAttributes() != 2)
      throw new KDDMLOperatorException(msg1, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);

    AttributeStatisticManager[] as = stat.getStatistics();
    if (!as[0].getType().equals(AttributeType.NUMERIC))
      throw new KDDMLOperatorException(msg1, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
    if (!as[1].getType().equals(AttributeType.NUMERIC))
      throw new KDDMLOperatorException(msg1, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);

    // no missing values are allowed
    if (as[0].getMissingCount() != 0)
      throw new KDDMLOperatorException(msg2, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
  }

  /**
   * Formatting of the input instances.
   * The input for the SPARROW algorithm is an ARFF file generated in the wording directory
   * of the agorithm.
   *
   * @param instances InternalTableManager
   * @throws KDDMLOperatorException
   */
  private static void inputSparrowFormatting(InternalTableManager instances)
      throws KDDMLOperatorException {
    FileWriter output = null;
    int num_tuple = 0;
    try {
      long opt = instances.optimize();
      output = new FileWriter(SPARROW_EXE + INPUT_DATA);

      // schema import...
      Instances tuple = (Instances) instances.load(0);
      output.write(tuple.toString());

      while (instances.hasNext()) {

        // instances import...
        tuple = (Instances) instances.load(opt);
        num_tuple += opt;
        for (int i=0; i<tuple.numInstances(); i++)
          output.write(tuple.instance(i).toString()+"\n");
      }
      output.close();
    }
    catch(Exception e) {
      throw new KDDMLOperatorException("I/O Error!", e.toString(), KDDMLOperatorErrorCodes.IO_ERROR);
    }
    System.out.println("Input SPARROW generated in: "+SPARROW_EXE + INPUT_DATA);

  }

  /**
   * Run the sparrow algorithm.
   *
   * @param input_arff_name String the ARFF file name containing the input instances.
   * @param output_arff_name String the ARFF file name containing the result. The output
   * instances have a new attribute at the end of the schema (cluster_number) containing
   * the cluster number related to each single instance.
   * @param na int a positive integer representing the number of agents.
   * @param ni int a positive integer representing the maximum number of iterations
   * @param density int a positive integere representing the density threshold
   * @param radius int a positive integer representing the radiuts.
   * @throws KDDMLOperatorException if a generic error occours.
   */
  private static void runSparrow(String input_arff_name, String output_arff_name,
                                 int na, int ni, int density, int radius)
      throws KDDMLOperatorException {

    Runtime runtime = Runtime.getRuntime();

    try {
      int result = -1;
      if (System.getProperty("os.name").indexOf("Windows") != -1) {

        // Windows operative System
        result = callSparrowForWindows(input_arff_name, output_arff_name, na, ni, density, radius);
      }

      if (System.getProperty("os.name").indexOf("Linux") != -1) {

        // Linux operative System
        result = callSparrowForLinux(input_arff_name, output_arff_name, na, ni, density, radius);

      }

      if (result != 0)
        throw new RuntimeException("Error executing SPARROW libraries: exit code " + result + ".");
    }
    catch (Exception ec) {
      String msg1 = "Error invoking SPARROW libraries: invalid command!!!";
      String msg2 = ec.toString();
      removeTmpFiles();
      throw new KDDMLOperatorException(msg1, msg2,
                                       KDDMLOperatorErrorCodes.ERROR_INVOKING_EXTERNAL_LIBRARIES);
    }
    System.out.println("SPARROW output created in: "+ SPARROW_EXE + OUTPUT_DATA);

  }

  /*private static String getCmdForLinux(String input_arff_name, String output_arff_name,
                                       int na, int ni, int density, int radius) {
    String cmd =
        "java "+JVM_OPTIONS+" -classpath \"" + SPARROW_EXE + "\" -jar "+SPARROW_EXE+"sparrow.jar" + " " +SPARROW_EXE+
        input_arff_name + " "
        +SPARROW_EXE+ output_arff_name + " " +
        na + " " +
        ni + " " +
        density + " " +
        radius;
    return cmd;
  }*/

  private static int callSparrowForWindows(String input_arff_name, String output_arff_name,
                                           int na, int ni, int density, int radius)
      throws IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();
    String exec =
        "java "+JVM_OPTIONS+
        //" -classpath \"" + SPARROW_EXE + "\""+
        " -jar "+"\""+SPARROW_EXE+"sparrow.jar" + "\"" + " " +
        "\"" + SPARROW_EXE + input_arff_name +"\""+ " "+
        "\"" + SPARROW_EXE + output_arff_name + "\"" + " " +
        na + " " +
        ni + " " +
        density + " " +
        radius;

    System.out.println("Invoking SPARROW libraries...");

    Process p = esegui.exec(exec);
    p.waitFor();

    return p.exitValue();
  }

  private static int callSparrowForLinux(String input_arff_name, String output_arff_name,
                                         int na, int ni, int density, int radius)
      throws IOException, InterruptedException {

    Runtime esegui = Runtime.getRuntime();
    String[] exec = {
        "java",
        JVM_OPTIONS,
        "-jar",
        SPARROW_EXE+"sparrow.jar",
        SPARROW_EXE + input_arff_name,
        SPARROW_EXE + output_arff_name,
        new Integer(na).toString(),
        new Integer(ni).toString(),
        new Integer(density).toString(),
        new Integer(radius).toString()};

    System.out.println("Invoking SPARROW libraries...");
    //System.out.println(exec);

    Process p = esegui.exec(exec);
    p.waitFor();

    return p.exitValue();

  }


  /**
   * Output formatting of the SPARROW algorithm.
   *
   * @param arff_name String
   * @param xml_dest File
   * @throws KDDMLCoreException
   * @return InternalTable
   */
  private static InternalTable outputSparrowFormatting(String arff_name, File xml_dest)
      throws KDDMLCoreException {

    String arff_source = SPARROW_EXE + arff_name;
    ARFFFactory f = new ARFFFactory(xml_dest, arff_source);
    f.setMaxNumberOfLoadedInstAtIteration(10000);
    f.setOverwrite(true);
    return (InternalTable) f.newInstance();
  }

  private static void removeTmpFiles() {
    File f = new File(SPARROW_EXE + INPUT_DATA);
    if (f.exists()) {
      boolean b = f.delete();
      System.out.println("Removing " + f.toString() + ": " + b);
      if (!b)
        System.err.println("Warning!!! Enable to remove: "+ f.toString());
    }

    f = new File(SPARROW_EXE + OUTPUT_DATA);
    if (f.exists()) {
      boolean b = f.delete();
      System.out.println("Removing " + f.toString() + ": " + b);
      if (!b)
        System.err.println("Warning!!! Enable to remove: "+ f.toString());
    }
  }
}
