/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.Postprocessing;

import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.KDDMLOperatorErrorCodes;
import java.util.Hashtable;
import kddml.Core.KDDMLScalarManager;

/**
 * This class contains the settings definition for the SPARROW algorithm.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class CLUSTER_PARTITION_SPARROW_SETTINGS extends PostProcessingSettings {

  /**
   * The number of agents used by algorithm.
   * It is a positive integer. Default: 100.
   */
  private int num_agents;

  /**
   * The maximum number of iterations before the algorithm converges.
   * It is a positive integer. Default: 200.
   */
  private int num_iterations;

  /**
   * The density threshold used by algorithm during the cluster mining.
   * It is a positive integer. Default: 19.
   */
  private int density_threshold;

  /**
   * The visibility radius used by the algorithm.
   * It is a positive integer. Default: 9.
   */
  private int visibility_radius;

  /**
   * Empty constructor.
   */
  public CLUSTER_PARTITION_SPARROW_SETTINGS() {
    super();
    num_agents = 100;
    num_iterations = 200;
    density_threshold = 19;
    visibility_radius = 9;
  }

  /**
   * Apply the input settings to this class.
   *
   * @throws KDDMLOperatorException
   */
  public void applySettings(java.util.Hashtable attributes)
      throws kddml.Operators.KDDMLOperatorException {

    super.applySettings(attributes);
    setNumAgents(attributes, "num_agents", false);
    this.setNumIterations(attributes, "num_iterations", false);
    this.setDensityThreshold(attributes, "density_threshold", false);
    this.setVisibilityRadius(attributes, "visibility_radius", false);
  }

  protected void setNumAgents(Hashtable attributes,
                              String attribute_name,
                              boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        num_agents = ((Double) s.getExprValue()).intValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (num_agents < 1) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + "  must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  protected void setNumIterations(Hashtable attributes,
                              String attribute_name,
                              boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        num_iterations = ((Double) s.getExprValue()).intValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (num_iterations < 1) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + "  must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  protected void setDensityThreshold(Hashtable attributes,
                                     String attribute_name,
                                     boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        density_threshold = ((Double) s.getExprValue()).intValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (density_threshold < 1) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + "  must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  protected void setVisibilityRadius(Hashtable attributes,
                                     String attribute_name,
                                     boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        visibility_radius = ((Double) s.getExprValue()).intValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + " must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (visibility_radius < 1) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" + "  must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  public int getNumAgents() {
    return num_agents;
  }

  public void setNumAgents(int num_agents) {
    this.num_agents = num_agents;
  }

  public int getNumIterations() {
    return num_iterations;
  }

  public void setNumIterations(int num_iter) {
    this.num_iterations = num_iter;
  }

  public int getDensityThreshold() {
    return density_threshold;
  }

  public void setDensityThreshold(int density) {
    this.density_threshold = density;
  }

  public int getVisibilityRadius() {
    return visibility_radius;
  }

  public void setVisibilityRadius(int radius) {
    this.visibility_radius = radius;
  }
}
