/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Core.DataMining.Taxonomy.*;
import java.util.Iterator;
import java.util.Vector;

public class RDA_PRESERVED_RESOLVER extends PostProcessingResolver {

  private HierarchyModelManager hierarchy;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException {

    // get the hierarchy
    hierarchy = (HierarchyModelManager) arguments.get(0);

    // get the first association model
    AssociationModelManager m = (AssociationModelManager) arguments.get(1);

    // get output destination
    java.io.File f = new java.io.File(kddml.Core.KDDMLObjectType.PMML_ASSOCIATION_MODEL.getRepositoryPath() +
                                      settings.getXMLDestination());

    // clone the first association model as output
    AssociationModel cloned = m.clone(f);

    // iteratore sul primo insieme di regole...
    Iterator iter1 = m.getAssociationRules();

    // questo array serve a contenere le regole estratte ai vari livelli
    //AssociationRule[] rules = new AssociationRule[arguments.size()-1];

    AssociationRuleManager rule1 = null;
    AssociationRuleManager rule2 = null;
    while (iter1.hasNext()) {
      rule1 = (AssociationRuleManager) iter1.next();

      // ciclo che scorre gli altri association models
      int i = 2;
      boolean unsupported = false;
      //rules[0] = rule1;
      while (i<arguments.size() && !unsupported) {
        rule2 = getGeneralization(rule1, (AssociationModelManager) arguments.get(i));

        if (rule2 == null)
          // questa regola non si supporta a questo livello --> exit
          unsupported = true;
        //else
          // la regola si supporta a questo livello --> aggiungo all'array
          //rules[i-1] = rule2;
        i++;
      }
      if (!unsupported)
        // devo aggiungere la regola foglia al modello
        cloned.addRule(rule1);
      else
        // rimuovo la regola dal modello e ricalcolo gli itemsets
        cloned.removeRule(rule1, true);
    }

    return cloned;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_HIERARCHY;
    if (i == 1)
      return KDDMLObjectType.PMML_ASSOCIATION_MODEL;
    if (i == 2)
      return KDDMLObjectType.PMML_ASSOCIATION_MODEL;
    return null;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  /**
   * Testa se una regola e' supportata da un insieme di regole all'interno di una
   * gerarchia. Una regola e' supportata se esiste la sua generalizzazione.
   * Restituisce la generalizzazione dell'insieme di regole in ingresso.
   *
   * @param hierarchy HierarchyModel
   * @param rule AssociationRule
   * @param rules AssociationModel
   * @return AssociationRule
   */
  private AssociationRuleManager getGeneralization(AssociationRuleManager rule,
                                            AssociationModelManager rules)
      throws KDDMLOperatorException {

    Iterator iter = rules.getAssociationRules();
    // scorre le regole fino a trovare la generalizzazione
    while (iter.hasNext()) {
      AssociationRuleManager r = (AssociationRuleManager) iter.next();
      if (isGeneralized(rule, r))
        return r;
    }
    return null;
  }

  /**
   * Testa se due regole estratte a due differenti livelli della gerarchia sono
   * generalizzabili.
   *
   * @param rule1 AssociationRule la regola al livello 0 della gerarchia.
   * @param rule2 AssociationRule la regola generalizzata.
   * @return boolean true se rule2 risulta una generalizzazione di rule1.
   */
  private boolean isGeneralized(AssociationRuleManager rule1,
                                AssociationRuleManager rule2)
      throws KDDMLOperatorException {

    ItemsetManager antecedent1 = (ItemsetManager) rule1.getAntecedent();
    ItemsetManager consequent1 = (ItemsetManager) rule1.getConsequent();

    ItemsetManager antecedent2 = (ItemsetManager) rule2.getAntecedent();
    ItemsetManager consequent2 = (ItemsetManager) rule2.getConsequent();

    if ((isGeneralized(antecedent1, antecedent2)) &&
        (isGeneralized(consequent1, consequent2)))
      return true;
    else
      return false;
  }

  private boolean isGeneralized(ItemsetManager itemset1,
                                ItemsetManager itemset2)
      throws KDDMLOperatorException {

    // ottiene la generalizzazione per ogni singolo item
    Iterator iter = itemset1.getItems();

    while(iter.hasNext()) {
      String item = ((Item) iter.next()).toString();
      HierarchyNodeManager node = (HierarchyNodeManager) hierarchy.getNode(item);
      if (node != null)
        if (!node.isLeaf()) {
          String msg = "The first set of association rules must be extracted at the "+
              "leaves of the hierarchy";
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
        }
      boolean found = false;
      while((node != null) && !found) {
        node = (HierarchyNodeManager) node.getParent();
        if (node == null)
          break;
        String gen = node.getName();
        Item item2 = new Item(gen);
        if (itemset2.contain(item2)) {
          // generalizzazione trovata per l'item corrente....
          found = true;
        }
      }
      if (!found)
        return false;
    }
    if (itemset1.getSize() < itemset2.getSize())
      return false;
    else
      return true;
  }

}

