/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.AssociationRules.*;
import kddml.Core.DataSources.*;
import weka.core.*;
import java.util.Iterator;
import java.util.Vector;

public class RDA_SATISFY_RESOLVER extends PostProcessingResolver {

  // input rda
  AssociationModelManager rules;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // si distingue l'esecuzione dell'operatore se la tabella in ingresso e'
    // di tipo relazionale oppure di tipo transazionale.

    // get association rules...
    rules = (AssociationModelManager) arguments.get(0);

    // get input internal table...
    InternalTableManager table = (InternalTableManager) arguments.get(1);

    // controlla la compatibilita' del modello con le istanze...
    if (!rules.isCompatible((DataStatisticsManager) table.getStatistic())) {
      String msg = "The input association model is not compatible with the data source!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INCOMPATIBLE_OBJECTS);
    }


    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_TABLE.getRepositoryPath() +
                                         settings.getXMLDestination());

    // get the evaluation type: rules or itemsets
    boolean b = ((RDA_SATISFY_SETTINGS) settings).evaluate_on_rules;

    if (!table.isTransactionalTable()) {
      System.out.println("Relational Table.");
      return executeRelational(table, dest, b);
    }
    else {
      System.out.println("Transactional Table.");
      return executeTransactional((TransactionalInternalTable) table, dest, b);
    }
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  /**
   * Return true if the operator performe an high number of access to the hard disk.
   *
   * @return boolean
   */
  public boolean highDiskAccessIsNeeded() {
    return true;
  }

  /**
   * Execute the operator given a relational internal table.
   *
   * @param settings KDDMLOperatorSettings
   * @param table InternalTable
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  private kddml.Core.KDDMLObject executeRelational(InternalTableManager table,
                                                   java.io.File output_dest,
                                                   boolean on_rules)
    throws KDDMLOperatorException, KDDMLCoreException {

    // get the instances schema
    Instances tuple = (Instances) table.load(0);
    // servono a causa dell'eliminazione di attributi numerici
    Instances tuple_cloned;
    WekaFactory fact = new WekaFactory(output_dest, tuple);
    // set overwrite to be true
    fact.setOverwrite(true);
    // create an internal table from input internal table with no instances
    InternalTable output = (InternalTable) fact.newInstance();

    long opt = table.optimize();
    while (table.hasNext()) {
      // get a subset of instances
      tuple = (Instances) table.load(opt);
      tuple_cloned = new Instances(tuple);
      Transaction t;

      // rimuove dalle tuple ogni attributo di tipo numerico
      tuple = deleteNumericAttributes(tuple);

      Vector supported = new Vector();

      for (int i=0; i<tuple.numInstances(); i++) {
        t = new Transaction(tuple.instance(i));

        boolean b;
        if (on_rules)
          b = isSupportedByRda(t);
        else
          b = isSupportedByItemset(t);

        supported.add(new Boolean(b));
      }
      for (int k=0; k<supported.size(); k++) {
        if (!((Boolean) supported.get(k)).booleanValue()) {
          // rimuove l'istanza k
          tuple_cloned.delete(k);
          // rimuove l'elemento del vettore in posizione k
          supported.remove(k);
        }
      }
      output.addInstances(tuple_cloned);
    }

    return output;
  }

  /**
   * Execute the operator given a transactional internal table.
   *
   * @param settings KDDMLOperatorSettings
   * @param table TransactionalInternalTable
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  private kddml.Core.KDDMLObject executeTransactional(TransactionalInternalTable table,
                                                     java.io.File output_dest,
                                                     boolean on_rules)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get the instances schema
    Instances tuple = (Instances) table.load(0);
    WekaFactory fact = new WekaFactory(output_dest, tuple);
    // set overwrite to be true
    fact.setOverwrite(true);
    // create an internal table from input internal table with no instances
    InternalTable output = (InternalTable) fact.newInstance();

    Transaction t;
    boolean b;
    int num_trans = 0;
    // scorre le singole transazioni
    while(table.hasNextTransaction()) {
      t = (Transaction) table.loadNextTransaction();
      num_trans++;
      if (on_rules)
        b = isSupportedByRda(t);
      else
        b = isSupportedByItemset(t);

      if (b) {
        // transazione supportata: ottiene le istanze dalla tabella in input
        tuple = (Instances) table.load(t.getNumberOfItems());
        // e le aggiunge alla tabella in output
        output.addInstances(tuple);
      }
      else
        // transazione non supportata: si scorre il puntatore
        table.increasePointer(t.getNumberOfItems());
    }
    System.out.println("Total number of transactions: "+num_trans);

    return output;
  }


  public KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_ASSOCIATION_MODEL;
    if (i == 1)
      return KDDMLObjectType.INTERNAL_TABLE;
    return null;
  }

  /**
   * Test if a transaction is supported by the set of association rules.
   *
   * @param transaction Transaction
   * @return boolean
   */
  protected boolean isSupportedByRda(TransactionManager transaction) {

    Iterator iter = rules.getAssociationRules();
    AssociationRuleManager r;
    while(iter.hasNext()) {
      r = (AssociationRuleManager) iter.next();
      ItemsetManager body = (ItemsetManager) r.getAntecedent();
      ItemsetManager head = (ItemsetManager) r.getConsequent();
      // scorre la transazione per controllare che tutti i suoi items appartengano alla regola
      Item i;
      Iterator iter2 = transaction.getItems();
      boolean supported = true;
      while ((iter2.hasNext()) && supported) {
        i = (Item) iter2.next();
        if (!body.contain(i) && !head.contain(i))
          // l'item non e' supportato dalla regola
          supported = false;
      }
      if (supported)
        return true;
    }
    return false;
  }

  /**
   * Test if a transaction is supported by the set of itemset.
   *
   * @param transaction Transaction
   * @return boolean
   */
  protected boolean isSupportedByItemset(TransactionManager transaction) {
    Iterator iter = rules.getItemsets();
    ItemsetManager r;
    while(iter.hasNext()) {
      r = (ItemsetManager) iter.next();
      // scorre la transazione per controllare che tutti i suoi items appartengano all'itemset
      Item i;
      Iterator iter2 = transaction.getItems();
      boolean supported = true;
      while ((iter2.hasNext()) && supported) {
        i = (Item) iter2.next();
        if (!r.contain(i))
          // l'item non e' supportato dall'itemset
          supported = false;
      }
      if (supported)
        return true;
    }
    return false;
  }

  /**
   * Rimuove tutti gli attributi numerici dalle istanze in ingresso.
   *
   * @param tuple Instances
   * @return Instances
   */
  private static Instances deleteNumericAttributes(Instances tuple) {

    Vector numeric_attr = new Vector();
    for (int i=0; i<tuple.numAttributes(); i++) {
      if (tuple.attribute(i).isNumeric())
        numeric_attr.add(tuple.attribute(i).name());
    }
    // il vettore contiene i nomi degli attributi numerici
    for (int j=0; j<numeric_attr.size(); j++) {
      System.out.println("Ignoring numeric attribute: "+(String) numeric_attr.get(j));
      Attribute attr = tuple.attribute((String) numeric_attr.get(j));
      tuple.deleteAttributeAt(attr.index());
    }
    return tuple;
  }
}

