/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.ProprietaryElement.Condition.*;
import kddml.Core.DataMining.*;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataSources.*;
import weka.core.*;
import java.util.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SEQUENCE_AGGREGATE_FILTER_RESOLVER extends PostProcessingResolver {

  private SequenceModelManager seq_model;

  private TimestampInternalTableManager table;

  //contains table schema
  private Instances meta_data;

  //loaded instances
  private Instances insts = null;
  //current instance
  private Instance inst = null;


  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get sequences...
    seq_model = (SequenceModelManager) arguments.get(0);

    // get input internal table...
    InternalTableManager table = (InternalTableManager) arguments.get(1);

    // get condition...
    Condition condition = (Condition) arguments.get(2);

    // controlla la compatibilita' del modello con le istanze...
    if (!seq_model.isCompatible((DataStatistics) table.getStatistic())) {
      String msg = "The input sequence model is not compatible with the data source!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INCOMPATIBLE_OBJECTS);
    }

    if (!table.isTimestampTable()) {
      throw new KDDMLOperatorException(
          "Error in SEQUENCE_AGGREGATE: the input table <"+seq_model.getModelName()+
          "> must be a timestamp table.", null,
          KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // get output destination
    java.io.File f = new java.io.File(kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL.getRepositoryPath() +
                                      settings.getXMLDestination());

    // numero di occorrenze minime affinche' la sequenza sia considerata frequente
    int min_occurrence =
        (int) (seq_model.getNumberOfTransactions().intValue() * seq_model.getMinimumSupport().doubleValue());

    //Hashtable for sequence support
    Hashtable sequences_support = new Hashtable();

    this.table = (TimestampInternalTable)table;

    // get the instances schema
    this.meta_data = (Instances) this.table.load(0);

    //Scandisce tutte le data-sequence della tabella
    while( hasNext() ){
      //get next data-sequence
      List dseq = next();
      if( dseq == null ) break;
      //calcola l'occorrenza della sequenza
      sequences_support = check_support(dseq, seq_model.getSequences(),
                                     sequences_support, condition);
    }

    TreeSet new_sequences = new TreeSet(new CompareSequences());
    Iterator iseq = seq_model.getSequences();
    while( iseq.hasNext() ){
      SequenceManager seq = (SequenceManager) iseq.next();
      //get new support for sequence
      Integer support = (Integer) sequences_support.get(seq.toString());
      //if sequence is still frequent
      if( support != null && support.intValue() >= min_occurrence ){
        //change support to sequence
        seq.setOccurrence(support);
        seq.setSupport( new Double((support.doubleValue()/seq_model.getNumberOfTransactions().doubleValue())*100.0) );
        //add new sequence
        new_sequences.add(seq);
      }
    }

    System.out.println("Sequences removed: "+
                       (seq_model.getNumberOfSequences()-new_sequences.size())+
                       " of "+seq_model.getNumberOfSequences());

    // create a new model
    SequenceFactory fact = new SequenceFactory(f,
                                               (DataDictionary)seq_model.getDataDictionary(),
                                               (MiningSchema)seq_model.getMiningSchema(),
                                               seq_model.getMinimumSupport().doubleValue(),
                                               seq_model.getMinimumConfidence().doubleValue(),
                                               seq_model.getNumberOfTransactions().intValue(),
                                               new_sequences);
    fact.setModelName( seq_model.getModelName()+"_aggregate" );
    fact.setAlgorithmName( seq_model.getAlgorithmName() );
    fact.setAvgNumberOfItemsPerTA( seq_model.getAvgNumberOfItemsPerTransaction().doubleValue() );
    fact.setMaxNumberOfItemsPerTA( seq_model.getMaxNumberOfItemsPerTransaction().intValue() );
    SequenceModel new_model = (SequenceModel)fact.newInstance();

    // return the new model
    return new_model;

  }

  /**
   * Aggiorna il support delle sequenze.
   */
  private Hashtable check_support(List dseq, Iterator seq_iter,
                               Hashtable seq_support,
                               Condition condition)
      throws KDDMLOperatorException, KDDMLCoreException {

    SequenceManager seq;

    while(seq_iter.hasNext()) {
      seq = (SequenceManager) seq_iter.next();

      //controlla se la data-sequence supporta la sequenza
      int[][] dseq_indexes = support(dseq, seq);
      boolean supported = (dseq_indexes != null);

      if( supported ){

        //build a set of instances containing all instances that support current sequence
        Instances dseq_instances = clone_schema(meta_data);//new Instances((Instances)dseq.get(0),-1);
        for( int dseq_cnt=0; dseq_cnt<dseq_indexes.length; dseq_cnt++ ){
          Instances trans_instances = (Instances) dseq.get(dseq_indexes[dseq_cnt][0]);
          for( int trans_cnt=1; trans_cnt<dseq_indexes[dseq_cnt].length; trans_cnt++ )
            dseq_instances.add(trans_instances.instance(dseq_indexes[dseq_cnt][trans_cnt]));
        }

        Instances ds_filtered = filter_schema(meta_data, table);
        //Build an Instance to be evalueted by Condition
        Instance inst_cond = new Instance(ds_filtered.numAttributes());
        inst_cond.setDataset(ds_filtered);
        for( int i=0; i<ds_filtered.numAttributes(); i++ ){
          String attr_name = ds_filtered.attribute(i).name();
          int attr_index = dseq_instances.attribute( attr_name ).index();
          double[] attr_values = dseq_instances.attributeToDoubleArray( attr_index );
          String str_attr_values = toString(attr_values);
          inst_cond.setValue( ds_filtered.attribute(i), str_attr_values);
        }
        //check aggregate constraint
        boolean constraint_ok = condition.evalCond(inst_cond, null);
        if( constraint_ok ) {
          Integer support = (Integer) seq_support.get(seq.toString());
          if( support == null ) seq_support.put(seq.toString(), new Integer(1) );
          else seq_support.put(seq.toString(), new Integer(support.intValue()+1) );
        }

      }

    }
    return seq_support;
  }

  /**
   * Se la data-sequence non supporta la sequenza ritorna null,
   * altrimenti ritorna un insieme di instanze.
   * Ogni istanza contiene i valori degli attributi di un item nella sequenza,
   * calcolati nella data-sequence.
   */
  protected int[][] support(List dseq,
                            SequenceManager sequence)
      throws KDDMLOperatorException, KDDMLCoreException {

    //se il numero di sequence element e' maggiore del numero di transazioni
    if (sequence.getNumberOfElements() > dseq.size())
      return null;

    //Per ogni transazione che supporta un elemento della sequenza,
    //viene costruito un array di double per contenere i valori, dell'attributo
    //specificato, che gli items hanno nella transazione in esame.
    //Il vettore contiene tali array per tutte quelle transazioni nella data-sequence
    //che supportano la sequenza in esame.

    Iterator elements_iterator = sequence.getElements();
    Iterator transactions_iterator = dseq.iterator();
    boolean transactions_ended = false;
    boolean element_supported = false;

        int transaction_index = 0;
        int[][] support_indexes = new int[sequence.getNumberOfElements()][];
        int index = 0;

    //scandisce tutti gli elementi della sequenza fino a quando:
    //1) ci sono altri elementi della sequenza da controllare, e
    //2) non sono state esaminate tutte le transazioni
    while (elements_iterator.hasNext() && !transactions_ended) {

      SequenceElementManager element = (SequenceElementManager) elements_iterator.next();
      if (elements_iterator.hasNext()) elements_iterator.next(); //skip delimiter

      element_supported = false;
      while (transactions_iterator.hasNext() && !element_supported) {

        Instances transaction = (Instances) transactions_iterator.next();
        transaction_index++;
        //se la transazione supporta l'elemento corrente, il metodo restituisce
        //i valori di "attr" per gli items in "elem", altrimenti restituisce null
        int[] indexes = support(transaction, element);
        element_supported = (indexes != null);
        if( element_supported ) {
          indexes[0] = transaction_index-1;
          support_indexes[index++] = indexes;
        }

      }
      if (!transactions_iterator.hasNext()) transactions_ended = true;

    }

    //Se il ciclo termina prima di scandire tutte le transazioni della
    //data-sequence, allora la data-sequence e' supportata dalla sequenza.
    if (!transactions_ended) {
      return support_indexes;
    }
    //Se il ciclo termina dopo aver scandito tutte le transazioni della
    //data-sequence, allora la data-sequence supporta la sequenza
    //solo se l'ultimo transazione supporta l'ultimo elemento della sequenza.
    if (!elements_iterator.hasNext() && element_supported){
      return support_indexes;
    }

    return null;
  }

  /**
   * Se la transazione non supporta l'elemento della sequenza ritorna null,
   * altrimenti ritorna un insieme di instanze.
   * Ogni istanza contiene i valori degli attributi di un item nell'elemento della
   * sequenza, calcolati nella transazione.
   */
  private int[] support(Instances trans_instances, SequenceElementManager elem)
      throws KDDMLOperatorException, KDDMLCoreException{

    int[] instances_indexes = new int[elem.getNumberOfItems()+1];
    int index = 1;

    Item[] items = new Item[elem.getNumberOfItems()];
    //scandisce tutti gli items nel sequence element
    Iterator e_iterator = elem.getItems();
    for(int i=0; e_iterator.hasNext();i++ )
      items[i] = (Item) e_iterator.next();

      //scandisce le istanze della transazione
    for( int i=0; i<trans_instances.numInstances(); i++ ){
      //Carica l'i-esima instanza della transazione
      Instance inst = trans_instances.instance(i);
      //Prende l'item dall'istanza della transazione
      String item = inst.stringValue( trans_instances.attribute(TimestampInternalTable.EVENT_ATTRIBUTE));

      //Se l'item della transazione e' uguale all'item corrente
      for( int j=0; j<items.length; j++)
        if( items[j].equals( new Item(item) ) ){
          //aggiunge l'istanza
          instances_indexes[index++] = i;
          //interrompe la scansione della transazione
          break;
        }

    }

    if( (index) == instances_indexes.length )
      return instances_indexes;

    return null;
  }

  private boolean hasNext(){
    return (insts != null && insts.numInstances() > 0) || table.hasNext();
  }

  private List next() throws KDDMLCoreException{

    if ( !hasNext() ) return null;

    //Instances data = new Instances(meta_data, -1);
    List d_sequence = new ArrayList();

    if (inst == null){
      if( insts == null || insts.numInstances() == 0 )
        insts = ( (Instances) table.load(table.optimize()));
      inst = insts.instance(0);
      insts.delete(0);
    }

    String current_sequence_id = inst.stringValue( table.getTransactionPosition() );
    String first_sequence_id = current_sequence_id;

    while ( (current_sequence_id.equals(first_sequence_id)) &&
            (inst != null)) {

      Instances data = clone_schema(meta_data);

      String current_timestamp = inst.toString( table.getTimestampPosition() );
      String first_timestamp = current_timestamp;

      while( (current_sequence_id.equals(first_sequence_id)) &&
             (current_timestamp.equals(first_timestamp)) &&
             (inst != null) ) {

        Instance new_inst = new Instance(inst.numAttributes());
        new_inst.setDataset(data);
        for (int c=0; c<inst.numAttributes(); c++) {
          if( inst.attribute(c).isNumeric() )
            new_inst.setValue(c, inst.value(c));
          else
            new_inst.setValue(c, inst.stringValue(c));
        }
        data.add(new_inst);

        if( (insts == null || insts.numInstances() == 0) && table.hasNext() )
          insts = (Instances) table.load(table.optimize());
        if( insts == null || insts.numInstances() == 0 )
          inst = null;
        else{
          inst = insts.instance(0);
          insts.delete(0);
        }

        if (inst != null){
          current_sequence_id = inst.stringValue( table.getTransactionPosition() );
          current_timestamp = inst.toString( table.getTimestampPosition() );
        }

      }

      d_sequence.add(data);

    }

    return d_sequence;

  }

  private static Instances clone_schema(Instances schema){
    FastVector fv = new FastVector();
    for (int i=0; i<schema.numAttributes(); i++ )
      if( schema.attribute(i).isNumeric() )
        fv.addElement( new Attribute(schema.attribute(i).name()) );
      else
        fv.addElement( new Attribute(schema.attribute(i).name(), (FastVector)null) );
    return new Instances("clone", fv, 0);
  }

  private static Instances filter_schema(Instances schema, TimestampInternalTableManager table){
    FastVector fv = new FastVector();
    for( int i=0; i<schema.numAttributes(); i++ )
      if( schema.attribute(i).isNumeric() &&
          schema.attribute(i).index() != table.getTransactionPosition() &&
          schema.attribute(i).index() != table.getTimestampPosition() &&
          schema.attribute(i).index() != table.getEventPosition() )
        fv.addElement( new Attribute(schema.attribute(i).name(), (FastVector)null) );
    return new Instances("filter", fv, 0);
  }

  private String toString(double[] array){
    String str = new String();
    for (int j = 0; j < array.length-1; j++)
      if( Double.isNaN(array[j]) )
        str += Double.toString(0) + ",";
      else
        str += Double.toString(array[j]) + ",";
    if( Double.isNaN(array[array.length-1]) )
      str += Double.toString(0);
    else
      str += Double.toString(array[array.length-1]);
    return str;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_SEQUENCE_MODEL;
    if (i == 1)
      return KDDMLObjectType.INTERNAL_TABLE;
    if (i == 2)
      return KDDMLObjectType.CONDITION;
    return null;
  }

  public static void main(String[] args) {
    try{
      java.io.File dest = new java.io.File("./wlog.xml");
      String file = "repository/data/arff/web_log.arff";
      ARFFFactory fact = new ARFFFactory(dest, file);
      fact.setOverwrite(true);
      TimestampInternalTable table = (TimestampInternalTable) fact.newInstance();
      table.saveToRepository();

      java.io.File f = new java.io.File(kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL.getRepositoryPath() + "web_log_mined.xml");
      SequenceRepositoryFactory sfact = new SequenceRepositoryFactory(f);
      SequenceModel model = (SequenceModel) sfact.newInstance();

      XMLUtility.initializeDocument("CONDITION");
      org.w3c.dom.Element and = XMLUtility.doc.createElement("AND_COND");
      org.w3c.dom.Element base1 = XMLUtility.doc.createElement("BASE_COND");
      org.w3c.dom.Element base2 = XMLUtility.doc.createElement("BASE_COND");

      base1.setAttribute("op_type", "greater_or_equal");
      base1.setAttribute("term1", "@content-length");
      base1.setAttribute("term2", "1024");
      base1.setAttribute("term3", "sum");

      base2.setAttribute("op_type", "greater");
      base2.setAttribute("term1", "@content-length");
      base2.setAttribute("term2", "0");
      base2.setAttribute("term3", "min");

      and.appendChild(base1);
      and.appendChild(base2);

      ConditionFactory cfact = new ConditionFactory(and,
                                                    "SEQUENCE_AGGREGATE_FILTER_CONDITION");
      Condition cond = (Condition) cfact.newInstance();

      Vector oargs = new Vector();
      oargs.add(model);
      oargs.add(table);
      oargs.add(cond);

      SEQUENCE_AGGREGATE_FILTER_SETTINGS sett = new
          SEQUENCE_AGGREGATE_FILTER_SETTINGS();

      Hashtable htable = new Hashtable();
      //htable.put("xml_dest", "result.xml");
      sett.setXMLDestination("result.xml");
      //sett.applySettings(htable);

      new SEQUENCE_AGGREGATE_FILTER_RESOLVER().execute(sett, oargs).
          saveToRepository();
    }
    catch(KDDMLCoreException e){ System.out.println(e.getErrorMessage()); e.printStackTrace();}
    catch(KDDMLOperatorException e){ System.out.println(e.getErrorMessage()); e.printStackTrace();}
    catch(Exception e){ System.out.println(e.getMessage()); e.printStackTrace();}
  }
}
