/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataMining.*;
import java.util.*;

/**
 * Given a set of sequential patterns, this operator returns all corresponding
 * sequential rules.
 * Note that this operator need the complete set of sequential
 * patterns to build the rules.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SEQUENCE_RULE_RESOLVER  extends PostProcessingResolver {

  SequenceModelManager seq_model;

  public SEQUENCE_RULE_RESOLVER() {
    super();
  }

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get sequences...
    seq_model = (SequenceModelManager) arguments.get(0);

    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL.getRepositoryPath() +
                                         settings.getXMLDestination());

    // get min_confidence
    Double min_confidence = ((SEQUENCE_RULE_SETTINGS)settings).getMinConfidence();

    // get max number of rules
    Integer max_rules = ((SEQUENCE_RULE_SETTINGS)settings).getMaxNumberOfRules();
    if( max_rules == null ) max_rules = new Integer(SequenceModel.MAX_NUMBER_OF_RULES);

    // get sequences from model
    TreeSet sequences = new TreeSet( new CompareSequences() );
    Iterator iterator = seq_model.getSequences();
    while( iterator.hasNext() )
      sequences.add( (SequenceManager)iterator.next() );

    // build rules from sequences
    TreeSet rules = compute(sequences,
                            min_confidence.doubleValue(),
                            seq_model.getNumberOfTransactions().intValue(),
                            max_rules.intValue());

    if( rules.isEmpty() )
      throw new KDDMLOperatorException(
        "No rules generated.",
        "Possible causes: missing (sub)sequence(s)/missing sequence(s) support",
        KDDMLOperatorErrorCodes.INVALID_SETTINGS);

    // build model
    DataDictionary dd = (DataDictionary)seq_model.getDataDictionary();
    MiningSchema ms = (MiningSchema)seq_model.getMiningSchema();
    RulesSequenceFactory rule_factory = new RulesSequenceFactory(
      dest, dd, ms, seq_model.getMinimumSupport().doubleValue(),
      min_confidence.doubleValue(), seq_model.getNumberOfTransactions().intValue(), sequences, rules);

    rule_factory.setModelName(seq_model.getModelName()+"_rules");
    rule_factory.setAlgorithmName(seq_model.getAlgorithmName());
    rule_factory.setAvgNumberOfItemsPerTA(seq_model.getAvgNumberOfItemsPerTransaction().doubleValue());
    rule_factory.setMaxNumberOfItemsPerTA(seq_model.getMaxNumberOfItemsPerTransaction().intValue());

    SequenceModel rules_model = (SequenceModel) rule_factory.newInstance();

    return rules_model;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_SEQUENCE_MODEL;
    return null;
  }

  private static TreeSet compute(TreeSet sequences, double min_conf, int num_transactions, int max_rules)
      throws SequenceModelException {

    Comparator comp = new CompareSequenceRules();
    TreeSet rules = new TreeSet(comp);
    // costruisce la tabella hash contenete i supporti delle sequenze...
    Hashtable support_table = buildHashtable(sequences, num_transactions);

    Iterator iterator = sequences.iterator();
    int i = 0;
    while(iterator.hasNext()) {

      if ((i%500) == 0)
        System.out.println("Sequences parsed: "+i+" of "+sequences.size());
      i++;

      SequenceManager sequence = (SequenceManager) iterator.next();

      /*
       From a Sequence s = <set1,...,setN> build correspondig rules as follows:
       for (I=1; I< s.length; I++){
          rule = <set1,...,setI> -> <set(I+1),...,setN>
          rule.support = s.support
          rule.confidence = s.confidence / <set1,...,setI>.confidence
       }
       */
      for (int j = 1; j < sequence.getNumberOfElements(); j++) {
        try{

          SequenceManager ant = sequence.subSequence(0, j);
          Double ant_supp = (Double) support_table.get(ant);
          ant.setSupport(ant_supp);

          if (ant_supp == null)
            throw new Exception(
                "[WARNING] SEQUENCE_RULE_RESOLVER:: Antecedent sub-sequence <" +
                ant + "> support is null; can't calculate rules.");

          DelimiterManager d = sequence.getPrecedingDelimiter(j);

          SequenceManager cons = sequence.subSequence(j ,
              sequence.getNumberOfElements());
          Double cons_supp = (Double) support_table.get(cons);
          cons.setSupport(cons_supp);

          if (cons_supp == null)
            throw new Exception(
                "[WARNING] SEQUENCE_RULE_RESOLVER:: Consequent sub-sequence <" +
                cons + "> support is null; can't calculate rules.");

          Double seq_support = sequence.getSupport();
          Integer seq_occurrence = sequence.getOccurrence();

          //se il supporto della sequenza e' nullo
          if (seq_support == null)

            //calcolo il supporto con l'occorrenza -- non nulla altrimenti buildHashtable() avrebbe dato errore
            seq_support = new Double( (double) ( (double) seq_occurrence.
                                                intValue() /
                                                (double) num_transactions));
            //se l'occorrenza della sequenza e' nulla
          else if (seq_occurrence == null)

            //calcolo l'occorrenza con il supporto
            seq_occurrence = new Integer( (int) (seq_support.doubleValue() *
                                                 (double) num_transactions));

          double confidence = sequence.getSupport().doubleValue() /
              ant.getSupport().doubleValue();

          SequenceRuleManager rule = new SequenceRule(ant, d, cons, confidence,
              seq_support.doubleValue(), seq_occurrence.intValue());

          if (confidence >= min_conf) {
            rules.add(rule);
            // controlla il numero di regole prodotte
            //sistemare rischio di perdere regole con confidenza piu' alta...
            if (rules.size() > max_rules)
              rules.remove( (SequenceRule) rules.last());
          }

        }catch(Exception e){ System.out.println(e.getMessage()); }
      }


    }

    support_table = null;
    iterator = null;
    System.gc();

    return rules;
  }


  /**
   * Costruisce la tabella hash per leggere i supporti associati alle sequenze.
   * Come chiave viene utilizzata la sequenza.
   * Se il supporto di una sequenza e' nullo, si prova a calcolare il supporto
   * utilizzando l'occorenza ed il numero di transazioni.
   * Se sia il supporto, sia l'occorrenza di una sequenza sono nulli,
   * l'operatore da errore.
   */
  private static Hashtable buildHashtable(TreeSet sequences, int num_transactions) throws SequenceModelException {
    Iterator iterator = sequences.iterator();
    Hashtable support_table = new Hashtable();
    while (iterator.hasNext()) {

      Sequence sequence = (Sequence) iterator.next();
      if (sequence.getSupport() == null) {
        if(sequence.getOccurrence() != null ){
          Double support = new Double(
                  (double)((double)sequence.getOccurrence().intValue() / (double)num_transactions) );
          support_table.put(sequence, support);
        }
        else{
          String msg = "[WARNING] SEQUENCE_RULE_RESOLVER:: sequence <" +
              sequence + "> support is null; can't calculate rules.";
          System.out.println(msg);
        }
      }
      else{
        support_table.put(sequence, sequence.getSupport());
      }

    }

    return support_table;
  }

}
