/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.AssociationRules.Item;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataSources.*;
import weka.core.*;
import java.util.Iterator;

/**
 * Given a set of sequential patterns and a table, this operator returns
 * those data-sequences that supports at least one pattern.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SEQUENCE_SATISFY_RESOLVER extends PostProcessingResolver {

  SequenceModelManager seq_model;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get sequences...
    seq_model = (SequenceModelManager) arguments.get(0);

    // get input internal table...
    InternalTableManager table = (InternalTableManager) arguments.get(1);

    // controlla la compatibilita' del modello con le istanze...
    if (!seq_model.isCompatible((DataStatistics) table.getStatistic())) {
      String msg = "The input sequence model is not compatible with the data source!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INCOMPATIBLE_OBJECTS);
    }


    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_TABLE.getRepositoryPath() +
                                         settings.getXMLDestination());

    if (!table.isTimestampTable()) {
      throw new KDDMLOperatorException(
          "Error in SEQUENCE_STATISFY: the input table <"+seq_model.getModelName()+
          "> must be a timestamp table.", null,
          KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
    else {
      System.out.println("Timestamp Table.");
      return executeTimestamp((TimestampInternalTable) table, dest);
    }

  }

  /**
   * Execute the operator given a timestamp internal table.
   *
   * @param settings KDDMLOperatorSettings
   * @param table TimestampInternalTable
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  private kddml.Core.KDDMLObject executeTimestamp(TimestampInternalTable table,
                                                     java.io.File output_dest)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get the instances schema
    Instances tuple = (Instances) table.load(0);
    WekaFactory fact = new WekaFactory(output_dest, tuple);
    // set overwrite to be true
    fact.setOverwrite(true);
    // create an internal table from input internal table with no instances
    InternalTable output = (InternalTable) fact.newInstance();

    DataSequenceManager dseq;
    boolean supported;
    int num_trans = 0;
    // scorre le singole transazioni
    while(table.hasNextDataSequence()) {
      dseq = (DataSequenceManager) table.loadNextDataSequence();
      num_trans++;
      supported = supportSequences(dseq);

      int num_instances = dseq.getNumberOfItems();

      if (supported) {
        // transazione supportata: ottiene le istanze dalla tabella in input
        tuple = (Instances) table.load(num_instances);
        // e le aggiunge alla tabella in output
        output.addInstances(tuple);
      }
      else
        // transazione non supportata: si scorre il puntatore
        table.increasePointer(num_instances);
    }

    System.out.println("Total number of transactions: "+num_trans);

    return output;
  }

  /**
   * Test if a data-sequence support at least one sequence in the model.
   *
   * @param dseq DataSequenceManager
   * @return boolean
   */
  protected boolean supportSequences(DataSequenceManager dseq) {
    Iterator iterator = seq_model.getSequences();
    while(iterator.hasNext()) {
      SequenceManager seq = (SequenceManager) iterator.next();
      if( seq.isSupported(dseq) ) return true;
    }
    return false;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_SEQUENCE_MODEL;
    if (i == 1)
      return KDDMLObjectType.INTERNAL_TABLE;
    return null;
  }
}
