/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.*;
import kddml.Core.DataMining.SequentialPatterns.*;
import kddml.Core.DataSources.*;
import weka.core.*;
import java.util.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SEQUENCE_TIMESTAMP_FILTER_RESOLVER extends PostProcessingResolver {

  SequenceModelManager seq_model;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get sequences...
    seq_model = (SequenceModelManager) arguments.get(0);

    // get input internal table...
    InternalTableManager table = (InternalTableManager) arguments.get(1);

    // controlla la compatibilita' del modello con le istanze...
    if (!seq_model.isCompatible((DataStatisticsManager) table.getStatistic())) {
      String msg = "The input sequence model is not compatible with the data source!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INCOMPATIBLE_OBJECTS);
    }

    // controlla la tabella in input...
    if (!table.isTimestampTable()) {
      throw new KDDMLOperatorException(
          "Error in SEQUENCE_TIMESTAMP_CONSTRAINT: the input table <"+seq_model.getModelName()+
          "> must be a timestamp table.", null,
          KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }

    // get output destination
    java.io.File f = new java.io.File(kddml.Core.KDDMLObjectType.PMML_SEQUENCE_MODEL.getRepositoryPath() +
                                      settings.getXMLDestination());

    // get constr_type
    String constr_type = ((SEQUENCE_TIMESTAMP_FILTER_SETTINGS)settings).getConstrType();

    // get interval_closure
    String interval_closure = ((SEQUENCE_TIMESTAMP_FILTER_SETTINGS)settings).getIntervalClosure();

    // get lower_bound
    Integer lower_bound = ((SEQUENCE_TIMESTAMP_FILTER_SETTINGS)settings).getLowerBound();

    // get upper_bound
    Integer upper_bound = ((SEQUENCE_TIMESTAMP_FILTER_SETTINGS)settings).getUpperBound();

    // get the instances schema
    Instances tuple = (Instances) table.load(0);

    // se l'attributo timestamp non e' di tipo numerico viene generata un'eccezione
    Attribute attr = tuple.attribute( ((TimestampInternalTableManager)table).getTimestampPosition() );
    if( !attr.isNumeric() )
      throw new KDDMLOperatorException(
        "Input table does not contains valid timestamp attribute",
        "Timestamp attribute must be a positive integer",
        kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);

    if( lower_bound == null && upper_bound == null ){
      System.out.println
          ("[WARNING] SEQUENCE_TIMESTAMP_CONSTRAINT: Operator execution skipped, " +
           "due to constraint specification (constraint always satisfied).");

      // clone this model...
      SequenceModel sequence_model = (SequenceModel) seq_model.clone(f);
      String model_name = seq_model.getModelName();
      sequence_model.setModelName(model_name+"_timestamp");

      return sequence_model;
    }

    // numero di occorrenze minime affinche' la sequenza sia considerata frequente
    int min_occurrence =
        (int) (seq_model.getNumberOfTransactions().intValue() * seq_model.getMinimumSupport().doubleValue());


    //Hashtable for sequence support
    Hashtable sequences_support = new Hashtable();

    DataSequenceManager dseq = null;

    //Scandisce tutte le data-sequence della tabella
    while( ((TimestampInternalTableManager)table).hasNextDataSequence() ){
      dseq = ((TimestampInternalTableManager)table).loadNextDataSequence();
      if( dseq == null ) break;

      //calcola l'occorrenza della sequenza
      sequences_support = occurrence(dseq,
                                     seq_model.getSequences(),
                                     sequences_support,
                                     constr_type,
                                     interval_closure,
                                     lower_bound,
                                     upper_bound);
    }

    TreeSet new_sequences = new TreeSet(new CompareSequences());
    SequenceManager seq;
    Iterator iseq = seq_model.getSequences();
    while( iseq.hasNext() ){
      seq = (SequenceManager) iseq.next();
      Integer support = (Integer) sequences_support.get(seq.toString());
      if( support != null && support.intValue() >= min_occurrence ){
        seq.setOccurrence(support);
        seq.setSupport( new Double((support.doubleValue()/seq_model.getNumberOfTransactions().doubleValue())*100.0) );
        new_sequences.add(seq);
      }
    }

    System.out.println("Sequences removed: "+
                       (seq_model.getNumberOfSequences()-new_sequences.size())+
                       " of "+seq_model.getNumberOfSequences());

    // create a new model
    SequenceFactory fact = new SequenceFactory(f,
                                               (DataDictionary)seq_model.getDataDictionary(),
                                               (MiningSchema)seq_model.getMiningSchema(),
                                               seq_model.getMinimumSupport().doubleValue(),
                                               seq_model.getMinimumConfidence().doubleValue(),
                                               seq_model.getNumberOfTransactions().intValue(),
                                               new_sequences);
    fact.setModelName( seq_model.getModelName()+"_timestamp" );
    fact.setAlgorithmName( seq_model.getAlgorithmName() );
    fact.setAvgNumberOfItemsPerTA( seq_model.getAvgNumberOfItemsPerTransaction().doubleValue() );
    fact.setMaxNumberOfItemsPerTA( seq_model.getMaxNumberOfItemsPerTransaction().intValue() );
    SequenceModel new_model = (SequenceModel)fact.newInstance();

    // return the new model
    return new_model;

  }

  /**
   * Conta le occorrenze della sequenza nella tabella in input.
   *
   */
  private Hashtable occurrence(DataSequenceManager dseq,
                               Iterator i_sequences,
                               Hashtable seq_support,
                               String constr,
                               String interval,
                               Integer lower,
                               Integer upper)
      throws KDDMLOperatorException, KDDMLCoreException {

    SequenceManager sequence;
    // scorre le sequenze
    while(i_sequences.hasNext()) {
      sequence = (SequenceManager) i_sequences.next();
      boolean constraint_ok = false;

      //execute duration constraint
      if( constr.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.CONSTR_TYPE_DURATION) ){
        // scandisce tutte le data-sequenze per contare le occorrenze della sequenza in input
        constraint_ok = support_duration(dseq, sequence, interval, lower, upper);
      }
      //execute gap constraint
      else if( constr.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.CONSTR_TYPE_GAP) ){
        // scandisce tutte le data-sequenze per contare le occorrenze della sequenza in input
        constraint_ok = support_gap(dseq, sequence, interval, lower, upper);
      }
      //check timestamp constraint

      if( constraint_ok ) {
        Integer support = (Integer) seq_support.get(sequence.toString());
        if( support == null ) seq_support.put(sequence.toString(), new Integer(1) );
        else seq_support.put(sequence.toString(), new Integer(support.intValue()+1) );
      }

    }

    return seq_support;
  }

  /**
   * Controlla se una data-sequence supporta una sequenza, rispettando il
   * "vincolo di durata".
   * Data una data-sequence d, una sequenza s, un operatore op ed una soglia limit,
   * d supporta s, se e solo se:
   * 1) s e' contenuta in d, e
   * 2) d[s(s.length)].time - d[s(1)].time op limit
   * s(i) ritorna l'indice della transazione in d che supporta l'i-esimo elemento
   * della sequenza s.
   *
   * @param dseq DataSequenceManager
   * @param sequence SequenceManager
   * @param op String
   * @param value Integer
   * @param op2 String
   * @param value2 Integer
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return boolean
   */
  private boolean support_duration(DataSequenceManager dseq,
                                   SequenceManager sequence,
                                   String interval,
                                   Integer lower,
                                   Integer upper)
      throws KDDMLOperatorException, KDDMLCoreException {

    //se il numero di elementi nella sequenza e' maggiore del numero di transazioni
    if (sequence.getNumberOfElements() > dseq.getNumberOfTransactions())
      //la sequenza non e' supportata
      return false;

    Iterator elements_iterator = sequence.getElements();
    Iterator transactions_iterator = dseq.getTransactions();
    boolean transactions_ended = false;
    boolean element_supported = false;

    TimestampedTransactionManager transaction = null;

    //tempo della prima transazione che supporta il primo elemento della sequenza
    String start_time = null;
    //tempo dell'ultimo transazione che supporta l'ultimo elemento della sequenza
    String end_time = null;

    //scandisce tutti gli elementi della sequenza fino a quando:
    //1) ci sono altri elementi della sequenza da controllare, e
    //2) non sono state esaminate tutte le transazioni
    while (elements_iterator.hasNext() && !transactions_ended) {

      SequenceElementManager element = (SequenceElementManager) elements_iterator.next();
      if (elements_iterator.hasNext()) elements_iterator.next(); //skip delimiter

      element_supported = false;
      while (transactions_iterator.hasNext() && !element_supported) {

        //transazione corrente
        transaction = (TimestampedTransactionManager) transactions_iterator.next();

        //se la transazione supporta l'elemento corrente,
        element_supported = element.isSupported(transaction);
        if( element_supported )
          //e l'elemento e' il primo
          if( start_time == null )
            //viene registrato il tempo della prima transazione
            start_time = transaction.getTimestamp();

      }

      if (!transactions_iterator.hasNext())
        transactions_ended = true;

    }

    //Se il ciclo termina prima di scandire tutte le transazioni della
    //data-sequence, allora la data-sequence non supportata la sequenza.
    if (!transactions_ended) {
      //registro il tempo dell'ultima transazione esaminata
      end_time = transaction.getTimestamp();
      //se la sequenza ha un solo elemento il vincolo e' soddisfatto
      if( sequence.getNumberOfElements() == 1 ) return true;
      //la sequenza e' supportata solo se viene soddisfatto il vincolo di durata
      return satisfy(start_time, end_time, interval, lower, upper);
    }
    //Se il ciclo termina dopo aver scandito tutte le transazioni della
    //data-sequence, allora la data-sequence supporta la sequenza
    //solo se l'ultima transazione supporta l'ultimo elemento della sequenza.
    if (!elements_iterator.hasNext() && element_supported){
      //registro il tempo dell'ultima transazione esaminata
      end_time = transaction.getTimestamp();
      //se la sequenza ha un solo elemento il vincolo e' soddisfatto
      if( sequence.getNumberOfElements() == 1 ) return true;
      //la sequenza e' supportata solo se viene soddisfatto il vincolo di durata
      return satisfy(start_time, end_time, interval, lower, upper);
    }

    //sequenza non supportata
    return false;
  }

  /**
   * Controlla se una data-sequence supporta una sequenza, rispettando il
   * "vincolo di intervallo".
   * Data una data-sequence d, una sequenza s, un operatore op ed una soglia limit,
   * d supporta s, se e solo se:
   * 1) s e' contenuta in d, e
   * 2) per ogni i, d[s(i+1)].time - d[s(i)].time op limit
   * s(i) ritorna l'indice della transazione in d che supporta l'i-esimo elemento
   * della sequenza s.
   *
   * @param dseq DataSequenceManager
   * @param sequence SequenceManager
   * @param op String
   * @param value Integer
   * @param op2 String
   * @param value2 Integer
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return boolean
   */
  private boolean support_gap(DataSequenceManager dseq,
                              SequenceManager sequence,
                              String interval,
                              Integer lower,
                              Integer upper)
      throws KDDMLOperatorException, KDDMLCoreException {

    //se il numero di elementi nella sequenza e' maggiore del numero di transazioni
    if (sequence.getNumberOfElements() > dseq.getNumberOfTransactions())
      //la sequenza non e' supportata
      return false;

    Iterator elements_iterator = sequence.getElements();
    Iterator transactions_iterator = dseq.getTransactions();
    boolean transactions_ended = false;
    boolean element_supported = false;

    TimestampedTransactionManager prev_transaction = null;
    TimestampedTransactionManager curr_transaction = null;

    //scandisce tutti gli elementi della sequenza fino a quando:
    //1) ci sono altri elementi della sequenza da controllare, e
    //2) non sono state esaminate tutte le transazioni
    while (elements_iterator.hasNext() && !transactions_ended) {

      SequenceElementManager element = (SequenceElementManager) elements_iterator.next();
      if (elements_iterator.hasNext()) elements_iterator.next(); //skip delimiter

      element_supported = false;
      while (transactions_iterator.hasNext() && !element_supported) {

        //transazione corrente
        curr_transaction = (TimestampedTransactionManager) transactions_iterator.next();

        //se la transazione supporta l'elemento corrente,
        element_supported = element.isSupported(curr_transaction);
        if( element_supported ){
          //e l'elemento non e' il primo
          if( prev_transaction != null ){
            //la sequenza e' supportata solo se viene soddisfatto il vincolo di intervallo
            if( !satisfy(prev_transaction.getTimestamp(), curr_transaction.getTimestamp(), interval, lower, upper) )
              return false;
          }
          //aggiorna la transazione precedente
          prev_transaction = curr_transaction;
        }

      }

      if (!transactions_iterator.hasNext())
        transactions_ended = true;

    }

    //Se il ciclo termina prima di scandire tutte le transazioni della
    //data-sequence, allora la data-sequence non supporta la sequenza.
    if (!transactions_ended)
      return true;
    //Se il ciclo termina dopo aver scandito tutte le transazioni della
    //data-sequence, allora la data-sequence supporta la sequenza
    //solo se l'ultima transazione supporta l'ultimo elemento della sequenza.
    if (!elements_iterator.hasNext() && element_supported)
      return true;

    //sequenza non supportata
    return false;
  }

  /**
   * Controllo del vincolo.
   *
   * @param time1 String
   * @param time2 String
   * @param op String
   * @param value Integer
   * @param op2 String
   * @param value2 Integer
   * @return boolean
   */
  private boolean satisfy(String time1, String time2, String interval_closure, Integer lower, Integer upper){
    int start = Double.valueOf(time1).intValue();
    int end = Double.valueOf(time2).intValue();
    int time = end - start;

    boolean first = true;
    boolean second = true;

    //closed_closed
    if( interval_closure.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.CLOSED_CLOSED) ){
      if( lower != null ) first = lower.intValue() < time;
      if( upper != null ) second = time < upper.intValue();
      return (first && second);
    }
    //open_open
    else if( interval_closure.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.OPEN_OPEN) ){
      if( lower != null ) first = lower.intValue() <= time;
      if( upper != null ) second = time <= upper.intValue();
      return (first && second);
    }
    //close_open
    else if( interval_closure.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.CLOSED_OPEN) ){
      if( lower != null ) first = lower.intValue() < time;
      if( upper != null ) second = time <= upper.intValue();
      return (first && second);
    }
    //open_closed
    else if( interval_closure.equals(SEQUENCE_TIMESTAMP_FILTER_SETTINGS.OPEN_CLOSED) ){
      if( lower != null ) first = lower.intValue() <= time;
      if( upper != null ) second = time < upper.intValue();
      return (first && second);
    }

   return false;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i == 0)
      return KDDMLObjectType.PMML_SEQUENCE_MODEL;
    if (i == 1)
      return KDDMLObjectType.INTERNAL_TABLE;
    return null;
  }
}
