/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */
package kddml.Operators.Postprocessing;

import java.util.Hashtable;
import kddml.Operators.*;
import kddml.Core.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Daniele Cerra (daniele@cerra.info)
 * @version 2.0.0 alfa
 */
public class SEQUENCE_TIMESTAMP_FILTER_SETTINGS extends PostProcessingSettings {

  private final static String CONSTR_TYPE_ATTRIBUTE = "constr_type";
  private final static String INTERVAL_CLOSURE_ATTRIBUTE = "interval_closure";
  private final static String LOWER_BOUND_ATTRIBUTE = "lower_bound";
  private final static String UPPER_BOUND_ATTRIBUTE = "upper_bound";

  //CONSTR_TYPE legal values (duretion|gap)
  protected final static String CONSTR_TYPE_DURATION = "duration";
  protected final static String CONSTR_TYPE_GAP = "gap";

  //INTERVAL_CLOSURE legal values (open_closed|closed_closed|open_open|closed_open)
  protected final static String OPEN_CLOSED = "open_closed";
  protected final static String CLOSED_CLOSED = "closed_closed";
  protected final static String OPEN_OPEN = "open_open";
  protected final static String CLOSED_OPEN = "closed_open";

  protected String constr_type;
  protected String interval_closure;
  protected Integer lower_bound;
  protected Integer upper_bound;

  public SEQUENCE_TIMESTAMP_FILTER_SETTINGS() {
    super();
  }

  /**
   * Apply the input settings to this class.
   *
   * @throws KDDMLOperatorException
   */
  public void applySettings(java.util.Hashtable attributes)
      throws kddml.Operators.KDDMLOperatorException {

    super.applySettings(attributes);
    setConstrType(attributes, CONSTR_TYPE_ATTRIBUTE, true);
    setIntervalClosure(attributes, INTERVAL_CLOSURE_ATTRIBUTE, true);
    setLowerBound(attributes, LOWER_BOUND_ATTRIBUTE, false);
    setUpperBound(attributes, UPPER_BOUND_ATTRIBUTE, false);

    if( lower_bound != null && upper_bound != null ){

      //open_open
      if( interval_closure.equals(OPEN_OPEN) ){
        // upper >= lower
        if (upper_bound.intValue() < lower_bound.intValue())
          throw new KDDMLOperatorException
              ("< " + UPPER_BOUND_ATTRIBUTE + " > must be greater or equal than < " + LOWER_BOUND_ATTRIBUTE + " >",
               null,
               kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
      }
      //open_closed, closed_open, closed_closed
      else{
        // upper > lower
        if (upper_bound.intValue() <= lower_bound.intValue())
          throw new KDDMLOperatorException
              ("< " + UPPER_BOUND_ATTRIBUTE + " > must be greater than < " + LOWER_BOUND_ATTRIBUTE + " >",
               null,
               kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
      }

    }

  }

  protected void setConstrType(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        constr_type = (String) s.getExprValue();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a string constant.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if( !(constr_type.equals(CONSTR_TYPE_DURATION) || constr_type.equals(CONSTR_TYPE_GAP)) )
        throw new KDDMLOperatorException(
          "< "+attribute_name+" > only support one of this type of constraints:  "+
          CONSTR_TYPE_DURATION+", "+CONSTR_TYPE_GAP,
          null,
          kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }
  }

  protected void setIntervalClosure(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        interval_closure = (String) s.getExprValue();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be string constant.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if( !( interval_closure.equals(OPEN_CLOSED) ||
             interval_closure.equals(CLOSED_CLOSED) ||
             interval_closure.equals(OPEN_OPEN) ||
             interval_closure.equals(CLOSED_OPEN)
             ) )
        throw new KDDMLOperatorException(
          "< "+attribute_name+" > only support one of this interval:  "+
          OPEN_CLOSED+", "+CLOSED_CLOSED+", "+OPEN_OPEN+", "+CLOSED_OPEN,
          null,
          kddml.Core.KDDMLCoreErrorCodes.INVALID_SETTINGS);
    }
  }

  protected void setLowerBound(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        lower_bound = new Integer( ((Double) s.getExprValue()).intValue() );
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (lower_bound.intValue() <= 0)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  protected void setUpperBound(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {
    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        upper_bound = new Integer( ((Double) s.getExprValue()).intValue() );
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
      if (upper_bound.intValue() <= 0)
        throw new KDDMLOperatorException
            ("< "+attribute_name+" >"+" must be a positive integer value.",
             null,
             KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  public String getConstrType(){
    return constr_type;
  }

  public void setConstrType(String constr){
    constr_type = constr;
  }

  public String getIntervalClosure(){
    return interval_closure;
  }

  public void setIntervalClosure(String interval){
    interval_closure = interval;
  }

  public Integer getLowerBound() {
    return lower_bound;
  }

  public void setLowerBound(int i) {
    lower_bound = new Integer(i);
  }

  public Integer getUpperBound() {
    return upper_bound;
  }

  public void setUpperBound(int i) {
    upper_bound = new Integer(i);
  }

  public String[] getSupportedParameters() {
    String[] s = {
        CONSTR_TYPE_ATTRIBUTE, INTERVAL_CLOSURE_ATTRIBUTE,
        LOWER_BOUND_ATTRIBUTE, UPPER_BOUND_ATTRIBUTE
    };
    return s;
  }

}
