/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Postprocessing;

import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataMining.ClassificationTrees.*;
import kddml.Core.DataMining.*;

public class TREE_META_CLASSIFIER_RESOLVER extends PostProcessingResolver {

  /**
   * List of classification trees.
   */
  private TreeModelManager[] trees;

  /**
   * Combination Type: and, or, committee.
   */
  private byte combination_type;

  /**
   * The positive class used for and/or combination.
   */
  private String positive_class;

  public kddml.Core.KDDMLObject execute(KDDMLOperatorSettings settings,
                                        java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    // get output destination
    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.PMML_CLASSIFICATION_MODEL.getRepositoryPath() +
                                         settings.getXMLDestination());

    // get the combination type and the positive class
    combination_type = ((TREE_META_CLASSIFIER_SETTINGS) settings).getCombinationType();
    positive_class = ((TREE_META_CLASSIFIER_SETTINGS) settings).getPositiveClass();

    // get the trees
    trees = new TreeModelManager[arguments.size()];
    for(int i=0; i<arguments.size(); i++)
      trees[i] = (TreeModelManager) arguments.get(i);

    // controlla che ogni albero sia compatibile --> stesso MiningSchema
    if (!areMiningSchemaCompatibles()) {
      String msg = "Input Trees are incompatible for combination: different mining schema!";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INCOMPATIBLE_OBJECTS);
    }

    // controlla che in caso di combinazione AND, OR l'attributo di classificazione sia di
    // tipo binario
    if (combination_type != TREE_META_CLASSIFIER_SETTINGS.COMMITTEE_TREE) {
      String[] values = trees[0].getTargetAttributeValues();
      if (values.length != 2) {
        // exception
        String msg = "AND/OR combination is allowed only for decision trees!";
        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }

    return (KDDMLObject) combineTrees(dest);
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    return KDDMLObjectType.PMML_CLASSIFICATION_MODEL;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  private boolean areMiningSchemaCompatibles() {

    MiningSchema m1, m2;
    m1 = (MiningSchema) trees[0].getMiningSchema();
    for (int i=1; i<trees.length; i++) {
      m2 = (MiningSchema) trees[i].getMiningSchema();
      if (!m1.equals(m2))
        return false;
    }
    return true;
  }

  private TreeModel combineTrees(java.io.File dest) throws KDDMLCoreException {

    // get mining schema from the first tree
    MiningSchema m = (MiningSchema) trees[0].getMiningSchema();
    // get the data dictionaryfrom the first tree
    DataDictionary d = (DataDictionary) trees[0].getDataDictionary();

    Node combination = getRootNode();

    // build the factory with the node combination as root and no confusion matrixes
    ClassificationProprietaryFactory fact =
        new ClassificationProprietaryFactory(dest, d, m, combination, null, null);

    // setta le proprieta' del factory...
    fact.setOverwrite(true);
    fact.setAlgorithmName(getAlgName());
    fact.setModelName(getModelName());
    fact.setBinaryNodeSplitting(getSplitting());

    return (TreeModel) fact.newInstance();
  }

  private String getAlgName() {
    String s = "";
    switch(combination_type) {
      case TREE_META_CLASSIFIER_SETTINGS.COMMITTEE_TREE:
        s = "Committee Combination";
        break;
      case TREE_META_CLASSIFIER_SETTINGS.AND_TREE:
        s = "And Combination";
        break;
      case TREE_META_CLASSIFIER_SETTINGS.OR_TREE:
        s = "Or Combination";
        break;
    }
    return s;
  }

  private String getModelName() {

    String s = "";
    switch(combination_type) {
      case TREE_META_CLASSIFIER_SETTINGS.COMMITTEE_TREE:
        s = "Committee";
        break;
      case TREE_META_CLASSIFIER_SETTINGS.AND_TREE:
        s = "And";
        break;
      case TREE_META_CLASSIFIER_SETTINGS.OR_TREE:
        s = "Or";
        break;
    }

    String ris = s+"(";
    for(int i=0; i<trees.length; i++)
      ris = ris + trees[i].getModelName() + ",";
    ris = ris.substring(0, ris.length()-1);
    return ris+")";
  }

  private boolean getSplitting() {
    for(int i=0; i<trees.length; i++)
      if (!trees[i].isBinarySplitting())
        return false;
    return true;
  }

  private Node getRootNode() throws KDDMLCoreException {

    CompoundTree ris = null;
    switch(combination_type) {
      case TREE_META_CLASSIFIER_SETTINGS.COMMITTEE_TREE:
        ris = new CompoundBaggingsTree();
        break;
      case TREE_META_CLASSIFIER_SETTINGS.AND_TREE:
        String neg = getNegativeClass();
        ris = new CompoundBooleanTree(positive_class, neg, CompoundBooleanTree.AND_TREE);
        break;
      case TREE_META_CLASSIFIER_SETTINGS.OR_TREE:
        neg = getNegativeClass();
        ris = new CompoundBooleanTree(positive_class, neg, CompoundBooleanTree.OR_TREE);
        break;
      default:
        return null;
    }

    // add the trees...
    for (int i=0; i<trees.length; i++) {
      ris.addTree(trees[i].getRoot());
    }

    return (Node) ris;
  }

  /**
   * Get the negative class.
   *
   * @return String
   */
  private String getNegativeClass() throws KDDMLCoreException {

    String[] target_values = trees[0].getTargetAttributeValues();
    if (target_values[0].equals(positive_class))
      return target_values[1];
    else
      return target_values[0];
  }
}
