package kddml.Operators.Preprocessing.DiscretizationAlgorithms;

import weka.core.*;
import kddml.Operators.Preprocessing.PP_NUMERIC_DISCRETIZATION_SETTINGS;
import java.util.Vector;
import java.util.Enumeration;
import kddml.Operators.KDDMLOperatorException;

/**
 * This class represents a numeric interval, such as (70, 80].
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
class NumericInterval {

  private double inf;

  private double sup;

  private boolean is_open_inf;

  private boolean is_open_sup;


  /**
   * Constructor.
   * Build the interval given a string representation of that interval and the minumum
   * and maximum values for the training instances.
   * The interval description can be:
   * '(-inf - value]'
   * '(value - value2]'
   * '(value2 - -inf)'
   * 'ALL'
   *
   * @param intervallo string representing the interval
   * @param min minimum value of the discretization attribute
   * @param max maximum value of the discretization attribute
   */
  protected NumericInterval(String intervallo, double min, double max) {

    double fine = max;
    double inizio = min;

    //se si ha un unico intervallo, il valore della stringa
    //intervallo  'ALL' quindi inizio e fine
    //devono assumere i valori del minimo e del massimo
    //rispettivamente

    //se non  stato ottenuto un unico intervallo
    if (!intervallo.equals("'All'")) {
      //divido la stringa in due pezzi con punto di
      //staccamento il "-"
      String[] splitted = intervallo.split("-");
      String primo = splitted[0];
      String secondo = splitted[1];

      //se dopo lo split , si ottiene nel primo pezzo
      // la stringa "'(", vuol dire che l'intervallo  del tipo "'(-inf-..')
      //quindi l'inizio dell'intervallo rimane il minimo mentre
      //la fine va calcolata
      if (primo.equals("'(")) {
        String secondo2 = intervallo.substring( (intervallo.indexOf("f") + 2),
                                                intervallo.length());
        fine = new Double(secondo2.substring(0, secondo2.length() - 2)).
            doubleValue();

      }
      else {
        //se dopo lo split , si ottiene nel secondo pezzo
        // la stringa "inf)'", vuol dire che l'intervallo  del tipo
        // "'(..-inf)'"
        //quindi la fine dell'intervallo rimane il massimo mentre
        //l'inizio va calcolato

        if (secondo.equals("inf)'")) {
          String primo2 = intervallo.substring(0, (intervallo.indexOf("-")));
          inizio = new Double(primo2.substring(2, primo2.length())).doubleValue();
        }
        else {
          //se ho un generico intervallo che non ha tra gli estremi il massimo
          //oppure il minimo
          //va calcolato l'inizio e la fine
          inizio = new Double(primo.substring(2, primo.length())).
              doubleValue();
          fine = new Double(secondo.substring(0, secondo.length() - 2)).
              doubleValue();
        }
      }
    }

    inf = inizio;
    sup = fine;

    // controlla se l'intervallo e' aperto o chiuso....
    if (intervallo.startsWith("'[") || intervallo.startsWith("'(-inf"))
      is_open_inf = false;
    else
      is_open_inf = true;
    if (intervallo.endsWith("]'") || intervallo.endsWith("inf)'"))
      is_open_sup = false;
    else
      is_open_sup = true;
  }

  /**
   * Get a string representation of the interval.
   *
   * @return String
   */
  public String toString() {
    String s = "";
    if (is_open_inf)
      s = s + "(";
    else
      s = s + "[";
    s = s + new Double(inf).toString() + " - ";
    s = s + new Double(sup).toString();
    if (is_open_sup)
      s = s + ")";
    else
      s = s + "]";
    return s;
  }


  /**
   * Tests if a value belongs to the interval.
   *
   * @param d double
   * @return boolean
   */
  protected boolean belongs(double d) {
    boolean inf_value = false;
    boolean sup_value = false;

    if (inf == sup)
      if (d == inf)
        return true;
      else
        return false;

    // controlla l'estremo inferiore
    if ((d > inf) ||
        ((d == inf) && (!is_open_inf)))
      inf_value = true;

    // controlla l'estremo superiore
    if ((d < sup) ||
        ((d == sup) && (!is_open_sup)))
      sup_value = true;

    return (inf_value && sup_value);
  }

  /**
   * Get the mean of the input values belonging to the interval considered.
   *
   * @param values double[] all values whose mean must be computed.
   * @return double
   */
  protected double getMean(double[] values) {
    double sum = 0;
    double count = 0;
    for (int i=0; i<values.length; i++) {
      if (belongs(values[i])) {
        sum = sum + values[i];
        count++;
      }
    }
    if (count != 0)
      return sum / count;
    else
      return 0;
  }

  /**
   * Get the median of the input values belonging to the interval considered.
   *
   * @param values double[] all values whose median must be computed.
   * @return double
   */
  protected double getMedian(double[] values) {
    double sum = 0;
    double count = 0;
    for (int i=0; i<values.length; i++) {
      if (belongs(values[i])) {
        sum = sum + values[i];
        count++;
      }
    }
    if (count == 0)
      return 0;
    else {
      double mean = sum / count;
      int position = 0;
      double current_min = Double.MAX_VALUE;
      for (int i=0; i<values.length; i++) {

        double distance = Math.abs(mean - values[i]);
        if (distance < current_min) {
          current_min = distance;
          position = i;
        }
      }

      return values[position];
    }
  }

  /**
   * Get the minimum of the input values belonging to the interval considered.
   *
   * @param values double[] all values whose minimum must be computed.
   * @return double
   */
  protected double getMin(double[] values) {
    double min = Double.MAX_VALUE;
    int position = 0;
    for (int i=0; i<values.length; i++) {
      if ((belongs(values[i])) && (values[i] < min)) {
        min = values[i];
        position = i;
      }
    }
    return values[position];
  }

  /**
   * Get the minimum of the input values belonging to the interval considered.
   *
   * @param values double[] all values whose minimum must be computed.
   * @return double
   */
  protected double getMax(double[] values) {
    double max = Double.MIN_VALUE;
    int position = 0;
    for (int i=0; i<values.length; i++) {
      if ((belongs(values[i])) && (values[i] > max)) {
        max = values[i];
        position = i;
      }
    }
    return values[position];
  }
}
