/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.Vector;
import java.io.File;
import weka.core.Instances;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import kddml.Core.DataSources.PPInternalTable;
import kddml.Core.KDDMLObject;
import kddml.Core.KDDMLCoreException;
import kddml.Operators.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description: classe INSTANCE_LEVEL_Dipendent_Transformation che implementa
 * l'interfaccia degli operatori di PreProcessing.
 * Superclasse per tutti gli operatori di preprocessing che producono trasformazioni
 * sulle istanze del dataset,  con risultati che sono dipendenti dai
 * valori assunti dalle altre istanze.
 * Gli operatori in questione sono:
 *
 * --PPDISCRETIZATION;
 * --PPMARK_DUPLICATES;
 * --PPMERGE_DUPLICATES;
 * --PPSAMPLING;
 * --PPSORTING_BY_FREQUENCY;
 *
 *</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */

public abstract class InstanceLevelDependentTransformation extends PreprocessingResolver {

  /**
   * Esegue l'operatore procedendo importando tutte le istanze in memoria.
   *
   * @return tabella estesa
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // importa tutte le istanze
    Instances[] old_inst = ((Instances[]) input.load(PPInternalTable.ALL));

    // richiama il metodo principale per ottenere le nuove istanze
    Instances[] new_inst = execute(old_inst[0], old_inst[1]);

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = new PPWekaFactory(dest, new_inst[0], new_inst[1]);
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();

    setAllTaxonomies(input, output);

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }


  /**
   * Metodo astratto che conterra' il corpo dell'operatore.
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  protected abstract Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLOperatorException, KDDMLCoreException;

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 1024;
  }

  /**
   * Return true if the operator performe an high number of access to the hard disk.
   *
   * @return boolean
   */
  public boolean highDiskAccessIsNeeded() {
    return false;
  }
}
