/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.Vector;
import weka.core.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import kddml.Core.DataSources.PPInternalTable;
import kddml.Core.KDDMLObject;
import kddml.Core.KDDMLCoreException;
import kddml.Operators.*;

/**
 * <p>Title: KDDML</p>
*
 * <p>Description: Classe che implementa
 * l'interfaccia degli operatori di PreProcessing.
 * Superclasse per tutti gli operatori di preprocessing che producono trasformazioni
 * sulle istanze del dataset,  con risultati che sono indipendenti dai
 * valori assunti dalle altre istanze.
 * Gli operatori in questione sono:
 *
 * -- PPNORMALIZATION;
 * -- PPMARKING_QUERY;
 * -- PPREWRITING;
 * -- PPREMOVE_ROWS;
 * -- PP_HIERARHCICAL_DISCRETIZATION;
 *
 *
 * Sulle trasformazioni locali ha senso definire una restrizione del tipo WHERE
 * in stile SQL. Per questa classe, lelaborazione consiste nell'applicazione di un operatore
 * su ogni tupla della tabella in ingresso ;
 * ogni tupla viene cos trasformata in modo indipendente dalle altre.
 * Si pu allora controllare lapplicazione di tale funzione locale  attraverso una condizione
 * di restrizione sulla tupla. Tale condizione, come nella clausola WHERE nelle interrogazioni SQL,
 *   rappresentata da unespressione booleana formata da operatori di confronto applicati ad attributi
 * e costanti, dalla composizione logica di espressioni booleane.
 * La condizione di restrizione, rappresentata dalla clausola WHERE (condizione), va semplicemente aggiunta alla fine del costrutto.
 *</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */


public abstract class InstanceLevelIndependentTransformation
    extends PreprocessingResolver {

  /**
   * Esegue l'operatore procedendo per blocchi di istanze.
   * Ciascun blocco di istanze  costituito da un numero di istanze "ottimale".
   *
   * @return tabella estesa
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // importa soltanto lo schema delle istanze...
    Instances[] inst = ((Instances[]) input.load(0));

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = new PPWekaFactory(dest, inst[0], inst[1]);
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();
    setAllTaxonomies(input, output);

    long opt = input.optimize();
    Instances tuple, metatuple;
    Instances new_tuple, new_metatuple;

    while (input.hasNext()) {

      // importazione delle istanze
      Instances[] inst2 = ((Instances[]) input.load(opt));
      tuple = inst2[0];
      metatuple = inst2[1];
      new_tuple = new Instances(tuple);
      new_metatuple = new Instances(metatuple);
      new_tuple.delete();
      new_metatuple.delete();

      // ciclo che scorre le tuple
      for (int i=0; i<tuple.numInstances(); i++) {
        // instance contiene la nuova istanza
        Instance[] instance = execute(tuple.instance(i), metatuple.instance(i));
        if (instance != null) {
          new_tuple.add(instance[0]);
          new_metatuple.add(instance[1]);
        }
      }

      output.addInstances(new_tuple, new_metatuple);
    }

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }

  /**
   * Return the new instance. Null if no instance must be add.
   *
   * @param tupla Instance
   * @param metatupla Instance
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return Instance[]
   */
  protected abstract Instance[] execute(Instance tupla, Instance metatupla)
      throws KDDMLOperatorException, KDDMLCoreException;

  /**
   * Return true if the operator performe an high number of access to the hard disk.
   *
   * @return boolean
   */
  public boolean highDiskAccessIsNeeded() {
    return true;
  }
}
