/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import weka.core.*;

import org.w3c.dom.Element;
import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;
import java.util.Hashtable;
import java.util.regex.*;

/**
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */

public class PP_DIVIDING_ATTRIBUTE_RESOLVER extends SchemaLevelTransformation {

  // attributo sorgente
  private String source;

  // attributo destinazione 1
  private String destination1;

  // attributo destinazione 2
  private String destination2;

  // espressione regolare
  private String regex;

  // marca
  private String mark_ex;

  // source attribute must be removed
  private boolean remove_source;

  // tuple originarie - inizializzate da execute
  private Instances tuple;

  // metatuple originarie - inizializzate da execute
  private Instances metatuple;

  // numero di attributi dello schema originario
  private int num_attributi;

  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    //sottoalbero relativo agli attributi da unificare
    source = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).getSourceAttributeName();
    destination1 = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).getDestinationAttribute1();
    destination2 = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).getDestinationAttribute2();
    regex = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).getRegularExpression();
    mark_ex = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).getMarkException();
    remove_source = ( (PP_DIVIDING_ATTRIBUTE_SETTINGS) settings).isAttributeSourceRemoved();

    // legge lo schema dei dati per controllare la presenza dell'attributo sorgente
    PPInternalTableManager t = (PPInternalTableManager) arguments.get(0);
    DataStatisticsManager stat = t.getStatistic();
    num_attributi = stat.getNumAttributes();

    if (stat.getAttributeIndex(source) == -1) {
      // genera eccezione se un attributo non viene individuato
      String msg = "The attribute < " + source + " > do not exist.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    // controlla che gli attributi di destinazione non esistano nello schema
    if (stat.getAttributeIndex(destination1) != -1) {
      String msg = "Attribute < " + destination1 + " > already exists in input table.";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
    if (stat.getAttributeIndex(destination2) != -1) {
      String msg = "Attribute < " + destination2 + " > already exists in input table.";
      throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
    }
  }

  /**
   * crea e restituisce la stringa che rappresenta l'attributo "Description"
   * per il nodo figlio della history
   * relativo a questa operazione di preprocessing.
   *
   * @return descrizione operazione da inserire nella History
   */
  protected String createDescription() {
    return "Split " + source + " on " + destination1 + " and " + destination2 +
        " using Reg.Expr.: " + regex;
  }

  protected Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLCoreException, KDDMLOperatorException {

    this.tuple = tuple;
    this.metatuple = metatuple;

    // esegue l'operatore
    makeNewAttributeSplitting();

    // rimuovo l'attributo sorgente se richiesto...
    if (remove_source) {
      tuple.deleteAttributeAt(tuple.attribute(source).index());
      metatuple.deleteAttributeAt(metatuple.attribute(source+"_metadata").index());
    }
    // costruisce l'output
    Instances[] ris = new Instances[2];
    ris[0] = tuple;
    ris[1] = metatuple;
    return ris;
  }

 /* protected Instances buildNewSchema(Instances old_schema)
      throws KDDMLCoreException, KDDMLOperatorException {

    System.out.println(old_schema.toString());
    Attribute a = old_schema.attribute(source);

    // tutto ok --> si generano gli attributi destinazione di tipo string
    Instances inst1 = PreprocessingUtility.addNewAttribute(old_schema, destination1,
        AttributeType.STRING, num_attributi, null);
    Instances inst2 = PreprocessingUtility.addNewAttribute(inst1, destination2,
        AttributeType.STRING, num_attributi+1, null);

    // rimuovo l'attributo sorgente se richiesto...
    if (remove_source)
      inst2.deleteAttributeAt(a.index());

    return inst2;
  }*/

  /**
   * Crea i nuovi valori da inserire nei due attributi relativi ai risultati dello split
   *
   * @param nome_primo String
   * @param nome_secondo String
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  private void makeNewAttributeSplitting()
      throws KDDMLOperatorException, KDDMLCoreException {

    FastVector fv = new FastVector();
    fv = null;
    Attribute attr1 = new Attribute(destination1, fv);
    Attribute attr2 = new Attribute(destination2, fv);

    Instances[] inst1 =
        PreprocessingUtility.addNewAttribute(tuple, metatuple, destination1,
                                             AttributeType.STRING, num_attributi, null);
    tuple = inst1[0];
    metatuple = inst1[1];

    Instances[] inst2 =
        PreprocessingUtility.addNewAttribute(inst1[0], inst1[1], destination2,
                                             AttributeType.STRING, num_attributi+1, null);

    tuple = inst2[0];
    metatuple = inst2[1];

    // esegue l'operatore vero e proprio
    setValuesWithSplitResult();
  }

  /**
   * riempie i valori delle tuple relative ai nuovi  attributi  inserendo
   * i valori del risultato dello split del valore dell'attributo origine di origine, e
   * riempie i valori dei metadati associati ai nuovi attributi.
   *
   * @param attributoDaSplittare Attribute
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  private void setValuesWithSplitResult() throws
      KDDMLOperatorException, KDDMLCoreException {

    Attribute a = tuple.attribute(source);

    String[] valori = new String[2];
    String origine;
    String Descrizione_marca = null;

    for (int i = 0; i < tuple.numInstances(); i++) {
      if (a.isNumeric())
        origine = tuple.instance(i).stringValue(a);
      else
        origine = tuple.instance(i).toString(a);

        // String pattern = RegEx.getAttribute("Pattern");
      Pattern p = Pattern.compile(regex);
      Matcher m = p.matcher(origine);
      boolean esiste_sottosequenza_ER = m.find();
      // System.out.println("esiste_sottosequenza_ER " + esiste_sottosequenza_ER);
      if (esiste_sottosequenza_ER) {
        int start_ER = m.start();
        valori = getSplitResults(origine, p);
        if (valori[0].compareTo("") == 0) {
          //se coincide con il pattern verra' spezzato in (""),("pattern")
          //quindi il primo valore del vettore risultante dallo split contiene la stringa vuota
          //che deve corrispondere con un valore missing
          PreprocessingUtility.setStringValue(num_attributi, "?", tuple, i);
        }
        else {
          PreprocessingUtility.setStringValue(num_attributi, valori[0], tuple, i);
        }

        PreprocessingUtility.setStringValue(num_attributi + 1, origine.substring(start_ER,
            origine.length()), tuple, i);

      }
      else {
        if (mark_ex != null) {
          // se e' previsto il tag Exception si decide di marcare i valori per i quali non e'
          // stata trovata alcuna occorrenza del pattern indicato
          PreprocessingUtility.addMarkInAttribute(num_attributi, mark_ex, metatuple.instance(i));
          PreprocessingUtility.addMarkInAttribute(num_attributi + 1, mark_ex, metatuple.instance(i));

        }
        //se non e' previsto il tag Exception gli attributi relativi ai metadati
        //dei nuovi attributi contenenti i risultati dello split
        //esistono (creati subito dopo aver controllato la validita' dei nomi scelti per i
        //nuovi attributi) ma hanno valori missing!!
      }
    }
  }

  /**
   * getSplitResults(String stringa,Pattern pattern)
   * restituisce un array di stringhe di dimensione due
   * nel quale saranno contenute le stringhe risultanti dallo spezzamento
   * del valore dell'attributo rispetto alla prima occorrenza del pattern.
   *  Si ha rispettivamente:
   *   -  in pos. [0] la sottosequenza del valore dell'attributo fino al carattere in posizione
   *     precedente rispetto a quella del primo carattere della prima occorrenza del pattern
   *  descritto dall'espressione regolare
   *   -  in pos [1] la sottosequenza del valore dell'attributo a partire dalla posizione
   *   immediatamente successiva a quella dell'ultimo carattere della prima occorrenza de pattern,
   *      fino all'ultimo carattere del valore dell'attributo
   * usando lo split del pacchetto di java.util.regex.pattern
   * si ha questi tipi di risultati:
   *  1. Esempio :getSplitResults(aaapippobbb,pippo)
   *      restituisce
   *     nuovi_valori_spezzati[0]="aaa"
   *      nuovi_valori_spezzati[1]="bbb"
   *
   *  2. Esempio : getSplitResults(pippo,pippo)
   *      nuovi_valori_spezzati[0]=""
   *      nuovi_valori_spezzati[1]=""
   *  3. Esempio : getSplitResults(apippo,pippo)
   *     nuovi_valori_spezzati[0]="a"
   *      nuovi_valori_spezzati[1]=""
   *  3. Esempio : getSplitResults(pippob,pippo)
   *      nuovi_valori_spezzati[0]=""
   *      nuovi_valori_spezzati[1]="b"
   */

  private String[] getSplitResults(String stringa, Pattern pattern) throws
      KDDMLOperatorException, KDDMLCoreException {

    String[] nuovi_valori_spezzati = new String[2];
    try {

      int fine_stringa = stringa.length();
      CharSequence sequenza = (stringa).subSequence(0, fine_stringa);
      nuovi_valori_spezzati = pattern.split(sequenza, 2);

    }
    catch (Exception e) {
      String msg = "PP_SPLITTING_ATTRIBUTE : pattern <" + pattern +
          "> is not valid!";

      new KDDMLOperatorException(msg, msg, 1);

    }
    return nuovi_valori_spezzati;
  }
}
