/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

/**
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */

import weka.core.*;

import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;

/**
 * <p>Title: KDDML</p>
 * <p>Description:  * Classe che risolve la rimozione di  una sequenza di attributi dal dataset
 * delle tuple e i corrispondenti metadati nelle metatuple.
 * L'operazione di rimozione e' una trasformazione a livello di schema.
 </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */

public class PP_FILTER_ATTRIBUTES_RESOLVER extends SchemaLevelTransformation {

  // lista dei nomi degli attributi
  private Vector names;

  // tiene o elimina
  private boolean take;

  /**
   *  -- prende dalla query la PPtable sorgente (figlio in posizione 0 )
   *    e ne estrae la rappresentazione interna.
   *
   * -- inizializza le instances tuple e metatuple come insiemi vuoti  delle istanze
   *    componenti la PPTable input
   * @param kddml_objects vettore degli oggetti kddml tra i parametri della query
   * @param elements vettore degli elementi che non sono kddml_objects ma sono parametri della query
   * @param attributes tabella hash contenente le associazioni nome-valore dei parametri della query
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */

  protected void readInput(KDDMLOperatorSettings settings,
                           java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    names = ((PP_FILTER_ATTRIBUTES_SETTINGS) settings).getAttributeNames();
    take = ((PP_FILTER_ATTRIBUTES_SETTINGS) settings).takeAttribute();

    // legge lo schema dei dati per controllare la presenza degli attributi
    PPInternalTableManager t = (PPInternalTableManager) arguments.get(0);
    DataStatisticsManager stat = t.getStatistic();

    for(int i=0; i< names.size(); i++) {

      String name = names.get(i).toString();
      if (stat.getAttributeIndex(name) == -1) {
        // genera eccezione se un attributo non viene individuato
        String msg = "The attribute < " + name + " > do not exist.";

        throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
      }
    }
  }

  /**
   * crea e restituisce la stringa che rappresenta l'attributo "Description"
   * per il nodo figlio della history
   * relativo a questa operazione di preprocessing.
   *
   * @return descrizione operazione da inserire nella History
   */
  protected String createDescription() {
    String ris = names.elementAt(0).toString();
    for (int i = 1; i < names.size(); i++) {
      ris = (ris + ", " + (names.elementAt(i)).toString());
    }
    if (take)
      return "Take attribute(s): "+ris;
    else
      return "Removed attribute(s): "+ris;

  }

  protected Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLCoreException, KDDMLOperatorException {
    Instances inst1, inst2;
    if (!take) {
      inst1 = removeAttributes(tuple, names);
      Vector names2 = new Vector();
      for (int i=0; i< names.size(); i++) {
        names2.add((String) names.get(i) + "_metadata");
      }
      inst2 = removeAttributes(metatuple, names2);
    }
    else {
      inst1 = takeAttributes(tuple, names);
      Vector names2 = new Vector();
      for (int i=0; i< names.size(); i++) {
        names2.add((String) names.get(i) + "_metadata");
      }
      inst2 = takeAttributes(metatuple, names2);
    }
    Instances[] ris = new Instances[2];
    ris[0] = inst1;
    ris[1] = inst2;
    return ris;
  }

  /*protected Instances buildNewSchema(Instances old_schema)
      throws KDDMLCoreException, KDDMLOperatorException {
    if (!take)
      return removeAttributes(old_schema, names);
    else
      return takeAttributes(old_schema, names);
  }*/

  private static Instances takeAttributes(Instances tuple, Vector names)
      throws KDDMLCoreException, KDDMLOperatorException {

    int num_attributi = tuple.numAttributes();
    boolean[] attr_to_del = new boolean[num_attributi];
    int deleted = 0;

    for (int j = 0; j < num_attributi; j++) {
      attr_to_del[j] = true;
    }

    for (int i = 0; i < names.size(); i++) {
      Attribute a = tuple.attribute(names.elementAt(i).toString());
      attr_to_del[a.index()] = false;

    }
    for (int j = 0; j < num_attributi; j++) {
      if (attr_to_del[j]) {

        tuple.deleteAttributeAt(j - deleted);
        deleted++;
      }
    }
    return tuple;
  }

  /**
   * Per ogni nome della sequenza specificata, elimina l'attributo con quel nome da tuple
   * e elimina i metadati corrispondenti in metatuple
   *
   * @param FastVector dei nomi degli attributi
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */

  private static Instances removeAttributes(Instances tuple, Vector names) throws
      KDDMLCoreException, KDDMLOperatorException {

    for (int i = 0; i < names.size(); i++) {
      Attribute a = tuple.attribute(names.elementAt(i).toString());
      try {
        // l'attributo viene rimosso dal dataset (tuple e metatuple)

        tuple.deleteAttributeAt(a.index());
      }
      catch (Exception e) {
        // l'attributo non pu essere rimosso: si solleva eccezione

        String msg = "PP_FILTER_ATTRIBUTES: Attribute < " + names.elementAt(i) +
            " > couldn't be removed.";

        throw new KDDMLOperatorException(msg, msg, 1);

      }
    }
    return tuple;
  }
}
