/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import weka.core.*;
import org.w3c.dom.Element;
import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;
import java.util.Hashtable;


/**
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */

public class PP_FOLDING_RESOLVER  extends SchemaLevelTransformation {

  private String source;

  private String destination;

  private int index_source;

  private int index_destination;


  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    source = ( (PP_FOLDING_SETTINGS) settings).getSourceAttributeName();
    destination = ( (PP_FOLDING_SETTINGS) settings).getDestinationAttributeName();

    // legge lo schema dei dati per controllare la presenza degli attributi
    PPInternalTableManager t = (PPInternalTableManager) arguments.get(0);
    DataStatisticsManager stat = t.getStatistic();

    index_source = stat.getAttributeIndex(source);
    // controlla che l'attributo sorgente sia presente nella tabella
    if (index_source == -1) {
      String msg = "The attribute < " + source +" > do not exist.";
      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    index_destination = stat.getAttributeIndex(destination);
    // controlla che l'attributo destinazione sia presente nella tabella
    if (index_destination == -1) {
      String msg = "The attribute < " + destination +" > do not exist.";
      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    // controlla che i due attributi siano dello stesso tipo
    AttributeStatisticManager stat1 = stat.getStatistic(source);
    AttributeStatisticManager stat2 = stat.getStatistic(destination);
    if (!stat1.getType().equals(stat2.getType())) {
      String msg = "The source and destination attributes must have the same type.";
      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
    }
  }

  protected String createDescription() {

    return "Folding of "+source+" into "+destination;
  }

  protected Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLCoreException, KDDMLOperatorException {

    Instance instance;
    Instance meta_instance;

    int num_attr = tuple.numAttributes();
    int num_inst = tuple.numInstances();

    // per ogni istanza aggiungo a tuple e metatuple i valori come duplicati,
    // tranne che per l'attributo destinazione del fold per il quale metto i
    // valori dell'attributo origine
    for (int i = 0; i < num_inst; i++) {

      instance = new Instance(num_attr);
      meta_instance = new Instance(num_attr);
      instance.setDataset(tuple);
      meta_instance.setDataset(metatuple);
      instance = setValues(tuple.instance(i), instance, index_source, index_destination);
      meta_instance = setValues(metatuple.instance(i), meta_instance, index_source, index_destination);
      metatuple.add(meta_instance);
      tuple.add(instance);

    }

    tuple.deleteAttributeAt(index_source);
    metatuple.deleteAttributeAt(index_source);

    Instances[] ris = new Instances[2];
    ris[0] = tuple;
    ris[1] = metatuple;
    return ris;
  }

  /*protected Instances buildNewSchema(Instances old_schema)
      throws KDDMLCoreException, KDDMLOperatorException {

    old_schema.deleteAttributeAt(index_source);
    return old_schema;
  }*/

  /**
   * setto l'istanza di destinazione come duplicato di quella di origine
   * tranne che per l'attributo in posizione indice2, che e' l'attributo destinazione del FOLD
   * Per questo attributo prendo come valore quello dell'attributo origine.
   *
   * @param istanza_origine istanza originale prima del fold
   * @param istanza_dest istanza da settare a seguito del fold
   * @indice1 indice colonna relativo all'attributo origine del fold
   * @indice2  indice colonna relativo all'attributo destinazione del fold
   * @return istanza come duplicato per il ripiegamento
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  private static Instance setValues(Instance istanza_origine,
                                    Instance istanza_dest,
                                    int indice1,
                                    int indice2)
      throws KDDMLOperatorException {

    //setto tutti i campi dell' "instance" (tranne quelli dell'attributo in posizione indice 2)
    //come copia dell'istanza che sto analizzando

    for (int i = 0; i < istanza_origine.numAttributes(); i++) {
      //per ogni attributo tranne quello in posizione indice2
      if (i != indice2) {
        //se e' un attributo nominale o una stringa
        if (! (istanza_origine.attribute(i).isNumeric())) {
          if (! (istanza_origine.isMissing(i))) {
            istanza_dest.setValue(istanza_origine.attribute(i),
                                  istanza_origine.stringValue(i));
          }
          else
            istanza_dest.setMissing(istanza_origine.attribute(i));
        }
       /* else { //se e' un attributo numerico
          if (! (instanza_origine.isMissing(i))) {
            istanza_dest.setValue(instanza_origine.attribute(i),
                                  instanza_origine.value(i));
          }
          else
            istanza_dest.setMissing(instanza_origine.attribute(i));
        }*/
      }
    }

    //a questo punto manca l'assegnamento al campo relativo all'attributo in
    //posizione indice2 che prima avevamo saltato
    if (! (istanza_origine.attribute(indice2).isNumeric())) {
      // se e' nominale o tipo stringa

      //setto il valore dell'attributo in posizione indice2
      // con il valore che assume in quella instanza l'attributo in posizione indice1

      //NOTA: se l'attributo in posizione indice2 ha tipo stringa,
      //accetto anche che il tipo dei valori che vado a mettere (quelli dell'attributo in pos indice1)
      //siano di tipo numerico o nominale
      //se invece l'attributo in posizione indice2 ha tipo nominale e quello di pos indice1 ha tipo
      //numerico , o comunque ha valori che non sono fra quelli ammessi nel nominale in cui voglio
      //inserirli, verra' sollevata un'eccezione

      if (! (istanza_origine.isMissing(indice1))) {
        try {
          istanza_dest.setValue(indice2, istanza_origine.stringValue(indice1));
        }
        catch(java.lang.IllegalArgumentException exc) {
          // incompatibilita' di categorie per attributi nominali...
          String msg = "Illegal category compatibility for given nominal attribute!";
          throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
        }
      }
      else
        istanza_dest.setMissing(indice2);
    }
    else {
      //se e' numerico
      if (istanza_origine.attribute(indice1).isNumeric()) {
        if (! (istanza_origine.isMissing(indice1))) {
          istanza_dest.setValue(indice2, istanza_origine.value(indice1));
        }
        else
          istanza_dest.setMissing(indice2);
      }
    }
    return istanza_dest;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }
}
