/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;
import kddml.Core.DataMining.Taxonomy.*;
import kddml.Operators.KDDMLOperatorException;
import weka.core.Instance;

/**
 * Effettua una discretizzazione sui valori assunti da un attributo categorico
 * in base ad una gerarchia.
 *
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Andrea Romei (romei@di.unipi.it)
 * @version 1.0
 */
public class PP_HIERARCHICAL_DISCRETIZATION_RESOLVER extends InstanceLevelIndependentTransformation {

  private int level;

  private String attribute_name;

  private String exception;

  private HierarchyModelManager hierarchy;

  public void readInput(KDDMLOperatorSettings settings , java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    level = ((PP_HIERARCHICAL_DISCRETIZATION_SETTINGS) settings).getLevel();
    attribute_name = ((PP_HIERARCHICAL_DISCRETIZATION_SETTINGS) settings).getAttributeName();
    exception = ((PP_HIERARCHICAL_DISCRETIZATION_SETTINGS) settings).getMarkException();

    // get the object hieararchy
    hierarchy = (HierarchyModelManager) arguments.get(1);

    // legge lo schema dei dati per controllare la presenza degli attributi
    PPInternalTableManager t = (PPInternalTableManager) arguments.get(0);
    DataStatisticsManager stat = t.getStatistic();

    // return error if the attribute do not exist
    if (stat.getAttributeIndex(attribute_name) == -1) {
      String msg = "The attribute < " + attribute_name + " > do not exist.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    // return error if the attribute is numeric
    if (stat.getStatistic(attribute_name).getType().equals(AttributeType.NUMERIC))
      throw new KDDMLOperatorException
          ("< " + attribute_name + " >" + " is numeric.",
           null,
           KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);

    // return error if the level is not compatible with the hierarchy depth
    if (level > hierarchy.getDepth())
      throw new KDDMLOperatorException
          ("Index of level "+level+" is not compatible with the hierarchy depth.",
           null,
           KDDMLOperatorErrorCodes.INVALID_SETTINGS);
  }

  /**
   * crea e restituisce la stringa che rappresenta l'attributo "Description"
   * per il nodo figlio della history
   * relativo a questa operazione di preprocessing.
   *
   * @return descrizione operazione da inserire nella History
   */
  protected  String createDescription () {
    String s = "Hierarchy generalization on attribute ";
    s = s + attribute_name + " at level "+level;
    return s;
  }

  protected Instance[] execute(Instance tupla, Instance metatupla)
      throws KDDMLOperatorException, KDDMLCoreException {

    // ricerca l'indice dell'attributo da discretizzare
    int index = 0;
    for (int i=0; i<tupla.numAttributes(); i++) {
      if (tupla.attribute(i).name().equalsIgnoreCase(attribute_name))
        index = i;
    }

    String item = tupla.stringValue(index);
    // get the generalization...
    String item2 = generalize(item);

    if (item2.equals("?")) {
      // attributo non generalizzabile: mette a missing l'istanza e marca i
      // metadati con l'eccezione...
      tupla.setMissing(index);
      PreprocessingUtility.addExceptionInAttribute(index, exception, metatupla);
    }
    else
      tupla.setValue(index, item2);

    Instance[] ris = new Instance[2];
    ris[0] = tupla;
    ris[1] = metatupla;
    return ris;
  }


  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE;
    if (i==1)
      return kddml.Core.KDDMLObjectType.PMML_HIERARCHY;
    return null;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  /**
   * Istanzia il metodo astratto con la chiamata all'operatore specifico
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  protected void doOperator()
      throws KDDMLOperatorException, KDDMLCoreException {

  }

   private String generalize(String item) throws KDDMLOperatorException {
     HierarchyNode node = (HierarchyNode) hierarchy.getNode(item);
     if (node == null)
       // nodo non trovato all'interno della gerarchia
       return "?";
     for (int i=0; i<level; i++) {
       if (node.isRoot())
         throw new KDDMLOperatorException
             ("Index of level "+level+" is not compatible with the hierarchy depth.",
              null,
              KDDMLOperatorErrorCodes.INVALID_SETTINGS);

       node = (HierarchyNode) node.getParent();
     }
     return node.getName();
   }
}


