/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import weka.core.*;
import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;

/**
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */

public class PP_MARK_DUPLICATES_RESOLVER extends InstanceLevelDependentTransformation{

  protected String mark;

  protected Vector key_names;

  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    key_names = ((PP_MARK_DUPLICATES_SETTINGS) settings).getKeyAttributes();
    mark = ((PP_MARK_DUPLICATES_SETTINGS) settings).getMark();

    DataStatisticsManager d = ((PPInternalTable) arguments.get(0)).getStatistic();

    if (key_names.size() == 0) {
      // si prendono tutti gli attributi della tabella
      AttributeStatisticManager[] s = d.getStatistics();
      for (int i=0; i<s.length; i++)
        key_names.add(s[i].getName());
    }

    else
      // verifica la presenza degli attributi nello schema...
      checkAttributes(d);

    System.out.println("Attributes: "+key_names.toString());
  }

  private void checkAttributes(DataStatisticsManager statistic) throws KDDMLOperatorException {

    for (int i=0; i<key_names.size(); i++) {
      String attribute_name = (String) key_names.get(i);
      int index = statistic.getAttributeIndex(attribute_name);
      // controlla che l'attributo sia presente nella tabella
      if (index == -1) {
        // genera eccezione se un attributo non viene individuato
        String msg = "The attribute < " + attribute_name +" > do not exist.";

        throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
      }
    }
  }


  protected  String createDescription () {

    String primo = (String) key_names.elementAt(0);
    String ris = primo;
    for (int i = 1; i < key_names.size();i++) {
      String name=(String) key_names.elementAt(i);
      ris = ris + "; " + name;
    }

    return "Mark duplicates based on key : "+ris;
  }


  protected Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLOperatorException, KDDMLCoreException {

    // ordina le tuple e metatuple sulla base della sequenza di attributi
    // a partire dall'attributo in posizione zero della sequenza
    Instances[] tuple_e_metatuple_ordinate = ordina(tuple, metatuple, 0);

    // marco nei metadati gli attributi chiave delle istanze duplicate
    markMetatuple(tuple_e_metatuple_ordinate[0], tuple_e_metatuple_ordinate[1]);

    //tuple e metatuple da restituire in output
    //tuple = tuple_e_metatuple_ordinate[0];
    //metatuple ora sono marcate
    //metatuple = tuple_e_metatuple_ordinate[1];

    return tuple_e_metatuple_ordinate;
  }


   /**
    * Restituisce le metatuple marcate se risultano duplicate sulla base dei
    * valori che assumono in tuple gli attributi  che compongono
    * la chiave.
    *
    *
    * @param tuple_ordinate
    * @param metatuple_ordinate
    * @throws KDDMLInterpreterException
    * @throws KDDMLCoreException
    */

   private void markMetatuple(Instances tuple_ordinate,
                              Instances metatuple_ordinate) throws
       KDDMLOperatorException, KDDMLCoreException {

     int num_Attr_seq =key_names.size();

     for (int i = 1; i < tuple_ordinate.numInstances(); i++) {
       if (isEqual(tuple_ordinate, tuple_ordinate.instance(i),
                   tuple_ordinate.instance(i - 1))) {

         for (int j = 0; j < num_Attr_seq; j++) {

           String nomeAttributo = (String)key_names.elementAt(j);
           Attribute a = tuple_ordinate.attribute(nomeAttributo);
           int indice = a.index();

           PreprocessingUtility.addMarkInAttribute(indice, mark, metatuple_ordinate.instance(i));
           PreprocessingUtility.addMarkInAttribute(indice, mark, metatuple_ordinate.instance(i-1));
         }

       }
     }
   }

   /**
    * Verifica se due istanze sono duplicate controllando l'ugualglianza
    * sui valori assunti dagli attributi della sequenza specificata.
    *
    * @param tuple_ordinate
    * @param istanza
    * @param istanza_precedente
    *
    * @return boolean :esito della verifica
    * @throws KDDMLInterpreterException
    * @throws KDDMLCoreException
    */
   protected boolean isEqual(Instances tuple_ordinate,
                             Instance istanza,
                             Instance istanza_precedente) throws
       KDDMLOperatorException, KDDMLCoreException {

     boolean ris = true;
     int key_cardinality = key_names.size();

     boolean[] attr = new boolean[key_cardinality];
     for (int j = 0; j < key_cardinality; j++)
       attr[j] = false;

     for (int i = 0; i < key_cardinality; i++) {

       String nomeAttributo = (String)key_names.elementAt(i);
       Attribute a = tuple_ordinate.attribute(nomeAttributo);
       if ( (!istanza.isMissing(a)) && (!istanza_precedente.isMissing(a))) {
         if (a.isNumeric()) { //numerico
           if (istanza.value(a) == istanza_precedente.value(a))
             attr[i] = true;
         }
         else { // NON numerico
           if (istanza.stringValue(a).equals(istanza_precedente.stringValue(a)))
             attr[i] = true;
         }
       }
     }
     //valuta in AND gli esiti ottenuti per il confronto  su tutti gli attributi
     for (int j = 0; j < key_cardinality; j++)
       ris = (attr[j] & ris);

     return ris;
   }

   /**
    * Metodo che ordina le Instances sulla base della chiave composta dagli attributi
    * specificati.
    * Ordina sulla base dell'attributo in posizione  "indiceAttributo"
    * e richiama se stesso ricorsivamente (con parametro l'attributo successivo),
    * su ogni gruppo di istanze che hanno lo stesso valore per l'attributo di ordinamento.
    *
    * @param istanzeDati
    * @param istanzeMeta
    * @param indiceAttributo
    * @return array di Instances di dimensione 2 contenente i dati e i metadati ordinati
    * @throws KDDMLInterpreterException
    * @throws KDDMLCoreException
    */
    Instances[] ordina(Instances istanzeDati,
                       Instances istanzeMeta,
                       int indiceAttributo)

        throws KDDMLOperatorException,KDDMLCoreException {

     //numero attributi sulla base dei quali effettuare l'ordinamento
     //num istanze contenute dalle istances da ordinare
     //( dati e metadati hanno la stessa cardinalit )
     int num = istanzeDati.numInstances();
     Instances[] ris = new Instances[2];

     String nomeAttr =(String) key_names.get(indiceAttributo);
     Attribute a = istanzeDati.attribute(nomeAttr);

     int indice_attr_in_tuple = a.index();

     //ordino tuple e metatuple sull'attributo

     istanzeDati.sort(indice_attr_in_tuple);


     int last_pos =istanzeMeta.numAttributes();
     istanzeMeta.insertAttributeAt(a,last_pos);
     istanzeMeta.sort(last_pos);
     istanzeMeta.deleteAttributeAt(last_pos);

     // prendo la cardinalit delle istanze che hanno lo stesso valore fino al primo valore distinto
     int numeroIstanzeUguali = getNumberOfInstanceHavingSameValues(istanzeDati, indice_attr_in_tuple);

     // costruisco  nuove instances per i dati e i metadati comprendenti solo le istanze con lo
     // stesso valore per l'attributo di ordinamento nell'istances delle tuple
     Instances newTuple = new Instances(istanzeDati, 0, numeroIstanzeUguali);
     Instances newMetaTuple = new Instances(istanzeMeta, 0, numeroIstanzeUguali);
     //e metto su altre instances le istanze rimanenti
     Instances altreTuple = new Instances(istanzeDati, numeroIstanzeUguali,
                                          num - numeroIstanzeUguali);
     Instances altreMeta = new Instances(istanzeMeta, numeroIstanzeUguali,
                                         num - numeroIstanzeUguali);
     if (indiceAttributo < key_names.size() - 1) {

       //finche' non finisco tutti gli attributi su cui devo effettuare l'ordinamento
       if (num - numeroIstanzeUguali > 0) {
         //se oltre alle prime istanze con valori uguali risultanti dall'ordinamento ne ho altre
         return add(ordina(newTuple, newMetaTuple, indiceAttributo + 1),
                    ordina(altreTuple, altreMeta, indiceAttributo + 1));
       }
       else {
         //se con il primo gruppo di istanze con valori  uguali
         //risultanti dall'ordinamento esaurisco il numero
         //di istanze totali dell'INSTANCES che sto analizzando
         return ordina(newTuple, newMetaTuple, indiceAttributo + 1);
       }
     }
     else { //se ho ordinato su tutti gli attributi dell'attribute_seq

       ris[0] = istanzeDati;
       ris[1] = istanzeMeta;
       return ris;

     }
   }

   /**
    * Conta quante istanze hanno lo stesso valore per uno degli attributi della sequenza
    * Le istanze sono ordinate pertanto i valori uguali sono contigui ,
    * e al primo valore distinto si e' sicuri di non avere piu' ulteriori valori
    * da conteggiare.
    * @param istanze
    * @param indice attributo da controllare
    *
    * @return numero di istanze che hanno lo stesso valore
    * @throws KDDMLInterpreterException
    * @throws KDDMLCoreException
    */

   private static int getNumberOfInstanceHavingSameValues(Instances istanze, int indice)
       throws  KDDMLCoreException {

     int i = 1;
     int cont = 1;
     while (i < istanze.numInstances()) {
       if (!istanze.attribute(indice).isNumeric()) {
         //se non e' numerico
         if (istanze.instance(i).stringValue(indice).
             equals(istanze.instance(i - 1).stringValue(indice))) {
           cont++;
           i++;
         }
         else
           i = istanze.numInstances();
       }
       else {
         //se l'attributo ha tipo numerico
         if (istanze.instance(i).value(indice) ==
             (istanze.instance(i - 1).value(indice))) {
           cont++;
           i++;
         }
         else
           i = istanze.numInstances();
       }
     }
     return cont;
   }

   /**
    * Unifica due insiemi di tuple e metatuple
    * @param PrimeIstanze
    * @param SecondeIstanze
    *
    * @return array di istanze (tuple e metatuple) risultati dopo l'unificazione
    * @throws KDDMLInterpreterException
    * @throws KDDMLCoreException
    */
   private static Instances[] add(Instances[] PrimeIstanze,
                                  Instances[] SecondeIstanze)

       throws  KDDMLCoreException {

     int num_tot = SecondeIstanze[0].numInstances();
     for (int i = 0; i < num_tot; i++) {
       PrimeIstanze[0].add(SecondeIstanze[0].instance(i));
       PrimeIstanze[1].add(SecondeIstanze[1].instance(i));
     }
     return PrimeIstanze;
   }
 }
