/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.*;
import kddml.Core.*;
import kddml.Operators.*;
import kddml.Core.DataSources.*;
import weka.core.*;
import kddml.Operators.Preprocessing.NormalizationAlgorithms.*;

/**
 * <p>Title: KDDML</p>
 * <p>Description: * Classe che risolve l'operatore di normalizzazione su una sequenza di attributi.
 * La normalizzazione e' una trasformazione a livello di istanza dei dati.
 * La normalizzazione puo' essere applicata solo su attributi che hanno tipo numerico.
 * I tipi di normalizzazione previsti sono:
 * -- decimale;
 * -- min_max ;
 * -- z_score.
 * Queste operazioni di normalizzazioni sono NON locali ,nel senso che hanno bisogno delle
 * informazioni riguardanti tutti i valori del dataset
 * per poter correttamente calcolare i valori:
 * ---di minimo e di massimo,
 * ---di deviazione standard e di media
 *
 * La normalizzazione decimale converte tutti i valori in un intervallo [-1,1].
 *
 * La normalizzazione min-max converte tutti i valori in un intervallo [min,max].
 *
 * La normalizzazione z-score converte tutti i valori secondo la trasformazione
 *  v = ( v  media A ) / dev_standard A   .
 * Questo metodo  di particolare utilit quando non sono noti il valore massimo
 * ed il valore minimo attuali, oppure in presenza di outliers che avrebbero un
 *  impatto troppo forte sulla normalizzazione minimo-massimo.
</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */
public class PP_NORMALIZATION_RESOLVER extends InstanceLevelIndependentTransformation {

  private KDDMLOperatorSettings settings;

  private DataStatisticsManager statistic;

  private String description;

  private Vector enum2;

  private int[] attributes_index;

  private NormalizationAlgorithmResolverTask as;

  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    // get arguments...
    this.settings = settings;
    statistic = ((PPInternalTable) arguments.get(0)).getStatistic();

    // controlla la presenza degli attributi...
    enum2 = ((PP_NORMALIZATION_SETTINGS) settings).getAttributes();
    // test if the attributes are in the table and are numeric...
    attributes_index = checkAttributes(enum2);
  }

  /**
   * Overloading of the main execute method.
   *
   * @param settings KDDMLOperatorSettings
   * @param arguments Vector
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // get the list of attributes...
    enum2 = ((PP_NORMALIZATION_SETTINGS) settings).getAttributes();

    // test if the algorithm is supported...
    String alg_name = ((PPAlgorithmSettings) settings).algorithm_name;

    NormalizationAlgorithmFactory factory = new NormalizationAlgorithmFactory(alg_name);
    factory.initialize();
    if (!factory.isSupported()) {
      throw new KDDMLOperatorException
          ("Algorithm < " + alg_name + " > not supported!", null,
           KDDMLOperatorErrorCodes.UNSUPPORTED_ALGORITHM);
    }

    // algoritmo di preprocessing individuato...
    as = (NormalizationAlgorithmResolverTask) factory.newResolverInstance();

    // lettura dei settaggi dell'algoritmo
    as.readSettings(((PPAlgorithmSettings) settings).alg_settings);

    // lettura delle statistiche
    as.readStatistics(statistic);

    // normalizzazione degli attributi
    description = as.getDescription();

    // importa soltanto lo schema delle istanze...
    Instances[] inst = ((Instances[]) input.load(0));

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = new PPWekaFactory(dest, inst[0], inst[1]);
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();

    long opt = input.optimize();
    Instances tuple, metatuple;
    Instances new_tuple, new_metatuple;

    while (input.hasNext()) {

      // importazione delle istanze
      Instances[] inst2 = ((Instances[]) input.load(opt));
      tuple = inst2[0];
      metatuple = inst2[1];
      new_tuple = new Instances(tuple);
      new_metatuple = new Instances(metatuple);
      new_tuple.delete();
      new_metatuple.delete();

      // ciclo che scorre le tuple
      for (int i=0; i<tuple.numInstances(); i++) {
        // instance contiene la nuova istanza
        Instance[] instance = execute(tuple.instance(i), metatuple.instance(i));
        if (instance != null) {
          new_tuple.add(instance[0]);
          new_metatuple.add(instance[1]);
        }
      }

      output.addInstances(new_tuple, new_metatuple);
    }

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }


  protected Instance[] execute(Instance tupla, Instance metatupla)
      throws KDDMLOperatorException, KDDMLCoreException {

    Instance ris[] = new Instance[2];
    Instance tupla2 = new Instance(tupla);
    tupla2.setDataset(tupla.dataset());
    Instance metatupla2 = new Instance(metatupla);
    metatupla2.setDataset(metatupla.dataset());

    for (int i = 0; i < attributes_index.length; i++) {
      if (tupla.isMissing(attributes_index[i]))

        // set missing value
        tupla2.setMissing(attributes_index[i]);
      else {
        double val = as.normalize( (String) enum2.get(i),
                                   tupla.value(attributes_index[i]));

        // set the normalized value
        tupla2.setValue(attributes_index[i], val);
      }
    }
    // else do nothing

    ris[0] = tupla2;
    ris[1] = metatupla2;
    return ris;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE;
    if (i==1)
      return kddml.Core.KDDMLObjectType.ALGORITHM;
    return null;
  }

  protected String createDescription()  {
    description += "Attributes: "+((PP_NORMALIZATION_SETTINGS) settings).getAttributes();;
    return description;
  }

  private int[] checkAttributes(Vector enum2) throws KDDMLOperatorException {

    int[] ris = new int[enum2.size()];

    for (int i=0; i<enum2.size(); i++) {
      String attribute_name = (String) enum2.get(i);
      int index = statistic.getAttributeIndex(attribute_name);
      // controlla che l'attributo sia presente nella tabella
      if (index == -1) {
        // genera eccezione se un attributo non viene individuato
        String msg = "The attribute < " + attribute_name +" > do not exist.";

        throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
      }

      AttributeStatisticManager a = statistic.getStatistic(attribute_name);
      if (!a.getType().equals(AttributeType.NUMERIC)) {
        // genera eccezione se un attributo non viene individuato
        String msg = "The attribute < " + attribute_name +" > is not numeric!";

        throw new KDDMLOperatorException(msg, null, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
      }

      ris[i] = index;
    }

    return ris;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

}
