/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.Vector;
import weka.core.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import kddml.Operators.*;
import kddml.Operators.Preprocessing.DiscretizationAlgorithms.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description:
 * Sottoclasse di INSTANCE_LEVEL_Dipendent_Transformation.
 * Discretizza i valori di un attributo specificato nel dataset di input.
 * La discretizzazione pu essere effettuata :
 * -- su dati numerici,
 * -- su dati categorici.
 *
 * La discretizzazione numerica :
 *
 * consiste nel suddividere i valori assunti
 * dall'attributo in un insieme di intervalli,
 * quindi si sostiutuisce ciascun valore assunto con
 * l'etichetta del suo intervallo di appartenenza.
 * Le due varianti sono:
 * ---Natural Binning,
 * ---Equal Frequency Binning.
 *
 * Nel Natural Binning la suddivisione mira a ottenere intervalli della stessa ampiezza.
 * (es [1..3], [4..6], [7..9], etc...)
 * Nel Equal Frequency Binning la suddivisione mira a ottenere intervalli nei quali ricade
 * la stessa quantit dei valori originali in termini di numero di elementi.
 *
 * L'etichetta da associare a ciascun intervallo ottenuto sar:
 * -- la media dei valori compresi nell'intervallo,
 * -- la mediana dei valori calcolata sui quei valori originali che ricadono nell'intervallo,
 * -- l'estremo superiore dell'intervallo,
 * -- l'estremo inferiore dell'intervallo,
 * -- scelta tra le etichette specificate dall'utente
 *    (ciascuna etichetta  associata all'intervallo
 *     seguendo la corrispondenza tra gli ordinamenti delle etichette fornite
 *     e degli interrvalli ottenuti).
 *
 *
 *
 * La discretizzazione categorica:
 *
 * consiste nel ridurre il numero di valori che l'attributo pu assumere.
 *
 * per i valori categorici non esiste un ordinamento pertanto  possibile:
 * -- usare una tabella con cui ottenere  un ordinamento sui possibili valori assunti, e successivamente
 *     applicare  una discretizzazione numerica basata su ordinamento;
 * -- usare una tabella in cui  specificata la gerarchia e sostituire i valori originali con
 *    il gruppo di appartenenza;
 * -- usare l'ordine indotto dalla scala nominale SOLO per gli attributi nominali.
</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */
public class PP_NUMERIC_DISCRETIZATION_RESOLVER extends InstanceLevelIndependentTransformation {

  //private KDDMLOperatorSettings settings;

  private DataStatisticsManager statistic;

  private String description;

  private String attribute_name;

  private int attribute_position;

  private Double[] values_to_discretize;

  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    //this.settings = settings;
    statistic = ((PPInternalTable) arguments.get(0)).getStatistic();

    attribute_name = ((PP_NUMERIC_DISCRETIZATION_SETTINGS) settings).getAttributeName();

    // controlla la presenza dell'attributo di discretizzatione
    if (statistic.getAttributeIndex(attribute_name) == -1) {
      // genera eccezione se un attributo non viene individuato
      String msg = "The attribute < " + attribute_name + " > do not exist.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    // controlla che l'attributo di discretizzation sia di tipo numerico
    AttributeStatisticManager attr = statistic.getStatistic(attribute_name);
    if (!attr.getType().equals(AttributeType.NUMERIC)) {
      String msg = "The attribute < " + attribute_name + " > in not numeric. Can't discretize.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
    }

    attribute_position = statistic.getAttributeIndex(attribute_name);

    // effettua una prima scansione dei dati per ottenere i valori dell'attributo di discretizzatione
    PPInternalTable input = (PPInternalTable) arguments.get(0);
    long opt = input.optimize();
    values_to_discretize = new Double[statistic.getNumInstances()];
    int index = 0;

    while (input.hasNext()) {

      // importazione delle istanze
      Vector values = getValuesToDiscretize(((Instances[]) input.load(opt))[0]);

      for (int i=0; i<values.size(); i++) {
        values_to_discretize[index] = (Double) values.get(i);
        index++;
      }
    }
    input.reset();



  }

  /**
   * Overloading of the main execute method.
   *
   * @param settings KDDMLOperatorSettings
   * @param arguments Vector
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // test if the algorithm is supported...
    String alg_name = ((PPAlgorithmSettings) settings).algorithm_name;

    DiscretizationAlgorithmFactory factory = new DiscretizationAlgorithmFactory(alg_name);
    factory.initialize();

    if (!factory.isSupported()) {
      throw new KDDMLOperatorException
          ("Algorithm < " + alg_name + " > not supported!", null,
           KDDMLOperatorErrorCodes.UNSUPPORTED_ALGORITHM);
    }

    // algoritmo di preprocessing individuato...
    DiscretizationAlgorithmResolverTask as =
        (DiscretizationAlgorithmResolverTask) factory.newResolverInstance();

    // lettura dei settaggi dell'algoritmo
    as.readSettings(((PPAlgorithmSettings) settings).alg_settings);

    // lettura delle statistiche
    as.readStatistics(statistic);

    // lettura della description
    description = as.getDescription();

    // discretizzazione dei valori
    Object[] discretized = as.discretize(values_to_discretize);
    Vector categories = null;
    if (!as.isNumericOutput())
      // ottiene le categorie per l'attributo nominale
      categories = getCategories(discretized);

    // importa soltanto lo schema delle istanze...
    Instances[] inst = ((Instances[]) input.load(0));

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = null;
    if (as.isNumericOutput())
      // risultato di tipo numerico -> nessuna conversione di tipo
      fact = new PPWekaFactory(dest, inst[0], inst[1]);
    else {
      // risultato di tipo enumerator -> converto il tipo dell'attributo
      numericToNominal(inst[0], categories.elements());
      fact = new PPWekaFactory(dest, inst[0], inst[1]);
    }
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();

    long opt = input.optimize();
    Instances tuple, metatuple;

    while (input.hasNext()) {

      // importazione delle istanze
      Instances[] inst2 = ((Instances[]) input.load(opt));
      tuple = inst2[0];
      metatuple = inst2[1];

      if (!as.isNumericOutput()) {
        numericToNominal(tuple, categories.elements());
      }

      // ciclo che scorre le tuple
      for (int i=0; i<tuple.numInstances(); i++) {
        // instance contiene la nuova istanza
        if (as.isNumericOutput())
          execute(tuple.instance(i), metatuple.instance(i), (Double)discretized[i]);
        else
          execute(tuple.instance(i), metatuple.instance(i), (String)discretized[i]);
      }

      output.addInstances(tuple, metatuple);
    }

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }

  protected Instance[] execute(Instance tupla, Instance metatupla, Double discretized_value)
      throws KDDMLOperatorException, KDDMLCoreException {

    Instance ris[] = new Instance[2];
    Instance tupla2 = new Instance(tupla);
    tupla2.setDataset(tupla.dataset());
    Instance metatupla2 = new Instance(metatupla);
    metatupla2.setDataset(metatupla.dataset());

    if (discretized_value == null)
      tupla.setMissing(attribute_position);
    else
      tupla.setValue(attribute_position, discretized_value.doubleValue());

    ris[0] = tupla2;
    ris[1] = metatupla2;
    return ris;
  }

  protected Instance[] execute(Instance tupla, Instance metatupla, String discretized_value)
      throws KDDMLOperatorException, KDDMLCoreException {

    Instance ris[] = new Instance[2];
    Instance tupla2 = new Instance(tupla);
    tupla2.setDataset(tupla.dataset());
    Instance metatupla2 = new Instance(metatupla);
    metatupla2.setDataset(metatupla.dataset());

    if (discretized_value == null)
      tupla.setMissing(attribute_position);
    else
      tupla.setValue(attribute_position, discretized_value);

    ris[0] = tupla2;
    ris[1] = metatupla2;
    return ris;
  }

  protected Instance[] execute(Instance tupla, Instance metatupla)
      throws KDDMLOperatorException, KDDMLCoreException {

    return null;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE;
    if (i==1)
      return kddml.Core.KDDMLObjectType.ALGORITHM;
    return null;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }

  protected String createDescription()  {
    return description;
  }

  /**
   * Filtra le tuple in ingresso con indice non contenuto in index.
   *
   * @param index int[] gli indici delle istanze campionate
   * @param tuple Instances
   * @return Instances
   */
  private Vector getValuesToDiscretize(Instances tuple) {

    Vector ris = new Vector();
    int index = tuple.attribute(attribute_name).index();
    for (int i=0; i<tuple.numInstances(); i++) {
      if (tuple.instance(i).isMissing(index)) {
        Double d = null;
        ris.add(d);
      }
      else
        ris.add(new Double(tuple.instance(i).value(index)));
    }
    return ris;
  }


  private static Vector getCategories(Object[] obj) {
    Vector categories = new Vector();
    for (int i=0; i<obj.length; i++)
      if ((obj[i] != null) && (!categories.contains((String) obj[i])))
        categories.add((String) obj[i]);
    return categories;
  }

  /**
   * Converte l'attributo di discretizzazione da numerico a enumerato.
   *
   * @param inst Instances lo schmea delle istanze
   * @param obj Object[]
   */
  private void numericToNominal(Instances inst, java.util.Enumeration categories)
      throws KDDMLOperatorException {

    int index = inst.attribute(attribute_name).index();
    // aggiunge un attributo di appoggio in cui copiare i valori dell'attributo da castare
    PreprocessingUtility.addNewAttribute(inst, attribute_name, AttributeType.NOMINAL, index+1, categories);

    inst.deleteAttributeAt(index);
  }

}
