/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import weka.core.*;
import kddml.Operators.KDDMLOperatorException;
import kddml.Operators.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import java.util.Vector;
import java.util.Hashtable;

public class PP_NUMERIC_LABELING_RESOLVER extends SchemaLevelTransformation {

  private String attribute_name;

  private Vector mapping;

  private String mark;

  /**
   * Se true utilizza il ranking espresso dal valore nominale. Se false
   * utilizza il rank dato in input dall'utente.
   */
  private boolean use_rank_order;

  public void readInput(KDDMLOperatorSettings settings,
                        java.util.Vector arguments) throws KDDMLCoreException,
      KDDMLOperatorException {

    attribute_name = ((PP_NUMERIC_LABELING_SETTINGS) settings).getAttributeName();
    mapping = ((PP_NUMERIC_LABELING_SETTINGS) settings).getMapping();
    if (mapping.size() == 0)
      use_rank_order = true;
    else
      use_rank_order = false;

    // legge lo schema dei dati per controllare la presenza degli attributi
    PPInternalTable t = (PPInternalTable) arguments.get(0);
    DataStatisticsManager stat = t.getStatistic();

    if (stat.getAttributeIndex(attribute_name) == -1) {
      // genera eccezione se un attributo non viene individuato
      String msg = "The attribute < " + attribute_name + " > do not exist.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    // controlla che l'attributo in input sia di tipo nominale
    AttributeStatisticManager a_s = stat.getStatistic(attribute_name);
    if (!a_s.getType().equals(AttributeType.NOMINAL)) {
      // genera eccezione se un attributo non viene individuato
      String msg = "The attribute < " + attribute_name + " > must be nominal.";
      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
    }
    if (!use_rank_order)
      if (((NominalStatisticManager) a_s).getNumberOfValues() != mapping.size()) {
      // genera eccezione se il numero di labels fornito e' diverso dal numero di categorie
      String msg = "The number of numeric labels is not equal to the number of categories for < " + attribute_name + " >.";
      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_TYPE_ERROR);
    }

  }

  protected String createDescription() {
    return "Numeric Labeling on attribute: "+attribute_name;
  }

  protected Instances[] execute(Instances tuple, Instances metatuple) throws
      KDDMLCoreException, KDDMLOperatorException {

    //creo un attributo numerico d'appoggio in cui inserire i valori numerici corrispondenti
    //a ciascun valore dell'attributo
    int pos = tuple.attribute(attribute_name).index();
    tuple.insertAttributeAt(new Attribute(attribute_name), pos+1);

    //if (use_rank_order) {
      //setto i valori dell'attributo d'appoggio con i valori numerici corrispondenti
      setWithHisNominalOrder(tuple, pos + 1);
      tuple.deleteAttributeAt(pos);
    //}
   // else {
      // ordine delle categorie dato in ingresso
    //  setWithHashTable(tuple, metatuple, pos + 1);
    //  tuple.deleteAttributeAt(pos);
    //}

    Instances[] ris = new Instances[2];
    ris[0] = tuple;
    ris[1] = metatuple;
    return ris;
  }

  /**
   * Setta i valori dell'attributo in posizione "posizione_dest "
   * con l'ordine numerico indotto dall'ordinamento che i valori nominali
   * hanno nell'attributo nominale da discretizzare
   *
   * @param posizione_dest indice dell'attributo d'appoggio per la discretizzazione
   */
  private void setWithHisNominalOrder(Instances tuple, int posizione_dest) {

    int index = tuple.attribute(attribute_name).index();
    for (int i = 0; i < tuple.numInstances(); i++) {
      double order;
      if (use_rank_order)
        order = tuple.instance(i).value(index);
      else
        order = ((Double) mapping.get((int) tuple.instance(i).value(index))).doubleValue();
      tuple.instance(i).setValue(posizione_dest, order);
    }
  }

  /**
   * Setta i valori dell'attributo in posizione "posizione_dest "
   * utilizzando la tabella hash.
   *
   * @param posizione_dest indice dell'attributo d'appoggio per la discretizzazione
   */
 /* private void setWithHashTable(Instances tuple, Instances metatuple, int posizione_dest)
      throws KDDMLOperatorException {

    int index = tuple.attribute(attribute_name).index();
    for (int i = 0; i < tuple.numInstances(); i++) {
      String category = tuple.instance(i).stringValue(index);
      if (mapping.containsKey(category)) {
        Double d = (Double) mapping.get(category);
        tuple.instance(i).setValue(posizione_dest, d.doubleValue());
      }
      else {
        // category not found: set the metadata with exception
        tuple.instance(i).setMissing(posizione_dest);
        PreprocessingUtility.addExceptionInAttribute(index, mark, metatuple.instance(i));
      }
    }
  }*/


  /**
 * Verifica se i valori assunti dall'attributo da specificare siano tutti presenti
 * nella tabella da utilizzare per ricavarne l'ordine numerico oppure la gerarchia
 * di appartenenza.
 * Setta i valori de nuovo attributo d'appoggio , creato in posizione successiva a
 * quella dell'attributo da discretizzare, con i valori corispondenti nella tabella.
 *
 * @param pos indice della colonna dell'attributo da discretizzare
 * @param newpos indice della colonna dell'attributo d'appoggio
 * @param mark stringa relativa la codice di marcatura
 * @return esito
 * @throws KDDMLInterpreterException
 * @throws KDDMLCoreException
 */
/*private boolean isInSpecifyTableAndSet(int pos,
                                       int newpos,
                                       String mark
                                       ) throws KDDMLInterpreterException,
    KDDMLCoreException {

  Instances tuple_ordering = (Instances)input_table_ordering.getInstances();
  int num_tuple_table_ordering = tuple_ordering.numInstances();
  boolean has_matches = true;
  int start = 0;

  int i = 0;
  while ( (i < num_tuple) && (has_matches)) {
    int j = 0;
    if (!tuple.instance(i).isMissing(pos)) {
     // System.out.println("tuple_ordering.instance(j) " + j + " valore " +
                        // tuple_ordering.instance(j).stringValue(0));
      while ( (j < num_tuple_table_ordering) &&
             (! (tuple.instance(i).stringValue(pos).
                 equals(tuple_ordering.instance(j).stringValue(0))))) {
        j++;
      }

      if (j < num_tuple_table_ordering) {
        if (tuple.attribute(newpos).isNumeric()) {
          //se sto settando il valore relativo all'ordine numerico
          //corrispondente all'attributo
          tuple.instance(i).setValue(newpos,
                                     tuple_ordering.instance(j).value(1));
        }
        else {
          //se invece sto settando il valore relativo alla sua gerarchia di appartenenza
          tuple.instance(i).setValue(newpos,
                                     tuple_ordering.instance(j).stringValue(1));
       //   System.out.println("setValue hierarchy " +
                           //  tuple_ordering.instance(j).stringValue(1));
        }
      }
      else { //non ho la gerarchia per l'istanza i-esima

      //  System.out.println("non ho la gerarchia per l'istanza i-esima " + i);
       // System.out.println("non ho la gerarchia per " +
                         //  tuple.instance(i).stringValue(pos));

       // System.out.println("mark .......... " + mark);

        PreprocessingUtility.addStringValueInAttribute(pos, mark, metatuple,
            i);

      }
    }
    i++;
  }

  //se  almeno un valore assunto dall'attributo, non esiste
  //nella tabella di riferimento per gli ordini
  if (!has_matches) {
    //elimino l'attributo d'appoggio che ho creato
  //  tuple.deleteAttributeAt(newpos);
    notify = "no such case for hierarchy";
  }
//  else {
    //altrimenti se ho tutti i valori assunti dall'attributo nella tabella
    //elimino il vecchio attributo e tengo il nuovo per portare
    //a buon fine la discretizzazione
    tuple.deleteAttributeAt(pos);

//  }
  return has_matches;

}
*/

}
