/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.*;
import kddml.Core.*;
import kddml.Operators.*;
import kddml.Core.DataSources.*;
import kddml.Operators.Preprocessing.RewritingAlgorithms.*;
import weka.core.*;
import kddml.Core.ProprietaryElement.Condition.*;

/**
 * <p>Title: KDDML</p>
 * <p>Description: Classe che risolve l'operatore di riscrittura.
 * Tramite questo operatore e' possibile ricercare, marcare e sostituire dei valori di un
 * attributo. Le regole di riscrittura previste sono:
 *
 * --RULE_S
 * La regola di riscrittura principale, ruleS, permette di sostituire tutte le stringhe
 * che rappresentano il valore di un attributo di una tabella estesa.
 * Lapplicazione della regola di riscrittura consiste nelleffettuare un pattern matching tra lespressione regolare
 * della regola, e le stringhe rappresentanti lattributo in questione.
 * Se esso dovesse fallire per qualche tupla, la sostituzione non avviene, e nel caso in cui fosse
 * specificato il tag EXCEPTION, la tupla viene marcata, ossia estesa con le informazioni sullavvenuta eccezione.
 *
 * --RULE_T
 * Ha un comportamento simile a quello di  rule_s ,
 * tuttavia in questo caso il pattern non si riferisce allintero contenuto
 * dellattributo. Il pattern matching viene effettuato ripetutamente e la stringa di sostituzione andr a rimpiazzare tutte le
 *  occorrenze del pattern incontrate nella stringa rappresentante il valore attuale dellattributo.
 * Se  specificato il tag EXCEPTION e nessuna occorrenza del pattern dovesse essere trovata allinterno di un attributo,
 * allora nellestensione della tabella la tupla verr estesa con le informazioni di marcatura/eccezione
 *
 *
 * --RULE_M
 * Essa non effettua alcuna modifica sui valori dellattributo in questione, ma serve solamente per verificare che sia rispettato
 * un dato schema di costruzione delle stringhe, chiamato pattern. Ogni qual volta il pattern matching dovesse fallire, la tupla viene
 * marcata e leccezione sollevata viene memorizzata nellestensione della tabella.
 *
 * Si pu allora controllare lapplicazione di tale funzione locale  attraverso una condizione
* di restrizione sulla tupla. Tale condizione, come nella clausola WHERE nelle interrogazioni SQL,
*   rappresentata da unespressione booleana formata da operatori di confronto applicati ad attributi
* e costanti, dalla composizione logica di espressioni booleane.
* La condizione di restrizione, rappresentata dalla clausola WHERE (condizione), va semplicemente aggiunta alla fine del costrutto.
*
* </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */

import kddml.Operators.Preprocessing.RewritingAlgorithms.RewritingAlgorithmResolverTask;

public class PP_REWRITING_RESOLVER extends InstanceLevelIndependentTransformation {

  /**
   * The list of settings for each algorithm.
   */
  private RewritingAlgorithmResolverTask[] algorithms;

  private DataStatisticsManager statistic;

  private String description = "";

  /**
   * The list of conditions.
   */
  private Condition[] cond;

  private String attribute_name;

  private int attribute_position;


  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException, KDDMLOperatorException {

    // get arguments...
    AlgorithmSettings[] alg_settings = ((PP_REWRITING_SETTINGS) settings).getSettings();
    statistic = ((PPInternalTableManager) arguments.get(0)).getStatistic();

    // controlla la presenza dell'attributo di riscrittura
    attribute_name = ((PP_REWRITING_SETTINGS) settings).getAttributeName();
    if (statistic.getAttributeIndex(attribute_name) == -1) {
      // genera eccezione se un attributo non viene individuato
      String msg = "The attribute < " + attribute_name + " > do not exist.";

      throw new KDDMLOperatorException(msg, msg, KDDMLOperatorErrorCodes.ATTRIBUTE_DO_NOT_EXIST);
    }

    attribute_position = statistic.getAttributeIndex(attribute_name);

    algorithms = new RewritingAlgorithmResolverTask[alg_settings.length];
    // test if the algorithms are supported...
    for (int k=0; k<alg_settings.length; k++) {
      String alg_name = alg_settings[k].algorithm_name;
      System.out.println("Alg Name: "+alg_name);
      RewritingAlgorithmFactory factory = new RewritingAlgorithmFactory(alg_name);
      factory.initialize();
      if (!factory.isSupported()) {
        throw new KDDMLOperatorException
            ("Algorithm < " + alg_name + " > not supported!", null,
             KDDMLOperatorErrorCodes.UNSUPPORTED_ALGORITHM);
      }

      // algoritmo di riscrittura individuato...
      RewritingAlgorithmResolverTask as = (RewritingAlgorithmResolverTask) factory.newResolverInstance();
      algorithms[k] = as;
    }

    // get the conditions...
    int i=0;
    int k=2;
    cond = new Condition[(int) arguments.size() / 2];
    while(k < arguments.size()) {
      Condition c = (Condition) arguments.get(k);
      cond[i] = c;
      i++;
      k = k + 2;
    }
  }

  /**
   * Overloading of the main execute method.
   *
   * @param settings KDDMLOperatorSettings
   * @param arguments Vector
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return KDDMLObject
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    java.io.File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // per ogni algoritmo di riscrittura...
    for (int i=0; i<algorithms.length; i++) {
      // get the i-esim settings for the algorithm
      RewritingAlgorithmResolverTask as = (RewritingAlgorithmResolverTask) algorithms[i];
      // lettura dei settaggi dell'algoritmo
      as.readSettings(((PP_REWRITING_SETTINGS) settings).getSettings()[i]);

      // lettura delle statistiche
      as.readStatistics(statistic);

      // lettura della description
      description += as.getDescription();
    }

    // importa soltanto lo schema delle istanze...
    Instances[] inst = ((Instances[]) input.load(0));

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = new PPWekaFactory(dest, inst[0], inst[1]);
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();

    long opt = input.optimize();
    Instances tuple, metatuple;
    Instances new_tuple, new_metatuple;

    while (input.hasNext()) {

      // importazione delle istanze
      Instances[] inst2 = ((Instances[]) input.load(opt));
      tuple = inst2[0];
      metatuple = inst2[1];
      new_tuple = new Instances(tuple);
      new_metatuple = new Instances(metatuple);
      new_tuple.delete();
      new_metatuple.delete();

      // ciclo che scorre le tuple
      for (int i=0; i<tuple.numInstances(); i++) {
        // instance contiene la nuova istanza
        Instance[] instance = execute(tuple.instance(i), metatuple.instance(i));
        if (instance != null) {
          new_tuple.add(instance[0]);
          new_metatuple.add(instance[1]);
        }
      }

      output.addInstances(new_tuple, new_metatuple);
    }

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }


  protected Instance[] execute(Instance tupla, Instance metatupla)
      throws KDDMLOperatorException, KDDMLCoreException {

    String value, value_m;

    if (tupla.isMissing(attribute_position))
      value = "?";
    else
      if (tupla.attribute(attribute_position).isNumeric())
        value = new Double(tupla.value(attribute_position)).toString();
      else
        value = tupla.stringValue(attribute_position);
    if (metatupla.isMissing(attribute_position))
      value_m = "?";
    else
      value_m = metatupla.stringValue(attribute_position);

    // per ogni algoritmo di riscrittura...
    for (int i=0; i<algorithms.length; i++) {

      // get the i-esim settings for the algorithm
      RewritingAlgorithmResolverTask as = (RewritingAlgorithmResolverTask)
          algorithms[i];

      // get the i-esim condition...
      Condition condition = cond[i];
      // valuta la condizione
      boolean esito = as.evalCondition(condition, tupla, metatupla);

      // esegue la riscrittura
      String[] result = as.rewrite(value, value_m, esito);
      value = result[0];
      value_m = result[1];

      //}

      if (value.equals("?"))
        tupla.setMissing(attribute_position);
      else {
        if (tupla.attribute(attribute_position).isNumeric())
          try {
            tupla.setValue(attribute_position, new Double(value).doubleValue());
          }
          catch (Exception e) {
            // genera eccezione se il risultato non e' un double
            String msg = "Can't convert " + value + " to double. Error during rewriting!";
            throw new KDDMLOperatorException(msg, msg,
                                             KDDMLOperatorErrorCodes.
                                             ILLEGAL_CASTING);
          }
        else
          try {
            tupla.setValue(attribute_position, value);
          }
          catch (Exception e) {

            // genera eccezione se il risultato non appartiene alla lista dei valori nominali
            String msg = "Can't convert " + value +
                " to a compatible nominal value. Error during rewriting!";
            throw new KDDMLOperatorException(msg, msg,
                                             KDDMLOperatorErrorCodes.
                                             ILLEGAL_CASTING);

          }
      }

      if (value_m.equals("?"))
        metatupla.setMissing(attribute_position);
      else
        metatupla.setValue(attribute_position, value_m);
    }


    Instance[] ris = new Instance[2];
    ris[0] = tupla;
    ris[1] = metatupla;
    return ris;
  }

  protected  String createDescription () {
    String ris =  "Rewriting on attribute "+attribute_name+". ";
    ris = ris + description;
    return ris;
  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE;
    // in posizione dispari le specifiche degli algoritmi
    if ((i%2) != 0)
      return kddml.Core.KDDMLObjectType.ALGORITHM;
    // in posizione pari le specifiche delle condizioni
    else
      return kddml.Core.KDDMLObjectType.CONDITION;
  }

  /**
   * The estimated RAM usage for the operator.
   * Must be a value between 0 and 1024 MB.
   *
   * @return boolean
   */
  public int estimatedRAMUsage() {
    return 512;
  }
}
