/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.Preprocessing;

import java.util.Hashtable;
import kddml.Operators.*;
import java.util.*;
import kddml.Core.KDDMLScalarManager;
import kddml.Core.KDDMLScalar;

/**
 * <p>Title: KDDML</p>
 * <p>Description: Knowledge Discovery in Database Environment</p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@di.unipi.it)
 * @version 1.0
 */
public class PP_REWRITING_SETTINGS extends PPAlgorithmSettings {

  private final static String ATTRIBUTE_NAME_ATTRIBUTE = "attribute_name";

  protected String attribute;

  protected String regular_expression;

  protected String substitution_then;

  protected AlgorithmSettings[] alg_settings;

  public PP_REWRITING_SETTINGS() throws KDDMLOperatorException {

    super();
    attribute = null;
    regular_expression = null;
    substitution_then = null;
  }

  public void applySettings(Hashtable attributes)
      throws KDDMLOperatorException {

    setXMLDestination(attributes, "xml_dest", true);
    setAttributeName(attributes, ATTRIBUTE_NAME_ATTRIBUTE, true);

    // hashtables contiene le varie tabelle hash relative ad ogni singolo algoritmo
    Vector hashtables = splitHashtable(attributes);

    // ottiene i vari settings associati ad ogni algoritmo di riscrittura
    alg_settings = new AlgorithmSettings[hashtables.size()];
    for(int i=0; i<hashtables.size(); i++) {

      Hashtable ht = (Hashtable) hashtables.get(i);
      System.out.println("HT "+i+": "+ht.toString());
      String alg_name = null;
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) ht.get("algorithm_name");
        alg_name = (String) s.getExprValue();
      }
      catch(Exception e) {
        throw new KDDMLOperatorException
            ("algorithm_name is required!", null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }

      AlgorithmFactory factory = new AlgorithmFactory(alg_name);
      if (!factory.isSupported()) {
        throw new KDDMLOperatorException
            ("Algorithm < " + algorithm_name + " > not supported!", null,
             KDDMLOperatorErrorCodes.UNSUPPORTED_ALGORITHM);
      }

      // settings per l'algoritmo di riscrittura...
      AlgorithmSettings alg_settings = (AlgorithmSettings) factory.newSettingsInstance();
      // setta l'xml_dest...
      alg_settings.setXMLDestination(this.xml_dest);

      // apply che algorithm settings
      ((AlgorithmSettingsTask) alg_settings).applyAlgorithmSettings(ht);
      // setta il nome dell'algoritmo
      alg_settings.setAlgorithmName(alg_name);

      this.alg_settings[i] = alg_settings;
    }
  }

  public AlgorithmSettings[] getSettings() {
    return alg_settings;
  }

  protected void setAttributeName(Hashtable attributes,
                                  String attribute_name,
                                  boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {

        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        this.attribute = (String) s.getExprValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" +
             " must be a string.", null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  public String getAttributeName() {
    return this.attribute;
  }

  public void setAttributeName(String name) {
    this.attribute = name;
  }

  protected void setRegularExpression(Hashtable attributes,
                                      String attribute_name,
                                      boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {

        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        this.regular_expression = (String) s.getExprValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" +
             " must be a string.", null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  public String getRegularExpression() {
    return this.regular_expression;
  }

  public void setRegularException(String re) {
    this.regular_expression = re;
  }

  protected void setSubstitutionThen(Hashtable attributes,
                                     String attribute_name,
                                     boolean error_if_not_found)
      throws KDDMLOperatorException {

    if (existAttribute(attributes, attribute_name, error_if_not_found)) {
      try {
        KDDMLScalarManager s = (KDDMLScalarManager) attributes.get(attribute_name);
        if (s.getExprType() == KDDMLScalar.NUMERIC)
          this.substitution_then = ((Double) s.getExprValue()).toString();
        else
          this.substitution_then = (String) s.getExprValue();
      }
      catch (Exception e) {
        throw new KDDMLOperatorException
            ("< " + attribute_name + " >" +
             " must be a string.", null, KDDMLOperatorErrorCodes.INVALID_SETTINGS);
      }
    }
  }

  public String getSubstitutionThen() {
    return this.substitution_then;
  }

  public void setSubstitutionThen(String substitution) {
    this.substitution_then = substitution;
  }

  /**
   * Splitta la tabella hash in input in base ai nomi degli algoritmi.
   *
   * @param attributes Hashtable
   * @return Vector a vector of hashtables
   */
  private static Vector splitHashtable(Hashtable attributes) {

    // vector of hashtable
    Vector v = new Vector();
    // supporta al max 9 algoritmi
    final int max_number_of_algorithm = 9;
    for (int i=1; i<max_number_of_algorithm; i++) {
      Enumeration enum2 = attributes.keys();
      Hashtable ht = new Hashtable();
      while(enum2.hasMoreElements()) {
        String key = (String) enum2.nextElement();
        if (key.endsWith("algorithm_name_"+i)) {
          // costruisce l'iesima tabella hash
          KDDMLScalarManager value = (KDDMLScalarManager) attributes.get(key);
          String new_key;
          if (key.startsWith("algorithm_name"))
            new_key = "algorithm_name";
          else
            new_key = key.substring(0, key.length()-17);
          ht.put(new_key, value);
          attributes.remove(key);
        }
      }
      // ht contiene la tabella hash splittata...
      if (!ht.isEmpty())
        v.add(ht);
    }
    // il rimante di attributes contiene la tabella hash che va inserita in
    // prima posizione del vettore
    v.add(0, attributes);

    // aggiunge ad ogni tabella attribute_name e xml_dest...
    for (int j=0; j<v.size(); j++) {
      KDDMLScalarManager s1 = (KDDMLScalarManager) attributes.get("xml_dest");
      KDDMLScalarManager s2 = (KDDMLScalarManager) attributes.get("attribute_name");
      ((Hashtable) v.get(j)).put("xml_dest", s1);
      ((Hashtable) v.get(j)).put("attribute_name", s2);
    }
    return v;
  }

  protected String getAlgorithmName(int i) {
    AlgorithmSettings s = alg_settings[i];
   // KDDMLScalarManager sc = (KDDMLScalarManager) s.get("algorithm_name");
   // this.substitution_then = (String) sc.getExprValue();
    return s.algorithm_name;
  }
}
