/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.Preprocessing;

import java.util.Vector;
import weka.core.*;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import kddml.Operators.*;
import kddml.Operators.Preprocessing.SamplingAlgorithms.*;

/**
 *
 * <p>Title: KDDML</p>
 * <p>Description: Classe che risolve l'operatore di campionamento.
 * Il campionamento e' una trasformazione dipendente perche' necessita
 * di tutte le informazioni sui dati.
 * Le modalita' di campionamento previste sono:
 *
 * --  campionamento semplice;
 * --  campionamento stratificato;
 * --  campionamento su clusters.
 *
 * Il campionamento semplice prevede di restituire un insieme di istanze
 * (con o senza rimpiazzo), di cardinalita' fissata,  a partire da quelle iniziali.
 *
 * Il campionamento stratificato costruisce  un insieme composto da una quantita'
 * fissata di istanze, per ciascun valore assunto da un attributo nominale specificato.
 *
 * Il campionamento su clusters restituisce un insieme composto da una cardinalita' fissata
 * di sottoinsiemi, ciascuno contenente tutte le istanze che hanno lo stesso valore di un
 * attributo fissato. Come caso particolare , se si fissa come attributo il cluster di appartenenza
 * e si chiede di campionare in modo da avere items="n" , si ottiene un insieme contenente tutti gli elementi
 * di "n" clusters.
 *
 * </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */

public class PP_SAMPLING_RESOLVER extends InstanceLevelDependentTransformation {

  private KDDMLOperatorSettings settings;

  private DataStatisticsManager statistic;

  private String description;

  public void readInput(KDDMLOperatorSettings settings, java.util.Vector arguments)
      throws KDDMLCoreException {

    this.settings = settings;
    statistic = ((PPInternalTableManager) arguments.get(0)).getStatistic();
  }

  /**
   * Execute tipico per un operatore di sampling.
   *
   * @param settings KDDMLOperatorSettings
   * @param arguments Vector
   * @return KDDMLObject
   */
  protected Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLOperatorException, KDDMLCoreException {

    String alg_name = ((PPAlgorithmSettings) settings).algorithm_name;
    SamplingAlgorithmFactory factory = new SamplingAlgorithmFactory(alg_name);
    factory.initialize();
    if (!factory.isSupported()) {
      throw new KDDMLOperatorException
          ("Algorithm < " + alg_name + " > not supported!", null,
           KDDMLOperatorErrorCodes.UNSUPPORTED_ALGORITHM);
    }

    // algoritmo di preprocessing individuato...
    SamplingAlgorithmResolverTask as = (SamplingAlgorithmResolverTask) factory.newResolverInstance();

    int[] index;

    // lettura dei settaggi dell'algoritmo
    as.readSettings(((PPAlgorithmSettings) settings).alg_settings);

    // lettura delle statistiche
    as.readStatistics(statistic);

    // campiona l'insieme di istanze ottenento l'insieme di indici come output
    index = as.campionate(tuple);

    description = as.getDescription();

    Instances ris[] = new Instances[2];
    ris[0] = filterTuple(index, tuple);
    ris[1] = filterTuple(index, metatuple);
    return ris;

  }

  public kddml.Core.KDDMLObjectType getParamType(int i) {
    if (i==0)
      return kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE;
    if (i==1)
      return kddml.Core.KDDMLObjectType.ALGORITHM;
    return null;
  }

  protected String createDescription()  {
    return description;
  }

  /**
   * Filtra le tuple in ingresso con indice non contenuto in index.
   *
   * @param index int[] gli indici delle istanze campionate
   * @param tuple Instances
   * @return Instances
   */
  private static Instances filterTuple(int[] index, Instances tuple) {

    Instances ris = new Instances(tuple, 0);
    Instance tupla;
    for (int i=0; i<index.length; i++) {
      tupla = tuple.instance(index[i]);
      tupla.setDataset(ris);
      ris.add(tupla);
    }
    return ris;
  }
}
