/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */

package kddml.Operators.Preprocessing;


/**
 * Classe di utilita' per gli operatori di preprocessing.
 */
import java.util.Enumeration;
import weka.core.*;
import kddml.Operators.KDDMLOperatorException;
import kddml.Core.DataSources.AttributeType;

class PreprocessingUtility {

  /**
   * Crea e aggiunge alle tuple e metatuple l'attributo specificato.
   *
   * @param type : tipo attributo
   * @param newattr : nuovo attributo
   * @param nome_nuovo : nome attributo
   * @param pos : posizione
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static Instances[] addNewAttribute(Instances tuple,
                                               Instances metatuple,
                                               String nome_nuovo,
                                               AttributeType type,
                                               int pos,
                                               Enumeration enum_values)
      throws KDDMLOperatorException {

    Instances inst1 = null;
    Instances inst2 = null;
    if (type.equals(AttributeType.STRING)) {
      // attributo stringa
      inst1 = makeStringAttribute(nome_nuovo, pos, tuple);
    }
    else {
      if (type.equals(AttributeType.NOMINAL)) {
        // attributo nominale
        inst1 = makeEnumeratedAttribute(nome_nuovo, pos, enum_values, tuple);
      }
      else {
        if (type.equals(AttributeType.NUMERIC)) {
          // attributo numerico
          inst1 = makeNumericAttribute(nome_nuovo, pos, tuple);
        }
      }
    }
    inst2 = makeStringAttribute(nome_nuovo+"_metadata", pos, metatuple);
    Instances[] ris = new Instances[2];
    ris[0] = inst1;
    ris[1] = inst2;
    return ris;
  }


  /**
   * Crea e aggiunge alle tuple e metatuple l'attributo specificato.
   *
   * @param type : tipo attributo
   * @param newattr : nuovo attributo
   * @param nome_nuovo : nome attributo
   * @param pos : posizione
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static Instances addNewAttribute(Instances tuple,
                                             String nome_nuovo,
                                             AttributeType type,
                                             int pos,
                                             Enumeration enum_values)
      throws KDDMLOperatorException {

    if (type.equals(AttributeType.STRING)) {
      // attributo stringa
      return makeStringAttribute(nome_nuovo, pos, tuple);
    }
    else {
      if (type.equals(AttributeType.NOMINAL)) {
        // attributo nominale
        return makeEnumeratedAttribute(nome_nuovo, pos, enum_values, tuple);
      }
      else {
        if (type.equals(AttributeType.NUMERIC)) {
          // attributo numerico
          return makeNumericAttribute(nome_nuovo, pos, tuple);
        }
      }
    }
    // no type
    return null;
  }

  /**
   * Metodo che aggiunge a un attributo string una stringa seguita dal carattere ";"
   * per sequenzializzare i valori.
   * Se l'attributo su cui si vuole aggiungere la stringa e' missing , esso verra'
   * preventivamente inizializzato a stringa vuota
   *
   *
   * @param  posAttributo,
   * @param  valore,
   * @param instanze,
   * @param indiceIstanza
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static void addMarkInAttribute(int posAttributo,
                                           String valore,
                                           Instance instanza)
      throws KDDMLOperatorException {

    if (instanza.isMissing(posAttributo))
    {
      instanza.setValue(posAttributo, valore+";");
    }
    else {
      String old_value = instanza.stringValue(posAttributo);
      instanza.setValue(posAttributo, old_value + " "+valore + ";");
    }
  }


  /**
   * Metodo che aggiunge a un attributo string una stringa seguita dal carattere ";"
   * per sequenzializzare i valori.
   * Se l'attributo su cui si vuole aggiungere la stringa e' missing , esso verra'
   * preventivamente inizializzato a stringa vuota
   *
   * @param  posAttributo,
   * @param  valore,
   * @param instanze,
   * @param indiceIstanza
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static void addExceptionInAttribute(int posAttributo,
                                                String valore,
                                                Instance instanza)
      throws KDDMLOperatorException {

    if (instanza.isMissing(posAttributo))
    {
      instanza.setValue(posAttributo, "Exception: "+valore+";");
    }
    else {
      String old_value = instanza.stringValue(posAttributo);
      instanza.setValue(posAttributo, old_value + " "+"Exception: "+valore + ";");
    }
  }

  /**
   * Metodo che cre un nuovo attributo di tipo string e lo aggiunge
   * all'instances specificata
   *
   * @param nome attributo
   * @param pos posizione su cui aggiungerlo
   * @param istanze Instances in cui inserirlo
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static Instances makeStringAttribute(String nome,
                                         int pos,
                                         Instances istanze) throws
    KDDMLOperatorException {
    FastVector FV = new FastVector(0);
    FV = null;
    Attribute nuovo_Attributo = new Attribute(nome, FV);
    istanze.insertAttributeAt(nuovo_Attributo, pos);
    return istanze;
  }

  /**
    * Metodo che cre un nuovo attributo di tipo numeric e lo aggiunge
    * all'instances specificata
    *
    * @param nome attributo
    * @param pos posizione su cui aggiungerlo
    * @param istanze Instances in cui inserirlo
    *
    * @throws KDDMLInterpreterException
    * @throws KDDMLOperatorException
    */

  protected static Instances makeNumericAttribute(String nome,
                                                  int pos,
                                                  Instances istanze) throws
    KDDMLOperatorException {
    Attribute nuovo_Attributo = new Attribute(nome);
    istanze.insertAttributeAt(nuovo_Attributo, pos);
    return istanze;
  }

  /**
   * Metodo che cre un nuovo attributo di tipo enumerated e lo aggiunge
   * all'instances specificata
   *
   * @param nome attributo
   * @param pos posizione su cui aggiungerlo
   * @param enum valori ammissibili per il nuovo attributo enumerato
   * @param istanze Instances in cui inserirlo
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */

  protected static Instances makeEnumeratedAttribute(String nome,
      int pos,
      Enumeration enum2,
      Instances istanze) throws
     KDDMLOperatorException {
    // creo il vettore dei valori che puo' assumere l'attributo
    FastVector my_nominal_values = new FastVector();
    while (enum2.hasMoreElements()) {
      String value = (String) enum2.nextElement();
      my_nominal_values.addElement(value);
    }
    //creo l'attributo nominale
    Attribute nuovo_Attributo = new Attribute(nome, my_nominal_values);
    istanze.insertAttributeAt(nuovo_Attributo, pos);

    return istanze;
  }

  /**
   * Metodo che unifica le prime "num" instances contenute in un array, in un'unica Instances
   * che avra` un numero di istanze pari alla somma della cardinalita` di istanze
   * di ciascun Instances da unificare
   *
   *
   * @param sub_instances array contenente le istanze da unificare
   * @param cardinalita' delle istanze da unificare
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */
  protected static Instances unify(Instances[] sub_instances,
                                int num) throws KDDMLOperatorException {

    Instances nuoveTuple = sub_instances[0];

    for (int i = 1; i < num; i++) {
     // System.out.println("i : " + i);
    //  System.out.println("sub_instances [i ].numInstances()" +
                  //       sub_instances[i].numInstances());
      for (int k = 0; k < sub_instances[i].numInstances(); k++) {
        nuoveTuple.add(sub_instances[i].instance(k));

      }
    }
  //  System.out.println("nuoveTuple.numInstances()" + nuoveTuple.numInstances());
    return nuoveTuple;
  }

  protected static void setAsMissing(int posAttributo, Instances istanze) throws
       KDDMLOperatorException {
    for (int i = 0; i < istanze.numInstances(); i++) {
      istanze.instance(i).setMissing(posAttributo);
    }
  }

  /**
   * Metodo che setta i valori di un attributo string oppure enumerato
   * usando i valori passati in un array
   * di stringhe
   * @param indiceAttributo : posizione dell'attributo nell'instances
   * @param valoriString array dei valori da assegnare all'attributo
   * @param istanze Instances che contiene l'attributo
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   *
   */
  protected static void setStringValues(int indiceAttributo,
                                     String[] valoriString,
                                     Instances istanze) throws
  KDDMLOperatorException, KDDMLOperatorException {
   // System.out.println("indice attributo da settare " + indiceAttributo);
    if (!istanze.attribute(indiceAttributo).isNumeric()){

      for (int i = 0; i < istanze.numInstances(); i++) {
       // System.out.println("valoriString[i] da settare... " + valoriString[i]);
        if (valoriString[i] != "?") {
          istanze.instance(i).setValue(indiceAttributo, valoriString[i]);
          //System.out.println("ho settato... la posizione " + i);

        }
        else {
          istanze.instance(i).setMissing(indiceAttributo);
         // System.out.println("ho settato.missing  ..... la posizione.. " + i);
        }

      }
    }else
    {
      String msg = " attribute < " + istanze.attribute(indiceAttributo).name() +
           " >has wrong type.";

       throw new  KDDMLOperatorException(msg, msg,1);

    }
  }

  /**
    * Metodo che setta il valore  di un attributo string oppure enumerato , su un'istanza
    * usando il valore string  passato
    * @param posAttributo : posizione dell'attributo nell'instances
    * @param valore stringa da assegnare all'attributo
    * @param istanze Instances che contiene l'attributo
    * @param indiceIstanza istanza di riferimento
    * @throws KDDMLInterpreterException
    * @throws KDDMLOperatorException
   */

  protected static void setStringValue(int posAttributo,
                                    String valore,
                                    Instances istanze,
                                    int indiceIstanza) throws
      KDDMLOperatorException, KDDMLOperatorException {

  if (!istanze.attribute(posAttributo).isNumeric()){
    if (valore != "?"){
      istanze.instance(indiceIstanza).setValue(posAttributo, valore);
    //  System.out.println("setStringValue di PrePrUtility " +istanze.instance(indiceIstanza).stringValue(posAttributo));

    }
    else{
      istanze.instance(indiceIstanza).setMissing(posAttributo);
    }
  }
  else
    {
      String msg = " attribute < " + istanze.attribute(posAttributo).name() +
           " >has wrong type.";

       throw new  KDDMLOperatorException(msg, msg,1);

    }
  }

  /**
   * Metodo che setta il valore  di un attributo numerico , su un'istanza
   * usando il valore double  passato
   * @param posAttributo : posizione dell'attributo nell'instances
   * @param valore stringa da assegnare all'attributo
   * @param istanze Instances che contiene l'attributo
   * @param indiceIstanza istanza di riferimento
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   */

  protected static void setDoubleValue(int posAttributo,
                                    double valore,
                                    Instances instanze,
                                    int indiceIstanza) throws
      KDDMLOperatorException {

    if (valore != -1)
      instanze.instance(indiceIstanza).setValue(posAttributo, valore);
    else
      instanze.instance(indiceIstanza).setMissing(posAttributo);

  }

  /**
   * Metodo che setta i valori di un attributo numerico
   * usando i valori passati in un array
   * di double
   * @param indiceAttributo : posizione dell'attributo nell'instances
   * @param valoriNumeirci array dei valori da assegnare all'attributo
   * @param istanze Instances che contiene l'attributo
   *
   * @throws KDDMLInterpreterException
   * @throws KDDMLOperatorException
   *
   */

  protected static void setDoubleValues(int indiceAttributo,
                                     double[] valoriNumerici,
                                     Instances istanze) throws
     KDDMLOperatorException {
    for (int i = 0; i < istanze.numInstances(); i++) {
      if (valoriNumerici[i] != -1)
        istanze.instance(i).setValue(indiceAttributo, valoriNumerici[i]);
      else
        istanze.instance(i).setMissing(indiceAttributo);
    }
  }
}
