/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    ------------------------------------------------------------------------
 *    Copyright (C) 2004-2005
 *    Dipartimento di Informatica, Universita' di Pisa
 *    Pisa, Italy
 *    ------------------------------------------------------------------------
 *
 */


package kddml.Operators.Preprocessing;

import java.util.Vector;
import kddml.Core.*;
import kddml.Core.DataSources.*;
import kddml.Core.DataSources.PPInternalTable;
import kddml.Core.KDDMLObject;
import kddml.Core.KDDMLCoreException;
import kddml.Operators.*;
import weka.core.*;
import java.io.File;

/**
 * <p>Title: KDDML</p>
 * <p>Description: classe SchemaLevelTransformationche implementa
 * l'interfaccia degli operatori di PreProcessing.
 * Superclasse per tutti gli operatori di preprocessing che producono trasformazioni
 * sulle schema del dataset.
 * Gli operatori in questione sono:
 *
 * --PP_FILTER_ATTRIBUTES
 * --PP_RENAME_ATTRIBUTES;
 * --PP_SPLITTING_ATTRIBUTE;
 * --PP_FOLDING;
 * --PP_NEW_ATTRIBUTE;
 * --PP_MERGE_ATTRIBUTE;
 </p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: Universita' di Pisa - Dipartimento di Informatica</p>
 * @author Sandra Zimei (zimei@cli.di.unipi.it)
 * @version 2.0
 */

public abstract class SchemaLevelTransformation extends PreprocessingResolver {

  /**
   * Esegue l'operatore procedendo per blocchi di istanze.
   * Ciascun blocco di istanze  costituito da un numero di istanze "ottimale".
   *
   * @return tabella estesa
   * @throws KDDMLInterpreterException
   * @throws KDDMLCoreException
   */
  public KDDMLObject execute(KDDMLOperatorSettings settings,
                             java.util.Vector arguments)
      throws KDDMLOperatorException, KDDMLCoreException {

    PPInternalTableManager input = (PPInternalTableManager) arguments.get(0);

    HistoryManager hist = input.getHistory();
    readInput(settings,arguments);

    File dest = new java.io.File(kddml.Core.KDDMLObjectType.INTERNAL_PPTABLE.getRepositoryPath() +
                                 settings.getXMLDestination());

    // importa soltanto lo schema delle istanze...
    Instances[] inst = ((Instances[]) input.load(0));
    //Instances old_schema = inst[0];

    // costruisce il nuovo schema di output
    //Instances new_schema = buildNewSchema(old_schema);
    // costruisce il nuovo metaschema di output con attributi di tipo string
    //Instances new_metaschema = buildMetaSchema(new_schema);
    Instances[] new_ = execute(inst[0], inst[1]);
    Instances new_schema = new_[0];
    Instances new_metaschema = new_[1];

    // preparazione dell'output
    PPInternalTable output;
    PPWekaFactory fact = new PPWekaFactory(dest, new_schema, new_metaschema);
    fact.setOverwrite(true);
    output = (PPInternalTable) fact.newInstance();
    setAllTaxonomies(input, output);

    long opt = input.optimize();
    Instances tuple, metatuple;

    while (input.hasNext()) {

      new_schema.delete();
      new_metaschema.delete();

      // importazione delle istanze
      Instances[] inst2 = ((Instances[]) input.load(opt));
      tuple = inst2[0];
      metatuple = inst2[1];

      if (dataTransformationIsNeeded()) {
        // modifica delle singole istanze
        Instances[] new_inst = execute(tuple, metatuple);
        tuple = new_inst[0];
        metatuple = new_inst[1];
        output.addInstances(tuple, metatuple);
      }
      else {
        // le singole istanze non vengono modificate
        for (int i=0; i< tuple.numInstances(); i++) {

          Instance cloned1 = new Instance(tuple.numAttributes());
          Instance cloned2 = new Instance(metatuple.numAttributes());
          cloned1.setDataset(new_schema);
          cloned2.setDataset(new_metaschema);
          for (int j=0; j<new_schema.numAttributes(); j++) {

            if (tuple.instance(i).isMissing(j))
              cloned1.setMissing(j);
            else {

              if (tuple.attribute(j).isNumeric()) {
                // attributo originario di tipo numerico
                if (new_schema.attribute(j).isNumeric())

                  // attributo destinazione di tipo numerico
                  cloned1.setValue(j, tuple.instance(i).value(j));
                else

                  // attributo destinazione di tipo stringa
                  cloned1.setValue(j, new Double(tuple.instance(i).value(j)).toString());
              }
              else

              // attributo originario di tipo string
              if (new_schema.attribute(j).isNumeric())

                // attributo destinazione di tipo numerico
                cloned1.setValue(j, new Double(tuple.instance(i).stringValue(j)).doubleValue());
              else

                // attributo destinazione di tipo string
                cloned1.setValue(j, tuple.instance(i).stringValue(j));
            }

            if (metatuple.instance(i).isMissing(j))
              cloned2.setMissing(j);
            else
              cloned2.setValue(j, metatuple.instance(i).stringValue(j));

          }

          new_schema.add(cloned1);
          new_metaschema.add(cloned2);
        }
        output.addInstances(new_schema, new_metaschema);
      }
    }

    // aggiunge la description alla history
    String desc = createDescription();
    HistoryManager new_hist = new History((History) hist);
    new_hist.addHistory(settings.getOperatorName(), desc);
    output.setHistory(new_hist);

    return output;
  }

  /**
   * Return true if not only schema transformation is needed, but also data
   * transformation.
   * Overwrite this method if the operator modify also the physical data. In
   * this case, execute method can return null.
   *
   * @return boolean
   */
  protected boolean dataTransformationIsNeeded() {
    return true;
  }

  /**
   * Execute the operator. Tuple e metatuple sono importate a blocchi; alla prima iterazione
   * contengono soltanto lo schema dei dati.
   *
   * @param tuple Instances
   * @param metatuple Instances
   * @throws KDDMLOperatorException
   * @throws KDDMLCoreException
   * @return Instances[]
   */
  protected abstract Instances[] execute(Instances tuple, Instances metatuple)
      throws KDDMLOperatorException, KDDMLCoreException;


  /**
   * Return true if the operator performe an high number of access to the hard disk.
   *
   * @return boolean
   */
  public boolean highDiskAccessIsNeeded() {
    return true;
  }
}
